Package: test.assessr
Title: Assessing Package Test Reliability and Quality
Version: 2.1.2
Authors@R: c(
    person("Edward", "Gillian", , "edward.gillian-ext@sanofi.com", role = c("cre", "aut"), comment = c(ORCID = "0000-0003-2732-5107")),
    person("Hugo", "Bottois", , "hugo.bottois@sanofi.com", role = "aut", comment = c(ORCID = "0000-0003-4674-0875")),
    person("Paulin", "Charliquart", , "paulin.charliquart@sanofi.com", role = "aut"),
    person("Andre", "Couturier", , "andre.couturier@sanofi.com", role = "aut"),
    person("Sanofi", role = c("cph", "fnd")))
Description: A reliable and validated tool that calculates unit test coverage for R packages with standard testing frameworks and non-standard testing frameworks.
License: GPL (>= 2)
BugReports: https://github.com/pharmaverse/test.assessr/issues
Depends: R (>= 4.1.0)
Imports: callr, checkmate, covr (>= 3.6.0), dplyr, fs, jsonlite,
        pkgload, remotes, rlang, rmarkdown, RUnit, stringr (>= 1.5.0),
        testthat (>= 3.0.0), tinytest, tidyr, utils, withr
Suggests: curl, devtools, DT, here, kableExtra (>= 1.4.0), knitr,
        methods, R6, S7, roxygen2, testit (>= 0.13), tidyselect, tools,
        mockery
Encoding: UTF-8
Config/testthat/edition: 3
Config/build/clean-inst-doc: false
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-06-10 05:31:57 UTC; u1004798
Author: Edward Gillian [cre, aut] (ORCID:
    <https://orcid.org/0000-0003-2732-5107>),
  Hugo Bottois [aut] (ORCID: <https://orcid.org/0000-0003-4674-0875>),
  Paulin Charliquart [aut],
  Andre Couturier [aut],
  Sanofi [cph, fnd]
Maintainer: Edward Gillian <edward.gillian-ext@sanofi.com>
Repository: CRAN
Date/Publication: 2026-06-10 09:10:02 UTC
Built: R 4.6.0; ; 2026-06-10 11:29:29 UTC; unix

# Generated by roxygen2: do not edit by hand

export(extract_short_path)
export(generate_test_report)
export(get_package_coverage)
export(get_pkg_name)
export(install_package_local)
export(run_coverage)
export(run_covr_modes)
export(set_up_pkg)
importFrom(covr,environment_coverage)
importFrom(covr,package_coverage)
importFrom(dplyr,"%>%")
importFrom(fs,dir_exists)
importFrom(fs,path_abs)
importFrom(rlang,"%||%")
importFrom(rmarkdown,render)
importFrom(stats,setNames)
importFrom(testthat,test_dir)
importFrom(utils,capture.output)
importFrom(utils,data)
importFrom(utils,getParseData)
importFrom(utils,head)
importFrom(utils,read.csv)
importFrom(utils,read.table)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# test.assessr <a><img src="man/figures/test_assessr.png" align="right" height="100"/></a>

<!-- badges: start -->
![R-CMD-check](https://img.shields.io/badge/R%20CMD%20check-Passing-brightgreen.svg)
![Test Coverage](https://img.shields.io/badge/coverage-94%25-brightgreen.svg)
![Pharmaverse](http://pharmaverse.org/shields/test.assessr.svg)
<!-- badges: end -->


# Overview

test.assessr helps in the assessment of a package's test reliability and validity. 

It calculates unit test coverage for packages with standard testing frameworks 
and non-standard testing frameworks, including Bioconductor packages.

# Description

This package executes the following tasks:

1.  upload the source package(`tar.gz` file)

2.  Unpack the `tar.gz` file

3.  Install the package locally

4.  Run code coverage


# Package Installation

## from [Github](https://github.com/pharmaverse/test.assessr)

-   Create a `Personal Access Token` (PAT) on `github`

    -   Log into your `github` account
    -   Go to the token settings URL using the [Token Settings URL](https://github.com/settings/tokens)
        -   (do not forget to add the SSH `GitHub` authorization)

-   Create a `.Renviron` file with your GITHUBTOKEN as:

<!-- -->

```         
# .Renviron
GITHUBTOKEN=dfdxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxfdf
```

-   restart R session
-   You can install the package with:

<!-- -->

```         
auth_token = Sys.getenv("GITHUBTOKEN")
devtools::install_github("pharmaverse/test.assessr", ref = "main", auth_token = auth_token)
```

# Usage

## Assessing your own package for test coverage

To assess your package's test coverage, do the following steps:

1 - save your package as a `tar.gz` file

-   This can be done in `RStudio` -\> `Build Tab` -\> `More` -\> `Build Source Package`

2 - Run the following code sample by loading or add path parameter to your `tar.gz` package source code

``` r
# for local tar.gz R package
dp <- (path/to/your/package)

test_package_coverage <- get_package_coverage(dp)

generate_test_report(test_package_coverage)

```


## `test.assessr` results


For a package with a standard testing framework such as `testthat` or `testit` 
or non-standard testing frameworks such as `BiocGenerics` and `RUnit`, 
`test_assessr` can provide data on which tests are passing and which tests are skipped.

The standard testing framework test report tells you which tests have passed: 

![](man/figures/tp0014-stf-tests-passing.png)

The report provides information on which test blocks have passed and the start 
line of each test block.

The standard testing framework test report tells you which tests have been skipped: 

![](man/figures/tp0014-stf-tests-skipped.png)

The report provides information on which test blocks have passed, 
expectation type, and the start line of each test block.

The non-standard testing framework test report tells you which functions have no tests:

![](man/figures/MASS-nstf-no-tests.png)

The report tells you which tests have passed: 

![](man/figures/MASS-nstf-tests-passing.png)

The test report tells you which tests have been skipped: 

![](man/figures/MASS-nstf-tests-skipped.png)


# Current/Future directions

- to develop methods for checking the validity of test data
- to develop methods of checking mocks

# Acknowledgements

The project is inspired by the [`covrpage`](https://github.com/yonicd/covrpage) and [`mpn.scorecard`](https://github.com/metrumresearchgroup/mpn.scorecard) packages and draws on some of their ideas and functions.

#' get package test coverage 
#'  
#' @description simplified input to assess package for test coverage
#' @param path (optional) path of locally stored package source code
#' 
#' @return 
#' An object of class \code{"coverage"} as produced by 
#' \code{test.assessr::run_covr_modes()}.  
#' This object is a structured list containing detailed test coverage
#' information, including:
#' \itemize{
#'   \item \strong{File-level coverage:} percentage of lines covered in each file.
#'   \item \strong{Line-by-line coverage:} number of executions for each line.
#'   \item \strong{Overall coverage:} aggregated coverage across the package.
#'   \item \strong{Metadata:} source file paths, expressions, and summary stats.
#' }
#'
#' The resulting object can be printed, summarized, or passed to 
#' \code{test.assessr::generate_test_report()} to produce a human-readable test
#' coverage report.
#'
#' Returns \code{NULL} if the package cannot be installed or if the specified
#' path does not exist.
#' 
#' @examples
#' \donttest{
#' # Example STF package included in test.assessr
#' pkg_source_path <- system.file(
#'   "test-data",
#'   "test.package.0001_0.1.0.tar.gz",
#'   package = "test.assessr"
#' )
#'
#' # Run get_package_coverage
#' pkg_test_coverage <- get_package_coverage(pkg_source_path)
#' }
#' @export
get_package_coverage <-function(path = NULL) {
  
  # record covr tests (temporarily set and restore on exit)
  old_covr_record_tests <- getOption("covr.record_tests")
  options(covr.record_tests = TRUE)
  on.exit(
    {
      # If the option was previously unset (NULL), remove it again;
      # otherwise, restore the original value.
      if (is.null(old_covr_record_tests)) {
        options(covr.record_tests = NULL)
      } else {
        options(covr.record_tests = old_covr_record_tests)
      }
    },
    add = TRUE
  )
  
  # save and set user's current working directory
  oldwd <- getwd()  
  on.exit(setwd(oldwd))
  
  # get user chosen file
  pkg_source_path <- if (is.null(path)) file.choose() else path
  
  if (!file.exists(pkg_source_path)) {
    warning(paste("The specified path", pkg_source_path, "does not exist. Returning NULL."))
    return(NULL)
  }
  
  # --- Temporarily set CRAN repo (restore immediately on exit) ---
  old_repos <- getOption("repos")
  tmp_repos <- old_repos
  tmp_repos["CRAN"] <- "http://cran.us.r-project.org"
  options(repos = tmp_repos)
  # Ensure options are restored even if the function errors
  on.exit(options(repos = old_repos), add = TRUE)
  
  # Set up the package using the temporary file
  install_list <- set_up_pkg(pkg_source_path)
  
  # Extract information from the installation list
  package_installed <- install_list$package_installed
  pkg_source_path <- install_list$pkg_source_path
  
  # check if the package needs to be installed locally
  package_installed <- install_package_local(pkg_source_path)
  
  # Check if the package was installed successfully
  if (package_installed == TRUE) {
    # Assess the package
    package_coverage <- run_covr_modes(pkg_source_path)
    # Output the assessment result
  } else {
    message("Package installation failed.")
    package_coverage <- NULL
  }
  
  #clean up cov env
  cleanup_and_return_null(env = cov_env)
  
  return(package_coverage)
} 

test_that("get_package_coverage works with mocked file.choose", {
  skip_on_cran()
  
  
  # Temporarily replace the cov_env binding for this test only
  testthat::local_mocked_bindings(
    .package = "test.assessr",  # your package name
    cov_env  = new.env(parent = baseenv())
  )
  
  
  # --- Mocks ---
  # 1) Mock file chooser to a known test tarball
  mock_file_choose <- mockery::mock(
    system.file("test-data", "test.package.0001_0.1.0.tar.gz", package = "test.assessr")
  )
  
  # 2) Mock run_covr_modes to avoid calling run_coverage() and other internals
  mock_run_covr_modes <- mockery::mock({
    pkg_name <- "test.package.0001"
    pkg_ver  <- "0.1.0"
    list(
      # ---- metadata fields (10) ----
      pkg_name = pkg_name,
      pkg_ver  = pkg_ver,
      date_time = "2025-01-01 12:00:00",  # any string is fine for the test
      executor  = "tester",
      sysname   = "Windows",
      version   = "10.0",
      release   = "build",
      machine   = "x86_64",
      r_version = "4.3.2",
      test_framework_type = "standard testing framework",
      # ---- coverage payload (2) ----
      total_cov = 1,  # 100% as a fraction, per your run_coverage()
      res_cov = list(
        name = pkg_name,
        coverage = list(
          filecoverage  = c(func1 = 100),
          totalcoverage = 100
        ),
        errors = NA,
        notes  = NA
      )
    )
  })
  
  # --- Stub the symbols used inside get_package_coverage() ---
  mockery::stub(get_package_coverage, "file.choose",     mock_file_choose)
  mockery::stub(get_package_coverage, "run_covr_modes",  mock_run_covr_modes)
  
  # --- Execute ---
  tpc <- get_package_coverage()
  
  # --- Assertions (as in your original test) ---
  testthat::expect_identical(length(tpc), 12L)
  testthat::expect_true(checkmate::check_class(tpc, "list"))
  
  # Optional: verify mocks were actually invoked
  mockery::expect_called(mock_file_choose, 1)
  mockery::expect_called(mock_run_covr_modes, 1)
})



test_that("get_package_coverage warns and returns NULL if chosen file does not exist", {
  skip_on_cran()
  
  # Mock file.choose to return a fake path
  fake_path <- "some/non/existing/path.tar.gz"
  mock_file_choose <- function() fake_path
  
  # Stub file.choose and file.exists
  mockery::stub(get_package_coverage, "file.choose", mock_file_choose)
  mockery::stub(get_package_coverage, "file.exists", function(path) FALSE)
  
  # Capture the warning
  expect_warning(
    result <- get_package_coverage(),
    regexp = paste("The specified path", fake_path, "does not exist. Returning NULL.")
  )
  expect_null(result)
})

test_that("get_package_coverage outputs message and returns NULL when package installation fails", {
  skip_on_cran()
  
  # Use a valid path so file.exists returns TRUE
  fake_path <- "some/existing/path.tar.gz"
  mock_file_choose <- function() fake_path
  
  # Stub file.choose to return the fake path
  mockery::stub(get_package_coverage, "file.choose", mock_file_choose)
  # Stub file.exists to return TRUE for our fake path
  mockery::stub(get_package_coverage, "file.exists", function(path) TRUE)
  # Stub set_up_pkg to return a minimal valid install_list
  mockery::stub(get_package_coverage, "set_up_pkg", function(pkg_source_path) {
    list(
      build_vignettes = FALSE,
      package_installed = FALSE,
      pkg_source_path = fake_path,
      rcmdcheck_args = NULL
    )
  })
  # Stub install_package_local to return FALSE (simulate installation failure)
  mockery::stub(get_package_coverage, "install_package_local", function(pkg_source_path) FALSE)
  
  # Capture the message and check result is NULL (since get_package_coverage never sets it in the else block)
  expect_message(
    result <- get_package_coverage(),
    regexp = "Package installation failed."
  )
  expect_null(result)
})

###########################






test_that("on.exit restores NULL when old covr.record_tests was NULL (IF branch)", {
  skip_on_cran()
  skip_if_not_installed("mockery")
  
  options_calls <- list()
  
  # Work on a local copy so stubs bind to it, then call that copy
  fn <- test.assessr::get_package_coverage
  
  ## --- Stub heavy internals as they appear in the body ---
  mockery::stub(fn, "set_up_pkg", function(path, ...) {
    list(package_installed = TRUE, pkg_source_path = path)
  })
  mockery::stub(fn, "test.assessr::set_up_pkg", function(path, ...) {
    list(package_installed = TRUE, pkg_source_path = path)
  })
  mockery::stub(fn, "install_package_local", function(...) TRUE)
  mockery::stub(fn, "test.assessr::install_package_local", function(...) TRUE)
  mockery::stub(fn, "run_covr_modes", function(...) "dummy-cov")
  mockery::stub(fn, "test.assessr::run_covr_modes", function(...) "dummy-cov")
  mockery::stub(fn, "cleanup_and_return_null", function(...) invisible(NULL))
  mockery::stub(fn, "test.assessr::cleanup_and_return_null", function(...) invisible(NULL))
  
  ## --- Control what the function *sees* for old value ---
  stub_getOption <- function(x, default = NULL) {
    if (identical(x, "covr.record_tests")) return(NULL)  # force IF branch
    if (identical(x, "repos")) return(c(CRAN = "https://example-cran", RSPM = "https://example-rspm"))
    base::getOption(x, default)
  }
  mockery::stub(fn, "getOption", stub_getOption)
  
  ## --- Capture options() arguments reliably (preserve NULL, evaluate others) ---
  stub_options <- function(...) {
    dots_expr <- as.list(substitute(list(...)))[-1]
    eval_env <- parent.frame()
    args <- lapply(dots_expr, function(expr) if (identical(expr, quote(NULL))) NULL else eval(expr, eval_env))
    names(args) <- names(dots_expr)
    options_calls <<- append(options_calls, list(args))
    invisible(NULL)
  }
  mockery::stub(fn, "options", stub_options)
  
  ## --- Make on.exit run immediately to avoid harness interference ---
  mockery::stub(fn, "on.exit", function(expr, add = FALSE) {
    eval.parent(substitute(expr), n = 1L)
    invisible(NULL)
  })
  
  # Call the stubbed copy
  res <- fn(path = tempdir())
  expect_identical(res, "dummy-cov")
  
  # Filter only options() calls that touched covr.record_tests
  covr_calls <- Filter(function(x) "covr.record_tests" %in% names(x), options_calls)
  
  # We expect:
  #   - set to TRUE in the body
  #   - restore to NULL via (immediately executed) on.exit
  expect_gte(length(covr_calls), 2L)
  expect_identical(covr_calls[[1]][["covr.record_tests"]], TRUE)
  expect_null(covr_calls[[length(covr_calls)]][["covr.record_tests"]])
})


test_that("on.exit restores previous non-NULL covr.record_tests (ELSE branch)", {
  skip_on_cran()
  skip_if_not_installed("mockery")
  
  options_calls <- list()
  old_value <- FALSE
  
  fn <- test.assessr::get_package_coverage
  
  ## --- Stub heavy internals as they appear in the body ---
  mockery::stub(fn, "set_up_pkg", function(path, ...) {
    list(package_installed = TRUE, pkg_source_path = path)
  })
  mockery::stub(fn, "test.assessr::set_up_pkg", function(path, ...) {
    list(package_installed = TRUE, pkg_source_path = path)
  })
  mockery::stub(fn, "install_package_local", function(...) TRUE)
  mockery::stub(fn, "test.assessr::install_package_local", function(...) TRUE)
  mockery::stub(fn, "run_covr_modes", function(...) "dummy-cov")
  mockery::stub(fn, "test.assessr::run_covr_modes", function(...) "dummy-cov")
  mockery::stub(fn, "cleanup_and_return_null", function(...) invisible(NULL))
  mockery::stub(fn, "test.assessr::cleanup_and_return_null", function(...) invisible(NULL))
  
  ## --- Control what the function *sees* as old value ---
  stub_getOption <- function(x, default = NULL) {
    if (identical(x, "covr.record_tests")) return(old_value)  # force ELSE branch
    if (identical(x, "repos")) return(c(CRAN = "https://example-cran", RSPM = "https://example-rspm"))
    base::getOption(x, default)
  }
  mockery::stub(fn, "getOption", stub_getOption)
  
  ## --- Capture options() arguments reliably (preserve NULL, evaluate others) ---
  stub_options <- function(...) {
    dots_expr <- as.list(substitute(list(...)))[-1]
    eval_env <- parent.frame()
    args <- lapply(dots_expr, function(expr) if (identical(expr, quote(NULL))) NULL else eval(expr, eval_env))
    names(args) <- names(dots_expr)
    options_calls <<- append(options_calls, list(args))
    invisible(NULL)
  }
  mockery::stub(fn, "options", stub_options)
  
  ## --- Make on.exit run immediately ---
  mockery::stub(fn, "on.exit", function(expr, add = FALSE) {
    eval.parent(substitute(expr), n = 1L)
    invisible(NULL)
  })
  
  res <- fn(path = tempdir())
  expect_identical(res, "dummy-cov")
  
  covr_calls <- Filter(function(x) "covr.record_tests" %in% names(x), options_calls)
  expect_gte(length(covr_calls), 2L)
  expect_identical(covr_calls[[1]][["covr.record_tests"]], TRUE)          # set in body
  expect_identical(covr_calls[[length(covr_calls)]][["covr.record_tests"]], old_value)  # restored
})

#' @title Summarize detailed testthat results
#' @description Internal helper function to generate a long-format summary of testthat results, including status and line references.
#' @param x list of test result objects
#' @return data.frame with detailed test results including status and source line links
#' @keywords internal

get_tests_long_sum_stf <- function(x) {
  if (length(x) == 0) return(NULL)
  
  test_rows <- list()
  
  for (i in seq_along(x)) {
    test_group <- x[[i]]
    results <- test_group$results
    for (j in seq_along(results)) {
      result <- results[[j]]
      
      # Determine flags based on result class or message
      failed_flag <- as.numeric(inherits(result, "expectation_failure"))
      error_flag <- as.numeric(inherits(result, "expectation_error"))
      skipped_flag <- as.numeric(inherits(result, "expectation_skip"))
      warning_flag <- as.numeric(grepl("warning", tolower(result$message)))
      
      # Extract start and end line numbers from srcref
      line_ref <- tryCatch({
        sr <- result$srcref
        if (!is.null(sr) && length(sr) >= 7) {
          start_line <- sr[1]
          end_line <- sr[7]
          paste0("L", start_line, "_L", end_line)
        } else {
          "L?"
        }
      }, error = function(e) "L?")
      
      test_rows[[length(test_rows) + 1]] <- list(
        file = paste0(test_group$file, "#", line_ref),
        context = test_group$context,
        test = test_group$test,
        failed = failed_flag,
        error = error_flag,
        skipped = skipped_flag,
        warning = warning_flag,
        nb = 1,
        real = test_group$real
      )
    }
  }
  
  test_summary_df <- do.call(rbind, lapply(test_rows, as.data.frame))
  
  # Compute pass/fail
  test_summary_df$pass <- as.numeric(
    (test_summary_df$failed + test_summary_df$error +
       test_summary_df$skipped + test_summary_df$warning) == 0
  )
  
  # Assign status
  test_summary_df$status <- "PASS"
  test_summary_df$status[test_summary_df$failed != 0] <- "FAILED"
  test_summary_df$status[test_summary_df$error != 0] <- "ERROR"
  test_summary_df$status[test_summary_df$skipped != 0] <- "SKIPPED"
  test_summary_df$status[test_summary_df$warning != 0] <- "WARNING"
  
  # Select and rename relevant columns
  result_df <- test_summary_df[, c("file", "context", "test", "status", "nb", "real")]
  names(result_df)[c(5, 6)] <- c("n", "time")
  
  return(result_df)
}

test_that("get_tests_long_sum_stf handles PASS and formats output correctly", {
  # Mocked expectation object
  mock_result <- list(
    message = "All good",
    srcref = c(10, 3, 10, 49, 3, 49, 10, 10),
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test A"
  )
  class(mock_result) <- c("expectation_success", "expectation", "condition")
  
  mock_input <- list(
    list(
      file = "test-file.R",
      context = "Context A",
      test = "Test A",
      real = 0.123,
      results = list(mock_result)
    )
  )
  
  result <- get_tests_long_sum_stf(mock_input)
  
  expect_type(result, "list")
  expect_equal(nrow(result), 1)
  expect_equal(result$status, "PASS")
  expect_equal(result$n, 1)
  expect_equal(result$time, 0.123)
  expect_true(grepl("test-file.R#L10_L10", result$file))
})




test_that("get_tests_long_sum_stf handles all statuses with varied srcref lines", {
  # PASS
  mock_result_pass <- list(
    message = "All good",
    srcref = c(5, 3, 5, 49, 3, 49, 6, 6),  # srcref[1] = 5, srcref[7] = 6
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test PASS"
  )
  class(mock_result_pass) <- c("expectation_success", "expectation", "condition")
  
  # FAILED
  mock_result_failed <- list(
    message = "Expected 2 but got 3",
    srcref = c(15, 3, 15, 49, 3, 49, 16, 17),  # srcref[1] = 15, srcref[7] = 16
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test FAILED"
  )
  class(mock_result_failed) <- c("expectation_failure", "expectation", "condition")
  
  # ERROR
  mock_result_error <- list(
    message = "Error: object not found",
    srcref = c(25, 3, 25, 49, 3, 49, 26, 28),  # srcref[1] = 25, srcref[7] = 26
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test ERROR"
  )
  class(mock_result_error) <- c("expectation_error", "expectation", "condition")
  
  # SKIPPED
  mock_result_skipped <- list(
    message = "Test skipped due to condition",
    srcref = c(35, 3, 35, 49, 3, 49, 36, 39),  # srcref[1] = 35, srcref[7] = 36
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test SKIPPED"
  )
  class(mock_result_skipped) <- c("expectation_skip", "expectation", "condition")
  
  # WARNING
  mock_result_warning <- list(
    message = "Warning: something might be wrong",
    srcref = c(45, 3, 45, 49, 3, 49, 46, 50),  # srcref[1] = 45, srcref[7] = 46
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test WARNING"
  )
  class(mock_result_warning) <- c("expectation_success", "expectation", "condition")
  
  # Combine all into mock input
  mock_input <- list(
    list(file = "test-file.R", context = "Context A", test = "Test A", real = 0.123, results = list(mock_result_pass)),
    list(file = "test-file.R", context = "Context B", test = "Test B", real = 0.234, results = list(mock_result_failed)),
    list(file = "test-file.R", context = "Context C", test = "Test C", real = 0.345, results = list(mock_result_error)),
    list(file = "test-file.R", context = "Context D", test = "Test D", real = 0.456, results = list(mock_result_skipped)),
    list(file = "test-file.R", context = "Context E", test = "Test E", real = 0.567, results = list(mock_result_warning))
  )
  
  result <- get_tests_long_sum_stf(mock_input)
  
  expect_equal(nrow(result), 5)
  expect_equal(result$status, c("PASS", "FAILED", "ERROR", "SKIPPED", "WARNING"))
  expect_equal(result$n, rep(1, 5))
  expect_equal(result$time, c(0.123, 0.234, 0.345, 0.456, 0.567))
  expect_true(grepl("test-file.R#L5_L6", result$file[1]))
  expect_true(grepl("test-file.R#L15_L16", result$file[2]))
  expect_true(grepl("test-file.R#L25_L26", result$file[3]))
  expect_true(grepl("test-file.R#L35_L36", result$file[4]))
  expect_true(grepl("test-file.R#L45_L46", result$file[5]))
})


test_that("get_tests_long_sum_stf returns NULL for empty input", {
  empty_input <- list()
  result <- get_tests_long_sum_stf(empty_input)
  expect_null(result)
})


test_that("else block is handled with correct srcref", {
  mock_result_else <- list(
    message = "Handled by else block",
    srcref = c(1, 3, 1, 49, 3, 49, 32, 32),  # srcref[1] = 1, srcref[7] = 32
    trace = NULL,
    start_frame = 36,
    end_frame = 38,
    test = "test ELSE"
  )
  class(mock_result_else) <- c("expectation_success", "expectation", "condition")
  
  mock_input <- list(
    list(file = "test-file.R", context = "Context F", test = "Test F", real = 0.678, results = list(mock_result_else))
  )
  
  result <- get_tests_long_sum_stf(mock_input)
  
  expect_equal(nrow(result), 1)
  expect_equal(result$status, "PASS")
  expect_equal(result$n, 1)
  expect_equal(result$time, 0.678)
  expect_true(grepl("test-file.R#L1_L32", result$file[1]))
})

#' Create Information on Local Package Installation
#'
#' This function unpacks a package tarball using \code{unpack_tarball()} and
#' returns information about whether the unpacked directory exists. It is a
#' lightweight preparatory step used before attempting a local package
#' installation.
#'
#' @param dp Character string. Path to a package tarball or package directory.
#'
#' @return
#' A named list with the following elements:
#' \itemize{
#'   \item \code{package_installed} — Logical. \code{TRUE} if the unpacked
#'         package directory exists, otherwise \code{FALSE}.
#'   \item \code{pkg_source_path} — Character string giving the unpacked
#'         package source directory, or \code{""} if unpacking failed.
#' }
#'
#' The returned object is always a base R \code{list}. It contains no side
#' effects besides calling \code{unpack_tarball()} and checking filesystem paths.
#'
#' @section Output Meaning:
#' A value of \code{package_installed = TRUE} indicates that the unpacked
#' directory exists on disk and can be used for local installation.  
#'
#' \code{package_installed = FALSE} indicates either:
#' \itemize{
#'   \item \code{unpack_tarball()} returned an empty result, or
#'   \item the unpacked directory does not exist on disk.
#' }
#'
#' @examples
#' \dontrun{
#' set_up_pkg(path/to/package, "mypackage")
#' }
#' @export
set_up_pkg <- function(dp) {
  
  suppressWarnings(pkg_source_path <- unpack_tarball(dp))
  
  # check if pkg_source_path exists
  if (length(pkg_source_path) == 0) {
    package_installed <- FALSE
    results <- ""
    pkg_source_path <- ""
    out_dir <- ""
  } else {
    if (fs::file_exists(pkg_source_path)) {
      package_installed <- TRUE
    }
  }
  
  install_list <- list(
    package_installed = package_installed,
    pkg_source_path = pkg_source_path
  )
  return(install_list)
}

test_that("get package description works correctly", {
  
  r = getOption("repos")
  r["CRAN"] = "http://cran.us.r-project.org"
  options(repos = r)
  
  dp <- system.file("test-data", "here-1.0.1.tar.gz", 
                    package = "test.assessr")
  
  # set up package
  install_list <- set_up_pkg(dp)
  
  package_installed <- install_list$package_installed
  pkg_source_path <- install_list$pkg_source_path
 
  if (package_installed == TRUE ) {	
    pkg_desc <- get_pkg_desc(pkg_source_path)
    
    expect_identical(length(pkg_desc), 17L)
    
    expect_true(checkmate::check_list(pkg_desc, 
                                      any.missing = FALSE)
    )
    
    expect_true(checkmate::check_list(pkg_desc, 
                                      types = "character")
    )
    
  }
  
})

test_that("get package name works correctly", {
  
  r = getOption("repos")
  r["CRAN"] = "http://cran.us.r-project.org"
  options(repos = r)
  
  dp <- system.file("test-data", "MASS_7.3-65.tar.gz", 
                    package = "test.assessr")
  
  # set up package
  install_list <- set_up_pkg(dp)
  
  package_installed <- install_list$package_installed
  pkg_source_path <- install_list$pkg_source_path
 
  if (package_installed == TRUE ) {	
    pkg <- get_pkg_name(pkg_source_path)
    
    expect_identical(length(pkg), 1L)
    
    expect_true(checkmate::check_character(pkg, 
                                           any.missing = FALSE)
    )
    
    expect_true(checkmate::check_character(pkg, 
                                           pattern = "MASS")
    )
    
  }
  
})

test_that("get_test_metadata returns correct metadata with executor", {
  # Mocked values
  fake_time <- "2025-09-30 12:00:00"
  fake_sysinfo <- list(
    sysname = "Linux",
    version = "#1 SMP Debian",
    release = "5.10.0-8-amd64",
    machine = "x86_64"
  )
  
  # Stub system functions
  mockery::stub(get_test_metadata, "Sys.time", function() fake_time)
  mockery::stub(get_test_metadata, "Sys.info", function() fake_sysinfo)
  
  result <- get_test_metadata(executor = "edward")
  
  expect_equal(result$datetime, fake_time)
  expect_equal(result$executor, "edward")
  expect_equal(result$info$sys, fake_sysinfo)
})

test_that("get_test_metadata returns correct metadata without executor", {
  # Mocked values
  fake_time <- "2025-09-30 12:00:00"
  fake_user <- "mock_user"
  fake_sysinfo <- list(
    sysname = "Darwin",
    version = "22.6.0",
    release = "macOS",
    machine = "arm64"
  )
  
  # Stub system functions
  mockery::stub(get_test_metadata, "Sys.time", function() fake_time)
  mockery::stub(get_test_metadata, "Sys.getenv", function(var) {
    if (var == "USER") return(fake_user)
  })
  mockery::stub(get_test_metadata, "Sys.info", function() fake_sysinfo)
  
  result <- get_test_metadata()
  
  expect_equal(result$datetime, fake_time)
  expect_equal(result$executor, fake_user)
  expect_equal(result$info$sys, fake_sysinfo)
})


test_that("extract_short_path handles forward slashes", {
  p <- "C:/Users/yyy/AppData/Local/Temp/RtmpXXXX/MASS/R/add.R"
  expect_equal(extract_short_path(p), "R/add.R")
})

test_that("extract_short_path handles backslashes", {
  p <- "C:\\Users\\yyy\\AppData\\Local\\Temp\\RtmpXXXX\\MASS\\R\\add.R"
  expect_equal(extract_short_path(p), "R/add.R")
})

test_that("extract_short_path handles mixed separators", {
  p <- "C:/Users\\yyy/AppData\\Local/Temp/RtmpXXXX/MASS/R/add.R"
  expect_equal(extract_short_path(p), "R/add.R")
})

test_that("extract_short_path returns last two components for general paths", {
  p <- "/opt/projects/pkgname/src/module/file.ext"
  expect_equal(extract_short_path(p), "module/file.ext")
})

test_that("extract_short_path handles single-component paths", {
  p <- "file.ext"
  expect_equal(extract_short_path(p), "file.ext")
})

test_that("extract_short_path handles empty string", {
  p <- ""
  # strsplit("", "[/\\\\]") returns character(0), function returns ""
  expect_equal(extract_short_path(p), "")
})

test_that("extract_short_path handles trailing separator", {
  # Trailing separator creates an empty last component
  p1 <- "dir/subdir/"
  expect_equal(extract_short_path(p1), "subdir/")  # last component is ""
  
  p2 <- "dir\\subdir\\"
  expect_equal(extract_short_path(p2), "subdir/")  # unified separator in output
})

test_that("extract_short_path works over a vector via vapply", {
  paths <- c(
    "C:/A/B/C/D.R",
    "C:\\A\\B\\C\\E.R",
    "/A/B/C/F.R",
    "file.ext"
  )
  out <- vapply(paths, extract_short_path, FUN.VALUE = character(1))
  expect_equal(unname(out), c("C/D.R", "C/E.R", "C/F.R", "file.ext"))
})

test_that("extract_short_path behavior for NA (optional policy)", {
  # If you want NA-in -> NA-out, you can wrap it:
  safe_extract <- function(x) if (is.na(x)) NA_character_ else extract_short_path(x)
  expect_true(is.na(safe_extract(NA_character_)))
})

test_that("set up package for tar file with default check type", {
  
  r = getOption("repos")
  r["CRAN"] = "http://cran.us.r-project.org"
  options(repos = r)
  
  # Copy test package to a temp file
  dp_orig <- system.file("test-data", 
                         "test.package.0001_0.1.0.tar.gz", 
                         package = "test.assessr")
  dp <- tempfile(fileext = ".tar.gz")
  file.copy(dp_orig, dp)
  
  # Defer cleanup of copied tarball
  withr::defer(unlink(dp), envir = parent.frame())
  
  # set up package
  install_list <- set_up_pkg(dp)
  
  expect_identical(length(install_list), 2L)
  
  expect_true(checkmate::check_list(install_list, 
                                    any.missing = FALSE)
  )
  
  expect_true(checkmate::check_list(install_list, 
                                    types = c("logical",
                                              "character",
                                              "list")
                                    )
  )
})

test_that("set_up_pkg handles missing pkg_source_path correctly (length 0)", {
  # Use a dummy dp (doesn't matter, as we stub the relevant function)
  dp <- "dummy/path/to/package.tar.gz"
  
  # stub the internal function that returns pkg_source_path to always return length 0
  mockery::stub(set_up_pkg, "unpack_tarball", character(0))
 
  # Run function
  install_list <- set_up_pkg(dp)
  
  # Expectation: install_list should match the state in the code block:
  # c(package_installed, results, pkg_source_path, out_dir, build_vignettes)
  expect_identical(length(install_list), 2L)
  
  expect_false(install_list[[1]])
  expect_identical(install_list[[2]], "")

})

#' Package-Private Coverage Environment
#'
#' `cov_env` is a package-internal environment used by test/coverage helpers
#' (e.g., `cleanup_and_return_null()`) to store temporary state without ever
#' touching the user's workspace (`.GlobalEnv`). It is created at package load
#' time so that a binding for `cov_env` exists *inside the locked package
#' namespace*, which allows tests to temporarily replace its value (mock) while
#' still complying with CRAN’s “no .GlobalEnv modifications” rule.
#'
#' @details
#' **When is it built?**  
#' During package setup, all `R/` files are parsed and evaluated as part of
#' namespace loading. At that point this line is executed:
#'
#' ```
#' cov_env <- new.env(parent = baseenv())
#' ```
#'
#' which creates the binding `cov_env` in the package namespace **before** the
#' namespace is locked by R. Once a namespace is locked, you cannot add or
#' remove bindings—you may only change the *values* of existing bindings. By
#' pre-creating `cov_env`, tests can safely and temporarily replace it (e.g.,
#' with `testthat::local_mocked_bindings()`), because they modify an *existing*
#' binding rather than creating a new one.  (See: R’s environment/namespace
#' locking rules and testthat’s guidance that mocked bindings must already
#' exist.)  [1](https://colinfay.me/writing-r-extensions/writing-r-documentation-files.html)[2](https://contributor.r-project.org/cran-cookbook/code_issues.html)
#'
#' **Why `parent = baseenv()`?**  
#' Using `baseenv()` isolates lookups so they do not fall through to the user’s
#' workspace. This keeps internal state separate and avoids accidental reads or
#' writes to `.GlobalEnv`.  [3](https://github.com/coolbutuseless/CRAN-checks)
#'
#' **Testing tip:**  
#' In unit tests, you can temporarily replace this binding without editing the
#' global environment:
#'
#' ```
#' testthat::local_mocked_bindings(
#'   .package = "yourpkg",
#'   cov_env  = new.env(parent = baseenv())
#' )
#' ```
#'
#' This works because the `cov_env` *binding already exists* in the namespace;
#' `local_mocked_bindings()` cannot create new bindings in a locked namespace.  [2](https://contributor.r-project.org/cran-cookbook/code_issues.html)
#'
#' @keywords internal
cov_env <- new.env(parent = baseenv())   # package-private state

#' Safely generate skip test mapping from problem tests and test map
#'
#' This function filters and joins test metadata to identify which tests
#' should be skipped based on failed or problematic test results.
#'
#' @param problem_tests A data frame of problematic tests, including columns `file` and `context`.
#' @param test_map A data frame representing the test structure, including columns `file` and `context`.
#'
#' @return A data frame containing the merged skip test mapping, or `NULL` if validation or processing fails.
#'
#' @keywords internal
get_tests_skip_stf <- function(problem_tests, test_map) {
  
  # Validate 'problem_tests'
  if (!checkmate::test_data_frame(problem_tests, any.missing = TRUE, min.rows = 1)) {
    warning("'problem_tests' is not a valid data frame or has no rows.")
    return(NULL)
  }
  if (!checkmate::test_names(names(problem_tests), must.include = c("file", "context"))) {
    warning("'problem_tests' must include 'file' and 'context' columns.")
    return(NULL)
  }
  
  # Validate 'test_map'
  if (!checkmate::test_data_frame(test_map, any.missing = FALSE, min.rows = 1)) {
    warning("'test_map' is not a valid data frame or has no rows.")
    return(NULL)
  }
  if (!checkmate::test_names(names(test_map), must.include = c("file", "context"))) {
    warning("'test_map' must include 'file' and 'context' columns.")
    return(NULL)
  }
  
  # Filter test_map to only include rows matching both context and file in problem_tests
  test_map_filtered <- tryCatch({
    test_map[test_map$context %in% problem_tests$context &
               test_map$file %in% problem_tests$file, ]
  }, error = function(e) {
    warning("Error filtering test_map: ", e$message)
    return(NULL)
  })
  
  if (is.null(test_map_filtered) || nrow(test_map_filtered) == 0) {
    warning("Filtered test_map is empty or invalid.")
    return(NULL)
  }
  
  # Perform join
  test_skip <- safe_left_join(problem_tests, 
                              test_map_filtered, 
                              by = c("context", "file"))
  
  
  if (is.null(test_skip) || nrow(test_skip) == 0) {
    warning("Join result is empty or invalid.")
    return(NULL)
  }
  
  return(test_skip)
}


#' Safely perform a left join with error handling
#'
#' This helper function attempts to join two data frames using
#' `dplyr::left_join()`. If an error occurs during the join (for example,
#' due to missing columns or incompatible types), the function issues a
#' warning and returns `NULL` instead of stopping execution.
#'
#' @param x A data frame or tibble. The left-hand side of the join.
#' @param y A data frame or tibble. The right-hand side of the join.
#' @param by A character vector specifying the join keys. Passed to
#'   `dplyr::left_join()`.
#' @param relationship A string specifying the expected relationship between keys:
#'   `"one-to-one"`, `"one-to-many"`, `"many-to-one"`, or `"many-to-many"`.
#'   Defaults to `NULL` (no check).
#'
#' @return A data frame resulting from the left join if successful,
#'   otherwise `NULL` when an error occurs.
#'
#' @keywords internal
safe_left_join <- function(x, y, by, relationship = "many-to-many", ...) {
  tryCatch({
    dplyr::left_join(x, y, by = by, ...)
  }, error = function(e) {
    # Signal warning and return NULL
    warning("Error joining data frames: ", e$message, call. = FALSE)
    return(NULL)
  })
}

test_that("get_tests_skip_stf performs safe join with valid input", {
  problem_tests <- data.frame(
    file = c("test1.R", "test2.R"),
    context = c("ctx1", "ctx2"),
    status = c("FAILED", "FAILED"),
    n = c(1, 1),
    time = c(0.1, 0.2),
    stringsAsFactors = FALSE
  )
  
  test_map <- data.frame(
    file = c("test1.R", "test1.R", "test2.R"),
    context = c("ctx1", "ctx1", "ctx2"),
    expectation = c("expect_equal", "expect_true", "expect_error"),
    line1 = c(10, 20, 30),
    line2 = c(10, 20, 30),
    stringsAsFactors = FALSE
  )
  
  result <- get_tests_skip_stf(problem_tests, test_map)
  
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 3)
  expect_true(all(c("file", "context", "expectation", "line1", "line2") %in% names(result)))
})

test_that("get_tests_skip_stf returns NULL for invalid problem_tests structure", {
  problem_tests <- data.frame(
    wrong_col = c("test1.R"),
    stringsAsFactors = FALSE
  )
  test_map <- data.frame(
    file = "test1.R",
    context = "ctx1",
    expectation = "expect_equal",
    line1 = 10,
    line2 = 10,
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})

test_that("get_tests_skip_stf returns NULL for invalid test_map structure", {
  problem_tests <- data.frame(
    file = "test1.R",
    context = "ctx1",
    status = "FAILED",
    n = 1,
    time = 0.1,
    stringsAsFactors = FALSE
  )
  test_map <- data.frame(
    wrong_col = "test1.R",
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})

test_that("get_tests_skip_stf returns NULL when join result is empty", {
  problem_tests <- data.frame(
    file = "nonexistent.R",
    context = "missing_ctx",
    status = "FAILED",
    n = 1,
    time = 0.1,
    stringsAsFactors = FALSE
  )
  
  test_map <- data.frame(
    file = "test1.R",
    context = "ctx1",
    expectation = "expect_equal",
    line1 = 10,
    line2 = 10,
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})


test_that("get_tests_skip_stf returns NULL when problem_tests is not a data frame", {
  problem_tests <- list(file = "test1.R", context = "ctx1")  # Not a data frame
  test_map <- data.frame(
    file = "test1.R",
    context = "ctx1",
    expectation = "expect_equal",
    line1 = 10,
    line2 = 10,
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})


test_that("get_tests_skip_stf returns NULL when problem_tests has no rows", {
  problem_tests <- data.frame(
    file = character(0),
    context = character(0),
    status = character(0),
    n = integer(0),
    time = numeric(0),
    stringsAsFactors = FALSE
  )
  
  test_map <- data.frame(
    file = "test1.R",
    context = "ctx1",
    expectation = "expect_equal",
    line1 = 10,
    line2 = 10,
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})


test_that("get_tests_skip_stf returns NULL when test_map is not a data frame", {
  problem_tests <- data.frame(
    file = "test1.R",
    context = "ctx1",
    status = "FAILED",
    n = 1,
    time = 0.1,
    stringsAsFactors = FALSE
  )
  
  test_map <- list(file = "test1.R", context = "ctx1")  # Not a data frame
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})

test_that("get_tests_skip_stf returns NULL when test_map has missing values", {
  problem_tests <- data.frame(
    file = "test1.R",
    context = "ctx1",
    status = "FAILED",
    n = 1,
    time = 0.1,
    stringsAsFactors = FALSE
  )
  
  test_map <- data.frame(
    file = c("test1.R", NA),
    context = c("ctx1", "ctx1"),
    expectation = c("expect_equal", "expect_true"),
    line1 = c(10, 20),
    line2 = c(10, 20),
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})

test_that("get_tests_skip_stf returns NULL when test_map has zero rows", {
  problem_tests <- data.frame(
    file = "test1.R",
    context = "ctx1",
    status = "FAILED",
    n = 1,
    time = 0.1,
    stringsAsFactors = FALSE
  )
  
  test_map <- data.frame(
    file = character(),
    context = character(),
    expectation = character(),
    line1 = integer(),
    line2 = integer(),
    stringsAsFactors = FALSE
  )
  
  result <- suppressWarnings(get_tests_skip_stf(problem_tests, test_map))
  expect_null(result)
})


test_that("get_tests_skip_stf handles error during test_map filtering", {
  # Prepare valid problem_tests and test_map data frames
  problem_tests <- data.frame(file = "test.R", context = "ctx", stringsAsFactors = FALSE)
  test_map <- data.frame(file = "test.R", context = "ctx", stringsAsFactors = FALSE)
  
  # Stub the subsetting operation to throw an error
  mockery::stub(get_tests_skip_stf, "[", function(...) stop("Simulated subsetting error"))
  
  # Capture warnings
  warnings <- capture_warnings({
    result <- get_tests_skip_stf(problem_tests, test_map)
  })
  
  # Assertions
  expect_null(result)
  expect_true(any(grepl("Error filtering test_map", warnings)))
})

test_that("successful join returns expected data frame", {
  df1 <- data.frame(id = 1:2, val = c("a", "b"))
  df2 <- data.frame(id = 1:2, extra = c("x", "y"))
  
  result <- safe_left_join(df1, df2, by = "id")
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_true("extra" %in% names(result))
})


test_that("safe_left_join returns NULL and warns on error", {
  df1 <- data.frame(id = 1)
  df2 <- data.frame(id_wrong = 1)
  
  expect_warning(
    result <- safe_left_join(df1, df2, by = "id"),
    regexp = "Error joining data frames"
  )
  
  expect_null(result)
})


test_that("empty data frames still join correctly", {
  df1 <- data.frame(id = integer())
  df2 <- data.frame(id = integer())
  
  result <- safe_left_join(df1, df2, by = "id")
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 0)
})

test_that("join with missing columns triggers warning and returns NULL", {
  df1 <- data.frame(id = 1:2)
  df2 <- data.frame(other = 1:2)
  
  expect_warning(
    result <- safe_left_join(df1, df2, by = "id"),
    regexp = "Error joining data frames"
  )
  expect_null(result)
})


test_that("safe_left_join handles many-to-many relationships correctly", {
  # Create data frames with duplicate keys
  df1 <- data.frame(id = c(1, 1, 2), val1 = c("A", "B", "C"))
  df2 <- data.frame(id = c(1, 1, 2), val2 = c("X", "Y", "Z"))
  
  # Perform join
  result <- suppressWarnings(safe_left_join(df1, df2, by = "id"))
  
  # Check structure
  expect_s3_class(result, "data.frame")
  expect_true(all(c("id", "val1", "val2") %in% names(result)))
  
  # Expected number of rows: Cartesian product for duplicates
  # df1 has 2 rows with id=1, df2 has 2 rows with id=1 → 4 rows for id=1
  # id=2 matches 1 row each → 1 row
  expect_equal(nrow(result), 5)
  
  # Check that all combinations exist for id=1
  combos_id1 <- result[result$id == 1, c("val1", "val2")]
  expect_true(all(c("A", "B") %in% combos_id1$val1))
  expect_true(all(c("X", "Y") %in% combos_id1$val2))
})

test_that("safe_left_join warns and returns NULL when many-to-many join fails", {
  # Force an error by stubbing left_join
  mockery::stub(safe_left_join, "dplyr::left_join", function(...) {
    stop("Simulated join error")
  })
  
  expect_warning(
    result <- safe_left_join(data.frame(id = 1), data.frame(id = 1), by = "id"),
    regexp = "Error joining data frames"
  )
  
  expect_null(result)
})


test_that("returns NULL and warns when safe_left_join returns NULL", {
  # Mock safe_left_join to return NULL
  mockery::stub(get_tests_skip_stf, "safe_left_join", NULL)
  
  problem_tests <- data.frame(file = "file1", context = "ctx1")
  test_map <- data.frame(file = "file1", context = "ctx1")
  
  expect_warning(
    result <- get_tests_skip_stf(problem_tests, test_map),
    "Join result is empty or invalid."
  )
  expect_null(result)
})


test_that("returns NULL and warns when safe_left_join returns empty data frame", {
  # Mock safe_left_join to return empty data frame
  mockery::stub(get_tests_skip_stf, "safe_left_join", data.frame())
  
  problem_tests <- data.frame(file = "file1", context = "ctx1")
  test_map <- data.frame(file = "file1", context = "ctx1")
  
  expect_warning(
    result <- get_tests_skip_stf(problem_tests, test_map),
    "Join result is empty or invalid."
  )
  expect_null(result)
})

#' Run Coverage Analysis on data.table Package Source and Tests
#'
#' This function maps source files to test files, checks for skipped/problematic tests,
#' runs the remaining test files manually, calculates coverage, and optionally generates a report.
#'
#' @param pkg_source_path Character string. Path to the root of the package source directory.
#' @param test_pkg_data List. Output from `check_pkg_tests_and_snaps()` indicating test framework presence.
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run.
#'
#'
#' @return returns a list with coverage object and source files with no tests.
#'
#' @keywords internal
run_covr_skip_dt_nstf <- function(pkg_source_path, 
                                  test_pkg_data,
                                  cov_env) {
  
  # --- Input checking (CRAN-safe and explicit) ---
  checkmate::assert_string(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_directory_exists(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_true(!missing(cov_env), .var.name = "cov_env")
  checkmate::assert_environment(cov_env, .var.name = "cov_env")
  
  pkg_name <- get_pkg_name(pkg_source_path)
  
  # 1) Resolve data.table test files (prefers tests/main.R)
  #    tests/main.R -> test.data.table() -> inst/tests/tests.Rraw (big suite)
  #    https://github.com/Rdatatable/data.table/blob/master/tests/main.R
  #    https://rdatatable.gitlab.io/data.table/reference/test.data.table.html
  test_files_clean <- compute_test_files_clean_dt(
    pkg_name        = pkg_name,
    pkg_source_path = pkg_source_path,
    clean_mapping   = NULL,            # mapping not used for dt
    test_pkg_data   = test_pkg_data,
    skip_tests      = NULL
  )
  test_files_clean <- normalizePath(test_files_clean, winslash = "/", mustWork = FALSE)
  
  if (!length(test_files_clean) || !all(file.exists(test_files_clean))) {
    return(cleanup_and_return_null(
      sprintf("No resolvable data.table tests for %s", pkg_name)
    ))
  }
  
  # 2) Load package + datasets into coverage env
  message(sprintf("loading package and datasets for %s", pkg_name))
  tryCatch({
    pkgload::load_all(pkg_source_path)
    
    pkg_all <- ls(envir = asNamespace(pkg_name), all.names = TRUE)
    for (obj in pkg_all) {
      tryCatch({
        assign(obj, get(obj, envir = asNamespace(pkg_name)), envir = cov_env)
      }, error = function(e) {
        message(sprintf("Failed to assign object %s : %s", obj, e$message))
      })
    }
    datasets <- load_package_datasets(pkg_name, env = cov_env)
    if (length(datasets)) message("Datasets loaded: ", paste(datasets, collapse = ", "))
  }, error = function(e) {
    return(cleanup_and_return_null(
      sprintf("Error loading package or datasets for %s : %s", pkg_name, e$message)
    ))
  })
  
  # 3) Ensure runtime layout for test.data.table(): MIRROR inst/tests/* -> tests/ if needed
  #    (CRAN-friendly; no .GlobalEnv mods)
  prep <- prepare_tests_for_package_dt(
    pkg_name        = pkg_name,
    pkg_source_path = pkg_source_path,
    test_files      = test_files_clean,
    mode            = "mirror"
  )
  on.exit(prep$teardown(), add = TRUE)
  
  # 4) Run coverage on the resolved tests
  message(sprintf("running coverage manually for %s", pkg_name))
  coverage <- tryCatch({
    covr::environment_coverage(
      env        = cov_env,
      test_files = test_files_clean
    )
  }, error = function(e) {
    return(cleanup_and_return_null(
      sprintf("Error running coverage for %s : %s", pkg_name, e$message),
      env = cov_env
    ))
  })
  
  if (is.null(coverage)) {
    message("No test coverage for this configuration")
    return(list(
      total_cov = 0,
      res_cov = list(
        name = pkg_name,
        coverage = list(
          filecoverage   = matrix(0, nrow = 1, dimnames = list("No functions tested")),
          totalcoverage  = 0
        ),
        errors = "No coverage generated",
        notes  = NA
      ),
      functions_no_tests = data.frame(function_name = character(), file = character()),
      tests_skipped      = character(),
      tests_passing      = character()
    ))
  }
  
  # 5) Convert coverage -> list, gather file coverage summary
  res_cov <- tryCatch({
    coverage_list <- covr::coverage_to_list(coverage)
    list(name = pkg_name, coverage = coverage_list, errors = NA, notes = NA)
  }, error = function(cond) {
    coverage_list <- list(filecoverage = NA, totalcoverage = NA_integer_)
    list(name = pkg_name, coverage = coverage_list, errors = cond, notes = NA)
  })
  
  # 6) Summaries specifically for data.table
  # 6a) tested_files = how many test files executed (tests/main.R => 1)
  tested_files <- length(unique(test_files_clean))
  
  # 6b) skipped_tests = known wrapper scripts present but not selected
  dt_wrapper_candidates <- dt_find_wrapper_tests(pkg_source_path)
  skipped_tests <- setdiff(dt_wrapper_candidates, test_files_clean)
  
  # 6c) untested_functions = functions whose srcref span has zero executed lines
  dt_no_tests_df <- dt_compute_untested_functions(pkg_name, coverage)
  
  # 7) Aggregate totals with your existing coverage combiner
  percent_cov <- as.numeric(res_cov$coverage$totalcoverage)
  calc_covr_list <- compute_total_coverage(
    percent_cov, 
    dt_no_tests_df,         # functions_no_tests_df
    test_files_clean,       # tests_passing
    skipped_tests           # tests_skipped
  )
  
  total_cov           <- calc_covr_list$total_cov
  percent_cov_round   <- round(total_cov, 2)
  n_testfiles         <- calc_covr_list$n_testfiles
  n_no_function_tests <- calc_covr_list$n_no_function_tests
  n_skipped_files     <- calc_covr_list$n_skipped_files
  
  message(sprintf("Total Coverage for %s is: %.2f%%", pkg_name, percent_cov_round))
  message(sprintf(
    "Approx total coverage: (%.2f%%), counts: tested_files=%d, untested_functions=%d, skipped_tests=%d",
    calc_covr_list$percent_cov, n_testfiles, n_no_function_tests, n_skipped_files
  ))
  
  cleanup_and_return_null(env = cov_env)
  
  list(
    total_cov = total_cov,
    res_cov = list(
      name     = res_cov$name,
      coverage = res_cov$coverage,  # contains $filecoverage and $totalcoverage
      errors   = res_cov$errors,
      notes    = res_cov$notes
    ),
    functions_no_tests = dt_no_tests_df,
    tests_skipped      = skipped_tests,
    tests_passing      = test_files_clean
  )
}




#' Locate data.table wrapper test scripts (internal)
#'
#' Returns absolute, normalized paths to known wrapper test scripts under
#' \code{tests/} for a data.table source checkout. This is used to determine
#' which wrapper files exist locally so that callers can distinguish between
#' the canonical entrypoint \code{tests/main.R} and other wrapper scripts
#' (e.g., \code{froll.R}, \code{programming.R}, etc.).
#'
#' @param pkg_source_path character(1). Path to the package source root that
#'   contains the \code{tests/} directory.
#'
#' @return character(). A vector of existing files among the known candidates,
#'   with absolute paths normalized via \code{normalizePath(..., winslash = "/")}.
#'   The vector is empty if none of the candidates exist.
#'
#' @section Notes:
#' The candidate list is intentionally small and stable. If the upstream
#' repository adds additional wrappers, extend this function’s list.
#'
#'
#' @keywords internal
dt_find_wrapper_tests <- function(pkg_source_path) {
  tests_dir  <- file.path(pkg_source_path, "tests")
  candidates <- file.path(tests_dir, c(
    "main.R", "froll.R", "programming.R", "types.R",
    "nafill.R", "other.R", "S4.R", "autoprint.R"
  ))
  normalizePath(candidates[file.exists(candidates)], winslash = "/", mustWork = FALSE)
}


#' Compute untested functions from coverage using srcref spans (internal)
#'
#' Identifies package functions that were not executed by the provided coverage
#' object. A function is considered \emph{untested} if \strong{no} line within
#' its \code{srcref} span (from \code{start_line} to \code{end_line}) has a
#' positive execution count in \code{covr::tally_coverage()}.
#'
#' @param pkg_name character(1). The package name whose namespace is inspected
#'   for functions (\code{asNamespace(pkg_name)}).
#' @param coverage A coverage object as returned by \code{covr} (e.g.,
#'   \code{covr::environment_coverage()} or \code{covr::package_coverage()}),
#'   consumable by \code{covr::tally_coverage()}.
#'
#' @return \link[base:data.frame]{data.frame} with one row per untested function
#'   and columns:
#'   \describe{
#'     \item{function_name}{Function symbol in the package namespace.}
#'     \item{file}{Absolute, normalized path to the source file containing the function.}
#'     \item{start_line}{Start line (inclusive) from the function's \code{srcref}.}
#'     \item{end_line}{End line (inclusive) from the function's \code{srcref}.}
#'   }
#'   Returns an empty data.frame with these columns if (a) the coverage table is
#'   empty, (b) no functions have an \code{srcref}, or (c) all functions had at
#'   least one executed line within their span.
#'
#' @section Details:
#' \itemize{
#'   \item Functions without an \code{srcref} are skipped.
#'   \item Files are matched by \emph{basename} between the coverage table and the
#'         function’s \code{srcref} source path to make the join resilient to
#'         differing absolute paths in temporary build trees.
#'   \item A line is considered executed when its coverage \code{value} is a finite
#'         number greater than zero.
#' }
#'
#' @seealso \code{\link[covr]{tally_coverage}}
#' @keywords internal
dt_compute_untested_functions <- function(pkg_name, coverage) {
  tab <- covr::tally_coverage(coverage)
  if (is.null(tab) || !nrow(tab)) {
    return(data.frame(function_name = character(), file = character(),
                      start_line = integer(), end_line = integer(),
                      stringsAsFactors = FALSE))
  }
  tab$file_base <- basename(tab$filename)
  
  ns   <- asNamespace(pkg_name)
  objs <- ls(ns, all.names = TRUE)
  
  out <- lapply(objs, function(nm) {
    obj <- tryCatch(get(nm, envir = ns), error = function(e) NULL)
    if (!is.function(obj)) return(NULL)
    sref <- attr(obj, "srcref")
    if (is.null(sref)) return(NULL)
    
    srcfile <- attr(sref, "srcfile")
    fpath   <- if (!is.null(srcfile) && !is.null(srcfile$filename)) srcfile$filename else NA_character_
    if (is.na(fpath)) return(NULL)
    
    data.frame(
      function_name = nm,
      file          = normalizePath(fpath, winslash = "/", mustWork = FALSE),
      file_base     = basename(fpath),
      start_line    = sref[[1]],  # start line
      end_line      = sref[[3]],  # end line
      stringsAsFactors = FALSE
    )
  })
  fun_df <- do.call(rbind, out)
  if (is.null(fun_df) || !nrow(fun_df)) {
    return(data.frame(function_name = character(), file = character(),
                      start_line = integer(), end_line = integer(),
                      stringsAsFactors = FALSE))
  }
  
  fun_df$tested <- FALSE
  for (i in seq_len(nrow(fun_df))) {
    sub <- tab[tab$file_base == fun_df$file_base[i] &
                 tab$line >= fun_df$start_line[i] &
                 tab$line <= fun_df$end_line[i], , drop = FALSE]
    # 'value' = hit count; treat missing/NA as zero
    fun_df$tested[i] <- any(is.finite(sub$value) & sub$value > 0)
  }
  
  fun_df[!fun_df$tested, c("function_name", "file", "start_line", "end_line")]
}



#' Prepare package-specific test environment for data.table (internal)
#'
#' For data.table, Only "mirror" mode is supported here: 
#' copy inst/tests/* -> tests/ and clean up.
#'
#' Returns a list with:
#'   - test_files: character() of wrapper files to execute
#'   - teardown:   function() to revert any changes (delete mirrored files, unset env)
#'
#' @keywords internal
prepare_tests_for_package_dt <- function(pkg_name,
                                      pkg_source_path,
                                      test_files,
                                      mode = c("mirror")) {
  mode <- match.arg(mode)
  teardown <- function() {}
  tests_to_run <- test_files
  
  if (identical(pkg_name, "data.table")) {
    inst_tests_dir <- file.path(pkg_source_path, "inst", "tests")
    tests_dir      <- file.path(pkg_source_path, "tests")
    main           <- file.path(tests_dir, "main.R")
    raw_in_inst    <- file.path(inst_tests_dir, "tests.Rraw")
    raw_in_tests   <- file.path(tests_dir, "tests.Rraw")
    
    created <- character(0)
    
    # If tests/tests.Rraw is missing but inst/tests/tests.Rraw exists,
    # mirror top-level files from inst/tests -> tests (non-recursive).
    if (!file.exists(raw_in_tests) && file.exists(raw_in_inst)) {
      if (mode == "mirror") {
        message("data.table: MIRROR mode -> copying inst/tests/* -> tests/ to satisfy test.data.table()")
        dir.create(tests_dir, showWarnings = FALSE, recursive = TRUE)
        inst_files <- list.files(inst_tests_dir, all.files = FALSE, full.names = TRUE, no.. = TRUE)
        
        finfo <- file.info(inst_files)
        inst_files <- inst_files[!is.na(finfo$isdir) & !finfo$isdir]  # files only, top-level
        
        for (src in inst_files) {
          dest <- file.path(tests_dir, basename(src))
          if (!file.exists(dest)) {
            ok <- tryCatch(file.copy(src, dest, overwrite = FALSE), error = function(e) FALSE)
            if (isTRUE(ok)) {
              created <- c(created, normalizePath(dest, winslash = "/", mustWork = FALSE))
            }
          }
        }
      } else {
        message("data.table: MIRROR disabled; tests/tests.Rraw missing. Full suite may fail.")
      }
    }
    
    # Prefer canonical entrypoint (runs 5,000+ tests)
    # https://github.com/Rdatatable/data.table/blob/master/tests/main.R
    # https://rdatatable.gitlab.io/data.table/reference/test.data.table.html
    if (file.exists(main)) {
      tests_to_run <- normalizePath(main, winslash = "/", mustWork = FALSE)
    } else {
      # If main.R is missing, leave tests_to_run as provided; compute_test_files_clean()
      # will pick wrappers (e.g., froll.R, programming.R, etc.) as a fallback.
      message("data.table: tests/main.R not found; will rely on wrapper selection.")
    }
    
    # Teardown: remove only what we created via MIRROR
    teardown <- function() {
      if (length(created)) {
        try(unlink(created, force = TRUE), silent = TRUE)
      }
    }
  }
  
  list(test_files = tests_to_run, teardown = teardown)
}




#' Compute wrapper data.table test files to run for coverage (internal)
#'
#' Selects and sanitizes the list of **wrapper** test files to execute during
#' coverage, supporting both the standard `tests/` layout and the `testit`
#' layout (`tests/testit/`). When `pkg_name` is `"data.table"`, the function
#' adds package-aware behavior: it falls back to `tests/main.R` if the mapping
#' yields no existing files and optionally prioritizes `froll.R` so rolling
#' function tests run early.
#'
#' This helper **only resolves file paths**; it does **not** modify the
#' evaluation environment, copy/mirror test assets, or toggle data.table's
#' test harness mode. Any package-specific preparation (e.g., mirroring
#' `inst/tests/` to `tests/` or exposing `test.data.table()` in dev mode)
#' must be handled elsewhere (see `prepare_tests_for_package()`).
#'
#' @param pkg_name        `character(1)`. Package name.
#' @param pkg_source_path `character(1)`. Absolute or relative path to the
#'   package source root.
#' @param clean_mapping   A data frame or list containing at least a
#'   `test_file` column/element with **relative** wrapper paths to run.
#'   Paths are resolved against `tests/` or `tests/testit/` depending on
#'   `test_pkg_data$has_testit`.
#' @param test_pkg_data   `list`. Must include logical flag `has_testit`
#'   indicating whether the package uses a `tests/testit/` layout.
#' @param skip_tests      `NULL` or `character()`. Relative wrapper paths to
#'   exclude (matched within the chosen base directory).
#'
#' @return `character()` - **normalized**, de-duplicated vector of wrapper
#'   file paths that **exist** on disk. May be zero length if no candidates
#'   exist and no fallback is available.
#'
#' @keywords internal
compute_test_files_clean_dt <- function(pkg_name,
                                     pkg_source_path,
                                     clean_mapping,
                                     test_pkg_data,
                                     skip_tests = NULL) {
  if (identical(pkg_name, "data.table")) {
    tests_dir <- file.path(pkg_source_path, "tests")
    main      <- file.path(tests_dir, "main.R")
    
    # If main.R exists, use it (this calls test.data.table() -> full suite)
    if (file.exists(main)) {
      return(normalizePath(main, winslash = "/", mustWork = FALSE))
    }
    
    # Otherwise, fall back to wrapper scripts present under tests/.
    # Keep froll.R as first priority (legacy behavior),
    # then add broader wrappers commonly present in data.table/tests:
    # programming.R, types.R, nafill.R, other.R, S4.R
    # (See repo listing for these files.)
    # https://github.com/Rdatatable/data.table/blob/master/tests/
    priority_wrappers <- c("froll.R", "programming.R", "types.R", "nafill.R", "other.R", "S4.R")
    existing <- file.path(tests_dir, priority_wrappers)
    existing <- existing[file.exists(existing)]
    
    # If a mapping was provided, use it but sort by our priority; else use the priority set.
    test_files_col <- clean_mapping$test_file
    mapped <- character()
    if (!is.null(test_files_col)) {
      mapped <- file.path(tests_dir, test_files_col)
      mapped <- mapped[file.exists(mapped)]
    }
    
    # Apply skip list if provided
    if (!is.null(skip_tests) && length(skip_tests) > 0) {
      mapped   <- setdiff(mapped, file.path(tests_dir, skip_tests))
      existing <- setdiff(existing, file.path(tests_dir, skip_tests))
    }
    
    # Merge (mapped first, then priority list), then order by our explicit priority
    candidates <- unique(c(mapped, existing))
    if (!length(candidates)) {
      # As a last-ditch fallback, if someone mirrored and raw exists, try running tests.Rraw directly
      raw <- file.path(tests_dir, "tests.Rraw")
      if (file.exists(raw)) return(normalizePath(raw, winslash = "/", mustWork = FALSE))
      return(character())
    }
    
    # Sort by explicit priority order (keep froll.R first if present)
    nm <- basename(candidates)
    ord <- order(match(nm, priority_wrappers, nomatch = length(priority_wrappers) + seq_along(nm)))
    return(unique(normalizePath(candidates[ord], winslash = "/", mustWork = FALSE)))
  }
  
  ## ---- generic logic for other packages (unchanged) ----
  base_dir <- if (isTRUE(test_pkg_data$has_testit)) {
    file.path(pkg_source_path, "tests", "testit")
  } else {
    file.path(pkg_source_path, "tests")
  }
  
  test_files_col <- clean_mapping$test_file
  if (is.null(test_files_col)) test_files_col <- character()
  candidates <- file.path(base_dir, test_files_col)
  
  if (!is.null(skip_tests) && length(skip_tests) > 0) {
    candidates <- setdiff(candidates, file.path(base_dir, skip_tests))
  }
  
  test_files <- candidates[file.exists(candidates)]
  unique(normalizePath(test_files, winslash = "/", mustWork = FALSE))
  
  return(test_files)
}

# Helper: make a simple mapping df/list
make_mapping <- function(files) {
  # accept vector -> data.frame(test_file=..)
  if (is.null(files)) files <- character()
  data.frame(test_file = files, stringsAsFactors = FALSE)
}

test_that("standard layout: returns only existing wrappers and obeys skip list", {
  pkg_source_path <- "mockpkg"
  test_pkg_data   <- list(has_testit = FALSE)
  clean_mapping   <- make_mapping(c("a.R", "b.R", "c.R"))
  skip_tests      <- "b.R"
  
  # Stub normalizePath to identity to keep paths predictable
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  # Only mockpkg/tests/a.R and mockpkg/tests/c.R "exist"
  existing <- file.path(pkg_source_path, "tests", c("a.R", "c.R"))
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p %in% existing)
  
  got <- compute_test_files_clean_dt(
    pkg_name        = "mypkg",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = test_pkg_data,
    skip_tests      = skip_tests
  )
  
  expect_setequal(got, file.path(pkg_source_path, "tests", c("a.R", "c.R")))
  expect_false(any(grepl("b\\.R$", got)))
})

test_that("testit layout: resolves under tests/testit and includes all existing wrappers", {
  pkg_source_path <- "mockpkg"
  test_pkg_data   <- list(has_testit = TRUE)
  clean_mapping   <- make_mapping(c("test-foo.R", "test-bar.R"))
  
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  existing <- file.path(pkg_source_path, "tests", "testit", c("test-foo.R", "test-bar.R"))
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p %in% existing)
  
  got <- compute_test_files_clean_dt(
    pkg_name        = "mypkg",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = test_pkg_data
  )
  
  expect_setequal(got, existing)
  expect_true(all(grepl("^mockpkg/tests/testit/", got)))
})

test_that("data.table: falls back to tests/main.R when mapped wrappers do not exist", {
  pkg_source_path <- "mockpkg"
  test_pkg_data   <- list(has_testit = FALSE)
  # mapping points to wrappers that don't exist
  clean_mapping   <- make_mapping(c("nonexistent-1.R", "nonexistent-2.R"))
  
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  # No mapped files exist; but mockpkg/tests/main.R exists
  fallback <- file.path(pkg_source_path, "tests", "main.R")
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p == fallback)
  
  got <- compute_test_files_clean_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = test_pkg_data
  )
  
  expect_length(got, 1L)
  expect_equal(got, fallback)
})

test_that("data.table: prioritizes froll.R first, keeps rest in alphabetical order", {
  pkg_source_path <- "mockpkg"
  test_pkg_data   <- list(has_testit = FALSE)
  clean_mapping   <- make_mapping(c("z.R", "froll.R", "a.R"))
  
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  # All mapped files exist
  existing <- file.path(pkg_source_path, "tests", c("z.R", "froll.R", "a.R"))
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p %in% existing)
  
  got <- compute_test_files_clean_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = test_pkg_data
  )
  
  # Expected order: froll.R first, then a.R, z.R 
  # preserves input order among non-priority items
  expect_equal(
    got,
    file.path(pkg_source_path, "tests", c("froll.R", "z.R", "a.R"))
  )
})

test_that("returns character(0) when neither mapped files nor fallback exist", {
  pkg_source_path <- "mockpkg"
  test_pkg_data   <- list(has_testit = FALSE)
  clean_mapping   <- make_mapping("does-not-exist.R")
  
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  # Nothing exists, including fallback
  mockery::stub(compute_test_files_clean_dt, "file.exists",
                function(p) rep(FALSE, length(p)))
  
  got <- compute_test_files_clean_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = test_pkg_data
  )
  
  expect_length(got, 0L)
  expect_identical(got, character(0))
})

test_that("deduplicates and normalizes paths", {
  pkg_source_path <- "mockpkg"
  test_pkg_data   <- list(has_testit = FALSE)
  clean_mapping   <- make_mapping(c("dup.R", "dup.R", "other.R"))
  
  # normalizePath returns varied styles to test dedup (simulate OS differences)
  np_calls <- 0L
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) {
    # alternate between original and a fake 'normalized' copy to ensure uniquing works
    np_calls <<- np_calls + 1L
    path
  })
  
  existing <- file.path(pkg_source_path, "tests", c("dup.R", "dup.R", "other.R"))
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p %in% existing)
  
  got <- compute_test_files_clean_dt(
    pkg_name        = "mypkg",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = test_pkg_data
  )
  
  expect_setequal(got, existing)
  expect_length(got, 3L)      
  expect_true(all(file.exists(gsub("^mockpkg", tempdir(), got)) | TRUE)) # trivial assertion to avoid CRAN notes
})



test_that("data.table MIRROR: copies inst/tests files, records created, and teardown unlinks them", {
  pkg_source_path <- "mockpkg"
  test_files_in   <- file.path(pkg_source_path, "tests", "wrapper.R") # pretend wrapper exists; not used here
  
  # --- Paths used by prepare_tests_for_package_dt() ---
  inst_tests_dir <- file.path(pkg_source_path, "inst", "tests")
  tests_dir      <- file.path(pkg_source_path, "tests")
  raw_in_inst    <- file.path(inst_tests_dir, "tests.Rraw")
  raw_in_tests   <- file.path(tests_dir, "tests.Rraw")
  main_path      <- file.path(tests_dir, "main.R")
  
  # --- Stubs on prepare_tests_for_package_dt itself ---
  
  # mkdir is a no-op that always succeeds
  mockery::stub(prepare_tests_for_package_dt, "dir.create", function(...) TRUE)
  
  # inst/tests contains mix of file & directory; only files should be copied
  src_files <- file.path(inst_tests_dir, c("froll.Rraw", "other.Rraw.bz2", "subdir"))
  mockery::stub(prepare_tests_for_package_dt, "list.files", function(path, all.files, full.names, no..) {
    stopifnot(path == inst_tests_dir)
    src_files
  })
  
  # Mark first two as files, third as directory
  mockery::stub(prepare_tests_for_package_dt, "file.info", function(paths) {
    data.frame(isdir = basename(paths) == "subdir", row.names = paths)
  })
  
  # File existence logic to trigger MIRROR:
  # - inst/tests/tests.Rraw exists (to trigger mirroring)
  # - tests/tests.Rraw does NOT exist (so mirroring is needed)
  # - tests/main.R does NOT exist (so tests_to_run is left as provided)
  # - destination copies do NOT exist yet (so copying occurs)
  mockery::stub(prepare_tests_for_package_dt, "file.exists", function(p) {
    if (identical(p, raw_in_inst))  return(TRUE)   # source driver exists
    if (identical(p, raw_in_tests)) return(FALSE)  # target driver missing -> mirror
    if (identical(p, main_path))    return(FALSE)  # no main.R -> passthrough test_files
    # For copied file destinations and any other path: report missing so copy proceeds
    FALSE
  })
  
  # Copy succeeds for both files
  mockery::stub(prepare_tests_for_package_dt, "file.copy", function(src, dest, overwrite) TRUE)
  
  # normalizePath returns the path unchanged for predictability
  mockery::stub(prepare_tests_for_package_dt, "normalizePath", function(path, ...) path)
  
  # Capture unlink calls from teardown
  unlinked <- NULL
  mockery::stub(prepare_tests_for_package_dt, "unlink", function(paths, ...) {
    unlinked <<- paths
    TRUE
  })
  
  # --- Call under test ---
  res <- prepare_tests_for_package_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    test_files      = test_files_in,
    mode            = "mirror"
  )
  
  # Returned test files are passed through unchanged (since tests/main.R "doesn't exist")
  expect_equal(res$test_files, test_files_in)
  
  # When we call teardown, it should unlink the files we "created"
  res$teardown()
  
  # Expect the two copied files to have been scheduled for unlink
  expected_created <- file.path(pkg_source_path, "tests", basename(src_files[1:2]))
  expect_setequal(unlinked, expected_created)
})



test_that("data.table: falls back to tests/main.R when inst/tests driver is absent (no message)", {
  pkg_source_path <- "mockpkg"
  test_files_in   <- character(0)
  
  inst_tests_dir <- file.path(pkg_source_path, "inst", "tests")
  tests_dir      <- file.path(pkg_source_path, "tests")
  raw_in_inst    <- file.path(inst_tests_dir, "tests.Rraw")
  raw_in_tests   <- file.path(tests_dir, "tests.Rraw")
  fallback       <- file.path(tests_dir, "main.R")
  
  # file.exists: main.R exists; both tests.Rraw do not
  mockery::stub(prepare_tests_for_package_dt, "file.exists", function(p) {
    if (identical(p, fallback))     return(TRUE)   # main.R present
    if (identical(p, raw_in_inst))  return(FALSE)  # inst driver missing
    if (identical(p, raw_in_tests)) return(FALSE)  # tests driver missing
    FALSE
  })
  
  # normalizePath is identity
  mockery::stub(prepare_tests_for_package_dt, "normalizePath", function(path, ...) path)
  
  # Should be silent (no message when inst/tests/tests.Rraw is absent)
  expect_silent({
    res <- prepare_tests_for_package_dt(
      pkg_name        = "data.table",
      pkg_source_path = pkg_source_path,
      test_files      = test_files_in,
      mode            = "mirror"
    )
  })
  
  # Falls back to tests/main.R
  expect_equal(res$test_files, fallback)
  
  # Teardown should be a no-op (no mirroring -> nothing created)
  unlinked <- "not-called"
  mockery::stub(prepare_tests_for_package_dt, "unlink", function(paths, ...) {
    unlinked <<- paths
    TRUE
  })
  res$teardown()
  expect_identical(unlinked, "not-called")
})


test_that("data.table MIRROR: does not copy when destination exists; teardown does not unlink", {
  pkg_source_path <- "mockpkg"
  test_files_in   <- file.path(pkg_source_path, "tests", "wrapper.R")
  
  inst_tests_dir  <- file.path(pkg_source_path, "inst", "tests")
  tests_dir       <- file.path(pkg_source_path, "tests")
  dests           <- file.path(tests_dir, c("froll.Rraw", "other.Rraw.bz2"))
  
  # inst/tests exists
  mockery::stub(prepare_tests_for_package_dt, "dir.exists", function(p) p == inst_tests_dir)
  
  mockery::stub(prepare_tests_for_package_dt, "dir.create", function(...) TRUE)
  mockery::stub(prepare_tests_for_package_dt, "list.files", function(...) {
    file.path(inst_tests_dir, c("froll.Rraw", "other.Rraw.bz2"))
  })
  mockery::stub(prepare_tests_for_package_dt, "file.info", function(paths) {
    data.frame(isdir = rep(FALSE, length(paths)), row.names = paths)
  })
  
  # Destination already exists; hence no copy should occur and 'created' remains empty
  mockery::stub(prepare_tests_for_package_dt, "file.exists", function(p) {
    # Return TRUE for the two destination files; FALSE otherwise (e.g., wrapper checks)
    p %in% dests
  })
  # file.copy should never be called, but if it is, make it fail loudly
  mockery::stub(prepare_tests_for_package_dt, "file.copy", function(...) stop("should not copy"))
  mockery::stub(prepare_tests_for_package_dt, "normalizePath", function(path, ...) path)
  
  # Track unlink – should not be called since 'created' is empty
  unlink_called <- FALSE
  mockery::stub(prepare_tests_for_package_dt, "unlink", function(paths, ...) {
    unlink_called <<- TRUE
    TRUE
  })
  
  res <- prepare_tests_for_package_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    test_files      = test_files_in,
    mode            = "mirror"
  )
  
  expect_equal(res$test_files, test_files_in)
  
  res$teardown()
  expect_false(unlink_called)  # nothing was created; nothing unlinked
})


test_that("prepare_tests_for_package_dt: emits message when MIRROR disabled and tests/tests.Rraw is missing", {
  pkg_source_path <- "mockpkg"
  test_files_in   <- file.path(pkg_source_path, "tests", "wrapper.R")  # arbitrary passthrough
  
  # Paths used by the function
  inst_tests_dir <- file.path(pkg_source_path, "inst", "tests")
  tests_dir      <- file.path(pkg_source_path, "tests")
  raw_in_inst    <- file.path(inst_tests_dir, "tests.Rraw")
  raw_in_tests   <- file.path(tests_dir, "tests.Rraw")
  main_path      <- file.path(tests_dir, "main.R")
  
  # --- Stubs: drive the exact branch we want ---
  
  # Force mode != "mirror" regardless of call input
  mockery::stub(prepare_tests_for_package_dt, "match.arg", function(x) "disabled")
  
  # File existence matrix:
  # - inst driver exists (so mirroring would be needed)
  # - tests driver missing (so branch considers mirroring)
  # - main.R missing (to ensure passthrough of input test_files)
  mockery::stub(prepare_tests_for_package_dt, "file.exists", function(p) {
    if (identical(p, raw_in_inst))  return(TRUE)
    if (identical(p, raw_in_tests)) return(FALSE)
    if (identical(p, main_path))    return(FALSE)
    FALSE
  })
  
  # Identity normalizer for deterministic assertions
  mockery::stub(prepare_tests_for_package_dt, "normalizePath", function(path, ...) path)
  
  # Guard: if mirroring is attempted (it should NOT be), fail the test
  mockery::stub(prepare_tests_for_package_dt, "dir.create",
                function(...) stop("dir.create should not be called when MIRROR is disabled"))
  mockery::stub(prepare_tests_for_package_dt, "list.files",
                function(...) stop("list.files should not be called when MIRROR is disabled"))
  mockery::stub(prepare_tests_for_package_dt, "file.info",
                function(...) stop("file.info should not be called when MIRROR is disabled"))
  mockery::stub(prepare_tests_for_package_dt, "file.copy",
                function(...) stop("file.copy should not be called when MIRROR is disabled"))
  
  # Capture unlink from teardown (should NOT be called because nothing was created)
  unlinked <- "not-called"
  mockery::stub(prepare_tests_for_package_dt, "unlink", function(paths, ...) {
    unlinked <<- paths
    TRUE
  })
  
  # --- Expect the specific message and call SUT ---
  expect_message(
    {
      res <- prepare_tests_for_package_dt(
        pkg_name        = "data.table",
        pkg_source_path = pkg_source_path,
        test_files      = test_files_in,
        mode            = "mirror"  # input is ignored by our match.arg stub above
      )
    },
    regexp = "MIRROR disabled; tests/tests\\.Rraw missing\\. Full suite may fail\\."
  )
  
  # Since main.R "doesn't exist", the function should passthrough the input test_files
  expect_identical(res$test_files, test_files_in)
  
  # Teardown should be a no-op (no created files -> unlink should not be called)
  res$teardown()
  expect_identical(unlinked, "not-called")
})



test_that("run_covr_skip_dt_nstf: happy path (main.R, mirror OK, coverage collected)", {
  # Create a real, existing package source directory + tests folder
  pkg_source_path <- tempfile("mockpkg_")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)     # NEW
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE)       # NEW
  test_main <- file.path(pkg_source_path, "tests", "main.R")
  file.create(test_main)                                                  # optional but harmless
  
  # Make sure the temp files are cleaned after the test
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  # --- Stubs (order matters: stub before calling the function) ---
  mockery::stub(run_covr_skip_dt_nstf, "get_pkg_name", function(...) "data.table")
  mockery::stub(run_covr_skip_dt_nstf, "compute_test_files_clean_dt",
                function(pkg_name, pkg_source_path, ..., skip_tests) test_main)
  mockery::stub(run_covr_skip_dt_nstf, "normalizePath", function(path, ...) path)
  mockery::stub(run_covr_skip_dt_nstf, "file.exists", function(p) TRUE)
  
  # Loading/datasets/namespace
  mockery::stub(run_covr_skip_dt_nstf, "pkgload::load_all", function(...) TRUE)
  mockery::stub(run_covr_skip_dt_nstf, "ls", function(envir, all.names) character(0))
  mockery::stub(run_covr_skip_dt_nstf, "load_package_datasets", function(...) character(0))
  
  # Mirror prep returns teardown no-op
  teardown_called <- FALSE
  mockery::stub(run_covr_skip_dt_nstf, "prepare_tests_for_package_dt", function(...) {
    list(teardown = function() { teardown_called <<- TRUE; invisible(TRUE) })
  })
  
  # covr pipeline
  mockery::stub(run_covr_skip_dt_nstf, "covr::environment_coverage",
                function(env, test_files) "COV")
  mockery::stub(run_covr_skip_dt_nstf, "covr::coverage_to_list",
                function(coverage) list(
                  filecoverage = matrix(1, 1, dimnames = list("R/file.R", NULL)),
                  totalcoverage = 97.82
                ))
  
  # wrappers present -> only main.R is selected, others become "skipped"
  mockery::stub(run_covr_skip_dt_nstf, "dt_find_wrapper_tests", function(pkg_source_path) {
    c(file.path(pkg_source_path, "tests", "main.R"),
      file.path(pkg_source_path, "tests", "froll.R"),
      file.path(pkg_source_path, "tests", "programming.R"))
  })
  
  # 39 untested functions (data.frame with rows = 39)
  untested_df <- data.frame(function_name = paste0("f", 1:39),
                            file = "R/foo.R", start_line = 1, end_line = 1,
                            stringsAsFactors = FALSE)
  mockery::stub(run_covr_skip_dt_nstf, "dt_compute_untested_functions",
                function(pkg_name, coverage) untested_df)
  
  # capture the inputs to compute_total_coverage to assert correctness
  cap <- new.env(parent = emptyenv())
  mockery::stub(run_covr_skip_dt_nstf, "compute_total_coverage",
                function(percent_cov, functions_no_tests_df, tests_passing, tests_skipped) {
                  cap$percent_cov   <- percent_cov
                  cap$no_tests_df   <- functions_no_tests_df
                  cap$tests_passing <- tests_passing
                  cap$tests_skipped <- tests_skipped
                  list(
                    total_cov           = percent_cov,
                    percent_cov         = percent_cov,
                    n_testfiles         = length(tests_passing),
                    n_no_function_tests = nrow(functions_no_tests_df),
                    n_skipped_files     = length(tests_skipped)
                  )
                })
  
  # cleanup stub (called at the end; we just no-op)
  mockery::stub(run_covr_skip_dt_nstf, "cleanup_and_return_null", function(...) TRUE)
  
  # --- Call under test (pass cov_env now) ---
  cov_env <- new.env(parent = emptyenv())   # NEW
  res <- run_covr_skip_dt_nstf(pkg_source_path, test_pkg_data = list(), cov_env = cov_env)  # NEW
  
  # --- Assertions ---
  expect_type(res, "list")
  expect_true(all(c("total_cov", "res_cov", "functions_no_tests",
                    "tests_skipped", "tests_passing") %in% names(res)))
  
  # coverage summary echoes 97.82
  expect_equal(res$total_cov, 97.82)
  expect_equal(res$res_cov$coverage$totalcoverage, 97.82)
  
  # tests_passing = main.R; tests_skipped = the other wrappers
  expect_identical(res$tests_passing, test_main)
  expect_setequal(res$tests_skipped,
                  c(file.path(pkg_source_path, "tests", "froll.R"),
                    file.path(pkg_source_path, "tests", "programming.R")))
  
  # functions_no_tests has 39 rows (as mocked)
  expect_equal(nrow(res$functions_no_tests), 39)
  
  # compute_total_coverage received the correct inputs
  expect_identical(cap$tests_passing, res$tests_passing)
  expect_setequal(cap$tests_skipped, res$tests_skipped)
  expect_equal(cap$percent_cov, 97.82)
  expect_equal(nrow(cap$no_tests_df), 39)
  
  # on.exit teardown executed
  expect_true(teardown_called)
})




test_that("run_covr_skip_dt_nstf: returns zeroed structure when coverage is NULL", {
  # Create a real, existing package source directory + tests folder
  pkg_source_path <- tempfile("mockpkg_")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)     # NEW
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE)       # NEW
  test_main <- file.path(pkg_source_path, "tests", "main.R")
  file.create(test_main)                                                  # optional but safe
  
  # Ensure cleanup after the test
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  # --- Stubs (set BEFORE calling the SUT) ---
  
  # This runner is data.table-only
  mockery::stub(run_covr_skip_dt_nstf, "get_pkg_name", function(...) "data.table")
  
  # Resolve tests: pretend tests/main.R is the chosen entrypoint
  mockery::stub(run_covr_skip_dt_nstf, "compute_test_files_clean_dt",
                function(pkg_name, pkg_source_path, ..., skip_tests) test_main)
  
  # filesystem helpers kept predictable
  mockery::stub(run_covr_skip_dt_nstf, "normalizePath", function(path, ...) path)
  mockery::stub(run_covr_skip_dt_nstf, "file.exists", function(p) TRUE)
  
  # Load package & datasets: no-op
  mockery::stub(run_covr_skip_dt_nstf, "pkgload::load_all", function(...) TRUE)
  mockery::stub(run_covr_skip_dt_nstf, "ls", function(envir, all.names) character(0))
  mockery::stub(run_covr_skip_dt_nstf, "load_package_datasets", function(...) character(0))
  
  # Prepare MIRROR: return a teardown function we can detect later
  teardown_called <- FALSE
  mockery::stub(run_covr_skip_dt_nstf, "prepare_tests_for_package_dt", function(...) {
    list(teardown = function() { teardown_called <<- TRUE; invisible(TRUE) })
  })
  
  # Force coverage to be NULL so the function returns the zeroed structure
  mockery::stub(run_covr_skip_dt_nstf, "covr::environment_coverage", function(...) NULL)
  
  # Make sure we DO NOT call coverage_to_list() on NULL (fail fast if we do)
  mockery::stub(run_covr_skip_dt_nstf, "covr::coverage_to_list",
                function(...) stop("should not be called when coverage is NULL"))
  
  # cleanup function should NOT be invoked on the NULL-coverage path
  cleanup_called <- FALSE
  mockery::stub(run_covr_skip_dt_nstf, "cleanup_and_return_null",
                function(...) { cleanup_called <<- TRUE; NULL })
  
  # --- Call under test; assert the user-facing message is emitted ---
  cov_env <- new.env(parent = emptyenv())  # NEW
  expect_message(
    {
      res <- run_covr_skip_dt_nstf(pkg_source_path, test_pkg_data = list(), cov_env = cov_env)  # NEW
    },
    regexp = "No test coverage for this configuration"
  )
  
  # --- Assertions for the returned zeroed structure (lines 118–134) ---
  
  # Structure and top-level fields
  expect_type(res, "list")
  expect_true(all(c("total_cov", "res_cov", "functions_no_tests",
                    "tests_skipped", "tests_passing") %in% names(res)))
  
  # total_cov = 0
  expect_identical(res$total_cov, 0)
  
  # res_cov sub-structure
  expect_true(all(c("name", "coverage", "errors", "notes") %in% names(res$res_cov)))
  expect_identical(res$res_cov$name, "data.table")
  expect_identical(res$res_cov$errors, "No coverage generated")
  expect_true(is.na(res$res_cov$notes))
  
  # coverage sub-structure
  cov_list <- res$res_cov$coverage
  expect_true(all(c("filecoverage", "totalcoverage") %in% names(cov_list)))
  expect_identical(cov_list$totalcoverage, 0)
  
  # filecoverage is a 1-row matrix with rowname "No functions tested"
  expect_true(is.matrix(cov_list$filecoverage))
  expect_identical(nrow(cov_list$filecoverage), 1L)
  expect_identical(rownames(cov_list$filecoverage), "No functions tested")
  
  # functions_no_tests: empty data.frame with the expected columns
  expect_true(is.data.frame(res$functions_no_tests))
  expect_identical(nrow(res$functions_no_tests), 0L)
  expect_true(all(c("function_name", "file") %in% names(res$functions_no_tests)))
  
  # tests_skipped / tests_passing are empty character vectors
  expect_identical(res$tests_skipped, character())
  expect_identical(res$tests_passing, character())
  
  # on.exit teardown ran
  expect_true(teardown_called)
  
  # cleanup path should NOT have been used in this branch
  expect_false(cleanup_called)
})




test_that("run_covr_skip_dt_nstf: early return when no test files are resolvable (length == 0)", {
  pkg_source_path <- tempfile("mockpkg_")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)   # NEW: ensure dir exists
  # Not required for this branch, but harmless if you want:
  # dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE)
  
  # Clean up temp dir after the test
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  # --- Stubs (defined BEFORE invoking SUT) ---
  
  # Ensure we are in the data.table-only code path
  mockery::stub(run_covr_skip_dt_nstf, "get_pkg_name", function(...) "data.table")
  
  # Force compute_test_files_clean_dt() to return an empty vector
  mockery::stub(run_covr_skip_dt_nstf, "compute_test_files_clean_dt", function(...) character(0))
  
  # normalizePath: pass-through (safe on character(0))
  mockery::stub(run_covr_skip_dt_nstf, "normalizePath", function(path, ...) path)
  
  # Guardrails: if heavy paths are reached, fail the test (they should NOT be)
  mockery::stub(run_covr_skip_dt_nstf, "pkgload::load_all",
                function(...) stop("load_all should not be called on early return"))
  mockery::stub(run_covr_skip_dt_nstf, "prepare_tests_for_package_dt",
                function(...) stop("prepare_tests_for_package_dt should not be called on early return"))
  mockery::stub(run_covr_skip_dt_nstf, "covr::environment_coverage",
                function(...) stop("environment_coverage should not be called on early return"))
  
  # Capture the cleanup message and return a recognizable sentinel value
  captured <- new.env(parent = emptyenv())
  mockery::stub(run_covr_skip_dt_nstf, "cleanup_and_return_null",
                function(msg, ...) {
                  captured$msg <- msg
                  structure(list(aborted = TRUE, reason = msg), class = "cleanup_result")
                })
  
  # --- Call SUT (pass cov_env now) ---
  cov_env <- new.env(parent = emptyenv())  # NEW
  res <- run_covr_skip_dt_nstf(pkg_source_path, test_pkg_data = list(), cov_env = cov_env)  # NEW
  
  # --- Assertions ---
  expect_s3_class(res, "cleanup_result")
  expect_true(res$aborted)
  expect_match(res$reason, "^No resolvable data\\.table tests for data\\.table$")
  # also confirm the stub captured the same message
  expect_identical(captured$msg, res$reason)
})



test_that("run_covr_skip_dt_nstf: early return when test files exist list is non-empty but all file.exists == FALSE", {
  pkg_source_path <- tempfile("mockpkg_")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)    # NEW: ensure dir exists
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))  # NEW: cleanup
  
  test_main <- "/tmp/does-not-exist/tests/main.R"
  
  # Ensure data.table path is taken
  mockery::stub(run_covr_skip_dt_nstf, "get_pkg_name", function(...) "data.table")
  
  # Return a non-empty vector of test files
  mockery::stub(run_covr_skip_dt_nstf, "compute_test_files_clean_dt",
                function(...) test_main)
  
  # Keep deterministic path values
  mockery::stub(run_covr_skip_dt_nstf, "normalizePath", function(path, ...) path)
  
  # Force file.exists to FALSE for the candidate path
  mockery::stub(run_covr_skip_dt_nstf, "file.exists",
                function(p) rep(FALSE, length(p)))
  
  # Guardrails: heavy paths must not be reached
  mockery::stub(run_covr_skip_dt_nstf, "pkgload::load_all",
                function(...) stop("load_all should not be called on early return"))
  mockery::stub(run_covr_skip_dt_nstf, "prepare_tests_for_package_dt",
                function(...) stop("prepare_tests_for_package_dt should not be called on early return"))
  mockery::stub(run_covr_skip_dt_nstf, "covr::environment_coverage",
                function(...) stop("environment_coverage should not be called on early return"))
  
  # Capture the cleanup and return sentinel
  captured <- new.env(parent = emptyenv())
  mockery::stub(run_covr_skip_dt_nstf, "cleanup_and_return_null",
                function(msg, ...) {
                  captured$msg <- msg
                  structure(list(aborted = TRUE, reason = msg), class = "cleanup_result")
                })
  
  # --- Call SUT (pass cov_env now) ---
  cov_env <- new.env(parent = emptyenv())  # NEW
  res <- run_covr_skip_dt_nstf(pkg_source_path, test_pkg_data = list(), cov_env = cov_env)  # NEW
  
  # --- Assertions ---
  expect_s3_class(res, "cleanup_result")
  expect_true(res$aborted)
  expect_match(res$reason, "^No resolvable data\\.table tests for data\\.table$")
  expect_identical(captured$msg, res$reason)
})



test_that("dt_find_wrapper_tests() returns only existing candidates (normalized)", {
  pkg_source_path <- "/tmp/mockpkg"
  tests_dir <- file.path(pkg_source_path, "tests")
  
  # Build the full candidate set as per the implementation
  candidates <- file.path(
    tests_dir,
    c("main.R", "froll.R", "programming.R", "types.R",
      "nafill.R", "other.R", "S4.R", "autoprint.R")
  )
  
  # Stub file.exists(): only some files exist
  exists_set <- candidates[c(1, 3)]  # main.R, programming.R
  mockery::stub(dt_find_wrapper_tests, "file.exists", function(p) p %in% exists_set)
  
  # Stub normalizePath to identity for deterministic expectations
  mockery::stub(dt_find_wrapper_tests, "normalizePath", function(path, ...) path)
  
  res <- dt_find_wrapper_tests(pkg_source_path)
  expect_type(res, "character")
  expect_equal(res, exists_set)
})

test_that("dt_find_wrapper_tests() returns empty when no candidates exist", {
  pkg_source_path <- "/tmp/mockpkg"
  
  mockery::stub(dt_find_wrapper_tests, "file.exists", function(p) FALSE)
  mockery::stub(dt_find_wrapper_tests, "normalizePath", function(path, ...) path)
  
  res <- dt_find_wrapper_tests(pkg_source_path)
  expect_identical(res, character())
})




# Helper used by the test only: create a function *without* srcref attributes
make_fun_without_srcref <- function() {
  f <- function(...) NULL
  attr(f, "srcref") <- NULL
  attr(f, "wholeSrcref") <- NULL
  f
}

make_fake_fun_with_srcref <- function(fname, start_line, end_line) {
  f <- function(...) NULL
  sref <- list()
  sref[[1]] <- as.integer(start_line)  # start line
  sref[[3]] <- as.integer(end_line)    # end line
  # scalar filename is critical; do not pass a vector here
  attr(sref, "srcfile") <- list(filename = as.character(fname)[1L])
  attr(f, "srcref") <- sref
  f
}

test_that("dt_compute_untested_functions(): returns only functions with zero covered lines in span", {
  pkg_name <- "data.table"
  
  # Use scalar file names for srcref
  file_a <- "/tmp/mockpkg/R/a.R"  # fun_tested (has one covered line in its span)
  file_b <- "/tmp/mockpkg/R/b.R"  # fun_untested (no covered lines)
  
  # Construct fake namespace env containing our objects
  fake_ns <- new.env(parent = emptyenv())
  assign("fun_tested",
         make_fake_fun_with_srcref(file_a, start_line = 10, end_line = 20),
         envir = fake_ns)
  assign("fun_untested",
         make_fake_fun_with_srcref(file_b, start_line = 5, end_line = 8),
         envir = fake_ns)
  # A function without srcref must have the attributes removed explicitly
  assign("fun_nosrc", make_fun_without_srcref(), envir = fake_ns)
  assign("not_a_fun", 123L, envir = fake_ns)
  
  # Stub asNamespace() to return our fake namespace
  mockery::stub(dt_compute_untested_functions, "asNamespace",
                function(pkg) { stopifnot(pkg == pkg_name); fake_ns })
  
  # Coverage table: only "a.R" has a covered line within [10,20]
  cov_tab <- data.frame(
    filename = c("R/a.R", "R/a.R"),
    line     = c(15L, 30L),   # 15 in span; 30 out of span (ignored)
    value    = c(2, 0),       # 2 hits at line 15; 0 hits at 30
    stringsAsFactors = FALSE
  )
  mockery::stub(dt_compute_untested_functions, "covr::tally_coverage",
                function(coverage) cov_tab)
  
  # normalizePath is identity for predictability
  mockery::stub(dt_compute_untested_functions, "normalizePath",
                function(path, ...) path)
  
  # ---- Call under test ----
  res <- dt_compute_untested_functions(pkg_name, coverage = "IGNORED")
  
  # ---- Assertions ----
  # Only fun_untested should appear:
  expect_true(is.data.frame(res))
  expect_equal(nrow(res), 1L)
  expect_setequal(names(res), c("function_name", "file", "start_line", "end_line"))
  expect_identical(res$function_name, "fun_untested")
  expect_identical(res$file, file_b)
  expect_identical(res$start_line, 5L)
  expect_identical(res$end_line, 8L)
})


test_that("dt_compute_untested_functions(): early return when tally_coverage() returns NULL", {
  pkg_name <- "data.table"
  
  # Stub covr::tally_coverage to return NULL (simulate no coverage table)
  mockery::stub(
    dt_compute_untested_functions,
    "covr::tally_coverage",
    function(coverage) NULL
  )
  
  # Ensure nothing else is called after the early return:
  # If these are called, the test should fail.
  mockery::stub(
    dt_compute_untested_functions,
    "asNamespace",
    function(pkg) stop("asNamespace should not be called on NULL coverage")
  )
  mockery::stub(
    dt_compute_untested_functions,
    "normalizePath",
    function(...) stop("normalizePath should not be called on NULL coverage")
  )
  
  res <- dt_compute_untested_functions(pkg_name, coverage = "IGNORED")
  
  # Returned structure is an empty data.frame with the expected columns
  expect_true(is.data.frame(res))
  expect_identical(nrow(res), 0L)
  expect_setequal(
    names(res),
    c("function_name", "file", "start_line", "end_line")
  )
})

test_that("dt_compute_untested_functions(): early return when tally_coverage() returns empty data.frame", {
  pkg_name <- "data.table"
  
  # Stub covr::tally_coverage to return a 0-row data.frame
  mockery::stub(
    dt_compute_untested_functions,
    "covr::tally_coverage",
    function(coverage) {
      data.frame(
        filename = character(),
        line     = integer(),
        value    = numeric(),
        stringsAsFactors = FALSE
      )
    }
  )
  
  # Ensure nothing else is called after the early return:
  mockery::stub(
    dt_compute_untested_functions,
    "asNamespace",
    function(pkg) stop("asNamespace should not be called on empty coverage")
  )
  mockery::stub(
    dt_compute_untested_functions,
    "normalizePath",
    function(...) stop("normalizePath should not be called on empty coverage")
  )
  
  res <- dt_compute_untested_functions(pkg_name, coverage = "IGNORED")
  
  # Returned structure is an empty data.frame with the expected columns
  expect_true(is.data.frame(res))
  expect_identical(nrow(res), 0L)
  expect_setequal(
    names(res),
    c("function_name", "file", "start_line", "end_line")
  )
})


test_that("dt_compute_untested_functions(): early return when fun_df is NULL", {
  pkg_name <- "data.table"
  
  # 1) coverage table is non-empty (pass earlier guard)
  mockery::stub(
    dt_compute_untested_functions,
    "covr::tally_coverage",
    function(coverage) data.frame(filename = "R/a.R", line = 1L, value = 1)
  )
  
  # 2) fake namespace with one symbol
  fake_ns <- new.env(parent = emptyenv())
  assign("x", 123L, envir = fake_ns)
  
  mockery::stub(
    dt_compute_untested_functions,
    "asNamespace",
    function(pkg) { stopifnot(pkg == pkg_name); fake_ns }
  )
  
  # 3) ensure we "see" one object in the namespace
  mockery::stub(
    dt_compute_untested_functions,
    "ls",
    function(envir, all.names) "x"
  )
  
  # 4) force get() to return a non-function (so lapply element becomes NULL)
  mockery::stub(
    dt_compute_untested_functions,
    "get",
    function(nm, envir) 123L
  )
  
  # 5) normalizePath shouldn't matter, but keep deterministic
  mockery::stub(
    dt_compute_untested_functions,
    "normalizePath",
    function(path, ...) path
  )
  
  # ---- call SUT ----
  res <- dt_compute_untested_functions(pkg_name, coverage = "IGNORED")
  
  # ---- assertions ----
  expect_true(is.data.frame(res))
  expect_identical(nrow(res), 0L)
  expect_setequal(names(res), c("function_name", "file", "start_line", "end_line"))
})


test_that("dt_compute_untested_functions(): early return when fun_df has zero rows", {
  pkg_name <- "data.table"
  
  # 1) non-empty coverage table again
  mockery::stub(
    dt_compute_untested_functions,
    "covr::tally_coverage",
    function(coverage) data.frame(filename = "R/a.R", line = 1L, value = 1)
  )
  
  # 2) asNamespace / ls / get can be anything; they won't affect do.call stub
  mockery::stub(
    dt_compute_untested_functions,
    "asNamespace",
    function(pkg) new.env(parent = emptyenv())
  )
  mockery::stub(
    dt_compute_untested_functions,
    "ls",
    function(envir, all.names) character(0)
  )
  mockery::stub(
    dt_compute_untested_functions,
    "get",
    function(nm, envir) NULL
  )
  
  # 3) Force do.call(rbind, out) to return a 0-row data.frame
  mockery::stub(
    dt_compute_untested_functions,
    "do.call",
    function(what, args) {
      # zero-row frame with expected columns would also work,
      # but any 0-row data.frame triggers !nrow(fun_df)
      data.frame(function_name = character(),
                 file          = character(),
                 start_line    = integer(),
                 end_line      = integer(),
                 stringsAsFactors = FALSE)
    }
  )
  
  # 4) keep normalizePath deterministic
  mockery::stub(
    dt_compute_untested_functions,
    "normalizePath",
    function(path, ...) path
  )
  
  # ---- call SUT ----
  res <- dt_compute_untested_functions(pkg_name, coverage = "IGNORED")
  
  # ---- assertions ----
  expect_true(is.data.frame(res))
  expect_identical(nrow(res), 0L)
  expect_setequal(names(res), c("function_name", "file", "start_line", "end_line"))
})


test_that("compute_test_files_clean_dt: applies skip_tests to both mapped and existing", {
  pkg_source_path <- "/tmp/mockpkg"
  tests_dir <- file.path(pkg_source_path, "tests")
  
  # clean_mapping lists two mapped test files; 'froll.R' is on the priority list,
  # 'alpha.R' is not (so its position comes after priority wrappers).
  clean_mapping <- list(test_file = c("alpha.R", "froll.R"))
  
  # We want existing wrappers (from priority set) to include 'programming.R' and 'types.R'.
  # We'll simulate existence via file.exists() stub.
  paths_exist <- c(
    file.path(tests_dir, "alpha.R"),
    file.path(tests_dir, "froll.R"),
    file.path(tests_dir, "programming.R"),
    file.path(tests_dir, "types.R")
  )
  # Important: Ensure tests/main.R does NOT exist so the function does not early-return there.
  main_path <- file.path(tests_dir, "main.R")
  
  # --- Stubs for isolation ---
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) {
    # vectorized truth for the arguments passed by the function
    p %in% paths_exist
  })
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  # Skip 'alpha.R' (mapped) and 'programming.R' (existing wrapper)
  skip_tests <- c("alpha.R", "programming.R")
  
  out <- compute_test_files_clean_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = list(),     # not used in DT-specific branch
    skip_tests      = skip_tests
  )
  
  # After skip:
  # - mapped   initially: {alpha.R, froll.R} -> remove alpha.R -> {froll.R}
  # - existing initially: {programming.R, types.R} -> remove programming.R -> {types.R}
  # The function sorts by explicit priority, so expected order: froll.R, types.R
  expected <- c(
    file.path(tests_dir, "froll.R"),
    file.path(tests_dir, "types.R")
  )
  expect_identical(out, expected)
})

test_that("compute_test_files_clean_dt: skip_tests=NULL does not filter mapped/existing", {
  pkg_source_path <- "/tmp/mockpkg"
  tests_dir <- file.path(pkg_source_path, "tests")
  
  clean_mapping <- list(test_file = c("alpha.R", "froll.R"))
  
  # Simulate these files as existing:
  paths_exist <- c(
    file.path(tests_dir, "alpha.R"),
    file.path(tests_dir, "froll.R"),
    file.path(tests_dir, "programming.R")
  )
  
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p %in% paths_exist)
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  out <- compute_test_files_clean_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = list(),
    skip_tests      = NULL  # <- no skipping
  )
  
  # mapped: {alpha.R, froll.R}
  # existing (priority-present): {programming.R}
  # Ordered by DT priority wrappers: froll.R (priority), programming.R (priority), alpha.R (non-priority -> after)
  expected <- c(
    file.path(tests_dir, "froll.R"),
    file.path(tests_dir, "programming.R"),
    file.path(tests_dir, "alpha.R")
  )
  expect_identical(out, expected)
})

test_that("compute_test_files_clean_dt: empty skip_tests vector does not filter", {
  pkg_source_path <- "/tmp/mockpkg"
  tests_dir <- file.path(pkg_source_path, "tests")
  
  clean_mapping <- list(test_file = c("alpha.R", "froll.R"))
  paths_exist <- c(
    file.path(tests_dir, "alpha.R"),
    file.path(tests_dir, "froll.R"),
    file.path(tests_dir, "types.R")
  )
  
  mockery::stub(compute_test_files_clean_dt, "file.exists", function(p) p %in% paths_exist)
  mockery::stub(compute_test_files_clean_dt, "normalizePath", function(path, ...) path)
  
  out <- compute_test_files_clean_dt(
    pkg_name        = "data.table",
    pkg_source_path = pkg_source_path,
    clean_mapping   = clean_mapping,
    test_pkg_data   = list(),
    skip_tests      = character(0)  # length == 0 => branch not taken
  )
  
  # mapped: {alpha.R, froll.R}
  # existing: {types.R}
  # expected priority order: froll.R, types.R, alpha.R
  expected <- c(
    file.path(tests_dir, "froll.R"),
    file.path(tests_dir, "types.R"),
    file.path(tests_dir, "alpha.R")
  )
  expect_identical(out, expected)
})

#' Run Coverage and Return Structured Coverage Results
#'
#' This function executes code coverage analysis for a package using the
#' `covr` framework. It is typically used after a package has been installed
#' locally and test files are available. The function runs coverage in an
#' isolated process (using `callr`) and returns a structured summary of
#' overall and file‑level coverage.
#'
#' @param pkg_source_path Character. Path to the installed package directory
#'   from which coverage should be computed.
#' @param timeout Numeric. Timeout (in seconds) passed to
#'   \code{callr::r_safe()} when running coverage. This limits the maximum
#'   time allowed for executing the tests underlying the coverage analysis.
#'
#' @details
#' The function invokes covr’s coverage evaluation in a clean R session and
#' extracts both:
#' \itemize{
#'   \item total coverage: percentage of lines covered across the package, and
#'   \item function/file‑level coverage: coverage data for individual files.
#' }
#'
#' It is used internally by higher‑level functions such as
#' \code{run_covr_modes()} and skip‑aware coverage wrappers in the Standard
#' Testing Framework (STF).
#'
#' @return
#' A named list containing:
#' \describe{
#'   \item{total_cov}{Numeric. Aggregated coverage percentage for the package.}
#'   \item{res_cov}{A list containing file‑level or function‑level coverage
#'         results as returned by `covr`. This includes per‑file coverage,
#'         errors, and diagnostic notes if present.}
#' }
#'
#' Returns \code{NULL} if coverage could not be computed.
#'
#' @examples
#' \donttest{
#' # Save and adjust CRAN mirror for reproducibility
#' r <- getOption("repos")
#' old <- options(repos = r)
#' r["CRAN"] <- "http://cran.us.r-project.org"
#' options(repos = r)
#'
#' # Example package from test.assessr
#' dp <- system.file(
#'   "test-data",
#'   "test.package.0001_0.1.0.tar.gz",
#'   package = "test.assessr"
#' )
#'
#' # Set up package source directory
#' install_list <- set_up_pkg(dp)
#' pkg_source_path <- install_list$pkg_source_path
#'
#' # Install locally to enable testing and coverage
#' package_installed <- install_package_local(pkg_source_path)
#'
#' if (isTRUE(package_installed)) {
#'   coverage_results <- run_coverage(pkg_source_path)
#' }
#'
#' # Restore user's original repository settings
#' options(old)
#' }
#'
#' @export
run_coverage <- function(pkg_source_path, timeout = Inf) {
  
  pkg_name <- basename(pkg_source_path)
  
  message(paste0("running code coverage for ", pkg_name))
  
  # run covr
  res_cov <- tryCatch({
    coverage_list <- run_covr(pkg_source_path, timeout)
    
    # If no testable functions are found in the package, `filecoverage` and `totalcoverage`
    # will yield logical(0) and NaN respectively. Coerce to usable format
    if(is.na(coverage_list$totalcoverage)){
      if(rlang::is_empty(coverage_list$filecoverage) && is.logical(coverage_list$filecoverage)){
        coverage_list$totalcoverage <- 0
        notes <- "no testable functions found"
      }else{
        message("Total coverage returned NaN. This likely means the package had non-standard characteristics.")
        notes <- NA
      }
    }else{
      notes <- NA
    }
    
    list(name = pkg_name, coverage = coverage_list, errors = NA, notes = notes)
  },
  error = function(cond){
    coverage_list <- list(filecoverage = NA, totalcoverage = NA_integer_)
    list(
      name = pkg_name, coverage = coverage_list,
      errors = cond,
      notes = NA
    )
  })
  
  if(is.na(res_cov$coverage$totalcoverage)) {
    message(paste0("code coverage for ", pkg_name, " unsuccessful"))
  } else {
    message(paste0("code coverage for ", pkg_name, " successful"))
  }
  
  # return total coverage as fraction
  total_cov <- as.numeric(res_cov$coverage$totalcoverage/100)
  
  if(is.na(total_cov)){
    message(paste0("R coverage for ", pkg_name, " failed. Read in the covr output to see what went wrong: "))
  }
  
  if(!is.na(res_cov$notes)){
    message(paste0("R coverage for ", pkg_name, " had notes: ", res_cov$notes))
  }
  
  covr_list <- list(
    total_cov = total_cov,
    res_cov = res_cov
  )
  return(covr_list)
}


#' Run covr in a Subprocess with a Timeout
#'
#' This internal helper runs coverage evaluation in a separate R subprocess
#' using `callr`. It allows coverage to be computed with a specified timeout,
#' ensuring that long‑running or hanging test processes do not block further
#' analysis.
#'
#' @param path Character. Path to the package source directory for which
#'   coverage should be computed.
#' @param timeout Numeric. Maximum time (in seconds) allowed for the coverage
#'   computation. Defaults to \code{Inf}, meaning no timeout is applied unless
#'   explicitly set.
#'
#' @details
#' The function launches a safe R session via \code{callr::r_safe()} and
#' executes the relevant covr routines within that isolated environment.
#' This is used internally by higher‑level STF/NSTF coverage functions to
#' ensure reproducible and controlled coverage execution.
#'
#' @return
#' A coverage object returned by the underlying covr call, or \code{NULL} if
#' the subprocess execution fails or times out.
#'
#' @keywords internal
run_covr <- function(path, timeout) {
  callr::r_safe(
    function(p) {
      covr::coverage_to_list(covr::package_coverage(p, type = "tests"))
    },
    args = list(path),
    libpath = .libPaths(),
    repos = NULL,
    package = FALSE,
    user_profile = FALSE,
    error = "error",
    timeout = timeout
  )
}

test_that("run_covr returns successful coverage list", {
  path <- "my/pkg/path"
  timeout <- 60
  expected_output <- list(totalcoverage = 80, filecoverage = c(a = 80))
  
  # Mock callr::r_safe to return the expected coverage list directly
  mockery::stub(run_covr, "callr::r_safe", expected_output)
  
  result <- run_covr(path, timeout)
  
  testthat::expect_equal(result, expected_output)
})

test_that("run_covr passes correct arguments to r_safe", {
  path <- "test_path"
  timeout <- 100
  
  m <- mockery::mock(list(totalcoverage = 0))
  mockery::stub(run_covr, "callr::r_safe", m)
  
  run_covr(path, timeout)
  
  # Verify callr was called with correct args and timeout
  args <- mockery::mock_args(m)[[1]]
  testthat::expect_equal(args$args, list(path))
  testthat::expect_equal(args$timeout, timeout)
})

test_that("run_covr propagates errors from callr", {
  # Mock callr to throw an error (e.g., timeout or process crash)
  mockery::stub(run_covr, "callr::r_safe", function(...) stop("Timeout reached"))
  
  testthat::expect_error(run_covr("path", 1), "Timeout reached")
})

test_that("the anonymous function inside run_covr calls covr functions correctly", {
  test_path <- "mock/pkg/path"
  
  # 1. Setup mocks for the functions called INSIDE the anonymous closure
  # We use mockery::mock to track if they were called
  m_pkg_cov <- mockery::mock("raw_coverage_object")
  m_cov_list <- mockery::mock(list(totalcoverage = 100))
  
  # 2. Stub r_safe so we can capture the anonymous function
  m_r_safe <- mockery::mock(TRUE)
  mockery::stub(run_covr, "callr::r_safe", m_r_safe)
  
  # Call the wrapper to trigger the stub
  run_covr(test_path, timeout = 10)
  
  # 3. Extract the anonymous function (it's the first argument to r_safe)
  r_safe_args <- mockery::mock_args(m_r_safe)[[1]]
  inner_fun <- r_safe_args[[1]] # The anonymous function(p) { ... }
  
  # 4. Execute the extracted function in an environment where covr is mocked
  # We use with_mocked_bindings from testthat or mockery::stub logic
  withr::with_package("covr", {
    mockery::stub(inner_fun, "covr::package_coverage", m_pkg_cov, depth = 1)
    mockery::stub(inner_fun, "covr::coverage_to_list", m_cov_list, depth = 1)
    
    inner_fun(test_path)
  })
  
  # 5. Assertions: Did the inner function pass the path to covr?
  mockery::expect_called(m_pkg_cov, 1)
  testthat::expect_equal(mockery::mock_args(m_pkg_cov)[[1]][[1]], test_path)
  
  # Did it pass the results of package_coverage to coverage_to_list?
  mockery::expect_called(m_cov_list, 1)
  testthat::expect_equal(mockery::mock_args(m_cov_list)[[1]][[1]], "raw_coverage_object")
})

test_that("run_coverage handles success, empty-tests, and failure paths", {
  pkg_source_path <- "dummy/path/test.package.0001"
  pkg_name <- basename(pkg_source_path)
  
  # -----------------------------
  # 1) SUCCESS: 100% coverage -> total_cov = 1
  # -----------------------------
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      # IMPORTANT: top-level fields (NOT nested under $coverage)
      list(
        filecoverage  = c(100, 100),  # any non-NULL vector is fine
        totalcoverage = 100
      )
    }
  )
  
  covr_list_success <- NULL
  testthat::expect_message({
    covr_list_success <- run_coverage(pkg_source_path)
  }, regexp = "successful")
  
  testthat::expect_equal(covr_list_success$total_cov, 1)                 # 100/100
  testthat::expect_equal(covr_list_success$res_cov$coverage$totalcoverage, 100)
  testthat::expect_true(is.na(covr_list_success$res_cov$notes))
  
  # -----------------------------
  # 2) EMPTY TESTS: logical(0) + NA -> coerced to 0 + note
  # -----------------------------
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      list(
        filecoverage  = logical(0),    # triggers "no testable functions found"
        totalcoverage = NA_real_
      )
    }
  )
  
  covr_list_empty <- NULL
  # Will print "successful" (after coercion to 0) and later "had notes: ..."
  testthat::expect_message({
    covr_list_empty <- run_coverage(pkg_source_path)
  }, regexp = "had notes: no testable functions found")
  
  testthat::expect_equal(covr_list_empty$total_cov, 0)                   # 0/100
  testthat::expect_identical(
    covr_list_empty$res_cov$notes, 
    "no testable functions found"
  )
  testthat::expect_equal(covr_list_empty$res_cov$coverage$totalcoverage, 0)
  
  # -----------------------------
  # 3) TRUE FAILURE: NA totalcoverage (non-empty/non-logical filecoverage)
  # -----------------------------
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      list(
        filecoverage  = c(NA_real_),   # not logical(0) -> won't be coerced to 0
        totalcoverage = NA_real_
      )
    }
  )
  
  covr_list_fail <- NULL
  testthat::expect_message({
    covr_list_fail <- run_coverage(pkg_source_path)
  }, regexp = "unsuccessful")
  
  testthat::expect_true(is.na(covr_list_fail$total_cov))
  testthat::expect_true(is.na(covr_list_fail$res_cov$coverage$totalcoverage))
  testthat::expect_true(is.na(covr_list_fail$res_cov$notes))
})





test_that("running coverage for created package with 1 note 1 warning - mocked", {
  # Define dummy paths and package name
  pkg_source_path <- "dummy/path/test.package.0002"
  pkg_name <- basename(pkg_source_path)
  
  # Pattern used by run_coverage() to announce success/unsuccessful
  msg_pattern <- paste0("code coverage for ", pkg_name, " (successful|unsuccessful)")
  
  # --- Success Path: 100% coverage -> total_cov == 1 ---
  mockery::stub(
    where = run_coverage,      # stub INSIDE the function under test
    what  = "run_covr",        # replace the dependency it calls
    how   = function(path, timeout) {
      # NOTE: run_coverage() expects a FLAT list with top-level fields:
      # totalcoverage (0..100) and filecoverage (non-NULL vector)
      list(
        filecoverage  = c(func1 = 100),
        totalcoverage = 100
      )
    }
  )
  
  covr_list <- NULL
  testthat::expect_message({
    covr_list <- run_coverage(pkg_source_path)
  }, regexp = msg_pattern)
  
  testthat::expect_true(checkmate::test_numeric(covr_list$total_cov))
  testthat::expect_equal(covr_list$total_cov, 1)  # 100 / 100
  testthat::expect_equal(covr_list$res_cov$coverage$totalcoverage, 100)
  
  # --- Failure Path: NA totalcoverage (non-empty, non-logical filecoverage) ---
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      list(
        filecoverage  = c(NA_real_),   # not logical(0) -> won't be coerced to 0
        totalcoverage = NA_real_
      )
    }
  )
  
  covr_list_fail <- NULL
  testthat::expect_message({
    covr_list_fail <- run_coverage(pkg_source_path)
  }, regexp = paste0("code coverage for ", pkg_name, " unsuccessful"), fixed = TRUE)
  
  testthat::expect_true(is.na(covr_list_fail$total_cov))
})




test_that("running coverage for created package with 1 note 1 error - mocked", {
  # Define dummy package info
  pkg_source_path <- "dummy/path/test.package.0003"
  pkg_name <- basename(pkg_source_path)
  
  # NOTE: In the current implementation, the error path in run_coverage()
  # sets `notes = NA`. You cannot have both a note and an error simultaneously.
  # We therefore simulate an rcmdcheck-like failure by making run_covr() throw.
  
  # Stub run_covr INSIDE run_coverage to throw an error
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      stop("Simulated error from rcmdcheck failure")
    }
  )
  
  # --- Test Execution ---
  covr_list <- NULL
  # run_coverage() emits multiple messages; assert the "failed" one
  testthat::expect_message({
    covr_list <- run_coverage(pkg_source_path)
  }, regexp = paste0("R coverage for ", pkg_name, " failed"))
  
  # --- Assertions ---
  results <- list(covr = covr_list$total_cov)
  
  # total_cov is numeric NA for a failed run
  testthat::expect_true(checkmate::test_numeric(results$covr))
  testthat::expect_true(is.na(results$covr))
  
  # Optional: also assert the earlier "unsuccessful" message is produced
  testthat::expect_message(
    run_coverage(pkg_source_path),
    regexp = paste0("code coverage for ", pkg_name, " unsuccessful"),
    fixed = TRUE
  )
  
  # The error object should be captured in res_cov$errors; notes should be NA
  testthat::expect_true(inherits(covr_list$res_cov$errors, "error"))
  testthat::expect_true(is.na(covr_list$res_cov$notes))
})


test_that("running coverage for created package with no tests - mocked", {
  
  
  # Define dummy package info
  pkg_source_path <- "dummy/path/test.package.0004"
  pkg_name <- basename(pkg_source_path)
  
  # --- Setup Mock ---
  
  # Mock run_coverage to simulate the "no tests" scenario
  # total_cov is 0 (coerced from NaN/empty logical)
  mock_cov_no_tests <- list(
    total_cov = 0,
    res_cov = list(
      name = pkg_name,
      coverage = list(totalcoverage = 0, filecoverage = logical(0)),
      errors = NA,
      notes = "no testable functions found"
    )
  )
  
  # Stub run_coverage to return the "no tests" object
  mockery::stub(test_that, "run_coverage", mock_cov_no_tests)
  
  # --- Test Execution ---
  
  covr_list <- run_coverage(pkg_source_path)
  
  # --- Assertions ---
  
  results <- list(covr = covr_list$total_cov)
  
  # Assertions for a package with 0 tests
  testthat::expect_true(checkmate::test_numeric(results$covr))
  testthat::expect_equal(results$covr, NaN)
  
})


test_that("running coverage for created package with no tests - mocked", {
  # Define dummy package info
  pkg_source_path <- "dummy/path/test.package.0004"
  pkg_name <- basename(pkg_source_path)
  
  # --- Setup Mock ---
  # Stub run_covr INSIDE run_coverage to simulate "no testable functions found":
  # totalcoverage = NA_real_ with filecoverage = logical(0) -> coerced to 0 with a note
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      list(
        filecoverage  = logical(0),
        totalcoverage = NA_real_
      )
    }
  )
  
  # --- Test Execution ---
  covr_list <- run_coverage(pkg_source_path)
  
  # --- Assertions ---
  results <- list(covr = covr_list$total_cov)
  
  # For a package with no tests, run_coverage coerces to 0 (not NaN)
  testthat::expect_true(checkmate::test_numeric(results$covr))
  testthat::expect_equal(results$covr, 0)
  
  # Optional: assert the messages your implementation emits
  # (1) overall status -> "successful" because totalcoverage was coerced to 0
  testthat::expect_message(
    run_coverage(pkg_source_path),
    regexp = paste0("code coverage for ", pkg_name, " successful"),
    fixed  = TRUE
  )
  
  # (2) note message emitted afterward
  testthat::expect_message(
    run_coverage(pkg_source_path),
    regexp = "had notes: no testable functions found"
  )
  
  # Optional: check the note is set in the result structure
  testthat::expect_identical(
    covr_list$res_cov$notes,
    "no testable functions found"
  )
  testthat::expect_equal(covr_list$res_cov$coverage$totalcoverage, 0)
})




test_that("run_coverage handles errors in coverage_list correctly", {
  # Create a temporary package source path (optional here since we mock run_covr)
  temp_pkg_path <- tempfile("tempPkg")
  dir.create(temp_pkg_path, recursive = TRUE)
  writeLines("Package: tempPkg\nVersion: 0.1.0\n", file.path(temp_pkg_path, "DESCRIPTION"))
  
  # --- Stub the dependency used by run_coverage() ---
  # Force the error path by making run_covr() throw.
  mockery::stub(
    where = run_coverage,
    what  = "run_covr",
    how   = function(path, timeout) {
      stop("Simulated error from run_covr / rcmdcheck failure")
    }
  )
  
  # --- Execute ---
  result <- NULL
  # Optional: assert the "failed" message your implementation emits
  testthat::expect_message({
    result <- run_coverage(temp_pkg_path, timeout = 1)
  }, regexp = paste0("R coverage for ", basename(temp_pkg_path), " failed"))
  
  # --- Assertions: structure from error branch ---
  # coverage NA values indicate error
  expect_true(is.na(result$res_cov$coverage$filecoverage))
  expect_true(is.na(result$res_cov$coverage$totalcoverage))
  
  # total_cov is computed from totalcoverage/100 -> remains NA on error
  expect_true(is.na(result$total_cov))
  
  # errors field captures the condition object
  expect_false(is.null(result$res_cov$errors))
  expect_true(inherits(result$res_cov$errors, "error"))
  
  # Optional: the earlier status message is "unsuccessful" in this path
  testthat::expect_message(
    run_coverage(temp_pkg_path, timeout = 1),
    regexp = paste0("code coverage for ", basename(temp_pkg_path), " unsuccessful"),
    fixed  = TRUE
  )
})



test_that("run_coverage captures all messages", {
  # Mock the run_covr function to simulate different scenarios
  mock_run_covr <- function(path, timeout) {
    if (path == "no_testable_functions") {
      # EXACT inputs for the "no tests" path:
      # - filecoverage must be logical(0)
      # - totalcoverage NA/NaN is fine; your code checks is.na()
      return(list(filecoverage = logical(0), totalcoverage = NA_real_))
    } else if (path == "nan_total_coverage") {
      # A "non-standard" NA/NaN path (not 'no tests'):
      # Keep filecoverage NOT logical(0). A list() works: is_empty = TRUE, is.logical = FALSE
      return(list(filecoverage = list(), totalcoverage = NaN))
    } else if (path == "successful_coverage") {
      # Success path: totalcoverage numeric (0..100); filecoverage content irrelevant here
      return(list(filecoverage = c(func1 = 80), totalcoverage = 80))
    } else {
      stop("Unexpected path")
    }
  }
  
  with_mocked_bindings(
    run_covr = mock_run_covr,
    {
      # --- Test case: No testable functions found ---
      messages <- capture_messages(run_coverage("no_testable_functions"))
      expect_true(any(grepl("running code coverage for no_testable_functions", messages)))
      # In this branch, your code coerces to 0 and prints "successful"
      expect_true(any(grepl("code coverage for no_testable_functions successful", messages)))
      expect_true(any(grepl("R coverage for no_testable_functions had notes: no testable functions found", messages)))
      
      # --- Test case: Total coverage returned NaN (non-standard) ---
      messages <- capture_messages(run_coverage("nan_total_coverage"))
      expect_true(any(grepl("running code coverage for nan_total_coverage", messages)))
      # This informational message is emitted in the NA/NaN-but-not-empty-logical path:
      expect_true(any(grepl("Total coverage returned NaN\\.", messages)))
      expect_true(any(grepl("code coverage for nan_total_coverage unsuccessful", messages)))
      expect_true(any(grepl("R coverage for nan_total_coverage failed\\. Read in the covr output to see what went wrong:", messages)))
      
      # --- Test case: Successful coverage ---
      messages <- capture_messages(run_coverage("successful_coverage"))
      expect_true(any(grepl("running code coverage for successful_coverage", messages)))
      expect_true(any(grepl("code coverage for successful_coverage successful", messages)))
    }
  )
})



test_that("running coverage for created package in tar file with no notes (mocked)", {
  pkg_name <- "mocked.package.0001"
  pkg_source_path <- file.path(tempdir(), pkg_name)
  
  # 1) Create mocks (mockery::mock returns a callable that tracks calls)
  mock_setup <- mockery::mock(list(
    build_vignettes   = FALSE,
    package_installed = TRUE,
    pkg_source_path   = pkg_source_path,
    rcmdcheck_args    = list(path = pkg_source_path)
  ))
  mock_install   <- mockery::mock(TRUE)
  mock_run_covr  <- mockery::mock(list(filecoverage = c(func1 = 100), totalcoverage = 100))
  
  # 2) Bind the mocks in this test environment
  set_up_pkg            <- mock_setup
  install_package_local <- mock_install
  run_covr              <- mock_run_covr
  
  # 3) Execute logic using the mocks
  install_list <- set_up_pkg("dummy_path.tar.gz")
  package_installed <- install_package_local(install_list$pkg_source_path)
  
  testthat::skip_if_not(package_installed)
  
  coverage_list <- NULL
  testthat::expect_message({
    tryCatch({
      coverage_list <- run_covr(install_list$pkg_source_path, timeout = Inf)
      message(paste0("code coverage for ", basename(install_list$pkg_source_path), " successful"))
    }, error = function(e) message("unsuccessful"))
  }, regexp = "successful")
  
  # 4) Assertions
  testthat::expect_equal(coverage_list$totalcoverage, 100)
  
  # Call count assertions
  mockery::expect_called(mock_setup, 1)
  mockery::expect_called(mock_install, 1)
  mockery::expect_called(mock_run_covr, 1)
  
  # Optional: verify arguments passed to the mocks
  mockery::expect_args(mock_setup,            1, "dummy_path.tar.gz")
  mockery::expect_args(mock_install,          1, install_list$pkg_source_path)
  mockery::expect_args(mock_run_covr,         1, install_list$pkg_source_path, Inf)
})

#' Check for tests/testthat and _snaps folder and count golden tests
#'
#' @param pkg_source_path Path to the root of the package source
#'
#' @return A list with:
#'   - `has_testthat`: Does tests/testthat exist?
#'   - `has_testit`: Does tests/testit exist?
#'   - `has_tests_base`: Do tests exist in tests subdirectory?
#'   - `has_BioG_test`: Do BioGenerics tests exist in inst/tests subdirectory?
#'   - `has_RUnit_test`: Do RUnit tests exist (inst/unitTests; optionally tests/run_unitTests.R)?
#'   - `bioc_unit_tests_dir` : location of BioGenerics unit tests
#'   - `bioc_run_ut_path`: location of BioGenerics `run_unitTests.R`
#'   - `has_snaps`: Does _snaps exist inside tests/testthat?
#'   - `n_golden_tests`: Number of snapshot files inside _snaps
#'   - `n_test_files`: Number of test-*.R files inside tests/testthat
#'   - `has_tinytest` : Does tinytest exist? 
#'   - `tinytest_dir` : location of tinytest  unit tests
#'   - `tinytest_runner_paths` : location of tinytest runner paths
#'   - `n_tinytest_files` :  Number of test_*.R files under inst/tinytest  
#'   - `has_data_table_tests`: Do data.table tests exist in directories?
#'   - `data_table_tests_path` : location of data.table unit tests
#' @keywords internal
check_pkg_tests_and_snaps <- function(pkg_source_path) {
  
  message("checking package test config")
  test_dir       <- file.path(pkg_source_path, "tests")
  testthat_path  <- file.path(test_dir, "testthat")
  snaps_path     <- file.path(testthat_path, "_snaps")
  
  testit_path    <- file.path(test_dir, "testit")
  test_ci_path   <- file.path(test_dir, "test-ci")
  test_cran_path <- file.path(test_dir, "test-cran")

  # --- tinytest discovery (inst/tinytest + common runner files in tests/) ---
  tinytest_dir <- file.path(pkg_source_path, "inst", "tinytest")
  # Some packages expose a tinytest runner in tests/ (names vary)
  tinytest_runner_candidates <- c(
    file.path(test_dir, "tinytest.R"),
    file.path(test_dir, "runTinyTests.R"),
    file.path(test_dir, "run_tinytest.R")
  )
  has_tinytest_dir    <- dir.exists(tinytest_dir)
  tinytest_runner_vec <- tinytest_runner_candidates[file.exists(tinytest_runner_candidates)]
  has_tinytest_runner <- length(tinytest_runner_vec) > 0
  has_tinytest        <- has_tinytest_dir || has_tinytest_runner
  
  
  # Check for testthat and testit (standard and nonstandard)
  has_testthat <- dir.exists(testthat_path)
  has_testit <- dir.exists(testit_path) || (dir.exists(test_ci_path) && dir.exists(test_cran_path))

  # ---------- data.table-specific detection ----------
  # Canonical entrypoint and raw drivers:
  # - tests/main.R   (calls test.data.table() -> full suite)  [preferred]
  # - tests/tests.Rraw or tests/tests.Rraw.bz2                [mirrored]
  # - inst/tests/tests.Rraw or inst/tests/tests.Rraw.bz2      [source driver]
  # that this detection is intentionally broad to support 
  # data.table-like test structures in other packages.
  dt_main            <- file.path(test_dir, "main.R")
  dt_raw_tests       <- file.path(test_dir, "tests.Rraw")
  dt_raw_tests_bz2   <- paste0(dt_raw_tests, ".bz2")
  dt_raw_inst        <- file.path(pkg_source_path, "inst", "tests", "tests.Rraw")
  dt_raw_inst_bz2    <- paste0(dt_raw_inst, ".bz2")
  
  has_dt_main        <- file.exists(dt_main)
  has_dt_raw_tests   <- file.exists(dt_raw_tests)    || file.exists(dt_raw_tests_bz2)
  has_dt_raw_inst    <- file.exists(dt_raw_inst)     || file.exists(dt_raw_inst_bz2)
  
  has_data_table_tests <- has_dt_main || has_dt_raw_tests || has_dt_raw_inst
  
  # Choose the best path to run:
  data_table_tests_path <- if (has_dt_main) {
    dt_main
  } else if (has_dt_raw_tests) {
    if (file.exists(dt_raw_tests)) dt_raw_tests else dt_raw_tests_bz2
  } else if (has_dt_raw_inst) {
    if (file.exists(dt_raw_inst)) dt_raw_inst else dt_raw_inst_bz2
  } else {
    NA_character_
  }
  
  # snapshots ----------
  
  has_snaps <- dir.exists(snaps_path)
  n_golden_tests <- if (has_snaps) {
    snapshot_files <- list.files(snaps_path, recursive = TRUE, full.names = TRUE)
    length(snapshot_files)
  } else {
    0
  }
  
  # ---------- testthat test-*.R count ----------
  n_test_files <- if (has_testthat) {
    test_files <- list.files(
      testthat_path,
      pattern = "^test-.*\\.R$",
      recursive = TRUE,
      full.names = TRUE
    )
    test_files <- test_files[!grepl("_snaps", test_files)]
    length(test_files)
  } else {
    0
  }
  

  # --- Count tinytest test files (any *.R under inst/tinytest) ---
  n_tinytest_files <- if (has_tinytest_dir) {
    list.files(tinytest_dir, pattern = "\\.[rR]$", recursive = TRUE, full.names = TRUE) |>
      length()
  } else {
    0
  }
  
  # ---------- Bioconductor RUnit (BiocGenerics-style) ----------
  bioc_unit_tests_dir <- file.path(pkg_source_path, "inst", "unitTests")
  has_bioc_unit_tests <- dir.exists(bioc_unit_tests_dir)
  bioc_unit_tests_dir <- if (has_bioc_unit_tests) bioc_unit_tests_dir else NA_character_
  
  bioc_run_ut_path <- file.path(pkg_source_path, "tests", "run_unitTests.R")
  has_bioc_run_unitTests <- file.exists(bioc_run_ut_path)
  bioc_run_ut_path <- if (has_bioc_run_unitTests) bioc_run_ut_path else NA_character_
  
  zzz_path <- file.path(pkg_source_path, "R", "zzz.R")
  has_bioc_dot_test <- FALSE
  if (file.exists(zzz_path)) {
    zzz_lines <- tryCatch(readLines(zzz_path, warn = FALSE), error = function(e) character())
    has_bioc_dot_test <- any(grepl("\\.test\\s*<-\\s*function\\b", zzz_lines))
  }
  has_BioG_test <- has_bioc_unit_tests && has_bioc_run_unitTests && has_bioc_dot_test
  
  # CRAN RUnit runner (e.g. fBasics, many legacy CRAN packages)
  doRUnit_path <- file.path(pkg_source_path, "tests", "doRUnit.R")
  has_doRUnit <- file.exists(doRUnit_path)
  doRUnit_path <- if (has_doRUnit) doRUnit_path else NA_character_
  
  # ---------- RUnit-only (inst/unitTests; optionally tests/run_unitTests.R) ----------
  # Broader than BioG: does not require .test <- function in zzz.R
  # Treat as "RUnit-only" by excluding full BioG/BiocGenerics layouts
  has_RUnit_test <- has_bioc_unit_tests && !has_BioG_test
  
  # ---------- base-R style tests (only if no known frameworks) ----------
  has_tests_base <- FALSE
  if (!has_data_table_tests && !has_tinytest && 
      !has_testthat && !has_testit && 
      !has_RUnit_test && !has_BioG_test &&
      !dir.exists(test_ci_path) && !dir.exists(test_cran_path)) {
    
    base_test_files <- list.files(
      path         = test_dir,
      pattern      = "\\.r$",        # match .R or .r
      full.names   = TRUE,
      recursive    = FALSE,          # only top-level
      ignore.case  = TRUE,           # handle .R vs .r
      include.dirs = FALSE,          # don’t return directories
      no..         = TRUE            # exclude '.' and '..'
    )
    
    # Exclude RUnit launchers (BioC and CRAN conventions)
    base_test_files <- base_test_files[
      !basename(base_test_files) %in% c("run_unitTests.R", "doRUnit.R")
    ]
    has_tests_base <- length(base_test_files) > 0
  }
  
  # set up list
  test_types <- list(
    has_testthat = has_testthat,
    has_snaps = has_snaps,
    has_testit = has_testit,
    has_tests_base = has_tests_base,
    has_BioG_test =  has_BioG_test,
    has_RUnit_test = has_RUnit_test,
    bioc_unit_tests_dir = bioc_unit_tests_dir,
    bioc_run_ut_path = bioc_run_ut_path,
    doRUnit_path = doRUnit_path,
    n_golden_tests = n_golden_tests,
    n_test_files = n_test_files,
    has_tinytest          = has_tinytest,
    tinytest_dir          = if (has_tinytest_dir) tinytest_dir else NA_character_,
    tinytest_runner_paths = if (has_tinytest_runner) tinytest_runner_vec else character(0),
    n_tinytest_files      = n_tinytest_files,
    has_data_table_tests  = has_data_table_tests,
    data_table_tests_path = data_table_tests_path
  )

  return(test_types)

}



#' Compute approximate total coverage combining line coverage and test breadth
#'
#' @description
#' Computes an approximate "total coverage" metric by combining:
#' 1) line coverage percentage (`percent_cov`, 0–100), and
#' 2) a conservative proxy for breadth of tests based on counts of passing
#'    test files, functions without tests, and skipped test files.
#'
#' The final metric scales line coverage (as a proportion) by the fraction of
#' tested over total function-space, where:
#' - `n_testfiles` = number of passing test files (if 0, treated as 1 to avoid divide-by-zero),
#' - `n_no_function_tests` = number of functions lacking tests (`nrow(functions_no_tests_df)`),
#' - `n_skipped_files` = number of skipped test files (`length(tests_skipped)`),
#' - `total_functions` = `n_tested + n_untested` with `n_tested = max(n_testfiles, 1)`,
#' - `total_cov` = (percent_cov / 100) * (n_tested / total_functions).
#'
#' @details
#' This function is intentionally conservative:
#' - Passing test *files* are used as a proxy for tested functions.
#' - Skipped test files contribute to "untested breadth".
#' - If there are zero passing test files, `n_tested` is set to 1 to avoid
#'   divide-by-zero and produce a small non-zero denominator.
#'
#' @param percent_cov Numeric scalar. Line coverage percentage (0–100).
#' @param functions_no_tests_df A `data.frame` (or `NULL`) of functions lacking tests.
#' @param tests_passing A character vector (or `NULL`) of passing test file paths/names.
#' @param tests_skipped A character vector (or `NULL`) of skipped test file paths/names.
#'
#' @return A named list with:
#' \itemize{
#'   \item \code{total_cov} (numeric): Approximate total coverage (proportion 0–1).
#'   \item \code{percent_cov_round} (numeric): \code{total_cov} rounded to 2 decimals (still 0–1).
#'   \item \code{n_testfiles} (integer): Count of passing test files.
#'   \item \code{n_no_function_tests} (integer): Count of functions lacking tests.
#'   \item \code{n_skipped_files} (integer): Count of skipped test files.
#' }
#'
#' @keywords internal
compute_total_coverage <- function(
    percent_cov,
    functions_no_tests_df,
    tests_passing,
    tests_skipped
) {
  
  # Clamp to [0, 100] to avoid pathological inputs
  percent_cov <- max(0, min(100, percent_cov))
  
  # --- Counts (approximate breadth) ---
  n_no_function_tests <- if (!is.null(functions_no_tests_df)) nrow(functions_no_tests_df) else 0L
  n_skipped_files     <- if (!is.null(tests_skipped)) length(tests_skipped) else 0L
  
  # Treat skipped test files as contributing to "untested breadth"
  n_untested <- n_no_function_tests + n_skipped_files
  
  # Proxy for "tested functions": number of passing test files (conservative)
  n_testfiles <- if (!is.null(tests_passing)) length(tests_passing) else 0L
  
  # Avoid divide-by-zero if coverage exists but counts are zero
  n_tested <- if (n_testfiles > 0L) n_testfiles else 1L
  
  # Approximate total function-space
  total_functions <- n_tested + n_untested
  
  # Rounded proportion (still 0..1)
  percent_cov_round <- round(percent_cov, 2)
  
  # Return the requested elements
  calc_cov_list <- list(
    total_cov = percent_cov,
    percent_cov_round = percent_cov_round,
    n_testfiles = n_testfiles,
    n_no_function_tests = n_no_function_tests,
    n_skipped_files = n_skipped_files
  )
  
  return(calc_cov_list)
}  

test_that("check_pkg_tests_and_snaps works as expected", {
  # Create a temp directory to act as our package
  pkg_dir <- tempfile("dummyPkg")
  dir.create(pkg_dir)
  
  # Create tests/testthat/_snaps structure
  testthat_dir <- file.path(pkg_dir, "tests", "testthat")
  snaps_dir <- file.path(testthat_dir, "_snaps")
  dir.create(snaps_dir, recursive = TRUE)
  
  # Create some dummy golden snapshot files
  file.create(file.path(snaps_dir, "snapshot1.md"))
  file.create(file.path(snaps_dir, "snapshot2.md"))
  
  # Call the function
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  # Check the results
  expect_true(result$has_testthat)
  expect_false(result$has_testit)
  expect_true(result$has_snaps)
  expect_equal(result$n_golden_tests, 2)
  
  # Clean up (optional)
  unlink(pkg_dir, recursive = TRUE)
})

test_that("works when both testthat and _snaps exist with files", {
  pkg_dir <- tempfile("dummyPkg1")
  dir.create(pkg_dir)
  
  testthat_dir <- file.path(pkg_dir, "tests", "testthat")
  snaps_dir <- file.path(testthat_dir, "_snaps")
  dir.create(snaps_dir, recursive = TRUE)
  
  file.create(file.path(snaps_dir, "snapshot1.md"))
  file.create(file.path(snaps_dir, "snapshot2.md"))
  
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  expect_true(result$has_testthat)
  expect_false(result$has_testit)  
  expect_true(result$has_snaps)
  expect_equal(result$n_golden_tests, 2)
  
  unlink(pkg_dir, recursive = TRUE)
})


test_that("works when both tinytest and _snaps exist with files", {
  pkg_dir <- tempfile("dummyPkg1")
  dir.create(pkg_dir)
  
  # tinytest (canonical): inst/tinytest with at least one test file
  tiny_dir <- file.path(pkg_dir, "inst", "tinytest")
  dir.create(tiny_dir, recursive = TRUE)
  writeLines("expect_true(TRUE)", file.path(tiny_dir, "test_basic.R"))
  
  # testthat snapshots live in tests/testthat/_snaps
  snaps_dir <- file.path(pkg_dir, "tests", "testthat", "_snaps")
  dir.create(snaps_dir, recursive = TRUE)
  file.create(file.path(snaps_dir, "snapshot1.md"))
  file.create(file.path(snaps_dir, "snapshot2.md"))
  
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  expect_true(result$has_tinytest)
  expect_equal(result$n_golden_tests, 2)     # presence of .md alone shouldn't bump this
  expect_true(result$has_snaps)              # optional but consistent with the test title
  expect_gt(result$n_tinytest_files, 0)      # optional: verifies tinytest detection
  
  unlink(pkg_dir, recursive = TRUE)
})



test_that("works when neither testthat nor _snaps exist", {
  pkg_dir <- tempfile("dummyPkg3")
  dir.create(pkg_dir)
  
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  expect_false(result$has_testthat)
  expect_false(result$has_testit)  
  expect_false(result$has_snaps)
  expect_equal(result$n_golden_tests, 0)
  
  unlink(pkg_dir, recursive = TRUE)
})


test_that("works when _snaps exists but is empty", {
  pkg_dir <- tempfile("dummyPkg4")
  dir.create(pkg_dir)
  
  testthat_dir <- file.path(pkg_dir, "tests", "testthat")
  snaps_dir <- file.path(testthat_dir, "_snaps")
  dir.create(snaps_dir, recursive = TRUE)
  
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  expect_true(result$has_testthat)
  expect_false(result$has_testit)  
  expect_true(result$has_snaps)
  expect_equal(result$n_golden_tests, 0)
  
  unlink(pkg_dir, recursive = TRUE)
})


test_that("counts nested snapshot files", {
  pkg_dir <- tempfile("dummyPkg5")
  dir.create(pkg_dir)
  
  testthat_dir <- file.path(pkg_dir, "tests", "testthat")
  snaps_dir <- file.path(testthat_dir, "_snaps")
  nested_dir <- file.path(snaps_dir, "nested")
  dir.create(nested_dir, recursive = TRUE)
  
  file.create(file.path(snaps_dir, "snapshot1.md"))
  file.create(file.path(nested_dir, "snapshot2.md"))
  
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  expect_true(result$has_testthat)
  expect_false(result$has_testit)  
  expect_true(result$has_snaps)
  expect_equal(result$n_golden_tests, 2)
  
  unlink(pkg_dir, recursive = TRUE)
})


# testit

test_that("detects testit framework presence", {
  pkg_dir <- tempfile("dummyPkg_testit")
  dir.create(pkg_dir)
  
  testit_dir <- file.path(pkg_dir, "tests", "testit")
  dir.create(testit_dir, recursive = TRUE)
  
  result <- check_pkg_tests_and_snaps(pkg_dir)
  
  expect_false(result$has_testthat)
  expect_true(result$has_testit)  
  unlink(pkg_dir, recursive = TRUE)
})



test_that("normal case: combines line coverage and breadth proxy correctly", {
  percent_cov <- 80  # 80% line coverage
  functions_no_tests_df <- data.frame(fn = c("a", "b", "c"))
  tests_passing <- c("test-a.R", "test-b.R")
  tests_skipped <- c("test-slow.R", "test-experimental.R")
  
  res <- compute_total_coverage(
    percent_cov = percent_cov,
    functions_no_tests_df = functions_no_tests_df,
    tests_passing = tests_passing,
    tests_skipped = tests_skipped
  )
  
  # Counts
  expect_equal(res$n_no_function_tests, 3L)
  expect_equal(res$n_skipped_files, 2L)
  expect_equal(res$n_testfiles, 2L)
  
  # Derived
  n_untested <- 3 + 2
  n_tested <- 2
  total_functions <- n_tested + n_untested
  expected_total_cov <- 80
  
  expect_equal(res$total_cov, expected_total_cov)
  expect_equal(res$percent_cov_round, round(expected_total_cov, 2))
})

test_that("NULL inputs handled: no rows, no lengths; n_tested falls back to 1", {
  res <- compute_total_coverage(
    percent_cov = 50,
    functions_no_tests_df = NULL,
    tests_passing = NULL,
    tests_skipped = NULL
  )
  
  expect_equal(res$n_no_function_tests, 0L)
  expect_equal(res$n_skipped_files, 0L)
  expect_equal(res$n_testfiles, 0L)
  
  # n_tested = 1 (fallback), total_functions = 1
  expected_total_cov <- 50
  expect_equal(res$total_cov, expected_total_cov)
  expect_equal(res$percent_cov_round, round(expected_total_cov, 2))
})

test_that("zero passing tests: fallback n_tested = 1 avoids divide-by-zero", {
  res <- compute_total_coverage(
    percent_cov = 75,
    functions_no_tests_df = data.frame(fn = character()),
    tests_passing = character(),  # empty
    tests_skipped = c("test-x.R")
  )
  
  expect_equal(res$n_testfiles, 0L)
  expect_equal(res$n_no_function_tests, 0L)
  expect_equal(res$n_skipped_files, 1L)
  
  n_untested <- 0 + 1
  n_tested <- 1L       # fallback
  total_functions <- n_tested + n_untested
  expected_total_cov <- 75
  
  expect_equal(res$total_cov, expected_total_cov)
})

test_that("percent_cov is clamped to [0, 100]", {
  # Above 100
  res_hi <- compute_total_coverage(
    percent_cov = 130,
    functions_no_tests_df = data.frame(fn = "a"),
    tests_passing = c("test-a.R", "test-b.R"),
    tests_skipped = NULL
  )
  
  n_untested <- 1
  n_tested <- 2
  total_functions <- n_tested + n_untested
  expected_total_cov_hi <- 100 
  expect_equal(res_hi$total_cov, expected_total_cov_hi)
  
  # Below 0
  res_lo <- compute_total_coverage(
    percent_cov = -20,
    functions_no_tests_df = NULL,
    tests_passing = c("t1.R"),
    tests_skipped = c("t_skip.R")
  )
  expect_equal(res_lo$total_cov, 0)
  expect_equal(res_lo$percent_cov_round, 0)
})

# --- Using mockery::stub to control internal dependencies ---

test_that("stub: simulate large number of functions_without_tests via nrow()", {
  fn <- compute_total_coverage
  mock_nrow <- function(x) 100L
  
  dummy_df <- data.frame(x = 1)
  
  mockery::stub(fn, "nrow", mock_nrow)
  
  res <- fn(
    percent_cov = 60,
    functions_no_tests_df = dummy_df,
    tests_passing = c("test-1.R", "test-2.R"),
    tests_skipped = character()
  )
  
  expect_equal(res$n_no_function_tests, 100L)
  
  n_untested <- 100
  n_tested <- 2
  total_functions <- n_tested + n_untested
  expected_total_cov <- 60 
  expect_equal(res$total_cov, expected_total_cov)
})

test_that("stub: simulate many skipped and passing files via length()", {
  fn <- compute_total_coverage
  
  mock_length <- function(x) {
    if (identical(x, "PASSING")) return(10L)
    if (identical(x, "SKIPPED")) return(7L)
    return(base::length(x))
  }
  
  mockery::stub(fn, "length", mock_length)
  
  res <- fn(
    percent_cov = 90,
    functions_no_tests_df = NULL,
    tests_passing = "PASSING",
    tests_skipped = "SKIPPED"
  )
  
  expect_equal(res$n_testfiles, 10L)
  expect_equal(res$n_skipped_files, 7L)
  n_untested <- 0 + 7
  n_tested <- 10
  total_functions <- n_tested + n_untested
  expected_total_cov <- 90
  expect_equal(res$total_cov, expected_total_cov)
})

test_that("stub: when tests_passing would be 0, fallback n_tested=1 is used", {
  fn <- compute_total_coverage
  
  mock_length_zero <- function(x) {
    if (identical(x, "PASSING")) return(0L)
    if (identical(x, "SKIPPED")) return(5L)
    return(base::length(x))
  }
  
  mockery::stub(fn, "length", mock_length_zero)
  
  res <- fn(
    percent_cov = 40,
    functions_no_tests_df = NULL,
    tests_passing = "PASSING",
    tests_skipped = "SKIPPED"
  )
  
  expect_equal(res$n_testfiles, 0L)
  expect_equal(res$n_skipped_files, 5L)
  
  n_untested <- 5
  n_tested <- 1L
  total_functions <- n_tested + n_untested
  expected_total_cov <- 40
  expect_equal(res$total_cov, expected_total_cov)
})

test_that("stub: rounding behavior can be validated separately", {
  fn <- compute_total_coverage
  
  mock_nrow <- function(x) 3L
  mock_length <- function(x) {
    if (identical(x, "PASSING")) return(2L)
    if (identical(x, "SKIPPED")) return(1L)
    return(base::length(x))
  }
  mockery::stub(fn, "nrow", mock_nrow)
  mockery::stub(fn, "length", mock_length)
  
  res <- fn(
    percent_cov = 83.3333,
    functions_no_tests_df = data.frame(x = 1),
    tests_passing = "PASSING",
    tests_skipped = "SKIPPED"
  )
  
  expected <- 83.3333 
  expect_equal(res$total_cov, expected)
  expect_equal(res$percent_cov_round, round(expected, 2))  # 0.28  expect_equal(res$percent_cov_round, round(expected, 2))  # 0.28
})  

#' Generate HTML Report for Package Test Assessment
#'
#' @description Generates an HTML report for the package test assessment results using rmarkdown.
#'
#' @param test_results List containing the results from get_package_coverage function.
#' @param output_dir (required) Character string path to an existing directory where the report will be saved. 
#'
#' @return Path to the generated HTML report.
#'
#' @examples
#' \dontrun{
#' test_results <- get_package_coverage()
#' # Always provide a directory; tempdir() is convenient in examples.
#' generate_test_report(test_results, output_dir = tempdir())
#' }
#' @importFrom rmarkdown render
#' @importFrom fs dir_exists
#' @importFrom fs path_abs
#' @export
generate_test_report <- function(test_results, output_dir = NULL) {
  
  # Check if rmarkdown is available
  if (!requireNamespace("rmarkdown", quietly = TRUE)) {
    message("Package 'rmarkdown' is required but not installed.")
    return(NULL)
  }
  
  # check input
  checkmate::assert_list(test_results, names = "named", any.missing = TRUE)
  is_multi <- isTRUE(test_results$multi_framework) &&
    !is.null(test_results$results) &&
    length(test_results$results) > 0L
  if (!is_multi) {
    checkmate::assert_list(test_results$res_cov, names = "named", any.missing = TRUE)
  }
  
  # Normalize (single unchanged; multi gets multi_framework + frameworks)
  test_results <- normalize_test_results(test_results)
  
  # output_dir is REQUIRED: must be a non-empty string, existing, and writable
  checkmate::assert_string(output_dir, min.chars = 1, na.ok = FALSE)
  output_dir <- fs::path_abs(output_dir)
  checkmate::assert_directory_exists(output_dir, access = "w")
  
  pkg_name <- test_results$pkg_name
  pkg_version <- test_results$pkg_ver
  date_time <- test_results$date_time
  multi_framework <- isTRUE(test_results$multi_framework)
  frameworks <- if (multi_framework) test_results$frameworks else character(0L)
  
  # Build report_env for unified template
  report_env <- new.env()
  report_env$pkg_name <- pkg_name
  report_env$pkg_version <- pkg_version
  report_env$multi_framework <- multi_framework
  report_env$frameworks <- frameworks
  
  test_pkg_summary_output <- generate_test_pkg_summary(test_results)
  report_env$test_pkg_summary_output <- test_pkg_summary_output
  
  if (!multi_framework) {
    # Single framework: current flow
    test_details_output <- generate_test_details(test_results)
    coverage_output <- generate_coverage_section(test_results, pkg_name)
    report_env$test_details_output <- test_details_output
    report_env$coverage_output <- coverage_output
    
    nstf_elements <- check_elements_simple(test_results, c("functions_no_tests", "tests_skipped", "tests_passing"))
    stf_elements <- check_elements_simple(test_results, c("long_summary", "test_skip"))
    nstf_none <- length(nstf_elements$present) == 0
    stf_none <- length(stf_elements$present) == 0
    
    nstf_df_list <- list()
    stf_df_list <- list()
    if ("functions_no_tests" %in% nstf_elements$present) {
      nstf_df_list$functions_no_tests <- test_results$functions_no_tests
      short_is_needed <- detect_path_shortening(nstf_df_list$functions_no_tests)
      if (short_is_needed) {
        nstf_df_list$functions_no_tests <- create_function_locations_df(nstf_df_list$functions_no_tests)
      }
    }
    if ("tests_skipped" %in% nstf_elements$present && length(test_results$tests_skipped) > 0) {
      nstf_df_list$tests_skipped_df <- convert_vectors_to_df(test_results, "tests_skipped")[[1]]
      short_is_needed <- detect_path_shortening(nstf_df_list$tests_skipped_df, "file")
      if (short_is_needed) {
        nstf_df_list$tests_skipped_df$file <- sapply(nstf_df_list$tests_skipped_df$file, extract_short_path)
      }
    }
    if ("tests_passing" %in% nstf_elements$present) {
      nstf_df_list$tests_passing_df <- convert_vectors_to_df(test_results, "tests_passing")[[1]]
      nstf_df_list$tests_passing_df$file <- sapply(nstf_df_list$tests_passing_df$file, extract_short_path)
    }
    if ("long_summary" %in% stf_elements$present) {
      stf_df_list$long_summary_df <- create_long_summary_details(test_results$long_summary)
    }
    if ("test_skip" %in% stf_elements$present) {
      stf_df_list$tests_skip_df <- create_test_skip_details(test_results$test_skip)
    }
    
    report_env$functions_no_tests <- nstf_df_list$functions_no_tests
    report_env$tests_skipped_df <- nstf_df_list$tests_skipped_df
    report_env$tests_passing_df <- nstf_df_list$tests_passing_df
    report_env$tests_skip_df <- stf_df_list$tests_skip_df
    report_env$long_summary_df <- stf_df_list$long_summary_df
    report_env$has_stf <- !stf_none
    report_env$has_nstf <- !nstf_none
  } else {
    # Multi-framework: build framework_results
    res_list <- test_results$results
    coverage_sections <- generate_coverage_sections(test_results, pkg_name)
    nstf_elements <- check_elements_simple(test_results, c("functions_no_tests", "tests_skipped", "tests_passing"), nested = TRUE)
    stf_elements <- check_elements_simple(test_results, c("long_summary", "test_skip"), nested = TRUE)
    has_stf <- length(stf_elements$present) > 0
    has_nstf <- length(nstf_elements$present) > 0
    report_env$has_stf <- has_stf
    report_env$has_nstf <- has_nstf
    
    framework_results <- setNames(lapply(frameworks, function(fw) {
      r <- res_list[[fw]]
      meta <- c(test_results, r)
      meta$date_time <- test_results$date_time
      meta$executor <- test_results$executor
      meta$sysname <- test_results$sysname
      meta$release <- test_results$release
      meta$machine <- test_results$machine
      meta$r_version <- test_results$r_version
      meta$pkg_name <- test_results$pkg_name
      meta$pkg_ver <- test_results$pkg_ver
      meta$total_cov <- r$total_cov
      
      out <- list(
        coverage_output = coverage_sections[[fw]],
        test_details_output = generate_test_details(meta, framework_id = fw)
      )
      if (has_nstf) {
        if (!is.null(r$functions_no_tests) && is.data.frame(r$functions_no_tests) && nrow(r$functions_no_tests) > 0) {
          fn <- r$functions_no_tests
          short_is_needed <- tryCatch(detect_path_shortening(fn), error = function(e) FALSE)
          if (short_is_needed) {
            fn <- tryCatch(create_function_locations_df(fn), error = function(e) r$functions_no_tests)
          }
          out$functions_no_tests <- fn
        }
        if (!is.null(r$tests_skipped) && length(r$tests_skipped) > 0) {
          df <- convert_vectors_to_df(list(tests_skipped = r$tests_skipped), "tests_skipped")[[1]]
          short_is_needed <- detect_path_shortening(df, "file")
          if (short_is_needed) df$file <- sapply(df$file, extract_short_path)
          out$tests_skipped_df <- df
        }
        if (!is.null(r$tests_passing) && length(r$tests_passing) > 0) {
          df <- convert_vectors_to_df(list(tests_passing = r$tests_passing), "tests_passing")[[1]]
          df$file <- sapply(df$file, extract_short_path)
          out$tests_passing_df <- df
        }
      }
      if (has_stf) {
        if (!is.null(r$long_summary) && is.data.frame(r$long_summary) && nrow(r$long_summary) > 0) {
          out$long_summary_df <- create_long_summary_details(r$long_summary)
        }
        if (!is.null(r$test_skip) && is.data.frame(r$test_skip) && nrow(r$test_skip) > 0) {
          out$tests_skip_df <- create_test_skip_details(r$test_skip)
        }
      }
      out
    }), frameworks)
    report_env$framework_results <- framework_results
  }
  
  template_path <- system.file(
    "report_templates",
    "test_report_template.Rmd",
    package = "test.assessr")
  
  
  # Compose output file path within the provided directory --------------------
  output_file <- fs::path(
    output_dir,
    paste0("test_report_", pkg_name, "_", pkg_version, ".html")
  )
  
  # Render if allowed
  if (identical(Sys.getenv("NOT_CRAN"), "true") || interactive()) {
    rmarkdown::render(
      input       = template_path,
      output_file = output_file,
      envir       = report_env,
      quiet       = TRUE
    )
  } else {
    message("Rendering skipped on CRAN or non-interactive environment.")
    return(NULL)
  }
  
  return(output_file)
  
}

#' Normalize test results for single or multi-framework structure
#'
#' @description For single-framework results, passes through unchanged. For
#'   multi-framework, sets \code{multi_framework} and \code{frameworks} while
#'   keeping \code{results} as a named list of per-framework blocks (no flattening).
#'
#' @param test_results List from \code{get_package_coverage} or \code{run_covr_modes}.
#'
#' @return A list with \code{multi_framework} (logical), \code{frameworks}
#'   (character, multi only), and \code{results} preserved for multi-framework.
#'
#' @keywords internal
normalize_test_results <- function(test_results) {
  is_multi <- isTRUE(test_results$multi_framework) &&
    !is.null(test_results$results) &&
    length(test_results$results) > 0L
  
  if (!is_multi) {
    return(test_results)
  }
  
  test_results$multi_framework <- TRUE
  test_results$frameworks <- names(test_results$results)
  test_results
}

#' Helper function to replace NULL with "N/A"
#'
#' @param x - input value
#'
#' @keywords internal
handle_null <- function(x) {
  if (is.null(x) || length(x) == 0) {
    return("N/A")
  } else if (is.logical(x)) {
    return(paste(as.character(x), collapse = ", "))
  } else if (is.numeric(x)) {
    return(paste(as.character(x), collapse = " "))
  } else {
    return(x)
  }
}


#' Convert number to abbreviation
#'
#' @param value 
#'
#' @keywords internal
convert_number_to_abbreviation <- function(value) {
  result <- NA
  
  if (!is.na(value) && is.numeric(value)) {
    abs_value <- abs(value)
    
    if (abs_value >= 1e6) {
      result <- paste0(round(value / 1e6, 1), "M")
    } else if (abs_value >= 1e3) {
      result <- paste0(round(value / 1e3, 1), "K")
    } else {
      result <- as.character(value)
    }
  }
  
  return(result)
}

#' Convert number to percent
#'
#' @param value 
#'
#' @keywords internal
convert_number_to_percent <- function(value) {
  
  result <- NA
  
  if (!is.na(value)) {
    # Try to coerce to numeric if it's a character
    if (is.character(value)) {
      value <- suppressWarnings(as.numeric(value))
    }
    
    # Proceed if it's now numeric and not NA
    if (!is.na(value) && is.numeric(value)) {
      
      percent_value <- round(
        ifelse(value <= 1, value * 100, value),
        1
      )
      
      result <- paste0(percent_value, "%")
    }
  }
  
  return(result)
}

#' Helper to conditionally apply handle_null or abbreviation
#'
#' @param x - value
#'
#' @keywords internal
safe_value <- function(x) {
  if (is.null(x)) {
    return(handle_null(x))
  } else if (is.numeric(x)) {
    return(convert_number_to_abbreviation(x))
  } else {
    return(handle_null(x))
  }
}



#' Generate Test Summary
#'
#' @description Generates the Risk Summary table for the HTML report.
#'
#' @param assessment_results - input data
#'
#' @keywords internal
generate_test_pkg_summary <- function(assessment_results) {
  
  test_pkg_summary_table <- data.frame(
    Metric = c('Package', 'Version'), 
    Value = c(
      assessment_results$pkg_name,
      assessment_results$pkg_ver
    )
  )
  
  return(test_pkg_summary_table)
}

#' Generate Test Details
#'
#' @description Generates the Test Details table for the HTML report.
#'
#' @param assessment_results - input data
#' @param framework_id Optional character; when provided (multi-framework), the
#'   "Testing Framework" row shows this framework name.
#'
#' @keywords internal
generate_test_details <- function(assessment_results, framework_id = NULL) {
  
  total_cov <- assessment_results$total_cov
  if (!is.numeric(total_cov)) {
    fallback <- tryCatch(
      as.numeric(assessment_results$results[["covr"]]),
      error = function(e) NA_real_
    )
    total_cov <- if (!is.na(fallback)) fallback else NA_real_
  }
  total_cov <- round(total_cov, 2)
  
  covr <- handle_null(total_cov)
  covr <- convert_number_to_percent(covr)
  
  fw_label <- if (!is.null(framework_id)) {
    handle_null(framework_id)
  } else {
    handle_null(assessment_results$test_framework_type)
  }
  
  test_details_table <- data.frame(
    Metric = c(
      'Test Coverage Score', 'Date Time', 'Executor',
      'OS Name', 'OS Release', 'OS Machine',
      'R version', 'Testing Framework'
    ),
    Value = c(
      covr,
      handle_null(assessment_results$date_time),
      handle_null(assessment_results$executor),
      handle_null(assessment_results$sysname),
      handle_null(assessment_results$release),
      handle_null(assessment_results$machine),
      handle_null(assessment_results$r_version),
      fw_label
    )
  )
  return(test_details_table)
}

#' Generate file coverage df
#'
#' @description Generates file coverage df when errors list created.
#' 
#' @param file_names - file names
#' @param file_coverage - test coverage for files
#' @param errors - test coverage errors
#' @param notes - test coverage notes
#' 
#' @keywords internal
create_file_coverage_df <- function(file_names, file_coverage, errors, notes) {
  
  # Convert errors to a character string if it is a complex structure
  if (is.list(errors)) {
    errors <- sapply(errors, function(x) {
      if (is.null(x)) {
        return("N/A")
      } else if (is.data.frame(x)) {
        return(paste(utils::capture.output(print(x)), collapse = "; "))
      } else if (is.list(x)) {
        return(paste(unlist(x), collapse = "; "))
      } else {
        return(as.character(x))
      }
    })
  }
  
  # Ensure the number of rows match
  max_len <- max(length(file_names), length(file_coverage), length(notes))
  
  # Extend lists to match the maximum length
  file_names <- c(file_names, rep("", max_len - length(file_names)))
  
  # extract short file name from long path
  file_names <- vapply(file_names, extract_short_path, FUN.VALUE = character(1))
  
  file_coverage <- c(file_coverage, rep(NA, max_len - length(file_coverage)))
  notes <- c(notes, rep("", max_len - length(notes)))
  errors <- rep(paste(errors, collapse = "; "), max_len)
  
  # Create the data frame
  file_coverage_df <- data.frame(
    File = file_names,
    Coverage = file_coverage,
    Errors = errors,
    Notes = notes,
    stringsAsFactors = FALSE
  )
  
  return(file_coverage_df)
}

#' Generate Coverage Section (single framework)
#'
#' @description Generates the Coverage section for the HTML report from a single
#'   framework's \code{res_cov}.
#' 
#' @param assessment_results - input data with \code{res_cov}
#' @param pkg_name - name of the package
#' 
#' @keywords internal
generate_coverage_section <- function(assessment_results, pkg_name) {
  
  total_coverage <- assessment_results$res_cov$coverage$totalcoverage
  file_coverage <- assessment_results$res_cov$coverage$filecoverage
  
  # Extract file names from the attributes
  file_names <- attr(file_coverage, "dimnames")[[1]]
  
  # extract short file name from long path
  file_names <- vapply(file_names, extract_short_path, FUN.VALUE = character(1))
  
  # Handle errors and notes
  errors <- assessment_results$res_cov$errors
  if (all(is.na(errors))) {
    errors <- "No test coverage errors"
  }
  
  notes <- assessment_results$res_cov$notes
  if (all(is.na(notes))) {
    notes <- "No test coverage notes"
  }
  
  # Create a data frame for file coverage
  if (is.list(errors) && all(c("message", "srcref", "status", "stdout", "stderr", "parent_trace", "call", "procsrcref", "parent") %in% names(errors))) {
    # Create the file coverage data frame
    file_coverage_df <- create_file_coverage_df(file_names, file_coverage, errors, notes)
  } else {
    # Handle the case where errors does not have the expected structure
    
    if (is.null(file_names) || is.null(file_coverage)) {
      file_coverage_df <- data.frame(
        Function = NA_character_,
        Coverage = NA_real_,
        Errors = NA,
        Notes = NA,
        stringsAsFactors = FALSE
      )
    } else {
      file_coverage_df <- data.frame(
        Function = file_names,
        Coverage = file_coverage,
        Errors = errors,
        Notes = notes,
        stringsAsFactors = FALSE
      )
    }
  }
  
  return(file_coverage_df)
}

#' Generate Coverage Sections (multi-framework)
#'
#' @description Returns a named list of coverage data frames, one per framework.
#' 
#' @param test_results Multi-framework list with \code{results} (named list of
#'   per-framework blocks, each with \code{res_cov}).
#' @param pkg_name - name of the package
#' 
#' @return Named list of data frames (Function, Coverage, Errors, Notes).
#' @keywords internal
generate_coverage_sections <- function(test_results, pkg_name) {
  frameworks <- names(test_results$results)
  res_list <- test_results$results
  out <- setNames(
    lapply(frameworks, function(fw) {
      r <- res_list[[fw]]
      if (is.null(r$res_cov) || is.null(r$res_cov$coverage)) {
        return(data.frame(
          Function = NA_character_,
          Coverage = NA_real_,
          Errors = "No coverage data",
          Notes = "No coverage notes",
          stringsAsFactors = FALSE
        ))
      }
      fake_assessment <- list(res_cov = r$res_cov)
      generate_coverage_section(fake_assessment, pkg_name)
    }),
    frameworks
  )
  out
}

#' Check Which Elements Exist in a List
#'
#' This internal function checks whether specified elements exist in a given list
#' and returns two vectors: elements that exist and elements that do not. When
#' \code{nested = TRUE} and \code{x} has \code{results} (multi-framework), checks
#' inside each \code{x$results$<framework>}.
#'
#' @param x A list object to check (e.g., `test_results`).
#' @param elems A character vector of element names to check.
#' @param nested If TRUE and \code{x} has \code{results}, check inside each
#'   framework block. Returns \code{present} (flattened unique) and
#'   \code{by_framework} (named list: framework -> present elements).
#'
#' @return A list with:
#' \describe{
#'   \item{present}{Character vector of elements that exist.}
#'   \item{missing}{Character vector of elements that do not exist (top-level only).}
#'   \item{by_framework}{Named list (framework -> present elements), when nested.}
#' }
#'
#' @keywords internal
check_elements_simple <- function(x, elems, nested = FALSE) {
  if (!nested || is.null(x$results)) {
    exists_flags <- elems %in% names(x)
    return(list(
      present = elems[exists_flags],
      missing = elems[!exists_flags]
    ))
  }
  frameworks <- names(x$results)
  by_fw <- setNames(
    lapply(frameworks, function(fw) {
      r <- x$results[[fw]]
      elems[elems %in% names(r)]
    }),
    frameworks
  )
  present <- unique(unlist(by_fw, use.names = FALSE))
  missing <- setdiff(elems, present)
  list(present = present, missing = missing, by_framework = by_fw)
}


' Convert Character Vectors to Data Frames
#'
#' This internal function takes named character vectors (e.g., `tests_skipped`, `tests_passing`)
#' from a list and converts them into data frames for easier reporting or processing.
#'
#' @param x A list containing character vectors (e.g., `test_results`).
#' @param elems A character vector of element names to convert (must exist in `x`).
#'
#' @return A named list of data frames, where each data frame has one column:
#' \describe{
#'   \item{file}{The file names or paths from the original character vector.}
#' }
#'
#' @keywords internal
convert_vectors_to_df <- function(x, elems) {
  # Filter only elements that exist
  elems <- elems[elems %in% names(x)]
  
  # Convert each to a data frame
  result <- lapply(elems, function(nm) {
    data.frame(file = x[[nm]], stringsAsFactors = FALSE)
  })
  
  names(result) <- elems
  return(result)
}

#' Generate Long Summary Details
#'
#' @description Generates the  Long Summary table for the HTML report.
#'
#' @param df - input data
#'
#' @keywords internal
create_long_summary_details <- function(df) {
  
  # validate input
  required <- c("file", "code_script", "status")
  if (!all(required %in% names(df))) {
    stop("Input df must contain columns: file, code_script, status")
  }
  
  file_orig <- as.character(df$file)
  func <- as.character(df$code_script)
  status_vec <- as.character(df$status)
  
  # extract portion before '#' (if any)
  before_hash <- sub("#.*$", "", file_orig)
  
  test <- before_hash  
  
  # extract the part after '#' (if any); then take up to first '_' (if any)
  after_hash <- ifelse(grepl("#", file_orig), sub(".*#", "", file_orig), NA_character_)
  start_line <- ifelse(is.na(after_hash) | after_hash == "",
                       NA_character_,
                       sub("_.*$", "", after_hash))
  
  # assemble output (use 'function' as requested; if you prefer a different name change here)
  out <- data.frame(
    `R function` = func,
    test = test,
    start_line = start_line,
    status = status_vec,
    stringsAsFactors = FALSE
  )
  
  # Filter: keep only rows with status == "PASS" (case/whitespace tolerant)
  keep <- toupper(trimws(out$status)) == "PASS"
  out <- out[keep, , drop = FALSE]
  
  rownames(out) <- NULL
  
  return(out)
}


#' Generate Test Skip Details
#'
#' @description Generates the Test Skip table for the HTML report.
#'
#' @param df - input data
#'
#' @keywords internal
create_test_skip_details <- function(df) {
  # validate input
  required <- c("file", "code_script", "status", "expectation", "line1")
  if (!all(required %in% names(df))) {
    stop("Input df must contain columns: file, code_script, status, expectation, line1")
  }
  
  file_orig    <- as.character(df$file)
  func         <- as.character(df$code_script)
  status_vec   <- as.character(df$status)
  expect_vec   <- as.character(df$expectation)
  line1_vec    <- as.character(df$line1)
  
  # test: portion of `file` before '#' (if present); keep digits before .R (e.g. test-myscript2.R)
  test <- sub("#.*$", "", file_orig)
  
  # start_line: rename of line1 (keeps as character)
  start_line <- line1_vec
  
  out <- data.frame(
    `Rfunction`    = func,
    test        = test,
    status      = status_vec,
    expectation = expect_vec,
    start_line  = start_line,
    stringsAsFactors = FALSE
  )
  
  rownames(out) <- NULL
  return(out)
}


#' Normalize a function-locations data frame
#'
#' Accept a data frame containing function locations and return a normalized
#' data frame with columns `file`, `function_name`, `start_line`, and `end_line`.
#' If the `file` column is missing or entirely `NA`, file names may be derived
#' from `attr(file_coverage, "dimnames")[[1]]`. File paths are shortened with
#' `basename()` by default; a custom shortener can be provided.
#'
#' @param df A data frame with at least columns `function_name`, `start_line`,
#'   `end_line`. Column `file` is recommended, but may be omitted if
#'   `file_coverage` is supplied.
#' @param file_coverage Optional coverage-like object whose first `dimnames`
#'   entry contains file paths; used only if `df$file` is missing or entirely `NA`.
#' @param extract_short_path A function to convert long paths to short names.
#'   Defaults to `basename`. Set to `NULL` to skip shortening.
#'
#' @return A `data.frame` with columns:
#'   `file` (character), `function_name` (character),
#'   `start_line` (integer), `end_line` (integer).
#'
#' @keywords internal
#' @noRd
create_function_locations_df <- function(
    df,
    file_coverage = NULL,
    extract_short_path = base::basename
) {
  if (!is.data.frame(df)) {
    stop("`df` must be a data.frame (or tibble).", call. = FALSE)
  }
  
  # --- Resolve required columns (allow a few common variants) ---
  cn <- names(df)
  tolower_cn <- tolower(cn)
  
  find_col <- function(candidates_regex) {
    idx <- grep(paste0("^(", candidates_regex, ")$"), tolower_cn, perl = TRUE)
    if (length(idx)) cn[idx[1]] else NA_character_
  }
  
  col_function_name <- find_col("function_name|function|fn|func")
  col_file          <- find_col("file|filepath|path")
  col_start_line    <- find_col("start_line|start|startline")
  col_end_line      <- find_col("end_line|end|endline")
  
  missing_core <- setdiff(
    c("function_name", "start_line", "end_line"),
    c(if (!is.na(col_function_name)) "function_name",
      if (!is.na(col_start_line))   "start_line",
      if (!is.na(col_end_line))     "end_line")
  )
  if (length(missing_core)) {
    stop(
      "Missing required column(s): ",
      paste(missing_core, collapse = ", "),
      ". Expected columns include: function_name, start_line, end_line.",
      call. = FALSE
    )
  }
  
  function_name <- df[[col_function_name]]
  start_line    <- df[[col_start_line]]
  end_line      <- df[[col_end_line]]
  file          <- if (!is.na(col_file)) df[[col_file]] else NULL
  
  # --- If file is missing or entirely NA, try to derive from coverage dimnames ---
  needs_file <- is.null(file) || all(is.na(file))
  if (needs_file) {
    dn <- tryCatch(attr(file_coverage, "dimnames"), error = function(e) NULL)
    if (!is.null(dn) && length(dn) >= 1L && !is.null(dn[[1]])) {
      fc_files <- as.character(dn[[1]])
      n <- nrow(df)
      if (length(fc_files) == n) {
        file <- fc_files
      } else if (length(fc_files) == 1L) {
        file <- rep(fc_files, n)
      } else {
        # pad/truncate to n rows
        file <- rep(NA_character_, n)
        file[seq_len(min(n, length(fc_files)))] <- fc_files[seq_len(min(n, length(fc_files)))]
      }
    } else {
      file <- rep(NA_character_, nrow(df))
    }
  }
  
  
  # --- Coerce types ---
  function_name <- as.character(function_name)
  file          <- as.character(file)
  start_line    <- as.integer(start_line)
  end_line      <- as.integer(end_line)
  
  # --- Shorten paths, if requested ---
  if (!is.null(extract_short_path)) {
    file <- vapply(
      file,
      function(p) {
        if (is.na(p) || !nzchar(p)) return(p)
        extract_short_path(p)
      },
      FUN.VALUE = character(1)
    )
  }
  
  # --- HARDEN AGAINST STRAY NAMES/ROW.NAMES ON VECTORS ---
  # 1) Drop any names on the vectors (these can be used as row.names by data.frame)
  file          <- unname(file)
  function_name <- unname(function_name)
  start_line    <- unname(start_line)
  end_line      <- unname(end_line)
  
  # 2) Also remove any stray 'row.names' attribute on the vectors (ultra defensive)
  attr(file, "row.names")          <- NULL
  attr(function_name, "row.names") <- NULL
  attr(start_line, "row.names")    <- NULL
  attr(end_line, "row.names")      <- NULL
  
  # --- Build normalized output, then force safe row names ---
  n <- length(function_name)
  
  out <- data.frame(
    file = file,
    function_name = function_name,
    start_line = start_line,
    end_line = end_line,
    stringsAsFactors = FALSE
  )
  
  # Force clean numeric row names (prevents "row names contain missing values")
  row.names(out) <- seq_len(n)
  
  return(out)
}  


#' Detect if path shortening is advisable (boolean, df input)
#'
#' Given a data frame and a column that contains file paths, decide whether
#' shortening those paths is advisable. Returns a single TRUE/FALSE.
#'
#' @param df A data frame that contains a path column.
#' @param file_col Character scalar **or vector** of candidate column names that hold
#'   paths. Defaults to `c("file", "source_file")`. The first existing column
#'   (case-insensitive) will be used.
#' @param max_len_threshold Consider paths "long" if any has length > this (default: 60).
#' @param gap_threshold Consider shortening if `max(nchar(full)) - max(nchar(basename))`
#'   is at least this many characters (default: 10).
#' @param min_common_prefix Consider a "long" common directory prefix if >= this many
#'   characters (default: 20).
#' @param min_dir_share Minimum fraction of entries that must contain a directory
#'   separator to treat them as true paths (default: 0.2).
#'
#' @return TRUE if shortening is advisable; otherwise FALSE.
#'
#' @examples
#' \dontrun{
#' # Explicit column name "file"
#' short_is_needed <- detect_path_shortening(nstf_df_list$functions_no_tests, "file")
#'
#' # Explicit column name "source_file"
#' short_is_needed <- detect_path_shortening(nstf_df_list$functions_no_tests, "source_file")
#'
#' # Let the function choose the first that exists among c("file","source_file")
#' short_is_needed <- detect_path_shortening(nstf_df_list$functions_no_tests)
#' }
#' @keywords internal
#' @noRd
detect_path_shortening <- function(
    df,
    file_col = c("file", "source_file"),
    max_len_threshold = 60L,
    gap_threshold = 10L,
    min_common_prefix = 20L,
    min_dir_share = 0.2
) {
  if (!is.data.frame(df)) {
    stop("`df` must be a data.frame.", call. = FALSE)
  }
  
  # Resolve the path column: accept a single name or a vector of candidates
  resolve_col <- function(df, candidates) {
    if (length(candidates) == 1L) {
      if (!(candidates %in% names(df))) {
        # try case-insensitive
        idx <- which(tolower(names(df)) == tolower(candidates))
        if (length(idx)) return(names(df)[idx[1]])
        return(NA_character_)
      }
      return(candidates)
    } else {
      # pick the first candidate that exists (case-insensitive)
      df_lower <- tolower(names(df))
      cand_lower <- tolower(candidates)
      for (cl in cand_lower) {
        hit <- which(df_lower == cl)
        if (length(hit)) return(names(df)[hit[1]])
      }
      return(NA_character_)
    }
  }
  
  col <- resolve_col(df, file_col)
  if (is.na(col)) {
    stop(
      "Could not find any of the specified path columns in `df`. Looked for: ",
      paste(file_col, collapse = ", "),
      call. = FALSE
    )
  }
  
  paths <- df[[col]]
  xp <- as.character(paths)
  # normalize to '/' for detection; keep logic simple
  xp_norm <- gsub("\\\\", "/", xp)
  xpn <- xp_norm[!is.na(xp_norm) & nzchar(xp_norm)]
  
  # If no usable paths, no shortening needed
  if (!length(xpn)) return(FALSE)
  
  # Share with directory separators
  dir_share <- mean(grepl("/", xpn, fixed = TRUE))
  
  # If hardly any directory separators, shortening not helpful
  if (!(dir_share >= min_dir_share)) return(FALSE)
  
  # Length heuristics
  base_names <- basename(xpn)
  max_full <- max(nchar(xpn), na.rm = TRUE)
  max_base <- max(nchar(base_names), na.rm = TRUE)
  big_gap  <- (max_full - max_base) >= gap_threshold
  too_long <- max_full > max_len_threshold
  
  # Longest common directory prefix (approximate, trimmed to dir boundary)
  common_prefix_raw <- {
    if (length(xpn) == 1L) xpn else {
      split_chars <- strsplit(xpn, "", fixed = TRUE)
      min_len <- min(lengths(split_chars))
      lcp_len <- 0L
      for (i in seq_len(min_len)) {
        chars_i <- vapply(split_chars, function(v) v[i], "", USE.NAMES = FALSE)
        if (length(unique(chars_i)) == 1L) lcp_len <- i else break
      }
      substr(xpn[1], 1L, lcp_len)
    }
  }
  cut_at <- regexpr(".*/", common_prefix_raw)
  common_prefix <- if (cut_at[1] > 0) substr(common_prefix_raw, 1, attr(cut_at, "match.length")) else ""
  long_common_prefix <- nchar(common_prefix) >= min_common_prefix
  
  # Final decision
  is_needed <- (too_long || big_gap || long_common_prefix)
  
  return(is_needed)
}

toy_assessment_results <- list(
  pkg_name = "here",
  pkg_ver = "1.0.1",
  date_time = "2026-01-07 15:48:31.338571",
  executor = "u1004798",
  sysname = "Linux",
  version = "#1 SMP PREEMPT_DYNAMIC Tue Sep  9 12:21:26 UTC 2025",
  release = "6.1.150-174.273.amzn2023.x86_64",
  machine = "x86_64",
  r_version = "4.4.1",
  total_cov = 0.987,
  res_cov = list(
    name = "here-1.0.1",
    coverage = list(
      filecoverage = structure(
        c(100, 100, 100, 95.8, 100, 97.5),
        dim = c(6L),
        dimnames = list(c("R/aaa.R", "R/dr_here.R", "R/here.R", "R/i_am.R", "R/utils.R", "R/zzz.R"))
      ),
      totalcoverage = 98.7
    ),
    errors = NA,
    notes = NA
  )
)


test_that("handle_null returns 'N/A' for NULL", {
  expect_equal(handle_null(NULL), "N/A")
})

test_that("normalize_test_results returns single-framework input unchanged", {
  tr <- toy_assessment_results
  out <- test.assessr:::normalize_test_results(tr)
  expect_identical(out, tr)
})

test_that("normalize_test_results preserves multi-framework structure and sets flags", {
  tr_multi <- list(
    pkg_name = "bookdown",
    pkg_ver = "0.43",
    date_time = "2026-03-19 09:29:21",
    executor = "u1004798",
    sysname = "Linux",
    r_version = "4.4.1",
    multi_framework = TRUE,
    frameworks = c("testthat", "testit"),
    total_cov = 0.225,
    results = list(
      testthat = list(
        total_cov = 0.45,
        res_cov = list(
          name = "bookdown-0.43",
          coverage = list(
            filecoverage = structure(c(84.3, 36.9), dim = 2L, dimnames = list(c("R/addins.R", "R/bs4_book.R"))),
            totalcoverage = 45
          ),
          errors = NA,
          notes = NA
        ),
        test_framework_type = "standard testing framework",
        framework_id = "testthat"
      ),
      testit = list(
        total_cov = 0,
        res_cov = list(
          name = "bookdown",
          coverage = list(
            filecoverage = structure(0, dim = 1L, dimnames = list("R/foo.R")),
            totalcoverage = 0
          ),
          errors = NA,
          notes = NA
        ),
        functions_no_tests = data.frame(source_file = "addins.R", test_file = NA_character_, comment = "No tests", stringsAsFactors = FALSE),
        tests_skipped = c("test-html.R", "test-latex.R"),
        tests_passing = character(0),
        test_framework_type = "non-standard testing framework",
        framework_id = "testit"
      )
    )
  )
  out <- test.assessr:::normalize_test_results(tr_multi)
  expect_true(out$multi_framework)
  expect_equal(out$frameworks, c("testthat", "testit"))
  expect_identical(out$results, tr_multi$results)
})

test_that("handle_null returns original value for non-NULL", {
  expect_equal(handle_null(123), "123")
  expect_equal(handle_null("text"), "text")
})

test_that("convert_number_to_abbreviation handles millions", {
  expect_equal(convert_number_to_abbreviation(1500000), "1.5M")
  expect_equal(convert_number_to_abbreviation(-2500000), "-2.5M")
})

test_that("convert_number_to_abbreviation handles thousands", {
  expect_equal(convert_number_to_abbreviation(1200), "1.2K")
  expect_equal(convert_number_to_abbreviation(-9999), "-10K")
})

test_that("convert_number_to_abbreviation handles small numbers", {
  expect_equal(convert_number_to_abbreviation(999), "999")
  expect_equal(convert_number_to_abbreviation(0), "0")
})

test_that("convert_number_to_abbreviation handles NA and non-numeric", {
  expect_true(is.na(convert_number_to_abbreviation(NA)))
  expect_true(is.na(convert_number_to_abbreviation("text")))
})

test_that("safe_value returns 'N/A' for NULL", {
  expect_equal(safe_value(NULL), "N/A")
})

test_that("safe_value abbreviates numeric values", {
  expect_equal(safe_value(1000), "1K")
  expect_equal(safe_value(1000000), "1M")
})

test_that("safe_value returns character for small numbers", {
  expect_equal(safe_value(42), "42")
})

# Define the toy assessment_results object
toy_assessment_results_1_1 <- list(
  pkg_name = "test.package.0001",
  pkg_version = "1.0.0"
)

# Test the generate_test_pkg_summary function
test_that("generate_risk_summary works correctly", {
  result <- generate_test_pkg_summary(toy_assessment_results_1_1)
  
  expect_equal(result$Metric[1], "Package")
  expect_equal(result$Value[1], "test.package.0001")
  expect_equal(result$Metric[2], "Version")
  expect_equal(result$Value[2], "1.0.0")
})

# Define the toy assessment_results object (must include total_cov and r_version for generate_test_details)
toy_assessment_results_1_2 <- list(
  check = 1L,
  covr = 1L,
  total_cov = 0.85,
  date_time = "2025-01-13 17:01:49",
  executor = "test.executor",
  sysname = "Linux",
  release = "5.4.0-42-generic",
  machine = "x86_64",
  r_version = "4.0.2",
  check_list = list(res_check = list(rversion = "4.0.2")),
  test_framework_type = "standard testing framework"
)

# Define the expected output (generate_test_details uses total_cov, r_version, etc.)
expected_test_details <- list(
  Metric = c(
    'Test Coverage Score', 'Date Time', 'Executor',
    'OS Name', 'OS Release', 'OS Machine',
    'R version', 'Testing Framework'
  ),
  Value = c(
    "85%",  # total_cov 0.85 -> convert_number_to_percent
    toy_assessment_results_1_2$date_time,
    toy_assessment_results_1_2$executor,
    toy_assessment_results_1_2$sysname,
    toy_assessment_results_1_2$release,
    toy_assessment_results_1_2$machine,
    toy_assessment_results_1_2$r_version,
    toy_assessment_results_1_2$test_framework_type
  )
)

# Test the generate_test_details function
test_that("generate_test_details works correctly", {
  result <- generate_test_details(toy_assessment_results_1_2)
  
  expect_equal(result$Metric, expected_test_details$Metric)
  expect_equal(result$Value, expected_test_details$Value)
})



test_that("generate_coverage_section calls create_file_coverage_df when errors have expected structure", {
  file_names <- c("pkg/R/file1.R", "pkg/R/file2.R")
  file_coverage <- matrix(c(85, 90), ncol = 1,
                          dimnames = list(file_names, NULL))
  
  errors <- list(
    message = "in callr subprocess.",
    srcref = NULL,
    status = 0,
    stdout = "",
    stderr = "",
    parent_trace = NULL,
    call = NULL,
    procsrcref = NULL,
    parent = NULL
  )
  
  notes <- c("Note1", "Note2")
  
  assessment_results <- list(
    res_cov = list(
      coverage = list(totalcoverage = mean(file_coverage), filecoverage = file_coverage),
      errors = errors,
      notes = notes
    )
  )
  
  mockery::stub(generate_coverage_section, "extract_short_path", function(x) basename(x))
  
  captured_args <- NULL
  mockery::stub(generate_coverage_section, "create_file_coverage_df", function(fn, fc, err, nt) {
    captured_args <<- list(file_names = fn, file_coverage = fc, errors = err, notes = nt)
    data.frame(Function = fn, Coverage = fc, Errors = paste0(err$message, "; ", err$status),
               Notes = nt, stringsAsFactors = FALSE, row.names = fn)
  })
  
  result <- generate_coverage_section(assessment_results, pkg_name = "mockpkg")
  
  expect_identical(rownames(result), c("file1.R", "file2.R"))
  expect_true(!is.null(captured_args))
  expect_identical(unname(captured_args$file_names), c("file1.R", "file2.R"))
  expect_equal(as.vector(captured_args$file_coverage), c(85, 90))
})

test_that("generate_coverage_section handles NA errors and notes", {
  file_names <- c("R/file1.R", "R/file2.R")
  file_coverage <- matrix(c(75, 82), ncol = 1,
                          dimnames = list(file_names, NULL))
  
  errors <- NA
  notes <- NA
  
  assessment_results <- list(
    res_cov = list(
      coverage = list(totalcoverage = mean(file_coverage), filecoverage = file_coverage),
      errors = errors,
      notes = notes
    )
  )
  
  mockery::stub(generate_coverage_section, "extract_short_path", function(x) basename(x))
  
  result <- generate_coverage_section(assessment_results, pkg_name = "mockpkg")
  
  expect_identical(result$Function, c("file1.R", "file2.R"))
  expect_equal(result$Coverage, c(75, 82))
  expect_true(all(!is.na(result$Errors)))
  expect_true(all(!is.na(result$Notes)))
})

test_that("generate_test_report errors when directory does not exist", {
  # Skip if rmarkdown not installed because the function checks it first
  testthat::skip_if_not_installed("rmarkdown")
  
  # Construct a path that does not exist
  output_dir <- file.path(tempdir(), paste0("mock_output_dir_", as.integer(Sys.time())))
  testthat::expect_false(fs::dir_exists(output_dir))
  
  # Expect a checkmate assertion error about non-existent directory
  testthat::expect_error(
    generate_test_report(toy_assessment_results, output_dir),
    regexp = "does not exist."
  )
})

test_that("generate_test_report creates an HTML report", {
  skip_on_cran()
  # Create a temporary directory for output
  tmp_dir <- withr::local_tempdir()
  
  # Ensure cleanup of all files in tmp_dir
  withr::defer(unlink(tmp_dir, recursive = TRUE), envir = parent.frame())
  
  # Force rendering in CRAN/win-builder
  withr::local_envvar(NOT_CRAN = "true")
  
  # Run the function
  generate_test_report(toy_assessment_results, tmp_dir)
  
  # Find any .html file in tmp_dir
  html_files <- list.files(tmp_dir, pattern = "\\.html$", full.names = TRUE)
  
  # Check that at least one HTML file exists and is non-empty
  expect_true(length(html_files) > 0, info = "No HTML report was generated")
  expect_true(file.info(html_files[1])$size > 0, info = "HTML report is empty")
  
})

test_that("generate_test_report accepts multi-framework input", {
  skip_on_cran()
  testthat::skip_if_not_installed("rmarkdown")
  tmp_dir <- withr::local_tempdir()
  withr::local_envvar(NOT_CRAN = "true")
  tr_multi <- list(
    pkg_name = "bookdown",
    pkg_ver = "0.43",
    date_time = "2026-03-19 09:29:21",
    executor = "u1004798",
    sysname = "Linux",
    version = "x",
    release = "x",
    machine = "x86_64",
    r_version = "4.4.1",
    multi_framework = TRUE,
    frameworks = c("testthat", "testit"),
    total_cov = 0.225,
    results = list(
      testthat = list(
        total_cov = 0.45,
        res_cov = list(
          name = "bookdown-0.43",
          coverage = list(
            filecoverage = structure(c(84.3, 36.9), dim = 2L, dimnames = list(c("R/addins.R", "R/bs4_book.R"))),
            totalcoverage = 45
          ),
          errors = NA,
          notes = NA
        ),
        framework_id = "testthat"
      ),
      testit = list(
        total_cov = 0,
        res_cov = list(
          name = "bookdown",
          coverage = list(
            filecoverage = structure(0, dim = 1L, dimnames = list("R/foo.R")),
            totalcoverage = 0
          ),
          errors = NA,
          notes = NA
        ),
        framework_id = "testit"
      )
    )
  )
  out <- generate_test_report(tr_multi, tmp_dir)
  expect_true(!is.null(out))
  expect_true(file.exists(out))
  expect_true(endsWith(out, "bookdown_0.43.html"))
})

test_that("generate_coverage_section handles unexpected errors structure (character)", {
  file_names <- c("src/a.R", "src/b.R")
  file_coverage <- matrix(c(50, 60), ncol = 1,
                          dimnames = list(file_names, NULL))
  
  errors <- "Some error occurred"
  notes <- c("n1", "n2")
  
  assessment_results <- list(
    res_cov = list(
      coverage = list(totalcoverage = mean(file_coverage), filecoverage = file_coverage),
      errors = errors,
      notes = notes
    )
  )
  
  mockery::stub(generate_coverage_section, "extract_short_path", function(x) basename(x))
  
  result <- generate_coverage_section(assessment_results, pkg_name = "mockpkg")
  
  expect_identical(result$Function, c("a.R", "b.R"))
  expect_equal(result$Coverage, c(50, 60))
  expect_identical(result$Errors, rep("Some error occurred", 2))
  expect_identical(result$Notes, notes)
})

test_that("generate_coverage_sections returns No coverage data when res_cov or coverage is NULL", {
  # Framework with res_cov = NULL
  tr_null_res_cov <- list(
    pkg_name = "pkg",
    results = list(
      fw_ok = list(
        res_cov = list(
          coverage = list(
            filecoverage = structure(90, dim = 1L, dimnames = list("R/foo.R")),
            totalcoverage = 90
          ),
          errors = NA,
          notes = NA
        )
      ),
      fw_null_cov = list(res_cov = NULL),
      fw_null_coverage = list(res_cov = list(coverage = NULL, errors = NA, notes = NA))
    )
  )
  
  mockery::stub(
    test.assessr:::generate_coverage_sections,
    "generate_coverage_section",
    function(x, pkg_name) data.frame(Function = "foo.R", Coverage = 90, Errors = NA, Notes = NA, stringsAsFactors = FALSE)
  )
  
  out <- test.assessr:::generate_coverage_sections(tr_null_res_cov, pkg_name = "pkg")
  
  expect_named(out, c("fw_ok", "fw_null_cov", "fw_null_coverage"))
  
  # fw_ok: stubbed coverage section
  expect_equal(out$fw_ok$Function, "R/foo.R")
  expect_equal(out$fw_ok$Coverage, 90)
  
  # fw_null_cov and fw_null_coverage: No coverage data branch
  no_cov_df <- data.frame(
    Function = NA_character_,
    Coverage = NA_real_,
    Errors = "No coverage data",
    Notes = "No coverage notes",
    stringsAsFactors = FALSE
  )
  expect_identical(out$fw_null_cov, no_cov_df)
  expect_identical(out$fw_null_coverage, no_cov_df)
})

test_that("generate_coverage_section returns NA-row when filecoverage is NULL", {
  assessment_results <- list(
    res_cov = list(
      coverage = list(totalcoverage = NA_real_, filecoverage = NULL),
      errors = list(message = "m", status = 1),
      notes = NA
    )
  )
  
  result <- generate_coverage_section(assessment_results, pkg_name = "mockpkg")
  
  expect_equal(nrow(result), 1)
  expect_true(all(is.na(result$Function)))
  expect_true(all(is.na(result$Coverage)))
  expect_true(all(is.na(result$Errors)))
  expect_true(all(is.na(result$Notes)))
})




test_that("create_file_coverage_df works correctly with toy dataset", {
  # Toy dataset
  file_names <- c("file1.R", "file2.R")
  file_coverage <- c(85.0, 90.0)
  notes <- c("Note1", "Note2")
  errors <- list(
    message = "in callr subprocess.",
    status = 0
  )
  
  # Expected output with matching row names
  expected_output <- data.frame(
    File = file_names,
    Coverage = file_coverage,
    Errors = rep("in callr subprocess.; 0", length(file_names)),
    Notes = notes,
    stringsAsFactors = FALSE,
    row.names = file_names
  )
  
  
  # Run the function
  result <- create_file_coverage_df(file_names, file_coverage, errors, notes)
  
  # Check if the result matches the expected output
  expect_equal(result, expected_output)
})


test_that("convert_vectors_to_df returns named list of data frames for existing elements", {
  # Mocked input list
  x <- list(
    tests_skipped  = c("glmmPQL.R", "loglm.R", "polr.R"),
    tests_passing  = c("/tmp/MASS/tests/fitdistr.R", "/tmp/MASS/tests/hubers.R"),
    functions_no_tests = c("add.R", "area.R")
  )
  
  # Call function
  res <- convert_vectors_to_df(x, c("tests_skipped", "tests_passing"))
  
  # Basic structure checks
  expect_type(res, "list")
  expect_named(res, c("tests_skipped", "tests_passing"))
  
  # Each element is a data.frame with column 'file'
  expect_s3_class(res$tests_skipped, "data.frame")
  expect_s3_class(res$tests_passing, "data.frame")
  expect_equal(names(res$tests_skipped), "file")
  expect_equal(names(res$tests_passing), "file")
  
  # Row content matches source vectors
  expect_equal(res$tests_skipped$file, x$tests_skipped)
  expect_equal(res$tests_passing$file, x$tests_passing)
})


test_that("convert_vectors_to_df filters out non-existent elements", {
  x <- list(
    tests_skipped = c("a.R", "b.R")
  )
  
  res <- convert_vectors_to_df(x, c("tests_skipped", "tests_passing", "functions_no_tests"))
  
  # Only existing element should be present
  expect_named(res, "tests_skipped")
  expect_equal(nrow(res$tests_skipped), 2)
  expect_equal(res$tests_skipped$file, x$tests_skipped)
})

test_that("convert_vectors_to_df returns empty list when no elems exist", {
  x <- list(
    tests_skipped = character(0)
  )
  res <- convert_vectors_to_df(x, c("tests_passing", "functions_no_tests"))
  
  expect_type(res, "list")
  expect_length(res, 0L)
  expect_equal(names(res), character(0))
})

test_that("convert_vectors_to_df handles empty vectors and NA values", {
  x <- list(
    tests_skipped = character(0),
    tests_passing = c(NA_character_, "foo.R", NA_character_)
  )
  
  res <- convert_vectors_to_df(x, c("tests_skipped", "tests_passing"))
  
  expect_named(res, c("tests_skipped", "tests_passing"))
  
  # Empty vector becomes 0-row data frame
  expect_s3_class(res$tests_skipped, "data.frame")
  expect_equal(nrow(res$tests_skipped), 0L)
  expect_equal(names(res$tests_skipped), "file")
  
  # NA values are kept as-is in the 'file' column
  expect_s3_class(res$tests_passing, "data.frame")
  expect_equal(nrow(res$tests_passing), length(x$tests_passing))
  expect_true(any(is.na(res$tests_passing$file)))
  expect_equal(res$tests_passing$file, x$tests_passing)
})

test_that("convert_vectors_to_df preserves element order as in 'elems' after filtering to existing names", {
  x <- list(
    a = c("a1.R", "a2.R"),
    b = c("b1.R"),
    c = c("c1.R", "c2.R", "c3.R")
  )
  
  res <- convert_vectors_to_df(x, c("c", "a", "missing", "b"))
  expect_named(res, c("c", "a", "b"))
  
  # Confirm each data frame content matches
  expect_equal(res$c$file, x$c)
  expect_equal(res$a$file, x$a)
  expect_equal(res$b$file, x$b)
})

test_that("convert_vectors_to_df works with non-character vectors by coercion (data.frame behavior)", {
  x <- list(
    tests_numeric = c(1, 2, 3),
    tests_logical = c(TRUE, FALSE)
  )
  
  res <- convert_vectors_to_df(x, c("tests_numeric", "tests_logical"))
  
  expect_named(res, c("tests_numeric", "tests_logical"))
  expect_s3_class(res$tests_numeric, "data.frame")
  expect_s3_class(res$tests_logical, "data.frame")
  
  # data.frame will coerce non-character to appropriate column types
  expect_true(is.numeric(res$tests_numeric$file))
  expect_true(is.logical(res$tests_logical$file))
  expect_equal(res$tests_numeric$file, x$tests_numeric)
  expect_equal(res$tests_logical$file, x$tests_logical)
})

test_that("convert_vectors_to_df handles elems = character(0)", {
  x <- list(
    tests_skipped = c("a.R")
  )
  res <- convert_vectors_to_df(x, character(0))
  
  expect_type(res, "list")
  expect_length(res, 0L)
  expect_equal(names(res), character(0))
})

test_that("convert_vectors_to_df handles list with additional names without affecting result", {
  x <- list(
    tests_skipped = c("a.R"),
    extra_matrix = matrix(1:4, nrow = 2),
    inner_list = list(foo = "bar")
  )
  
  res <- convert_vectors_to_df(x, c("tests_skipped", "inner_list"))
  
  # Only elements matching elems and existing in x are included
  expect_named(res, c("tests_skipped", "inner_list"))
  expect_s3_class(res$tests_skipped, "data.frame")
  expect_s3_class(res$inner_list, "data.frame")
  
  # For list element, data.frame(file = x[[nm]]) will create a list-column; ensure it exists
  expect_true("foo" %in% names(res$inner_list))
  expect_equal(nrow(res$inner_list), length(x$inner_list))
})


# Minimal res_cov structure that satisfies assert_list(... names="named")
make_res_cov <- function() {
  list(
    coverage = list(
      totalcoverage = 0.42,
      filecoverage  = data.frame(file = "R/foo.R", cov = 0.9, stringsAsFactors = FALSE)
    )
  )
}

make_test_results_stf <- function(pkg = "pkg", ver = "1.0.0") {
  list(
    pkg_name  = pkg,
    pkg_ver   = ver,
    date_time = Sys.time(),
    total_cov = 0.42,
    res_cov   = make_res_cov()
  )
}

make_test_results_nstf <- function(pkg = "pkg", ver = "1.0.0") {
  list(
    pkg_name  = pkg,
    pkg_ver   = ver,
    date_time = Sys.time(),
    res_cov   = make_res_cov(),
    # NSTF payloads
    functions_no_tests = data.frame(`function` = c("fn_a", "fn_b"), stringsAsFactors = FALSE),
    tests_skipped      = c("test-a", "test-b"),
    tests_passing      = c("path/to/test-1.R", "path/to/test-2.R")
  )
}

make_test_results_multi <- function(pkg = "bookdown", ver = "0.43") {
  list(
    pkg_name = pkg,
    pkg_ver = ver,
    date_time = "2026-03-19 09:29:21",
    executor = "u1004798",
    sysname = "Linux",
    r_version = "4.4.1",
    multi_framework = TRUE,
    frameworks = c("testthat", "testit"),
    total_cov = 0.225,
    results = list(
      testthat = list(
        total_cov = 0.45,
        res_cov = list(
          name = paste0(pkg, "-", ver),
          coverage = list(
            filecoverage = structure(c(84.3, 36.9), dim = 2L, dimnames = list(c("R/addins.R", "R/bs4_book.R"))),
            totalcoverage = 45
          ),
          errors = NA,
          notes = NA
        ),
        test_framework_type = "standard testing framework",
        framework_id = "testthat"
      ),
      testit = list(
        total_cov = 0,
        res_cov = list(
          name = pkg,
          coverage = list(
            filecoverage = structure(0, dim = 1L, dimnames = list("R/foo.R")),
            totalcoverage = 0
          ),
          errors = NA,
          notes = NA
        ),
        test_framework_type = "non-standard testing framework",
        framework_id = "testit"
      )
    )
  )
}

# Multi-framework with NSTF and STF data in testit (for has_nstf/has_stf branch coverage)
make_test_results_multi_nstf_stf <- function(pkg = "bookdown", ver = "0.43") {
  long_summary_df <- data.frame(
    file = "tests/testthat/test-addins.R",
    code_script = "addin_fn",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  test_skip_df <- data.frame(
    file = "tests/testthat/test-skip.R",
    code_script = "skip_fn",
    status = "SKIP",
    expectation = "skip reason",
    line1 = "42",
    stringsAsFactors = FALSE
  )
  functions_no_tests_df <- data.frame(
    source_file = "R/addins.R",
    test_file = NA_character_,
    comment = "No tests",
    stringsAsFactors = FALSE
  )
  tr <- make_test_results_multi(pkg, ver)
  tr$results$testit$functions_no_tests <- functions_no_tests_df
  tr$results$testit$tests_skipped <- c("test-html.R", "test-latex.R")
  tr$results$testit$tests_passing <- c("long/path/to/test-pass.R")
  tr$results$testit$long_summary <- long_summary_df
  tr$results$testit$test_skip <- test_skip_df
  tr
}

# A general stub for system.file that returns a temp path for the requested filename

stub_system_file <- function(...) {
  args <- list(...)
  
  # Extract the requested filename (2nd positional arg)
  filename <- if (length(args) >= 2) args[[2]] else "template.Rmd"
  
  # Create a real file in tempdir()
  tmp <- file.path(tempdir(), filename)
  writeLines(
    c(
      "---",
      "title: 'Stub Template'",
      "output: html_document",
      "---",
      "",
      "Dummy content"
    ),
    tmp
  )
  
  return(tmp)
}


# A simple path_abs stub that mimics fs::path_abs(path, start=output_dir)
stub_path_abs <- function(path, start = NULL) {
  if (!is.null(start)) {
    file.path(start, basename(path))
  } else {
    normalizePath(path, winslash = "/", mustWork = FALSE)
  }
}


# Basic stubs for coverage and other section generators
stub_generate_test_pkg_summary    <- function(test_results) "pkg_summary_stub"
stub_generate_test_details        <- function(test_results, framework_id = NULL) "details_stub"
stub_generate_coverage_section    <- function(test_results, pkg_name) "coverage_stub"

# ---- TESTS ----

test_that("returns NULL with message when rmarkdown is not available", {
  tr <- make_test_results_stf()
  
  # First requireNamespace call should return FALSE -> early exit
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) FALSE)
  
  # Output dir doesn't matter because we should exit first
  expect_message(
    res <- generate_test_report(tr, output_dir = tempdir()),
    "Package 'rmarkdown' is required but not installed\\."
  )
  expect_null(res)
})





test_that("standard testing framework (nstf_none) uses STF template and renders", {
  tr     <- make_test_results_stf()
  outdir <- tempdir()
  
  withr::local_envvar(NOT_CRAN = "true")
  
  # ---- CRITICAL: match how the function calls these symbols ----
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  mockery::stub(generate_test_report, "dir_exists", function(d) TRUE)
  mockery::stub(generate_test_report, "fs::path_abs", stub_path_abs)
  mockery::stub(generate_test_report, "system.file", stub_system_file)
  mockery::stub(generate_test_report, "interactive", function() FALSE)
  
  # NSTF none -> has_stf TRUE, has_nstf FALSE; unified template
  mockery::stub(
    generate_test_report,
    "check_elements_simple",
    function(test_results, elements_to_check, ...) list(present = character(0), missing = elements_to_check)
  )
  
  # Generators
  mockery::stub(generate_test_report, "generate_test_pkg_summary", stub_generate_test_pkg_summary)
  mockery::stub(generate_test_report, "generate_test_details",     stub_generate_test_details)
  mockery::stub(generate_test_report, "generate_coverage_section", stub_generate_coverage_section)
  
  # ---- CRITICAL: stub correct render symbol ----
  # generate_test_report calls rmarkdown::render, not render
  mock_render <- mockery::mock("ok")
  mockery::stub(generate_test_report, "rmarkdown::render", mock_render)
  
  # ---- Act ----
  res <- generate_test_report(tr, output_dir = outdir)
  
  # ---- Assert return file path ----
  
  expect_identical(
    normalizePath(dirname(res), winslash = "/", mustWork = FALSE),
    normalizePath(outdir, winslash = "/", mustWork = FALSE)
  )
  
  
  expect_identical(
    basename(res),
    paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html")
  )
  
  
  expected_outfile <- file.path(outdir, paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html"))
  
  # ---- Assert render() was called ----
  mockery::expect_called(mock_render, 1)
  
  args_list <- mockery::mock_args(mock_render)
  expect_equal(length(args_list), 1)
  
  args <- args_list[[1]]
  expect_true(is.list(args))
  
  # ---- Assert render() arguments ----
  expect_true(grepl("test_report_template\\.Rmd$", as.character(args$input)))
  expect_true(is.environment(args$envir))
  expect_identical(args$quiet, TRUE)
  
  # ---- Assert report_env for single-framework STF ----
  env <- args$envir
  expect_false(env$multi_framework)
  expect_false(env$has_stf)
  expect_false(env$has_nstf)
})



test_that("multi-framework uses unified template and populates framework_results", {
  tr <- make_test_results_multi()
  outdir <- tempdir()
  
  withr::local_envvar(NOT_CRAN = "true")
  
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  mockery::stub(generate_test_report, "dir_exists", function(d) TRUE)
  mockery::stub(generate_test_report, "fs::path_abs", stub_path_abs)
  mockery::stub(generate_test_report, "system.file", stub_system_file)
  mockery::stub(generate_test_report, "interactive", function() FALSE)
  
  # No STF/NSTF elements in any framework
  mockery::stub(
    generate_test_report,
    "check_elements_simple",
    function(test_results, elements_to_check, ...) list(present = character(0), missing = elements_to_check)
  )
  
  mockery::stub(generate_test_report, "generate_test_pkg_summary", stub_generate_test_pkg_summary)
  mockery::stub(generate_test_report, "generate_test_details", stub_generate_test_details)
  mockery::stub(
    generate_test_report,
    "generate_coverage_sections",
    function(test_results, pkg_name) list(testthat = "cov_stub", testit = "cov_stub2")
  )
  
  mock_render <- mockery::mock("ok")
  mockery::stub(generate_test_report, "rmarkdown::render", mock_render)
  
  res <- generate_test_report(tr, output_dir = outdir)
  
  expect_identical(
    normalizePath(dirname(res), winslash = "/", mustWork = FALSE),
    normalizePath(outdir, winslash = "/", mustWork = FALSE)
  )
  expect_identical(basename(res), paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html"))
  
  mockery::expect_called(mock_render, 1)
  args <- mockery::mock_args(mock_render)[[1]]
  expect_true(grepl("test_report_template\\.Rmd$", as.character(args$input)))
  
  env <- args$envir
  expect_true(env$multi_framework)
  expect_equal(env$frameworks, c("testthat", "testit"))
  expect_true(env$has_stf == FALSE)
  expect_true(env$has_nstf == FALSE)
  expect_named(env$framework_results, c("testthat", "testit"))
  expect_equal(env$framework_results$testthat$coverage_output, "cov_stub")
  expect_equal(env$framework_results$testit$coverage_output, "cov_stub2")
})

test_that("multi-framework has_nstf and has_stf exercises NSTF and STF data paths in framework_results", {
  tr <- make_test_results_multi_nstf_stf()
  outdir <- tempdir()
  
  withr::local_envvar(NOT_CRAN = "true")
  
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  mockery::stub(generate_test_report, "dir_exists", function(d) TRUE)
  mockery::stub(generate_test_report, "fs::path_abs", stub_path_abs)
  mockery::stub(generate_test_report, "system.file", stub_system_file)
  mockery::stub(generate_test_report, "interactive", function() FALSE)
  
  # has_nstf=TRUE and has_stf=TRUE: both NSTF and STF elements present (nested mode)
  nstf_set <- c("functions_no_tests", "tests_skipped", "tests_passing")
  stf_set <- c("long_summary", "test_skip")
  mockery::stub(
    generate_test_report,
    "check_elements_simple",
    function(test_results, elements_to_check, ...) {
      if (setequal(elements_to_check, nstf_set)) {
        list(present = nstf_set, missing = character(0))
      } else if (setequal(elements_to_check, stf_set)) {
        list(present = stf_set, missing = character(0))
      } else {
        list(present = character(0), missing = elements_to_check)
      }
    }
  )
  
  mockery::stub(generate_test_report, "generate_test_pkg_summary", stub_generate_test_pkg_summary)
  mockery::stub(generate_test_report, "generate_test_details", stub_generate_test_details)
  mockery::stub(
    generate_test_report,
    "generate_coverage_sections",
    function(test_results, pkg_name) list(testthat = "cov_stub", testit = "cov_stub2")
  )
  
  # Stub convert_vectors_to_df for tests_skipped and tests_passing (multi-framework calls with list(x), "name")
  mockery::stub(
    generate_test_report,
    "convert_vectors_to_df",
    function(x, elems) {
      nm <- elems[elems %in% names(x)]
      if (length(nm) == 0) return(list())
      nm <- nm[1]
      if (nm == "tests_skipped") {
        list(data.frame(file = c("test-html.R", "test-latex.R"), stringsAsFactors = FALSE))
      } else if (nm == "tests_passing") {
        list(data.frame(file = c("long/path/to/test-pass.R"), stringsAsFactors = FALSE))
      } else {
        list(data.frame(file = if (nm %in% names(x) && length(x[[nm]]) > 0) x[[nm]] else character(0), stringsAsFactors = FALSE))
      }
    }
  )
  
  # detect_path_shortening: TRUE for dfs with path columns to exercise shortening
  mockery::stub(
    generate_test_report,
    "detect_path_shortening",
    function(df, file_col = c("file", "source_file"), ...) {
      if (!is.data.frame(df)) return(FALSE)
      any(tolower(file_col) %in% tolower(names(df)))
    }
  )
  
  # create_function_locations_df: return transformed df
  mockery::stub(
    generate_test_report,
    "create_function_locations_df",
    function(df, ...) {
      data.frame(
        file = "addins.R",
        function_name = "fn",
        start_line = 1L,
        end_line = 10L,
        stringsAsFactors = FALSE
      )
    }
  )
  
  mockery::stub(generate_test_report, "extract_short_path", function(x) basename(x))
  
  # create_long_summary_details and create_test_skip_details stubs
  mockery::stub(
    generate_test_report,
    "create_long_summary_details",
    function(df) data.frame(`R function` = "addin_fn", test = "test-addins.R", start_line = NA_character_, status = "PASS", stringsAsFactors = FALSE)
  )
  mockery::stub(
    generate_test_report,
    "create_test_skip_details",
    function(df) data.frame(Rfunction = "skip_fn", test = "test-skip.R", status = "SKIP", expectation = "skip reason", start_line = "42", stringsAsFactors = FALSE)
  )
  
  mock_render <- mockery::mock("ok")
  mockery::stub(generate_test_report, "rmarkdown::render", mock_render)
  
  res <- generate_test_report(tr, output_dir = outdir)
  
  expect_identical(basename(res), paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html"))
  
  mockery::expect_called(mock_render, 1)
  args <- mockery::mock_args(mock_render)[[1]]
  env <- args$envir
  
  expect_true(env$multi_framework)
  expect_true(env$has_nstf)
  expect_true(env$has_stf)
  expect_named(env$framework_results, c("testthat", "testit"))
  
  # testit has NSTF data: functions_no_tests, tests_skipped_df, tests_passing_df
  testit <- env$framework_results$testit
  expect_true(is.data.frame(testit$functions_no_tests))
  expect_equal(testit$functions_no_tests$file, "addins.R")
  expect_true(is.data.frame(testit$tests_skipped_df))
  expect_equal(testit$tests_skipped_df$file, c("test-html.R", "test-latex.R"))
  expect_true(is.data.frame(testit$tests_passing_df))
  expect_equal(testit$tests_passing_df$file, "test-pass.R")
  
  # testit has STF data: long_summary_df, tests_skip_df
  expect_true(is.data.frame(testit$long_summary_df))
  expect_true(is.data.frame(testit$tests_skip_df))
  expect_equal(testit$tests_skip_df$Rfunction, "skip_fn")
})

testthat::test_that("NSTF: renders NSTF template, converts vectors->df, shortens file paths, and returns HTML in output_dir (basic render arg checks)", {
  tr     <- make_test_results_nstf()
  outdir <- tempdir()
  
  withr::local_envvar(NOT_CRAN = "true")
  
  # --- CRITICAL: stubs must match unqualified names used inside generate_test_report ---
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  
  # NOTE: generate_test_report() uses fs::path_abs(), not path_abs()
  # Keep your stub if you like, but this one actually matches the call site:
  if (exists("stub_path_abs", inherits = TRUE)) {
    mockery::stub(generate_test_report, "fs::path_abs", stub_path_abs)
  } else {
    # Fallback: identity if you don't have a custom stub
    mockery::stub(generate_test_report, "fs::path_abs", function(p) normalizePath(p, winslash = "/", mustWork = FALSE))
  }
  
  # You don't need dir_exists() because the code uses checkmate::assert_directory_exists()
  # But harmless if you want to keep it:
  mockery::stub(generate_test_report, "dir_exists", function(d) TRUE)
  
  mockery::stub(generate_test_report, "system.file", stub_system_file)
  mockery::stub(generate_test_report, "interactive", function() FALSE)
  
  # NSTF present (functions_no_tests, tests_skipped, tests_passing); STF absent (long_summary, test_skip)
  mockery::stub(
    generate_test_report,
    "check_elements_simple",
    function(test_results, elements_to_check, ...) {
      nstf <- c("functions_no_tests", "tests_skipped", "tests_passing")
      stf <- c("long_summary", "test_skip")
      if (identical(sort(elements_to_check), sort(nstf))) {
        list(present = nstf, missing = character(0))
      } else if (identical(sort(elements_to_check), sort(stf))) {
        list(present = character(0), missing = stf)
      } else {
        list(present = character(0), missing = elements_to_check)
      }
    }
  )
  
  # Generators
  mockery::stub(generate_test_report, "generate_test_pkg_summary", stub_generate_test_pkg_summary)
  mockery::stub(generate_test_report, "generate_test_details",     stub_generate_test_details)
  mockery::stub(generate_test_report, "generate_coverage_section", stub_generate_coverage_section)
  
  # Convert vectors + shorten paths
  mockery::stub(
    generate_test_report,
    "convert_vectors_to_df",
    function(test_results, name)
      list(data.frame(
        file = c("long/path/to/file_one.R", "another/deep/path/file_two.R"),
        x = 1:2,
        stringsAsFactors = FALSE
      ))
  )
  mockery::stub(generate_test_report, "extract_short_path", function(x) basename(x))
  
  # --- NEW: Tolerant stub for detect_path_shortening -------------------------
  # Return FALSE if there is no candidate path column; TRUE otherwise.
  mockery::stub(
    generate_test_report,
    "detect_path_shortening",
    function(df, file_col = c("file", "source_file"), ...) {
      stopifnot(is.data.frame(df))
      # case-insensitive check for the presence of any candidate
      has_path_col <- any(tolower(file_col) %in% tolower(names(df)))
      if (!has_path_col) return(FALSE)
      # If there IS a path column, you can choose TRUE (to exercise shortening)
      # or FALSE (to skip). Either is fine for this test because "tests_passing"
      # is shortened unconditionally later. Let's return TRUE:
      TRUE
    }
  )
  # ---------------------------------------------------------------------------
  
  # --- CRITICAL: stub correct symbol for render ---
  mock_render <- mockery::mock("ok")
  mockery::stub(generate_test_report, "rmarkdown::render", mock_render)
  
  # --- Act ---
  res <- generate_test_report(tr, output_dir = outdir)
  
  # --- Assert return path ---
  testthat::expect_identical(
    normalizePath(dirname(res), winslash = "/", mustWork = FALSE),
    normalizePath(outdir, winslash = "/", mustWork = FALSE)
  )
  
  testthat::expect_identical(
    basename(res),
    paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html")
  )
  
  expected_outfile <- file.path(outdir, paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html"))
  
  # --- Assert render() was called ---
  mockery::expect_called(mock_render, 1)
  
  args_list <- mockery::mock_args(mock_render)
  testthat::expect_equal(length(args_list), 1)
  
  args <- args_list[[1]]
  testthat::expect_true(is.list(args))
  
  # --- Assert NSTF template was used ---
  testthat::expect_true(grepl("test_report_template\\.Rmd$", as.character(args$input)))
  
  # --- Assert render args ---
  testthat::expect_true(is.environment(args$envir))
  testthat::expect_identical(args$quiet, TRUE)
  
  # --- Assert NSTF environment values (data frames processed properly) ---
  env <- args$envir
  testthat::expect_true(is.data.frame(env$functions_no_tests))
  testthat::expect_true(is.data.frame(env$tests_skipped_df))
  testthat::expect_true(is.data.frame(env$tests_passing_df))
  
  testthat::expect_identical(env$tests_passing_df$file, c("file_one.R", "file_two.R"))
  
  # --- Assert report_env for single-framework NSTF ----
  testthat::expect_false(env$multi_framework)
  testthat::expect_false(env$has_stf)
  testthat::expect_true(env$has_nstf)
})



test_that("errors on NULL output_dir with tolerant message match (no render spy)", {
  tr <- make_test_results_stf()
  
  # Ensure we don't early-return due to missing rmarkdown
  withr::local_envvar(NOT_CRAN = "true")
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  
  # Now, because output_dir must be a non-empty string, NULL should error
  expect_error(
    generate_test_report(tr, output_dir = NULL),
    regexp = "(non-empty.*string|Assertion on 'output_dir')",
    fixed  = FALSE
  )
})


test_that("errors when required columns are missing", {
  df_bad <- data.frame(a = 1)
  expect_error(create_long_summary_details(df_bad), "Input df must contain columns")
})

test_that("parses file without '#' and returns NA start_line", {
  df <- data.frame(
    file = "path/to/file.R",
    code_script = "fun",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "path/to/file.R")
  expect_equal(res$`R.function`, "fun")
  expect_true(is.na(res$start_line))
})

test_that("extracts start_line from after '#' up to the first underscore", {
  df <- data.frame(
    file = "dir/file.R#123_abc",
    code_script = "do_stuff",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "dir/file.R")
  expect_equal(res$start_line, "123")
  expect_equal(res$`R.function`, "do_stuff")
})

test_that("trailing '#' with empty after_hash yields NA start_line", {
  df <- data.frame(
    file = "dir/file.R#",
    code_script = "f",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_true(is.na(res$start_line))
})

test_that("filters only rows with status == PASS (case and whitespace tolerant)", {
  df <- data.frame(
    file = c("a.R", "b.R", "c.R", "d.R"),
    code_script = c("f1", "f2", "f3", "f4"),
    status = c(" pass ", "FAIL", "PASS", "pAsS"),
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 3)
  expect_setequal(res$`R.function`, c("f1", "f3", "f4"))
})

test_that("using mockery::stub to override sub allows testing parsing branches", {
  # Replacement inspects the pattern argument to decide what to return,
  # so we can simulate the three different sub() usages inside the function.
  replacement_sub <- function(pattern, replacement, x, ...) {
    if (identical(pattern, "#.*$")) return("patched_before")
    if (identical(pattern, ".*#")) return("patched_after")
    if (identical(pattern, "_.*$")) return("patched_start")
    return(NA_character_)
  }
  
  # Stub `sub` inside create_long_summary_details, then call the function
  mockery::stub(create_long_summary_details, "sub", replacement_sub)
  
  df <- data.frame(
    file = "anything",           # value doesn't matter because sub is stubbed
    code_script = "foo",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "patched_before")
  expect_equal(res$status, "PASS")
})

test_that("stubbing grepl to FALSE simulates no '#' present and results in NA start_line", {
  # Force grepl to always say there is no '#'
  mockery::stub(create_long_summary_details, "grepl", function(pattern, x, ...) FALSE)
  
  df <- data.frame(
    file = "dir/file.R#1_abc",   # original string contains '#', but grepl is stubbed
    code_script = "foo",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_true(is.na(res$start_line))
})

test_that("errors when required columns are missing", {
  df_bad <- data.frame(a = 1)
  expect_error(
    create_test_skip_details(df_bad),
    "Input df must contain columns: file, code_script, status, expectation, line1"
  )
})

test_that("errors when only some required columns are present", {
  df_partial <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP"
  )
  expect_error(
    create_test_skip_details(df_partial),
    "Input df must contain columns: file, code_script, status, expectation, line1"
  )
})

test_that("processes valid dataframe with all required columns", {
  df <- data.frame(
    file = "test-myscript.R",
    code_script = "my_func",
    status = "SKIP",
    expectation = "function should handle NA",
    line1 = "42",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 1)
  expect_equal(res$`Rfunction`, "my_func")
  expect_equal(res$test, "test-myscript.R")
  expect_equal(res$status, "SKIP")
  expect_equal(res$expectation, "function should handle NA")
  expect_equal(res$start_line, "42")
})

test_that("extracts portion before '#' from file column", {
  df <- data.frame(
    file = "test-example.R#100_context",
    code_script = "example_func",
    status = "SKIP",
    expectation = "should skip this",
    line1 = "100",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "test-example.R")
})

test_that("handles file names without '#' separator", {
  df <- data.frame(
    file = "test-nomodifier.R",
    code_script = "func",
    status = "SKIP",
    expectation = "no hash present",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "test-nomodifier.R")
})

test_that("preserves digits in test file names", {
  df <- data.frame(
    file = "test-utils2.R#200",
    code_script = "util_func",
    status = "SKIP",
    expectation = "test expectation",
    line1 = "200",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "test-utils2.R")
})

test_that("handles multiple rows correctly", {
  df <- data.frame(
    file = c("test-a.R#1", "test-b.R", "test-c.R#50_info"),
    code_script = c("func_a", "func_b", "func_c"),
    status = c("SKIP", "SKIP", "SKIP"),
    expectation = c("expect a", "expect b", "expect c"),
    line1 = c("1", "20", "50"),
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 3)
  expect_equal(res$test[1], "test-a.R")
  expect_equal(res$test[2], "test-b.R")
  expect_equal(res$test[3], "test-c.R")
  expect_equal(res$`Rfunction`, c("func_a", "func_b", "func_c"))
})

test_that("converts input columns to character type", {
  df <- data.frame(
    file = "test. R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = 123,  # numeric input
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$start_line, "123")
  expect_true(is.character(res$start_line))
})

test_that("returns dataframe with correct column names and order", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expected_cols <- c("Rfunction", "test", "status", "expectation", "start_line")
  expect_equal(colnames(res), expected_cols)
})

test_that("removes row names (resets to default numeric indexing)", {
  df <- data.frame(
    file = c("test-a.R", "test-b.R"),
    code_script = c("f1", "f2"),
    status = c("SKIP", "SKIP"),
    expectation = c("e1", "e2"),
    line1 = c("1", "2"),
    row.names = c("custom_1", "custom_2"),
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(rownames(res), c("1", "2"))
})

test_that("using mockery::stub to override sub allows testing file parsing", {
  # Stub sub() to return predictable values based on pattern
  replacement_sub <- function(pattern, replacement, x, ...) {
    if (identical(pattern, "#.*$")) return("stubbed_file.R")
    return(x)
  }
  
  mockery::stub(create_test_skip_details, "sub", replacement_sub)
  
  df <- data.frame(
    file = "original_file.R#some_hash",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "42",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "stubbed_file.R")
})

test_that("stubbing sub to identity preserves full file path including '#'", {
  # Stub sub() to act as identity (no replacement)
  replacement_sub <- function(pattern, replacement, x, .) {
    return(x)  # Always return x unchanged
  }
  
  mockery::stub(create_test_skip_details, "sub", replacement_sub)
  
  df <- data.frame(
    file = "test-preserve. R#100_suffix",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "100",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  # Since sub is stubbed to return x unchanged, test should be full file string
  expect_equal(res$test, "test-preserve. R#100_suffix")
})

test_that("handles empty/whitespace in expectation and status columns", {
  df <- data.frame(
    file = "test. R",
    code_script = "func",
    status = "  ",
    expectation = "",
    line1 = "5",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 1)
  expect_equal(res$status, "  ")
  expect_equal(res$expectation, "")
})

test_that("handles special characters in expectation column", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "function should throw error:  'value < 0' is invalid; see #123",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(
    res$expectation,
    "function should throw error:  'value < 0' is invalid; see #123"
  )
})

test_that("line1 column is directly assigned to start_line without modification", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "999",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$start_line, "999")
})

test_that("stringsAsFactors=FALSE is handled correctly for character conversion", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  # Check that all expected columns are character class
  expect_true(is.character(res$`Rfunction`))
  expect_true(is.character(res$test))
  expect_true(is.character(res$status))
  expect_true(is.character(res$expectation))
  expect_true(is.character(res$start_line))
})

test_that("handles very long file paths and special delimiters", {
  df <- data.frame(
    file = "path/to/deeply/nested/test-complex_name.R#line_456_with_underscores",
    code_script = "complex_function_name",
    status = "SKIP",
    expectation = "should handle complex paths",
    line1 = "456",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "path/to/deeply/nested/test-complex_name.R")
  expect_equal(res$`Rfunction`, "complex_function_name")
})

test_that("stubbing as.character allows testing type coercion behavior", {
  # Mock as.character to prefix all values with "MOCK_"
  call_count <- 0
  
  replacement_as_char <- function(x) {
    call_count <<- call_count + 1
    paste0("MOCK_", x)
  }
  
  # Note: We can only stub one instance; here we stub the first as.character call
  # In practice, you might test the overall behavior instead
  mockery::stub(create_test_skip_details, "as.character", replacement_as_char)
  
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  # After stubbing, all character conversions should be prefixed
  expect_match(res$`Rfunction`, "^MOCK_")
})



testthat::test_that("NSTF: verifies fs path handling and render(output_file), exercises create_function_locations_df, and processes data frames with path shortening", {
  tr     <- make_test_results_nstf()
  outdir <- tempdir()
  
  withr::local_envvar(NOT_CRAN = "true")
  
  # --- Stubs that must match the *actual call sites* in generate_test_report ---
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  
  # generate_test_report() uses fs::path_abs(), so stub that (not path_abs)
  if (exists("stub_path_abs", inherits = TRUE)) {
    mockery::stub(generate_test_report, "fs::path_abs", stub_path_abs)
  } else {
    mockery::stub(generate_test_report, "fs::path_abs", function(p) normalizePath(p, winslash = "/", mustWork = FALSE))
  }
  
  # Optional: this isn't actually called (assert_directory_exists is used),
  # but harmless to leave
  mockery::stub(generate_test_report, "dir_exists", function(d) TRUE)
  
  mockery::stub(generate_test_report, "system.file", stub_system_file)
  mockery::stub(generate_test_report, "interactive", function() FALSE)
  
  # NSTF present (functions_no_tests, tests_skipped, tests_passing); STF absent (long_summary, test_skip)
  nstf_set <- c("functions_no_tests", "tests_skipped", "tests_passing")
  stf_set <- c("long_summary", "test_skip")
  mockery::stub(
    generate_test_report, "check_elements_simple",
    function(test_results, elements_to_check, ...) {
      if (setequal(elements_to_check, nstf_set)) {
        list(present = nstf_set, missing = character(0))
      } else if (setequal(elements_to_check, stf_set)) {
        list(present = character(0), missing = stf_set)
      } else {
        list(present = character(0), missing = elements_to_check)
      }
    }
  )
  
  # Generators
  mockery::stub(generate_test_report, "generate_test_pkg_summary", stub_generate_test_pkg_summary)
  mockery::stub(generate_test_report, "generate_test_details",     stub_generate_test_details)
  mockery::stub(generate_test_report, "generate_coverage_section", stub_generate_coverage_section)
  
  # Convert vectors -> df with a 'file' column
  mockery::stub(
    generate_test_report, "convert_vectors_to_df",
    function(test_results, name)
      list(data.frame(
        file = c("long/path/to/file_one.R", "another/deep/path/file_two.R"),
        x    = 1:2,
        stringsAsFactors = FALSE
      ))
  )
  
  # Tolerant path-shortening decision:
  # - If there's no path column, return FALSE to skip shortening (avoids errors)
  # - If there is, return TRUE to exercise the branch
  mockery::stub(
    generate_test_report, "detect_path_shortening",
    function(df, file_col = c("file", "source_file"), ...) {
      stopifnot(is.data.frame(df))
      has_path_col <- any(tolower(file_col) %in% tolower(names(df)))
      if (!has_path_col) return(FALSE)
      TRUE
    }
  )
  
  # Ensure functions_no_tests can be transformed into a file-located data frame
  mockery::stub(
    generate_test_report, "create_function_locations_df",
    function(df, ...) {
      n <- NROW(df)
      files <- if ("file" %in% names(df)) df$file else rep("file_one.R", max(1L, n))
      data.frame(
        file          = files,
        function_name = if ("function_name" %in% names(df)) as.character(df$function_name) else paste0("fn_", seq_len(length(files))),
        start_line    = if ("start_line" %in% names(df)) as.integer(df$start_line) else seq_len(length(files)),
        end_line      = if ("end_line" %in% names(df)) as.integer(df$end_line) else seq_len(length(files)) + 1L,
        stringsAsFactors = FALSE
      )
    }
  )
  
  mockery::stub(generate_test_report, "extract_short_path", function(x) basename(x))
  
  # --- Spy for render: must match fully qualified symbol used in the function
  mock_render <- mockery::mock("ok")
  mockery::stub(generate_test_report, "rmarkdown::render", mock_render)
  
  # --- Act ---
  res <- generate_test_report(tr, output_dir = outdir)
  
  # Build expected path the same way the function does or coerce to character
  expected_outfile <- fs::path(outdir, paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html"))
  
  # Coerce to character to avoid class mismatch (fs_path vs character)
  expected_basename <- paste0("test_report_", tr$pkg_name, "_", tr$pkg_ver, ".html")
  testthat::expect_identical(
    basename(as.character(res)),
    expected_basename
  )
  
  
  # --- Assert render() was called and inspect its arguments ---
  mockery::expect_called(mock_render, 1)
  args <- mockery::mock_args(mock_render)[[1]]
  
  testthat::expect_true(grepl("test_report_template\\.Rmd$", as.character(args$input)))
  
  testthat::expect_identical(
    basename(as.character(args$output_file)),
    expected_basename
  )
  
  testthat::expect_true(is.environment(args$envir))
  testthat::expect_identical(args$quiet, TRUE)
  
  # --- Assert NSTF environment values (data frames processed properly) ---
  env <- args$envir
  testthat::expect_true(is.data.frame(env$functions_no_tests))
  testthat::expect_true(is.data.frame(env$tests_skipped_df))
  testthat::expect_true(is.data.frame(env$tests_passing_df))
  
  testthat::expect_identical(env$tests_passing_df$file, c("file_one.R", "file_two.R"))
  
  # --- Assert report_env for single-framework NSTF ----
  testthat::expect_false(env$multi_framework)
  testthat::expect_false(env$has_stf)
  testthat::expect_true(env$has_nstf)
})




testthat::test_that("errors on NULL output_dir and fail-fast prevents rendering (exact message; render not called)", {
  tr <- make_test_results_stf()
  
  # Ensure we don't try to render even if the env var is set
  withr::local_envvar(NOT_CRAN = "true")
  
  # Only the minimal stubs needed before the assertion fires
  mockery::stub(generate_test_report, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  
  # Spy on render to ensure it is NOT called (must match the call site in the function)
  mock_render <- mockery::mock("ok")
  mockery::stub(generate_test_report, "rmarkdown::render", mock_render)
  
  # Act + Assert: because of `assert_string`, this must error immediately
  testthat::expect_error(
    generate_test_report(tr, output_dir = NULL),
    "Assertion on 'output_dir' failed"
  )
  
  # Render should not be called because we fail fast on input validation
  mockery::expect_called(mock_render, 0)
})


test_that("errors when required columns are missing", {
  df_bad <- data.frame(a = 1)
  expect_error(create_long_summary_details(df_bad), "Input df must contain columns")
})

test_that("parses file without '#' and returns NA start_line", {
  df <- data.frame(
    file = "path/to/file.R",
    code_script = "fun",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "path/to/file.R")
  expect_equal(res$`R.function`, "fun")
  expect_true(is.na(res$start_line))
})

test_that("extracts start_line from after '#' up to the first underscore", {
  df <- data.frame(
    file = "dir/file.R#123_abc",
    code_script = "do_stuff",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "dir/file.R")
  expect_equal(res$start_line, "123")
  expect_equal(res$`R.function`, "do_stuff")
})

test_that("trailing '#' with empty after_hash yields NA start_line", {
  df <- data.frame(
    file = "dir/file.R#",
    code_script = "f",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_true(is.na(res$start_line))
})

test_that("filters only rows with status == PASS (case and whitespace tolerant)", {
  df <- data.frame(
    file = c("a.R", "b.R", "c.R", "d.R"),
    code_script = c("f1", "f2", "f3", "f4"),
    status = c(" pass ", "FAIL", "PASS", "pAsS"),
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 3)
  expect_setequal(res$`R.function`, c("f1", "f3", "f4"))
})

test_that("using mockery::stub to override sub allows testing parsing branches", {
  # Replacement inspects the pattern argument to decide what to return,
  # so we can simulate the three different sub() usages inside the function.
  replacement_sub <- function(pattern, replacement, x, ...) {
    if (identical(pattern, "#.*$")) return("patched_before")
    if (identical(pattern, ".*#")) return("patched_after")
    if (identical(pattern, "_.*$")) return("patched_start")
    return(NA_character_)
  }
  
  # Stub `sub` inside create_long_summary_details, then call the function
  mockery::stub(create_long_summary_details, "sub", replacement_sub)
  
  df <- data.frame(
    file = "anything",           # value doesn't matter because sub is stubbed
    code_script = "foo",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "patched_before")
  expect_equal(res$status, "PASS")
})

test_that("stubbing grepl to FALSE simulates no '#' present and results in NA start_line", {
  # Force grepl to always say there is no '#'
  mockery::stub(create_long_summary_details, "grepl", function(pattern, x, ...) FALSE)
  
  df <- data.frame(
    file = "dir/file.R#1_abc",   # original string contains '#', but grepl is stubbed
    code_script = "foo",
    status = "PASS",
    stringsAsFactors = FALSE
  )
  
  res <- create_long_summary_details(df)
  expect_equal(nrow(res), 1)
  expect_true(is.na(res$start_line))
})

test_that("errors when required columns are missing", {
  df_bad <- data.frame(a = 1)
  expect_error(
    create_test_skip_details(df_bad),
    "Input df must contain columns: file, code_script, status, expectation, line1"
  )
})

test_that("errors when only some required columns are present", {
  df_partial <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP"
  )
  expect_error(
    create_test_skip_details(df_partial),
    "Input df must contain columns: file, code_script, status, expectation, line1"
  )
})

test_that("processes valid dataframe with all required columns", {
  df <- data.frame(
    file = "test-myscript.R",
    code_script = "my_func",
    status = "SKIP",
    expectation = "function should handle NA",
    line1 = "42",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 1)
  expect_equal(res$`Rfunction`, "my_func")
  expect_equal(res$test, "test-myscript.R")
  expect_equal(res$status, "SKIP")
  expect_equal(res$expectation, "function should handle NA")
  expect_equal(res$start_line, "42")
})

test_that("extracts portion before '#' from file column", {
  df <- data.frame(
    file = "test-example.R#100_context",
    code_script = "example_func",
    status = "SKIP",
    expectation = "should skip this",
    line1 = "100",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 1)
  expect_equal(res$test, "test-example.R")
})

test_that("handles file names without '#' separator", {
  df <- data.frame(
    file = "test-nomodifier.R",
    code_script = "func",
    status = "SKIP",
    expectation = "no hash present",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "test-nomodifier.R")
})

test_that("preserves digits in test file names", {
  df <- data.frame(
    file = "test-utils2.R#200",
    code_script = "util_func",
    status = "SKIP",
    expectation = "test expectation",
    line1 = "200",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "test-utils2.R")
})

test_that("handles multiple rows correctly", {
  df <- data.frame(
    file = c("test-a.R#1", "test-b.R", "test-c.R#50_info"),
    code_script = c("func_a", "func_b", "func_c"),
    status = c("SKIP", "SKIP", "SKIP"),
    expectation = c("expect a", "expect b", "expect c"),
    line1 = c("1", "20", "50"),
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 3)
  expect_equal(res$test[1], "test-a.R")
  expect_equal(res$test[2], "test-b.R")
  expect_equal(res$test[3], "test-c.R")
  expect_equal(res$`Rfunction`, c("func_a", "func_b", "func_c"))
})

test_that("converts input columns to character type", {
  df <- data.frame(
    file = "test. R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = 123,  # numeric input
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$start_line, "123")
  expect_true(is.character(res$start_line))
})

test_that("returns dataframe with correct column names and order", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expected_cols <- c("Rfunction", "test", "status", "expectation", "start_line")
  expect_equal(colnames(res), expected_cols)
})

test_that("removes row names (resets to default numeric indexing)", {
  df <- data.frame(
    file = c("test-a.R", "test-b.R"),
    code_script = c("f1", "f2"),
    status = c("SKIP", "SKIP"),
    expectation = c("e1", "e2"),
    line1 = c("1", "2"),
    row.names = c("custom_1", "custom_2"),
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(rownames(res), c("1", "2"))
})

test_that("using mockery::stub to override sub allows testing file parsing", {
  # Stub sub() to return predictable values based on pattern
  replacement_sub <- function(pattern, replacement, x, ...) {
    if (identical(pattern, "#.*$")) return("stubbed_file.R")
    return(x)
  }
  
  mockery::stub(create_test_skip_details, "sub", replacement_sub)
  
  df <- data.frame(
    file = "original_file.R#some_hash",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "42",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "stubbed_file.R")
})

test_that("stubbing sub to identity preserves full file path including '#'", {
  # Stub sub() to act as identity (no replacement)
  replacement_sub <- function(pattern, replacement, x, .) {
    return(x)  # Always return x unchanged
  }
  
  mockery::stub(create_test_skip_details, "sub", replacement_sub)
  
  df <- data.frame(
    file = "test-preserve. R#100_suffix",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "100",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  # Since sub is stubbed to return x unchanged, test should be full file string
  expect_equal(res$test, "test-preserve. R#100_suffix")
})

test_that("handles empty/whitespace in expectation and status columns", {
  df <- data.frame(
    file = "test. R",
    code_script = "func",
    status = "  ",
    expectation = "",
    line1 = "5",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(nrow(res), 1)
  expect_equal(res$status, "  ")
  expect_equal(res$expectation, "")
})

test_that("handles special characters in expectation column", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "function should throw error:  'value < 0' is invalid; see #123",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(
    res$expectation,
    "function should throw error:  'value < 0' is invalid; see #123"
  )
})

test_that("line1 column is directly assigned to start_line without modification", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "999",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$start_line, "999")
})

test_that("stringsAsFactors=FALSE is handled correctly for character conversion", {
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  # Check that all expected columns are character class
  expect_true(is.character(res$`Rfunction`))
  expect_true(is.character(res$test))
  expect_true(is.character(res$status))
  expect_true(is.character(res$expectation))
  expect_true(is.character(res$start_line))
})

test_that("handles very long file paths and special delimiters", {
  df <- data.frame(
    file = "path/to/deeply/nested/test-complex_name.R#line_456_with_underscores",
    code_script = "complex_function_name",
    status = "SKIP",
    expectation = "should handle complex paths",
    line1 = "456",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  expect_equal(res$test, "path/to/deeply/nested/test-complex_name.R")
  expect_equal(res$`Rfunction`, "complex_function_name")
})

test_that("stubbing as.character allows testing type coercion behavior", {
  # Mock as.character to prefix all values with "MOCK_"
  call_count <- 0
  
  replacement_as_char <- function(x) {
    call_count <<- call_count + 1
    paste0("MOCK_", x)
  }
  
  # Note: We can only stub one instance; here we stub the first as.character call
  # In practice, you might test the overall behavior instead
  mockery::stub(create_test_skip_details, "as.character", replacement_as_char)
  
  df <- data.frame(
    file = "test.R",
    code_script = "func",
    status = "SKIP",
    expectation = "expect",
    line1 = "10",
    stringsAsFactors = FALSE
  )
  
  res <- create_test_skip_details(df)
  
  # After stubbing, all character conversions should be prefixed
  expect_match(res$`Rfunction`, "^MOCK_")
})


# Helper: construct a long string of n chars
.str_n <- function(n) paste(rep("x", n), collapse = "")

test_that("errors if df is not a data.frame", {
  expect_error(detect_path_shortening(1:3), "`df` must be a data.frame")
})

test_that("errors when requested column is not found", {
  df <- data.frame(other = "a", stringsAsFactors = FALSE)
  expect_error(
    detect_path_shortening(df, "file"),
    "Could not find any of the specified path columns"
  )
})

test_that("auto-resolves the first present among c('file','source_file')", {
  # 'file' (first candidate) has no slashes, 'source_file' would trigger TRUE if used
  df <- data.frame(
    file        = c("a.R", "b.R"),
    source_file = c("/long/path/a.R", "/long/path/b.R"),
    stringsAsFactors = FALSE
  )
  # Should pick 'file' and return FALSE (no slashes -> dir_share < 0.2)
  expect_false(detect_path_shortening(df))
})

test_that("accepts 'source_file' explicitly (case-insensitive)", {
  df <- data.frame(
    Source_File = paste0("/", .str_n(70), "/file.R"), # long > 60
    stringsAsFactors = FALSE
  )
  # Explicitly request source_file (case-insensitive); too_long should trigger TRUE
  expect_true(detect_path_shortening(df, "source_file"))
})

test_that("returns FALSE when no usable paths (all NA or empty)", {
  df <- data.frame(file = c(NA_character_, ""), stringsAsFactors = FALSE)
  expect_false(detect_path_shortening(df, "file"))
})

test_that("returns FALSE when few or no directory separators (dir_share < min_dir_share)", {
  df <- data.frame(
    file = c("file1.R", "file2.R", "file3.R"),
    stringsAsFactors = FALSE
  )
  expect_false(detect_path_shortening(df, "file"))
})

test_that("returns TRUE when 'too_long' (any path length > max_len_threshold)", {
  long_path <- paste0("/", .str_n(61), "/a.R") # > 60 chars
  df <- data.frame(file = c(long_path, "/short/b.R"), stringsAsFactors = FALSE)
  expect_true(detect_path_shortening(df, "file"))
})

test_that("returns TRUE when 'big_gap' (max_full - max_base >= gap_threshold)", {
  # Make a path with a long directory part but short basename
  p1 <- paste0("/", .str_n(30), "/x.R") # full ~ 33, base ~ 3, diff ~ 30
  p2 <- "/short/y.R"
  df <- data.frame(file = c(p1, p2), stringsAsFactors = FALSE)
  # Ensure 'too_long' doesn't dominate (keep default 60) -> big_gap should trigger TRUE
  expect_true(detect_path_shortening(df, "file"))
})

test_that("returns TRUE when 'long_common_prefix' alone (with len and gap thresholds high)", {
  # Two paths sharing a long common prefix of length >= 20, but not too long otherwise
  prefix20 <- .str_n(20)
  df <- data.frame(
    file = c(paste0("/", prefix20, "/a.R"),
             paste0("/", prefix20, "/b.R")),
    stringsAsFactors = FALSE
  )
  # Raise thresholds so only long_common_prefix is considered
  expect_true(detect_path_shortening(
    df,
    file_col = "file",
    max_len_threshold = 1e9,
    gap_threshold = 1e9,
    min_common_prefix = 20L
  ))
})

test_that("handles Windows-style paths (normalization to '/')", {
  df <- data.frame(
    file = c("C:\\temp\\pkg\\R\\file.R", "D:\\w\\pkg\\src\\c.cpp"),
    stringsAsFactors = FALSE
  )
  # Ensure 'too_long' triggers by lowering threshold
  expect_true(detect_path_shortening(df, "file", max_len_threshold = 10L))
})

test_that("mockery::stub - force dir_share=0 even when slashes are present", {
  df <- data.frame(
    file = c("/has/slash/a.R", "/has/slash/b.R"),
    stringsAsFactors = FALSE
  )
  
  # Local copy so global function isn't mutated for other tests
  fn <- detect_path_shortening
  
  # Stub grepl(pattern, x, ...) to always return FALSE (no entries have '/')
  fake_grepl <- function(pattern, x, ...) rep(FALSE, length(x))
  mockery::stub(fn, "grepl", fake_grepl)
  
  # Now even with slashes in data, dir_share will be 0 and function should return FALSE
  expect_false(fn(df, "file"))
})

test_that("candidate vector works when only source_file exists", {
  df <- data.frame(
    source_file = c("/pkg/R/a.R", "/pkg/R/b.R"),
    stringsAsFactors = FALSE
  )
  # Should pick source_file automatically (default candidates) and return FALSE by default
  # unless it meets other heuristics; make it 'too_long' to be deterministic TRUE.
  df$source_file <- paste0("/", .str_n(61), "/a.R")
  expect_true(detect_path_shortening(df))  # uses default candidates c("file","source_file")
})

# Test helper: shorten after '/R/' else keep last 2 segments
shortener_keep_after_R <- function(p) {
  s <- gsub("\\\\", "/", p)
  out <- sub("^.*?/R/", "", s)
  no_anchor <- (out == s)
  if (any(no_anchor)) {
    s2 <- s[no_anchor]
    out[no_anchor] <- vapply(s2, function(z) {
      seg <- strsplit(z, "/", fixed = TRUE)[[1]]
      paste(utils::tail(seg, 2L), collapse = "/")
    }, FUN.VALUE = character(1))
  }
  out
}


test_that("returns normalized df with expected columns and types", {
  df_in <- data.frame(
    function_name = c(".onLoad", ".onAttach"),
    file         = c("/tmp/x/pkg/R/aaa.R", "/tmp/x/pkg/R/bbb.R"),
    start_line   = c(10L, 20L),
    end_line     = c(15L, 30L),
    stringsAsFactors = FALSE
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = NULL,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_s3_class(out, "data.frame")
  expect_named(out, c("file", "function_name", "start_line", "end_line"))
  expect_type(out$file, "character")
  expect_type(out$function_name, "character")
  expect_type(out$start_line, "integer")
  expect_type(out$end_line, "integer")
  
  # Shortener applied
  expect_equal(out$file, c("aaa.R", "bbb.R"))
})

test_that("when extract_short_path = NULL, file paths are not shortened", {
  df_in <- data.frame(
    function_name = "f",
    file         = "/a/very/long/path/R/foo.R",
    start_line   = 1L,
    end_line     = 2L,
    stringsAsFactors = FALSE
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = NULL,
    extract_short_path = NULL
  )
  
  expect_equal(out$file, df_in$file)
})

test_that("derives file from file_coverage dimnames when file column is missing", {
  df_in <- data.frame(
    function_name = c("f1", "f2", "f3"),
    start_line   = c(1L, 2L, 3L),
    end_line     = c(2L, 3L, 4L),
    stringsAsFactors = FALSE
  )
  
  fc <- structure(
    1:3,
    dim = c(3L, 1L),
    dimnames = list(
      c("/tmp/pkg/R/a.R", "/tmp/pkg/R/b.R", "/tmp/pkg/src/c.c"),
      "dummy"
    )
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = fc,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_equal(out$file, c("a.R", "b.R", "src/c.c")) # note: strip_until_R fallback keeps 'src/c.c'
})

test_that("recycles single-length file_coverage dimnames", {
  df_in <- data.frame(
    function_name = c("f1", "f2"),
    start_line   = c(10L, 20L),
    end_line     = c(11L, 21L),
    stringsAsFactors = FALSE
  )
  
  fc <- structure(
    1,
    dim = c(1L, 1L),
    dimnames = list("/tmp/pkg/R/only.R", "dummy")
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = fc,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_equal(out$file, rep("only.R", 2))
})

test_that("pads/truncates file_coverage dimnames when lengths mismatch", {
  df_in <- data.frame(
    function_name = c("f1", "f2", "f3", "f4"),
    start_line   = 1:4,
    end_line     = 2:5,
    stringsAsFactors = FALSE
  )
  
  fc <- structure(
    1:2,
    dim = c(2L, 1L),
    dimnames = list(
      c("/x/R/a.R", "/x/R/b.R"),
      "dummy"
    )
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = fc,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_equal(out$file, c("a.R", "b.R", NA_character_, NA_character_))
})

test_that("accepts case-insensitive and alias column names", {
  df_in <- data.frame(
    FUNCtion = c("g1", "g2"),
    filepath = c("/root/R/x.R", "/root/R/y.R"),
    StartLine = c(5, 6),
    ENDLINE   = c(9, 10),
    stringsAsFactors = FALSE
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = NULL,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_equal(out$function_name, c("g1", "g2"))
  expect_equal(out$start_line, c(5L, 6L))
  expect_equal(out$end_line, c(9L, 10L))
  expect_equal(out$file, c("x.R", "y.R"))
})

test_that("errors when required columns are missing", {
  df_in <- data.frame(
    file = "/a/b/c.R",
    x = 1:2,
    stringsAsFactors = FALSE
  )
  
  expect_error(
    create_function_locations_df(df_in),
    "Missing required column\\(s\\): .*function_name.*start_line.*end_line"
  )
})

test_that("handles NA and empty file elements gracefully", {
  df_in <- data.frame(
    function_name = c("f1", "f2", "f3"),
    file         = c("/tmp/R/a.R", NA_character_, ""),
    start_line   = c(1L, 2L, 3L),
    end_line     = c(4L, 5L, 6L),
    stringsAsFactors = FALSE
  )
  
  out <- create_function_locations_df(
    df = df_in,
    file_coverage = NULL,
    extract_short_path = shortener_keep_after_R
  )
  # First shortened, others unchanged (NA / empty)
  expect_equal(out$file, c("a.R", NA_character_, ""))
})

# ---- Tests using mockery::stub ----

test_that("mockery::stub: attr() failure in file_coverage is caught and yields NA files", {
  # No file column; it should try to read dimnames via attr(), which we'll force to fail.
  df_in <- data.frame(
    function_name = c("h1", "h2", "h3"),
    start_line   = c(1L, 2L, 3L),
    end_line     = c(2L, 3L, 4L),
    stringsAsFactors = FALSE
  )
  
  # Dummy object—attr() call will be stubbed to error regardless of contents
  fc <- structure(1:3, dim = c(3L, 1L))
  
  # Local copy to stub (so we don't alter the original binding for other tests)
  fn <- create_function_locations_df
  
  # Force attr() to fail to exercise the tryCatch error path
  mockery::stub(fn, "attr", function(...) stop("attr failure"))
  
  out <- fn(
    df = df_in,
    file_coverage = fc,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_true(all(is.na(out$file)))
  expect_equal(out$function_name, df_in$function_name)
  expect_equal(out$start_line, df_in$start_line)
  expect_equal(out$end_line, df_in$end_line)
})

test_that("mockery::stub: attr() returns custom dimnames vector used as files", {
  df_in <- data.frame(
    function_name = c("k1", "k2"),
    start_line   = c(10L, 20L),
    end_line     = c(11L, 21L),
    stringsAsFactors = FALSE
  )
  
  fc <- structure(1:2, dim = c(2L, 1L))
  
  # Local copy to stub
  fn <- create_function_locations_df
  
  # Return a controlled dimnames list to ensure we exercise that branch w/o building a full object
  mockery::stub(
    fn,
    "attr",
    function(object, which) {
      if (identical(which, "dimnames")) list(c("/tmp/R/A.R", "/tmp/R/B.R")) else NULL
    }
  )
  
  out <- fn(
    df = df_in,
    file_coverage = fc,
    extract_short_path = shortener_keep_after_R
  )
  
  expect_equal(out$file, c("A.R", "B.R"))
})

#' Detect which test frameworks are present in a package
#'
#' @param test_pkg_data List from \code{check_pkg_tests_and_snaps}.
#' @return Character vector of framework IDs: "testthat", "testit", "tests_base",
#'   "tinytest", "data.table", "bioconductor".
#' @keywords internal
detect_frameworks <- function(test_pkg_data) {
  fw <- character(0)
  if (isTRUE(test_pkg_data$has_testthat))         fw <- c(fw, "testthat")
  if (isTRUE(test_pkg_data$has_testit))           fw <- c(fw, "testit")
  if (isTRUE(test_pkg_data$has_tests_base))       fw <- c(fw, "tests_base")
  if (isTRUE(test_pkg_data$has_tinytest))         fw <- c(fw, "tinytest")
  if (isTRUE(test_pkg_data$has_data_table_tests)) fw <- c(fw, "data.table")
  if (isTRUE(test_pkg_data$has_BioG_test))        fw <- c(fw, "bioconductor")
  if (isTRUE(test_pkg_data$has_RUnit_test) &&
      !isTRUE(test_pkg_data$has_BioG_test))       fw <- c(fw, "runit")
  unique(fw)
}


#' Compute aggregate total coverage across multiple frameworks
#'
#' @param per_framework Named list of covr results from each framework.
#' @return Numeric. Mean of per-framework total_cov (0-100), or 0 if none.
#' @keywords internal
compute_multi_framework_total <- function(per_framework) {
  out <- 0
  if (length(per_framework) > 0) {
    covs <- vapply(per_framework, function(x) {
      tryCatch(as.numeric(x$total_cov)[1L], error = function(e) 0)
    }, FUN.VALUE = numeric(1))
    non_na_covs <- covs[!is.na(covs)]
    if (length(non_na_covs) > 0) {
      out <- mean(non_na_covs)
    }
  }
  return(out)
}


#' Get human-readable test framework type label
#'
#' @param test_pkg_data List from \code{check_pkg_tests_and_snaps}.
#' @param framework_id Character. Framework ID from \code{detect_frameworks}.
#' @return Character. Label for test_framework_type.
#' @keywords internal
get_test_framework_type <- function(test_pkg_data, framework_id) {
  switch(
    framework_id,
    testthat = "standard testing framework",
    tinytest = "non-standard testing framework - tinytest",
    "data.table" = "non-standard testing framework - data.table",
    runit = "non-standard testing framework - RUnit",
    bioconductor = "non-standard testing framework - Bioconductor",
    testit       = "non-standard testing framework - testit",
    "non-standard testing framework"
  )
}


#' Initialise a temporary coverage environment (framework-aware)
#' @param pkg_source_path Character; package root.
#' @param pkg             Character; package name.
#' @param framework       Character; e.g. "testthat", "testit", "bioc", "none".
#' @param work_dir        Optional explicit temp workspace; if NULL it is composed from inputs.
#' @return environment cov_env with $work_dir and $pkg_root set.
#' @keywords internal
initialise_covr_env <- function(pkg_source_path, pkg, framework = "none", work_dir = NULL) {
  stopifnot(is.character(pkg_source_path), length(pkg_source_path) == 1L)
  stopifnot(is.character(pkg),            length(pkg)            == 1L)
  stopifnot(is.character(framework),      length(framework)      == 1L)
  
  if (is.null(work_dir) || !nzchar(work_dir)) {
    work_dir <- file.path(
      tempdir(),
      sprintf("covr_%s_%s_%s", pkg, framework, Sys.getpid())
    )
  }
  dir.create(work_dir, recursive = TRUE, showWarnings = FALSE)
  
  
  # Use the pre-existing binding in the package namespace
  e <- cov_env   # this resolves to test.assessr:::cov_env inside package code
  # Optionally clear only the keys you own to avoid leaking prior state across runs
  rm(list = intersect(ls(e), c("work_dir","pkg_root","package","framework","pkg_path")), envir = e)
  
  e$work_dir  <- work_dir
  e$pkg_root  <- normalizePath(pkg_source_path, winslash = "/", mustWork = TRUE)
  e$package   <- pkg
  e$framework <- framework
  e$pkg_path  <- pkg_source_path
  e
  
}

#' Decide and run the appropriate coverage path for a test framework
#'
#' Self-contained helper used by `run_one_framework()` to decide whether to run
#' standard (testthat) coverage via `run_coverage()` or non-standard (testit)
#' coverage via `run_covr_skip_nstf()`. The decision is made by inspecting files
#' under `<pkg>/tests/` directly (no reliance on `check_pkg_tests_and_snaps()`).
#'
#' Precedence (when framework == "testit"):
#'   1) If a testthat-style driver is detected -> run_coverage()   
#'   2) Else if a testit driver/structure is detected -> NSTF     
#'
#' @param framework Character scalar (e.g., "testthat", "testit").
#' @param pkg_source_path Character scalar; path to the package root.
#' @param test_pkg_data List; kept for signature compatibility.
#' @param cov_env Environment; coverage execution environment (forwarded).
#' @param covr_timeout Numeric/integer; timeout for `run_coverage()`.
#' @return list(covr_list = <coverage>, tf_type = <label>)
#' @keywords internal
#' @noRd
run_covr_decider <- function(framework,
                             pkg_source_path,
                             test_pkg_data,   
                             cov_env,
                             covr_timeout) {
  
  pkg_source_path <- normalizePath(pkg_source_path, mustWork = TRUE)
  tests_dir       <- file.path(pkg_source_path, "tests")
  test_all_file   <- file.path(tests_dir, "test-all.R")
  testthat_file   <- file.path(tests_dir, "testthat.R")
  testthat_dir    <- file.path(tests_dir, "testthat")
  testit_dir      <- file.path(tests_dir, "testit")
  
  read_safely <- function(path) {
    if (!file.exists(path)) return(character())
    tryCatch(readLines(path, warn = FALSE), error = function(e) character())
  }
  
  # --- Detect drivers directly from test files (no external state) -----------
  txt_all   <- read_safely(test_all_file)
  txt_tthat <- read_safely(testthat_file)
  
  # testit driver: accept namespaced or bare test_pkg(...)
  detects_test_pkg <- function(x) any(grepl("\\b(?:testit\\s*::\\s*)?test_pkg\\s*\\(", x))
  # testthat driver: test_package(...) or test_check(...)
  detects_test_package <- function(x) any(grepl("(?:testthat\\s*::\\s*)?(test_package|test_check)\\s*\\(", x))
  
  has_testit_driver_text    <- detects_test_pkg(txt_all) || detects_test_pkg(txt_tthat)
  has_testthat_driver_text  <- detects_test_package(txt_all) || detects_test_package(txt_tthat)
  has_test_all              <- file.exists(test_all_file)
  has_testthat_structure    <- dir.exists(testthat_dir)
  has_testit_structure      <- dir.exists(testit_dir)
  
  # ---------------------- Decision precedence --------------------------------
  # testthat -> always standard
  if (identical(framework, "testthat")) {
    message("decider: testthat -> run_coverage()")
    covr_list <- run_coverage(pkg_source_path, covr_timeout)
    return(list(covr_list = covr_list, tf_type = "standard testing framework"))
  }
  
  # testit -> prefer "line 65" (run_coverage) for DT-like layout:
  #           tests/test-all.R present AND NO tests/testthat/ directory
  #        -> else NSTF if we see explicit testit signals
  #        -> else standard if we see explicit testthat driver
  #        -> else NSTF default
  if (identical(framework, "testit")) {
    if (has_test_all && !has_testthat_structure) {
      message("decider: tests/test-all.R present and no tests/testthat/ -> run_coverage()  [DT-style]")
      covr_list <- run_coverage(pkg_source_path, covr_timeout)
      return(list(covr_list = covr_list, tf_type = "standard testing framework - testit"))
    }
    if (has_testit_driver_text || has_testit_structure) {
      message("decider: testit driver/structure detected -> run_covr_skip_nstf()")
      covr_list <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env, framework = "testit")
      return(list(covr_list = covr_list, tf_type = "non-standard testing framework - testit"))
    }
    if (has_testthat_driver_text || has_testthat_structure) {
      message("decider: testthat-style driver/structure detected -> run_coverage()")
      covr_list <- run_coverage(pkg_source_path, covr_timeout)
      return(list(covr_list = covr_list, tf_type = "standard testing framework"))
    }
    message("decider: no explicit driver found for testit -> default run_covr_skip_nstf()")
    covr_list <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env, framework = "testit")
    return(list(covr_list = covr_list, tf_type = "non-standard testing framework - testit"))
  }
  
}

#' Run coverage for a single test framework
#'
#' @param framework Character. Framework ID.
#' @param pkg Character. Package name.
#' @param test_pkg_data List from \code{check_pkg_tests_and_snaps}.
#' @param pkg_source_path Character. Path to package source.
#' @param covr_timeout Numeric. Timeout for coverage.
#' @return List with total_cov, res_cov, test_framework_type, framework_id.
#' @keywords internal
run_one_framework <- function(framework,
                              pkg,
                              test_pkg_data,
                              pkg_source_path,
                              covr_timeout) {
  
  safe_unload_package(pkg)
  
  cov_env <- initialise_covr_env(pkg_source_path, pkg, framework = framework)
  on.exit(
    cleanup_and_return_null(
      msg = sprintf("Cleaning coverage workspace for %s", pkg),
      env = cov_env
    ),
    add = TRUE
  )
  
  
  if (framework %in% c("testthat", "testit")) {
    # Decide and execute the correct runner
    res <- run_covr_decider(
      framework       = framework,
      pkg_source_path = pkg_source_path,
      test_pkg_data   = test_pkg_data,
      cov_env         = cov_env,
      covr_timeout    = covr_timeout
    )
    
    covr_list <- res$covr_list
    tf_type   <- res$tf_type
    
    # Keep your original "skip" fallback for testthat only
    if (identical(framework, "testthat")) {
      fc <- tryCatch(covr_list$res_cov$coverage$filecoverage, error = function(e) NULL)
      if (isTRUE(is.na(covr_list$total_cov)) && (is.null(fc) || all(is.na(fc)))) {
        message("running skip testing with standard testing framework")
        covr_list <- run_covr_skip_stf(pkg_source_path, test_pkg_data, cov_env)
      }
    }
    
    
  } else if (framework == "tests_base") {
    message("running non-standard testing framework")
    covr_list <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env, framework = framework)
    tf_type <- "non-standard testing framework"
    
  } else if (framework == "tinytest") {
    message("running non-standard tinytest testing framework")
    covr_list <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env, framework = framework)
    tf_type <- "non-standard testing framework - tinytest"
    
  } else if (framework == "data.table") {
    message("running non-standard data.table testing framework")
    covr_list <- run_covr_skip_dt_nstf(pkg_source_path, test_pkg_data, cov_env)
    tf_type <- "non-standard testing framework - data.table"
  
  }  else if (test_pkg_data$has_RUnit_test && !test_pkg_data$has_BioG_test) {
    message("running non-standard RUnit testing framework")
    covr_list <- run_covr_skip_runit_nstf(pkg_source_path, test_pkg_data, cov_env, framework = framework)    
    tf_type <- "non-standard testing framework - RUnit"  
    
  } else if (framework == "bioconductor") {
    message("running non-standard Bioconductor testing framework")
    covr_list <- run_covr_skip_bioc_nstf(pkg_source_path, test_pkg_data, cov_env, framework = framework)
    tf_type <- "non-standard testing framework - Bioconductor"
    
  } else {
    message("No recognised standard or non-standard testing configuration")
    covr_list <- list(
      total_cov = 0,
      res_cov = list(
        name = pkg,
        coverage = list(
          filecoverage = matrix(0, nrow = 1, dimnames = list("No functions tested")),
          totalcoverage = 0
        ),
        errors = "No recognised standard or non-standard testing configuration",
        notes = NA
      )
    )
    tf_type <- "non-standard testing framework"
  }
  
  if (is.null(covr_list)) {
    covr_list <- create_empty_covr_list(pkg, paste0("Coverage failed for framework: ", framework))
  }
  
  covr_list$test_framework_type <- tf_type
  covr_list$framework_id <- framework
  
  covr_list
}


#' Run Coverage Analysis with Test Detection
#'
#' This function inspects the test configuration of an R package and runs 
#' code coverage analysis using any available testing framework, including 
#' `testthat`, `testit`, base R test scripts, or Bioconductor-style tests. 
#' If no recognised testing configuration is found, a default zero‑coverage
#' result is returned.
#'
#' @param pkg_source_path Character. Path to the root directory of the 
#'   package source.
#' @param covr_timeout Numeric. Timeout in seconds for running coverage 
#'   analysis. Default is 60.
#'
#' @return 
#' A named list containing coverage results, package metadata, and test
#' configuration details. The returned list includes (but is not limited to):
#' \describe{
#'   \item{pkg_name}{Character. Package name extracted from the DESCRIPTION file.}
#'   \item{pkg_ver}{Character. Package version.}
#'   \item{date_time}{Timestamp of when the analysis was run.}
#'   \item{executor}{User or environment running the analysis.}
#'   \item{sysname, version, release, machine}{System metadata.}
#'   \item{r_version}{R version used during analysis.}
#'   \item{test_framework_type}{Character. Detected testing framework type.}
#'   \item{covr_list}{A nested list containing: 
#'     \describe{
#'       \item{total_cov}{Numeric. Aggregated coverage percentage.}
#'       \item{res_cov}{File‑level and line‑level coverage details.}
#'       \item{errors, notes}{Any warnings or notes detected during testing.}
#'     }
#'   }
#'   \item{test_pkg_data}{A list describing the test configuration of the
#'     package (presence of testthat, testit, base tests, snapshots, etc.).}
#' }
#'
#' This function always returns a value. When no supported testing framework is
#' detected, a default object with zero coverage and diagnostic information is
#' returned.
#'
#' @examples
#' \donttest{
#'
#' dp <- system.file("test-data", 
#'   "test.package.0001_0.1.0.tar.gz", 
#'   package = "test.assessr")
#' 
#' # set up package
#' install_list <- set_up_pkg(dp)
#'
#' package_installed <- install_list$package_installed
#' pkg_source_path <- install_list$pkg_source_path
#'
#' # install package locally to ensure test works
#' package_installed <- install_package_local(pkg_source_path)
#' package_installed <- TRUE
#'
#' covr_mode_list <- run_covr_modes(pkg_source_path)
#'
#' }
#'
#' @export
run_covr_modes <- function(pkg_source_path,
                           covr_timeout = 60) {
  
  # --- Temporarily set CRAN repo (restore immediately on exit) ---
  old_repos <- getOption("repos")
  tmp_repos <- old_repos
  tmp_repos["CRAN"] <- "http://cran.us.r-project.org"
  options(repos = tmp_repos)
  on.exit(options(repos = old_repos), add = TRUE)
  
  # ---- metadata ------------------------------------------------------------
  pkg <- get_pkg_name(pkg_source_path)
  pkg_desc <- get_pkg_desc(pkg_source_path, fields = c("Package", "Version"))
  pkg_name <- pkg_desc$Package
  pkg_ver <- pkg_desc$Version
  metadata <- get_test_metadata()
  r_version <- as.character(getRversion())
  
  # ---- detect frameworks ---------------------------------------------------
  test_pkg_data <- check_pkg_tests_and_snaps(pkg_source_path)
  frameworks_found <- detect_frameworks(test_pkg_data)
  
  # ---- no framework: original fallback -------------------------------------
  if (length(frameworks_found) == 0) {
    message("No recognised standard or non-standard testing configuration")
    covr_list <- list(
      total_cov = 0,
      res_cov = list(
        name = pkg,
        coverage = list(
          filecoverage = matrix(0, nrow = 1, dimnames = list("No functions tested")),
          totalcoverage = 0
        ),
        errors = "No recognised standard or non-standard testing configuration",
        notes = NA
      )
    )
    covr_list <- c(
      list(
        pkg_name = pkg_name,
        pkg_ver = pkg_ver,
        date_time = metadata$datetime,
        executor = metadata$executor,
        sysname = metadata$info$sys$sysname,
        version = metadata$info$sys$version,
        release = metadata$info$sys$release,
        machine = metadata$info$sys$machine,
        r_version = r_version,
        test_framework_type = "non-standard testing framework"
      ),
      covr_list
    )
    return(covr_list)
  }

  # ---- single framework: preserve original return shape --------------------
  if (length(frameworks_found) == 1) {
    one <- run_one_framework(
      framework = frameworks_found[[1]],
      pkg = pkg,
      test_pkg_data = test_pkg_data,
      pkg_source_path = pkg_source_path,
      covr_timeout = covr_timeout
    )
    one$test_framework_type <- NULL
    one$framework_id <- NULL
    
    covr_list <- c(
      list(
        pkg_name = pkg_name,
        pkg_ver = pkg_ver,
        date_time = metadata$datetime,
        executor = metadata$executor,
        sysname = metadata$info$sys$sysname,
        version = metadata$info$sys$version,
        release = metadata$info$sys$release,
        machine = metadata$info$sys$machine,
        r_version = r_version,
        test_framework_type = get_test_framework_type(test_pkg_data, frameworks_found[[1]])
      ),
      one
    )
    return(covr_list)
  }
  
  # ---- multiple frameworks: run each independently -------------------------
  message("Multiple test frameworks detected: ", paste(frameworks_found, collapse = ", "))
  
  per_framework <- lapply(frameworks_found, function(fr) {
    
    run_one_framework(
      framework = fr,
      pkg = pkg,
      test_pkg_data = test_pkg_data,
      pkg_source_path = pkg_source_path,
      covr_timeout = covr_timeout
    )
  })
  names(per_framework) <- frameworks_found
  
  total_cov <- compute_multi_framework_total(per_framework)
  
  out <- c(
    list(
      pkg_name = pkg_name,
      pkg_ver = pkg_ver,
      date_time = metadata$datetime,
      executor = metadata$executor,
      sysname = metadata$info$sys$sysname,
      version = metadata$info$sys$version,
      release = metadata$info$sys$release,
      machine = metadata$info$sys$machine,
      r_version = r_version,
      multi_framework = TRUE,
      frameworks = frameworks_found,
      total_cov = total_cov
    ),
    list(results = per_framework)
  )
  
  return(out)
}


#' Set up a coverage environment for isolated test execution
#'
#' Prepare an existing coverage environment for use with coverage workflows
#' that evaluate tests in an isolated environment, typically one created
#' with `new.env(parent = baseenv())`.
#'
#' The setup process:
#'
#' - loads helper files from `tests/testthat`,
#' - loads package namespace objects into `cov_env`,
#' - resolves the project root,
#' - loads package datasets into `cov_env`,
#' - loads Base R datasets and exports into `cov_env`,
#' - optionally loads `nycflights13` datasets if available,
#' - loads dependencies into `cov_env`, and
#' - unloads the package namespace afterward to reduce conflicts.
#'
#' This function does **not** create `cov_env`; it expects an existing
#' environment to be supplied.
#'
#' @param pkg_source_path Character scalar. Path to the source package root.
#' @param pkg_name Character scalar. Package name.
#' @param test_path Character scalar. Path to the test directory.
#'   Defaults to `"tests/testthat"`.
#' @param cov_env Environment. A pre-existing coverage environment, usually
#'   created by an initializer such as `initialise_covr_env()`.
#'
#' @return A named list with components:
#' \describe{
#'   \item{cov_env}{The populated coverage environment.}
#'   \item{datasets_loaded}{Character vector of dataset objects loaded into
#'   `cov_env` from the target package, `datasets`, and optionally
#'   `nycflights13`.}
#'   \item{base_r_datasets_loaded}{Character vector of datasets loaded from
#'   the `datasets` package.}
#'   \item{base_r_exports_loaded}{Character vector of exported objects copied
#'   from Base R packages other than `base` and `datasets`.}
#'   \item{root_dir}{Resolved project root path.}
#'   \item{initial_state}{Character vector of object names present in
#'   `cov_env` before setup.}
#'   \item{helpers_loaded}{Character vector of helper files successfully
#'   sourced into `cov_env`.}
#' }
#'
#' @details
#' This function is designed for workflows where `cov_env` is deliberately
#' isolated (for example with `parent = baseenv()`). In such cases, objects
#' from packages like `stats` and `datasets` are not automatically visible,
#' so they are explicitly loaded into `cov_env`.
#'
#' The package `nycflights13` is treated as optional: if it is installed,
#' its datasets are loaded into `cov_env`; otherwise it is silently skipped.
#'
#' @keywords internal
setup_covr_env <- function(pkg_source_path,
                           pkg_name,
                           test_path = "tests/testthat",
                           cov_env) {
  message(paste0("Setting up coverage environment for: ", pkg_name))
  
  # --- Require an existing cov_env (this function no longer creates one) ---
  if (missing(cov_env) || !is.environment(cov_env)) {
    stop("`cov_env` must be provided and must be an environment (e.g., from initialise_covr_env()).")
  }
  
  checkmate::assert_string(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_string(pkg_name, .var.name = "pkg_name")
  checkmate::assert_directory_exists(test_path, .var.name = "test_path")
  checkmate::assert_environment(cov_env, .var.name = "cov_env")
  
  datasets_loaded <- character()
  base_r_datasets_loaded <- character()
  base_r_exports_loaded <- character()
  root_dir <- NULL
  helpers_loaded <- character()
  
  # Record initial state before setup
  initial_state <- ls(envir = cov_env, all.names = TRUE)
  
  tryCatch({
    # ------------------------------------------------------------------
    # Load helper files (READ-ONLY from tests/, no writes there)
    # ------------------------------------------------------------------
    message(paste0("Loading helpers into coverage environment for: ", pkg_name))
    helper_files <- list.files(
      test_path,
      pattern = "^helper.*\\.R$",
      full.names = TRUE
    )
    
    for (file in helper_files) {
      tryCatch({
        sys.source(file, envir = cov_env)
        helpers_loaded <- c(helpers_loaded, file)
      }, error = function(e) {
        message("Failed to source helper file: ", file, " : ", e$message)
      })
    }
    
    # ------------------------------------------------------------------
    # Load package code into namespace, then copy objects into cov_env
    # ------------------------------------------------------------------
    message(paste0("Loading package into coverage environment for: ", pkg_name))
    pkgload::load_all(
      pkg_source_path,
      helpers = FALSE,
      attach = FALSE,
      quiet = TRUE
    )
    
    pkg_ns <- asNamespace(pkg_name)
    pkg_all <- ls(envir = pkg_ns, all.names = TRUE)
    
    for (obj in pkg_all) {
      tryCatch({
        assign(
          obj,
          get(obj, envir = pkg_ns, inherits = FALSE),
          envir = cov_env
        )
      }, error = function(e) {
        message("Failed to assign object ", obj, " : ", e$message)
      })
    }
    
    # ------------------------------------------------------------------
    # Locate project root (read-only)
    # ------------------------------------------------------------------
    message(paste0("Locating project root for: ", pkg_name))
    root_dir <- find_project_root()
    if (is.null(root_dir)) {
      stop("Project root not found. Coverage cannot be run.")
    }
    
    # ------------------------------------------------------------------
    # Load package datasets into cov_env
    # ------------------------------------------------------------------
    message(paste0("Loading package datasets into coverage environment for: ", pkg_name))
    datasets_loaded <- load_package_datasets(
      pkg_names = pkg_name,
      env = cov_env,
      source_paths = stats::setNames(list(pkg_source_path), pkg_name),
      overwrite = FALSE
    )
    
    # ------------------------------------------------------------------
    # Load Base R package surface into cov_env
    # ------------------------------------------------------------------
    message("Loading Base R packages into coverage environment")
    base_r_loaded <- load_base_r_packages_into_env(
      env = cov_env,
      overwrite = FALSE
    )
    
    base_r_exports_loaded <- base_r_loaded$exports_loaded
    base_r_datasets_loaded <- base_r_loaded$datasets_loaded
    
    # ------------------------------------------------------------------
    # Optional: load nycflights13 datasets if installed
    # ------------------------------------------------------------------
    message("Loading optional nycflights13 datasets into coverage environment")
    nycflights_loaded <- load_package_datasets(
      pkg_names = "nycflights13",
      env = cov_env,
      source_paths = NULL,
      overwrite = FALSE
    )
    
    datasets_loaded <- unique(c(
      datasets_loaded,
      base_r_datasets_loaded,
      nycflights_loaded
    ))
    
    if (length(datasets_loaded) > 0L) {
      message("Datasets loaded: ", paste(datasets_loaded, collapse = ", "))
    } else {
      message("No datasets were loaded into coverage environment.")
    }
    
    message("Base R exports loaded: ", length(base_r_exports_loaded))
    
    # ------------------------------------------------------------------
    # Load dependencies into env
    # ------------------------------------------------------------------
    message(paste0("Loading dependencies into coverage environment for: ", pkg_name))
    load_dependencies_into_env(pkg_source_path, cov_env)
    
    # ------------------------------------------------------------------
    # Optional sanity checks
    # ------------------------------------------------------------------
    if (!exists("rnorm", envir = cov_env, inherits = FALSE)) {
      message("Note: `rnorm` was not loaded into cov_env.")
    }
    if (!exists("mtcars", envir = cov_env, inherits = FALSE)) {
      message("Note: `mtcars` was not loaded into cov_env.")
    }
    if (!exists("iris", envir = cov_env, inherits = FALSE)) {
      message("Note: `iris` was not loaded into cov_env.")
    }
    
    # ------------------------------------------------------------------
    # Unload package to avoid conflicts after copying objects into cov_env
    # ------------------------------------------------------------------
    message(paste0("Unloading package: ", pkg_name))
    safe_unload_package(pkg_name)
    
  }, error = function(e) {
    cleanup_and_return_null(
      paste0("Error during setup for ", pkg_name, " : ", e$message),
      env = cov_env
    )
  })
  
  return(list(
    cov_env = cov_env,
    datasets_loaded = datasets_loaded,
    base_r_datasets_loaded = base_r_datasets_loaded,
    base_r_exports_loaded = base_r_exports_loaded,
    root_dir = root_dir,
    initial_state = initial_state,
    helpers_loaded = helpers_loaded
  ))
}





#' @title Load Datasets from a Package
#' @description 
#' Loads all datasets from a specified installed package into the cov environment.
#' It first attempts to load `.rda` files from the package's `data/` directory.
#' If none are found, it falls back to using `data()` to load datasets listed in the package metadata.
#'
#' @param pkg_names A character string specifying the name of the package.
#' @param env A character string specifying the covr env
#' @param source_paths A character string specifying the source paths
#' @param overwrite A logical specifying whether datasets should be overwritten
#'
#' @return A character vector of dataset names that were attempted to be loaded.
#' Returns `NULL` if the package is not installed.
#'
#' @details 
#' This function is intended for internal use. It silently attempts to load each dataset
#' and suppresses errors if loading fails. Datasets are loaded into the cov environment.
#'
#' @keywords internal
#' @importFrom utils read.table
#' @importFrom utils read.csv
#' @importFrom utils data
load_package_datasets <- function(pkg_names,
                                  env,
                                  source_paths = NULL,
                                  overwrite = FALSE) {
  checkmate::assert_character(
    pkg_names,
    min.len = 1,
    any.missing = FALSE,
    .var.name = "pkg_names"
  )
  checkmate::assert_environment(env, .var.name = "env")
  checkmate::assert_flag(overwrite, .var.name = "overwrite")
  
  # source_paths can be:
  # - NULL
  # - a named character vector/list, e.g. c("mypkg" = "/path/to/src")
  if (!is.null(source_paths)) {
    if (!(is.character(source_paths) || is.list(source_paths))) {
      stop("`source_paths` must be NULL, a named character vector, or a named list.")
    }
    if (is.null(names(source_paths)) || any(names(source_paths) == "")) {
      stop("`source_paths` must be named by package name.")
    }
    source_paths <- as.list(source_paths)
  }
  
  loaded <- character()
  
  load_from_file <- function(file, env, overwrite = FALSE) {
    ext <- tolower(tools::file_ext(file))
    before <- ls(envir = env, all.names = TRUE)
    
    out <- tryCatch({
      if (ext %in% c("rda", "rdata")) {
        load(file, envir = env)
        
      } else if (ext %in% c("tab", "tsv")) {
        nm <- tools::file_path_sans_ext(basename(file))
        if (overwrite || !exists(nm, envir = env, inherits = FALSE)) {
          assign(
            nm,
            utils::read.table(file, header = TRUE, sep = "\t"),
            envir = env
          )
        }
        
      } else if (ext == "csv") {
        nm <- tools::file_path_sans_ext(basename(file))
        if (overwrite || !exists(nm, envir = env, inherits = FALSE)) {
          assign(
            nm,
            utils::read.csv(file, header = TRUE),
            envir = env
          )
        }
        
      } else if (ext == "txt") {
        nm <- tools::file_path_sans_ext(basename(file))
        if (overwrite || !exists(nm, envir = env, inherits = FALSE)) {
          assign(
            nm,
            utils::read.table(file, header = TRUE),
            envir = env
          )
        }
        
      } else if (ext == "json") {
        nm <- tools::file_path_sans_ext(basename(file))
        if (!requireNamespace("jsonlite", quietly = TRUE)) {
          stop("jsonlite package required to load JSON datasets.")
        }
        if (overwrite || !exists(nm, envir = env, inherits = FALSE)) {
          assign(
            nm,
            jsonlite::fromJSON(file),
            envir = env
          )
        }
        
      } else if (ext == "r") {
        sys.source(file, envir = env)
        
      } else {
        return(character(0))
      }
      
      after <- ls(envir = env, all.names = TRUE)
      setdiff(after, before)
    }, error = function(e) {
      message("Failed to load file ", file, " : ", e$message)
      character(0)
    })
    
    out
  }
  
  load_from_source_package <- function(pkg, pkg_source_path, env, overwrite = FALSE) {
    if (is.null(pkg_source_path) || !nzchar(pkg_source_path)) {
      return(character(0))
    }
    
    data_path <- file.path(pkg_source_path, "data")
    if (!dir.exists(data_path)) {
      return(character(0))
    }
    
    files <- list.files(data_path, full.names = TRUE)
    out <- character()
    
    for (file in files) {
      new_objs <- load_from_file(file, env = env, overwrite = overwrite)
      out <- c(out, new_objs)
    }
    
    unique(out)
  }
  
  load_from_installed_package <- function(pkg, env, overwrite = FALSE) {
    if (!requireNamespace(pkg, quietly = TRUE)) {
      message("Package ", pkg, " is not installed; skipping installed datasets.")
      return(character(0))
    }
    
    info <- utils::data(package = pkg)$results
    if (is.null(info) || NROW(info) == 0L) {
      return(character(0))
    }
    
    item_col <- intersect(c("Item", "item"), colnames(info))
    if (length(item_col) == 0L) {
      message("Could not identify dataset column for package: ", pkg)
      return(character(0))
    }
    item_col <- item_col[[1]]
    
    dataset_items <- unique(info[, item_col])
    out <- character()
    
    for (dataset in dataset_items) {
      if (!overwrite && exists(dataset, envir = env, inherits = FALSE)) {
        out <- c(out, paste0(pkg, "::", dataset))
        next
      }
      
      before <- ls(envir = env, all.names = TRUE)
      
      ok <- tryCatch({
        utils::data(list = dataset, package = pkg, envir = env)
        TRUE
      }, error = function(e) {
        message("Failed to load dataset ", pkg, "::", dataset, " : ", e$message)
        FALSE
      })
      
      if (isTRUE(ok)) {
        after <- ls(envir = env, all.names = TRUE)
        new_objs <- setdiff(after, before)
        
        # In most cases utils::data() creates an object with the dataset name,
        # but record actual new objects if names differ.
        if (exists(dataset, envir = env, inherits = FALSE)) {
          out <- c(out, paste0(pkg, "::", dataset))
        } else if (length(new_objs) > 0L) {
          out <- c(out, paste0(pkg, "::", new_objs))
        }
      }
    }
    
    unique(out)
  }
  
  for (pkg in unique(pkg_names)) {
    pkg_loaded <- character()
    
    # 1) Try source package data first if a source path is provided for this package
    pkg_source_path <- NULL
    if (!is.null(source_paths) && pkg %in% names(source_paths)) {
      pkg_source_path <- source_paths[[pkg]]
    }
    
    source_loaded <- load_from_source_package(
      pkg = pkg,
      pkg_source_path = pkg_source_path,
      env = env,
      overwrite = overwrite
    )
    pkg_loaded <- c(pkg_loaded, source_loaded)
    
    # 2) Also try installed package datasets
    installed_loaded <- load_from_installed_package(
      pkg = pkg,
      env = env,
      overwrite = overwrite
    )
    pkg_loaded <- c(pkg_loaded, installed_loaded)
    
    loaded <- unique(c(loaded, pkg_loaded))
  }
  
  unique(loaded)
}

#' Remove New Globals from Environment Based on Initial Snapshot
#'
#' This function removes objects from an environment that were added after
#' a recorded initial state. Optionally filters by pattern.
#'
#' @param env Environment to clean.
#' @param initial_state Character vector of object names before setup.
#' @param patterns Optional regex patterns to filter objects.
#' @param dry_run Logical. If TRUE, only prints what would be removed.
#'
#' @return Invisibly returns a character vector of removed object names.
#'
#' @keywords internal
remove_new_globals <- function(env, initial_state, patterns = NULL, dry_run = FALSE) {
  current_state <- ls(envir = env, all.names = TRUE)
  new_objs <- setdiff(current_state, initial_state)
  
  if (!is.null(patterns)) {
    new_objs <- unique(unlist(lapply(patterns, function(p) grep(p, new_objs, value = TRUE))))
  }
  
  if (length(new_objs) == 0) {
    message("No new globals detected.")
    return(invisible(character()))
  }
  
  if (dry_run) {
    message("New globals detected (dry run):")
    print(new_objs)
  } else {
    rm(list = new_objs, envir = env)
    message("Removed new globals:")
    print(new_objs)
  }
  
  invisible(new_objs)
}



#' Clean up cov environment and return NULL
#'
#' This internal utility function is used to clean up the cov environment
#' by removing all objects and optionally displaying a message before returning `NULL`.
#' It is primarily used within `run_coverage_base()` to ensure consistent cleanup
#' in error-handling scenarios.
#'
#' @param msg Optional character string. If provided, the message will be printed before cleanup.
#' @param env A character string specifying the covr env
#'
#' @return NULL (invisibly)
#'
#' @keywords internal
cleanup_and_return_null <- function(msg = NULL, 
                                    env = cov_env) {
  if (!is.null(msg)) message(msg)
  
  # --- NEW: Snapshot coverage context BEFORE clearing env ---
  pkg_root_val <- try({
    if (base::exists("pkg_root", envir = env, inherits = FALSE)) {
      base::get("pkg_root", envir = env, inherits = FALSE)
    } else {
      NULL
    }
  }, silent = TRUE)
  if (inherits(pkg_root_val, "try-error")) pkg_root_val <- NULL
  
  work_dir_val <- try({
    if (base::exists("work_dir", envir = env, inherits = FALSE)) {
      base::get("work_dir", envir = env, inherits = FALSE)
    } else {
      NULL
    }
  }, silent = TRUE)
  if (inherits(work_dir_val, "try-error")) work_dir_val <- NULL
  # --- END NEW ---
  
  # 1) Clear all bindings from the provided env (as before)
  to_rm <- tryCatch(
    base::ls(envir = env, all.names = TRUE),
    error = function(e) base::ls(envir = env) # fall back if a stub doesn't accept all.names
  )
  if (length(to_rm)) {
    for (sym in to_rm) {
      if (base::exists(sym, envir = env, inherits = FALSE)) {
        base::rm(list = sym, envir = env)
      }
    }
  }
  
  # 2) ALWAYS try to remove the temporary workspace referenced by work_dir_val
  try({
    if (is.character(work_dir_val) && nzchar(work_dir_val)) {
      td_norm <- normalizePath(tempdir(), winslash = "/", mustWork = FALSE)
      wd_norm <- tryCatch(
        normalizePath(work_dir_val, winslash = "/", mustWork = FALSE),
        error = function(e) work_dir_val
      )
      # Only remove if the workspace lives under the session tempdir()
      if (startsWith(paste0(wd_norm, "/"), paste0(td_norm, "/")) && dir.exists(work_dir_val)) {
        unlink(work_dir_val, recursive = TRUE, force = TRUE)
      }
    }
  }, silent = TRUE)
  
  # 3) Orphan cleanup with a safer scope
  try({
    safe_roots <- character(0)
    
    # 3a) Preferred root: pkg_root_val (from snapshot)
    pkg_root <- NULL
    if (is.character(pkg_root_val) && length(pkg_root_val) == 1L && nzchar(pkg_root_val) && dir.exists(pkg_root_val)) {
      pkg_root <- tryCatch(
        normalizePath(pkg_root_val, winslash = "/", mustWork = FALSE),
        error = function(e) NULL
      )
    }
    if (!is.null(pkg_root)) {
      safe_roots <- c(safe_roots, pkg_root)
    }
    
    # 3b) Include getwd() only if it is a descendant of pkg_root (safe & complete)
    cwd_norm <- tryCatch(
      normalizePath(getwd(), winslash = "/", mustWork = FALSE),
      error = function(e) NULL
    )
    if (!is.null(pkg_root) && !is.null(cwd_norm)) {
      pkg_root_slash <- paste0(pkg_root, "/")
      cwd_slash <- paste0(cwd_norm, "/")
      if (startsWith(cwd_slash, pkg_root_slash)) {
        safe_roots <- c(safe_roots, cwd_norm)
      }
    }
    
    # 3c) Include work_dir_val if it is under tempdir()
    td_norm <- tryCatch(
      normalizePath(tempdir(), winslash = "/", mustWork = FALSE),
      error = function(e) NULL
    )
    if (is.character(work_dir_val) && nzchar(work_dir_val)) {
      wd_norm <- tryCatch(
        normalizePath(work_dir_val, winslash = "/", mustWork = FALSE),
        error = function(e) work_dir_val
      )
      if (!is.null(td_norm) &&
          startsWith(paste0(wd_norm, "/"), paste0(td_norm, "/")) &&
          dir.exists(work_dir_val)) {
        safe_roots <- c(safe_roots, wd_norm)
      }
    }
    
    safe_roots <- unique(safe_roots[!is.na(safe_roots) & nzchar(safe_roots)])
    
    if (!length(safe_roots)) {
      message("No safe roots found for orphan cleanup; skipping.")
    } else {
      # Known offenders from MASS and survival
      known_names <- c("polr.pdf", "silhouette-ex.ps")
      
      # Extensions removable only in ephemeral roots (e.g., temp work dirs)
      extensions_to_remove <- c(".pdf", ".ps")
      pat_ext <- paste0("\\", extensions_to_remove, "$", collapse = "|")
      
      for (root in safe_roots) {
        files_in_root <- tryCatch(
          list.files(root, full.names = TRUE),
          error = function(e) character(0)
        )
        
        # Always allow deleting the known filenames inside pkg_root or its subdir
        by_name <- files_in_root[basename(files_in_root) %in% known_names]
        
        # Allow deleting any *.pdf/*.ps only if the root is under tempdir()
        by_ext <- character(0)
        if (!is.null(td_norm) &&
            startsWith(paste0(root, "/"), paste0(td_norm, "/"))) {
          by_ext <- files_in_root[grepl(pat_ext, files_in_root)]
        }
        
        files_to_remove <- unique(c(by_name, by_ext))
        
        if (length(files_to_remove)) {
          invisible(vapply(files_to_remove, function(file) {
            if (file.exists(file)) {
              try(file.remove(file), silent = TRUE)
            }
            TRUE
          }, logical(1)))
        }
      }
    }
  }, silent = TRUE)
  
  # 4) Exit silently
  invisible(NULL)
}




#' @title Find R project root directory
#' @description Searches upward from a starting path to locate the root of an R package or project.
#' @param start character, starting path (e.g., test directory or script path)
#' @return character path to project root, or NULL if not found
#' @keywords internal
find_project_root <- function(start = getwd()) {
  
  current <- normalizePath(start, winslash = "/", mustWork = FALSE)
  while (TRUE) {
    if (file.exists(file.path(current, "DESCRIPTION")) ||
        file.exists(file.path(current, ".Rproj")) ||
        file.exists(file.path(current, ".git"))) {
      return(current)
    }
    parent <- dirname(current)
    if (parent == current) break
    current <- parent
  }
  return(NULL)
}


#' get functions with no tests
#'
#' @param mapping_df - data frame with source files and test files
#'
#' @return no_tests_df - data frame with source files with no tests
#' 
#' @keywords internal
get_function_no_tests <- function(mapping_df) {
  # Filter rows where test_file is NA
  no_tests_df <- subset(mapping_df, is.na(test_file))
  
  # Check if filtered_df is empty
  # If no rows are filtered, create a single-row data frame with the comment
  if (nrow(no_tests_df) == 0) {
    no_tests_df <- data.frame(
      source_file = NA,
      test_file = NA,
      comment = "all functions have at least 1 test",
      stringsAsFactors = FALSE
    )
  } else {
    # Add a comment to the filtered rows
    no_tests_df$comment <- "No tests found"
  }
  
  return(no_tests_df)
}


#' Create an empty covr list for fallback
#'
#' This internal utility function returns a default coverage list structure
#' used when test execution or mapping fails. It ensures consistent output
#' format for downstream processing even when no valid test results are available.
#'
#' @param pkg_name A character string representing the name of the package.
#' @param error_message A character string describing the reason for the fallback.
#'
#' @return A named list containing default coverage values, the provided error message,
#' and placeholder notes.
#'
#' @keywords internal
create_empty_covr_list <- function(pkg_name, error_message) {
  list(
    total_cov = 0,
    res_cov = list(
      name = pkg_name,
      coverage = list(
        filecoverage = matrix(0, nrow = 1, dimnames = list("No functions tested")),
        totalcoverage = 0
      ),
      errors = paste(error_message),
      notes = NA
    )
  )
}


#' Safely unload a package to avoid coverage conflicts
#'
#' This internal function attempts to unload a specified package if it is currently loaded.
#' It is used to prevent coverage conflicts when reloading the package in a custom environment.
#'
#' @param pkg Character string. The name of the package to unload.
#' @keywords internal
#' @return NULL (invisibly). Messages are printed to indicate success or failure.
safe_unload_package <- function(pkg) {
  tryCatch({
    if (pkg %in% loadedNamespaces()) {
      message(paste0("Unloading ", pkg, " package to avoid coverage conflict..."))
      unloadNamespace(pkg)
    }
  }, error = function(e) {
    message(paste0("Warning: Failed to unload ", pkg, " package. Proceeding anyway."))
    message("Details: ", e$message)
  })
}


#' Load package objects into a coverage environment
#'
#' This internal function loads all objects from a package namespace into a specified environment.
#' It is used by `create_covr_list_no_skip()` to prepare the environment for coverage analysis.
#'
#' @param pkg_source_path Path to the package source directory.
#' @param pkg Name of the package to load.
#' @param cov_env Environment into which package objects will be loaded.
#'
#' @return Invisibly returns TRUE if successful, otherwise stops with an error.
#' @keywords internal
load_package_into_env <- function(pkg_source_path, pkg, cov_env) {
  tryCatch({
    pkgload::load_all(pkg_source_path)
    pkg_all <- ls(envir = asNamespace(pkg), all.names = TRUE)
    for (obj in pkg_all) {
      tryCatch({
        assign(obj, get(obj, envir = asNamespace(pkg)), envir = cov_env)
      }, error = function(e) {
        message(paste0("Failed to assign object ", obj, " : ", e$message))
      })
    }
    invisible(TRUE)
  }, error = function(e) {
    stop(paste0("Failed to load package ", pkg, " : ", e$message))
  })
}


#' Load Testthat Helper Files into Environment
#'
#' This function searches for all files in the `tests/testthat` directory
#' that match the pattern `helper-*.R` and sources them into the provided environment.
#' This is useful when running coverage with `covr::environment_coverage()` to ensure
#' helper functions are available during test execution.
#'
#' @param env An environment into which the helper files should be sourced.
#'            Typically this is the same environment used for coverage, e.g., `cov_env`.
#' @param helper_dir Path to the directory containing helper files. Defaults to `"tests/testthat"`.
#'
#' @return Invisibly returns `TRUE` if all files are sourced successfully.
#' @keywords internal
load_test_helpers <- function(env, helper_dir = "tests/testthat") {
  helper_files <- list.files(helper_dir, pattern = "^helper.*\\.R$", full.names = TRUE)
  for (file in helper_files) {
    tryCatch({
      sys.source(file, envir = env)
    }, error = function(e) {
      message(paste("Failed to source helper file:", file, ":", e$message))
    })
  }
  invisible(TRUE)
}

#' Load Base R packages into an isolated coverage environment
#'
#' Populate an isolated environment (typically created with
#' `new.env(parent = baseenv())`) with objects needed to mimic the
#' standard Base R package surface during coverage or test evaluation.
#'
#' The `base` package itself is already available via `baseenv()` and is
#' therefore not copied into `env`. The function instead:
#'
#' - loads datasets from the `datasets` package into `env`, and
#' - copies exported objects from other Base R packages into `env`
#'   (e.g. `stats`, `utils`, `graphics`, `grDevices`, `methods`, etc.).
#'
#' This is useful when tests are evaluated in a deliberately isolated
#' environment whose parent is `baseenv()`, because objects from packages
#' such as `stats` (e.g. `rnorm`) and `datasets` (e.g. `mtcars`, `iris`)
#' are not automatically visible there.
#'
#' @param env An environment to populate.
#' @param overwrite Logical; if `TRUE`, overwrite objects already present in
#'   `env`. If `FALSE` (default), existing objects are left unchanged.
#'
#' @return A named list with two character vectors:
#' \describe{
#'   \item{exports_loaded}{Exported objects copied from Base R packages
#'   other than `base` and `datasets`.}
#'   \item{datasets_loaded}{Dataset objects loaded from the `datasets`
#'   package.}
#' }
#'
#' @details
#' The Base R package set handled here is:
#' \itemize{
#'   \item `base`
#'   \item `compiler`
#'   \item `datasets`
#'   \item `graphics`
#'   \item `grDevices`
#'   \item `grid`
#'   \item `methods`
#'   \item `parallel`
#'   \item `splines`
#'   \item `stats`
#'   \item `stats4`
#'   \item `tcltk`
#'   \item `tools`
#'   \item `utils`
#' }
#'
#' Because `base` is already reachable through `baseenv()`, it is excluded
#' from the explicit copy step.
#'
#' @keywords internal
load_base_r_packages_into_env <- function(env,
                                          overwrite = FALSE) {
  checkmate::assert_environment(env, .var.name = "env")
  checkmate::assert_flag(overwrite, .var.name = "overwrite")
  
  # Base R package set (excluding "translations" per user request)
  base_r_pkgs <- c(
    "base",
    "compiler",
    "datasets",
    "graphics",
    "grDevices",
    "grid",
    "methods",
    "parallel",
    "splines",
    "stats",
    "stats4",
    "tcltk",
    "tools",
    "utils"
  )
  
  # `base` is already available via parent = baseenv()
  export_pkgs <- setdiff(base_r_pkgs, c("base", "datasets"))
  
  exports_loaded <- load_package_exports_into_env(
    pkg_names = export_pkgs,
    env = env,
    overwrite = overwrite
  )
  
  datasets_loaded <- load_package_datasets(
    pkg_names = "datasets",
    env = env,
    source_paths = NULL,
    overwrite = overwrite
  )
  
  list(
    exports_loaded = exports_loaded,
    datasets_loaded = datasets_loaded
  )
}


#' Load exported objects from installed packages into an environment
#'
#' Copy exported objects (typically functions, but also exported constants or
#' other objects) from one or more installed package namespaces into a target
#' environment.
#'
#' This is useful when evaluating code in an isolated environment, such as one
#' created with `new.env(parent = baseenv())`, where package exports that are
#' normally available on the search path (for example from `stats` or `utils`)
#' are not automatically visible.
#'
#' The function does not attach packages to the search path. Instead, it copies
#' exported objects directly into `env`.
#'
#' @param pkg_names Character vector of package names whose exported objects
#'   should be copied into `env`.
#' @param env Environment into which exported objects should be assigned.
#' @param overwrite Logical; if `TRUE`, overwrite objects already present in
#'   `env`. If `FALSE` (default), existing objects are left unchanged.
#'
#' @return A character vector of fully qualified object names that were loaded,
#'   in the form `"pkg::name"`.
#'
#' @details
#' For each package in `pkg_names`, this function:
#' \enumerate{
#'   \item checks whether the package namespace is available,
#'   \item retrieves the package's exported names,
#'   \item obtains each exported object with `getExportedValue()`,
#'   \item assigns the object into `env`.
#' }
#'
#' Packages that are not installed or cannot be loaded are skipped with a
#' message.
#' 
#' @keywords internal
load_package_exports_into_env <- function(pkg_names,
                                          env,
                                          overwrite = FALSE) {
  checkmate::assert_character(
    pkg_names,
    min.len = 1,
    any.missing = FALSE,
    .var.name = "pkg_names"
  )
  checkmate::assert_environment(env, .var.name = "env")
  checkmate::assert_flag(overwrite, .var.name = "overwrite")
  
  loaded <- character()
  
  for (pkg in unique(pkg_names)) {
    if (!requireNamespace(pkg, quietly = TRUE)) {
      message("Package ", pkg, " is not installed or not available; skipping export load.")
      next
    }
    
    exports <- tryCatch(
      getNamespaceExports(pkg),
      error = function(e) {
        message("Failed to retrieve exports for package ", pkg, " : ", e$message)
        character(0)
      }
    )
    
    if (length(exports) == 0L) {
      next
    }
    
    for (nm in exports) {
      if (!overwrite && exists(nm, envir = env, inherits = FALSE)) {
        next
      }
      
      ok <- tryCatch({
        obj <- getExportedValue(pkg, nm)
        assign(nm, obj, envir = env)
        TRUE
      }, error = function(e) {
        message("Failed to load export ", pkg, "::", nm, " : ", e$message)
        FALSE
      })
      
      if (isTRUE(ok)) {
        loaded <- c(loaded, paste0(pkg, "::", nm))
      }
    }
  }
  
  unique(loaded)
}

test_that("loads datasets using utils::data() when installed package datasets are available", {
  dummy_env <- new.env(parent = emptyenv())
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  mockery::stub(ld, "requireNamespace", function(pkg, quietly = TRUE) TRUE)
  
  testthat::local_mocked_bindings(
    data = function(..., list = NULL, package = NULL, envir = NULL) {
      if (is.null(list)) {
        list(
          results = matrix(
            c("dataset1", "dataset2"),
            ncol = 1,
            dimnames = list(NULL, "Item")
          )
        )
      } else {
        assign(list, paste0("loaded_", list), envir = envir)
        invisible(NULL)
      }
    },
    .package = "utils"
  )
  
  result <- ld(
    pkg_names = "fakepkg",
    env = dummy_env,
    source_paths = NULL,
    overwrite = FALSE
  )
  
  expect_setequal(result, c("fakepkg::dataset1", "fakepkg::dataset2"))
  expect_equal(get("dataset1", envir = dummy_env, inherits = FALSE), "loaded_dataset1")
  expect_equal(get("dataset2", envir = dummy_env, inherits = FALSE), "loaded_dataset2")
})






test_that("run_covr_modes handles testit framework correctly (DT-style -> standard path)", {
  cov_env <- new.env(parent = baseenv())
  
  # --- Create DT-style temp package layout: tests/test-all.R, NO tests/testthat/ ---
  pkg_dir <- withr::local_tempdir()
  dir.create(file.path(pkg_dir, "tests"), recursive = TRUE, showWarnings = FALSE)
  writeLines(c("library(testit)", "test_pkg('mockpkg')"),
             con = file.path(pkg_dir, "tests", "test-all.R"))
  
  # Call counters to validate routing
  coverage_calls <- 0L
  stf_calls      <- 0L
  
  # --- Mocks ------------------------------------------------------------------
  mock_initialise_covr_env <- function(pkg_source_path, pkg, framework, ...) {
    assign("pkg_path",   pkg_source_path, envir = cov_env)
    assign("package",    pkg,             envir = cov_env)
    assign("framework",  framework,       envir = cov_env)
    cov_env
  }
  
  # Minimal detector output: we’ll force a 'testit' framework
  mock_check_pkg_tests_and_snaps <- function(pkg_source_path, ...) {
    list(
      has_testthat         = FALSE,
      has_tinytest         = FALSE,
      has_data_table_tests = FALSE,
      has_BioG_test        = FALSE,
      has_RUnit_test       = FALSE,
      has_testit           = TRUE,
      has_tests_base       = FALSE
    )
  }
  
  # Make STANDARD coverage succeed so no fallback is needed
  mock_run_coverage <- function(pkg_source_path, covr_timeout, ...) {
    coverage_calls <<- coverage_calls + 1L
    list(
      total_cov = 42,
      res_cov = list(coverage = list(filecoverage = matrix(1)))
    )
  }
  
  # STF fallback: SHOULD NOT be called in this scenario
  mock_run_covr_skip_stf <- function(pkg_source_path, test_pkg_data, cov_env, ...) {
    stf_calls <<- stf_calls + 1L
    list(
      total_cov = 99,
      res_cov = list(coverage = list(filecoverage = matrix(1)), notes = "STF coverage run")
    )
  }
  
  mock_get_pkg_desc    <- function(pkg_source_path, fields, ...) list(Package = "mockpkg", Version = "0.1.0")
  mock_unloadNamespace <- function(pkg, ...) invisible(NULL)
  mock_get_pkg_name    <- function(pkg_source_path, ...) "mockpkg"
  mock_detect_frameworks <- function(test_pkg_data, ...) list("testit")
  
  # --- Bind functions looked up INSIDE nested calls (package namespace) -------
  # Important: bind run_coverage here (not just stub on run_covr_modes)
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env            = cov_env,
    initialise_covr_env = mock_initialise_covr_env,
    run_covr_skip_stf   = mock_run_covr_skip_stf,
    run_coverage        = mock_run_coverage
  )
  
  # --- Stub top-level collaborators on run_covr_modes() -----------------------
  mockery::stub(run_covr_modes, "get_pkg_name",              mock_get_pkg_name)
  mockery::stub(run_covr_modes, "detect_frameworks",         mock_detect_frameworks)
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", mock_check_pkg_tests_and_snaps)
  mockery::stub(run_covr_modes, "get_pkg_desc",              mock_get_pkg_desc)
  mockery::stub(run_covr_modes, "unloadNamespace",           mock_unloadNamespace)
  # Do NOT stub run_coverage here; it won’t reach nested calls. We already bound it above.
  
  expect_true(is.environment(cov_env))
  
  # --- Execute on a REAL path and (optionally) assert the message -------------
  # If your decider emits the DT-style message, capture and assert it:
  expect_message(
    object = { result <- run_covr_modes(pkg_dir) },
    regexp = "DT-style",
    all = FALSE
  )
  
  # --- Assertions -------------------------------------------------------------
  expect_identical(coverage_calls, 1L)   # standard path used
  expect_identical(stf_calls, 0L)        # fallback NOT used
  expect_equal(result$total_cov, 42)
  # Depending on your get_test_framework_type mapping:
  expect_true(grepl("non-standard testing framework - testit", result$test_framework_type))
})




test_that("loads various file types from source data directory", {
  dummy_env <- new.env(parent = emptyenv())
  
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  # Source package branch should run from source_paths.
  # Installed-package branch should be skipped for mockpkg,
  # but jsonlite must still appear available for JSON loading.
  mockery::stub(
    ld,
    "requireNamespace",
    function(pkg, quietly = TRUE) {
      if (identical(pkg, "jsonlite")) {
        TRUE
      } else if (identical(pkg, "mockpkg")) {
        FALSE
      } else {
        TRUE
      }
    }
  )
  
  # Simulate source package data directory existing
  mockery::stub(ld, "dir.exists", function(path) TRUE)
  
  # Simulate files in /mock/path/data
  mockery::stub(
    ld,
    "list.files",
    function(path, full.names = TRUE) {
      c(
        "/mock/path/data/data1.rda",
        "/mock/path/data/data2.csv",
        "/mock/path/data/data3.tsv",
        "/mock/path/data/data4.txt",
        "/mock/path/data/data5.json",
        "/mock/path/data/data6.R"
      )
    }
  )
  
  # Stub extension detection exactly as used in the function
  mockery::stub(
    ld,
    "tools::file_ext",
    function(file) sub(".*\\.([^.]+)$", "\\1", file)
  )
  
  # Stub file_path_sans_ext exactly as used in the function
  mockery::stub(
    ld,
    "tools::file_path_sans_ext",
    function(file) sub("\\.[^.]*$", "", basename(file))
  )
  
  # Stub file loaders used inside load_from_file()
  mockery::stub(
    ld,
    "load",
    function(file, envir) {
      assign("data1", "rda_loaded", envir = envir)
      invisible("data1")
    }
  )
  
  mockery::stub(
    ld,
    "utils::read.csv",
    function(file, header = TRUE) {
      data.frame(a = 1)
    }
  )
  
  mockery::stub(
    ld,
    "utils::read.table",
    function(file, header = TRUE, sep = "") {
      data.frame(b = 2)
    }
  )
  
  mockery::stub(
    ld,
    "jsonlite::fromJSON",
    function(file) {
      list(c = 3)
    }
  )
  
  mockery::stub(
    ld,
    "sys.source",
    function(file, envir) {
      assign("data6", "R_loaded", envir = envir)
      invisible(NULL)
    }
  )
  
  result <- ld(
    pkg_names = "mockpkg",
    env = dummy_env,
    source_paths = c(mockpkg = "/mock/path"),
    overwrite = FALSE
  )
  
  expect_setequal(result, c("data1", "data2", "data3", "data4", "data5", "data6"))
  
  expect_equal(get("data1", envir = dummy_env, inherits = FALSE), "rda_loaded")
  expect_equal(get("data2", envir = dummy_env, inherits = FALSE)$a, 1)
  expect_equal(get("data3", envir = dummy_env, inherits = FALSE)$b, 2)
  expect_equal(get("data4", envir = dummy_env, inherits = FALSE)$b, 2)
  expect_equal(get("data5", envir = dummy_env, inherits = FALSE)$c, 3)
  expect_equal(get("data6", envir = dummy_env, inherits = FALSE), "R_loaded")
})





test_that("load_package_datasets returns empty character when package is not installed", {
  dummy_env <- new.env(parent = emptyenv())
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  # Simulate package not installed
  mockery::stub(
    ld,
    "requireNamespace",
    function(pkg, quietly = TRUE) FALSE
  )
  
  result <- ld(
    pkg_names = "fakepkg",
    env = dummy_env,
    source_paths = NULL,
    overwrite = FALSE
  )
  
  expect_type(result, "character")
  expect_length(result, 0)
  expect_equal(result, character(0))
})




test_that("cleanup_and_return_null finds project root and removes files by extension", {
  # Create a dummy environment
  dummy_env <- new.env()
  dummy_env$x <- 1
  dummy_env$y <- 2
  
  # Stub getwd and normalizePath
  mockery::stub(cleanup_and_return_null, "getwd", function() "/mock/project")
  mockery::stub(cleanup_and_return_null, "normalizePath", function(path) path)
  
  # Simulate files in root directory
  mock_files <- c("/mock/project/report.pdf", "/mock/project/diagram.ps", "/mock/project/readme.txt")
  
  # Stub list.files to return mock files
  mockery::stub(cleanup_and_return_null, "list.files", function(path, full.names) mock_files)
  
  # Stub file.exists to return TRUE for all mock files
  mockery::stub(cleanup_and_return_null, "file.exists", function(path) TRUE)
  
  # Track calls to file.remove
  removed_files <- character()
  mockery::stub(cleanup_and_return_null, "file.remove", function(path) {
    removed_files <<- c(removed_files, path)
    TRUE
  })
  
  # Stub ls and rm for environment cleanup
  mockery::stub(cleanup_and_return_null, "ls", function(envir) c("x", "y"))
  mockery::stub(cleanup_and_return_null, "rm", function(list, envir) {
    for (var in list) rm(list = var, envir = envir)
  })
  
  # Call function
  result <- cleanup_and_return_null(env = dummy_env)
  
  # Assertions
  expect_false(exists("x", envir = dummy_env))
  expect_false(exists("y", envir = dummy_env))
  
  # Check that only .pdf and .ps files were removed
  expect_true(all(grepl("\\.(pdf|ps)$", removed_files)))
})


test_that("cleanup_and_return_null returns NULL when project root not found", {
  dummy_env <- new.env()
  dummy_env$x <- 1
  
  mockery::stub(cleanup_and_return_null, "getwd", function() "/mock/project")
  mockery::stub(cleanup_and_return_null, "file.exists", function(path) FALSE)
  mockery::stub(cleanup_and_return_null, "normalizePath", function(path) path)
  mockery::stub(cleanup_and_return_null, "ls", function(envir) "x")
  mockery::stub(cleanup_and_return_null, "rm", function(list, envir) {
    for (var in list) rm(list = var, envir = envir)
  })
  
  result <- cleanup_and_return_null(env = dummy_env)
  
  expect_false(exists("x", envir = dummy_env))
})



test_that("cleanup removes work_dir under tempdir() when it exists", {
  # --- Arrange ---
  env <- new.env(parent = emptyenv())
  work_dir <- "/TD/work"
  assign("work_dir", work_dir, envir = env)
  
  # Prevent initial env wipe from deleting 'work_dir'
  mockery::stub(cleanup_and_return_null, "base::ls",
                function(envir, all.names = TRUE) character(0))
  mockery::stub(cleanup_and_return_null, "base::rm", function(...) NULL)
  
  # Make tempdir and normalizePath deterministic
  mockery::stub(cleanup_and_return_null, "tempdir", function() "/TD")
  mockery::stub(cleanup_and_return_null, "normalizePath",
                function(path, winslash = "/", mustWork = FALSE) {
                  if (identical(path, "/TD")) return("/TD")           # td_norm
                  if (identical(path, work_dir)) return("/TD/work")   # wd_norm
                  path                                                 # fallback for getwd()
                })
  
  # Ensure dir.exists reports TRUE for the work_dir
  mockery::stub(cleanup_and_return_null, "dir.exists",
                function(p) identical(p, work_dir))
  
  # Capture unlink calls
  fake_unlink <- mockery::mock(TRUE)
  mockery::stub(cleanup_and_return_null, "unlink", fake_unlink)
  
  # Neutralize project-root cleanups
  mockery::stub(cleanup_and_return_null, "getwd", function() "/ROOT")
  mockery::stub(cleanup_and_return_null, "list.files", function(...) character(0))
  
  # --- Act & Assert ---
  expect_silent(cleanup_and_return_null(NULL, env = env))
  mockery::expect_called(fake_unlink, 1)
  mockery::expect_args(fake_unlink, 1, work_dir, recursive = TRUE, force = TRUE)
})



test_that("no removal when work_dir under tempdir() but directory does not exist", {
  env <- new.env(parent = emptyenv())
  work_dir <- "/TD/work"
  assign("work_dir", work_dir, envir = env)
  
  mockery::stub(cleanup_and_return_null, "base::ls",
                function(envir, all.names = TRUE) character(0))
  mockery::stub(cleanup_and_return_null, "base::rm", function(...) NULL)
  
  mockery::stub(cleanup_and_return_null, "tempdir", function() "/TD")
  mockery::stub(cleanup_and_return_null, "normalizePath",
                function(path, winslash = "/", mustWork = FALSE) {
                  if (identical(path, "/TD")) return("/TD")
                  if (identical(path, work_dir)) return("/TD/work")
                  path
                })
  
  # dir.exists is FALSE → do not unlink
  mockery::stub(cleanup_and_return_null, "dir.exists", function(p) FALSE)
  
  fake_unlink <- mockery::mock(TRUE)
  mockery::stub(cleanup_and_return_null, "unlink", fake_unlink)
  
  mockery::stub(cleanup_and_return_null, "getwd", function() "/ROOT")
  mockery::stub(cleanup_and_return_null, "list.files", function(...) character(0))
  
  expect_message(
    cleanup_and_return_null(NULL, env = env), 
    "No safe roots found for orphan cleanup; skipping.")
  mockery::expect_called(fake_unlink, 0)
})

test_that("no removal when work_dir is outside tempdir()", {
  env <- new.env(parent = emptyenv())
  work_dir <- "/OTHER/work"
  assign("work_dir", work_dir, envir = env)
  
  mockery::stub(cleanup_and_return_null, "base::ls",
                function(envir, all.names = TRUE) character(0))
  mockery::stub(cleanup_and_return_null, "base::rm", function(...) NULL)
  
  mockery::stub(cleanup_and_return_null, "tempdir", function() "/TD")
  mockery::stub(cleanup_and_return_null, "normalizePath",
                function(path, winslash = "/", mustWork = FALSE) {
                  if (identical(path, "/TD")) return("/TD")
                  if (identical(path, work_dir)) return("/OTHER/work")  # NOT under /TD
                  path
                })
  
  mockery::stub(cleanup_and_return_null, "dir.exists", function(p) TRUE)
  
  fake_unlink <- mockery::mock(TRUE)
  mockery::stub(cleanup_and_return_null, "unlink", fake_unlink)
  
  mockery::stub(cleanup_and_return_null, "getwd", function() "/ROOT")
  mockery::stub(cleanup_and_return_null, "list.files", function(...) character(0))
  
  expect_message(
    cleanup_and_return_null(NULL, env = env), 
    "No safe roots found for orphan cleanup; skipping.")
  mockery::expect_called(fake_unlink, 0)
})



test_that("no removal when normalizePath(work_dir) errors (fallback to raw path)", {
  env <- new.env(parent = emptyenv())
  work_dir <- "/TD/work"
  assign("work_dir", work_dir, envir = env)
  
  mockery::stub(cleanup_and_return_null, "base::ls",
                function(envir, all.names = TRUE) character(0))
  mockery::stub(cleanup_and_return_null, "base::rm", function(...) NULL)
  
  mockery::stub(cleanup_and_return_null, "tempdir", function() "/TD")
  mockery::stub(cleanup_and_return_null, "normalizePath",
                function(path, winslash = "/", mustWork = FALSE) {
                  if (identical(path, "/TD")) return("/TD")     # td_norm
                  if (identical(path, work_dir)) stop("boom")   # wd_norm falls back to raw path
                  path
                })
  
  # Even if startsWith("/TD/work", "/TD") could be TRUE via fallback,
  # gate unlink by making dir.exists FALSE so we assert no unlink occurred.
  mockery::stub(cleanup_and_return_null, "dir.exists", function(p) FALSE)
  
  fake_unlink <- mockery::mock(TRUE)
  mockery::stub(cleanup_and_return_null, "unlink", fake_unlink)
  
  mockery::stub(cleanup_and_return_null, "getwd", function() "/ROOT")
  mockery::stub(cleanup_and_return_null, "list.files", function(...) character(0))
  
  expect_message(
    cleanup_and_return_null(NULL, env = env), 
    "No safe roots found for orphan cleanup; skipping.")
  mockery::expect_called(fake_unlink, 0)
})



test_that("returns functions with no tests", {
  mock_df <- data.frame(
    source_file = c("file1.R", "file2.R"),
    test_file = c(NA, "test_file2.R"),
    stringsAsFactors = FALSE
  )
  
  result <- get_function_no_tests(mock_df)
  
  expect_equal(nrow(result), 1)
  expect_equal(result$source_file, "file1.R")
  expect_equal(result$comment, "No tests found")
})

test_that("returns comment when all functions have tests", {
  mock_df <- data.frame(
    source_file = c("file1.R", "file2.R"),
    test_file = c("test_file1.R", "test_file2.R"),
    stringsAsFactors = FALSE
  )
  
  result <- get_function_no_tests(mock_df)
  
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$source_file))
  expect_equal(result$comment, "all functions have at least 1 test")
})

test_that("handles empty input gracefully", {
  mock_df <- data.frame(
    source_file = character(0),
    test_file = character(0),
    stringsAsFactors = FALSE
  )
  
  result <- get_function_no_tests(mock_df)
  
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$source_file))
  expect_equal(result$comment, "all functions have at least 1 test")
})




test_that("run_covr_modes handles standard testing framework with valid coverage", {
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mock_run_coverage <- function(pkg_source_path, covr_timeout, ...) {
    list(
      total_cov = 85,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("some_function")),
          totalcoverage = 85
        )
      )
    )
  }
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    run_coverage = mock_run_coverage
  )
  
  # Stub check_pkg_tests_and_snaps - has testthat
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", function(...) {
    list(has_testthat = TRUE, has_tests_base = FALSE, has_testit = FALSE)
  })
  
  # Stub get_pkg_desc
  mock_get_pkg_desc <- function(pkg_source_path, fields) {
    list(Package = "mockpkg", Version = "0.1.0")
  }
  mockery::stub(run_covr_modes, "get_pkg_desc", mock_get_pkg_desc)
  
  # Stub unloadNamespace
  mock_unloadNamespace <- function(pkg) {}
  mockery::stub(run_covr_modes, "unloadNamespace", mock_unloadNamespace)
  
  
  # Run
  result <- run_covr_modes(pkg_source_path)
  
  # Assertions
  expect_true(is.list(result))
  expect_equal(result$total_cov, 85)
})



test_that("run_covr_modes handles Bioconductor testing framework correctly", {
  local_cov_env <- new.env(parent = baseenv())
  
  mock_initialise_covr_env <- function(pkg_source_path, pkg, framework, ...) {
    assign("pkg_path", pkg_source_path, envir = local_cov_env)
    assign("package", pkg, envir = local_cov_env)
    assign("framework", framework, envir = local_cov_env)
    assign("work_dir", tempdir(), envir = local_cov_env)
    assign("pkg_root", pkg_source_path, envir = local_cov_env)
    local_cov_env
  }
  
  mock_run_covr_skip_bioc_nstf <- function(pkg_source_path, test_pkg_data, cov_env,
                                            framework = NULL, ...) {
    list(
      total_cov = 77,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("bioc_fun")),
          totalcoverage = 77
        ),
        notes = "Bioc NSTF run"
      ),
      tests_skipped = character(),
      tests_passing = character()
    )
  }
  
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env = local_cov_env,
    initialise_covr_env = mock_initialise_covr_env,
    run_covr_skip_bioc_nstf = mock_run_covr_skip_bioc_nstf
  )
  
  pkg_source_path <- file.path(tempdir(), "mockpkg_bioc")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mockery::stub(run_covr_modes, "get_pkg_name", function(path, ...) "mockpkg")
  mockery::stub(run_covr_modes, "get_pkg_desc",
                function(path, fields, ...) list(Package = "mockpkg", Version = "1.0.0")
  )
  mockery::stub(run_covr_modes, "get_test_metadata",
                function(...) {
                  list(
                    datetime = "2026-01-01 12:00:00",
                    executor = "test-user",
                    info = list(sys = list(
                      sysname = "Linux",
                      version = "5.15",
                      release = "Ubuntu",
                      machine = "x86_64"
                    ))
                  )
                }
  )
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps",
                function(path, ...) {
                  list(
                    has_testthat = FALSE,
                    has_testit = FALSE,
                    has_tests_base = FALSE,
                    has_tinytest = FALSE,
                    has_data_table_tests = FALSE,
                    has_RUnit_test = FALSE,
                    has_BioG_test = TRUE,
                    has_snaps = FALSE,
                    bioc_unit_tests_dir = "inst/unitTests",
                    n_golden_tests = 0L,
                    n_test_files = 1L
                  )
                }
  )
  mockery::stub(safe_unload_package, "loadedNamespaces", function() character(0))
  
  result <- run_covr_modes(pkg_source_path)
  
  expect_true(is.list(result))
  expect_equal(result$total_cov, 77)
  expect_equal(result$res_cov$coverage$totalcoverage, 77)
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "bioc_fun")
  expect_equal(
    result$test_framework_type,
    "non-standard testing framework - Bioconductor"
  )
  expect_equal(result$pkg_name, "mockpkg")
  expect_equal(result$pkg_ver, "1.0.0")
  expect_equal(result$executor, "test-user")
  expect_equal(result$sysname, "Linux")
})

test_that("run_covr_modes triggers fallback to run_covr_skip_stf when coverage is NA", {
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mock_run_coverage <- function(pkg_source_path, covr_timeout, ...) {
    list(
      total_cov = NA,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = NA,
          totalcoverage = NA
        )
      )
    )
  }
  
  mock_run_covr_skip_stf <- function(pkg_source_path, test_pkg_data, cov_env) {
    list(
      total_cov = 42,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("fallback_function")),
          totalcoverage = 42
        )
      )
    )
  }
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    run_coverage = mock_run_coverage,
    run_covr_skip_stf = mock_run_covr_skip_stf
  )
  
  # Stub test framework detection → has testthat
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", function(...) {
    list(has_testthat = TRUE, has_tests_base = FALSE, has_testit = FALSE)
  })
  
  # Stub get_pkg_desc
  mock_get_pkg_desc <- function(pkg_source_path, fields) {
    list(Package = "mockpkg", Version = "0.1.0")
  }
  mockery::stub(run_covr_modes, "get_pkg_desc", mock_get_pkg_desc)
  
  # Stub unloadNamespace (important to avoid real unloading)
  mockery::stub(run_covr_modes, "unloadNamespace", function(...) {})
  
  # Run
  result <- run_covr_modes(pkg_source_path)
  
  # Assertions
  expect_true(is.list(result))
  expect_equal(result$total_cov, 42)
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "fallback_function")
})

test_that("run_covr_modes handles non-standard testing framework", {
  local_cov_env <- new.env(parent = baseenv())
  
  mock_initialise_covr_env <- function(pkg_source_path, pkg, framework, ...) {
    assign("pkg_path", pkg_source_path, envir = local_cov_env)
    assign("package", pkg, envir = local_cov_env)
    assign("framework", framework, envir = local_cov_env)
    assign("work_dir", tempdir(), envir = local_cov_env)
    assign("pkg_root", pkg_source_path, envir = local_cov_env)
    local_cov_env
  }
  
  mock_run_covr_skip_nstf <- function(pkg_source_path, test_pkg_data, cov_env,
                                      framework = NULL, ...) {
    list(
      total_cov = 50,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = matrix(1),
          totalcoverage = 50
        ),
        notes = "NSTF coverage run"
      )
    )
  }
  
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env = local_cov_env,
    initialise_covr_env = mock_initialise_covr_env,
    run_covr_skip_nstf = mock_run_covr_skip_nstf
  )
  
  pkg_source_path <- file.path(tempdir(), "mockpkg_nstf_base")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mockery::stub(run_covr_modes, "get_pkg_name", function(path, ...) "mockpkg")
  mockery::stub(run_covr_modes, "get_pkg_desc",
                function(path, fields, ...) list(Package = "mockpkg", Version = "0.1.0")
  )
  mockery::stub(run_covr_modes, "get_test_metadata",
                function(...) {
                  list(
                    datetime = "2026-01-01 12:00:00",
                    executor = "test-user",
                    info = list(sys = list(
                      sysname = "Linux",
                      version = "5.15",
                      release = "Ubuntu",
                      machine = "x86_64"
                    ))
                  )
                }
  )
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps",
                function(...) {
                  list(
                    has_testthat = FALSE,
                    has_tinytest = FALSE,
                    has_data_table_tests = FALSE,
                    has_BioG_test = FALSE,
                    has_RUnit_test = FALSE,
                    has_tests_base = TRUE,
                    has_testit = FALSE,
                    has_snaps = FALSE,
                    bioc_unit_tests_dir = NA_character_,
                    n_golden_tests = 0L,
                    n_test_files = 1L
                  )
                }
  )
  mockery::stub(safe_unload_package, "loadedNamespaces", function() character(0))
  
  result <- run_covr_modes(pkg_source_path)
  
  expect_true(is.list(result))
  expect_equal(result$total_cov, 50)
  expect_equal(result$res_cov$notes, "NSTF coverage run")
})

test_that("run_covr_modes multiple frameworks branch runs each framework and aggregates total_cov", {
  pkg_source_path <- file.path(tempdir(), "mockpkg_multi")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mock_run_one_framework <- function(framework, pkg, test_pkg_data, pkg_source_path, covr_timeout) {
    cov_val <- switch(framework, testthat = 80, testit = 40, 0)
    list(
      total_cov = cov_val,
      res_cov = list(
        name = pkg,
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("fn")),
          totalcoverage = cov_val
        )
      ),
      test_framework_type = "framework",
      framework_id = framework
    )
  }
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    run_one_framework = mock_run_one_framework
  )
  
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", function(...) {
    list(
      has_testthat = TRUE,
      has_testit = TRUE,
      has_tests_base = FALSE,
      has_tinytest = FALSE,
      has_data_table_tests = FALSE,
      has_BioG_test = FALSE
    )
  })
  mockery::stub(run_covr_modes, "get_pkg_name", function(path, ...) "mockpkg")
  mockery::stub(run_covr_modes, "get_pkg_desc", function(path, fields, ...) {
    list(Package = "mockpkg", Version = "1.0.0")
  })
  mockery::stub(run_covr_modes, "get_test_metadata", function(...) {
    list(
      datetime = "2026-01-15 12:00:00",
      executor = "test-user",
      info = list(sys = list(sysname = "Linux", version = "5.15", release = "Ubuntu", machine = "x86_64"))
    )
  })
  mockery::stub(run_covr_modes, "unloadNamespace", function(pkg, ...) NULL)
  
  expect_message(
    result <- run_covr_modes(pkg_source_path, covr_timeout = 10),
    "Multiple test frameworks detected"
  )
  
  expect_true(result$multi_framework)
  expect_equal(result$frameworks, c("testthat", "testit"))
  expect_equal(result$total_cov, 60)
  expect_equal(result$pkg_name, "mockpkg")
  expect_equal(result$pkg_ver, "1.0.0")
  expect_named(result$results, c("testthat", "testit"))
  expect_equal(result$results$testthat$total_cov, 80)
  expect_equal(result$results$testit$total_cov, 40)
})

test_that("run_covr_modes handles no testing configuration", {
  pkg_source_path <- "mock/path"
  
  # Isolate a clean cov_env for this test
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env  = new.env(parent = baseenv())
  )
  
  # --- Stubs ---
  
  # Prevent normalizePath() and filesystem access. Return an environment.
  mockery::stub(run_covr_modes, "initialise_covr_env",
                function(pkg_source_path, pkg, framework, ...) {
                  assign("pkg_path",  pkg_source_path, envir = cov_env)
                  assign("package",   pkg,             envir = cov_env)
                  assign("framework", framework,       envir = cov_env)
                  cov_env
                }
  )
  
  # Mocked get_pkg_desc to return a package name
  mock_get_pkg_desc <- function(pkg_source_path, fields, ...) {
    list(Package = "mockpkg", Version = "0.1.0")
  }
  mockery::stub(run_covr_modes, "get_pkg_desc", mock_get_pkg_desc)
  
  # Mock unloadNamespace to avoid side effects
  mockery::stub(run_covr_modes, "unloadNamespace", function(...) {})
  
  # Return a complete "no tests present" structure
  mock_check_pkg_tests_and_snaps <- function(...) {
    list(
      has_testthat       = FALSE,
      has_snaps          = FALSE,
      has_testit         = FALSE,
      has_tinytests      = FALSE,
      has_data_table_tests = FALSE,
      has_tests_base     = FALSE,
      has_BioG_test      = FALSE,
      has_RUnit_test     = FALSE,   
      bioc_unit_tests_dir = NA_character_,
      n_golden_tests     = 0L,
      n_test_files       = 0L
    )
  }
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", mock_check_pkg_tests_and_snaps)
  
  # --- Execute ---
  result <- run_covr_modes(pkg_source_path)
  
  # --- Assertions ---
  expect_true(is.list(result))
  expect_equal(result$total_cov, 0)
  # The function should explain why coverage is 0 when no configs are found.
  expect_match(
    result$res_cov$errors,
    "No recognised standard or non-standard testing configuration",
    fixed = FALSE
  )
})

test_that("run_covr_modes unloads package when loaded", {
  pkg_source_path <- "mock/path"
  
  # Isolate a clean cov_env for this test
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env  = new.env(parent = baseenv())
  )
  
  # ---- Stubs ----
  
  # Stub initialise_covr_env: avoid normalizePath() and return the environment
  mockery::stub(run_covr_modes, "initialise_covr_env",
                function(pkg_source_path, pkg, framework, ...) {
                  assign("pkg_path",  pkg_source_path, envir = cov_env)
                  assign("package",   pkg,             envir = cov_env)
                  assign("framework", framework,       envir = cov_env)
                  cov_env
                }
  )
  
  # Mock get_pkg_desc to return a known package
  mock_get_pkg_desc <- function(pkg_source_path, fields, ...) {
    list(Package = "mockpkg", Version = "0.1.0")
  }
  mockery::stub(run_covr_modes, "get_pkg_desc", mock_get_pkg_desc)
  
  # Stub loadedNamespaces to simulate that the package is loaded
  mockery::stub(run_covr_modes, "loadedNamespaces", function(...) {
    c("mockpkg", "stats", "utils")
  })
  
  # Mock unloadNamespace to simulate successful unload (no side effects)
  mockery::stub(run_covr_modes, "unloadNamespace", function(pkg, ...) {
    # Optional: track or message if you want to assert it was called
    message(sprintf("Mock unloadNamespace called for %s", pkg))
    invisible(NULL)
  })
  
  # Return a COMPLETE "no tests present" structure
  mock_check_pkg_tests_and_snaps <- function(...) {
    list(
      has_testthat        = FALSE,
      has_snaps           = FALSE,
      has_testit          = FALSE,
      has_tests_base      = FALSE,
      has_tinytests       = FALSE,
      has_data_table_tests = FALSE,
      has_BioG_test       = FALSE,
      has_RUnit_test      = FALSE,   
      bioc_unit_tests_dir = NA_character_,
      n_golden_tests      = 0L,
      n_test_files        = 0L
    )
  }
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", mock_check_pkg_tests_and_snaps)
  
  # ---- Execute ----
  result <- run_covr_modes(pkg_source_path)
  
  # ---- Assertions ----
  expect_true(is.list(result))
  expect_equal(result$total_cov, 0)
  expect_match(
    result$res_cov$errors,
    "No recognised standard or non-standard testing configuration",
    fixed = FALSE
  )
})



test_that("run_covr_modes unloads package when loaded", {
  pkg_source_path <- "mock/path"
  
  # Isolate a clean cov_env for this test
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env  = new.env(parent = baseenv())
  )
  
  # ---- Stubs ----
  
  # Stub initialise_covr_env: avoid normalizePath() and return the environment
  mockery::stub(run_covr_modes, "initialise_covr_env",
                function(pkg_source_path, pkg, framework, ...) {
                  assign("pkg_path",  pkg_source_path, envir = cov_env)
                  assign("package",   pkg,             envir = cov_env)
                  assign("framework", framework,       envir = cov_env)
                  cov_env
                }
  )
  
  # Mock get_pkg_desc to return a known package
  mockery::stub(run_covr_modes, "get_pkg_desc",
                function(pkg_source_path, fields, ...) {
                  list(Package = "mockpkg", Version = "0.1.0")
                }
  )
  
  # Stub loadedNamespaces to simulate that the package is loaded
  mockery::stub(run_covr_modes, "loadedNamespaces",
                function(...) {
                  c("mockpkg", "stats", "utils")
                }
  )
  
  # Mock unloadNamespace to simulate successful unload (no side effects)
  mockery::stub(run_covr_modes, "unloadNamespace",
                function(pkg, ...) {
                  message(sprintf("Mock unloadNamespace called for %s", pkg))
                  invisible(NULL)
                }
  )
  
  # ---- CRITICAL: Return COMPLETE structure expected by run_covr_modes ----
  mock_check_pkg_tests_and_snaps <- function(...) {
    list(
      has_testthat          = FALSE,
      has_snaps             = FALSE,
      has_testit            = FALSE,
      has_tests_base        = FALSE,
      has_tinytest          = FALSE,
      has_data_table_tests  = FALSE,
      has_BioG_test         = FALSE,
      has_RUnit_test        = FALSE,         
      bioc_unit_tests_dir   = NA_character_,
      n_golden_tests        = 0L,
      n_test_files          = 0L
    )
  }
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", mock_check_pkg_tests_and_snaps)
  
  # ---- Execute ----
  result <- run_covr_modes(pkg_source_path)
  
  # ---- Assertions ----
  expect_true(is.list(result))
  expect_equal(result$total_cov, 0)
  expect_match(
    result$res_cov$errors,
    "No recognised standard or non-standard testing configuration",
    fixed = FALSE
  )
})

test_that("run_covr_modes handles error when unloading package", {
  pkg_source_path <- "mock/path"
  
  # Isolate clean cov_env for this test
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    cov_env = new.env(parent = baseenv())
  )
  
  # ---- Stub initialise_covr_env ----
  mockery::stub(run_covr_modes, "initialise_covr_env",
                function(pkg_source_path, pkg, framework, ...) {
                  assign("pkg_path",  pkg_source_path, envir = cov_env)
                  assign("package",   pkg,             envir = cov_env)
                  assign("framework", framework,       envir = cov_env)
                  cov_env
                }
  )
  
  # ---- get_pkg_desc ----
  mockery::stub(run_covr_modes, "get_pkg_desc",
                function(pkg_source_path, fields, ...) {
                  list(Package = "mockpkg", Version = "0.1.0")
                }
  )
  
  # ---- Simulate package being loaded ----
  mockery::stub(run_covr_modes, "loadedNamespaces",
                function(...) c("mockpkg", "stats", "utils")
  )
  
  # ---- Simulate unloadNamespace() failing ----
  mockery::stub(run_covr_modes, "unloadNamespace",
                function(pkg, ...) stop("Simulated unload failure")
  )
  
  # ---- REQUIRED: complete no-test configuration ----
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps",
                function(...) {
                  list(
                    has_testthat          = FALSE,
                    has_snaps             = FALSE,
                    has_testit            = FALSE,
                    has_tests_base        = FALSE,
                    has_tinytest          = FALSE,
                    has_data_table_tests  = FALSE,
                    has_BioG_test         = FALSE,
                    has_RUnit_test        = FALSE,  # <-- REQUIRED to avoid length-0 if()
                    bioc_unit_tests_dir   = NA_character_,
                    n_golden_tests        = 0L,
                    n_test_files          = 0L
                  )
                }
  )
  
  # ---- Execute ----
  result <- run_covr_modes(pkg_source_path)
  
  # ---- Assertions ----
  expect_true(is.list(result))
  expect_equal(result$total_cov, 0)
  expect_match(
    result$res_cov$errors,
    "No recognised standard or non-standard testing configuration",
    fixed = FALSE
  )
})

test_that("run_covr_modes handles error when unloading package", {
  mock_run_one_framework <- function(framework, pkg, test_pkg_data, pkg_source_path, covr_timeout) {
    safe_unload_package(pkg)
    list(
      total_cov = 85,
      res_cov = list(
        name = pkg,
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("some_function")),
          totalcoverage = 85
        )
      )
    )
  }
  
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    run_one_framework = mock_run_one_framework
  )
  
  pkg_source_path <- file.path(tempdir(), "mockpkg_unload_msg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mockery::stub(safe_unload_package, "loadedNamespaces", function() c("mockpkg", "stats", "utils"))
  mockery::stub(safe_unload_package, "unloadNamespace", function(pkg, ...) stop("Simulated unload failure"))
  
  mockery::stub(run_covr_modes, "get_pkg_name", function(path, ...) "mockpkg")
  mockery::stub(run_covr_modes, "get_pkg_desc",
                function(path, fields, ...) list(Package = "mockpkg", Version = "1.0.0")
  )
  
  mockery::stub(run_covr_modes, "get_test_metadata",
                function(...) {
                  list(
                    datetime = "2026-01-07 12:00:00",
                    executor = "test-user",
                    info = list(sys = list(
                      sysname = "Linux",
                      version = "5.15",
                      release = "Ubuntu",
                      machine = "x86_64"
                    ))
                  )
                }
  )
  
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps",
                function(path, ...) {
                  list(
                    has_testthat          = TRUE,
                    has_testit            = FALSE,
                    has_tests_base        = FALSE,
                    has_tinytest          = FALSE,
                    has_data_table_tests  = FALSE,
                    has_BioG_test         = FALSE,
                    has_RUnit_test        = FALSE,
                    has_snaps             = FALSE,
                    bioc_unit_tests_dir   = NA_character_,
                    n_golden_tests        = 0L,
                    n_test_files          = 0L
                  )
                }
  )
  
  msgs <- testthat::capture_messages({
    result <- run_covr_modes(pkg_source_path)
  })
  expect_true(any(grepl("Warning: Failed to unload mockpkg package", msgs, fixed = TRUE)))
  expect_true(any(grepl("Details: Simulated unload failure", msgs, fixed = TRUE)))

})


test_that("create_empty_covr_list returns correct structure", {
  pkg_name <- "mockpkg"
  error_message <- "No coverage data available"
  
  result <- create_empty_covr_list(pkg_name, error_message)
  
  expect_type(result, "list")
  expect_named(result, c("total_cov", "res_cov"))
  
  expect_equal(result$total_cov, 0)
  expect_type(result$res_cov, "list")
  expect_equal(result$res_cov$name, pkg_name)
  expect_equal(result$res_cov$coverage$totalcoverage, 0)
  expect_true(is.matrix(result$res_cov$coverage$filecoverage))
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "No functions tested")
  expect_equal(result$res_cov$errors, error_message)
  expect_true(is.na(result$res_cov$notes))
})

test_that("safe_unload_package unloads a loaded package", {
  mockery::stub(safe_unload_package, "loadedNamespaces", function() c("mockpkg"))
  mockery::stub(safe_unload_package, "unloadNamespace", function(pkg) TRUE)
  
  expect_message(safe_unload_package("mockpkg"), "Unloading mockpkg package to avoid coverage conflict...")
})

test_that("safe_unload_package handles unloadNamespace error gracefully", {
  mockery::stub(safe_unload_package, "loadedNamespaces", function() c("mockpkg"))
  mockery::stub(safe_unload_package, "unloadNamespace", function(pkg) stop("mock error"))
  
  expect_message(safe_unload_package("mockpkg"), "Warning: Failed to unload mockpkg package. Proceeding anyway.")
  expect_message(safe_unload_package("mockpkg"), "Details: mock error")
})

test_that("load_package_into_env loads package objects into environment", {
  cov_env <- new.env()
  
  # Stub pkgload::load_all to simulate successful load
  mockery::stub(load_package_into_env, "pkgload::load_all", function(path) TRUE)
  
  # Stub ls to simulate package objects
  mockery::stub(load_package_into_env, "ls", function(envir, all.names) c("obj1", "obj2"))
  
  # Stub get to return dummy objects
  mockery::stub(load_package_into_env, "get", function(name, envir) paste0("value_of_", name))
  
  # Stub assign to actually assign into cov_env
  mockery::stub(load_package_into_env, "assign", function(x, value, envir) assign(x, value, envir = envir))
  
  result <- load_package_into_env("mock/path", "mockpkg", cov_env)
  
  expect_true(exists("obj1", envir = cov_env))
  expect_equal(get("obj1", envir = cov_env), "value_of_obj1")
  expect_true(inherits(result, "logical"))
})


test_that("load_package_into_env stops on load_all error", {
  cov_env <- new.env()
  
  mockery::stub(load_package_into_env, "pkgload::load_all", function(path) stop("load error"))
  
  expect_error(load_package_into_env("mock/path", "mockpkg", cov_env),
               "Failed to load package mockpkg : load error")
})


test_that("load_package_into_env handles get() error gracefully", {
  cov_env <- new.env()
  
  mockery::stub(load_package_into_env, "pkgload::load_all", function(path) TRUE)
  mockery::stub(load_package_into_env, "ls", function(envir, all.names) c("obj1", "obj2"))
  
  # Simulate get() failing for obj2
  mockery::stub(load_package_into_env, "get", function(name, envir) {
    if (name == "obj2") stop("get error")
    paste0("value_of_", name)
  })
  
  # Stub assign to assign into cov_env
  mockery::stub(load_package_into_env, "assign", function(x, value, envir) assign(x, value, envir = envir))
  
  expect_message(load_package_into_env("mock/path", "mockpkg", cov_env),
                 "Failed to assign object obj2 : get error")
  
  expect_true(exists("obj1", envir = cov_env))
  expect_false(exists("obj2", envir = cov_env))
})


test_that("find_project_root returns NULL when no project files are found", {
  # Create a temporary directory with no DESCRIPTION, .Rproj, or .git
  temp_root <- tempfile("no_project_root_")
  dir.create(temp_root)
  
  # Stub file.exists to always return FALSE to simulate no project markers
  mockery::stub(find_project_root, "file.exists", function(path) FALSE)
  
  # Run the function starting from the temp directory
  result <- find_project_root(start = temp_root)
  
  # Expect NULL because no project root indicators exist
  expect_null(result)
})




test_that("cleanup_and_return_null skips orphan cleanup when no safe root can be determined", {
  cov_env <- new.env()
  assign("x", 1, envir = cov_env)
  
  # Make sure there is no pkg_root and no work_dir so safe_roots stays empty.
  # (You can keep your normalizePath/getwd stubs if you want, but they're no longer required.)
  msgs <- testthat::capture_messages({
    result <- cleanup_and_return_null("Test message", env = cov_env)
  })
  
  expect_null(result)
  expect_true(
    any(grepl("No safe roots found for orphan cleanup; skipping\\.", msgs)),
    info = paste("Captured messages:\n", paste(msgs, collapse = "\n"))
  )
  expect_equal(base::ls(envir = cov_env, all.names = TRUE), character(0))
})


test_that("returns character(0) when package is not installed", {
  dummy_env <- new.env(parent = emptyenv())
  
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  mockery::stub(
    ld,
    "requireNamespace",
    function(pkg, quietly = TRUE) FALSE
  )
  
  expect_message(
    result <- ld(
      pkg_names = "fakepkg",
      env = dummy_env,
      source_paths = NULL,
      overwrite = FALSE
    ),
    "Package fakepkg is not installed; skipping installed datasets."
  )
  
  expect_identical(result, character(0))
  expect_equal(ls(envir = dummy_env, all.names = TRUE), character(0))
})



test_that("returns datasets from source data directory with supported extensions", {
  fake_env <- new.env(parent = emptyenv())

  # Create a fake package root inside tempdir()
  pkg_root_dir <- file.path(tempdir(), "pkgroot_test")
  dir.create(pkg_root_dir, recursive = TRUE, showWarnings = FALSE)
  
  # Create a working directory *inside* pkg_root
  child_dir <- file.path(pkg_root_dir, "subdir")
  dir.create(child_dir, recursive = TRUE, showWarnings = FALSE)
  
  # Save original working directory
  old_wd <- getwd()
  on.exit(setwd(old_wd), add = TRUE)
  
  # Set working directory to child directory (so cwd_norm is inside pkg_root)
  setwd(child_dir)
  
  # Build cleanup environment
  fake_env <- new.env(parent = baseenv())
  fake_env$pkg_root <- pkg_root_dir
  fake_env$work_dir <- NULL
  
  # --- Mock filesystem functions -------------------------------------------
  # (No library(mockery) is used; only direct mockery::stub calls.)
  
  # Pretend all paths exist
  mockery::stub(cleanup_and_return_null, "dir.exists", function(path) TRUE)
  
  # Make list.files return a fixed empty set (we don't test deletion)
  mockery::stub(cleanup_and_return_null, "list.files", function(...) character(0))
  
  # Mock normalizePath so that pkg_root and cwd resolve exactly
  mockery::stub(cleanup_and_return_null, "normalizePath",
                function(path, ...) {
                  # Return path unchanged (good enough for prefix checks)
                  path
                }
  )
  
  # --- Run cleanup ----------------------------------------------------------
  # Capture the message that indicates safe_roots is non-empty
  # Specifically: we should NOT see "No safe roots found..."
  
  output <- capture.output({
    cleanup_and_return_null(env = fake_env)
  })
  
  # --- Assertions -----------------------------------------------------------
  
  # The function should *not* say "No safe roots found"
  expect_false(any(grepl("No safe roots found", output)))
  
  # And no error should occur
  expect_true(TRUE)  # reached successfully
})


test_that("falls back to data() if no files loaded", {
  fake_env <- new.env()
  mockery::stub(load_package_datasets, "requireNamespace", TRUE)
  mockery::stub(load_package_datasets, "find.package", "/fake/path")
  mockery::stub(load_package_datasets, "dir.exists", TRUE)
  mockery::stub(load_package_datasets, "list.files", character(0))
 
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  # For this test we want to exercise only the source data directory branch.
  # So treat the source package path as available, but the installed package
  # as unavailable.
  mockery::stub(
    ld,
    "requireNamespace",
    function(pkg, quietly = TRUE) FALSE
  )
  
  # Simulate /fake/path/data existing
  mockery::stub(ld, "dir.exists", function(path) TRUE)
  
  # Simulate supported files in the source data directory
  mockery::stub(
    ld,
    "list.files",
    function(path, full.names = TRUE) {
      c(
        "/fake/path/data/data1.csv",
        "/fake/path/data/data2.rda"
      )
    }
  )
  
  # Match the symbols used in the production code
  mockery::stub(
    ld,
    "tools::file_ext",
    function(x) {
      if (grepl("\\.csv$", x)) "csv" else "rda"
    }
  )
  
  mockery::stub(
    ld,
    "tools::file_path_sans_ext",
    function(x) {
      sub("\\.[^.]*$", "", basename(x))
    }
  )
  
  # Stub file loaders
  mockery::stub(
    ld,
    "utils::read.csv",
    function(file, header = TRUE) {
      data.frame(a = 1:3)
    }
  )
  
  mockery::stub(
    ld,
    "load",
    function(file, envir) {
      assign("data2", data.frame(x = 1), envir = envir)
      invisible("data2")
    }
  )
  
  result <- ld(
    pkg_names = "fakepkg",
    env = fake_env,
    source_paths = c(fakepkg = "/fake/path"),
    overwrite = FALSE
  )
  
  expect_setequal(result, c("data1", "data2"))
  expect_true(exists("data1", envir = fake_env, inherits = FALSE))
  expect_true(exists("data2", envir = fake_env, inherits = FALSE))
  expect_equal(get("data1", envir = fake_env, inherits = FALSE)$a, 1:3)
  expect_equal(get("data2", envir = fake_env, inherits = FALSE)$x, 1)
})



test_that("falls back to utils::data() if no files are loaded from source data directory", {
  fake_env <- new.env(parent = emptyenv())
  
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  # Treat fakepkg as installed so the installed-package fallback branch is used
  mockery::stub(
    ld,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  # Simulate source data directory exists but contains no files
  mockery::stub(ld, "dir.exists", function(path) TRUE)
  mockery::stub(ld, "list.files", function(path, full.names = TRUE) character(0))
  
  # Mock utils::data() in the utils namespace, because the function
  # calls utils::data(...) explicitly.
  testthat::local_mocked_bindings(
    data = function(..., list = NULL, package = NULL, envir = NULL) {
      if (is.null(list)) {
        # Simulate utils::data(package = "fakepkg")$results
        return(list(
          results = matrix(
            "fallback_data",
            ncol = 1,
            dimnames = list(NULL, "Item")
          )
        ))
      } else {
        # Simulate utils::data(list = "fallback_data", package = "fakepkg", envir = fake_env)
        assign(list, data.frame(x = 1), envir = envir)
        invisible(NULL)
      }
    },
    .package = "utils"
  )
  
  result <- ld(
    pkg_names = "fakepkg",
    env = fake_env,
    source_paths = c(fakepkg = "/fake/path"),
    overwrite = FALSE
  )
  
  expect_identical(result, "fakepkg::fallback_data")
  expect_true(exists("fallback_data", envir = fake_env, inherits = FALSE))
  expect_equal(get("fallback_data", envir = fake_env, inherits = FALSE)$x, 1)
})




test_that("handles errors during file loading gracefully", {
  fake_env <- new.env(parent = emptyenv())
  
  ld <- get("load_package_datasets", envir = asNamespace("test.assessr"))
  
  # For this test, exercise only the source data directory branch.
  # Skip installed-package loading by making fakepkg appear unavailable.
  mockery::stub(
    ld,
    "requireNamespace",
    function(pkg, quietly = TRUE) FALSE
  )
  
  mockery::stub(ld, "dir.exists", function(path) TRUE)
  
  mockery::stub(
    ld,
    "list.files",
    function(path, full.names = TRUE) {
      "/fake/path/data/bad.csv"
    }
  )
  
  mockery::stub(
    ld,
    "tools::file_ext",
    function(x) "csv"
  )
  
  mockery::stub(
    ld,
    "tools::file_path_sans_ext",
    function(x) "bad"
  )
  
  mockery::stub(
    ld,
    "utils::read.csv",
    function(...) stop("read error")
  )
  
  expect_message(
    result <- ld(
      pkg_names = "fakepkg",
      env = fake_env,
      source_paths = c(fakepkg = "/fake/path"),
      overwrite = FALSE
    ),
    "Failed to load file /fake/path/data/bad.csv : read error"
  )
  
  expect_identical(result, character(0))
  expect_false("bad" %in% result)
  expect_false(exists("bad", envir = fake_env, inherits = FALSE))
})



test_that("returns empty when no new globals", {
  env <- new.env()
  initial <- ls(envir = env)
  expect_message(result <- remove_new_globals(env, initial), "No new globals detected.")
  expect_equal(result, character())
})

test_that("removes new globals", {
  env <- new.env()
  initial <- ls(envir = env)
  assign("new_var", 42, envir = env)
  expect_message(result <- remove_new_globals(env, initial), "Removed new globals:")
  expect_false("new_var" %in% ls(envir = env))
  expect_equal(result, "new_var")
})

test_that("filters new globals by pattern", {
  env <- new.env()
  initial <- ls(envir = env)
  assign("temp_data", 1, envir = env)
  assign("keep_me", 2, envir = env)
  result <- remove_new_globals(env, initial, patterns = "^temp")
  expect_false("temp_data" %in% ls(envir = env))
  expect_true("keep_me" %in% ls(envir = env))
  expect_equal(result, "temp_data")
})

test_that("dry run does not remove globals", {
  env <- new.env()
  initial <- ls(envir = env)
  assign("dry_var", 99, envir = env)
  expect_message(result <- remove_new_globals(env, initial, dry_run = TRUE), "New globals detected")
  expect_true("dry_var" %in% ls(envir = env))
  expect_equal(result, "dry_var")
})


test_that("does nothing when no helper files found", {
  env <- new.env()
  mockery::stub(load_test_helpers, "list.files", character(0))
  expect_message(result <- load_test_helpers(env), NA)  # no message expected
  expect_equal(result, invisible(TRUE))
})

test_that("sources helper files successfully", {
  env <- new.env()
  helper_files <- c("tests/testthat/helper_one.R", "tests/testthat/helper_two.R")
  
  mockery::stub(load_test_helpers, "list.files", helper_files)
  mockery::stub(load_test_helpers, "sys.source", function(file, envir) {
    assign(basename(file), TRUE, envir = envir)
  })
  
  result <- load_test_helpers(env)
  expect_true("helper_one.R" %in% ls(env))
  expect_true("helper_two.R" %in% ls(env))
  expect_equal(result, invisible(TRUE))
})

test_that("handles errors during sourcing", {
  env <- new.env()
  helper_files <- c("tests/testthat/helper_bad.R")
  
  mockery::stub(load_test_helpers, "list.files", helper_files)
  mockery::stub(load_test_helpers, "sys.source", function(file, envir) {
    stop("source error")
  })
  
  expect_message(load_test_helpers(env), "Failed to source helper file:")
})

test_that("cleanup_and_return_null removes only .pdf/.ps files and only if they exist (ephemeral root)", {
  cov_env <- new.env(parent = emptyenv())
  assign("dummy", 1, envir = cov_env)
  
  # Put work_dir under tempdir() so extension-based cleanup is allowed
  cov_env$work_dir <- "/tmp/project"
  
  # Ensure pkg_root is not set for this test
  if (exists("pkg_root", envir = cov_env, inherits = FALSE)) {
    rm("pkg_root", envir = cov_env)
  }
  
  # Make tempdir and normalizePath consistent with our paths
  mockery::stub(cleanup_and_return_null, "tempdir", function() "/tmp")
  mockery::stub(cleanup_and_return_null, "normalizePath", function(path, ...) path)
  
  # Ensure the work_dir is considered present and prevent its deletion in step 2
  mockery::stub(cleanup_and_return_null, "dir.exists", function(path) path == "/tmp/project")
  mockery::stub(cleanup_and_return_null, "unlink", function(path, recursive = FALSE, force = FALSE) TRUE)
  
  # Mock files in the ephemeral root
  mock_files <- c(
    "/tmp/project/report.pdf",  # exists = TRUE => should be removed
    "/tmp/project/diagram.ps",  # exists = FALSE => should NOT be removed
    "/tmp/project/readme.txt"   # wrong extension => ignored
  )
  mockery::stub(cleanup_and_return_null, "list.files", function(path, full.names = TRUE, ...) mock_files)
  
  # Only the .pdf exists
  mockery::stub(cleanup_and_return_null, "file.exists", function(path) path == "/tmp/project/report.pdf")
  
  # Track deletions
  removed_files <- character(0)
  mockery::stub(cleanup_and_return_null, "file.remove", function(path) { removed_files <<- c(removed_files, path); TRUE })
  
  # Run
  result <- cleanup_and_return_null(env = cov_env)
  expect_null(result)
  
  # Only the existing .pdf should have been removed
  expect_identical(removed_files, "/tmp/project/report.pdf")
})


test_that("setup_covr_env completes successfully", {
  # -------------------------------
  # Create a mock cov_env
  # -------------------------------
  cov_env <- new.env(parent = emptyenv())
  
  # setup_covr_env() asserts that test_path exists
  test_path <- withr::local_tempdir()
  
  # Pull function from namespace so stubs apply to the function object used here
  sce <- get("setup_covr_env", envir = asNamespace("test.assessr"))
  
  # -------------------------------
  # Stubs for helpers / dependencies
  # -------------------------------
  
  # Simulate a single helper file
  mockery::stub(
    sce,
    "list.files",
    function(path, pattern = NULL, full.names = TRUE) "helper_one.R"
  )
  
  # Pretend sourcing a helper sets a flag inside cov_env
  mockery::stub(
    sce,
    "sys.source",
    function(file, envir) {
      assign("helper_loaded", TRUE, envir = envir)
      invisible(NULL)
    }
  )
  
  # Skip loading the real package
  mockery::stub(
    sce,
    "pkgload::load_all",
    function(...) NULL
  )
  
  # Fake namespace with one object to copy into cov_env
  mockery::stub(
    sce,
    "asNamespace",
    function(pkg) {
      e <- new.env(parent = emptyenv())
      assign("fun", function() NULL, envir = e)
      e
    }
  )
  
  # Force find_project_root() to return a known value
  mockery::stub(
    sce,
    "find_project_root",
    function(...) "/mock/root"
  )
  
  # Fake package dataset loading
  mockery::stub(
    sce,
    "load_package_datasets",
    function(pkg_names, env, source_paths = NULL, overwrite = FALSE) {
      if (identical(pkg_names, "mockpkg")) {
        c("data1", "data2")
      } else if (identical(pkg_names, "nycflights13")) {
        "flights"
      } else {
        character(0)
      }
    }
  )
  
  # Fake Base R package surface loading
  mockery::stub(
    sce,
    "load_base_r_packages_into_env",
    function(env, overwrite = FALSE) {
      list(
        exports_loaded = c("stats::rnorm", "utils::head"),
        datasets_loaded = c("datasets::mtcars", "datasets::iris")
      )
    }
  )
  
  # Neutralize dependency loader and safe unload
  mockery::stub(sce, "load_dependencies_into_env", function(...) NULL)
  mockery::stub(sce, "safe_unload_package", function(...) NULL)
  
  # -------------------------------
  # Run the function
  # -------------------------------
  result <- sce(
    pkg_source_path = "mock/path",
    pkg_name = "mockpkg",
    test_path = test_path,
    cov_env = cov_env
  )
  
  # -------------------------------
  # Assertions
  # -------------------------------
  expect_type(result, "list")
  
  expect_named(
    result,
    c(
      "cov_env",
      "datasets_loaded",
      "base_r_datasets_loaded",
      "base_r_exports_loaded",
      "root_dir",
      "initial_state",
      "helpers_loaded"
    )
  )
  
  expect_identical(result$cov_env, cov_env)
  expect_equal(result$root_dir, "/mock/root")
  expect_equal(result$helpers_loaded, "helper_one.R")
  
  # Combined datasets = package datasets + base R datasets + nycflights13 datasets
  expect_setequal(
    result$datasets_loaded,
    c("data1", "data2", "datasets::mtcars", "datasets::iris", "flights")
  )
  
  expect_setequal(
    result$base_r_datasets_loaded,
    c("datasets::mtcars", "datasets::iris")
  )
  
  expect_setequal(
    result$base_r_exports_loaded,
    c("stats::rnorm", "utils::head")
  )
  
  # Confirm the helper actually "loaded"
  expect_true(isTRUE(get("helper_loaded", envir = cov_env, inherits = FALSE)))
  
  # Confirm namespace objects were copied into cov_env
  expect_true(exists("fun", envir = cov_env, inherits = FALSE))
  expect_true(is.function(get("fun", envir = cov_env, inherits = FALSE)))
})


test_that("load_base_r_packages_into_env returns exports and datasets from helper functions", {
  lb <- get("load_base_r_packages_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  seen_pkg_names <- NULL
  seen_env_exports <- NULL
  seen_overwrite_exports <- NULL
  
  seen_pkg_names_datasets <- NULL
  seen_env_datasets <- NULL
  seen_source_paths_datasets <- NULL
  seen_overwrite_datasets <- NULL
  
  mockery::stub(
    lb,
    "load_package_exports_into_env",
    function(pkg_names, env, overwrite = FALSE) {
      seen_pkg_names <<- pkg_names
      seen_env_exports <<- env
      seen_overwrite_exports <<- overwrite
      c("stats::rnorm", "utils::head")
    }
  )
  
  mockery::stub(
    lb,
    "load_package_datasets",
    function(pkg_names, env, source_paths = NULL, overwrite = FALSE) {
      seen_pkg_names_datasets <<- pkg_names
      seen_env_datasets <<- env
      seen_source_paths_datasets <<- source_paths
      seen_overwrite_datasets <<- overwrite
      c("datasets::mtcars", "datasets::iris")
    }
  )
  
  result <- lb(env = fake_env, overwrite = FALSE)
  
  expect_type(result, "list")
  expect_named(result, c("exports_loaded", "datasets_loaded"))
  
  expect_identical(result$exports_loaded, c("stats::rnorm", "utils::head"))
  expect_identical(result$datasets_loaded, c("datasets::mtcars", "datasets::iris"))
  
  expect_identical(seen_env_exports, fake_env)
  expect_false(seen_overwrite_exports)
  
  expect_identical(seen_env_datasets, fake_env)
  expect_null(seen_source_paths_datasets)
  expect_false(seen_overwrite_datasets)
  
  expect_identical(seen_pkg_names_datasets, "datasets")
})


test_that("load_base_r_packages_into_env excludes base and datasets from export package loading", {
  lb <- get("load_base_r_packages_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  seen_pkg_names <- NULL
  
  mockery::stub(
    lb,
    "load_package_exports_into_env",
    function(pkg_names, env, overwrite = FALSE) {
      seen_pkg_names <<- pkg_names
      character(0)
    }
  )
  
  mockery::stub(
    lb,
    "load_package_datasets",
    function(pkg_names, env, source_paths = NULL, overwrite = FALSE) {
      character(0)
    }
  )
  
  result <- lb(env = fake_env, overwrite = FALSE)
  
  expect_identical(result$exports_loaded, character(0))
  expect_identical(result$datasets_loaded, character(0))
  
  expect_setequal(
    seen_pkg_names,
    c(
      "compiler",
      "graphics",
      "grDevices",
      "grid",
      "methods",
      "parallel",
      "splines",
      "stats",
      "stats4",
      "tcltk",
      "tools",
      "utils"
    )
  )
  
  expect_false("base" %in% seen_pkg_names)
  expect_false("datasets" %in% seen_pkg_names)
})


test_that("load_base_r_packages_into_env passes overwrite to helper functions", {
  lb <- get("load_base_r_packages_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  seen_overwrite_exports <- NULL
  seen_overwrite_datasets <- NULL
  
  mockery::stub(
    lb,
    "load_package_exports_into_env",
    function(pkg_names, env, overwrite = FALSE) {
      seen_overwrite_exports <<- overwrite
      character(0)
    }
  )
  
  mockery::stub(
    lb,
    "load_package_datasets",
    function(pkg_names, env, source_paths = NULL, overwrite = FALSE) {
      seen_overwrite_datasets <<- overwrite
      character(0)
    }
  )
  
  result <- lb(env = fake_env, overwrite = TRUE)
  
  expect_identical(result$exports_loaded, character(0))
  expect_identical(result$datasets_loaded, character(0))
  
  expect_true(seen_overwrite_exports)
  expect_true(seen_overwrite_datasets)
})


test_that("load_base_r_packages_into_env errors if env is not an environment", {
  expect_error(
    load_base_r_packages_into_env(env = "not_an_env"),
    "env"
  )
})


test_that("load_base_r_packages_into_env errors if overwrite is not a logical flag", {
  fake_env <- new.env(parent = emptyenv())
  
  expect_error(
    load_base_r_packages_into_env(env = fake_env, overwrite = "yes"),
    "overwrite"
  )
})


test_that("load_package_exports_into_env loads exports into env", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) c("foo", "bar")
  )
  
  mockery::stub(
    lpe,
    "getExportedValue",
    function(pkg, nm) {
      if (nm == "foo") {
        function() "foo"
      } else if (nm == "bar") {
        123
      } else {
        stop("unexpected export")
      }
    }
  )
  
  result <- lpe(
    pkg_names = "fakepkg",
    env = fake_env,
    overwrite = FALSE
  )
  
  expect_setequal(result, c("fakepkg::foo", "fakepkg::bar"))
  expect_true(exists("foo", envir = fake_env, inherits = FALSE))
  expect_true(exists("bar", envir = fake_env, inherits = FALSE))
  expect_true(is.function(get("foo", envir = fake_env, inherits = FALSE)))
  expect_identical(get("bar", envir = fake_env, inherits = FALSE), 123)
})


test_that("load_package_exports_into_env skips packages that are not installed", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) FALSE
  )
  
  expect_message(
    result <- lpe(
      pkg_names = "missingpkg",
      env = fake_env,
      overwrite = FALSE
    ),
    "Package missingpkg is not installed or not available; skipping export load."
  )
  
  expect_identical(result, character(0))
  expect_identical(ls(envir = fake_env, all.names = TRUE), character(0))
})


test_that("load_package_exports_into_env handles getNamespaceExports errors gracefully", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) stop("namespace failure")
  )
  
  expect_message(
    result <- lpe(
      pkg_names = "brokenpkg",
      env = fake_env,
      overwrite = FALSE
    ),
    "Failed to retrieve exports for package brokenpkg : namespace failure"
  )
  
  expect_identical(result, character(0))
  expect_identical(ls(envir = fake_env, all.names = TRUE), character(0))
})


test_that("load_package_exports_into_env skips packages with no exports", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) character(0)
  )
  
  result <- lpe(
    pkg_names = "emptyexports",
    env = fake_env,
    overwrite = FALSE
  )
  
  expect_identical(result, character(0))
  expect_identical(ls(envir = fake_env, all.names = TRUE), character(0))
})


test_that("load_package_exports_into_env does not overwrite existing objects when overwrite is FALSE", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  assign("foo", "existing_value", envir = fake_env)
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) c("foo", "bar")
  )
  
  mockery::stub(
    lpe,
    "getExportedValue",
    function(pkg, nm) {
      if (nm == "foo") "new_value" else 999
    }
  )
  
  result <- lpe(
    pkg_names = "fakepkg",
    env = fake_env,
    overwrite = FALSE
  )
  
  # only bar should be loaded because foo already exists
  expect_identical(result, "fakepkg::bar")
  expect_identical(get("foo", envir = fake_env, inherits = FALSE), "existing_value")
  expect_identical(get("bar", envir = fake_env, inherits = FALSE), 999)
})


test_that("load_package_exports_into_env overwrites existing objects when overwrite is TRUE", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  assign("foo", "existing_value", envir = fake_env)
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) c("foo")
  )
  
  mockery::stub(
    lpe,
    "getExportedValue",
    function(pkg, nm) "replacement_value"
  )
  
  result <- lpe(
    pkg_names = "fakepkg",
    env = fake_env,
    overwrite = TRUE
  )
  
  expect_identical(result, "fakepkg::foo")
  expect_identical(get("foo", envir = fake_env, inherits = FALSE), "replacement_value")
})


test_that("load_package_exports_into_env handles getExportedValue errors gracefully", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) c("good", "bad")
  )
  
  mockery::stub(
    lpe,
    "getExportedValue",
    function(pkg, nm) {
      if (nm == "good") {
        42
      } else {
        stop("cannot load export")
      }
    }
  )
  
  expect_message(
    result <- lpe(
      pkg_names = "fakepkg",
      env = fake_env,
      overwrite = FALSE
    ),
    "Failed to load export fakepkg::bad : cannot load export"
  )
  
  expect_identical(result, "fakepkg::good")
  expect_true(exists("good", envir = fake_env, inherits = FALSE))
  expect_false(exists("bad", envir = fake_env, inherits = FALSE))
  expect_identical(get("good", envir = fake_env, inherits = FALSE), 42)
})


test_that("load_package_exports_into_env deduplicates repeated package names", {
  lpe <- get("load_package_exports_into_env", envir = asNamespace("test.assessr"))
  
  fake_env <- new.env(parent = emptyenv())
  seen_pkgs <- character()
  
  mockery::stub(
    lpe,
    "requireNamespace",
    function(pkg, quietly = TRUE) TRUE
  )
  
  mockery::stub(
    lpe,
    "getNamespaceExports",
    function(pkg) {
      seen_pkgs <<- c(seen_pkgs, pkg)
      "foo"
    }
  )
  
  mockery::stub(
    lpe,
    "getExportedValue",
    function(pkg, nm) 1
  )
  
  result <- lpe(
    pkg_names = c("fakepkg", "fakepkg", "fakepkg"),
    env = fake_env,
    overwrite = FALSE
  )
  
  expect_identical(seen_pkgs, "fakepkg")
  expect_identical(result, "fakepkg::foo")
  expect_identical(get("foo", envir = fake_env, inherits = FALSE), 1)
})


test_that("load_package_exports_into_env validates env", {
  expect_error(
    load_package_exports_into_env(
      pkg_names = "fakepkg",
      env = "not_an_environment"
    ),
    "env"
  )
})


test_that("load_package_exports_into_env validates overwrite", {
  fake_env <- new.env(parent = emptyenv())
  
  expect_error(
    load_package_exports_into_env(
      pkg_names = "fakepkg",
      env = fake_env,
      overwrite = "yes"
    ),
    "overwrite"
  )
})


test_that("load_package_exports_into_env validates pkg_names", {
  fake_env <- new.env(parent = emptyenv())
  
  expect_error(
    load_package_exports_into_env(
      pkg_names = character(0),
      env = fake_env
    ),
    "pkg_names"
  )
  
  expect_error(
    load_package_exports_into_env(
      pkg_names = c("ok", NA_character_),
      env = fake_env
    ),
    "pkg_names"
  )
})




test_that("initialise_covr_env uses default work_dir when NULL", {
  # Arrange
  pkg_source_path <- "dummy/pkg"  # not used because normalizePath is stubbed
  pkg <- "apkg"
  framework <- "testthat"
  
  # Deterministic default work_dir path
  mockery::stub(initialise_covr_env, "tempdir", function() "/TD")
  mockery::stub(initialise_covr_env, "Sys.getpid", function() 12345L)
  mockery::stub(initialise_covr_env, "file.path", function(...){
    paste0(paste0(c(...), collapse = "/"))
  })
  
  # Avoid FS writes and assert call
  fake_dir_create <- mockery::mock(TRUE)
  mockery::stub(initialise_covr_env, "dir.create", fake_dir_create)
  
  # Avoid real path resolution
  mockery::stub(initialise_covr_env, "normalizePath", function(path, winslash="/", mustWork=TRUE) {
    # Should be called for pkg_source_path only
    if (identical(path, pkg_source_path)) "/ABS/PATH"
    else path
  })
  
  # Act
  env <- initialise_covr_env(pkg_source_path, pkg, framework, work_dir = NULL)
  
  # Assert
  expect_true(is.environment(env))
  expect_identical(env$pkg_root, "/ABS/PATH")
  expect_identical(env$work_dir, "/TD/covr_apkg_testthat_12345")
  mockery::expect_called(fake_dir_create, 1)
  # dir.create(work_dir, recursive=TRUE, showWarnings=FALSE)
  mockery::expect_args(fake_dir_create, 1, "/TD/covr_apkg_testthat_12345",
                       recursive = TRUE, showWarnings = FALSE)
})

test_that("initialise_covr_env uses default work_dir when empty string", {
  pkg_source_path <- "dummy/pkg"
  pkg <- "apkg"
  framework <- "none"
  
  mockery::stub(initialise_covr_env, "tempdir", function() "/TMP")
  mockery::stub(initialise_covr_env, "Sys.getpid", function() 99999L)
  mockery::stub(initialise_covr_env, "file.path", function(...){
    paste0(paste0(c(...), collapse = "/"))
  })
  mockery::stub(initialise_covr_env, "dir.create", function(...) TRUE)
  mockery::stub(initialise_covr_env, "normalizePath", function(path, winslash="/", mustWork=TRUE) {
    "/ABS/PATH2"
  })
  
  env <- initialise_covr_env(pkg_source_path, pkg, framework, work_dir = "")
  
  expect_true(is.environment(env))
  expect_identical(env$work_dir, "/TMP/covr_apkg_none_99999")
  expect_identical(env$pkg_root, "/ABS/PATH2")
})

test_that("initialise_covr_env respects provided non-empty work_dir and does not use tempdir/file.path", {
  pkg_source_path <- "dummy/pkg"
  pkg <- "apkg"
  framework <- "testit"
  provided_work_dir <- "/CUSTOM/WORKDIR"
  
  # Ensure the default-path branch is NOT used
  mockery::stub(initialise_covr_env, "tempdir", function() stop("tempdir() should not be called"))
  mockery::stub(initialise_covr_env, "file.path", function(...) stop("file.path() should not be called"))
  
  fake_dir_create <- mockery::mock(TRUE)
  mockery::stub(initialise_covr_env, "dir.create", fake_dir_create)
  
  mockery::stub(initialise_covr_env, "normalizePath", function(path, winslash="/", mustWork=TRUE) {
    if (identical(path, pkg_source_path)) "/ABS/PATH3" else path
  })
  
  env <- initialise_covr_env(pkg_source_path, pkg, framework, work_dir = provided_work_dir)
  
  expect_true(is.environment(env))
  expect_identical(env$work_dir, provided_work_dir)
  expect_identical(env$pkg_root, "/ABS/PATH3")
  mockery::expect_called(fake_dir_create, 1)
  mockery::expect_args(fake_dir_create, 1, provided_work_dir,
                       recursive = TRUE, showWarnings = FALSE)
})

test_that("initialise_covr_env validates arguments (stopifnot)", {
  # Wrong types or lengths should error
  expect_error(initialise_covr_env(123, "pkg", "none"), "is.character")
  expect_error(initialise_covr_env(c("a", "b"), "pkg", "none"), "length")
  expect_error(initialise_covr_env("path", 1, "none"), "is.character")
  expect_error(initialise_covr_env("path", "pkg", c("a","b")), "length")
})

test_that("initialise_covr_env propagates normalizePath error when pkg_source_path does not exist", {
  pkg_source_path <- "missing/path"
  pkg <- "apkg"
  framework <- "none"
  
  # The function calls normalizePath(..., mustWork=TRUE), so error should bubble up
  mockery::stub(initialise_covr_env, "normalizePath", function(path, winslash="/", mustWork=TRUE) {
    stop(sprintf("path[%s] does not exist", deparse(path)))
  })
  # Avoid side effects before error
  mockery::stub(initialise_covr_env, "dir.create", function(...) TRUE)
  mockery::stub(initialise_covr_env, "tempdir", function() "/TD")
  mockery::stub(initialise_covr_env, "file.path", function(...) paste0(c(...), collapse="/"))
  mockery::stub(initialise_covr_env, "Sys.getpid", function() 1L)
  
  expect_error(
    initialise_covr_env(pkg_source_path, pkg, framework, work_dir = NULL),
    "does not exist"
  )
})



testthat::test_that("run_covr_modes handles tinytest non-standard framework", {
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mock_run_covr_skip_nstf <- function(pkg_source_path, test_pkg_data, cov_env, ...) {
    list(
      total_cov = 66,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("some_function")),
          totalcoverage = 66
        )
      )
    )
  }
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    run_covr_skip_nstf = mock_run_covr_skip_nstf
  )
  
  # Select the tinytest branch explicitly
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", function(...) {
    list(
      has_testthat = FALSE,
      has_tests_base = FALSE,
      has_testit = FALSE,
      has_tinytest = TRUE,
      has_data_table_tests = FALSE,
      has_BioG_test = FALSE
    )
  })
  
  # Minimal required stubs, mirroring your working pattern
  mockery::stub(run_covr_modes, "get_pkg_desc", function(pkg_source_path, fields) {
    list(Package = "mockpkg", Version = "0.1.0")
  })
  
  mockery::stub(run_covr_modes, "unloadNamespace", function(pkg) {})
  
  # Run
  result <- run_covr_modes(pkg_source_path, covr_timeout = 10)
  
  # Assertions (kept simple, like in your successful test)
  testthat::expect_true(is.list(result))
  testthat::expect_equal(result$total_cov, 66)
  testthat::expect_equal(result$test_framework_type, "non-standard testing framework - tinytest")
})


testthat::test_that("run_covr_modes handles data.table non-standard framework", {
  pkg_source_path <- file.path(tempdir(), "mockpkg_dt")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  mock_run_covr_skip_dt_nstf <- function(pkg_source_path, test_pkg_data, cov_env) {
    list(
      total_cov = 42,
      res_cov = list(
        name = "mockpkg",
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("some_function")),
          totalcoverage = 42
        )
      )
    )
  }
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    run_covr_skip_dt_nstf = mock_run_covr_skip_dt_nstf
  )
  
  # Select the data.table branch explicitly
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", function(...) {
    list(
      has_testthat = FALSE,
      has_tests_base = FALSE,
      has_testit = FALSE,
      has_tinytest = FALSE,
      has_data_table_tests = TRUE,
      has_BioG_test = FALSE
    )
  })
  
  # Minimal required stubs, mirroring your working pattern
  mockery::stub(run_covr_modes, "get_pkg_desc", function(pkg_source_path, fields) {
    list(Package = "mockpkg", Version = "0.1.0")
  })
  mockery::stub(run_covr_modes, "unloadNamespace", function(pkg) {})
  
  # Run
  result <- run_covr_modes(pkg_source_path, covr_timeout = 10)
  
  # Assertions (kept simple, like in your successful test)
  testthat::expect_true(is.list(result))
  testthat::expect_equal(result$total_cov, 42)
  testthat::expect_equal(result$test_framework_type, "non-standard testing framework - data.table")
})

test_that("compute_multi_framework_total returns 0 for empty list", {
  expect_equal(compute_multi_framework_total(list()), 0)
})

test_that("compute_multi_framework_total returns mean for multiple frameworks", {
  per_framework <- list(
    testthat = list(total_cov = 80),
    testit = list(total_cov = 40)
  )
  expect_equal(compute_multi_framework_total(per_framework), 60)
})

test_that("compute_multi_framework_total excludes NA from mean", {
  per_framework <- list(
    testthat = list(total_cov = 50),
    testit = list(total_cov = NA_real_)
  )
  expect_equal(compute_multi_framework_total(per_framework), 50)
})

test_that("compute_multi_framework_total returns NaN when all total_cov are NA", {
  per_framework <- list(
    testthat = list(total_cov = NA_real_),
    testit = list(total_cov = NA_real_)
  )
  result <- compute_multi_framework_total(per_framework)
  expect_equal(result, 0)
})

test_that("compute_multi_framework_total handles single framework", {
  per_framework <- list(testthat = list(total_cov = 100))
  expect_equal(compute_multi_framework_total(per_framework), 100)
})

test_that("compute_multi_framework_total returns 0 for framework with invalid total_cov (error path)", {
  per_framework <- list(
    testthat = list(total_cov = 50),
    testit = list(total_cov = list())  # as.numeric(list()) errors -> tryCatch returns 0
  )
  result <- compute_multi_framework_total(per_framework)
  expect_equal(result, 50)  # mean(50, 0)
})

test_that("compute_multi_framework_total handles missing total_cov (excluded as NA)", {
  per_framework <- list(
    testthat = list(total_cov = 50),
    testit = list()  # no total_cov -> x$total_cov is NULL -> as.numeric(NULL) is numeric(0) -> [1L] is NA
  )
  result <- compute_multi_framework_total(per_framework)
  expect_equal(result, 50)  # NA excluded, mean of 50
})

test_that("compute_multi_framework_total with stubbed as.numeric returns 0 when conversions fail", {
  fn <- getFromNamespace("compute_multi_framework_total", "test.assessr")
  mockery::stub(fn, "as.numeric", function(x, ...) stop("stub"))
  per_framework <- list(a = list(total_cov = 100))
  expect_equal(fn(per_framework), 0)
})

test_that("run_one_framework else branch returns zero coverage for unrecognised framework", {
  run_fn <- getFromNamespace("run_one_framework", "test.assessr")
  pkg_source_path <- file.path(tempdir(), "mockpkg_unknown")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  # has_RUnit_test / has_BioG_test required: raw && on NULL errors; FALSE keeps final else branch
  test_pkg_data <- list(
    has_testthat = FALSE,
    has_testit = FALSE,
    has_RUnit_test = FALSE,
    has_BioG_test = FALSE
  )
  
  mockery::stub(run_fn, "safe_unload_package", function(pkg) NULL)
  mockery::stub(run_fn, "initialise_covr_env", function(pkg_source_path, pkg, framework, ...) {
    env <- new.env(parent = baseenv())
    env$work_dir <- tempdir()
    env$pkg_root <- pkg_source_path
    env
  })
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) NULL)
  
  expect_message(
    result <- run_fn(
      framework = "runit",
      pkg = "mockpkg",
      test_pkg_data = test_pkg_data,
      pkg_source_path = pkg_source_path,
      covr_timeout = 10
    ),
    "No recognised standard or non-standard testing configuration"
  )
  
  expect_equal(result$total_cov, 0)
  expect_equal(result$res_cov$errors, "No recognised standard or non-standard testing configuration")
  expect_equal(result$test_framework_type, "non-standard testing framework")
  expect_equal(result$framework_id, "runit")
})


test_that("run_one_framework selects RUnit non-standard testing branch", {
  pkg <- "dummyPkg"
  pkg_source_path <- tempdir()
  
  test_pkg_data <- list(
    has_RUnit_test = TRUE,
    has_BioG_test  = FALSE
  )
  
  # framework must NOT match any earlier branches, so use something generic
  framework <- "some_framework"
  
  # Mock safe_unload_package to avoid side effects
  mockery::stub(run_one_framework, "safe_unload_package", function(...) NULL)
  
  # Mock initialise_covr_env so cov_env becomes predictable
  mockery::stub(run_one_framework, "initialise_covr_env", function(...) list(env = "FAKE_ENV"))
  
  # Mock run_covr_skip_runit_nstf to return fake coverage list
  mockery::stub(run_one_framework, "run_covr_skip_runit_nstf", function(...) {
    list(total_cov = 42, res_cov = list())
  })
  
  # Mock cleanup in on.exit() to avoid real cleanup
  mockery::stub(run_one_framework, "cleanup_and_return_null", function(...) NULL)
  
  result <- run_one_framework(
    framework       = framework,
    pkg             = pkg,
    test_pkg_data   = test_pkg_data,
    pkg_source_path = pkg_source_path,
    covr_timeout    = 10
  )
  
  # Assertions
  expect_equal(result$total_cov, 42)
  expect_equal(result$test_framework_type, "non-standard testing framework - RUnit")
  expect_equal(result$framework_id, framework)
})


test_that("run_one_framework uses create_empty_covr_list when covr_list is NULL", {
  run_fn <- getFromNamespace("run_one_framework", "test.assessr")
  
  pkg_source_path <- file.path(tempdir(), "mockpkg_null")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  test_pkg_data <- list(has_testit = TRUE)
  
  # -- Stubs/mocks on run_one_framework's direct callees --
  mockery::stub(run_fn, "safe_unload_package", function(pkg) NULL)
  
  mockery::stub(run_fn, "initialise_covr_env",
                function(pkg_source_path, pkg, framework, ...) {
                  env <- new.env(parent = baseenv())
                  env$work_dir <- tempdir()
                  env$pkg_root <- pkg_source_path
                  env
                }
  )
  
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) NULL)
  
  # Force the primary coverage attempt to "fail" (NA) so we enter the STF fallback
  mockery::stub(run_fn, "run_coverage",
                function(pkg_source_path, covr_timeout, ...) {
                  list(
                    total_cov = NA_real_,
                    res_cov   = list(coverage = list(filecoverage = c(NA_real_, NA_real_)))
                  )
                }
  )
  
  # CRITICAL: for testit you now call run_covr_skip_stf(), not ..._nstf()
  # Return NULL to force run_one_framework() to use create_empty_covr_list()
  mockery::stub(run_fn, "run_covr_skip_stf", function(...) NULL)
  
  # (You can keep this if other tests depend on it; harmless here)
  mockery::stub(run_fn, "run_covr_skip_nstf", function(...) NULL)
  
  result <- run_fn(
    framework      = "testit",
    pkg            = "mockpkg",
    test_pkg_data  = test_pkg_data,
    pkg_source_path = pkg_source_path,
    covr_timeout   = 10
  )
  
  # Assertions: create_empty_covr_list() path was used
  expect_equal(result$total_cov, 0)
  expect_match(result$res_cov$errors, "Coverage failed for framework: testit")
  expect_equal(result$framework_id, "testit")
})

empty_pkg_flags <- function() {
  list(
    has_testthat = FALSE,
    has_testit = FALSE,
    has_tests_base = FALSE,
    has_tinytest = FALSE,
    has_data_table_tests = FALSE,
    has_BioG_test = FALSE,
    has_RUnit_test = FALSE
  )
}

test_that("detect_frameworks adds runit when has_RUnit_test is TRUE", {
  fn <- getFromNamespace("detect_frameworks", "test.assessr")
  d <- empty_pkg_flags()
  d$has_RUnit_test <- TRUE
  expect_equal(fn(d), "runit")
})

test_that("detect_frameworks RUnit branch is skipped when has_RUnit_test is FALSE or NULL", {
  fn <- getFromNamespace("detect_frameworks", "test.assessr")
  d <- empty_pkg_flags()
  d$has_RUnit_test <- FALSE
  expect_equal(fn(d), character(0))
  d2 <- empty_pkg_flags()
  d2$has_RUnit_test <- NULL
  expect_equal(fn(d2), character(0))
})

test_that("detect_frameworks keeps bioconductor and runit distinct when one flag TRUE", {
  fn <- getFromNamespace("detect_frameworks", "test.assessr")
  d <- empty_pkg_flags()
  d$has_RUnit_test <- TRUE
  d$has_BioG_test <- FALSE
  expect_setequal(fn(d), "runit")
  expect_length(fn(d), 1L)
})

test_that("detect_frameworks combines RUnit with other frameworks", {
  fn <- getFromNamespace("detect_frameworks", "test.assessr")
  d <- empty_pkg_flags()
  d$has_RUnit_test <- TRUE
  d$has_testthat <- TRUE
  out <- fn(d)
  expect_setequal(out, c("testthat", "runit"))
  expect_length(out, 2L)
})

test_that("run_covr_modes passes runit framework when check_pkg has has_RUnit_test (mockery::stub)", {
  pkg_source_path <- file.path(tempdir(), "pkg_runit_detect")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  
  captured_framework <- NULL
  mock_run_one <- function(framework, pkg, test_pkg_data, pkg_source_path, covr_timeout) {
    captured_framework <<- framework
    list(
      total_cov = 99,
      res_cov = list(
        name = pkg,
        coverage = list(
          filecoverage = matrix(1, nrow = 1, dimnames = list("runit_fn")),
          totalcoverage = 99
        )
      )
    )
  }
  
  mockery::stub(run_covr_modes, "check_pkg_tests_and_snaps", function(path, ...) {
    list(
      has_testthat = FALSE,
      has_testit = FALSE,
      has_tests_base = FALSE,
      has_tinytest = FALSE,
      has_data_table_tests = FALSE,
      has_BioG_test = FALSE,
      has_RUnit_test = TRUE,
      has_snaps = FALSE,
      bioc_unit_tests_dir = NA_character_,
      n_golden_tests = 0L,
      n_test_files = 0L
    )
  })
  mockery::stub(run_covr_modes, "run_one_framework", mock_run_one)
  mockery::stub(run_covr_modes, "get_pkg_name", function(path, ...) "mockpkg")
  mockery::stub(run_covr_modes, "get_pkg_desc",
                function(path, fields, ...) list(Package = "mockpkg", Version = "1.0.0")
  )
  mockery::stub(run_covr_modes, "get_test_metadata",
                function(...) {
                  list(
                    datetime = "2026-01-01 12:00:00",
                    executor = "test",
                    info = list(sys = list(
                      sysname = "Linux", version = "1", release = "1", machine = "x86_64"
                    ))
                  )
                }
  )
  
  result <- run_covr_modes(pkg_source_path)
  
  expect_equal(captured_framework, "runit")
  expect_equal(result$total_cov, 99)
})



test_that("decider chooses testit NSTF when testit driver detected", {
  pkg <- tempdir()
  
  mockery::stub(run_covr_decider, "normalizePath", function(x, ...) x)
  
  # Simulate testit driver text in test-all.R
  mockery::stub(run_covr_decider, "readLines", function(path, ...) {
    if (grepl("test-all\\.R$", path)) {
      return("testit::test_pkg('mypkg')")
    }
    character()
  })
  
  # Avoid DT-style: test-all.R does NOT exist
  mockery::stub(run_covr_decider, "file.exists", function(path) FALSE)
  
  # testit directory exists, testthat does not
  mockery::stub(run_covr_decider, "dir.exists", function(path) {
    grepl("testit$", path)
  })
  
  # Mock the NSTF call
  mockery::stub(run_covr_decider, "run_covr_skip_nstf", function(...) "FAKE_NSTF_RESULT")
  
  res <- run_covr_decider(
    framework       = "testit",
    pkg_source_path = pkg,
    test_pkg_data   = list(),
    cov_env         = list(),
    covr_timeout    = 10
  )
  
  expect_equal(res$covr_list, "FAKE_NSTF_RESULT")
  expect_equal(res$tf_type, "non-standard testing framework - testit")
})


test_that("decider chooses testthat branch when testthat structure exists", {
  pkg <- tempdir()
  
  mockery::stub(run_covr_decider, "normalizePath", function(x, ...) x)
  
  # No driver content
  mockery::stub(run_covr_decider, "readLines", function(...) character())
  
  # testthat directory exists
  mockery::stub(run_covr_decider, "dir.exists", function(path) {
    grepl("testthat$", path)
  })
  
  # Avoid DT-style: test-all.R does not exist
  mockery::stub(run_covr_decider, "file.exists", function(...) FALSE)
  
  # Mock run_coverage
  mockery::stub(run_covr_decider, "run_coverage", function(...) "FAKE_COV_RESULT")
  
  res <- run_covr_decider(
    framework       = "testit",
    pkg_source_path = pkg,
    test_pkg_data   = list(),
    cov_env         = list(),
    covr_timeout    = 10
  )
  
  expect_equal(res$covr_list, "FAKE_COV_RESULT")
  expect_equal(res$tf_type, "standard testing framework")
})

# Helper: minimal mapping DF used by downstream steps
make_mapping <- function() {
  data.frame(
    source_file = c("a.R", "b.R"),
    test_file   = c("test-a.R", NA_character_),  # one with a test, one without
    evidence    = c("filename_match", NA_character_),
    has_tests   = c(TRUE, FALSE),
    stringsAsFactors = FALSE
  )
}

test_that("diagnostic: coverage block functions are invoked (covr + reducer)", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  # Same setup as above (mapping, normalize, prepare, etc.)
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_bioc_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_biocgenerics_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE) x)
  
  mapping <- data.frame(
    source_file = "a.R", test_file = "test-a.R", evidence = "filename_match",
    has_tests = TRUE, stringsAsFactors = FALSE
  )
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf", mockery::mock(mapping))
  mockery::stub(fn, "prepare_for_bioc_runit",     mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_bioc_nstf",  mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests",      mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na",             function(df) df)
  
  mockery::stub(fn, "ls",               mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace",      mockery::mock(new.env()))
  mockery::stub(fn, "assign",           mockery::mock(NULL))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace",    mockery::mock(NULL))
  # Must stub library() or you’ll abort here!
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "library",          mockery::mock(NULL))
  mockery::stub(fn, "tempfile",         mockery::mock("/tmp/runner.R"))
  mockery::stub(fn, "writeLines",       mockery::mock(NULL))
  
  
  # avoid error from as.environment("package:mypkg")
  mockery::stub(fn, "as.environment", function(name) {
    # when calling as.environment("package:mypkg"), return a fresh env
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    # delegate other calls to base (rare here)
    base::as.environment(name)
  })
  
  
  # Spy/record calls to covr & reducer
  spy_env_cov <- mockery::mock(structure(list(), class = "covr_coverage"))
  spy_cov_to_list <- mockery::mock(list(totalcoverage = 75.0))
  spy_compute <- mockery::mock(list(
    total_cov = 0.50, percent_cov = 75.0,
    n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L
  ))
  
  mockery::stub(fn, "covr::environment_coverage", spy_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",     spy_cov_to_list)
  mockery::stub(fn, "compute_total_coverage",     spy_compute)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  mockery::expect_called(spy_env_cov,   1)
  mockery::expect_called(spy_cov_to_list, 1)
  mockery::expect_called(spy_compute,   1)
  
  expect_identical(res$total_cov, 0.50)
  expect_identical(res$res_cov$coverage$totalcoverage, 75.0)
})

test_that("messages when package is not using BioC RUnit layout and returns default coverage when covr fails", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # --- Stubs for early steps ---
  mock_get_pkg_name <- mockery::mock(pkg_name)
  mock_is_layout    <- mockery::mock(FALSE)  # trigger layout message
  
  mock_get_paths <- mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests")))
  mock_normalize <- mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE)
  
  # --- Mapping: succeed to progress into coverage path ---
  mock_get_mapping <- mockery::mock(make_mapping())
  
  # --- Cleaning branch: only run when bioc_run_ut_path is not NA ---
  # In this test, set bioc_run_ut_path = NA so the cleaning branch is skipped
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  # --- prepare_for_bioc_runit: succeed ---
  mock_prepare <- mockery::mock(TRUE)
  
  # --- Problems: return "No tests skipped" (default/no skip) ---
  mock_problems <- mockery::mock(data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  ))
  
  # --- Post-load mapping: functions with no tests ---
  mock_no_tests <- mockery::mock(data.frame(
    `function` = "b",
    stringsAsFactors = FALSE
  ))
  
  # --- tidyr::drop_na: identity (no change) ---
  mock_drop_na <- function(df) df
  
  # --- Env preparation (assign exported/internal objects + datasets) ---
  mock_ls <- mockery::mock(c("obj1", "obj2"))
  mock_asNamespace <- mockery::mock(new.env())
  mock_assign <- mockery::mock(NULL)
  mock_load_datasets <- mockery::mock(c("ds1", "ds2"))
  mock_message <- mockery::mock(NULL, cycle = TRUE)
  
  # --- Coverage block stubs: force an error in covr to take default path ---
  mock_loadedNamespaces <- mockery::mock(c("base", "stats", "utils", "methods"))  # so loadNamespace("methods") skipped
  mock_loadNamespace <- mockery::mock(NULL)
  mock_detach <- mockery::mock(TRUE)
  mock_library <- mockery::mock(NULL)
  mock_tempfile <- mockery::mock("/tmp/runner.R")
  mock_writeLines <- mockery::mock(NULL)
  
  # This triggers the error branch -> default coverage result
  mock_env_cov <- function(env, test_files) stop("instrumentation failed")
  # coverage_to_list is never reached when env_cov errors, but stub for safety
  mock_cov_to_list <- mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0))
  
  # --- cleanup_and_return_null used by error branches ---
  mock_cleanup <- mockery::mock(NULL)
  
  # Wire stubs
  mockery::stub(fn, "get_pkg_name",               mock_get_pkg_name)
  mockery::stub(fn, "is_bioc_runit_layout",       mock_is_layout)
  mockery::stub(fn, "get_biocgenerics_test_paths",mock_get_paths)
  mockery::stub(fn, "normalizePath",              mock_normalize)
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf", mock_get_mapping)
  mockery::stub(fn, "prepare_for_bioc_runit",     mock_prepare)
  mockery::stub(fn, "check_covr_skip_bioc_nstf",  mock_problems)
  mockery::stub(fn, "get_function_no_tests",      mock_no_tests)
  mockery::stub(fn, "tidyr::drop_na",             mock_drop_na)
  mockery::stub(fn, "ls",                         mock_ls)
  mockery::stub(fn, "asNamespace",                mock_asNamespace)
  mockery::stub(fn, "assign",                     mock_assign)
  mockery::stub(fn, "load_package_datasets",      mock_load_datasets)
  mockery::stub(fn, "message",                    mock_message)
  mockery::stub(fn, "loadedNamespaces",           mock_loadedNamespaces)
  mockery::stub(fn, "loadNamespace",              mock_loadNamespace)
  mockery::stub(fn, "detach_pkg_if_attached",     mock_detach)
  mockery::stub(fn, "library",                    mock_library)
  mockery::stub(fn, "tempfile",                   mock_tempfile)
  mockery::stub(fn, "writeLines",                 mock_writeLines)
  mockery::stub(fn, "covr::environment_coverage", mock_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",     mock_cov_to_list)
  mockery::stub(fn, "cleanup_and_return_null",    mock_cleanup)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  # Assertions: default coverage path
  expect_type(res, "list")
  expect_true("res_cov" %in% names(res))
  expect_true(is.list(res$res_cov))
  expect_identical(res$total_cov, 0)
  expect_identical(res$res_cov$name, pkg_name)
  expect_true(is.list(res$res_cov$coverage))
  expect_identical(res$res_cov$coverage$totalcoverage, 0)
  expect_identical(res$res_cov$errors, "No RUnit or BiocGenerics configuration")
  
  # The first message should hint about BioC RUnit layout
  msgs <- mockery::mock_args(mock_message)
  # Flatten first invocation into a single string to check substring
  first_msg <- paste0(msgs[[1]], collapse = "")
  expect_match(first_msg, "BioC RUnit layout")
})


test_that("returns NULL when source-test mapping errors", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_bioc_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_biocgenerics_test_paths", mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath",        mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE))
  # Mapping: throw error so tryCatch invokes cleanup_and_return_null -> returns NULL -> early exit
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf", function(...) stop("boom-mapping"))
  mockery::stub(fn, "cleanup_and_return_null",           mockery::mock(NULL))
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  expect_null(res)
})


test_that("returns NULL when prepare_for_bioc_runit fails", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_bioc_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_biocgenerics_test_paths", mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath",        mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE))
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf", mockery::mock(make_mapping()))
  # Force prepare to error; tryCatch should call cleanup_and_return_null and return NULL
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) stop("prep-error"))
  mockery::stub(fn, "cleanup_and_return_null", mockery::mock(NULL))
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  expect_null(res)
})


test_that("coverage error path yields a default covr_list even when mapping & problems succeed", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)  # skip cleaning branch
  
  # Layout ok
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_bioc_runit_layout", mockery::mock(TRUE))
  
  # Paths, normalization
  mockery::stub(fn, "get_biocgenerics_test_paths", mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath",        mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE))
  
  # Mapping and later pipeline
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf", mockery::mock(make_mapping()))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_bioc_nstf", mockery::mock(data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(`function` = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na",        function(df) df)
  
  # Env prep for coverage
  mockery::stub(fn, "ls",               mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace",      mockery::mock(new.env()))
  mockery::stub(fn, "assign",           mockery::mock(NULL))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "message",          mockery::mock(NULL, cycle = TRUE))
  
  # Coverage: trigger failure to default path
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace",    mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "library",          mockery::mock(NULL))
  mockery::stub(fn, "tempfile",         mockery::mock("/tmp/runner.R"))
  mockery::stub(fn, "writeLines",       mockery::mock(NULL))
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("covr-failed"))
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0)))
  mockery::stub(fn, "cleanup_and_return_null", mockery::mock(NULL))
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  expect_type(res, "list")
  expect_identical(res$total_cov, 0)
  expect_identical(res$res_cov$name, pkg_name)
  expect_identical(res$res_cov$coverage$totalcoverage, 0)
  expect_identical(res$res_cov$errors, "No RUnit or BiocGenerics configuration")
})

# Minimal helper mapping (one test file to pass mustWork normalization)
make_mapping <- function() {
  data.frame(
    source_file = c("a.R", "b.R"),
    test_file   = c("test-a.R", NA_character_),
    evidence    = c("filename_match", NA_character_),
    has_tests   = c(TRUE, FALSE),
    stringsAsFactors = FALSE
  )
}




test_that("error in coverage_to_list hits error branch but lines still produce messages and covr_list", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  # Upstream stubs to reach the block
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_bioc_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_biocgenerics_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE) x)
  
  # Mapping with one test to ensure test_files_clean is non-empty
  make_mapping <- function() {
    data.frame(
      source_file = c("a.R", "b.R"),
      test_file   = c("test-a.R", NA_character_),
      evidence    = c("filename_match", NA_character_),
      has_tests   = c(TRUE, FALSE),
      stringsAsFactors = FALSE
    )
  }
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf", mockery::mock(make_mapping()))
  
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_bioc_nstf", mockery::mock(
    data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)
  ))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  
  # Env prep / coverage stubs
  mockery::stub(fn, "ls",               mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace",      mockery::mock(new.env()))
  mockery::stub(fn, "assign",           mockery::mock(NULL))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace",    mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "library",          mockery::mock(NULL))
  mockery::stub(fn, "tempfile",         mockery::mock("/tmp/runner.R"))
  mockery::stub(fn, "writeLines",       mockery::mock(NULL))
  
  # keep coverage block alive by returning a benign env for "package:mypkg"
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) {
      return(new.env(parent = emptyenv()))
    }
    base::as.environment(name)
  })
  
  # Force error in coverage_to_list -> error branch sets res_cov$errors and NA coverage fields
  mockery::stub(fn, "covr::environment_coverage", mockery::mock(structure(list(), class = "covr_coverage")))
  mockery::stub(fn, "covr::coverage_to_list", function(...) stop("explode"))
  
  # Stub compute_total_coverage to still return values (so the lines proceed)
  mockery::stub(fn, "compute_total_coverage", function(percent_cov, ...) {
    # percent_cov will be NA_real_ due to error branch; ignore and return defaults
    list(
      total_cov            = 0.00,
      percent_cov          = 0.00,
      n_testfiles          = 1L,
      n_no_function_tests  = 1L,
      n_skipped_files      = 0L
    )
  })
  
  # Capture messages & cleanup calls
  mock_msg   <- mockery::mock(NULL, cycle = TRUE)
  mock_clean <- mockery::mock(NULL)
  mockery::stub(fn, "message",                 mock_msg)
  mockery::stub(fn, "cleanup_and_return_null", mock_clean)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  # Messages exist (Total Coverage 0.00%, Approx total coverage 0.00%)
  msg_calls <- mockery::mock_args(mock_msg)
  all_msgs  <- vapply(msg_calls, function(x) paste0(x, collapse = ""), character(1))
  total_line  <- all_msgs[grepl("^Total Coverage for mypkg is:", all_msgs)]
  approx_line <- all_msgs[grepl("^Approx total coverage:",       all_msgs)]
  
  expect_true(length(total_line)  >= 1)
  expect_true(length(approx_line) >= 1)
  expect_match(total_line[1],  "0%")
  expect_match(approx_line[1], "\\(0\\.00%\\)")
  
  # cleanup called once
  mockery::expect_called(mock_clean, 1)
  
  # covr_list assembled
  expect_type(res, "list")
  expect_true("res_cov" %in% names(res))
  # In error branch, res_cov$coverage$'totalcoverage' == NA_integer_
  expect_true(is.na(res$res_cov$coverage$totalcoverage))
  expect_true(inherits(res$res_cov$errors, "error") || is.character(res$res_cov$errors))
})


test_that("returns TRUE when test_pkg_data$has_BioG_test is TRUE regardless of files", {
  fn <- is_bioc_runit_layout
  
  # Mock filesystem as FALSE to ensure the first clause is sufficient
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(pkg_source_path = "/fake/path", test_pkg_data = list(has_BioG_test = TRUE))
  expect_true(res)
  
  # Both stubs should still be called once due to evaluation sequence of ||
  # (In R, || is short-circuiting, but since the first is TRUE, the others may not be called.
  # To be robust across implementations, we don't assert call counts here.)
})

test_that("returns FALSE when no BioC-RUnit indicators are present", {
  fn <- is_bioc_runit_layout
  
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  # test_pkg_data NULL or missing field -> isTRUE(NULL) => FALSE
  res1 <- fn(pkg_source_path = "/fake/path", test_pkg_data = NULL)
  expect_false(res1)
  
})





test_that("detaches target package and BioC infra packages in order", {
  fn <- prepare_for_bioc_runit
  
  # Mock the detacher to record calls
  mock_detach <- mockery::mock(NULL, NULL, NULL, NULL)  # 4 total calls expected
  
  # Mock load_all to prevent real loading
  mock_load_all <- mockery::mock(NULL)
  
  # Stub functions called inside prepare_for_bioc_runit
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  # Call with explicit detach_bioc_infra = TRUE since it's now a parameter
  res <- fn(
    pkg_name = "mypkg",
    pkg_source_path = "/tmp/mypkg",
    detach_bioc_infra = TRUE
  )
  
  expect_true(res)
  
  # Expect 4 detach calls in order: mypkg, IRanges, S4Vectors, BiocGenerics
  mockery::expect_called(mock_detach, 4)
  calls <- mockery::mock_args(mock_detach)
  
  expect_identical(calls[[1]][[1]], "mypkg")
  expect_identical(calls[[2]][[1]], "IRanges")
  expect_identical(calls[[3]][[1]], "S4Vectors")
  expect_identical(calls[[4]][[1]], "BiocGenerics")
  
  # Verify load_all was called with the correct parameters
  mockery::expect_called(mock_load_all, 1)
  la_args <- mockery::mock_args(mock_load_all)
  
  expect_identical(la_args[[1]][[1]], "/tmp/mypkg")
  expect_true(isTRUE(la_args[[1]][["quiet"]]))
})



test_that("options(warn) is temporarily set to 1 and restored after function exits", {
  fn <- prepare_for_bioc_runit
  
  # Mocks to avoid side effects
  mock_detach <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  # Preserve original warn and auto-restore within test using withr
  old_warn <- getOption("warn")
  withr::local_options(list(warn = old_warn))
  
  expect_identical(getOption("warn"), old_warn)  # sanity check
  
  # Must now pass detach_bioc_infra explicitly
  res <- fn(
    pkg_name = "mypkg",
    pkg_source_path = "/tmp/mypkg",
    detach_bioc_infra = TRUE
  )
  
  expect_true(res)
  
  # After fn() returns, warn should be restored
  expect_identical(getOption("warn"), old_warn)
})


test_that('R_TESTS is empty during test (local_envvar)', {
  is_windows <- identical(.Platform$OS.type, "windows")
  
  # Ensure clean slate first
  Sys.unsetenv("R_TESTS")
  
  # Temporarily set for the duration of the test; withr restores automatically
  withr::local_envvar(if (is_windows) list("R_TESTS" = "") else list(R_TESTS = ""))
  
  # ... run things that depend on R_TESTS="", or assertions ...
  expect_identical(Sys.getenv("R_TESTS"), "")
  
})




test_that("POSIX branch: sets R_TESTS='' and restores to UNSET when originally unset", {
  # Only run on Linux/macOS to exercise the POSIX branch
  skip_if(
    condition = .Platform$OS.type == "windows",
    message = paste(
      "Skipped on Windows: this test exercises the POSIX-only behaviour of",
      "prepare_for_bioc_runit where R_TESTS is set to '' and restored to UNSET.",
      "Windows does not reliably distinguish empty vs UNSET environment variables."
    )
  )
  
  fn <- prepare_for_bioc_runit
  
  # Mock collaborators
  mock_detach   <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  # Ensure R_TESTS is UNSET before the call
  Sys.unsetenv("R_TESTS")
  
  # Always restore to UNSET at end of test
  withr::defer(Sys.unsetenv("R_TESTS"))
  
  # Sanity check: must be UNSET
  expect_identical(
    Sys.getenv("R_TESTS", unset = NA_character_),
    NA_character_
  )
  
  # Call function; must pass detach_bioc_infra explicitly
  res <- fn(
    pkg_name = "mypkg",
    pkg_source_path = "/tmp/mypkg",
    detach_bioc_infra = TRUE
  )
  expect_true(res)
  
})




test_that("POSIX branch: sets R_TESTS='' and restores to previous value when originally set (mocked)", {
  
  skip_if(
    condition = .Platform$OS.type == "windows",
    message = paste(
      "Skipped on Windows: this test exercises POSIX-specific behaviour where R_TESTS must be",
      "set to '' then restored. Windows environment handling cannot reliably distinguish",
      "empty vs unset values."
    )
  )
  
  fn <- prepare_for_bioc_runit
  
  # --- Seed original R_TESTS value ---
  original <- "previous-value"
  Sys.setenv(R_TESTS = original)
  
  # --- Mocks ---
  # Sys.setenv should be called twice:
  #   1. R_TESTS = ""
  #   2. R_TESTS = original
  mock_setenv   <- mockery::mock(NULL, NULL)
  
  # Sys.unsetenv should NOT be called for "originally set" case
  mock_unsetenv <- mockery::mock()
  
  # Other collaborators
  mock_detach   <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "Sys.setenv",              mock_setenv)
  mockery::stub(fn, "Sys.unsetenv",            mock_unsetenv)
  mockery::stub(fn, "detach_pkg_if_attached",  mock_detach)
  mockery::stub(fn, "pkgload::load_all",       mock_load_all)
  
  # --- Execute ---
  res <- fn(
    pkg_name         = "mypkg",
    pkg_source_path  = "/tmp/mypkg",
    detach_bioc_infra = TRUE
  )
  expect_true(res)
  
  # --- Assertions ---
  set_calls <- mockery::mock_args(mock_setenv)
  
  # Expect first call: Sys.setenv(R_TESTS = "")
  expect_identical(
    set_calls[[1]],
    list(R_TESTS = "")
  )
  
  # Sys.unsetenv must NOT be called in this branch
  mockery::expect_called(mock_unsetenv, 0)
})




test_that("run_covr_skip_bioc_nstf: cleaning branch -> happy path updates mapping and proceeds", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # --- Stubs for early steps ---
  mock_get_pkg_name <- mockery::mock(pkg_name)
  mock_is_layout    <- mockery::mock(FALSE)  # trigger layout message path
  mock_get_paths    <- mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests")))
  mock_normalize    <- mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE)
  
  # Mapping before cleaning; keep simple to avoid make_mapping() dependency
  mock_get_mapping  <- mockery::mock(list(files = c("a.R", "b.R")))
  
  # --- Cleaning branch enabled: non-NA path ---
  test_pkg_data <- list(bioc_run_ut_path = file.path(pkg_source_path, "inst", "unitTests", "runTests.R"))
  
  # On happy path, cleaning returns a *non-NULL* updated mapping
  mock_remove_map <- mockery::mock(list(files = c("a.R")))
  
  # --- prepare_for_bioc_runit: succeed ---
  mock_prepare <- mockery::mock(TRUE)
  
  # --- Problems: default/no skip ---
  mock_problems <- mockery::mock(data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  ))
  
  # --- Post-load mapping: functions with no tests ---
  mock_no_tests <- mockery::mock(data.frame(`function` = "b", stringsAsFactors = FALSE))
  
  # --- tidyr::drop_na: identity (no change) ---
  mock_drop_na <- function(df) df
  
  # --- Env prep ---
  mock_ls            <- mockery::mock(c("obj1", "obj2"))
  mock_asNamespace   <- mockery::mock(new.env())
  mock_assign        <- mockery::mock(NULL)
  mock_load_datasets <- mockery::mock(c("ds1", "ds2"))
  mock_message       <- mockery::mock(NULL, cycle = TRUE)
  
  # --- Coverage block stubs: force covr error -> default coverage result ---
  mock_loadedNamespaces <- mockery::mock(c("base", "stats", "utils", "methods"))
  mock_loadNamespace    <- mockery::mock(NULL)
  mock_detach           <- mockery::mock(TRUE)
  mock_library          <- mockery::mock(NULL)
  mock_tempfile         <- mockery::mock("/tmp/runner.R")
  mock_writeLines       <- mockery::mock(NULL)
  mock_env_cov          <- function(env, test_files) stop("instrumentation failed")
  mock_cov_to_list      <- mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0))
  
  # --- cleanup called by error branches (should NOT be called here) ---
  mock_cleanup <- mockery::mock(NULL)
  
  # Wire stubs
  mockery::stub(fn, "get_pkg_name",                        mock_get_pkg_name)
  mockery::stub(fn, "is_bioc_runit_layout",                mock_is_layout)
  mockery::stub(fn, "get_biocgenerics_test_paths",         mock_get_paths)
  mockery::stub(fn, "normalizePath",                       mock_normalize)
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf",   mock_get_mapping)
  mockery::stub(fn, "remove_run_unitTests_mapping",        mock_remove_map)
  mockery::stub(fn, "prepare_for_bioc_runit",              mock_prepare)
  mockery::stub(fn, "check_covr_skip_bioc_nstf",           mock_problems)
  mockery::stub(fn, "get_function_no_tests",               mock_no_tests)
  mockery::stub(fn, "tidyr::drop_na",                      mock_drop_na)
  mockery::stub(fn, "ls",                                  mock_ls)
  mockery::stub(fn, "asNamespace",                         mock_asNamespace)
  mockery::stub(fn, "assign",                              mock_assign)
  mockery::stub(fn, "load_package_datasets",               mock_load_datasets)
  mockery::stub(fn, "message",                             mock_message)
  mockery::stub(fn, "loadedNamespaces",                    mock_loadedNamespaces)
  mockery::stub(fn, "loadNamespace",                       mock_loadNamespace)
  mockery::stub(fn, "detach_pkg_if_attached",              mock_detach)
  mockery::stub(fn, "library",                             mock_library)
  mockery::stub(fn, "tempfile",                            mock_tempfile)
  mockery::stub(fn, "writeLines",                          mock_writeLines)
  mockery::stub(fn, "covr::environment_coverage",          mock_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",              mock_cov_to_list)
  mockery::stub(fn, "cleanup_and_return_null",             mock_cleanup)
  
  # Execute
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  # Assertions specific to the cleaning branch
  
  # Ensure the cleaning message was emitted
  msg_args <- mockery::mock_args(mock_message)
  expect_true(any(vapply(msg_args, function(x) {
    any(grepl(paste0("performing source test map cleaning for ", pkg_name),
              paste0(unlist(x), collapse = "")))
  }, logical(1))))
  
  
  # Function should proceed (non-NULL result expected given our covr stubs)
  expect_false(is.null(res))
})  


test_that("run_covr_skip_bioc_nstf: cleaning branch -> error calls cleanup_and_return_null and returns NULL", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # --- Stubs for early steps ---
  mock_get_pkg_name <- mockery::mock(pkg_name)
  mock_is_layout    <- mockery::mock(FALSE)
  mock_get_paths    <- mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests")))
  mock_normalize    <- mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE)
  mock_get_mapping  <- mockery::mock(list(files = c("a.R", "b.R")))
  
  # --- Cleaning branch enabled ---
  test_pkg_data <- list(bioc_run_ut_path = file.path(pkg_source_path, "inst", "unitTests", "runTests.R"))
  
  # remove_run_unitTests_mapping throws -> tryCatch error handler runs
  mock_remove_map <- mockery::mock(stop("boom"))
  mock_cleanup     <- mockery::mock(NULL)  # will be called by error handler
  
  # --- Prepare rest (won't be reached if function returns NULL early) ---
  mock_prepare <- mockery::mock(TRUE)
  mock_problems <- mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE))
  mock_no_tests <- mockery::mock(data.frame(`function` = "b", stringsAsFactors = FALSE))
  mock_drop_na  <- function(df) df
  mock_ls            <- mockery::mock(c("obj1", "obj2"))
  mock_asNamespace   <- mockery::mock(new.env())
  mock_assign        <- mockery::mock(NULL)
  mock_load_datasets <- mockery::mock(c("ds1", "ds2"))
  mock_message       <- mockery::mock(NULL, cycle = TRUE)
  mock_loadedNamespaces <- mockery::mock(c("base", "stats", "utils", "methods"))
  mock_loadNamespace    <- mockery::mock(NULL)
  mock_detach           <- mockery::mock(TRUE)
  mock_library          <- mockery::mock(NULL)
  mock_tempfile         <- mockery::mock("/tmp/runner.R")
  mock_writeLines       <- mockery::mock(NULL)
  mock_env_cov          <- function(env, test_files) stop("instrumentation failed")
  mock_cov_to_list      <- mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0))
  
  # Wire stubs
  mockery::stub(fn, "get_pkg_name",                        mock_get_pkg_name)
  mockery::stub(fn, "is_bioc_runit_layout",                mock_is_layout)
  mockery::stub(fn, "get_biocgenerics_test_paths",         mock_get_paths)
  mockery::stub(fn, "normalizePath",                       mock_normalize)
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf",   mock_get_mapping)
  mockery::stub(fn, "remove_run_unitTests_mapping",        mock_remove_map)
  mockery::stub(fn, "prepare_for_bioc_runit",              mock_prepare)
  mockery::stub(fn, "check_covr_skip_bioc_nstf",           mock_problems)
  mockery::stub(fn, "get_function_no_tests",               mock_no_tests)
  mockery::stub(fn, "tidyr::drop_na",                      mock_drop_na)
  mockery::stub(fn, "ls",                                  mock_ls)
  mockery::stub(fn, "asNamespace",                         mock_asNamespace)
  mockery::stub(fn, "assign",                              mock_assign)
  mockery::stub(fn, "load_package_datasets",               mock_load_datasets)
  mockery::stub(fn, "message",                             mock_message)
  mockery::stub(fn, "loadedNamespaces",                    mock_loadedNamespaces)
  mockery::stub(fn, "loadNamespace",                       mock_loadNamespace)
  mockery::stub(fn, "detach_pkg_if_attached",              mock_detach)
  mockery::stub(fn, "library",                             mock_library)
  mockery::stub(fn, "tempfile",                            mock_tempfile)
  mockery::stub(fn, "writeLines",                          mock_writeLines)
  mockery::stub(fn, "covr::environment_coverage",          mock_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",              mock_cov_to_list)
  mockery::stub(fn, "cleanup_and_return_null",             mock_cleanup)
  
  # Execute
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  # Result should be NULL due to error path + cleanup handler return
  expect_null(res)
  
  # Assert error message composed and passed to cleanup
  cleanup_args <- mockery::mock_args(mock_cleanup)
  expect_true(
    any(vapply(cleanup_args, function(x) {
      is.character(x[[1]]) &&
        grepl(paste0("^Error in cleaning mapping for ", pkg_name, " : boom$"), x[[1]])
    }, logical(1))),
    info = "cleanup_and_return_null should receive composed error message"
  )
  
  # Ensure the cleaning message was emitted
  msg_args <- mockery::mock_args(mock_message)
  
  
  expect_true(
    any(
      vapply(
        msg_args,
        function(x) {
          any(
            grepl(
              paste0("performing source test map cleaning for ", pkg_name),
              paste0(unlist(x), collapse = "")
            )
          )
        },
        logical(1)
      )
    )
  )
  
  
})


test_that("run_covr_skip_bioc_nstf: cleaning branch -> NULL mapping returns NULL early", {
  fn <- run_covr_skip_bioc_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # --- Stubs for early steps ---
  mock_get_pkg_name <- mockery::mock(pkg_name)
  mock_is_layout    <- mockery::mock(FALSE)
  mock_get_paths    <- mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests")))
  mock_normalize    <- mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE)
  mock_get_mapping  <- mockery::mock(list(files = c("a.R", "b.R")))
  
  # --- Cleaning branch enabled ---
  test_pkg_data <- list(bioc_run_ut_path = file.path(pkg_source_path, "inst", "unitTests", "runTests.R"))
  
  # Cleaning returns NULL -> branch should return(NULL) immediately
  mock_remove_map <- mockery::mock(NULL)
  mock_cleanup     <- mockery::mock(NULL)  # not used in this path
  
  # --- Prepare rest (won't be reached if returns early) ---
  mock_prepare <- mockery::mock(TRUE)
  mock_problems <- mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE))
  mock_no_tests <- mockery::mock(data.frame(`function` = "b", stringsAsFactors = FALSE))
  mock_drop_na  <- function(df) df
  mock_ls            <- mockery::mock(c("obj1", "obj2"))
  mock_asNamespace   <- mockery::mock(new.env())
  mock_assign        <- mockery::mock(NULL)
  mock_load_datasets <- mockery::mock(c("ds1", "ds2"))
  mock_message       <- mockery::mock(NULL, cycle = TRUE)
  mock_loadedNamespaces <- mockery::mock(c("base", "stats", "utils", "methods"))
  mock_loadNamespace    <- mockery::mock(NULL)
  mock_detach           <- mockery::mock(TRUE)
  mock_library          <- mockery::mock(NULL)
  mock_tempfile         <- mockery::mock("/tmp/runner.R")
  mock_writeLines       <- mockery::mock(NULL)
  mock_env_cov          <- function(env, test_files) stop("instrumentation failed")
  mock_cov_to_list      <- mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0))
  
  # Wire stubs
  mockery::stub(fn, "get_pkg_name",                        mock_get_pkg_name)
  mockery::stub(fn, "is_bioc_runit_layout",                mock_is_layout)
  mockery::stub(fn, "get_biocgenerics_test_paths",         mock_get_paths)
  mockery::stub(fn, "normalizePath",                       mock_normalize)
  mockery::stub(fn, "get_source_test_mapping_bioc_nstf",   mock_get_mapping)
  mockery::stub(fn, "remove_run_unitTests_mapping",        mock_remove_map)
  mockery::stub(fn, "prepare_for_bioc_runit",              mock_prepare)
  mockery::stub(fn, "check_covr_skip_bioc_nstf",           mock_problems)
  mockery::stub(fn, "get_function_no_tests",               mock_no_tests)
  mockery::stub(fn, "tidyr::drop_na",                      mock_drop_na)
  mockery::stub(fn, "ls",                                  mock_ls)
  mockery::stub(fn, "asNamespace",                         mock_asNamespace)
  mockery::stub(fn, "assign",                              mock_assign)
  mockery::stub(fn, "load_package_datasets",               mock_load_datasets)
  mockery::stub(fn, "message",                             mock_message)
  mockery::stub(fn, "loadedNamespaces",                    mock_loadedNamespaces)
  mockery::stub(fn, "loadNamespace",                       mock_loadNamespace)
  mockery::stub(fn, "detach_pkg_if_attached",              mock_detach)
  mockery::stub(fn, "library",                             mock_library)
  mockery::stub(fn, "tempfile",                            mock_tempfile)
  mockery::stub(fn, "writeLines",                          mock_writeLines)
  mockery::stub(fn, "covr::environment_coverage",          mock_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",              mock_cov_to_list)
  mockery::stub(fn, "cleanup_and_return_null",             mock_cleanup)
  
  # Execute
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  # Should return NULL due to NULL mapping early return
  expect_null(res)
  
  # Ensure cleaning message was emitted
  msg_args <- mockery::mock_args(mock_message)
  expect_true(any(vapply(msg_args, function(x) {
    any(grepl(paste0("performing source test map cleaning for ", pkg_name),
              paste0(unlist(x), collapse = "")))
  }, logical(1))))
})


test_that("returns invisible(TRUE)", {
  fn <- prepare_for_bioc_runit
  
  mock_detach   <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  v <- withVisible(
    fn(
      pkg_name        = "mypkg",
      pkg_source_path = "/tmp/mypkg",
      detach_bioc_infra = TRUE   # required with new signature
    )
  )
  
  expect_true(v$value)
  expect_false(v$visible)   # ensure invisibility
})



test_that("propagates quietly (no messages) and does not error when mocked", {
  fn <- prepare_for_bioc_runit
  
  mock_detach   <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  expect_silent(
    res <- fn(
      pkg_name         = "mypkg",
      pkg_source_path  = "/tmp/mypkg",
      detach_bioc_infra = TRUE   # required with new signature
    )
  )
  
  expect_true(res)
})


test_that("load_all called even if detaches throw; try-catch not inside but we still see error", {
  fn <- prepare_for_bioc_runit
  
  # If a detach throws, since the function doesn't try-catch around it, the error
  # should propagate and load_all should not be called. We can still validate behavior.
  throwing_detach <- function(...) stop("detach failure")
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "detach_pkg_if_attached", throwing_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  expect_error(fn(pkg_name = "mypkg", pkg_source_path = "/tmp/mypkg"), "detach failure")
  # Ensure load_all was not called due to early error
  mockery::expect_called(mock_load_all, 0)
}) 


test_that("errors if RUnit is not available", {
  fn <- check_covr_skip_bioc_nstf
  
  mock_req_ns <- mockery::mock(FALSE)
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  
  expect_error(
    fn(pkg_source_path = "/pkg", mapping = data.frame(test_file = character()), test_path = "/tests"),
    "RUnit must be installed"
  )
  
  mockery::expect_called(mock_req_ns, 1)
  args <- mockery::mock_args(mock_req_ns)
  expect_identical(args[[1]][[1]], "RUnit")
  expect_true(isTRUE(args[[1]][["quietly"]]))
})

test_that("cleans mapping via tidyr::drop_na when test_file contains NA", {
  fn <- check_covr_skip_bioc_nstf
  
  mapping <- data.frame(test_file = c("test-a.R", NA_character_, "test-b.R"), stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  # drop_na should remove the NA row
  mock_drop_na <- function(df, test_file) df[!is.na(df$test_file), , drop = FALSE]
  
  # Minimal stubs to allow the function to proceed to the lapply without real side effects
  mock_file_exists <- mockery::mock(TRUE, TRUE, cycle = TRUE) # two files exist
  mock_search <- mockery::mock("package:stats", "package:stats", cycle = TRUE) # no RUnit attached
  mock_attach <- mockery::mock(NULL, NULL, cycle = TRUE)
  mock_detach <- mockery::mock(NULL, cycle = TRUE)  # may or may not be called by on.exit
  mock_run <- mockery::mock(structure(list(), class = "RUnitTestData"),
                            structure(list(), class = "RUnitTestData"), 
                            cycle = TRUE)
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0),
                                list(nErr = 0, nFail = 0, nDeactivated = 0), 
                                cycle = TRUE)
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", mock_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  # Both tests pass -> "No tests skipped"
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 1L)
  expect_true(is.na(res$test_file[1]))
  expect_identical(res$issue_type[1], "No tests skipped")
})


test_that("removes exact filename from mapping when bioc_run_ut_path provided (case-insensitive)", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("run_unitTests.R", "test-foo.R", "RUN_UNITTESTS.R", "test-bar.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/run_unitTests.R")
  
  # Stub basename() to a specific value to assert exact-match branch is used
  mock_basename <- mockery::mock("run_unitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  
  # Expected: exact "run_unitTests.R" removed via exact-match branch; regex also removes RUN_UNITTESTS.R
  expect_true("test-foo.R" %in% res$test_file)
  expect_true("test-bar.R" %in% res$test_file)
  expect_false("run_unitTests.R" %in% res$test_file)
  expect_false("RUN_UNITTESTS.R" %in% res$test_file)
  
  # basename should be called once with the provided path
  mockery::expect_called(mock_basename, 1)
  args <- mockery::mock_args(mock_basename)
  expect_identical(args[[1]][[1]], "/pkg/tests/run_unitTests.R")
})

test_that("regex fallback removes variations when bioc_run_ut_path is missing/NULL/NA", {
  fn <- remove_run_unitTests_mapping
  
  # No exact name available -> only regex path applies
  mapping <- data.frame(
    test_file = c("run-unitTests.R", "Run_UnitTests.R", "test-x.R", "sub/run_unitTests.R"),
    stringsAsFactors = FALSE
  )
  
  # Case 1: test_pkg_data missing entirely
  res1 <- fn(mapping, test_pkg_data = NULL)
  expect_setequal(res1$test_file, "test-x.R") # note: regex only matches basename
  
  # Case 2: field present but NULL
  res2 <- fn(mapping, test_pkg_data = list(bioc_run_ut_path = NULL))
  expect_setequal(res2$test_file, "test-x.R")
  
  # Case 3: field NA
  res3 <- fn(mapping, test_pkg_data = list(bioc_run_ut_path = NA_character_))
  expect_setequal(res3$test_file, "test-x.R")
})

test_that("does not remove unrelated files; only exact filename or regex variations dropped", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("test-a.R", "runUnitTest.R", "unitTests_run.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/NotRunUnitTests.R")
  
  # basename returns a non-matching value -> exact match won't drop anything
  mock_basename <- mockery::mock("NotRunUnitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  
  # None match the strict regex "run[_-]?unitTests\\.R$"
  expect_setequal(res$test_file, c("test-a.R", "runUnitTest.R", "unitTests_run.R", "test-b.R"))
})

test_that("handles NA entries in mapping$test_file gracefully (no warnings)", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c(NA_character_, "test-a.R", "run_unitTests.R", NA_character_, "test-b.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/whatever.R")
  
  # basename stub not strictly needed but keep deterministic
  mock_basename <- mockery::mock("whatever.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  
  # Only run_unitTests.R is removed; NAs are retained
  expect_setequal(res$test_file, c(NA_character_, "test-a.R", NA_character_, "test-b.R"))
})

test_that("exact match removal works even if mapping has different case than basename", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("RUN_UNITTESTS.R", "test-a.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/run_unitTests.R")
  
  mock_basename <- mockery::mock("run_unitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  expect_setequal(res$test_file, "test-a.R")
})

test_that("regex removes both underscore and hyphen variants, any case", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("run_unitTests.R", "run-unitTests.R", "RUN_UNITTESTS.R", "test-a.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  res <- fn(mapping, test_pkg_data = NULL)
  expect_setequal(res$test_file, c("test-a.R", "test-b.R"))
})

test_that("empty mapping returns empty data.frame (no errors)", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(test_file = character(), stringsAsFactors = FALSE)
  res <- fn(mapping, test_pkg_data = NULL)
  
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 0L)
})

test_that("basename is consulted only when bioc_run_ut_path is usable", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(test_file = c("run_unitTests.R", "test-a.R"), stringsAsFactors = FALSE)
  
  # Case: usable path -> basename called
  mock_basename <- mockery::mock("run_unitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  res1 <- fn(mapping, list(bioc_run_ut_path = "/pkg/tests/run_unitTests.R"))
  mockery::expect_called(mock_basename, 1)
  expect_setequal(res1$test_file, "test-a.R")
  
  # Case: NA path -> basename should not be called
  mock_basename2 <- mockery::mock("should-not-be-called")
  mockery::stub(fn, "basename", mock_basename2)
  res2 <- fn(mapping, list(bioc_run_ut_path = NA_character_))
  mockery::expect_called(mock_basename2, 0)
  expect_setequal(res2$test_file, c("test-a.R"))
  
  # Case: NULL path -> basename should not be called
  mock_basename3 <- mockery::mock("should-not-be-called")
  mockery::stub(fn, "basename", mock_basename3)
  res3 <- fn(mapping, list(bioc_run_ut_path = NULL))
  mockery::expect_called(mock_basename3, 0)
  expect_setequal(res3$test_file, c("test-a.R"))
})  

test_that("classifies missing, error, failure, skipped, and passing files correctly", {
  fn <- check_covr_skip_bioc_nstf
  
  mapping <- data.frame(
    test_file = c(
      "test-missing.R",
      "test-error-throw.R",
      "test-error-count.R",
      "test-failure.R",
      "test-skipped.R",
      "test-pass.R",
      NA_character_
    ),
    stringsAsFactors = FALSE
  )
  
  mock_req_ns <- mockery::mock(TRUE)
  
  # drop_na should be called and remove NA
  mock_drop_na <- function(df, test_file) df[!is.na(df$test_file), , drop = FALSE]
  
  # file.exists -> FALSE only for test-missing.R
  mock_file_exists <- function(path) {
    if (grepl("test-missing\\.R$", path)) return(FALSE)
    TRUE
  }
  
  # Pretend RUnit is already on the search path so we don't re-attach each time
  mock_search <- mockery::mock("package:RUnit", "package:RUnit", "package:RUnit",
                               "package:RUnit", "package:RUnit", "package:RUnit")
  
  # runTestFile: throws for test-error-throw, otherwise returns a testdata object
  mock_run <- function(absFileName, useOwnErrorHandler, testFuncRegexp) {
    bn <- basename(absFileName)
    if (bn == "test-error-throw.R") stop("parse/runtime error")
    structure(list(file = bn), class = "RUnitTestData")
  }
  
  # getErrors: return different counts per file name
  mock_get_err <- function(td) {
    bn <- td$file
    switch(
      bn,
      "test-error-count.R" = list(nErr = 1, nFail = 0, nDeactivated = 0),
      "test-failure.R"     = list(nErr = 0, nFail = 2, nDeactivated = 0),
      "test-skipped.R"     = list(nErr = 0, nFail = 0, nDeactivated = 3),
      # default: passing
      list(nErr = 0, nFail = 0, nDeactivated = 0)
    )
  }
  
  mock_attach <- mockery::mock(NULL) # not used because search() includes RUnit
  mock_detach <- mockery::mock(NULL)
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", mock_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  # Expected issues (order follows mapping after drop_na)
  expected <- data.frame(
    test_file  = c("test-missing.R", "test-error-throw.R", "test-error-count.R", "test-failure.R", "test-skipped.R"),
    issue_type = c("missing",        "error",              "error",               "failure",         "skipped"),
    stringsAsFactors = FALSE
  )
  
  expect_s3_class(res, "data.frame")
  expect_identical(res[, c("test_file", "issue_type")], expected)
})

test_that("when all files are passing, returns 'No tests skipped' row", {
  fn <- check_covr_skip_bioc_nstf
  
  mapping <- data.frame(test_file = c("t1.R", "t2.R"), stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  # anyNA(mapping$test_file) is FALSE -> ensure we don't call tidyr::drop_na (we make it throw if called)
  throwing_drop_na <- function(...) stop("drop_na should not be called when no NA present")
  
  mock_file_exists <- mockery::mock(TRUE, TRUE)
  mock_search <- mockery::mock("package:RUnit", "package:RUnit")
  mock_run <- mockery::mock(structure(list(file = "t1.R"), class = "RUnitTestData"),
                            structure(list(file = "t2.R"), class = "RUnitTestData"))
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0),
                                list(nErr = 0, nFail = 0, nDeactivated = 0))
  mock_attach <- mockery::mock(NULL)
  mock_detach <- mockery::mock(NULL)
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", throwing_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 1L)
  expect_true(is.na(res$test_file[1]))
  expect_identical(res$issue_type[1], "No tests skipped")
})



test_that("attaches RUnit namespace only when not already on search path", {
  fn <- check_covr_skip_bioc_nstf
  
  mapping <- data.frame(test_file = "pass.R", stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  # First call to search() -> no RUnit; inside on.exit check may call again, so provide two values
  mock_search <- mockery::mock("package:stats", "package:RUnit")
  mock_attach <- mockery::mock(NULL)  # should be called once
  mock_detach <- mockery::mock(NULL)
  
  mock_file_exists <- mockery::mock(TRUE)
  mock_run <- mockery::mock(structure(list(file = "pass.R"), class = "RUnitTestData"))
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0))
  id_drop_na <- function(df, test_file) df
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", id_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  expect_s3_class(res, "data.frame")
  
  mockery::expect_called(mock_attach, 1)
})

test_that("runTestFile error path is captured as 'error' without crashing", {
  fn <- check_covr_skip_bioc_nstf
  
  mapping <- data.frame(test_file = "test-error-throw.R", stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  id_drop_na <- function(df, test_file) df
  mock_file_exists <- mockery::mock(TRUE)
  mock_search <- mockery::mock("package:RUnit")
  mock_attach <- mockery::mock(NULL)
  mock_detach <- mockery::mock(NULL)
  
  # Force runTestFile to throw -> tryCatch will return class "RUnitTestData-ERROR"
  mock_run <- function(...) stop("boom")
  # getErrors should not be called in this path
  mock_get_err <- mockery::mock()
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", id_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  expect_identical(nrow(res), 1L)
  expect_identical(res$test_file[1], "test-error-throw.R")
  expect_identical(res$issue_type[1], "error")
  mockery::expect_called(mock_get_err, 0)
})


test_that("returns bioc_unitTests when testdir already exists", {
  fn <- get_biocgenerics_test_paths
  
  # Normalized testdir value
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) {
    # Return a consistent normalized path regardless of input
    if (grepl("^/work/pkg/tests$", path)) return("/work/pkg/tests")
    path
  }
  mock_dir_exists <- function(path) path == "/work/pkg/tests"
  
  # Stubs
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  
  res <- fn(test_pkg_data = list(), testdir = "/work/pkg/tests")
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, "/work/pkg/tests")
  expect_false("fallback" %in% names(res))
})

test_that("finds package root at testdir and returns inst/unitTests if it exists", {
  fn <- get_biocgenerics_test_paths
  
  # Paths
  raw_testdir <- "/work/pkg/tests"
  normalized_testdir <- "/work/pkg/tests"
  pkg_root <- "/work/pkg"
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) {
    # Normalize testdir and pkg_root and bioc_unit deterministically
    if (path == raw_testdir) return(normalized_testdir)
    if (path == pkg_root) return(pkg_root)
    if (path == bioc_unit) return(bioc_unit)
    path
  }
  
  # testdir does NOT exist; bioc unit dir DOES
  mock_dir_exists <- function(path) {
    path == bioc_unit
  }
  
  # file.exists(DESCRIPTION) returns TRUE only at testdir's root (/work/pkg)
  mock_file_exists <- function(path) {
    path == file.path(pkg_root, "DESCRIPTION")
  }
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, bioc_unit)
  expect_false("fallback" %in% names(res))
})

test_that("finds package root at dirname(testdir) and returns inst/unitTests", {
  fn <- get_biocgenerics_test_paths
  
  # Layout:
  # testdir: /work/pkg/tests/unit
  # dirname(testdir): /work/pkg/tests
  # dirname(dirname(testdir)): /work/pkg
  raw_testdir <- "/work/pkg/tests/unit"
  normalized_testdir <- raw_testdir
  pkg_root <- "/work/pkg"
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) {
    if (path %in% c(raw_testdir, pkg_root, bioc_unit)) return(path)
    path
  }
  
  mock_dir_exists <- function(path) {
    # testdir does not exist; bioc_unit exists
    path == bioc_unit
  }
  
  mock_file_exists <- function(path) {
    # Mark only /work/pkg as a package root
    path == file.path(pkg_root, "DESCRIPTION")
  }
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, bioc_unit)
})

test_that("finds package root at dirname(dirname(testdir)) and returns inst/unitTests", {
  fn <- get_biocgenerics_test_paths
  
  raw_testdir <- "/work/pkg/a/b"
  pkg_root <- "/work/pkg"
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) path
  mock_dir_exists <- function(path) path == bioc_unit
  mock_file_exists <- function(path) path == file.path(pkg_root, "DESCRIPTION")
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, bioc_unit)
})

test_that("returns fallback when testdir doesn't exist and no package root found", {
  fn <- get_biocgenerics_test_paths
  
  raw_testdir <- "/nowhere/tests"
  normalized_testdir <- "/nowhere/tests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) normalized_testdir
  mock_dir_exists <- function(path) FALSE
  mock_file_exists <- function(path) FALSE
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("fallback" %in% names(res))
  expect_identical(res$fallback, normalized_testdir)
  expect_false("bioc_unitTests" %in% names(res))
})

test_that("returns fallback if package root found but inst/unitTests does not exist", {
  fn <- get_biocgenerics_test_paths
  
  raw_testdir <- "/work/pkg/tests"
  pkg_root <- "/work/pkg"
  normalized_testdir <- raw_testdir
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) path
  mock_dir_exists <- function(path) FALSE  # inst/unitTests not present
  mock_file_exists <- function(path) path == file.path(pkg_root, "DESCRIPTION")
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("fallback" %in% names(res))
  expect_identical(res$fallback, normalized_testdir)
})

test_that("handles duplicate candidate paths via unique()", {
  fn <- get_biocgenerics_test_paths
  
  # If testdir == dirname(testdir), unique() collapses duplicates.
  raw_testdir <- "/flat"
  # Simulate a flat structure where dirname("/flat") == "/"
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) path
  mock_dir_exists <- function(path) FALSE
  mock_file_exists <- function(path) FALSE
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  
  expect_true("fallback" %in% names(res))
  expect_identical(res$fallback, raw_testdir)
})



test_that("errors when no R source files are found", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/proj/pkg"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/proj/pkg/tests"
  
  # Stub list.files so there are NO source files
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(character())  # no .R under R/
    if (path == test_dir)   return(character(0)) # doesn't matter
    character(0)
  }
  mockery::stub(fn, "list.files", mock_list_files)
  
  # Use a simple pattern (or regexp = NA) to avoid brittle matching
  expect_error(
    fn(pkg_source_path = pkg_source_path, test_dir = test_dir),
    regexp = "No R files found under:"
  )
})

test_that("warns when no .R test files are found and maps NA tests", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/proj/pkg"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/proj/pkg/tests"
  
  src_files <- file.path(source_dir, c("alpha.R", "beta.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)    # two sources
    if (path == test_dir)   return(character())  # no tests -> warning
    character()
  }
  # No test content will be read because length(test_files) == 0
  mock_readLines <- function(con, warn = FALSE) stop("readLines should not be called")
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines", mock_readLines)
  
  expect_warning(
    res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir),
    regexp = "No R test files found under:"
  )
  
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 2L)
  expect_identical(res$source_file, c("alpha.R", "beta.R"))
  expect_true(all(is.na(res$test_file)))
  expect_true(all(is.na(res$evidence)))
  expect_identical(res$has_tests, c(FALSE, FALSE))
})

test_that("filename-based mapping: tokenized, case-insensitive, picks first match", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/pkg"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/pkg/tests"
  
  src_files  <- file.path(source_dir, c("foo.R", "bar.R"))
  test_files <- file.path(test_dir, c("test-foo.R", "ZZZ_foo_extra.R", "TEST-bar.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  # No content-based checks needed
  mock_readLines <- function(con, warn = FALSE) character()
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines", mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_s3_class(res, "data.frame")
  expect_identical(res$source_file, c("foo.R", "bar.R"))
  expect_identical(res$test_file,   c("test-foo.R", "TEST-bar.R")) # first match wins
  expect_identical(res$evidence,    c("filename_match", "filename_match"))
  expect_identical(res$has_tests,   c(TRUE, TRUE))
})

test_that("content-based mapping via RUnit test function names when filename matching fails", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/p"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/p/tests"
  
  # One source 'widget.R'; one test 'test-something.R'
  src_files  <- file.path(source_dir, "widget.R")
  test_files <- file.path(test_dir,   "test-something.R")
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  
  # Provide a FUNCTION DEF in source (ensures symbol extraction works)
  # Provide an RUnit test function line that contains the symbol
  mock_readLines <- function(con, warn = FALSE) {
    f <- as.character(con)
    if (identical(basename(f), "widget.R")) {
      return(c("widget <- function(x) { x }"))
    }
    if (identical(basename(f), "test-something.R")) {
      return(c("test_widget <- function() {",
               "  # assertions here",
               "}"))
    }
    character()
  }
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, "widget.R")
  expect_identical(res$test_file,   "test-something.R")
  expect_identical(res$evidence,    "content_match")
  expect_true(res$has_tests)
})

test_that("content-based mapping uses symbols from function defs (avoid setGeneric parsing)", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/root"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/root/tests"
  
  src_files  <- file.path(source_dir, c("generics.R", "funcs.R"))
  test_files <- file.path(test_dir,   c("t-generic.R", "t-func.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  
  # IMPORTANT: Because the function's setGeneric grep uses fixed=TRUE with an escaped quote,
  # it's safer to provide FUNCTION DEFS in sources so symbol extraction surely works.
  mock_readLines <- function(con, warn = FALSE) {
    f <- as.character(con)
    switch(
      basename(f),
      "generics.R" = c(
        "doStuff <- function(...) TRUE"   # symbol: doStuff
      ),
      "funcs.R"    = c(
        "my_func <- function(a, b) { a + b }",
        "other <- function() 1"
      ),
      "t-generic.R" = c(
        "test_doStuff <- function() { TRUE }"  # detects 'doStuff' via RUnit name
      ),
      "t-func.R" = c(
        "test_my_func <- function() { TRUE }"
      ),
      character()
    )
  }
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, c("generics.R", "funcs.R"))
  expect_identical(res$test_file,   c("t-generic.R", NA))
  expect_identical(res$evidence,    c("content_match", NA))
  expect_identical(res$has_tests,   c(TRUE, FALSE))
})

test_that("short symbol (<= 2 chars) matches via filename tokens", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/a"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/a/tests"
  
  src_files  <- file.path(source_dir, c("t.R", "x1.R"))
  test_files <- file.path(test_dir,   c("test-t.R", "we-test-x1.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  mock_readLines <- function(con, warn = FALSE) character()
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, c("t.R", "x1.R"))
  expect_identical(res$test_file,   c("test-t.R", "we-test-x1.R"))
  expect_identical(res$evidence,    c("filename_match", "filename_match"))
  expect_identical(res$has_tests,   c(TRUE, TRUE))
})

test_that("returns NA mapping for source with no filename or content evidence", {
  fn <- get_source_test_mapping_bioc_nstf
  
  pkg_source_path <- "/no"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/no/tests"
  
  src_files  <- file.path(source_dir, "lonely.R")
  test_files <- file.path(test_dir,   c("test-a.R", "test-b.R"))  # no relevant content
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  
  mock_readLines <- function(con, warn = FALSE) {
    f <- as.character(con)
    if (identical(basename(f), "lonely.R")) return(c("lonely <- function() 1"))
    if (basename(f) %in% c("test-a.R", "test-b.R")) return(c("test_other <- function() 1"))
    character()
  }
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, "lonely.R")
  expect_true(is.na(res$test_file))
  expect_true(is.na(res$evidence))
  expect_false(res$has_tests)
})

# Helper: minimal mapping DF used by downstream steps
make_mapping <- function() {
  data.frame(
    source_file = c("a.R", "b.R"),
    test_file   = c("test-a.R", NA_character_),  # one with a test, one without
    evidence    = c("filename_match", NA_character_),
    has_tests   = c(TRUE, FALSE),
    stringsAsFactors = FALSE
  )
}

test_that("diagnostic: coverage block functions are invoked (covr + reducer)", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  # Same setup as above (mapping, normalize, prepare, etc.)
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE) x)
  
  mapping <- data.frame(
    source_file = "a.R", test_file = "test-a.R", evidence = "filename_match",
    has_tests = TRUE, stringsAsFactors = FALSE
  )
  mockery::stub(fn, "get_source_test_mapping_runit_nstf", mockery::mock(mapping))
  mockery::stub(fn, "prepare_for_bioc_runit",     mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf",  mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests",      mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na",             function(df) df)
  
  mockery::stub(fn, "ls",               mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace",      mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign",           mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace",    mockery::mock(NULL))
  # Must stub library() or you’ll abort here!
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "pkgload::load_all", mockery::mock(invisible(NULL)))
  mockery::stub(fn, "library",          mockery::mock(NULL))
  mockery::stub(fn, "tempfile",         mockery::mock("/tmp/runner.R", cycle = TRUE))
  mockery::stub(fn, "writeLines",       mockery::mock(NULL, cycle = TRUE))
  
  
  # avoid error from as.environment("package:mypkg")
  mockery::stub(fn, "as.environment", function(name) {
    # when calling as.environment("package:mypkg"), return a fresh env
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    # delegate other calls to base (rare here)
    base::as.environment(name)
  })
  
  
  # Spy/record calls to covr & reducer
  spy_env_cov <- mockery::mock(structure(list(), class = "covr_coverage"))
  spy_cov_to_list <- mockery::mock(list(totalcoverage = 75.0))
  spy_compute <- mockery::mock(list(
    total_cov = 0.50, percent_cov = 75.0,
    n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L
  ))
  # Return > 0 so we skip package_coverage fallback (which would need readLines/real path)
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(75.0))
  
  # Bypass build_runit_custom_runner to avoid normalizePath(mustWork=TRUE) on fake path
  mock_runner_path <- tempfile(fileext = ".R")
  writeLines(c("library(mypkg)", "1 + 1"), mock_runner_path)
  mockery::stub(fn, "build_runit_custom_runner", mockery::mock(mock_runner_path))
  
  mockery::stub(fn, "covr::environment_coverage", spy_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",     spy_cov_to_list)
  mockery::stub(fn, "compute_total_coverage",     spy_compute)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  expect_identical(res$total_cov, 0.50)
  expect_identical(res$res_cov$coverage$totalcoverage, 75.0)
})


test_that("messages when package is not using RUnit layout and returns default coverage when covr fails", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # --- Stubs for early steps ---
  mock_get_pkg_name <- mockery::mock(pkg_name)
  mock_is_layout    <- mockery::mock(FALSE)  # trigger layout message
  
  mock_get_paths <- mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests")))
  mock_normalize <- mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE)
  
  # --- Mapping: succeed to progress into coverage path ---
  mock_get_mapping <- mockery::mock(make_mapping())
  
  # --- Cleaning branch: only run when bioc_run_ut_path is not NA ---
  # In this test, set bioc_run_ut_path = NA so the cleaning branch is skipped
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  # --- prepare_for_bioc_runit: succeed ---
  mock_prepare <- mockery::mock(TRUE)
  
  # --- Problems: return "No tests skipped" (default/no skip) ---
  mock_problems <- mockery::mock(data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  ))
  
  # --- Post-load mapping: functions with no tests ---
  mock_no_tests <- mockery::mock(data.frame(
    `function` = "b",
    stringsAsFactors = FALSE
  ))
  
  # --- tidyr::drop_na: identity (no change) ---
  mock_drop_na <- function(df) df
  
  # --- Env preparation (assign exported/internal objects + datasets) ---
  mock_ls <- mockery::mock(c("obj1", "obj2"))
  mock_asNamespace <- mockery::mock(new.env())
  mock_assign <- mockery::mock(NULL)
  mock_load_datasets <- mockery::mock(c("ds1", "ds2"))
  mock_message <- mockery::mock(NULL, cycle = TRUE)
  
  # --- Coverage block stubs: force an error in covr to take default path ---
  mock_loadedNamespaces <- mockery::mock(c("base", "stats", "utils", "methods"))  # so loadNamespace("methods") skipped
  mock_loadNamespace <- mockery::mock(NULL)
  mock_detach <- mockery::mock(TRUE)
  mock_library <- mockery::mock(NULL)
  mock_tempfile <- mockery::mock("/tmp/runner.R")
  mock_writeLines <- mockery::mock(NULL)
  
  # This triggers the error branch -> default coverage result
  mock_env_cov <- function(env, test_files) stop("instrumentation failed")
  # coverage_to_list is never reached when env_cov errors, but stub for safety
  mock_cov_to_list <- mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0))
  
  # --- cleanup_and_return_null used by error branches ---
  mock_cleanup <- mockery::mock(NULL)
  
  # Wire stubs
  mockery::stub(fn, "get_pkg_name",               mock_get_pkg_name)
  mockery::stub(fn, "is_runit_layout",       mock_is_layout)
  mockery::stub(fn, "get_runit_test_paths",mock_get_paths)
  mockery::stub(fn, "normalizePath",              mock_normalize)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf", mock_get_mapping)
  mockery::stub(fn, "prepare_for_bioc_runit",     mock_prepare)
  mockery::stub(fn, "check_covr_skip_runit_nstf",  mock_problems)
  mockery::stub(fn, "get_function_no_tests",      mock_no_tests)
  mockery::stub(fn, "tidyr::drop_na",             mock_drop_na)
  mockery::stub(fn, "ls",                         mock_ls)
  mockery::stub(fn, "asNamespace",                mock_asNamespace)
  mockery::stub(fn, "assign",                     mock_assign)
  mockery::stub(fn, "load_package_datasets",      mock_load_datasets)
  mockery::stub(fn, "message",                    mock_message)
  mockery::stub(fn, "loadedNamespaces",           mock_loadedNamespaces)
  mockery::stub(fn, "loadNamespace",              mock_loadNamespace)
  mockery::stub(fn, "detach_pkg_if_attached",     mock_detach)
  mockery::stub(fn, "library",                    mock_library)
  mockery::stub(fn, "tempfile",                   mock_tempfile)
  mockery::stub(fn, "writeLines",                 mock_writeLines)
  mockery::stub(fn, "covr::environment_coverage", mock_env_cov)
  mockery::stub(fn, "covr::coverage_to_list",     mock_cov_to_list)
  mockery::stub(fn, "cleanup_and_return_null",    mock_cleanup)
  
  suppressWarnings(res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data))
  
  # Assertions: default coverage path
  expect_type(res, "list")
  expect_true("res_cov" %in% names(res))
  expect_true(is.list(res$res_cov))
  expect_identical(res$total_cov, 0)
  expect_identical(res$res_cov$name, pkg_name)
  expect_true(is.list(res$res_cov$coverage))
  expect_identical(res$res_cov$coverage$totalcoverage, 0)
  expect_identical(res$res_cov$errors, "No RUnit or BiocGenerics configuration")
  
  # The first message should hint about RUnit layout
  msgs <- mockery::mock_args(mock_message)
  # Flatten first invocation into a single string to check substring
  first_msg <- paste0(msgs[[1]], collapse = "")
  expect_match(first_msg, "RUnit layout")
})


test_that("returns NULL when source-test mapping errors", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths", mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath",        mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE))
  # Mapping: throw error so tryCatch invokes cleanup_and_return_null -> returns NULL -> early exit
  mockery::stub(fn, "get_source_test_mapping_runit_nstf", function(...) stop("boom-mapping"))
  mockery::stub(fn, "cleanup_and_return_null",           mockery::mock(NULL))
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  expect_null(res)
})


test_that("returns NULL when prepare_for_bioc_runit fails", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths", mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath",        mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE))
  mockery::stub(fn, "get_source_test_mapping_runit_nstf", mockery::mock(make_mapping()))
  # Force prepare to error; tryCatch should call cleanup_and_return_null and return NULL
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) stop("prep-error"))
  mockery::stub(fn, "cleanup_and_return_null", mockery::mock(NULL))
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  expect_null(res)
})


test_that("coverage error path yields a default covr_list even when mapping & problems succeed", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)  # skip cleaning branch
  
  # Layout ok
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  
  # Paths, normalization
  mockery::stub(fn, "get_runit_test_paths", mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath",        mockery::mock(function(x, winslash = "/", mustWork = FALSE) x, cycle = TRUE))
  
  # Mapping and later pipeline
  mockery::stub(fn, "get_source_test_mapping_runit_nstf", mockery::mock(make_mapping()))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf", mockery::mock(data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(`function` = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na",        function(df) df)
  
  # Env prep for coverage
  mockery::stub(fn, "ls",               mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace",      mockery::mock(new.env()))
  mockery::stub(fn, "assign",           mockery::mock(NULL))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "message",          mockery::mock(NULL, cycle = TRUE))
  
  # Coverage: trigger failure to default path
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace",    mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "library",          mockery::mock(NULL))
  mockery::stub(fn, "tempfile",         mockery::mock("/tmp/runner.R"))
  mockery::stub(fn, "writeLines",       mockery::mock(NULL))
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("covr-failed"))
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0)))
  mockery::stub(fn, "cleanup_and_return_null", mockery::mock(NULL))
  
  suppressWarnings(res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data))
  
  expect_type(res, "list")
  expect_identical(res$total_cov, 0)
  expect_identical(res$res_cov$name, pkg_name)
  expect_identical(res$res_cov$coverage$totalcoverage, 0)
  expect_identical(res$res_cov$errors, "No RUnit or BiocGenerics configuration")
})

# Minimal helper mapping (one test file to pass mustWork normalization)
make_mapping <- function() {
  data.frame(
    source_file = c("a.R", "b.R"),
    test_file   = c("test-a.R", NA_character_),
    evidence    = c("filename_match", NA_character_),
    has_tests   = c(TRUE, FALSE),
    stringsAsFactors = FALSE
  )
}

test_that("error in coverage_to_list hits error branch but lines still produce messages and covr_list", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  # Upstream stubs to reach the block
  mockery::stub(fn, "get_pkg_name",         mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE) x)
  
  # Mapping with one test to ensure test_files_clean is non-empty
  make_mapping <- function() {
    data.frame(
      source_file = c("a.R", "b.R"),
      test_file   = c("test-a.R", NA_character_),
      evidence    = c("filename_match", NA_character_),
      has_tests   = c(TRUE, FALSE),
      stringsAsFactors = FALSE
    )
  }
  mockery::stub(fn, "get_source_test_mapping_runit_nstf", mockery::mock(make_mapping()))
  
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf", mockery::mock(
    data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)
  ))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  
  # Env prep / coverage stubs
  mockery::stub(fn, "ls",               mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace",      mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign",           mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace",    mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "pkgload::load_all", function(...) invisible(NULL))
  mockery::stub(fn, "library",          mockery::mock(NULL))
  mockery::stub(fn, "tempfile",         mockery::mock("/tmp/runner.R", cycle = TRUE))
  mockery::stub(fn, "writeLines",       mockery::mock(NULL, cycle = TRUE))
  
  # keep coverage block alive by returning a benign env for "package:mypkg"
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) {
      return(new.env(parent = emptyenv()))
    }
    base::as.environment(name)
  })
  
  # Bypass build to avoid normalizePath(mustWork=TRUE) on fake path
  mock_runner_path <- tempfile(fileext = ".R")
  writeLines(c("library(mypkg)", "1 + 1"), mock_runner_path)
  mockery::stub(fn, "build_runit_custom_runner", mockery::mock(mock_runner_path))
  
  # Return > 0 to skip package_coverage fallback
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(75.0))
  
  # Force error in coverage_to_list -> error branch sets res_cov$errors and NA coverage fields
  mockery::stub(fn, "covr::environment_coverage", mockery::mock(structure(list(), class = "covr_coverage")))
  mockery::stub(fn, "covr::coverage_to_list", function(...) stop("explode"))
  
  # Stub compute_total_coverage to still return values (so the lines proceed)
  mockery::stub(fn, "compute_total_coverage", function(percent_cov, ...) {
    # percent_cov will be NA_real_ due to error branch; ignore and return defaults
    list(
      total_cov            = 0.00,
      percent_cov          = 0.00,
      n_testfiles          = 1L,
      n_no_function_tests  = 1L,
      n_skipped_files      = 0L
    )
  })
  
  # Capture messages & cleanup calls
  mock_msg   <- mockery::mock(NULL, cycle = TRUE)
  mock_clean <- mockery::mock(NULL)
  mockery::stub(fn, "message",                 mock_msg)
  mockery::stub(fn, "cleanup_and_return_null", mock_clean)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  # Messages exist (Total Coverage 0.00%, Approx total coverage 0.00%)
  msg_calls <- mockery::mock_args(mock_msg)
  all_msgs  <- vapply(msg_calls, function(x) paste0(x, collapse = ""), character(1))
  total_line  <- all_msgs[grepl("^Total Coverage for mypkg is:", all_msgs)]
  approx_line <- all_msgs[grepl("^Approx total coverage:",       all_msgs)]
  
  expect_true(length(total_line)  >= 1)
  expect_true(length(approx_line) >= 1)
  expect_match(total_line[1],  "0%")
  expect_match(approx_line[1], "\\(0\\.00%\\)")
  
  # cleanup called once
  mockery::expect_called(mock_clean, 1)
  
  # covr_list assembled
  expect_type(res, "list")
  expect_true("res_cov" %in% names(res))
  # In error branch, res_cov$coverage$'totalcoverage' == NA_integer_
  expect_true(is.na(res$res_cov$coverage$totalcoverage))
  expect_true(inherits(res$res_cov$errors, "error") || is.character(res$res_cov$errors))
})


test_that("returns TRUE when test_pkg_data$has_BioG_test is TRUE regardless of files", {
  fn <- is_bioc_runit_layout
  
  # Mock filesystem as FALSE to ensure the first clause is sufficient
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(pkg_source_path = "/fake/path", test_pkg_data = list(has_BioG_test = TRUE))
  expect_true(res)
  
  # Both stubs should still be called once due to evaluation sequence of ||
  # (In R, || is short-circuiting, but since the first is TRUE, the others may not be called.
  # To be robust across implementations, we don't assert call counts here.)
})

test_that("returns FALSE when no BioC-RUnit indicators are present", {
  fn <- is_runit_layout
  
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  # test_pkg_data NULL or missing field -> isTRUE(NULL) => FALSE
  res1 <- fn(pkg_source_path = "/fake/path", test_pkg_data = NULL)
  expect_false(res1)
  
})

test_that("returns TRUE when inst/unitTests directory exists", {
  fn <- is_runit_layout
  
  # dir.exists should return TRUE; file.exists shouldn't need to be consulted,
  # but we mock it to a known value anyway.
  mock_dir_exists  <- mockery::mock(TRUE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(pkg_source_path = "/fake/path", test_pkg_data = list(has_BioG_test = FALSE))
  expect_true(res)
  
  # Verify dir.exists was called once with constructed path
  mockery::expect_called(mock_dir_exists, 1)
  # We can also check the argument passed to dir.exists (optional)
  called_args <- mockery::mock_args(mock_dir_exists)
  expect_match(called_args[[1]][[1]], "/fake/path/inst/unitTests$")
})

test_that("returns TRUE when tests/run_unitTests.R file exists", {
  fn <- is_runit_layout
  
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(TRUE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(pkg_source_path = "/fake/path", test_pkg_data = list(has_BioG_test = FALSE))
  expect_true(res)
  
  # Verify file.exists was called once with constructed path
  mockery::expect_called(mock_file_exists, 1)
  called_args <- mockery::mock_args(mock_file_exists)
  expect_match(called_args[[1]][[1]], "/fake/path/tests/run_unitTests\\.R$")
})

test_that("short-circuit behavior: when has_BioG_test is TRUE, other checks need not be TRUE", {
  fn <- is_runit_layout
  
  # We set dir.exists and file.exists to throw if called; if the function
  # short-circuits correctly, these should not be invoked.
  throwing_dir_exists <- function(...) stop("dir.exists should not be called")
  throwing_file_exists <- function(...) stop("file.exists should not be called")
  
  mockery::stub(fn, "dir.exists", throwing_dir_exists)
  mockery::stub(fn, "file.exists", throwing_file_exists)
  
  expect_true(fn(pkg_source_path = "/fake/path", test_pkg_data = list(has_BioG_test = TRUE)))
})

test_that("handles has_BioG_test = NULL or missing field as FALSE", {
  fn <- is_runit_layout
  
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  # Explicit NULL value
  res_null <- fn(pkg_source_path = "/fake/path", test_pkg_data = list(has_BioG_test = NULL))
  expect_false(res_null)
  
})

test_that("supports atomic pkg_source_path; only path construction matters", {
  fn <- is_runit_layout
  
  mock_dir_exists  <- mockery::mock(FALSE)
  mock_file_exists <- mockery::mock(FALSE)
  
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  # Try with a different path format; the stubs ensure no actual filesystem usage
  res <- fn(pkg_source_path = "C:\\proj", test_pkg_data = NULL)
  expect_false(res)
})


test_that("does nothing when package not attached and namespace not loaded", {
  fn <- detach_pkg_if_attached
  
  # search() contains no package:pkg
  mock_search <- mockery::mock(c("package:stats", "package:utils"))
  mock_detach <- mockery::mock(NULL)             # should not be called
  mock_is_ns   <- mockery::mock(FALSE)
  mock_unload  <- mockery::mock(NULL)            # should not be called
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  # Result is invisible TRUE
  v <- withVisible(fn("mypkg", unload_namespace = TRUE, quiet = TRUE))
  expect_true(v$value)
  expect_false(v$visible)
  
  # detach and unload should not be called
  mockery::expect_called(mock_detach, 0)
  mockery::expect_called(mock_unload, 0)
  
  # no messages because quiet = TRUE
  mockery::expect_called(mock_message, 0)
})

test_that("detaches when package attached; quiet=TRUE suppresses message", {
  fn <- detach_pkg_if_attached
  
  mock_search <- mockery::mock(c("package:mypkg", "package:stats"))
  mock_detach <- mockery::mock(NULL)
  mock_is_ns  <- mockery::mock(FALSE)     # namespace not loaded
  mock_unload <- mockery::mock(NULL)      # not called
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  res <- fn("mypkg", unload_namespace = TRUE, quiet = TRUE)
  expect_true(res)
  
  # detach should be called once with constructed search name and flags
  mockery::expect_called(mock_detach, 1)
  args <- mockery::mock_args(mock_detach)
  expect_equal(args[[1]][[1]], "package:mypkg")
  expect_true(isTRUE(args[[1]][["unload"]]))
  expect_true(isTRUE(args[[1]][["character.only"]]))
  
  # namespace unload should not be called (isNamespaceLoaded == FALSE)
  mockery::expect_called(mock_unload, 0)
  
  # quiet=TRUE means no message
  mockery::expect_called(mock_message, 0)
})

test_that("unloads namespace when loaded; quiet=TRUE suppresses message", {
  fn <- detach_pkg_if_attached
  
  mock_search <- mockery::mock(c("package:stats"))  # not attached
  mock_detach <- mockery::mock(NULL)                # not called
  mock_is_ns  <- mockery::mock(TRUE)                # namespace loaded
  mock_unload <- mockery::mock(NULL)
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  res <- fn("mypkg", unload_namespace = TRUE, quiet = TRUE)
  expect_true(res)
  
  mockery::expect_called(mock_detach, 0)
  mockery::expect_called(mock_unload, 1)
  unload_args <- mockery::mock_args(mock_unload)
  expect_equal(unload_args[[1]][[1]], "mypkg")
  
  mockery::expect_called(mock_message, 0)
})

test_that("both detach and unload occur when attached and namespace loaded", {
  fn <- detach_pkg_if_attached
  
  mock_search <- mockery::mock(c("package:mypkg", "package:stats"))
  mock_detach <- mockery::mock(NULL)
  mock_is_ns  <- mockery::mock(TRUE)
  mock_unload <- mockery::mock(NULL)
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  res <- fn("mypkg", unload_namespace = TRUE, quiet = TRUE)
  expect_true(res)
  
  mockery::expect_called(mock_detach, 1)
  mockery::expect_called(mock_unload, 1)
  mockery::expect_called(mock_message, 0)
})

test_that("unload_namespace = FALSE prevents namespace unloading", {
  fn <- detach_pkg_if_attached
  
  mock_search <- mockery::mock(c("package:mypkg"))
  mock_detach <- mockery::mock(NULL)
  mock_is_ns  <- mockery::mock(TRUE)    # would be loaded, but flag stops unload
  mock_unload <- mockery::mock(NULL)    # should not be called
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  res <- fn("mypkg", unload_namespace = FALSE, quiet = TRUE)
  expect_true(res)
  
  mockery::expect_called(mock_detach, 1)
  mockery::expect_called(mock_unload, 0)
})

test_that("try() respects silent flag; errors in detach/unload are swallowed when quiet=TRUE", {
  fn <- detach_pkg_if_attached
  
  mock_search <- mockery::mock(c("package:mypkg"))
  mock_detach <- function(...) stop("detach failed")  # throws
  mock_is_ns  <- mockery::mock(TRUE)
  mock_unload <- function(...) stop("unload failed")  # throws
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  # quiet=TRUE => try(..., silent = TRUE): errors are swallowed, function still returns TRUE
  expect_true(fn("mypkg", unload_namespace = TRUE, quiet = TRUE))
})

test_that("returns invisible(TRUE) in all paths", {
  fn <- detach_pkg_if_attached
  
  mock_search <- mockery::mock(c("package:mypkg"))
  mock_detach <- mockery::mock(NULL)
  mock_is_ns  <- mockery::mock(TRUE)
  mock_unload <- mockery::mock(NULL)
  mock_message <- mockery::mock(NULL)
  
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "isNamespaceLoaded", mock_is_ns)
  mockery::stub(fn, "unloadNamespace", mock_unload)
  mockery::stub(fn, "message", mock_message)
  
  v <- withVisible(fn("mypkg", unload_namespace = TRUE, quiet = TRUE))
  expect_true(v$value)
  expect_false(v$visible)
})


test_that("detaches target package and BioC infra packages in order", {
  fn <- prepare_for_bioc_runit
  
  # Mock the detacher to record calls
  mock_detach <- mockery::mock(NULL, NULL, NULL, NULL)  # 4 calls total expected
  # Mock load_all to prevent real load and to inspect args
  mock_load_all <- mockery::mock(NULL)
  
  # Stub symbols referenced inside the function
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  # Use a fake path/pkg
  res <- fn(pkg_name = "mypkg", pkg_source_path = "/tmp/mypkg", detach_bioc_infra = TRUE)
  
  expect_true(res)
  
  # Expect 4 detach calls: mypkg, IRanges, S4Vectors, BiocGenerics
  mockery::expect_called(mock_detach, 4)
  calls <- mockery::mock_args(mock_detach)
  
  expect_identical(calls[[1]][[1]], "mypkg")
  expect_identical(calls[[2]][[1]], "IRanges")
  expect_identical(calls[[3]][[1]], "S4Vectors")
  expect_identical(calls[[4]][[1]], "BiocGenerics")
  
  # Verify load_all called once, with path and quiet=TRUE
  mockery::expect_called(mock_load_all, 1)
  la_args <- mockery::mock_args(mock_load_all)
  expect_identical(la_args[[1]][[1]], "/tmp/mypkg")
  expect_true(isTRUE(la_args[[1]][["quiet"]]))
})

test_that("options(warn) is temporarily set to 1 and restored after function exits", {
  fn <- prepare_for_bioc_runit
  
  # Mocks to avoid real side effects
  mock_detach <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  # Preserve and auto-restore original options/env using withr
  old_warn <- getOption("warn")
  withr::local_options(list(warn = old_warn))
  
  # Before call: whatever the test harness has
  expect_identical(getOption("warn"), old_warn)
  
  # Call the function (it will set warn=1 internally and restore on exit)
  res <- fn(pkg_name = "mypkg", pkg_source_path = "/tmp/mypkg", detach_bioc_infra = TRUE)
  expect_true(res)
  
  # After return, warn should be restored to old_warn
  expect_identical(getOption("warn"), old_warn)
})

test_that('R_TESTS is empty during test (local_envvar)', {
  is_windows <- identical(.Platform$OS.type, "windows")
  
  # Ensure clean slate first
  Sys.unsetenv("R_TESTS")
  
  # Temporarily set for the duration of the test; withr restores automatically
  withr::local_envvar(if (is_windows) list("R_TESTS" = "") else list(R_TESTS = ""))
  
  # ... run things that depend on R_TESTS="", or assertions ...
  expect_identical(Sys.getenv("R_TESTS"), "")
  
})



test_that("POSIX branch: sets R_TESTS='' and restores to UNSET when originally unset", {
  # Only run on Linux/macOS to exercise the POSIX branch
  
  # Skip on Windows with explanation
  skip_if(
    condition = .Platform$OS.type == "windows",
    message = paste(
      "Skipped on Windows: this test exercises the POSIX-only behaviour of",
      "prepare_for_bioc_runit where R_TESTS is set to '' and restored to UNSET.",
      "Windows does not reliably distinguish empty vs UNSET environment variables."
    )
  )
  
  
  fn <- prepare_for_bioc_runit
  
  # Mock non-essential collaborators
  mock_detach   <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  # Ensure R_TESTS is UNSET to begin with
  Sys.unsetenv("R_TESTS")
  
  # Belt & suspenders: ensure we leave R_TESTS unset after the test, even if assertions fail
  withr::defer({
    Sys.unsetenv("R_TESTS")
  })
  
  # Call the function under test (should set R_TESTS="" and register on.exit restoration)
  res <- fn(pkg_name = "mypkg", pkg_source_path = "/tmp/mypkg", detach_bioc_infra = TRUE)
  expect_true(res)
  
  # After the function returns, on.exit should have restored to UNSET
  # expect_true(is.na(Sys.getenv("R_TESTS", unset = NA_character_)))
})


test_that("POSIX branch: sets R_TESTS='' and restores to previous value when originally set (mocked)", {
  
  skip_if(
    condition = .Platform$OS.type == "windows",
    message = paste("Skipped on Windows: this test exercises POSIX-specific behaviour where R_TESTS must be ",
                    "set to '' then restored. Windows environment handling (cmd.exe/PowerShell) treats empty ",
                    "environment variables differently and cannot reliably distinguish '' from prior values, ",
                    "making the restoration semantics untestable.")
  )
  
  
  fn <- prepare_for_bioc_runit
  
  # Seed original value so old_R_TESTS inside fn is non-NA
  original <- "previous-value"
  Sys.setenv(R_TESTS = original)
  
  # --- Mocks ---
  mock_setenv   <- mockery::mock(NULL, NULL)   # first for "", then for restore to original
  mock_unsetenv <- mockery::mock(NULL)         # should NOT be called in this path
  
  mock_detach   <- mockery::mock(NULL, NULL, NULL, NULL)
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "Sys.setenv",   mock_setenv)
  mockery::stub(fn, "Sys.unsetenv", mock_unsetenv)
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all",      mock_load_all)
  
  # Execute
  res <- fn(pkg_name = "mypkg", pkg_source_path = "/tmp/mypkg", detach_bioc_infra = TRUE)
  expect_true(res)
  
  # Assertions:
  set_calls <- mockery::mock_args(mock_setenv)
  
  # 1) It sets R_TESTS to empty string initially
  expect_true(any(vapply(set_calls, function(x) identical(x, list(R_TESTS = "")), logical(1))),
              info = "Sys.setenv(R_TESTS = \"\") should be called in POSIX branch")
})  

test_that("run_covr_skip_runit_nstf: cleaning branch -> happy path updates mapping and proceeds", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # --- test_pkg_data with cleaning branch active ---
  test_pkg_data <- list(
    bioc_run_ut_path = file.path(pkg_source_path, "inst", "unitTests", "run_unitTests.R"),
    has_BioG_test    = TRUE
  )
  
  cov_env <- new.env(parent = baseenv())
  
  # --- Stubs for early steps -------------------------------------------------
  
  mockery::stub(fn, "get_pkg_name", function(...) pkg_name)
  mockery::stub(fn, "is_runit_layout", function(...) TRUE)
  
  mockery::stub(fn, "get_runit_test_paths",
                function(...) list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))
  )
  
  mockery::stub(fn, "normalizePath",
                function(path, winslash = "/", mustWork = FALSE, ...) path
  )
  
  # mapping BEFORE cleaning (contains run_unitTests.R)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                function(...) data.frame(
                  source_file = c("a.R", "b.R"),
                  test_file   = c("run_unitTests.R", "test-a.R"),
                  evidence    = c("filename_match", "filename_match"),
                  has_tests   = TRUE,
                  stringsAsFactors = FALSE
                )
  )
  
  # mapping AFTER cleaning
  mockery::stub(fn, "remove_run_unitTests_mapping",
                function(mapping, ...) {
                  mapping[mapping$test_file != "run_unitTests.R", , drop = FALSE]
                }
  )
  
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) TRUE)
  
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                function(...) data.frame(
                  test_file  = NA_character_,
                  issue_type = "No tests skipped",
                  stringsAsFactors = FALSE
                )
  )
  
  mockery::stub(fn, "get_function_no_tests",
                function(...) data.frame(`function` = "b", stringsAsFactors = FALSE)
  )
  
  # drop_na identity
  mockery::stub(fn, "tidyr::drop_na", function(df, ...) df)
  
  # Coverage environment mocks
  mockery::stub(fn, "ls", function(...) c("obj1", "obj2"))
  mockery::stub(fn, "asNamespace", function(...) new.env())
  mockery::stub(fn, "assign", function(...) NULL)
  mockery::stub(fn, "load_package_datasets", function(...) c("ds1", "ds2"))
  
  # coverage block -> fail -> fallback return
  mockery::stub(fn, "loadedNamespaces", function(...) c("base"))
  mockery::stub(fn, "loadNamespace", function(...) NULL)
  mockery::stub(fn, "detach_pkg_if_attached", function(...) TRUE)
  mockery::stub(fn, "library", function(...) NULL)
  mockery::stub(fn, "tempfile", function(...) "/tmp/runner.R")
  mockery::stub(fn, "writeLines", function(...) NULL)
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("instrumentation failed"))
  mockery::stub(fn, "covr::coverage_to_list",
                function(...) list(filecoverage = matrix(0,1), totalcoverage = 0)
  )
  
  # cleanup (should be called on coverage error)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  # --- Execute, capturing messages because message() was stubbed away earlier ---
  output <- testthat::capture_messages(
    suppressWarnings(res <- fn(pkg_source_path = pkg_source_path,
              test_pkg_data   = test_pkg_data,
              cov_env         = cov_env))
  )
  
  # --- Assertions -----------------------------------------------------------
  
  # Ensure cleaning message was emitted
  expect_true(any(grepl("performing source test map cleaning for mypkg", output)))
  
  # Should not be NULL (because fn returns fallback structure)
  expect_false(is.null(res))
})


test_that("run_covr_skip_runit_nstf builds correct run_test_line for prefix = 'runit'", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_pkg_data <- list(has_BioG_test = FALSE, bioc_run_ut_path = NA)
  cov_env <- new.env()
  
  # ---- ESSENTIAL STUBS ----
  mockery::stub(fn, "get_pkg_name", function(...) "mypkg")
  mockery::stub(fn, "is_runit_layout", function(...) TRUE)
  mockery::stub(fn, "get_runit_test_paths",
                function(...) list(bioc_unitTests = "/fake/pkg/inst/unitTests")
  )
  mockery::stub(fn, "normalizePath", function(path, ...) path)
  
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                function(...) data.frame(
                  source_file="a.R",
                  test_file="test-a.R",
                  evidence="filename_match",
                  has_tests=TRUE
                )
  )
  mockery::stub(fn, "remove_run_unitTests_mapping", identity)
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) TRUE)
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                function(...) data.frame(test_file=NA_character_, issue_type="No tests skipped")
  )
  mockery::stub(fn, "get_function_no_tests", function(...) data.frame(fun="a"))
  mockery::stub(fn, "tidyr::drop_na", function(df, ...) df)
  
  mockery::stub(fn, "ls", function(...) character(0))
  mockery::stub(fn, "asNamespace", function(...) new.env())
  mockery::stub(fn, "assign", function(...) NULL)
  mockery::stub(fn, "load_package_datasets", function(...) character(0))
  mockery::stub(fn, "file.exists", function(...) TRUE)
  mockery::stub(fn, "dir.exists", function(...) TRUE)
  
  # force prefix = runit
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  
  # ---- Capture runner file created inside writeLines() ----
  runner_path <- NULL
  
  mockery::stub(fn, "writeLines", function(lines, path, ...) {
    runner_path <<- path
    base::writeLines(lines, path)
  })
  
  # fail coverage
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("fail"))
  mockery::stub(fn, "covr::coverage_to_list",
                function(...) list(filecoverage=matrix(0,1), totalcoverage=0)
  )
  mockery::stub(fn, "loadedNamespaces", function(...) "base")
  mockery::stub(fn, "detach_pkg_if_attached", function(...) TRUE)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  # ---- RUN ----
  out <- testthat::capture_messages(
    suppressWarnings(fn(pkg_source_path = pkg_source_path,
       test_pkg_data = test_pkg_data,
       cov_env = cov_env))
  )
  
  # ---- ASSERTIONS ----
  expect_true(is.character(out))
  expect_false(is.null(out))
})


test_that("run_covr_skip_runit_nstf builds correct run_test_line for prefix = 'Test'", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_pkg_data <- list(has_BioG_test = FALSE, bioc_run_ut_path = NA)
  cov_env <- new.env()
  
  # ---- ESSENTIAL STUBS (same pattern as runit test) ----
  mockery::stub(fn, "get_pkg_name", function(...) "mypkg")
  mockery::stub(fn, "is_runit_layout", function(...) TRUE)
  mockery::stub(fn, "get_runit_test_paths",
                function(...) list(bioc_unitTests = "/fake/pkg/inst/unitTests"))
  mockery::stub(fn, "normalizePath", function(path, ...) path)
  
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                function(...) data.frame(
                  source_file="a.R",
                  test_file="Test-a.R",
                  evidence="filename_match",
                  has_tests=TRUE,
                  stringsAsFactors = FALSE
                ))
  
  mockery::stub(fn, "remove_run_unitTests_mapping", identity)
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) TRUE)
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                function(...) data.frame(test_file=NA_character_, issue_type="No tests skipped"))
  mockery::stub(fn, "get_function_no_tests", function(...) data.frame(fun="a"))
  mockery::stub(fn, "tidyr::drop_na", function(df, ...) df)
  mockery::stub(fn, "ls", function(...) character(0))
  mockery::stub(fn, "asNamespace", function(...) new.env())
  mockery::stub(fn, "assign", function(...) NULL)
  mockery::stub(fn, "load_package_datasets", function(...) character(0))
  mockery::stub(fn, "file.exists", function(...) TRUE)
  mockery::stub(fn, "dir.exists", function(...) TRUE)
  
  # FORCE "Test" prefix
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "Test")
  
  # Capture runner file
  runner_path <- NULL
  mockery::stub(fn, "writeLines", function(lines, path, ...) {
    runner_path <<- path
    base::writeLines(lines, path)
  })
  
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("fail"))
  mockery::stub(fn, "covr::coverage_to_list",
                function(...) list(filecoverage=matrix(0,1), totalcoverage=0))
  mockery::stub(fn, "loadedNamespaces", function(...) "base")
  mockery::stub(fn, "detach_pkg_if_attached", function(...) TRUE)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  # ---- RUN ----
  out <- testthat::capture_messages(
    suppressWarnings(fn(pkg_source_path = pkg_source_path,
       test_pkg_data   = test_pkg_data,
       cov_env         = cov_env))
  )
  
  # ---- ASSERTIONS ----
  expect_true(is.character(out))
  expect_false(is.null(out))
})


test_that("run_covr_skip_runit_nstf builds correct run_test_line for default prefix 'test'", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_pkg_data <- list(has_BioG_test = FALSE, bioc_run_ut_path = NA)
  cov_env <- new.env()
  
  # ---- ESSENTIAL STUBS ----
  mockery::stub(fn, "get_pkg_name", function(...) "mypkg")
  mockery::stub(fn, "is_runit_layout", function(...) TRUE)
  mockery::stub(fn, "get_runit_test_paths",
                function(...) list(bioc_unitTests = "/fake/pkg/inst/unitTests"))
  mockery::stub(fn, "normalizePath", function(path, ...) path)
  
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                function(...) data.frame(
                  source_file="a.R",
                  test_file="test-a.R",
                  evidence="filename_match",
                  has_tests=TRUE,
                  stringsAsFactors = FALSE
                ))
  
  mockery::stub(fn, "remove_run_unitTests_mapping", identity)
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) TRUE)
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                function(...) data.frame(test_file=NA_character_, issue_type="No tests skipped"))
  mockery::stub(fn, "get_function_no_tests", function(...) data.frame(fun="a"))
  mockery::stub(fn, "tidyr::drop_na", function(df, ...) df)
  mockery::stub(fn, "ls", function(...) character(0))
  mockery::stub(fn, "asNamespace", function(...) new.env())
  mockery::stub(fn, "assign", function(...) NULL)
  mockery::stub(fn, "load_package_datasets", function(...) character(0))
  mockery::stub(fn, "file.exists", function(...) TRUE)
  mockery::stub(fn, "dir.exists", function(...) TRUE)
  
  # FORCE any non-runit, non-Test → fall into default branch
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "xyz")
  
  # Capture runner file
  runner_path <- NULL
  mockery::stub(fn, "writeLines", function(lines, path, ...) {
    runner_path <<- path
    base::writeLines(lines, path)
  })
  
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("fail"))
  mockery::stub(fn, "covr::coverage_to_list",
                function(...) list(filecoverage=matrix(0,1), totalcoverage=0))
  mockery::stub(fn, "loadedNamespaces", function(...) "base")
  mockery::stub(fn, "detach_pkg_if_attached", function(...) TRUE)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  # ---- RUN ----
  out <- testthat::capture_messages(
    suppressWarnings(fn(pkg_source_path = pkg_source_path,
       test_pkg_data   = test_pkg_data,
       cov_env         = cov_env))
  )
  
  # ---- ASSERTIONS ----
  expect_true(is.character(out))
  expect_false(is.null(out))
})
  

test_that("run_covr_skip_runit_nstf: cleaning branch -> happy path updates mapping and proceeds", {
  fn <- run_covr_skip_runit_nstf

  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"

  # Cleaning branch must activate (non-NA bioc_run_ut_path)
  test_pkg_data <- list(
    bioc_run_ut_path = "/fake/pkg/inst/unitTests/run_unitTests.R",
    has_BioG_test = TRUE
  )

  cov_env <- new.env(parent = baseenv())

  # ---------------------------------------------------------------------------
  # ESSENTIAL STUBS ONLY
  # ---------------------------------------------------------------------------

  mockery::stub(fn, "get_pkg_name", function(...) pkg_name)
  mockery::stub(fn, "is_runit_layout", function(...) TRUE)

  mockery::stub(fn, "get_runit_test_paths",
                function(...) list(bioc_unitTests = "/fake/pkg/inst/unitTests")
  )

  mockery::stub(fn, "normalizePath", function(path, ...) path)

  # Mapping BEFORE cleaning (contains run_unitTests.R)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                function(...) data.frame(
                  source_file = c("a.R", "b.R"),
                  test_file   = c("run_unitTests.R", "test-b.R"),
                  evidence    = c("filename_match", "filename_match"),
                  has_tests   = TRUE,
                  stringsAsFactors = FALSE
                )
  )

  # Cleaning removes run_unitTests.R
  mockery::stub(fn, "remove_run_unitTests_mapping",
                function(mapping, ...) {
                  mapping[mapping$test_file != "run_unitTests.R", , drop = FALSE]
                }
  )

  mockery::stub(fn, "prepare_for_bioc_runit", function(...) TRUE)

  # No problematic tests → no skipping
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                function(...) data.frame(
                  test_file  = NA_character_,
                  issue_type = "No tests skipped",
                  stringsAsFactors = FALSE
                )
  )

  mockery::stub(fn, "get_function_no_tests",
                function(...) data.frame(fun = "b", stringsAsFactors = FALSE)
  )

  mockery::stub(fn, "tidyr::drop_na", function(df, ...) df)

  # Minimal env population stubs
  mockery::stub(fn, "ls", function(...) character(0))
  mockery::stub(fn, "asNamespace", function(...) new.env())
  mockery::stub(fn, "assign", function(...) NULL)
  mockery::stub(fn, "load_package_datasets", function(...) character(0))

  # Force coverage instrumentation to fail → fallback return
  mockery::stub(fn, "loadedNamespaces", function(...) c("base"))
  mockery::stub(fn, "detach_pkg_if_attached", function(...) TRUE)
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("instrumentation failed"))
  mockery::stub(fn, "covr::coverage_to_list",
                function(...) list(filecoverage = matrix(0, 1), totalcoverage = 0)
  )

  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)

  # ---------------------------------------------------------------------------
  # Run the function and CAPTURE MESSAGE() output (CRITICAL!)
  # ---------------------------------------------------------------------------

  output <- testthat::capture_messages(
    res <- suppressWarnings(fn(pkg_source_path = pkg_source_path,
              test_pkg_data   = test_pkg_data,
              cov_env         = cov_env))
  )


  # ---------------------------------------------------------------------------
  # Assertions
  # ---------------------------------------------------------------------------

  # Cleaning message MUST appear
  expect_true(any(grepl(
     "performing source test map cleaning for mypkg",
     output
  )))
  expect_false(is.null(output))

  # Function returns a non-null result (successful continuation)
  expect_false(is.null(res))
})

test_that("run_covr_skip_runit_nstf: cleaning branch -> NULL mapping returns NULL early", {
  fn <- run_covr_skip_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  
  # Cleaning branch must trigger -> bioc_run_ut_path is non-NA
  test_pkg_data <- list(
    bioc_run_ut_path = "/fake/pkg/inst/unitTests/run_unitTests.R",
    has_BioG_test = TRUE
  )
  
  # --- ESSENTIAL STUBS ONLY --------------------------------------------------
  
  mockery::stub(fn, "get_pkg_name", function(...) pkg_name)
  mockery::stub(fn, "is_runit_layout", function(...) TRUE)
  
  # valid test path list
  mockery::stub(fn, "get_runit_test_paths",
                function(...) list(bioc_unitTests = "/fake/pkg/inst/unitTests")
  )
  
  mockery::stub(fn, "normalizePath", function(path, ...) path)
  
  # Mapping BEFORE cleaning (must be a data.frame)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                function(...) data.frame(
                  source_file = c("a.R","b.R"),
                  test_file   = c("run_unitTests.R","test-b.R"),
                  evidence    = "filename_match",
                  has_tests   = TRUE,
                  stringsAsFactors = FALSE
                )
  )
  
  # Cleaning returns NULL → EARLY RETURN expected
  mockery::stub(fn, "remove_run_unitTests_mapping", function(...) NULL)
  
  # These will not execute due to early return but must exist to avoid missing stubs
  mockery::stub(fn, "prepare_for_bioc_runit", function(...) TRUE)
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                function(...) data.frame(test_file=NA_character_, issue_type="No tests skipped"))
  mockery::stub(fn, "get_function_no_tests",
                function(...) data.frame(fun="b"))
  mockery::stub(fn, "tidyr::drop_na", function(df, ...) df)
  mockery::stub(fn, "ls", function(...) character(0))
  mockery::stub(fn, "asNamespace", function(...) new.env())
  mockery::stub(fn, "assign", function(...) NULL)
  mockery::stub(fn, "load_package_datasets", function(...) character(0))
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  # ---------------------------------------------------------------------------
  # Capture messages reliably
  # ---------------------------------------------------------------------------
  output <- testthat::capture_messages(
    res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  )
  
  # ---- ASSERTIONS -----------------------------------------------------------
  
  # Must exit early with NULL because cleaned mapping is NULL
  expect_null(res)
  
  # Cleaning message MUST appear
  expect_true(any(grepl(
    "performing source test map cleaning for mypkg",
    output
  )))
})

test_that("doRUnit branch: message 'Trying doRUnit.R' and build_doRUnit_wrapper called when doRUnit_path exists", {
  fn <- run_covr_skip_runit_nstf
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  doRUnit_path <- "/fake/pkg/tests/doRUnit.R"
  test_pkg_data <- list(
    bioc_run_ut_path = NA_character_,
    doRUnit_path = doRUnit_path
  )
  
  mockery::stub(fn, "get_pkg_name", mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE, ...) x)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                mockery::mock(data.frame(source_file = "a.R", test_file = "test-a.R", evidence = "match", has_tests = TRUE, stringsAsFactors = FALSE)))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  mockery::stub(fn, "ls", mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace", mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign", mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace", mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "pkgload::load_all", mockery::mock(invisible(NULL)))
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    base::as.environment(name)
  })
  
  mock_file_exists <- function(path) path == doRUnit_path
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  mock_runner <- tempfile(fileext = ".R")
  writeLines("1+1", mock_runner)
  mock_build_doRUnit <- mockery::mock(mock_runner)
  mockery::stub(fn, "build_doRUnit_wrapper", mock_build_doRUnit)
  
  mock_env_cov <- mockery::mock(structure(list(), class = "covr_coverage"))
  mockery::stub(fn, "covr::environment_coverage", mock_env_cov)
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(75.0))
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(totalcoverage = 75.0)))
  mockery::stub(fn, "compute_total_coverage", mockery::mock(list(total_cov = 0.75, percent_cov = 75.0, n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L)))
  mockery::stub(fn, "build_runit_runTests_runner", mockery::mock(mock_runner))
  mockery::stub(fn, "build_runit_custom_runner", mockery::mock(mock_runner))
  
  mock_msg <- mockery::mock(NULL, cycle = TRUE)
  mockery::stub(fn, "message", mock_msg)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  msg_calls <- mockery::mock_args(mock_msg)
  all_msgs <- vapply(msg_calls, function(x) paste0(x, collapse = ""), character(1))
  expect_true(any(grepl("Trying doRUnit.R for mypkg", all_msgs)))
  mockery::expect_called(mock_build_doRUnit, 1)
  args <- mockery::mock_args(mock_build_doRUnit)
  expect_identical(args[[1]][[1]], doRUnit_path)
})

test_that("doRUnit branch: when env_coverage throws, message 'doRUnit.R failed' and fallback to runTests-style runner", {
  fn <- run_covr_skip_runit_nstf
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  doRUnit_path <- "/fake/pkg/tests/doRUnit.R"
  test_pkg_data <- list(
    bioc_run_ut_path = NA_character_,
    doRUnit_path = doRUnit_path
  )
  
  mockery::stub(fn, "get_pkg_name", mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE, ...) x)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                mockery::mock(data.frame(source_file = "a.R", test_file = "test-a.R", evidence = "match", has_tests = TRUE, stringsAsFactors = FALSE)))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  mockery::stub(fn, "ls", mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace", mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign", mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace", mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "pkgload::load_all", mockery::mock(invisible(NULL)))
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    base::as.environment(name)
  })
  
  mock_file_exists <- function(path) path == doRUnit_path
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  mock_runner <- tempfile(fileext = ".R")
  writeLines("1+1", mock_runner)
  mockery::stub(fn, "build_doRUnit_wrapper", mockery::mock(mock_runner))
  
  mockery::stub(fn, "covr::environment_coverage", function(...) stop("doRUnit instrumentation failed"))
  mock_runTests_runner <- mockery::mock(mock_runner)
  mockery::stub(fn, "build_runit_runTests_runner", mock_runTests_runner)
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(75.0))
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(totalcoverage = 75.0)))
  mockery::stub(fn, "compute_total_coverage", mockery::mock(list(total_cov = 0.75, percent_cov = 75.0, n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L)))
  
  mock_msg <- mockery::mock(NULL, cycle = TRUE)
  mockery::stub(fn, "message", mock_msg)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  msg_calls <- mockery::mock_args(mock_msg)
  all_msgs <- vapply(msg_calls, function(x) paste0(x, collapse = ""), character(1))
  expect_true(any(grepl("doRUnit.R failed:.*falling back to custom runner", all_msgs)))
  mockery::expect_called(mock_runTests_runner, 1)
})

test_that("doRUnit branch: when percent_coverage returns 0, message and fallback to runTests-style runner", {
  fn <- run_covr_skip_runit_nstf
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  doRUnit_path <- "/fake/pkg/tests/doRUnit.R"
  test_pkg_data <- list(
    bioc_run_ut_path = NA_character_,
    doRUnit_path = doRUnit_path
  )
  
  mockery::stub(fn, "get_pkg_name", mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE, ...) x)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                mockery::mock(data.frame(source_file = "a.R", test_file = "test-a.R", evidence = "match", has_tests = TRUE, stringsAsFactors = FALSE)))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  mockery::stub(fn, "ls", mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace", mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign", mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace", mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "pkgload::load_all", mockery::mock(invisible(NULL)))
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    base::as.environment(name)
  })
  
  mock_file_exists <- function(path) path == doRUnit_path
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  mock_runner <- tempfile(fileext = ".R")
  writeLines("1+1", mock_runner)
  mockery::stub(fn, "build_doRUnit_wrapper", mockery::mock(mock_runner))
  
  mockery::stub(fn, "covr::environment_coverage", mockery::mock(structure(list(), class = "covr_coverage")))
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(0.0))
  mock_runTests_runner <- mockery::mock(mock_runner)
  mockery::stub(fn, "build_runit_runTests_runner", mock_runTests_runner)
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(totalcoverage = 75.0)))
  mockery::stub(fn, "compute_total_coverage", mockery::mock(list(total_cov = 0.75, percent_cov = 75.0, n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L)))
  
  mock_msg <- mockery::mock(NULL, cycle = TRUE)
  mockery::stub(fn, "message", mock_msg)
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  msg_calls <- mockery::mock_args(mock_msg)
  all_msgs <- vapply(msg_calls, function(x) paste0(x, collapse = ""), character(1))
  expect_true(any(grepl("doRUnit.R produced 0% coverage.*retrying with runTests-style runner", all_msgs)))
  mockery::expect_called(mock_runTests_runner, 1)
})

test_that("package_coverage fallback: when env_coverage returns 0%, message and detach/readLines/package_coverage called", {
  fn <- run_covr_skip_runit_nstf
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  mock_runner <- tempfile(fileext = ".R")
  writeLines(c("library(mypkg)", "1 + 1"), mock_runner)
  
  mockery::stub(fn, "get_pkg_name", mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE, ...) x)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                mockery::mock(data.frame(source_file = "a.R", test_file = "test-a.R", evidence = "match", has_tests = TRUE, stringsAsFactors = FALSE)))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  mockery::stub(fn, "ls", mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace", mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign", mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace", mockery::mock(NULL))
  mockery::stub(fn, "pkgload::load_all", mockery::mock(invisible(NULL)))
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    base::as.environment(name)
  })
  
  mock_detach <- mockery::mock(TRUE, TRUE)
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "build_runit_custom_runner", mockery::mock(mock_runner))
  
  mockery::stub(fn, "covr::environment_coverage", mockery::mock(structure(list(), class = "covr_coverage")))
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(0.0, 50.0))
  mock_pkg_cov <- structure(list(pkg = "mypkg"), class = "covr_coverage")
  mock_pkg_cov_fn <- mockery::mock(mock_pkg_cov)
  mockery::stub(fn, "covr::package_coverage", mock_pkg_cov_fn)
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(totalcoverage = 50.0)))
  mockery::stub(fn, "compute_total_coverage", mockery::mock(list(total_cov = 0.50, percent_cov = 50.0, n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L)))
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  mock_msg <- mockery::mock(NULL, cycle = TRUE)
  mockery::stub(fn, "message", mock_msg)
  
  res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  
  msg_calls <- mockery::mock_args(mock_msg)
  all_msgs <- vapply(msg_calls, function(x) paste0(x, collapse = ""), character(1))
  expect_true(any(grepl("environment_coverage produced 0%.*retrying with covr::package_coverage", all_msgs)))
  mockery::expect_called(mock_detach, 2L)
  mockery::expect_called(mock_pkg_cov_fn, 1)
})


test_that("package_coverage fallback: when package_coverage fails, cov stays 0% and message or fallback path exercised", {
  fn <- run_covr_skip_runit_nstf
  pkg_source_path <- "/fake/pkg"
  pkg_name <- "mypkg"
  test_pkg_data <- list(bioc_run_ut_path = NA_character_)
  
  mock_runner <- tempfile(fileext = ".R")
  writeLines(c("library(mypkg)", "1 + 1"), mock_runner)
  
  mockery::stub(fn, "get_pkg_name", mockery::mock(pkg_name))
  mockery::stub(fn, "is_runit_layout", mockery::mock(TRUE))
  mockery::stub(fn, "get_runit_test_paths",
                mockery::mock(list(bioc_unitTests = file.path(pkg_source_path, "inst", "unitTests"))))
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = FALSE, ...) x)
  mockery::stub(fn, "get_source_test_mapping_runit_nstf",
                mockery::mock(data.frame(source_file = "a.R", test_file = "test-a.R", evidence = "match", has_tests = TRUE, stringsAsFactors = FALSE)))
  mockery::stub(fn, "prepare_for_bioc_runit", mockery::mock(TRUE))
  mockery::stub(fn, "check_covr_skip_runit_nstf",
                mockery::mock(data.frame(test_file = NA_character_, issue_type = "No tests skipped", stringsAsFactors = FALSE)))
  mockery::stub(fn, "get_function_no_tests", mockery::mock(data.frame(fn = "b", stringsAsFactors = FALSE)))
  mockery::stub(fn, "tidyr::drop_na", function(df) df)
  mockery::stub(fn, "ls", mockery::mock(c("obj1")))
  mockery::stub(fn, "asNamespace", mockery::mock(new.env(), cycle = TRUE))
  mockery::stub(fn, "assign", mockery::mock(NULL, cycle = TRUE))
  mockery::stub(fn, "load_package_datasets", mockery::mock(character()))
  mockery::stub(fn, "loadedNamespaces", mockery::mock(c("methods")))
  mockery::stub(fn, "loadNamespace", mockery::mock(NULL))
  mockery::stub(fn, "detach_pkg_if_attached", mockery::mock(TRUE))
  mockery::stub(fn, "pkgload::load_all", mockery::mock(invisible(NULL)))
  mockery::stub(fn, "as.environment", function(name) {
    if (identical(name, paste0("package:", pkg_name))) return(new.env(parent = emptyenv()))
    base::as.environment(name)
  })
  
  mockery::stub(fn, "build_runit_custom_runner", mockery::mock(mock_runner))
  mockery::stub(fn, "covr::environment_coverage", mockery::mock(structure(list(), class = "covr_coverage")))
  mockery::stub(fn, "covr::percent_coverage", mockery::mock(0.0))
  mockery::stub(fn, "readLines", function(con, ...) c("library(mypkg)", "1 + 1"))
  mockery::stub(fn, "covr::package_coverage", function(...) stop("package_coverage instrumentation failed"))
  mockery::stub(fn, "covr::coverage_to_list", mockery::mock(list(filecoverage = matrix(0, nrow = 1), totalcoverage = 0)))
  mockery::stub(fn, "compute_total_coverage", mockery::mock(list(total_cov = 0.0, percent_cov = 0.0, n_testfiles = 1L, n_no_function_tests = 1L, n_skipped_files = 0L)))
  mockery::stub(fn, "cleanup_and_return_null", function(...) NULL)
  
  out <- testthat::capture_messages(
    res <- fn(pkg_source_path = pkg_source_path, test_pkg_data = test_pkg_data)
  )
  
  expect_identical(res$res_cov$coverage$totalcoverage, 0)
  expect_true(
    any(grepl("package_coverage fallback failed", out, fixed = TRUE)) ||
      any(grepl("environment_coverage produced 0%", out, fixed = TRUE)),
    info = "package_coverage fallback path should be entered (env 0% message) and either inner error message or outer fallback"
  )
})

test_that("prepare_for_bioc_runit returns invisible(TRUE)", {
  fn <- prepare_for_bioc_runit
  
  # detach_pkg_if_attached is called 4 times:
  #   1. detach target package
  #   3× if detach_bioc_infra = TRUE
  mock_detach <- mockery::mock(NULL, NULL, NULL, NULL)
  
  mock_load_all <- mockery::mock(NULL)   # pkgload::load_all
  
  mockery::stub(fn, "detach_pkg_if_attached", mock_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  v <- withVisible(fn(pkg_name = "mypkg",
                      pkg_source_path = "/tmp/mypkg",
                      detach_bioc_infra = TRUE))
  
  expect_true(v$value)        # TRUE returned
  expect_false(v$visible)     # invisibly() applied
})


test_that("load_all called even if detaches throw; try-catch not inside but we still see error", {
  fn <- prepare_for_bioc_runit
  
  # If a detach throws, since the function doesn't try-catch around it, the error
  # should propagate and load_all should not be called. We can still validate behavior.
  throwing_detach <- function(...) stop("detach failure")
  mock_load_all <- mockery::mock(NULL)
  
  mockery::stub(fn, "detach_pkg_if_attached", throwing_detach)
  mockery::stub(fn, "pkgload::load_all", mock_load_all)
  
  expect_error(fn(pkg_name = "mypkg", pkg_source_path = "/tmp/mypkg"), "detach failure")
  # Ensure load_all was not called due to early error
  mockery::expect_called(mock_load_all, 0)
}) 


test_that("errors if RUnit is not available", {
  fn <- check_covr_skip_runit_nstf
  
  mock_req_ns <- mockery::mock(FALSE)
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  
  expect_error(
    fn(pkg_source_path = "/pkg", mapping = data.frame(test_file = character()), test_path = "/tests"),
    "RUnit must be installed"
  )
  
  mockery::expect_called(mock_req_ns, 1)
  args <- mockery::mock_args(mock_req_ns)
  expect_identical(args[[1]][[1]], "RUnit")
  expect_true(isTRUE(args[[1]][["quietly"]]))
})

test_that("cleans mapping via tidyr::drop_na when test_file contains NA", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(test_file = c("test-a.R", NA_character_, "test-b.R"), stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  # drop_na should remove the NA row
  mock_drop_na <- function(df, test_file) df[!is.na(df$test_file), , drop = FALSE]
  
  # Minimal stubs to allow the function to proceed to the lapply without real side effects
  mock_file_exists <- mockery::mock(TRUE, TRUE, cycle = TRUE) # two files exist
  mock_search <- mockery::mock("package:stats", "package:stats", cycle = TRUE) # no RUnit attached
  mock_attach <- mockery::mock(NULL, NULL, cycle = TRUE)
  mock_detach <- mockery::mock(NULL, cycle = TRUE)  # may or may not be called by on.exit
  mock_run <- mockery::mock(structure(list(), class = "RUnitTestData"),
                            structure(list(), class = "RUnitTestData"), 
                            cycle = TRUE)
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0),
                                list(nErr = 0, nFail = 0, nDeactivated = 0), 
                                cycle = TRUE)
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", mock_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  # Both tests pass -> "No tests skipped"
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 1L)
  expect_true(is.na(res$test_file[1]))
  expect_identical(res$issue_type[1], "No tests skipped")
})

test_that("Windows branch: Sys.setenv uses quoted R_TESTS and on.exit restores via Sys.unsetenv when originally unset", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(test_file = "pass.R", stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  mock_sysinfo <- c(sysname = "windows", nodename = "PC", release = "10", version = "", machine = "x86")
  mock_getenv <- mockery::mock(NA_character_)
  mock_setenv <- mockery::mock(NULL, NULL)
  mock_unsetenv <- mockery::mock(NULL)
  
  mock_file_exists <- mockery::mock(TRUE)
  mock_search <- mockery::mock("package:RUnit")
  mock_run <- mockery::mock(structure(list(file = "pass.R"), class = "RUnitTestData"))
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0))
  id_drop_na <- function(df, test_file) df
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "Sys.info", function() mock_sysinfo)
  mockery::stub(fn, "Sys.getenv", mock_getenv)
  mockery::stub(fn, "Sys.setenv", mock_setenv)
  mockery::stub(fn, "Sys.unsetenv", mock_unsetenv)
  mockery::stub(fn, "tidyr::drop_na", id_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  expect_s3_class(res, "data.frame")
  expect_identical(res$issue_type[1], "No tests skipped")
  
  set_calls <- mockery::mock_args(mock_setenv)
  expect_true(
    any(vapply(set_calls, function(x) identical(x, list("R_TESTS" = "")), logical(1))),
    info = "Windows branch: Sys.setenv uses quoted R_TESTS = ''"
  )
  mockery::expect_called(mock_unsetenv, 1)
  unset_args <- mockery::mock_args(mock_unsetenv)
  expect_identical(unset_args[[1]], list("R_TESTS"))
})

test_that("Windows branch: Sys.setenv uses quoted R_TESTS and on.exit restores via Sys.setenv when originally set", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(test_file = "pass.R", stringsAsFactors = FALSE)
  original_value <- "original-R_TESTS-value"
  
  mock_req_ns <- mockery::mock(TRUE)
  mock_sysinfo <- c(sysname = "windows", nodename = "PC", release = "10", version = "", machine = "x86")
  mock_getenv <- mockery::mock(original_value)
  mock_setenv <- mockery::mock(NULL, NULL)
  mock_unsetenv <- mockery::mock(NULL)
  
  mock_file_exists <- mockery::mock(TRUE)
  mock_search <- mockery::mock("package:RUnit")
  mock_run <- mockery::mock(structure(list(file = "pass.R"), class = "RUnitTestData"))
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0))
  id_drop_na <- function(df, test_file) df
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "Sys.info", function() mock_sysinfo)
  mockery::stub(fn, "Sys.getenv", mock_getenv)
  mockery::stub(fn, "Sys.setenv", mock_setenv)
  mockery::stub(fn, "Sys.unsetenv", mock_unsetenv)
  mockery::stub(fn, "tidyr::drop_na", id_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  expect_s3_class(res, "data.frame")
  
  set_calls <- mockery::mock_args(mock_setenv)
  expect_true(
    any(vapply(set_calls, function(x) identical(x, list("R_TESTS" = "")), logical(1))),
    info = "Windows branch: Sys.setenv uses quoted R_TESTS = ''"
  )
  expect_true(
    any(vapply(set_calls, function(x) identical(x, list(R_TESTS = original_value)), logical(1))),
    info = "Windows branch: on.exit restores R_TESTS to original value"
  )
  mockery::expect_called(mock_unsetenv, 0)
})

test_that("removes exact filename from mapping when bioc_run_ut_path provided (case-insensitive)", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("run_unitTests.R", "test-foo.R", "RUN_UNITTESTS.R", "test-bar.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/run_unitTests.R")
  
  # Stub basename() to a specific value to assert exact-match branch is used
  mock_basename <- mockery::mock("run_unitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  
  # Expected: exact "run_unitTests.R" removed via exact-match branch; regex also removes RUN_UNITTESTS.R
  expect_true("test-foo.R" %in% res$test_file)
  expect_true("test-bar.R" %in% res$test_file)
  expect_false("run_unitTests.R" %in% res$test_file)
  expect_false("RUN_UNITTESTS.R" %in% res$test_file)
  
  # basename should be called once with the provided path
  mockery::expect_called(mock_basename, 1)
  args <- mockery::mock_args(mock_basename)
  expect_identical(args[[1]][[1]], "/pkg/tests/run_unitTests.R")
})

test_that("regex fallback removes variations when bioc_run_ut_path is missing/NULL/NA", {
  fn <- remove_run_unitTests_mapping
  
  # No exact name available -> only regex path applies
  mapping <- data.frame(
    test_file = c("run-unitTests.R", "Run_UnitTests.R", "test-x.R", "sub/run_unitTests.R"),
    stringsAsFactors = FALSE
  )
  
  # Case 1: test_pkg_data missing entirely
  res1 <- fn(mapping, test_pkg_data = NULL)
  expect_setequal(res1$test_file, "test-x.R") # note: regex only matches basename
  
  # Case 2: field present but NULL
  res2 <- fn(mapping, test_pkg_data = list(bioc_run_ut_path = NULL))
  expect_setequal(res2$test_file, "test-x.R")
  
  # Case 3: field NA
  res3 <- fn(mapping, test_pkg_data = list(bioc_run_ut_path = NA_character_))
  expect_setequal(res3$test_file, "test-x.R")
})

test_that("does not remove unrelated files; only exact filename or regex variations dropped", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("test-a.R", "runUnitTest.R", "unitTests_run.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/NotRunUnitTests.R")
  
  # basename returns a non-matching value -> exact match won't drop anything
  mock_basename <- mockery::mock("NotRunUnitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  
  # None match the strict regex "run[_-]?unitTests\\.R$"
  expect_setequal(res$test_file, c("test-a.R", "runUnitTest.R", "unitTests_run.R", "test-b.R"))
})

test_that("handles NA entries in mapping$test_file gracefully (no warnings)", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c(NA_character_, "test-a.R", "run_unitTests.R", NA_character_, "test-b.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/whatever.R")
  
  # basename stub not strictly needed but keep deterministic
  mock_basename <- mockery::mock("whatever.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  
  # Only run_unitTests.R is removed; NAs are retained
  expect_setequal(res$test_file, c(NA_character_, "test-a.R", NA_character_, "test-b.R"))
})

test_that("exact match removal works even if mapping has different case than basename", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("RUN_UNITTESTS.R", "test-a.R"),
    stringsAsFactors = FALSE
  )
  test_pkg_data <- list(bioc_run_ut_path = "/pkg/tests/run_unitTests.R")
  
  mock_basename <- mockery::mock("run_unitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  
  res <- fn(mapping, test_pkg_data)
  expect_setequal(res$test_file, "test-a.R")
})

test_that("regex removes both underscore and hyphen variants, any case", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(
    test_file = c("run_unitTests.R", "run-unitTests.R", "RUN_UNITTESTS.R", "test-a.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  res <- fn(mapping, test_pkg_data = NULL)
  expect_setequal(res$test_file, c("test-a.R", "test-b.R"))
})

test_that("empty mapping returns empty data.frame (no errors)", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(test_file = character(), stringsAsFactors = FALSE)
  res <- fn(mapping, test_pkg_data = NULL)
  
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 0L)
})

test_that("basename is consulted only when bioc_run_ut_path is usable", {
  fn <- remove_run_unitTests_mapping
  
  mapping <- data.frame(test_file = c("run_unitTests.R", "test-a.R"), stringsAsFactors = FALSE)
  
  # Case: usable path -> basename called
  mock_basename <- mockery::mock("run_unitTests.R")
  mockery::stub(fn, "basename", mock_basename)
  res1 <- fn(mapping, list(bioc_run_ut_path = "/pkg/tests/run_unitTests.R"))
  mockery::expect_called(mock_basename, 1)
  expect_setequal(res1$test_file, "test-a.R")
  
  # Case: NA path -> basename should not be called
  mock_basename2 <- mockery::mock("should-not-be-called")
  mockery::stub(fn, "basename", mock_basename2)
  res2 <- fn(mapping, list(bioc_run_ut_path = NA_character_))
  mockery::expect_called(mock_basename2, 0)
  expect_setequal(res2$test_file, c("test-a.R"))
  
  # Case: NULL path -> basename should not be called
  mock_basename3 <- mockery::mock("should-not-be-called")
  mockery::stub(fn, "basename", mock_basename3)
  res3 <- fn(mapping, list(bioc_run_ut_path = NULL))
  mockery::expect_called(mock_basename3, 0)
  expect_setequal(res3$test_file, c("test-a.R"))
})  

test_that("classifies missing, error, failure, skipped, and passing files correctly", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(
    test_file = c(
      "test-missing.R",
      "test-error-throw.R",
      "test-error-count.R",
      "test-failure.R",
      "test-skipped.R",
      "test-pass.R",
      NA_character_
    ),
    stringsAsFactors = FALSE
  )
  
  mock_req_ns <- mockery::mock(TRUE)
  
  # drop_na should be called and remove NA
  mock_drop_na <- function(df, test_file) df[!is.na(df$test_file), , drop = FALSE]
  
  # file.exists -> FALSE only for test-missing.R
  mock_file_exists <- function(path) {
    if (grepl("test-missing\\.R$", path)) return(FALSE)
    TRUE
  }
  
  # Pretend RUnit is already on the search path so we don't re-attach each time
  mock_search <- mockery::mock("package:RUnit", "package:RUnit", "package:RUnit",
                               "package:RUnit", "package:RUnit", "package:RUnit")
  
  # runTestFile: throws for test-error-throw, otherwise returns a testdata object
  mock_run <- function(absFileName, useOwnErrorHandler, testFuncRegexp) {
    bn <- basename(absFileName)
    if (bn == "test-error-throw.R") stop("parse/runtime error")
    structure(list(file = bn), class = "RUnitTestData")
  }
  
  # getErrors: return different counts per file name
  mock_get_err <- function(td) {
    bn <- td$file
    switch(
      bn,
      "test-error-count.R" = list(nErr = 1, nFail = 0, nDeactivated = 0),
      "test-failure.R"     = list(nErr = 0, nFail = 2, nDeactivated = 0),
      "test-skipped.R"     = list(nErr = 0, nFail = 0, nDeactivated = 3),
      # default: passing
      list(nErr = 0, nFail = 0, nDeactivated = 0)
    )
  }
  
  mock_attach <- mockery::mock(NULL) # not used because search() includes RUnit
  mock_detach <- mockery::mock(NULL)
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", mock_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  # Expected issues (order follows mapping after drop_na)
  expected <- data.frame(
    test_file  = c("test-missing.R", "test-error-throw.R", "test-error-count.R", "test-failure.R", "test-skipped.R"),
    issue_type = c("missing",        "error",              "error",               "failure",         "skipped"),
    stringsAsFactors = FALSE
  )
  
  expect_s3_class(res, "data.frame")
  expect_identical(res[, c("test_file", "issue_type")], expected)
})

test_that("when all files are passing, returns 'No tests skipped' row", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(test_file = c("t1.R", "t2.R"), stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  # anyNA(mapping$test_file) is FALSE -> ensure we don't call tidyr::drop_na (we make it throw if called)
  throwing_drop_na <- function(...) stop("drop_na should not be called when no NA present")
  
  mock_file_exists <- mockery::mock(TRUE, TRUE)
  mock_search <- mockery::mock("package:RUnit", "package:RUnit")
  mock_run <- mockery::mock(structure(list(file = "t1.R"), class = "RUnitTestData"),
                            structure(list(file = "t2.R"), class = "RUnitTestData"))
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0),
                                list(nErr = 0, nFail = 0, nDeactivated = 0))
  mock_attach <- mockery::mock(NULL)
  mock_detach <- mockery::mock(NULL)
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", throwing_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 1L)
  expect_true(is.na(res$test_file[1]))
  expect_identical(res$issue_type[1], "No tests skipped")
})



test_that("attaches RUnit namespace only when not already on search path", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(test_file = "pass.R", stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  # First call to search() -> no RUnit; inside on.exit check may call again, so provide two values
  mock_search <- mockery::mock("package:stats", "package:RUnit")
  mock_attach <- mockery::mock(NULL)  # should be called once
  mock_detach <- mockery::mock(NULL)
  
  mock_file_exists <- mockery::mock(TRUE)
  mock_run <- mockery::mock(structure(list(file = "pass.R"), class = "RUnitTestData"))
  mock_get_err <- mockery::mock(list(nErr = 0, nFail = 0, nDeactivated = 0))
  id_drop_na <- function(df, test_file) df
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", id_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  expect_s3_class(res, "data.frame")
  
  mockery::expect_called(mock_attach, 1)
})

test_that("runTestFile error path is captured as 'error' without crashing", {
  fn <- check_covr_skip_runit_nstf
  
  mapping <- data.frame(test_file = "test-error-throw.R", stringsAsFactors = FALSE)
  
  mock_req_ns <- mockery::mock(TRUE)
  id_drop_na <- function(df, test_file) df
  mock_file_exists <- mockery::mock(TRUE)
  mock_search <- mockery::mock("package:RUnit")
  mock_attach <- mockery::mock(NULL)
  mock_detach <- mockery::mock(NULL)
  
  # Force runTestFile to throw -> tryCatch will return class "RUnitTestData-ERROR"
  mock_run <- function(...) stop("boom")
  # getErrors should not be called in this path
  mock_get_err <- mockery::mock()
  
  mockery::stub(fn, "requireNamespace", mock_req_ns)
  mockery::stub(fn, "tidyr::drop_na", id_drop_na)
  mockery::stub(fn, "file.exists", mock_file_exists)
  mockery::stub(fn, "search", mock_search)
  mockery::stub(fn, "attachNamespace", mock_attach)
  mockery::stub(fn, "detach", mock_detach)
  mockery::stub(fn, "RUnit::runTestFile", mock_run)
  mockery::stub(fn, "RUnit::getErrors", mock_get_err)
  
  res <- fn(pkg_source_path = "/pkg", mapping = mapping, test_path = "/tests")
  
  expect_identical(nrow(res), 1L)
  expect_identical(res$test_file[1], "test-error-throw.R")
  expect_identical(res$issue_type[1], "error")
  mockery::expect_called(mock_get_err, 0)
})


test_that("returns bioc_unitTests when testdir already exists", {
  fn <- get_runit_test_paths
  
  # Normalized testdir value
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) {
    # Return a consistent normalized path regardless of input
    if (grepl("^/work/pkg/tests$", path)) return("/work/pkg/tests")
    path
  }
  mock_dir_exists <- function(path) path == "/work/pkg/tests"
  
  # Stubs
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  
  res <- fn(test_pkg_data = list(), testdir = "/work/pkg/tests")
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, "/work/pkg/tests")
  expect_false("fallback" %in% names(res))
})

test_that("finds package root at testdir and returns inst/unitTests if it exists", {
  fn <- get_runit_test_paths
  
  # Paths
  raw_testdir <- "/work/pkg/tests"
  normalized_testdir <- "/work/pkg/tests"
  pkg_root <- "/work/pkg"
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) {
    # Normalize testdir and pkg_root and bioc_unit deterministically
    if (path == raw_testdir) return(normalized_testdir)
    if (path == pkg_root) return(pkg_root)
    if (path == bioc_unit) return(bioc_unit)
    path
  }
  
  # testdir does NOT exist; bioc unit dir DOES
  mock_dir_exists <- function(path) {
    path == bioc_unit
  }
  
  # file.exists(DESCRIPTION) returns TRUE only at testdir's root (/work/pkg)
  mock_file_exists <- function(path) {
    path == file.path(pkg_root, "DESCRIPTION")
  }
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, bioc_unit)
  expect_false("fallback" %in% names(res))
})

test_that("finds package root at dirname(testdir) and returns inst/unitTests", {
  fn <- get_runit_test_paths
  
  # Layout:
  # testdir: /work/pkg/tests/unit
  # dirname(testdir): /work/pkg/tests
  # dirname(dirname(testdir)): /work/pkg
  raw_testdir <- "/work/pkg/tests/unit"
  normalized_testdir <- raw_testdir
  pkg_root <- "/work/pkg"
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) {
    if (path %in% c(raw_testdir, pkg_root, bioc_unit)) return(path)
    path
  }
  
  mock_dir_exists <- function(path) {
    # testdir does not exist; bioc_unit exists
    path == bioc_unit
  }
  
  mock_file_exists <- function(path) {
    # Mark only /work/pkg as a package root
    path == file.path(pkg_root, "DESCRIPTION")
  }
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, bioc_unit)
})

test_that("finds package root at dirname(dirname(testdir)) and returns inst/unitTests", {
  fn <- get_runit_test_paths
  
  raw_testdir <- "/work/pkg/a/b"
  pkg_root <- "/work/pkg"
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) path
  mock_dir_exists <- function(path) path == bioc_unit
  mock_file_exists <- function(path) path == file.path(pkg_root, "DESCRIPTION")
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("bioc_unitTests" %in% names(res))
  expect_identical(res$bioc_unitTests, bioc_unit)
})

test_that("returns fallback when testdir doesn't exist and no package root found", {
  fn <- get_runit_test_paths
  
  raw_testdir <- "/nowhere/tests"
  normalized_testdir <- "/nowhere/tests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) normalized_testdir
  mock_dir_exists <- function(path) FALSE
  mock_file_exists <- function(path) FALSE
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("fallback" %in% names(res))
  expect_identical(res$fallback, normalized_testdir)
  expect_false("bioc_unitTests" %in% names(res))
})

test_that("returns fallback if package root found but inst/unitTests does not exist", {
  fn <- get_runit_test_paths
  
  raw_testdir <- "/work/pkg/tests"
  pkg_root <- "/work/pkg"
  normalized_testdir <- raw_testdir
  bioc_unit <- "/work/pkg/inst/unitTests"
  
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) path
  mock_dir_exists <- function(path) FALSE  # inst/unitTests not present
  mock_file_exists <- function(path) path == file.path(pkg_root, "DESCRIPTION")
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  expect_true("fallback" %in% names(res))
  expect_identical(res$fallback, normalized_testdir)
})

test_that("handles duplicate candidate paths via unique()", {
  fn <- get_runit_test_paths
  
  # If testdir == dirname(testdir), unique() collapses duplicates.
  raw_testdir <- "/flat"
  # Simulate a flat structure where dirname("/flat") == "/"
  mock_normalize <- function(path, winslash = "/", mustWork = FALSE) path
  mock_dir_exists <- function(path) FALSE
  mock_file_exists <- function(path) FALSE
  
  mockery::stub(fn, "normalizePath", mock_normalize)
  mockery::stub(fn, "dir.exists", mock_dir_exists)
  mockery::stub(fn, "file.exists", mock_file_exists)
  
  res <- fn(test_pkg_data = list(), testdir = raw_testdir)
  
  expect_true("fallback" %in% names(res))
  expect_identical(res$fallback, raw_testdir)
})



test_that("errors when no R source files are found", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/proj/pkg"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/proj/pkg/tests"
  
  # Stub list.files so there are NO source files
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(character())  # no .R under R/
    if (path == test_dir)   return(character(0)) # doesn't matter
    character(0)
  }
  mockery::stub(fn, "list.files", mock_list_files)
  
  # Use a simple pattern (or regexp = NA) to avoid brittle matching
  expect_error(
    fn(pkg_source_path = pkg_source_path, test_dir = test_dir),
    regexp = "No R files found under:"
  )
})

test_that("warns when no .R test files are found and maps NA tests", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/proj/pkg"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/proj/pkg/tests"
  
  src_files <- file.path(source_dir, c("alpha.R", "beta.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)    # two sources
    if (path == test_dir)   return(character())  # no tests -> warning
    character()
  }
  # No test content will be read because length(test_files) == 0
  mock_readLines <- function(con, warn = FALSE) stop("readLines should not be called")
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines", mock_readLines)
  
  expect_warning(
    res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir),
    regexp = "No R test files found under:"
  )
  
  expect_s3_class(res, "data.frame")
  expect_identical(nrow(res), 2L)
  expect_identical(res$source_file, c("alpha.R", "beta.R"))
  expect_true(all(is.na(res$test_file)))
  expect_true(all(is.na(res$evidence)))
  expect_identical(res$has_tests, c(FALSE, FALSE))
})

test_that("filename-based mapping: tokenized, case-insensitive, picks first match", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/pkg"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/pkg/tests"
  
  src_files  <- file.path(source_dir, c("foo.R", "bar.R"))
  test_files <- file.path(test_dir, c("test-foo.R", "ZZZ_foo_extra.R", "TEST-bar.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  # No content-based checks needed
  mock_readLines <- function(con, warn = FALSE) character()
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines", mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_s3_class(res, "data.frame")
  expect_identical(res$source_file, c("foo.R", "bar.R"))
  expect_identical(res$test_file,   c("test-foo.R", "TEST-bar.R")) # first match wins
  expect_identical(res$evidence,    c("filename_match", "filename_match"))
  expect_identical(res$has_tests,   c(TRUE, TRUE))
})

test_that("content-based mapping via RUnit test function names when filename matching fails", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/p"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/p/tests"
  
  # One source 'widget.R'; one test 'test-something.R'
  src_files  <- file.path(source_dir, "widget.R")
  test_files <- file.path(test_dir,   "test-something.R")
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  
  # Provide a FUNCTION DEF in source (ensures symbol extraction works)
  # Provide an RUnit test function line that contains the symbol
  mock_readLines <- function(con, warn = FALSE) {
    f <- as.character(con)
    if (identical(basename(f), "widget.R")) {
      return(c("widget <- function(x) { x }"))
    }
    if (identical(basename(f), "test-something.R")) {
      return(c("test_widget <- function() {",
               "  # assertions here",
               "}"))
    }
    character()
  }
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, "widget.R")
  expect_identical(res$test_file,   "test-something.R")
  expect_identical(res$evidence,    "content_match")
  expect_true(res$has_tests)
})

test_that("content-based mapping uses symbols from function defs (avoid setGeneric parsing)", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/root"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/root/tests"
  
  src_files  <- file.path(source_dir, c("generics.R", "funcs.R"))
  test_files <- file.path(test_dir,   c("t-generic.R", "t-func.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  
  # IMPORTANT: Because the function's setGeneric grep uses fixed=TRUE with an escaped quote,
  # it's safer to provide FUNCTION DEFS in sources so symbol extraction surely works.
  mock_readLines <- function(con, warn = FALSE) {
    f <- as.character(con)
    switch(
      basename(f),
      "generics.R" = c(
        "doStuff <- function(...) TRUE"   # symbol: doStuff
      ),
      "funcs.R"    = c(
        "my_func <- function(a, b) { a + b }",
        "other <- function() 1"
      ),
      "t-generic.R" = c(
        "test_doStuff <- function() { TRUE }"  # detects 'doStuff' via RUnit name
      ),
      "t-func.R" = c(
        "test_my_func <- function() { TRUE }"
      ),
      character()
    )
  }
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, c("generics.R", "funcs.R"))
  expect_identical(res$test_file,   c("t-generic.R", NA))
  expect_identical(res$evidence,    c("content_match", NA))
  expect_identical(res$has_tests,   c(TRUE, FALSE))
})

test_that("short symbol (<= 2 chars) matches via filename tokens", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/a"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/a/tests"
  
  src_files  <- file.path(source_dir, c("t.R", "x1.R"))
  test_files <- file.path(test_dir,   c("test-t.R", "we-test-x1.R"))
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  mock_readLines <- function(con, warn = FALSE) character()
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, c("t.R", "x1.R"))
  expect_identical(res$test_file,   c("test-t.R", "we-test-x1.R"))
  expect_identical(res$evidence,    c("filename_match", "filename_match"))
  expect_identical(res$has_tests,   c(TRUE, TRUE))
})

test_that("returns NA mapping for source with no filename or content evidence", {
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/no"
  source_dir <- file.path(pkg_source_path, "R")
  test_dir <- "/no/tests"
  
  src_files  <- file.path(source_dir, "lonely.R")
  test_files <- file.path(test_dir,   c("test-a.R", "test-b.R"))  # no relevant content
  
  mock_list_files <- function(path, pattern, full.names, recursive) {
    if (path == source_dir) return(src_files)
    if (path == test_dir)   return(test_files)
    character()
  }
  
  mock_readLines <- function(con, warn = FALSE) {
    f <- as.character(con)
    if (identical(basename(f), "lonely.R")) return(c("lonely <- function() 1"))
    if (basename(f) %in% c("test-a.R", "test-b.R")) return(c("test_other <- function() 1"))
    character()
  }
  
  mockery::stub(fn, "list.files", mock_list_files)
  mockery::stub(fn, "readLines",  mock_readLines)
  
  res <- fn(pkg_source_path = pkg_source_path, test_dir = test_dir)
  
  expect_identical(res$source_file, "lonely.R")
  expect_true(is.na(res$test_file))
  expect_true(is.na(res$evidence))
  expect_false(res$has_tests)
})


test_that("detect_runit_test_prefix detects runit.<name>.R pattern", {
  files <- c(
    "/fake/dir/runit.myfun.R",
    "/fake/dir/runit.other_test.R"
  )
  expect_equal(detect_runit_test_prefix(files), "runit")
})


test_that("detect_runit_test_prefix detects Test.<name>.R pattern", {
  files <- c(
    "/fake/dir/Test.myFun.R",
    "/fake/dir/Test.Other.R"
  )
  expect_equal(detect_runit_test_prefix(files), "Test")
})


test_that("detect_runit_test_prefix detects test_<name>.R pattern (underscore)", {
  files <- c(
    "/path/to/test_myFunc.R",
    "/path/to/test_otherThing.R"
  )
  expect_equal(detect_runit_test_prefix(files), "test")
})


test_that("detect_runit_test_prefix detects test.<name>.R pattern (dot)", {
  files <- c(
    "/path/test.myFunc.R",
    "/path/test.thing_more.R"
  )
  expect_equal(detect_runit_test_prefix(files), "test")
})


test_that("detect_runit_test_prefix returns first detected prefix among many", {
  files <- c(
    "/fake/zzz/NOT_A_TEST.R",
    "/fake/dir/runit.someFun.R",   # should win immediately
    "/fake/dir/Test.foo.R"         # never reached
  )
  expect_equal(detect_runit_test_prefix(files), "runit")
})


test_that("detect_runit_test_prefix falls back to default 'test' when nothing matches", {
  files <- c(
    "alpha.R",
    "beta.R",
    "not_a_test_file.txt"
  )
  expect_equal(detect_runit_test_prefix(files), "test")
})



test_that("short symbol triggers filename token match via get_source_test_mapping_runit_nstf", {
  
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_dir <- "/fake/tests"
  
  src_files <- c("/fake/pkg/R/x.R")
  test_files <- c("/fake/tests/test_x.R")
  
  # ---- CORRECT list.files STUB (dispatch on path) ----
  mockery::stub(fn, "list.files", function(path, pattern, full.names, recursive) {
    if (grepl("/R$", path)) {
      return(src_files)
    } else {
      return(test_files)
    }
  })
  
  # extract_symbols returns the symbol being searched
  mockery::stub(fn, "extract_symbols", function(...) "x")
  
  # readLines must return NON-empty or function returns FALSE early
  mockery::stub(fn, "readLines", function(...) c("dummy line"))
  
  # Tokenizer: filename tokens = c("x"), matches
  mockery::stub(fn, "tokenize_stem", function(stem) c("x"))
  
  # ---- RUN ----
  mapping <- fn(pkg_source_path, test_dir)
  
  # ---- ASSERTIONS ----
  expect_equal(mapping$test_file, "test_x.R")
  expect_equal(mapping$evidence, "filename_match")  # correct Pass 1 match
  expect_true(mapping$has_tests)
})


test_that("long symbol triggers filename token match via get_source_test_mapping_runit_nstf", {
  
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_dir <- "/fake/tests"
  
  src_files  <- "/fake/pkg/R/alpha.R"
  test_files <- "/fake/tests/test_alpha.R"
  
  # CORRECT list.files dispatcher
  mockery::stub(fn, "list.files", function(path, pattern, full.names, recursive) {
    if (grepl("/R$", path)) src_files else test_files
  })
  
  mockery::stub(fn, "extract_symbols", function(...) "alpha")
  
  mockery::stub(fn, "readLines", function(...) c("ignored"))
  mockery::stub(fn, "tokenize_stem", function(stem) c("alpha"))
  
  mapping <- fn(pkg_source_path, test_dir)
  
  expect_equal(mapping$test_file, "test_alpha.R")
  expect_equal(mapping$evidence, "filename_match")
  expect_true(mapping$has_tests)
})


test_that("test_that description triggers content_match in get_source_test_mapping_runit_nstf", {
  
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_dir <- "/fake/tests"
  
  src_files  <- "/fake/pkg/R/compute.R"
  test_files <- "/fake/tests/test_misc.R"
  
  mockery::stub(fn, "list.files", function(path, pattern, full.names, recursive) {
    if (grepl("/R$", path)) src_files else test_files
  })
  
  mockery::stub(fn, "extract_symbols", function(...) "compute")
  
  # Force description match
  mockery::stub(fn, "readLines", function(...) c(
    'test_that("checks compute functionality", {',
    "  expect_true(TRUE)",
    "})"
  ))
  
  # Force filename mismatch → Pass 1 fails → Pass 2 content match
  mockery::stub(fn, "tokenize_stem", function(stem) c("nothing"))
  
  mapping <- fn(pkg_source_path, test_dir)
  
  expect_equal(mapping$test_file, "test_misc.R")
  expect_equal(mapping$evidence, "content_match")
  expect_true(mapping$has_tests)
})


test_that("RUnit-style test_<symbol> triggers content_match in get_source_test_mapping_runit_nstf", {
  
  fn <- get_source_test_mapping_runit_nstf
  
  pkg_source_path <- "/fake/pkg"
  test_dir <- "/fake/tests"
  
  src_files  <- "/fake/pkg/R/compute.R"
  test_files <- "/fake/tests/test_compute.R"
  
  mockery::stub(fn, "list.files", function(path, pattern, full.names, recursive) {
    if (grepl("/R$", path)) src_files else test_files
  })
  
  mockery::stub(fn, "extract_symbols", function(...) "compute")
  
  mockery::stub(fn, "readLines", function(...) c(
    "test_compute <- function() {",
    "  checkTrue(TRUE)",
    "}"
  ))
  
  # No filename match
  mockery::stub(fn, "tokenize_stem", function(stem) c("unrelated"))
  
  mapping <- fn(pkg_source_path, test_dir)
  
  expect_equal(mapping$test_file, "test_compute.R")
  expect_equal(mapping$evidence, "filename_match")
  expect_true(mapping$has_tests)
})

test_that("build_runit_custom_runner returns path and writes correct content for prefix runit", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  out <- fn(tests)
  
  expect_identical(out, mock_runner)
  expect_true(any(grepl("RUnit::runTestFile", written_lines)))
  expect_true(any(grepl("requireNamespace.*RUnit", written_lines)))
  expect_true(any(grepl('testFuncRegexp = "\\^runit', written_lines)))
})

test_that("build_runit_custom_runner uses runit regex when prefix is runit", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests)
  
  run_line <- written_lines[grepl("runTestFile", written_lines)]
  expect_true(any(grepl('testFuncRegexp = "\\^runit\\[\\.\\]\\|\\^test\\.\\+"', run_line)))
})

test_that("build_runit_custom_runner uses Test regex when prefix is Test", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/Test.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "Test")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests)
  
  run_line <- written_lines[grepl("runTestFile", written_lines)]
  expect_true(any(grepl('testFuncRegexp = "\\^Test\\[\\.\\]\\|\\^test\\.\\+"', run_line)))
})

test_that("build_runit_custom_runner uses default test regex when prefix is neither runit nor Test", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/test_foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "test")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests)
  
  run_line <- written_lines[grepl("runTestFile", written_lines)]
  expect_true(any(grepl('testFuncRegexp = "\\^test\\.\\+"', run_line)))
})

test_that("build_runit_custom_runner includes library line when pkg_name is provided", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests, pkg_name = "mypkg")
  
  expect_true(any(grepl('library\\(package = "mypkg"', written_lines)))
})

test_that("build_runit_custom_runner omits library line when pkg_name is NULL", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests, pkg_name = NULL)
  
  expect_false(any(grepl('library\\(package = ', written_lines)))
})

test_that("build_runit_custom_runner omits library line when pkg_name is empty string", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests, pkg_name = "")
  
  expect_false(any(grepl('library\\(package = ', written_lines)))
})

test_that("build_runit_custom_runner includes wd_lines when pkg_root exists", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  pkg_root <- "/fake/pkg"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  mockery::stub(fn, "dir.exists", function(x) x == pkg_root)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests, pkg_root = pkg_root)
  
  expect_true(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_true(any(grepl("on.exit\\(setwd\\(owd\\)", written_lines)))
  expect_true(any(grepl('setwd\\(', written_lines)))
})

test_that("build_runit_custom_runner omits wd_lines when pkg_root is NULL", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests, pkg_root = NULL)
  
  expect_false(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_false(any(grepl("setwd\\(", written_lines)))
})

test_that("build_runit_custom_runner omits wd_lines when dir.exists(pkg_root) is FALSE", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  pkg_root <- "/nonexistent/pkg"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  mockery::stub(fn, "dir.exists", function(x) FALSE)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests, pkg_root = pkg_root)
  
  expect_false(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_false(any(grepl("setwd\\(", written_lines)))
})

test_that("build_runit_custom_runner writes one runTestFile line per test path", {
  fn <- build_runit_custom_runner
  tests <- c("/fake/pkg/inst/unitTests/runit.a.R", "/fake/pkg/inst/unitTests/runit.b.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(tests)
  
  run_lines <- written_lines[grepl("runTestFile", written_lines)]
  expect_length(run_lines, 2L)
  expect_true(any(grepl("runit.a.R", run_lines)))
  expect_true(any(grepl("runit.b.R", run_lines)))
})

test_that("build_runit_runTests_runner returns path and writes correct content for prefix runit", {
  fn <- build_runit_runTests_runner
  pkg_name <- "mypkg"
  test_files <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  out <- fn(pkg_name, test_files)
  
  expect_identical(out, mock_runner)
  expect_true(any(grepl("RUnit::runTestFile", written_lines)))
  expect_true(any(grepl("requireNamespace.*RUnit", written_lines)))
  expect_true(any(grepl('library\\(package = "mypkg"', written_lines)))
  expect_true(any(grepl('testFuncRegexp = "\\^runit', written_lines)))
})

test_that("build_runit_runTests_runner uses runit regex when prefix is runit", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files)
  
  run_line <- written_lines[grepl("runTestFile", written_lines)]
  expect_true(any(grepl('testFuncRegexp = "\\^runit\\[\\.\\]\\|\\^test\\.\\+"', run_line)))
})

test_that("build_runit_runTests_runner uses Test regex when prefix is Test", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/Test.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "Test")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files)
  
  run_line <- written_lines[grepl("runTestFile", written_lines)]
  expect_true(any(grepl('testFuncRegexp = "\\^Test\\[\\.\\]\\|\\^test\\.\\+"', run_line)))
})

test_that("build_runit_runTests_runner uses default test regex when prefix is neither runit nor Test", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/test_foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "test")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files)
  
  run_line <- written_lines[grepl("runTestFile", written_lines)]
  expect_true(any(grepl('testFuncRegexp = "\\^test\\.\\+"', run_line)))
})

test_that("build_runit_runTests_runner always includes library line", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("fBasics", test_files)
  
  expect_true(any(grepl('library\\(package = "fBasics"', written_lines)))
})

test_that("build_runit_runTests_runner includes wd_lines when pkg_root exists", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  pkg_root <- "/fake/pkg"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  mockery::stub(fn, "dir.exists", function(x) x == pkg_root)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files, pkg_root = pkg_root)
  
  expect_true(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_true(any(grepl("on.exit\\(setwd\\(owd\\)", written_lines)))
  expect_true(any(grepl('setwd\\(', written_lines)))
})

test_that("build_runit_runTests_runner omits wd_lines when pkg_root is NULL", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files, pkg_root = NULL)
  
  expect_false(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_false(any(grepl("setwd\\(", written_lines)))
})

test_that("build_runit_runTests_runner omits wd_lines when dir.exists(pkg_root) is FALSE", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/runit.foo.R")
  pkg_root <- "/nonexistent/pkg"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  mockery::stub(fn, "dir.exists", function(x) FALSE)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files, pkg_root = pkg_root)
  
  expect_false(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_false(any(grepl("setwd\\(", written_lines)))
})

test_that("build_runit_runTests_runner writes one runTestFile line per test path", {
  fn <- build_runit_runTests_runner
  test_files <- c("/fake/pkg/inst/unitTests/runit.a.R", "/fake/pkg/inst/unitTests/runit.b.R")
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  mockery::stub(fn, "detect_runit_test_prefix", function(...) "runit")
  mockery::stub(fn, "normalizePath", function(x, winslash = "/", mustWork = TRUE) x)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn("mypkg", test_files)
  
  run_lines <- written_lines[grepl("runTestFile", written_lines)]
  expect_length(run_lines, 2L)
  expect_true(any(grepl("runit.a.R", run_lines)))
  expect_true(any(grepl("runit.b.R", run_lines)))
})

test_that("build_doRUnit_wrapper returns runner path and writes correct content", {
  fn <- build_doRUnit_wrapper
  doRUnit_path <- "/fake/pkg/tests/doRUnit.R"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  out <- fn(doRUnit_path)
  
  expect_identical(out, mock_runner)
  expect_true(any(grepl("owd <- getwd\\(\\)", written_lines)))
  expect_true(any(grepl("on.exit\\(setwd\\(owd\\)", written_lines)))
  expect_true(any(grepl('setwd\\("/fake/pkg/tests"\\)', written_lines)))
  expect_true(any(grepl('source\\("/fake/pkg/tests/doRUnit.R"\\)', written_lines)))
})

test_that("build_doRUnit_wrapper uses dirname of doRUnit_path for setwd", {
  fn <- build_doRUnit_wrapper
  doRUnit_path <- "/some/nested/tests/dir/doRUnit.R"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(doRUnit_path)
  
  expect_true(any(grepl('setwd\\("/some/nested/tests/dir"\\)', written_lines)))
})

test_that("build_doRUnit_wrapper sources doRUnit_path exactly", {
  fn <- build_doRUnit_wrapper
  doRUnit_path <- "/path/to/doRUnit.R"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(doRUnit_path)
  
  expect_true(any(grepl('source\\("/path/to/doRUnit.R"\\)', written_lines)))
})

test_that("build_doRUnit_wrapper writes exactly four lines", {
  fn <- build_doRUnit_wrapper
  doRUnit_path <- "/fake/tests/doRUnit.R"
  mock_runner <- tempfile(fileext = ".R")
  
  mockery::stub(fn, "tempfile", function(fileext = "") mock_runner)
  
  written_lines <- NULL
  mockery::stub(fn, "writeLines", function(text, con, ...) {
    written_lines <<- text
    base::writeLines(text, con)
  })
  
  fn(doRUnit_path)
  
  expect_length(written_lines, 4L)
})

#' Install a Package from a Local Source Directory
#'
#' Attempts to install an R package from a local source directory using
#' \code{remotes::install_local()}. The function reports on whether installation
#' succeeded, whether the package was already installed, or whether the provided
#' source path does not exist.
#'
#' The display name of the package is derived from the input path using
#' \code{get_pkg_name()}.
#'
#' @param pkg_source_path Character string. Path to the local package source
#'   directory (e.g., an unpacked package or extracted tarball path).
#'
#' @return
#' A logical value indicating whether the package is installed after running
#' the function.  
#'
#' The returned object is always of class \code{logical}:
#' \itemize{
#'   \item \code{TRUE} — The package is already installed or was successfully installed.
#'   \item \code{FALSE} — Installation failed or the path does not exist.
#' }
#'
#' @section Output Meaning:
#' \code{TRUE} does not necessarily imply that the installation occurred during
#' this function call—it may also mean the package was already installed.  
#'
#' \code{FALSE} indicates a failure to install or an invalid path. All diagnostic
#' messages are printed via \code{message()} for user visibility.
#'
#' @examples
#' \dontrun{
#' results <- install_package_local("pkg_source_path")
#' print(results)
#' }
#' @export
install_package_local <- function(pkg_source_path) {
  
  pkg_disp <- get_pkg_name(pkg_source_path)
  message(paste0("installing ", pkg_disp, " locally"))
  
  # Check if the package source path exists
  if (!dir.exists(pkg_source_path)) {
    message(paste0("No such file or directory: ", pkg_source_path))
    package_installed <- FALSE
  } else if (requireNamespace(pkg_disp, quietly = TRUE)) {
    message(paste0(pkg_disp, " is already installed"))
    package_installed <- TRUE
  } else {
    tryCatch(
      {
        remotes::install_local(
          pkg_source_path,
          dependencies = TRUE,
          upgrade = "never",
          force = TRUE,
          quiet = TRUE,
          INSTALL_opts = "--with-keep.source"
        )
        message(paste0(pkg_disp, " installed locally"))
        package_installed <- TRUE
      },
      error = function(cond) {
        message(paste0("Local installation issue is: ", cond))
        message(paste0(pkg_disp, " not installed locally"))
        package_installed <- FALSE
      }
    )
  }
  
  return(package_installed)
}

test_that("install_package_local works correctly", {
  # get system file
  r = getOption("repos")
  r["CRAN"] = "http://cran.us.r-project.org"
  options(repos = r)
  
  # Copy test package to a temp file
  dp_orig <- system.file("test-data", 
                         "test.package.0001_0.1.0.tar.gz", 
                         package = "test.assessr")
  dp <- tempfile(fileext = ".tar.gz")
  file.copy(dp_orig, dp)
  
  # Defer cleanup of copied tarball
  withr::defer(unlink(dp), envir = parent.frame())
  
  # Defer cleanup of unpacked source directory
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE),
               envir = parent.frame())
  
  suppressWarnings(pkg_source_path <-   
                     unpack_tarball(dp))
  
  
  # Test the install_package_local function
  result <- suppressWarnings(install_package_local(pkg_source_path))
  expect_true(result)
  
  # Clean up
  withr::defer(unlink(pkg_source_path, recursive = TRUE))  # Test directory cleanup
  
  # Attempt to unload the namespace without checking installed.packages()
  try(unloadNamespace("test.package.0001"), silent = TRUE)
  
})



test_that("install_package_local handles errors correctly", {
  
  temp_dir <- tempdir()
  
  # Create an invalid path to generate an error
  invalid_pkg_path <- file.path(temp_dir, "invalidPackage")
  
  expect_message(
    result <- install_package_local(invalid_pkg_path),
    regexp = "No such file or directory"
  )
  
  expect_false(result)
  # Clean up
  unlink(temp_dir, recursive = TRUE)
})

# Mock function for remotes::install_local
mock_install_local <- function(pkg_source_path, dependencies, upgrade, force, quiet, INSTALL_opts) {
  message(paste0(basename(pkg_source_path), " installed locally"))
}

test_that("install_package_local installs package successfully", {
  # Mock the remotes::install_local function
  mockery::stub(install_package_local, "remotes::install_local", mock_install_local)
  
  # Create a temporary directory to simulate the package source path
  temp_dir <- tempdir()
  
  if (!dir.exists(temp_dir)) {
    dir.create(temp_dir, recursive = TRUE)  # Ensure parent directory exists
  }
  if (!dir.exists(file.path(temp_dir, "mock_package"))) {
    dir.create(file.path(temp_dir, "mock_package"))
  }
  
  # Test the function
  result <- install_package_local(file.path(temp_dir, "mock_package"))
  expect_true(result)
})

test_that("install_package_local handles non-existent directory", {
  # Test the function with a non-existent directory
  result <- install_package_local("non_existent_directory")
  expect_false(result)
})

test_that("install_package_local handles already installed package", {
  # Mock the requireNamespace function to simulate an already installed package
  mockery::stub(install_package_local, "requireNamespace", TRUE)
  
  # Create a temporary directory to simulate the package source path
  temp_dir <- tempdir()
  if (!dir.exists(file.path(temp_dir, "mock_package"))) {
    dir.create(file.path(temp_dir, "mock_package"))
  }
  
  # Test the function
  result <- install_package_local(file.path(temp_dir, "mock_package"))
  expect_true(result)
})

#' Run Coverage Analysis on Package Source and Non‑Standard Tests (NSTF)
#'
#' This function maps package source files to non‑standard test scripts, detects
#' failing or skipped tests, executes the remaining tests manually, computes code
#' coverage, and returns structured results. It is intended for packages that do
#' not use the standard `testthat` framework (e.g., base R tests under `tests/`,
#' Bioconductor‑style tests in `inst/tests/`, or other legacy/custom scripts).
#'
#' @param pkg_source_path Character. Path to the root of the package source
#'   directory.
#' @param test_pkg_data List. Output from \code{check_pkg_tests_and_snaps()}
#'   indicating the detected testing infrastructure (e.g., base tests,
#'   Bioconductor tests) and the presence of snapshots or related artifacts.
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run.
#' @param framework Optional character. When provided (e.g. from
#'   \code{run_one_framework}), restricts test paths to the framework-specific
#'   directory. Use when a package has multiple frameworks (e.g. testthat + testit)
#'   to avoid mixing paths across frameworks.
#'   
#' @details
#' This NSTF workflow does not rely on `testthat` reporters. Instead, it:
#' \itemize{
#'   \item maps source files to available NSTF test scripts,
#'   \item identifies failing or problematic tests and derives a skip map,
#'   \item runs the remaining test scripts, and
#'   \item computes coverage on the successfully executed regions.
#' }
#' The output mirrors the structure returned by the STF skip‑aware workflow so
#' that downstream reporting remains consistent across frameworks.
#'
#' @return
#' A named list containing coverage results and diagnostics:
#' \describe{
#'   \item{total_cov}{Numeric. Overall percentage of lines covered across the package.}
#'   \item{res_cov}{List. File‑ or element‑level coverage results (as derived from
#'         covr output), including per‑file coverage, any errors, and diagnostic notes.}
#'   \item{long_summary}{Long‑format summary of test outcomes (e.g., PASS/FAIL/ERROR/SKIPPED),
#'         when available from the NSTF execution.}
#'   \item{test_skip}{A data frame describing lines or regions skipped during coverage
#'         (if any). \code{NULL} if no skipping was required.}
#'   \item{untested_sources}{Character vector of source files for which no tests were found.}
#' }
#' If no valid NSTF test directory is detected, an “empty” coverage object may be
#' returned via \code{create_empty_covr_list()}.
#'
#'
#' @family nstf_utility
#' @rdname run_covr_skip_nstf
#' @keywords internal
run_covr_skip_nstf <- function(pkg_source_path, 
                               test_pkg_data,
                               cov_env,
                               framework = NULL
                               ) {
  
  # --- Input checking (CRAN-safe and explicit) ---
  checkmate::assert_string(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_directory_exists(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_true(!missing(cov_env), .var.name = "cov_env")
  checkmate::assert_environment(cov_env, .var.name = "cov_env")
  
  pkg_name <- get_pkg_name(pkg_source_path)
  
  # set up path to test directory 
  path    <- as.character(pkg_source_path)
  testdir <- file.path(path, "tests")
  
  # Get a *list* of candidate paths
  test_path_list <- get_nstf_test_path(test_pkg_data, testdir)
  
  # restrict paths 
  test_path_list <- restrict_test_paths_to_framework(test_path_list, framework)
  
  # Normalize *all* candidates and keep names (e.g., "testthat", "testit", etc.)
  test_path <- normalizePath(
    unlist(test_path_list, use.names = TRUE),
    winslash = "/",
    mustWork = FALSE
  )
  
  message(paste0("performing source test mapping for ", pkg_name))
  mapping <- tryCatch({
    get_source_test_mapping_nstf(pkg_source_path, test_path)
  }, error = function(e) {
    cleanup_and_return_null(
      paste0("Error in source-test mapping for ", pkg_name, " : ", e$message), 
      env = cov_env
      )
  })
  if (is.null(mapping)) return(NULL)
  
  message(paste0("identifying problem tests for ", pkg_name))
  
  # check for testit testing framework (unchanged)
  if (isTRUE(test_pkg_data$has_testit)) {
    problems <- tryCatch({
      check_covr_skip_testit(pkg_name, mapping, test_path)
    }, error = function(e) {
      cleanup_and_return_null(
        paste0("Error identifying skipped/problematic tests for ", pkg_name, " : ", e$message), 
        env = cov_env)
    })
  } else {
    problems <- tryCatch({
      check_covr_skip_nstf(pkg_source_path, mapping)
    }, error = function(e) {
      cleanup_and_return_null(
        paste0("Error identifying skipped/problematic tests for ", pkg_name, " : ", e$message),
        env = cov_env)
    })
  }
  if (is.null(problems)) return(NULL)
  
  # check for tests to skip
  skip_tests <- problems$test_file[
    !is.na(problems$test_file) &
      !startsWith(problems$issue_type, "warning:") &
      problems$issue_type != "No tests skipped"
  ]
  
  # create df with source files with no tests
  no_tests_df <- get_function_no_tests(mapping) 
  
  # remove source files with no tests
  clean_mapping <- if (anyNA(mapping)) tidyr::drop_na(mapping) else mapping
   
  # ---- build test files list per framework ----
  if (isTRUE(test_pkg_data$has_testit)) {
    test_files_clean <- file.path(pkg_source_path, "tests", "testit", clean_mapping$test_file)
    if (!is.null(skip_tests) | (length(skip_tests) > 0)) {
      test_files_clean <- setdiff(test_files_clean, file.path(pkg_source_path, "tests", "testit", skip_tests))
    }
  } else if (isTRUE(test_pkg_data$has_tinytest)) {
    # tinytest lives in inst/tinytest
    test_files_clean <- file.path(pkg_source_path, "inst", "tinytest", clean_mapping$test_file)
    if (!is.null(skip_tests) | (length(skip_tests) > 0)) {
      test_files_clean <- setdiff(test_files_clean, file.path(pkg_source_path, "inst", "tinytest", skip_tests))
    }
    # do not include any tinytest runners
    tinytest_runner_basenames <- c("tinytest.R", "runTinyTests.R", "run_tinytest.R")
    test_files_clean <- test_files_clean[!(basename(test_files_clean) %in% tinytest_runner_basenames)]
    if (!is.null(test_pkg_data$tinytest_runner_paths)) {
      runner_paths_norm <- normalizePath(
        test_pkg_data$tinytest_runner_paths,
        winslash = "/",
        mustWork = FALSE
      )
      test_files_clean <- setdiff(test_files_clean, runner_paths_norm)
    }
    # keep only files that actually exist under inst/tinytest
    test_files_clean <- test_files_clean[file.exists(test_files_clean)]
  } else {
    # base / MASS / testthat
    test_files_clean <- file.path(pkg_source_path, "tests", clean_mapping$test_file)
    if (!is.null(skip_tests) | (length(skip_tests) > 0)) {
      test_files_clean <- setdiff(test_files_clean, file.path(pkg_source_path, "tests", skip_tests))
    }
  }
   
  # --------------------------------------------
  
  message(paste0("loading package and datasets for ", pkg_name))
  tryCatch({
    pkgload::load_all(pkg_source_path)
    
    pkg_all <- ls(envir = asNamespace(pkg_name), all.names = TRUE)
    
    for (obj in pkg_all) {
      tryCatch({
        assign(obj, get(obj, envir = asNamespace(pkg_name)), envir = cov_env)
      }, error = function(e) {
        message(paste0("Failed to assign object ", obj, " : ", e$message))
      })
    }
    
    # Load dependencies
    message(paste0("Loading dependencies into coverage environment for: ", pkg_name))
    load_dependencies_into_env(pkg_source_path, cov_env)
    
    message("Loading Base R packages into coverage environment")
    base_r_loaded <- load_base_r_packages_into_env(
      env = cov_env,
      overwrite = FALSE
    )
    
    datasets <- load_package_datasets(pkg_name, env = cov_env)
    message("Datasets loaded: ", paste(datasets, collapse = ", "))
  }, error = function(e) {
    cleanup_and_return_null(
      paste0("Error loading package or datasets for ", pkg_name, " : ", e$message),
      env = cov_env)
  })
  
  # ---- compute coverage: tinytest vs. base/MASS vs. fallback ----
  if (isTRUE(test_pkg_data$has_tinytest)) {
    message(sprintf("running tinytest coverage for %s", pkg_name))
    tiny_cov <- tryCatch({
      create_tinytest_coverage(
        pkg_source_path = pkg_source_path,
        pkg_name        = pkg_name,
        tiny_dir        = file.path(pkg_source_path, "inst", "tinytest"),
        at_home         = TRUE,
        color           = FALSE
      )
    }, error = function(e) {
      cleanup_and_return_null(
        paste0("Error running tinytest coverage for ", pkg_name, " : ", e$message),
        env = cov_env
      )
    })
    if (is.null(tiny_cov)) return(NULL)
    coverage <- tiny_cov$coverage
    
  } else {
    # Detect a base/MASS style layout: tests/ present and no tinytest/testit declared
    has_testthat_dir <- dir.exists(file.path(pkg_source_path, "tests", "testthat"))
    is_base_layout   <- !isTRUE(test_pkg_data$has_testit) &&
      !isTRUE(test_pkg_data$has_tinytest) &&
      !has_testthat_dir
    
    if (is_base_layout) {
      message(sprintf("running base-style tests coverage for %s", pkg_name))
      base_cov <- tryCatch({
        # Ensure absolute paths for sys.source() calls
        abs_tests <- normalizePath(test_files_clean, winslash = "/", mustWork = FALSE)
        create_base_tests_coverage(pkg_source_path, pkg_name, abs_tests)
      }, error = function(e) {
        cleanup_and_return_null(
          paste0("Error running base-style tests for ", pkg_name, " : ", e$message),
          env = cov_env
        )
      })
      if (is.null(base_cov)) return(NULL)
      coverage <- base_cov$coverage
    } else {
      # Fallback: environment_coverage for other layouts (e.g., testthat)
      message(paste0("running coverage manually for ", pkg_name))
      coverage <- tryCatch({
        covr::environment_coverage(
          env        = cov_env,
          test_files = test_files_clean
        )
      }, error = function(e) {
        cleanup_and_return_null(
          paste0("Error running coverage for ", pkg_name, " : ", e$message),
          env = cov_env
        )
      })
    }
  }
  
  # -----------------------------------------------------------
  
  if (!is.null(coverage)) {
    
    covr_list <- create_nstf_covr_list(coverage, pkg_name, 
                                       no_tests_df, test_files_clean,
                                       skip_tests, problems)
    
  } else {
    message("No env coverage generated. Falling back to covr::package_coverage(type = 'tests').")
    
    covr <- tryCatch({
      
        withr::with_dir(
          pkg_source_path,
          covr::package_coverage(
            path       = pkg_source_path,
            type       = "tests",
            quiet      = FALSE
          )
        )
        
      
    }, error = function(e) {
      
      message("package_coverage() failed: ", conditionMessage(e))
      NULL
      
    })
    
    
    covr_list <- create_nstf_covr_list(covr, pkg_name, 
                                       no_tests_df, test_files_clean,
                                       skip_tests, problems)
  }  
    
  if (is.null(covr_list$res_cov)) {
    message("No test coverage for this configuration")
    covr_list <- list(
      total_cov = 0,
      res_cov = list(
        name = pkg_name,
        coverage = list(
          filecoverage   = matrix(0, nrow = 1, dimnames = list("No functions tested")),
          totalcoverage  = 0
        ),
        errors = "No testthat or testit configuration",
        notes  = NA
      )
    )
  }
  
  # Final cleanup (env bindings + temp workspace under tempdir())
  cleanup_and_return_null(env = cov_env)
  
  return(covr_list)
}


#' Check for skipped or errored test files (non-testit)
#'
#' This internal function attempts to run test files mapped to source files and
#' identifies any that produce warnings or errors. It does not handle `testit`.
#'
#' @param pkg_source_path Path to the root of the package source.
#' @param mapping A data frame with a column `test_file` listing test file names.
#' @param base_dir Directory containing the test files to probe. Defaults to
#'   `<pkg>/tests`, but may also be `inst/tinytest` or `tests/testthat`
#'   depending on package layout.
#'
#' @return A data frame listing test files and any issues encountered.
#' @keywords internal
#' @importFrom utils capture.output
check_covr_skip_nstf <- function(pkg_source_path, mapping,
                                 base_dir = file.path(pkg_source_path, "tests")) {
   
  test_dir <- base_dir
  pkg_name <- get_pkg_name(pkg_source_path)
  
  clean_mapping <- if (anyNA(mapping)) tidyr::drop_na(mapping) else mapping
  
  # Detect test layout from base_dir
  norm_base   <- normalizePath(test_dir, winslash = "/", mustWork = FALSE)
  is_tinytest <- grepl("(?:/|\\\\)inst(?:/|\\\\)tinytest$",  norm_base, ignore.case = TRUE)
  is_testthat <- grepl("(?:/|\\\\)tests(?:/|\\\\)testthat$", norm_base, ignore.case = TRUE)
  # Base/MASS style: tests/ (without testthat subdir)
  is_base     <- grepl("(?:/|\\\\)tests$", norm_base, ignore.case = TRUE) && !is_testthat
  
  # Parent env for non-tinytest probing
  parent_env <- if (is_testthat) {
    asNamespace("testthat")
  } else if (is_base) {
    # set up asNamespace to load env properly
    # Use the package namespace so unqualified calls hit the loaded source build
    asNamespace(pkg_name)
  } else {
      baseenv()
  }
  
  problem_list <- lapply(clean_mapping$test_file, function(test) {
    # Absolute path so the file is located before chdir takes effect
    
    abs_test_path <- normalizePath(file.path(test_dir, test), winslash = "/", mustWork = FALSE)
    
    if (!file.exists(abs_test_path)) {
      return(data.frame(
        test_file  = test,
        issue_type = paste0("error: file not found: ", abs_test_path),
        stringsAsFactors = FALSE
      ))
    }
    
    res <- tryCatch(
      {
        if (is_tinytest) {
          # Run with tinytest runner (captures expectations as data; runs from file dir)
          tt <- tinytest::run_test_file(abs_test_path, at_home = TRUE, color = FALSE)
          if (tinytest::any_fail(tt)) "failed expectation" else NULL
        } else {
          # Probe by sourcing from the file's directory so relative paths behave as expected
          suppressMessages(suppressWarnings(
            capture.output(
              sys.source(abs_test_path, envir = new.env(parent = parent_env), chdir = TRUE),
              file = NULL
            )
          ))
          NULL
        }
      },
      warning = function(w) paste0("warning: ", conditionMessage(w)),
      error   = function(e) paste0("error: ",   conditionMessage(e))
    )
    
    if (!is.null(res)) {
      data.frame(test_file = test, issue_type = res, stringsAsFactors = FALSE)
    } else {
      NULL
    }
  })
  
  problems <- do.call(rbind, problem_list)
  
  if (is.null(problems) || nrow(problems) == 0) {
    problems <- data.frame(
      test_file  = NA_character_,
      issue_type = "No tests skipped",
      stringsAsFactors = FALSE
    )
  }
  problems
}

#' Run testit tests individually, capturing all warnings and errors, with dynamic package root detection
#'
#' @param pkg_name The package name (not used in this function, but kept for compatibility)
#' @param mapping A data frame with a column `test_file` listing test file names
#' @param test_path The path to the test directory (e.g., tests/testit)
#'
#' @return A data frame with test_file and issue_type columns
#' @keywords internal
#' @importFrom utils capture.output
#'
#' @family nstf_utility
check_covr_skip_testit <- function(pkg_name, mapping, test_path) {
  
  # Helper to find the package root by searching for DESCRIPTION upward
  find_pkg_root <- function(start_path) {
    current <- normalizePath(start_path, mustWork = TRUE)
    while (!file.exists(file.path(current, "DESCRIPTION"))) {
      parent <- dirname(current)
      if (parent == current) stop("Package root (with DESCRIPTION) not found.")
      current <- parent
    }
    current
  }
  
  pkg_root <- find_pkg_root(test_path)
  
  # Prepare mapping
  clean_mapping <- if (anyNA(mapping)) tidyr::drop_na(mapping) else mapping
  
  problems <- data.frame(test_file = character(), issue_type = character(), stringsAsFactors = FALSE)
  
  problem_list <- lapply(clean_mapping$test_file, function(test) {
    test_file_path <- file.path(test_path, test)
    # Set working directory to package root for each test
    original_wd <- getwd()
    setwd(pkg_root)
    on.exit(setwd(original_wd), add = TRUE)
    
    result <- tryCatch(
      {
        capture.output(
          source(test_file_path, local = new.env()),
          file = NULL
        )
        NULL
      },
      warning = function(w) paste("warning:", conditionMessage(w)),
      error = function(e) paste("error:", conditionMessage(e))
    )
    
    if (!is.null(result)) {
      data.frame(test_file = test, issue_type = result, stringsAsFactors = FALSE)
    } else {
      NULL
    }
  })
  
  problems <- do.call(rbind, problem_list)
  
  if (is.null(problems) || nrow(problems) == 0) {
    problems <- data.frame(
      test_file = NA_character_,
      issue_type = "No tests skipped",
      stringsAsFactors = FALSE
    )
  }
  
  return(problems)
}


#' Map R Source Files to Corresponding Test Files
#'
#' This internal helper function scans the `R/` and `tests/` directories of a package
#' and attempts to match each source file with a corresponding test file based on filename similarity.
#'
#' @param pkg_source_path Character string. Path to the root of the package source directory.
#' @param test_dir character string. directory where tests are
#'
#' @return A data frame with two columns:
#' \describe{
#'   \item{source_file}{Name of the source file in the `R/` directory.}
#'   \item{test_file}{Name of the matched test file in the `tests/` directory, or \code{NA} if no match is found.}
#' }
#'
#' @keywords internal
#'
#' @family nstf_utility
get_source_test_mapping_nstf <- function(pkg_source_path, test_dir) {
  source_dir <- file.path(pkg_source_path, "R")
  # test_dir <- file.path(pkg_source_path, "tests")
  
  src_files <- list.files(source_dir, pattern = "\\.R$", full.names = FALSE)
  test_files <- list.files(test_dir, pattern = "\\.R$", full.names = FALSE)
  
  mapping <- do.call(rbind, lapply(tools::file_path_sans_ext(src_files), function(src) {
    match <- test_files[grepl(src, tools::file_path_sans_ext(test_files), ignore.case = TRUE)]
    data.frame(
      source_file = paste0(src, ".R"),
      test_file = ifelse(length(match) == 0, NA, match[1]),
      stringsAsFactors = FALSE
    )
  }))
  
  return(mapping)
}


#' Collect all applicable nonstandard test paths for a package (internal)
#'
#' Detects and returns all relevant testing directories used by an R package,
#' combining framework-specific and filesystem-based checks. Supported layouts:
#' \itemize{
#'   \item \strong{testthat}: \code{tests/testthat/}
#'   \item \strong{tinytest}: \code{inst/tinytest/} (primary location).
#'         Some packages also include a small runner under \code{tests/}
#'         (e.g., \code{tests/tinytest.R}) which this function does not return
#'         as a path, but it will still detect \code{inst/tinytest/}.
#'   \item \strong{testit} (standard): \code{tests/testit/}
#'   \item \strong{testit} (nonstandard): both \code{tests/test-ci/} and
#'         \code{tests/test-cran/} present, using \code{tests/} as the root
#'   \item \strong{base R tests}: \code{*.R} files directly under \code{tests/}
#'   \item \strong{Bioconductor unit tests}: \code{inst/unitTests/}
#' }
#'
#' The function returns a named list of paths for all detected testing setups.
#' It uses both the flags contained in \code{test_pkg_data} and direct filesystem
#' checks for robustness. If no paths are detected, the returned list contains
#' a single entry \code{default_tests_root = tests/}.
#'
#' @param test_pkg_data \code{list}. A list of booleans/fields describing detected
#'   test frameworks and layouts. All fields are optional and default to \code{FALSE}
#'   or \code{NULL} if missing:
#'   \itemize{
#'     \item \code{has_testthat}: logical; whether \code{tests/testthat/} is present.
#'     \item \code{has_tinytest} (or \code{has_tiny_test}): logical; whether
#'           tinytest is present. If true, the function will look for
#'           \code{inst/tinytest/} (or use \code{tinytest_dir} if provided).
#'     \item \code{tinytest_dir}: character; optional explicit path to the tinytest
#'           directory. May be absolute or relative to the package root; when
#'           missing, defaults to \code{inst/tinytest/}.
#'     \item \code{has_testit}: logical; whether \code{tests/testit/} or the
#'           nonstandard \code{tests/test-ci/ + tests/test-cran/} layout is present.
#'     \item \code{has_tests_base}: logical; whether \code{*.R} test files exist
#'           directly under \code{tests/}.
#'     \item \code{has_bioc_unit_tests}: logical; whether \code{inst/unitTests/}
#'           exists (Bioconductor layout).
#'   }
#'
#' @param testdir \code{character}. Path to the package \code{tests/} directory. The function
#'   also derives the package root as \code{dirname(testdir)} to locate paths such as
#'   \code{inst/unitTests/} and \code{inst/tinytest/}.
#'
#' @return \code{list}. A named list of character paths to all applicable test directories found.
#'   Possible names include:
#'   \itemize{
#'     \item \code{tinytest} — \code{inst/tinytest/}
#'     \item \code{testthat} — \code{tests/testthat/}
#'     \item \code{testit} — \code{tests/testit/}
#'     \item \code{nonstandard_testit_root} — \code{tests/} (when both \code{test-ci} and \code{test-cran} exist)
#'     \item \code{base_tests_root} — \code{tests/} (when base \code{*.R} files exist in \code{tests/} root)
#'     \item \code{bioc_unitTests} — \code{inst/unitTests/}
#'     \item \code{default_tests_root} — \code{tests/} (returned only if no other paths are detected)
#'   }
#'
#' @details
#' Identical paths arising from multiple detection routes (e.g., \code{tests/} playing
#' multiple roles) are de-duplicated while preserving the first descriptive label.
#' If \code{tinytest_dir} is provided and relative, it is resolved against the
#' package root (i.e., \code{dirname(testdir)}).
#'
#' @keywords internal
#'
#' @family nstf_utility
get_nstf_test_path <- function(test_pkg_data, testdir) {
  
  # Normalize key dirs
  testdir  <- normalizePath(testdir, winslash = "/", mustWork = FALSE)
  pkg_root <- normalizePath(file.path(testdir, ".."), winslash = "/", mustWork = FALSE)
  
  # Helper: absolute-path check (POSIX and Windows)
  is_absolute_path <- function(p) {
    is.character(p) && length(p) == 1L && nzchar(p) &&
      grepl("^(/|[A-Za-z]:[/\\])", p)
  }
  
  # Resolve tinytest directory from test_pkg_data or default to inst/tinytest
  resolve_tinytest_dir <- function() {
    # prefer explicit tinytest_dir if provided
    td <- test_pkg_data$tinytest_dir
    if (!is.null(td) &&
        !is.na(td) &&
        is.character(td) &&
        length(td) == 1L &&
        nzchar(td)) {
      if (!is_absolute_path(td)) td <- file.path(pkg_root, td)
      return(normalizePath(td, winslash = "/", mustWork = FALSE))
    }
    # default: <pkg_root>/inst/tinytest
    normalizePath(file.path(pkg_root, "inst", "tinytest"), winslash = "/", mustWork = FALSE)
  }
  
  # Candidate paths
  paths <- list(
    testthat       = file.path(testdir, "testthat"),
    testit         = file.path(testdir, "testit"),
    test_ci        = file.path(testdir, "test-ci"),
    test_cran      = file.path(testdir, "test-cran"),
    tests_root     = testdir,
    bioc_unitTests = file.path(pkg_root, "inst", "unitTests"),
    tinytest       = resolve_tinytest_dir()
  )
  
  # Helper: detect base R scripts directly in tests/ root
  has_base_R_in_tests_root <- function(tdir) {
    if (!dir.exists(tdir)) return(FALSE)
    files <- list.files(tdir, pattern = "\\.R$", full.names = TRUE, recursive = FALSE)
    length(files) > 0
  }
  
  # Build the output list of active test paths
  test_path_list <- list()
  
  # --- tinytest (new) ---
  # accept both has_tinytest and has_tiny_test for robustness
  if (isTRUE(test_pkg_data$has_tinytest) || isTRUE(test_pkg_data$has_tiny_test) || dir.exists(paths$tinytest)) {
    test_path_list$tinytest <- paths$tinytest
  }
  
  # testthat
  if (isTRUE(test_pkg_data$has_testthat) || dir.exists(paths$testthat)) {
    test_path_list$testthat <- paths$testthat
  }
  
  # testit (standard)
  if (isTRUE(test_pkg_data$has_testit) || dir.exists(paths$testit)) {
    test_path_list$testit <- paths$testit
  }
  
  # nonstandard testit pattern: both test-ci and test-cran present => use tests/ as the root
  if ((dir.exists(paths$test_ci) && dir.exists(paths$test_cran)) ||
      isTRUE(test_pkg_data$has_testit)) {
    test_path_list$nonstandard_testit_root <- paths$tests_root
  }
  
  # base R tests in tests/ root (when *.R directly under tests/)
  if (!(isTRUE(test_pkg_data$has_tinytest) || dir.exists(paths$tinytest))) {
    if (isTRUE(test_pkg_data$has_tests_base) || has_base_R_in_tests_root(paths$tests_root)) {
      test_path_list$base_tests_root <- paths$tests_root
    }
  }
  
  # Bioconductor unit tests (inst/unitTests)
  # Note: caller may pass test_pkg_data$has_bioc_unit_tests (your previous function used has_bioc_unit_tests)
  if (isTRUE(test_pkg_data$has_bioc_unit_tests) || dir.exists(paths$bioc_unitTests)) {
    test_path_list$bioc_unitTests <- paths$bioc_unitTests
  }
  
  # De-duplicate identical paths (e.g., tests/ might appear for multiple roles)
  if (length(test_path_list) > 0) {
    uniq_paths <- !duplicated(unname(unlist(test_path_list)))
    test_path_list <- test_path_list[uniq_paths]
  }
  
  # If nothing detected, return the provided testdir as a conservative default
  if (length(test_path_list) == 0) {
    test_path_list$default_tests_root <- paths$tests_root
  }
  
  return(test_path_list)
}


#' Create coverage from tinytest under covr instrumentation (internal)
#'
#' Instruments the package **namespace** with \pkg{covr}, runs all
#' \pkg{tinytest} files in \code{inst/tinytest}, and returns a covr coverage
#' object together with the tinytest results. This mirrors tinytest's execution
#' model (expectation helpers available; tests run from each file's directory)
#' while ensuring calls into the package namespace are traced.
#'
#' @details
#' Intended for **source-tree** coverage of tinytest suites:
#' \itemize{
#'   \item The package namespace is traced via \code{covr::trace_environment()}
#'         so that calls made by tests to package functions are recorded
#'         for coverage.
#'   \item \code{NOT_CRAN} is set to \code{"true"} during the run (restored on exit)
#'         so tinytest \code{at_home()} logic treats the run as local.
#'   \item Test files are driven by \code{tinytest::run_test_dir()} from
#'         \code{file.path(pkg_source_path, "inst", "tinytest")}.
#'   \item Coverage is built from covr's counters using \code{covr::as_coverage()}.
#' }
#'
#' \strong{Preconditions}
#' \itemize{
#'   \item The package under test should already be \emph{loaded} (e.g.,
#'         \code{pkgload::load_all(pkg_source_path)}), so that
#'         \code{asNamespace(pkg_name)} exists and reflects the current source.
#'   \item The directory \code{inst/tinytest} must exist and contain test files.
#' }
#'
#' @param pkg_source_path Character scalar. Path to the package source root.
#' @param pkg_name        Character scalar. Package name (as in \code{DESCRIPTION}).
#' @param tiny_dir        Optional override for the tinytest directory. Defaults
#'                        to \code{file.path(pkg_source_path, "inst", "tinytest")}.
#' @param at_home         Logical; passed to \code{tinytest::run_test_dir()}.
#'                        Defaults to \code{TRUE}.
#' @param color           Logical; color output from tinytest. Defaults to \code{FALSE}.
#'
#' @return A list with components:
#' \describe{
#'   \item{coverage}{An object of class \code{"coverage"} (from \pkg{covr}).}
#'   \item{tinytests}{The \code{tinytests} object returned by tinytest.}
#' }
#'
#' @keywords internal
#'
#' @family nstf_utility
create_tinytest_coverage <- function(pkg_source_path,
                                     pkg_name,
                                     tiny_dir = file.path(pkg_source_path, "inst", "tinytest"),
                                     at_home = TRUE,
                                     color = FALSE) {
  
  # Basic checks
  if (!dir.exists(pkg_source_path)) {
    stop("Package source path does not exist: ", pkg_source_path, call. = FALSE)
  }
  if (!dir.exists(tiny_dir)) {
    stop("tinytest directory not found: ", tiny_dir, call. = FALSE)
  }
  if (!pkg_name %in% loadedNamespaces()) {
    stop("Namespace '", pkg_name, "' is not loaded. ",
         "Call pkgload::load_all(pkg_source_path) before running this helper.",
         call. = FALSE)
  }
  
  message(sprintf("running tinytest coverage for %s", pkg_name))
  
  
  # Resolve covr internals in a version-agnostic way (exported or not)
  covr_ns <- asNamespace("covr")
  trace_env     <- get("trace_environment",  envir = covr_ns)  # exported in newer covr
  reset_traces  <- get("reset_traces",       envir = covr_ns)
  clear_counters<- get("clear_counters",     envir = covr_ns)
  as_coverage   <- get("as_coverage",        envir = covr_ns)
  
  # 1) Trace the *namespace* so tinytest calls are instrumented
  ns_env <- asNamespace(pkg_name)
  trace_env(ns_env)  # covr::trace_environment(ns_env) when exported
  
  # Always clean up covr traces/counters
  on.exit({
    reset_traces()
    clear_counters()
  }, add = TRUE)
  
  # 2) Ensure "local run" semantics (restored on exit)
  old_notcran <- Sys.getenv("NOT_CRAN", unset = NA)
  on.exit({
    if (is.na(old_notcran)) Sys.unsetenv("NOT_CRAN") else Sys.setenv(NOT_CRAN = old_notcran)
  }, add = TRUE)
  Sys.setenv(NOT_CRAN = "true")
  
  # 3) Run tinytest from canonical directory
  tt <- tinytest::run_test_dir(dir = tiny_dir, at_home = at_home, color = color)
  
  # 4) Convert covr’s recorded counters into a coverage object
  counters <- get(".counters", envir = covr_ns)
  coverage <- as_coverage(counters)
  
  list(coverage = coverage, tinytests = tt)
}


#' Create coverage by running base-style tests under covr (internal)
#'
#' Traces the package namespace with covr, then sources each test script
#' from its own directory (so relative file references like 'load("solr.rda")'
#' in MASS/tests work). Converts covr counters to a coverage object.
#'
#' @param pkg_source_path Character; package source root.
#' @param pkg_name        Character; package name.
#' @param test_files      Character vector of test file paths (absolute or relative).
#' @return list(coverage = <covr coverage object>)
#' @keywords internal
#'
#' @family nstf_utility
create_base_tests_coverage <- function(pkg_source_path, pkg_name, test_files) {
  
  if (!pkg_name %in% loadedNamespaces()) {
    stop("Namespace '", pkg_name, "' is not loaded. ",
         "Call pkgload::load_all(pkg_source_path) first.", call. = FALSE)
  }
  
  covr_ns        <- asNamespace("covr")
  trace_env      <- get("trace_environment",  envir = covr_ns)
  reset_traces   <- get("reset_traces",       envir = covr_ns)
  clear_counters <- get("clear_counters",     envir = covr_ns)
  as_coverage    <- get("as_coverage",        envir = covr_ns)
  
  # trace the *package namespace* so calls from tests are instrumented
  trace_env(asNamespace(pkg_name))
  on.exit({ reset_traces(); clear_counters() }, add = TRUE)
  
  # source each test from its own directory so relative files (e.g., 'solr.rda') are found
  # MASS has tests/glmmPQL.R which does load("solr.rda") from tests/. 
  
  for (tf in test_files) {
    tf_abs <- normalizePath(tf, winslash = "/", mustWork = TRUE)
    try(
      sys.source(tf_abs, envir = new.env(parent = asNamespace(pkg_name)), chdir = TRUE),
      silent = TRUE
    )
  }
  
  coverage <- as_coverage(get(".counters", envir = covr_ns))
  list(coverage = coverage)
}

#' Restrict a list of test paths to a single framework
#'
#' Filters a named list of candidate test directories (typically produced by
#' \code{get_nstf_test_path()}) down to the entry that corresponds to a requested
#' framework. If no usable mapping can be made, the input list is returned
#' unchanged.
#'
#' The function maps \code{framework} to a primary key in \code{test_path_list}
#' as follows:
#' \itemize{
#'   \item \code{"testthat"}   → \code{"testthat"}
#'   \item \code{"testit"}     → \code{"testit"} (falls back to
#'         \code{"nonstandard_testit_root"} if \code{"testit"} is \code{NULL})
#'   \item \code{"tests_base"} → \code{"base_tests_root"}
#'   \item \code{"tinytest"}   → \code{"tinytest"}
#' }
#'
#' @param test_path_list \code{list}. Named list of candidate test paths. Known
#'   names include \code{"testthat"}, \code{"testit"}, \code{"tinytest"},
#'   \code{"base_tests_root"}, and optionally \code{"nonstandard_testit_root"}.
#' @param framework \code{character(1)} or \code{NULL}. Framework identifier
#'   (e.g., \code{"testthat"}, \code{"testit"}, \code{"tests_base"},
#'   \code{"tinytest"}). If \code{NULL} or empty (\code{""}), the list is
#'   returned unchanged.
#'
#' @return \code{list}. Either:
#' \itemize{
#'   \item a single-element named list containing only the resolved framework
#'         path (name is the selected key), or
#'   \item the original \code{test_path_list} if \code{framework} is
#'         \code{NULL}/empty, maps to an unknown key, or the resolved key is
#'         absent in \code{test_path_list}.
#' }
#'
#' @details
#' For \code{framework == "testit"}, if \code{test_path_list[["testit"]]} is
#' \code{NULL} and \code{test_path_list[["nonstandard_testit_root"]]} is present,
#' the latter is used instead. No directories are created or validated here; the
#' function only selects from the provided list.
#'
#' @seealso \code{\link{get_nstf_test_path}}
#' @family nstf_utility
#' @keywords internal
restrict_test_paths_to_framework <- function(test_path_list, framework) {
  out <- test_path_list
  if (!is.null(framework) && nzchar(framework)) {
    primary_key <- switch(
      framework,
      testit     = "testit",
      tests_base = "base_tests_root",
      tinytest   = "tinytest",
      testthat   = "testthat",
      NULL
    )
    if (!is.null(primary_key)) {
      fw_key <- primary_key
      if (identical(framework, "testit") &&
          is.null(test_path_list[["testit"]]) &&
          !is.null(test_path_list[["nonstandard_testit_root"]])) {
        fw_key <- "nonstandard_testit_root"
      }
      if (!is.null(test_path_list[[fw_key]])) {
        out <- setNames(list(test_path_list[[fw_key]]), fw_key)
      }
    }
  }
  return(out)
}
  
#' Create the NSTF Coverage List
#'
#' Internal helper that constructs the final `covr_list` structure from a
#' coverage object. It wraps extraction of coverage information, performs
#' total coverage computation, and assembles all diagnostic components.
#'
#' @param coverage A coverage object returned from covr.
#' @param pkg_name A Character string - package name.
#' @param no_tests_df A dataframe - functions with no tests.
#' @param test_files_clean A Character vector - functions with working tests
#' @param skip_tests A Character vector - functions skipped
#' @param problems A Character vector - test problems
#' 
#' @return A named list (`covr_list`) containing:
#'   \describe{
#'     \item{total_cov}{Numeric; overall package coverage percentage.}
#'     \item{res_cov}{List; raw coverage extraction results with coverage data, errors, notes.}
#'     \item{functions_no_tests}{Data frame of functions lacking tests.}
#'     \item{tests_skipped}{Character vector of skipped test files.}
#'     \item{tests_passing}{Character vector of passing test files.}
#'     \item{tests_problems}{Character vector describing test problems.}
#'   }
#'   
#' @family nstf_utility
#' @keywords internal
create_nstf_covr_list <- function(coverage, pkg_name,
                                  no_tests_df, 
                                  test_files_clean,
                                  skip_tests,
                                  problems
                                  ) {
 
  # safely extract coverage to list
  res_cov <- tryCatch({
    coverage_list <- covr::coverage_to_list(coverage)
    list(
      name     = coverage$package,
      coverage = coverage_list,
      errors   = NA,
      notes    = NA
    )
  },
  error = function(cond) {
    coverage_list <- list(filecoverage = NA, totalcoverage = NA_integer_)
    list(
      name     = coverage$package,
      coverage = coverage_list,
      errors   = cond,
      notes    = NA
    )
  })
  
  # extract percent coverage
  percent_cov <- as.numeric(res_cov$coverage$totalcoverage)
  
  # inputs expected by compute_total_coverage()
  functions_no_tests_df <- no_tests_df
  tests_passing         <- test_files_clean
  tests_skipped         <- skip_tests
  problems              <- problems
  
  # compute combined coverage information
  calc_covr_list <- compute_total_coverage(
    percent_cov,
    functions_no_tests_df,
    tests_passing,
    tests_skipped
  )
  
  total_cov <- calc_covr_list$total_cov
  percent_cov_round <- round(total_cov, 2)
  message(paste0("Total Coverage for ", pkg_name, " is: ", percent_cov_round, "%"))
    
  percent_cov         <- calc_covr_list$percent_cov
  n_testfiles         <- calc_covr_list$n_testfiles
  n_no_function_tests <- calc_covr_list$n_no_function_tests
  n_skipped_files     <- calc_covr_list$n_skipped_files  
    
  message(sprintf(
    "Approx total coverage: (%.2f%%), counts: tested_files=%d, untested_functions=%d, skipped_tests=%d",
    percent_cov, n_testfiles, n_no_function_tests, n_skipped_files
  ))
  
  # structure the return list
  covr_list <- list(
    total_cov = total_cov,
    res_cov = list(
      name     = res_cov$name,
      coverage = res_cov$coverage,
      errors   = res_cov$errors,
      notes    = res_cov$notes
    ),
    functions_no_tests = no_tests_df,
    tests_skipped      = skip_tests,
    tests_passing      = test_files_clean,
    tests_problems     = problems
  )
  
  return(covr_list)
}

test_that("run_covr_skip_nstf returns correct structure on success", {
  
  # ------------------------------------------
  # Setup mock package structure
  # ------------------------------------------
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testthat"),
             recursive = TRUE, showWarnings = FALSE)
  
  test_pkg_data <- list(
    has_testit = FALSE,
    has_tinytest = FALSE
  )
  
  mapping <- data.frame(
    source_file = "file1.R",
    test_file   = "test-file1.R",
    stringsAsFactors = FALSE
  )
  
  problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  no_tests_df <- data.frame(
    source_file = "file2.R",
    stringsAsFactors = FALSE
  )
  
  coverage_obj <- structure(list(), class = "coverage")
  
  cov_env <- new.env(parent = emptyenv())
  
  # ------------------------------------------
  # STUB ALL DIRECT COLLABORATORS
  # ------------------------------------------
  
  mockery::stub(run_covr_skip_nstf, "get_pkg_name", "mockpkg")
  
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path",
                "mock/tests/testthat")
  
  mockery::stub(run_covr_skip_nstf, "restrict_test_paths_to_framework",
                function(x, ...) x)
  
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", mapping)
  
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_nstf", problems)
  
  mockery::stub(run_covr_skip_nstf, "get_function_no_tests", no_tests_df)
  
  mockery::stub(run_covr_skip_nstf, "pkgload::load_all", NULL)
  
  mockery::stub(run_covr_skip_nstf, "ls", c("fun1", "fun2"))
  
  mockery::stub(run_covr_skip_nstf, "get", function(obj, envir) obj)
  
  mockery::stub(run_covr_skip_nstf, "load_package_datasets", c("dataset1"))
  
  mockery::stub(run_covr_skip_nstf, "load_dependencies_into_env", NULL)
  
  mockery::stub(run_covr_skip_nstf, "load_base_r_packages_into_env", TRUE)
  
  mockery::stub(run_covr_skip_nstf, "normalizePath",
                function(path, ...) path)
  
  mockery::stub(run_covr_skip_nstf, "file.exists", TRUE)
  
  # FORCE covr::environment_coverage TO RETURN A VALID COVERAGE OBJECT
  mockery::stub(run_covr_skip_nstf, "covr::environment_coverage", coverage_obj)
  
  # ------------------------------------------
  # CRITICAL STUB: create_nstf_covr_list
  # (this MUST be stubbed using mockery::stub inside run_covr_skip_nstf)
  # ------------------------------------------
  mock_create_nstf_covr_list <- function(coverage, pkg_name,
                                         no_tests_df, test_files_clean,
                                         skip_tests, problems) {
    list(
      total_cov = 80,
      res_cov   = list(
        name     = pkg_name,
        coverage = list(
          filecoverage  = matrix(1, nrow = 1),
          totalcoverage = 80
        ),
        errors   = NA,
        notes    = NA
      ),
      functions_no_tests = no_tests_df,
      tests_skipped      = character(0),
      tests_passing      = test_files_clean,
      tests_problems     = problems
    )
  }
  
  mockery::stub(
    run_covr_skip_nstf,
    "create_nstf_covr_list",
    mock_create_nstf_covr_list
  )
  
  # DO NOT RETURN NULL FROM cleanup_and_return_null
  mockery::stub(
    run_covr_skip_nstf,
    "cleanup_and_return_null",
    function(...) invisible(NULL)
  )
  
  # ------------------------------------------
  # ACT
  # ------------------------------------------
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env)
  
  # ------------------------------------------
  # ASSERT
  # ------------------------------------------
  
  expect_type(result, "list")
  
  expect_true(all(c(
    "total_cov", "res_cov", "functions_no_tests",
    "tests_skipped", "tests_passing", "tests_problems"
  ) %in% names(result)))
  
  expect_equal(result$total_cov, 80)
  expect_equal(result$res_cov$coverage$totalcoverage, 80)
  
  expect_identical(result$functions_no_tests, no_tests_df)
  expect_identical(result$tests_skipped, character(0))
  
  expect_true(length(result$tests_passing) >= 1)
})





test_that("run_covr_skip_nstf returns testit nonzero coverage when at least one test passes", {
  # Arrange: create a temp pkg layout with tests/testit
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(file.path(pkg_source_path, "tests", "testit"), recursive = TRUE, showWarnings = FALSE)
  
  test_pkg_data <- list(has_testit = TRUE, has_tinytest = FALSE)
  pkg_name <- "mockpkg"
  
  # A simple mapping with one test file (keep it unskipped)
  mapping <- data.frame(source_file = "foo.R", test_file = "test-foo.R", stringsAsFactors = FALSE)
  
  # IMPORTANT: get_nstf_test_path should return a named list for frameworks
  test_path <- list(testit = file.path(pkg_source_path, "tests", "testit"))
  
  # --- Stubs on run_covr_skip_nstf's direct collaborators ---
  mockery::stub(run_covr_skip_nstf, "get_pkg_name",                     pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path",               test_path)
  mockery::stub(run_covr_skip_nstf, "restrict_test_paths_to_framework", function(x, ...) x)
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf",     mapping)
  mockery::stub(run_covr_skip_nstf, "cleanup_and_return_null",          NULL)
  
  # No skipped tests
  fake_problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_testit", fake_problems)
  
  mockery::stub(run_covr_skip_nstf, "get_function_no_tests", data.frame())
  mockery::stub(run_covr_skip_nstf, "pkgload::load_all",     NULL)
  mockery::stub(run_covr_skip_nstf, "ls",                    c("fun1"))
  mockery::stub(run_covr_skip_nstf, "get",                   function(obj, envir) obj)
  mockery::stub(run_covr_skip_nstf, "load_package_datasets", character(0))
  mockery::stub(run_covr_skip_nstf, "load_dependencies_into_env",    NULL)
  mockery::stub(run_covr_skip_nstf, "load_base_r_packages_into_env", TRUE)
  mockery::stub(run_covr_skip_nstf, "normalizePath",         function(path, ...) path)
  
  # Simulate a coverage object produced by environment_coverage()
  fake_coverage <- structure(list(), class = "coverage")
  mockery::stub(run_covr_skip_nstf, "covr::environment_coverage", fake_coverage)
  
  # ---- CRITICAL: Mock the collaborator create_nstf_covr_list() at package level ----
  mock_create_nstf_covr_list <- function(coverage, pkg_name,
                                         no_tests_df, test_files_clean,
                                         skip_tests, problems) {
    # Return a deterministic, non-NaN result
    list(
      total_cov = 100,
      res_cov   = list(
        name     = pkg_name,
        coverage = list(
          filecoverage  = matrix(1, nrow = 1, dimnames = list("foo.R", NULL)),
          totalcoverage = 100
        ),
        errors   = NA,
        notes    = NA
      ),
      functions_no_tests = no_tests_df,
      tests_skipped      = skip_tests,
      tests_passing      = test_files_clean,
      tests_problems     = problems
    )
  }
  
  # Bind the collaborator into the package namespace so *inner* calls pick it up
  testthat::local_mocked_bindings(
    .package = "test.assessr",
    create_nstf_covr_list = mock_create_nstf_covr_list
  )
  
  cov_env <- new.env(parent = emptyenv())
  
  # Act
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # Assert
  expect_type(result, "list")
  expect_true(all(c("total_cov","res_cov","functions_no_tests",
                    "tests_skipped","tests_passing","tests_problems") %in% names(result)))
  
  expect_equal(result$total_cov, 100)
  expect_equal(result$res_cov$coverage$totalcoverage, 100)
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "foo.R")
  expect_identical(result$tests_skipped, character(0))
  expect_identical(result$tests_passing, file.path(pkg_source_path, "tests", "testit", "test-foo.R"))
})






test_that("run_covr_skip_nstf handles multiple problems from check_covr_skip_testit", {
  
  # ---- Setup mock package ----
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testit"), recursive = TRUE, showWarnings = FALSE)
  
  test_pkg_data <- list(has_testit = TRUE)
  pkg_name <- "mockpkg"
  
  test_path <- list(testit = file.path(pkg_source_path, "tests", "testit"))
  
  mapping <- data.frame(
    source_file = "foo.R",
    test_file   = "test1.R",
    stringsAsFactors = FALSE
  )
  
  fake_problems <- data.frame(
    test_file  = c("test1.R", "test2.R"),
    issue_type = c("error: Test error!", "warning: Test warning!"),
    stringsAsFactors = FALSE
  )
  
  # Expected return object from the fallback path
  fallback_result <- list(
    total_cov = 0,
    res_cov = list(
      name = pkg_name,
      coverage = list(
        filecoverage  = matrix(0, nrow = 1, dimnames = list("No functions tested")),
        totalcoverage = 0
      ),
      errors = "No testthat or testit configuration",
      notes  = NA
    )
  )
  
  # ---- CRITICAL STUB: Force the function to return fallback structure ----
  mockery::stub(run_covr_skip_nstf, "create_nstf_covr_list", fallback_result)
  
  # ---- Other supporting stubs ----
  mockery::stub(run_covr_skip_nstf, "get_pkg_name", pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", test_path)
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", mapping)
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_testit", fake_problems)
  
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)
  mockery::stub(run_covr_skip_nstf, "get_function_no_tests", data.frame())
  mockery::stub(run_covr_skip_nstf, "pkgload::load_all", NULL)
  mockery::stub(run_covr_skip_nstf, "ls", character(0))
  mockery::stub(run_covr_skip_nstf, "get", function(obj, envir) obj)
  mockery::stub(run_covr_skip_nstf, "load_package_datasets", character(0))
  
  # ensure coverage step produces NULL
  mockery::stub(run_covr_skip_nstf, "covr::environment_coverage", NULL)
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Run ----
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assertions ----
  expect_type(result, "list")
  expect_equal(result$total_cov, 0)
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$coverage$totalcoverage, 0)
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "No functions tested")
  expect_equal(as.numeric(result$res_cov$coverage$filecoverage[1, 1]), 0)
  expect_equal(result$res_cov$errors, "No testthat or testit configuration")
  expect_true(is.na(result$res_cov$notes))
})



test_that("run_covr_skip_nstf handles multiple problems from check_covr_skip_testit", {
  
  # ---- Setup mock package ----
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testit"), recursive = TRUE, showWarnings = FALSE)
  
  test_pkg_data <- list(has_testit = TRUE)
  pkg_name <- "mockpkg"
  
  test_path <- list(testit = file.path(pkg_source_path, "tests", "testit"))
  
  mapping <- data.frame(
    source_file = "foo.R",
    test_file   = "test1.R",
    stringsAsFactors = FALSE
  )
  
  fake_problems <- data.frame(
    test_file  = c("test1.R", "test2.R"),
    issue_type = c("error: Test error!", "warning: Test warning!"),
    stringsAsFactors = FALSE
  )
  
  # Expected fallback return structure
  fallback_result <- list(
    total_cov = 0,
    res_cov = list(
      name = pkg_name,
      coverage = list(
        filecoverage  = matrix(0, nrow = 1, dimnames = list("No functions tested")),
        totalcoverage = 0
      ),
      errors = "No testthat or testit configuration",
      notes  = NA
    )
  )
  
  # ---- CRITICAL STUB: ensure function returns expected result ----
  mockery::stub(run_covr_skip_nstf, "create_nstf_covr_list", fallback_result)
  
  # ---- Supporting stubs ----
  mockery::stub(run_covr_skip_nstf, "get_pkg_name", pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", test_path)
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", mapping)
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_testit", fake_problems)
  
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)
  mockery::stub(run_covr_skip_nstf, "get_function_no_tests", data.frame())
  mockery::stub(run_covr_skip_nstf, "pkgload::load_all", NULL)
  mockery::stub(run_covr_skip_nstf, "ls", character(0))
  mockery::stub(run_covr_skip_nstf, "get", function(obj, envir) obj)
  mockery::stub(run_covr_skip_nstf, "load_package_datasets", character(0))
  
  # ---- FORCE ALL COVERAGE PATHS TO PRODUCE NULL ----
  mockery::stub(run_covr_skip_nstf, "create_base_tests_coverage", NULL)
  mockery::stub(run_covr_skip_nstf, "create_tinytest_coverage", NULL)
  mockery::stub(run_covr_skip_nstf, "covr::environment_coverage", NULL)
  mockery::stub(run_covr_skip_nstf, "covr::package_coverage", NULL)
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Run ----
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assertions ----
  expect_type(result, "list")
  expect_equal(result$total_cov, 0)
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$coverage$totalcoverage, 0)
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "No functions tested")
  expect_equal(as.numeric(result$res_cov$coverage$filecoverage[1, 1]), 0)
  expect_equal(result$res_cov$errors, "No testthat or testit configuration")
  expect_true(is.na(result$res_cov$notes))
  
})


test_that("find_pkg_root finds DESCRIPTION after traversing up", {
  # Simulate directory structure
  test_path <- "/home/user/project/tests/testit"
  pkg_root <- "/home/user/project"

  # Mock normalizePath to just return the input
  mockery::stub(check_covr_skip_testit, "normalizePath", function(x, mustWork = TRUE) x)

  # Mock file.exists: first call FALSE, second call TRUE
  file_exists_calls <- 0
  mockery::stub(check_covr_skip_testit, "file.exists", function(path) {
    file_exists_calls <<- file_exists_calls + 1
    if (file_exists_calls == 1) return(FALSE) # Not found in test_path
    if (file_exists_calls == 2) return(TRUE)  # Found in pkg_root
    FALSE
  })

  # Mock dirname: first call returns parent, then returns itself
  mockery::stub(check_covr_skip_testit, "dirname", function(path) {
    if (path == test_path) return(pkg_root)
    return(pkg_root)
  })

  # Minimal mapping and test file
  mapping <- data.frame(test_file = "test-foo.R", stringsAsFactors = FALSE)

  # Mock getwd/setwd/source to avoid side effects
  mockery::stub(check_covr_skip_testit, "getwd", function() pkg_root)
  mockery::stub(check_covr_skip_testit, "setwd", function(x) NULL)
  mockery::stub(check_covr_skip_testit, "source", function(file, local) NULL)

  # Should not error and should return "No tests skipped"
  result <- check_covr_skip_testit("mockpkg", mapping, test_path)
  expect_s3_class(result, "data.frame")
  expect_equal(result$issue_type[1], "No tests skipped")
})

test_that("find_pkg_root errors if DESCRIPTION never found", {
  test_path <- "/home/user/project/tests/testit"

  # Mock normalizePath to just return the input
  mockery::stub(check_covr_skip_testit, "normalizePath", function(x, mustWork = TRUE) x)

  # Mock file.exists: always FALSE
  mockery::stub(check_covr_skip_testit, "file.exists", function(path) FALSE)

  # Mock dirname: returns parent until root
  mockery::stub(check_covr_skip_testit, "dirname", function(path) {
    if (path == "/") return("/")
    return("/")
  })

  mapping <- data.frame(test_file = "test-foo.R", stringsAsFactors = FALSE)
  # Mock getwd/setwd/source to avoid side effects
  mockery::stub(check_covr_skip_testit, "getwd", function() "/")
  mockery::stub(check_covr_skip_testit, "setwd", function(x) NULL)
  mockery::stub(check_covr_skip_testit, "source", function(file, local) NULL)

  # Should error with "Package root (with DESCRIPTION) not found."
  expect_error(
    check_covr_skip_testit("mockpkg", mapping, test_path),
    "Package root \\(with DESCRIPTION\\) not found."
  )
})

test_that("returns NULL if get_source_test_mapping_nstf errors", {

  # Create a temporary directory to satisfy the assertion
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "mockpkg"), recursive = TRUE, showWarnings = FALSE)

  test_pkg_data <- list(has_testit = FALSE)
  pkg_name <- "mockpkg"


  mockery::stub(run_covr_skip_nstf, "get_pkg_name", pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", test_path)
  # Simulate error in get_source_test_mapping_nstf
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", function(...) stop("Simulated mapping error"))
  mockery::stub(run_covr_skip_nstf, "cleanup_and_return_null", NULL)
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)

  cov_env <- new.env(parent = emptyenv())
  # Run function
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  expect_null(result)
})

test_that("returns NULL if check_covr_skip_testit errors", {

  # Create a temporary directory to satisfy the assertion
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "mockpkg/tests", "testit"), recursive = TRUE, showWarnings = FALSE)

  test_pkg_data <- list(has_testit = TRUE)
  pkg_name <- "mockpkg"
  mapping <- data.frame(source_file = "foo.R", test_file = "test-foo.R", stringsAsFactors = FALSE)

  mockery::stub(run_covr_skip_nstf, "get_pkg_name", pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", test_path)
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", mapping)
  # Simulate error in check_covr_skip_testit
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_testit", function(...) stop("Simulated testit error"))
  mockery::stub(run_covr_skip_nstf, "cleanup_and_return_null", NULL)
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)

  cov_env <- new.env(parent = emptyenv())
  # Run function
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  expect_null(result)
})

test_that("returns NULL if check_covr_skip_nstf errors", {
  # Create a temporary directory to satisfy the assertion
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "mockpkg", "tests"), recursive = TRUE, showWarnings = FALSE)

  test_pkg_data <- list(has_testit = FALSE)
  pkg_name <- "mockpkg"
  mapping <- data.frame(source_file = "foo.R", test_file = "test-foo.R", stringsAsFactors = FALSE)

  mockery::stub(run_covr_skip_nstf, "get_pkg_name", pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", test_path)
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", mapping)
  # Simulate error in check_covr_skip_nstf
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_nstf", function(...) stop("Simulated nstf error"))
  mockery::stub(run_covr_skip_nstf, "cleanup_and_return_null", NULL)
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)

  cov_env <- new.env(parent = emptyenv())
  # Run function
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  expect_null(result)
})

test_that("run_covr_skip_nstf returns NULL if problems fails", {

  # Create a temporary fake package directory
  pkg_source_path <- file.path(tempdir(), "mockpkg_fail")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testthat"),
             recursive = TRUE, showWarnings = FALSE)

  test_pkg_data <- list(has_testit = FALSE)

  # Stubbing dependencies
  mockery::stub(run_covr_skip_nstf, "get_pkg_name", "mockpkg")
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", "tests/testthat")
  mockery::stub(run_covr_skip_nstf,
                "get_source_test_mapping_nstf",
                data.frame(source_file = "file1.R", test_file = "test-file1.R"))

  # Force the problems checker to fail (return NULL)
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_nstf", NULL)

  # cleanup should also return NULL but not error
  mockery::stub(run_covr_skip_nstf, "cleanup_and_return_null", NULL)

  # Avoid filesystem normalization issues
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)

  # Provide cov_env to satisfy assertion
  cov_env <- new.env()

  # Run function
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)

  # Expected behaviour: return NULL
  expect_null(result)
})




test_that("run_covr_skip_nstf handles multiple problems from check_covr_skip_testit", {
  
  # ---- Setup mock package ----
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testit"), recursive = TRUE, showWarnings = FALSE)
  
  test_pkg_data <- list(has_testit = TRUE)
  pkg_name <- "mockpkg"
  
  test_path <- list(testit = file.path(pkg_source_path, "tests", "testit"))
  
  mapping <- data.frame(
    source_file = "foo.R",
    test_file   = "test1.R",
    stringsAsFactors = FALSE
  )
  
  fake_problems <- data.frame(
    test_file  = c("test1.R", "test2.R"),
    issue_type = c("error: Test error!", "warning: Test warning!"),
    stringsAsFactors = FALSE
  )
  
  # Expected fallback structure
  fallback_result <- list(
    total_cov = 0,
    res_cov = list(
      name = pkg_name,
      coverage = list(
        filecoverage  = matrix(0, nrow = 1, dimnames = list("No functions tested")),
        totalcoverage = 0
      ),
      errors = "No testthat or testit configuration",
      notes  = NA
    )
  )
  
  # ------------------------------------------------------------------
  # IMPORTANT:
  # mockery::stub() REWRITES THE FUNCTION BODY.
  # Stubs PERSIST for entire R session.
  # → Therefore this entire test MUST be isolated in its own test file.
  # ------------------------------------------------------------------
  
  # ---- Stubs for deterministic behavior ----
  mockery::stub(run_covr_skip_nstf, "create_nstf_covr_list", fallback_result)
  mockery::stub(run_covr_skip_nstf, "get_pkg_name", pkg_name)
  mockery::stub(run_covr_skip_nstf, "get_nstf_test_path", test_path)
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf", mapping)
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_testit", fake_problems)
  
  mockery::stub(run_covr_skip_nstf, "normalizePath", function(path, ...) path)
  mockery::stub(run_covr_skip_nstf, "get_function_no_tests", data.frame())
  mockery::stub(run_covr_skip_nstf, "pkgload::load_all", NULL)
  mockery::stub(run_covr_skip_nstf, "ls", character(0))
  mockery::stub(run_covr_skip_nstf, "get", function(obj, envir) obj)
  mockery::stub(run_covr_skip_nstf, "load_package_datasets", character(0))
  
  # Force ALL coverage paths to NULL
  mockery::stub(run_covr_skip_nstf, "create_base_tests_coverage", NULL)
  mockery::stub(run_covr_skip_nstf, "create_tinytest_coverage", NULL)
  mockery::stub(run_covr_skip_nstf, "covr::environment_coverage", NULL)
  mockery::stub(run_covr_skip_nstf, "covr::package_coverage", NULL)
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Run ----
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assertions ----
  expect_type(result, "list")
  expect_equal(result$total_cov, 0)
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$coverage$totalcoverage, 0)
  expect_equal(rownames(result$res_cov$coverage$filecoverage), "No functions tested")
  expect_equal(as.numeric(result$res_cov$coverage$filecoverage[1, 1]), 0)
  expect_equal(result$res_cov$errors, "No testthat or testit configuration")
  expect_true(is.na(result$res_cov$notes))
  
})

test_that("returns problems for test files with errors and warnings", {
  dummy_mapping <- data.frame(
    source_file = c("modA.R", "modB.R"),
    test_file = c("test-modA.R", "test-modB.R"),
    stringsAsFactors = FALSE
  )

  # Stub file.path to return predictable paths
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # Stub asNamespace so check_covr_skip_nstf does not try loading real packages
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Stub source to simulate error and warning
  mockery::stub(check_covr_skip_nstf, "source", function(file, local) {
    if (grepl("test-modA", file)) stop("error in test-modA")
    if (grepl("test-modB", file)) warning("warning in test-modB")
  })

  result <- check_covr_skip_nstf("/mock/pkg", dummy_mapping)

  expect_equal(nrow(result), 2)
  expect_equal(result$test_file, c("test-modA.R", "test-modB.R"))
  expect_true(all(grepl("^error:", result$issue_type)))
})





test_that("returns default message when no problems found", {
  dummy_mapping <- data.frame(
    source_file = c("modA.R"),
    test_file   = c("test-modA.R"),
    stringsAsFactors = FALSE
  )

  # Fake namespace
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Predictable paths
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # File exists (prevent lookup failures)
  mockery::stub(check_covr_skip_nstf, "file.exists",
                function(...) TRUE)

  # Path normalization safe
  mockery::stub(check_covr_skip_nstf, "normalizePath",
                function(path, ..., mustWork = FALSE) path)

  # Source and sys.source must accept arbitrary args
  mockery::stub(check_covr_skip_nstf, "source",
                function(...) invisible(NULL))

  mockery::stub(check_covr_skip_nstf, "sys.source",
                function(...) invisible(NULL))

  # Act
  result <- check_covr_skip_nstf("/mock/pkg", dummy_mapping)

  # Assert
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$test_file))
  expect_equal(result$issue_type, "No tests skipped")
})

test_that("handles NA in mapping and drops them", {
  dummy_mapping <- data.frame(
    source_file = c("modA.R", "modB.R"),
    test_file = c(NA, "test-modB.R"),
    stringsAsFactors = FALSE
  )

  # Fake namespace
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Predictable paths
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # File exists (prevent lookup failures)
  mockery::stub(check_covr_skip_nstf, "file.exists",
                function(...) TRUE)

  # Path normalization safe
  mockery::stub(check_covr_skip_nstf, "normalizePath",
                function(path, ..., mustWork = FALSE) path)

  # Source and sys.source must accept arbitrary args
  mockery::stub(check_covr_skip_nstf, "source",
                function(...) invisible(NULL))

  mockery::stub(check_covr_skip_nstf, "sys.source",
                function(...) invisible(NULL))

  result <- check_covr_skip_nstf("/mock/pkg", dummy_mapping)

  expect_equal(nrow(result), 1)
  expect_true(is.na(result$test_file))
  expect_equal(result$issue_type, "No tests skipped")
})

test_that("returns 'No tests skipped' if all tests source fine", {
  mapping <- data.frame(test_file = c("test1.R", "test2.R"), stringsAsFactors = FALSE)
  pkg_source_path <- "mockpkg"

  # Fake namespace
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Predictable paths
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # File exists (prevent lookup failures)
  mockery::stub(check_covr_skip_nstf, "file.exists",
                function(...) TRUE)

  # Path normalization safe
  mockery::stub(check_covr_skip_nstf, "normalizePath",
                function(path, ..., mustWork = FALSE) path)

  # Source and sys.source must accept arbitrary args
  mockery::stub(check_covr_skip_nstf, "source",
                function(...) invisible(NULL))

  mockery::stub(check_covr_skip_nstf, "sys.source",
                function(...) invisible(NULL))

  result <- check_covr_skip_nstf(pkg_source_path, mapping)
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$test_file[1]))
  expect_equal(result$issue_type[1], "No tests skipped")
})

test_that("returns problems for tests that error or warn", {
  mapping <- data.frame(
    test_file = c("test1.R", "test2.R", "test3.R"),
    stringsAsFactors = FALSE
  )
  pkg_source_path <- "mockpkg"

  # ---- Required stubs ----

  # Prevent loading a real namespace
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Stable paths: e.g., "mockpkg/tests/testthat/test1.R"
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # Pretend all directories and files exist
  mockery::stub(check_covr_skip_nstf, "dir.exists",
                function(...) TRUE)
  mockery::stub(check_covr_skip_nstf, "file.exists",
                function(...) TRUE)

  # Avoid normalizePath errors
  mockery::stub(check_covr_skip_nstf, "normalizePath",
                function(path, ..., mustWork = FALSE) path)

  # Helper to pull `file` arg robustly from ... (handles named/unnamed)
  get_file_arg <- function(...) {
    args <- list(...)
    if (!is.null(args$file)) return(args$file)
    if (length(args) >= 1) return(args[[1]])
    stop("No file argument provided to stub")
  }

  # Stub both source and sys.source, accept arbitrary args
  mockery::stub(check_covr_skip_nstf, "source", function(...) {
    path <- get_file_arg(...)
    if (grepl("test2\\.R$", path)) stop("fail")
    if (grepl("test3\\.R$", path)) warning("warn")
    invisible(NULL)
  })

  mockery::stub(check_covr_skip_nstf, "sys.source", function(...) {
    path <- get_file_arg(...)
    if (grepl("test2\\.R$", path)) stop("fail")
    if (grepl("test3\\.R$", path)) warning("warn")
    invisible(NULL)
  })

  # ---- Run ----
  result <- check_covr_skip_nstf(pkg_source_path, mapping)

  # ---- Expectations ----
  expect_equal(nrow(result), 1)                 # Only test2 should be captured
  expect_setequal(result$test_file, "test2.R")  # Error case only
  expect_true(grepl("^error", result$issue_type))
})
 
test_that("drops NA rows in mapping", {
  mapping <- data.frame(test_file = c("test1.R", NA), stringsAsFactors = FALSE)
  pkg_source_path <- "mockpkg"

  # ---- Required stubs ----

  # Prevent loading a real namespace
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Stable paths: e.g., "mockpkg/tests/testthat/test1.R"
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # Pretend all directories and files exist
  mockery::stub(check_covr_skip_nstf, "dir.exists",
                function(...) TRUE)
  mockery::stub(check_covr_skip_nstf, "file.exists",
                function(...) TRUE)

  # Avoid normalizePath errors
  mockery::stub(check_covr_skip_nstf, "normalizePath",
                function(path, ..., mustWork = FALSE) path)

  # Helper to pull `file` arg robustly from ... (handles named/unnamed)
  get_file_arg <- function(...) {
    args <- list(...)
    if (!is.null(args$file)) return(args$file)
    if (length(args) >= 1) return(args[[1]])
    stop("No file argument provided to stub")
  }

  # Stub both source and sys.source, accept arbitrary args
  mockery::stub(check_covr_skip_nstf, "source", function(...) {
    path <- get_file_arg(...)
    if (grepl("test2\\.R$", path)) stop("fail")
    if (grepl("test3\\.R$", path)) warning("warn")
    invisible(NULL)
  })

  mockery::stub(check_covr_skip_nstf, "sys.source", function(...) {
    path <- get_file_arg(...)
    if (grepl("test2\\.R$", path)) stop("fail")
    if (grepl("test3\\.R$", path)) warning("warn")
    invisible(NULL)
  })

  result <- check_covr_skip_nstf(pkg_source_path, mapping)
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$test_file[1]))
  expect_equal(result$issue_type[1], "No tests skipped")
})

test_that("handles empty mapping", {
  mapping <- data.frame(test_file = character(0), stringsAsFactors = FALSE)
  pkg_source_path <- "mockpkg"

  # ---- Required stubs ----

  # Prevent loading a real namespace
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(check_covr_skip_nstf, "asNamespace",
                function(name) fake_ns)

  # Stable paths: e.g., "mockpkg/tests/testthat/test1.R"
  mockery::stub(check_covr_skip_nstf, "file.path",
                function(...) paste(..., sep = "/"))

  # Pretend all directories and files exist
  mockery::stub(check_covr_skip_nstf, "dir.exists",
                function(...) TRUE)
  mockery::stub(check_covr_skip_nstf, "file.exists",
                function(...) TRUE)

  # Avoid normalizePath errors
  mockery::stub(check_covr_skip_nstf, "normalizePath",
                function(path, ..., mustWork = FALSE) path)

  # Helper to pull `file` arg robustly from ... (handles named/unnamed)
  get_file_arg <- function(...) {
    args <- list(...)
    if (!is.null(args$file)) return(args$file)
    if (length(args) >= 1) return(args[[1]])
    stop("No file argument provided to stub")
  }

  # Stub both source and sys.source, accept arbitrary args
  mockery::stub(check_covr_skip_nstf, "source", function(...) {
    path <- get_file_arg(...)
    if (grepl("test2\\.R$", path)) stop("fail")
    if (grepl("test3\\.R$", path)) warning("warn")
    invisible(NULL)
  })

  mockery::stub(check_covr_skip_nstf, "sys.source", function(...) {
    path <- get_file_arg(...)
    if (grepl("test2\\.R$", path)) stop("fail")
    if (grepl("test3\\.R$", path)) warning("warn")
    invisible(NULL)
  })

  result <- check_covr_skip_nstf(pkg_source_path, mapping)
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$test_file[1]))
  expect_equal(result$issue_type[1], "No tests skipped")
})

test_that("returns problems for test files with errors", {
  # Mock mapping
  mapping <- data.frame(test_file = c("test1.R", "test2.R"), stringsAsFactors = FALSE)
  test_path <- "mock/tests/testit"
  pkg_name <- "mockpkg"

  # Mock find_pkg_root to just return a dummy path
  mockery::stub(check_covr_skip_testit, "find_pkg_root", "mockpkgroot")

  # Mock normalizePath to just return the input
  mockery::stub(check_covr_skip_testit, "normalizePath", function(x, mustWork = TRUE) x)

  # Mock file.exists to always return TRUE for DESCRIPTION
  mockery::stub(check_covr_skip_testit, "file.exists", function(x) TRUE)

  # Mock dirname to just return the parent directory
  mockery::stub(check_covr_skip_testit, "dirname", function(x) "mockpkgroot")

  # Mock getwd/setwd to do nothing
  mockery::stub(check_covr_skip_testit, "getwd", function() "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "setwd", function(x) NULL)

  # Mock source to throw error for test1.R, warning for test2.R
  mockery::stub(check_covr_skip_testit, "source", function(file, local) {
    if (grepl("test1.R", file)) stop("Test error!")
    if (grepl("test2.R", file)) warning("Test warning!")
    NULL
  })

  # Run function
  result <- check_covr_skip_testit(pkg_name, mapping, test_path)

  # Check output
  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 2)
  expect_equal(result$test_file, c("test1.R", "test2.R"))
  expect_true(any(grepl("error:", result$issue_type)))
  expect_true(any(grepl("warning:", result$issue_type)))
})

test_that("returns 'No tests skipped' if all tests pass", {
  mapping <- data.frame(test_file = c("test1.R"), stringsAsFactors = FALSE)
  test_path <- "mock/tests/testit"
  pkg_name <- "mockpkg"

  mockery::stub(check_covr_skip_testit, "find_pkg_root", "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "normalizePath", function(x, mustWork = TRUE) x)
  mockery::stub(check_covr_skip_testit, "file.exists", function(x) TRUE)
  mockery::stub(check_covr_skip_testit, "dirname", function(x) "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "getwd", function() "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "setwd", function(x) NULL)
  mockery::stub(check_covr_skip_testit, "source", function(file, local) NULL)

  result <- check_covr_skip_testit(pkg_name, mapping, test_path)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 1)
  expect_true(is.na(result$test_file[1]))
  expect_equal(result$issue_type[1], "No tests skipped")
})
 
test_that("handles mapping with NA values", {
  mapping <- data.frame(test_file = c(NA, "test2.R"), stringsAsFactors = FALSE)
  test_path <- "mock/tests/testit"
  pkg_name <- "mockpkg"

  mockery::stub(check_covr_skip_testit, "find_pkg_root", "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "normalizePath", function(x, mustWork = TRUE) x)
  mockery::stub(check_covr_skip_testit, "file.exists", function(x) TRUE)
  mockery::stub(check_covr_skip_testit, "dirname", function(x) "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "getwd", function() "mockpkgroot")
  mockery::stub(check_covr_skip_testit, "setwd", function(x) NULL)
  mockery::stub(check_covr_skip_testit, "source", function(file, local) warning("Test warning!"))

  result <- check_covr_skip_testit(pkg_name, mapping, test_path)

  expect_s3_class(result, "data.frame")
  expect_equal(nrow(result), 1)
  expect_equal(result$test_file[1], "test2.R")
  expect_true(grepl("warning:", result$issue_type[1]))
})
 
test_that("mapping matches test files to source files", {
  pkg_source_path <- "mockpkg"
  test_dir <- "mockpkg/tests"

  # Mock source and test files
  src_files <- c("foo.R", "bar.R", "baz.R")
  test_files <- c("test-foo.R", "test-bar.R")

  # Mock list.files for source_dir and test_dir
  mockery::stub(get_source_test_mapping_nstf, "list.files", function(path, pattern, full.names) {
    if (grepl("R$", path)) return(src_files)
    if (grepl("tests", path)) return(test_files)
    character(0)
  })

  mapping <- get_source_test_mapping_nstf(pkg_source_path, test_dir)

  expect_s3_class(mapping, "data.frame")
  expect_equal(nrow(mapping), 3)
  expect_equal(mapping$source_file, c("foo.R", "bar.R", "baz.R"))
  expect_equal(mapping$test_file, c("test-foo.R", "test-bar.R", NA))
})
 
test_that("returns empty mapping if no source files", {
  pkg_source_path <- "mockpkg"
  test_dir <- "mockpkg/tests"

  mockery::stub(get_source_test_mapping_nstf, "list.files", function(path, pattern, full.names) {
    if (grepl("R$", path)) return(character(0))
    if (grepl("tests", path)) return(c("test-foo.R"))
    character(0)
  })

  mapping <- get_source_test_mapping_nstf(pkg_source_path, test_dir)

  expect_equal(nrow(mapping), NULL)
})
 
test_that("returns NA for test_file if no match", {
  pkg_source_path <- "mockpkg"
  test_dir <- "mockpkg/tests"

  mockery::stub(get_source_test_mapping_nstf, "list.files", function(path, pattern, full.names) {
    if (grepl("R$", path)) return(c("foo.R"))
    if (grepl("tests", path)) return(c("test-bar.R"))
    character(0)
  })

  mapping <- get_source_test_mapping_nstf(pkg_source_path, test_dir)

  expect_equal(mapping$source_file, "foo.R")
  expect_true(is.na(mapping$test_file))
})
 
test_that("case-insensitive matching works", {
  pkg_source_path <- "mockpkg"
  test_dir <- "mockpkg/tests"

  mockery::stub(get_source_test_mapping_nstf, "list.files", function(path, pattern, full.names) {
    if (grepl("R$", path)) return(c("Foo.R"))
    if (grepl("tests", path)) return(c("test-foo.R"))
    character(0)
  })

  mapping <- get_source_test_mapping_nstf(pkg_source_path, test_dir)

  expect_equal(mapping$source_file, "Foo.R")
  expect_equal(mapping$test_file, "test-foo.R")
})
 
 
# Identity normalizePath stub to avoid "too many calls" and path surprises
normalizePath_identity <- function(path, ..., mustWork = FALSE) path

# dir.exists mock factory
dir_exists_fun <- function(truthy) {
  function(p) p %in% truthy
}

# list.files mock factory
list_files_fun <- function(mapping = list()) {
  function(path, pattern = NULL, full.names = FALSE, recursive = FALSE, ...) {
    out <- mapping[[path]]
    if (is.null(out)) character(0) else out
  }
}


# Minimal test_pkg_data builder
tpd <- function(
    has_testthat = FALSE,
    has_testit = FALSE,
    has_tests_base = FALSE,
    has_bioc_unit_tests = FALSE,
    has_BioG_test = FALSE,
    has_tinytest = FALSE,
    tinytest_runner_paths = NULL,
    has_data_table_tests = FALSE
) {
  list(
    has_testthat = has_testthat,
    has_testit = has_testit,
    has_tests_base = has_tests_base,
    has_bioc_unit_tests = has_bioc_unit_tests,
    has_BioG_test = has_BioG_test,
    has_tinytest = has_tinytest,
    tinytest_runner_paths = tinytest_runner_paths,
    has_data_table_tests = has_data_table_tests
  )
}


test_that("tinytest branch builds files under inst/tinytest, filters runner and skipped tests, and returns coverage", {
  
  run_fn <- run_covr_skip_nstf   # Or getFromNamespace() if needed
  
  # --- Arrange: mock package directory ---
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  
  normalizePath_passthru <- function(path, ...) path
  
  mapping <- data.frame(
    source_file = c("x.R", "y.R", "z.R"),
    test_file   = c("tinytest.R", "test-a.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  problems <- data.frame(
    test_file  = c("test-b.R", NA_character_),
    issue_type = c("skipped",  "No tests skipped"),
    stringsAsFactors = FALSE
  )
  
  no_tests_df <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  
  # coverage_to_list() return structure
  coverage_list <- list(
    filecoverage  = matrix(1),
    totalcoverage = 75
  )
  
  # total coverage summary
  compute_totals <- list(
    total_cov           = 75,
    percent_cov         = 75,
    n_testfiles         = 1,
    n_no_function_tests = 0,
    n_skipped_files     = 1
  )
  
  # Paths under inst/tinytest
  runner_path <- file.path(pkg_source_path, "inst", "tinytest", "tinytest.R")
  test_a_path <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  
  # ---- Stubs ----
  mockery::stub(run_fn, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(run_fn, "get_nstf_test_path",
                function(...) list(tinytest = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_fn, "normalizePath", normalizePath_passthru)
  
  mockery::stub(run_fn, "dir.exists",
                function(path) path %in% c(pkg_source_path,
                                           file.path(pkg_source_path, "inst"),
                                           file.path(pkg_source_path, "inst", "tinytest"))
  )
  
  mockery::stub(run_fn, "file.exists",
                function(paths) paths %in% c(runner_path, test_a_path)
  )
  
  mockery::stub(run_fn, "dir.create",  function(...) TRUE)
  mockery::stub(run_fn, "file.create", function(...) TRUE)
  mockery::stub(run_fn, "writeLines",  function(...) TRUE)
  
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_nstf",          function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests",         function(...) no_tests_df)
  
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_fn, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_fn, "asNamespace",       function(name) fake_ns)
  mockery::stub(run_fn, "ls",                function(envir, all.names = FALSE) c("fun1", "fun2"))
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_package_datasets",     function(...) c("dataset1"))
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE)
  
  # Capture arguments passed to create_tinytest_coverage
  captured_args <- new.env(parent = emptyenv())
  mockery::stub(
    run_fn, "create_tinytest_coverage",
    function(pkg_source_path, pkg_name, tiny_dir, at_home, color) {
      captured_args$pkg_source_path <- pkg_source_path
      captured_args$pkg_name        <- pkg_name
      captured_args$tiny_dir        <- tiny_dir
      captured_args$at_home         <- at_home
      captured_args$color           <- color
      list(coverage = list(), tinytests = list(ok = TRUE))
    }
  )
  
  mockery::stub(run_fn, "covr::coverage_to_list", function(coverage) coverage_list)
  mockery::stub(run_fn, "compute_total_coverage", function(...) compute_totals)
  
  # ---- CRITICAL: stub create_nstf_covr_list so we avoid NaN ----
  mockery::stub(
    run_fn, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      list(
        res_cov = list(
          name = pkg_name,
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 75
          )
        ),
        total_cov = 75,
        functions_no_tests = no_tests_df,
        tests_skipped = "test-b.R",
        tests_passing = file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
      )
    }
  )
  
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) invisible(NULL))
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Act ----
  test_pkg_data <- tpd(has_tinytest = TRUE)
  result <- run_fn(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assert ----
  expect_type(result, "list")
  expect_true(all(c("total_cov", "res_cov", "functions_no_tests",
                    "tests_skipped", "tests_passing") %in% names(result)))
  
  expected_passing <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  expect_setequal(result$tests_passing, expected_passing)
  expect_setequal(result$tests_skipped, "test-b.R")
  
  expect_equal(result$total_cov, 75)
  expect_equal(result$res_cov$coverage$totalcoverage, 75)
  
  expect_identical(captured_args$pkg_source_path, pkg_source_path)
  expect_identical(captured_args$pkg_name,        "mockpkg")
  expect_identical(captured_args$tiny_dir,        file.path(pkg_source_path, "inst", "tinytest"))
  expect_true(captured_args$at_home)
  expect_false(captured_args$color)
  
})



test_that("tinytest branch: warnings in problems do not produce skips", {
  
  # Arrange
  pkg_source_path <- file.path(tempdir(), "mockpkg_warns")
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  normalizePath_passthru <- function(path, ...) path
  
  mapping <- data.frame(
    source_file = c("x.R", "y.R"),
    test_file   = c("test-a.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  # WARNING ONLY → NO SKIPS EXPECTED
  problems <- data.frame(
    test_file  = c("test-a.R",      "test-b.R"),
    issue_type = c("warning: slow", "warning: flaky"),
    stringsAsFactors = FALSE
  )
  
  no_tests_df <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  
  coverage_list <- list(
    filecoverage  = matrix(1),
    totalcoverage = 60
  )
  
  compute_totals <- list(
    total_cov           = 60,
    percent_cov         = 60,
    n_testfiles         = 2,
    n_no_function_tests = 0,
    n_skipped_files     = 0
  )
  
  # Expected tinytest file paths
  test_a_path <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  test_b_path <- file.path(pkg_source_path, "inst", "tinytest", "test-b.R")
  
  # ---- Stubs ----
  mockery::stub(run_covr_skip_nstf, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(
    run_covr_skip_nstf, "get_nstf_test_path",
    function(...) list(tinytest = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_covr_skip_nstf, "normalizePath", normalizePath_passthru)
  
  mockery::stub(run_covr_skip_nstf, "get_source_test_mapping_nstf",
                function(...) mapping)
  
  mockery::stub(run_covr_skip_nstf, "check_covr_skip_nstf",
                function(...) problems)
  
  mockery::stub(run_covr_skip_nstf, "get_function_no_tests",
                function(...) no_tests_df)
  
  # Both files "exist"
  mockery::stub(
    run_covr_skip_nstf, "file.exists",
    function(paths) paths %in% c(test_a_path, test_b_path)
  )
  
  # Loader phase hermetic
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_covr_skip_nstf, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_covr_skip_nstf, "asNamespace",       function(name) fake_ns)
  mockery::stub(run_covr_skip_nstf, "ls",                function(...) "f1")
  mockery::stub(run_covr_skip_nstf, "get",               function(obj, envir) obj)
  mockery::stub(run_covr_skip_nstf, "load_package_datasets",
                function(...) character(0))
  mockery::stub(run_covr_skip_nstf, "load_dependencies_into_env",
                function(...) TRUE)
  
  # Prevent display/tcltk warnings
  mockery::stub(run_covr_skip_nstf, "load_base_r_packages_into_env",
                function(...) TRUE)
  
  # Tinytest coverage path
  mockery::stub(
    run_covr_skip_nstf, "create_tinytest_coverage",
    function(...) list(coverage = list(), tinytests = list(ok = TRUE))
  )
  
  mockery::stub(run_covr_skip_nstf, "covr::coverage_to_list",
                function(coverage) coverage_list)
  
  mockery::stub(run_covr_skip_nstf, "compute_total_coverage",
                function(...) compute_totals)
  
  # CRITICAL: avoid real normalization (produces NaN)
  mockery::stub(
    run_covr_skip_nstf, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      list(
        res_cov = list(
          name = pkg_name,
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 60
          )
        ),
        total_cov = 60,
        functions_no_tests = no_tests_df,
        tests_skipped = character(0),
        tests_passing = c(test_a_path, test_b_path)
      )
    }
  )
  
  mockery::stub(run_covr_skip_nstf, "cleanup_and_return_null",
                function(...) invisible(NULL))
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Act ----
  test_pkg_data <- tpd(has_tinytest = TRUE)
  result <- run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assert ----
  expect_setequal(result$tests_passing, c(test_a_path, test_b_path))
  expect_length(result$tests_skipped, 0L)
  
  expect_equal(result$total_cov, 60)
  expect_equal(result$res_cov$coverage$totalcoverage, 60)
  
})


# ---- Test 3: error in tiny coverage -> cleanup + NULL ----
test_that("tinytest branch: error in tiny coverage triggers cleanup + returns NULL", {
  
  # Use the real namespaced function so stubs attach correctly
  run_fn <- getFromNamespace("run_covr_skip_nstf", "test.assessr")
  
  # Arrange
  pkg_source_path <- file.path(tempdir(), "mockpkg_err")
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  normalizePath_passthru <- function(path, ...) path
  
  mapping <- data.frame(
    source_file = c("x.R", "y.R"),
    test_file   = c("test-a.R", "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  # Warnings only → do not skip (not actually relevant for this test)
  problems <- data.frame(
    test_file  = c("test-a.R", "test-b.R"),
    issue_type = c("warning: slow", "warning: flaky"),
    stringsAsFactors = FALSE
  )
  
  no_tests_df <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  
  # Expected tinytest test file paths
  test_a_path <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  test_b_path <- file.path(pkg_source_path, "inst", "tinytest", "test-b.R")
  
  # ---- Stubs ----
  mockery::stub(run_fn, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(
    run_fn, "get_nstf_test_path",
    function(...) list(tinytest = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_fn, "normalizePath", normalizePath_passthru)
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_nstf",          function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests",         function(...) no_tests_df)
  
  # Tinytest files "exist"
  mockery::stub(
    run_fn, "file.exists",
    function(paths) paths %in% c(test_a_path, test_b_path)
  )
  
  # Hermetic loading stubs
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_fn, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_fn, "asNamespace",       function(...) fake_ns)
  mockery::stub(run_fn, "ls",                function(...) "fun1")
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_package_datasets",     function(...) character(0))
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE)
  
  # ---- CRITICAL: Force tinytest coverage to ERROR ----
  mockery::stub(
    run_fn, "create_tinytest_coverage",
    function(...) stop("tinytest internal failure")
  )
  
  # ---- cleanup handler returns NULL (expected output for this test) ----
  mockery::stub(
    run_fn, "cleanup_and_return_null",
    function(...) NULL
  )
  
  # Important: prevent downstream coverage logic from running
  mockery::stub(run_fn, "covr::coverage_to_list",
                function(...) stop("should not be called"))
  mockery::stub(run_fn, "compute_total_coverage",
                function(...) stop("should not be called"))
  mockery::stub(run_fn, "create_nstf_covr_list",
                function(...) stop("should not be called"))
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Act ----
  test_pkg_data <- tpd(has_tinytest = TRUE)
  result <- run_fn(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assert ----
  expect_null(result)             # CLEANUP + NULL return expected
})


# ---- Test 4: runner excluded even with no problems ----

test_that("tinytest branch: runner is excluded even when no problems / warnings / skips", {
  
  run_fn <- getFromNamespace("run_covr_skip_nstf", "test.assessr")
  
  # Arrange
  pkg_source_path <- file.path(tempdir(), "mockpkg_runner")
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  normalizePath_passthru <- function(path, ...) path
  
  # Mapping includes a RUNNER + two real tinytest files
  mapping <- data.frame(
    source_file = c("x.R",        "y.R",        "z.R"),
    test_file   = c("tinytest.R", "test-a.R",   "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  # No problems (no warnings, no skips)
  problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  no_tests_df <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  
  coverage_list <- list(
    filecoverage  = matrix(1),
    totalcoverage = 88
  )
  
  compute_totals <- list(
    total_cov           = 88,
    percent_cov         = 88,
    n_testfiles         = 2,   # runner excluded
    n_no_function_tests = 0,
    n_skipped_files     = 0
  )
  
  # Expected tinytest file paths
  runner_path <- file.path(pkg_source_path, "inst", "tinytest", "tinytest.R")
  test_a_path <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  test_b_path <- file.path(pkg_source_path, "inst", "tinytest", "test-b.R")
  
  # --- Stubs ---
  mockery::stub(run_fn, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(
    run_fn, "get_nstf_test_path",
    function(...) list(tinytest = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_fn, "normalizePath", normalizePath_passthru)
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_nstf",          function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests",         function(...) no_tests_df)
  
  # Only test-a and test-b should “exist”; runner is ignored even if it exists
  mockery::stub(
    run_fn, "file.exists",
    function(paths) paths %in% c(test_a_path, test_b_path, runner_path)
  )
  
  # Hermetic loader
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_fn, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_fn, "asNamespace",       function(...) fake_ns)
  mockery::stub(run_fn, "ls",                function(...) "fun1")
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_package_datasets",     function(...) character(0))
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE)
  
  # Tinytest coverage OK
  mockery::stub(
    run_fn, "create_tinytest_coverage",
    function(...) list(coverage = list(), tinytests = list(ok = TRUE))
  )
  
  mockery::stub(run_fn, "covr::coverage_to_list",
                function(...) coverage_list)
  
  mockery::stub(run_fn, "compute_total_coverage",
                function(...) compute_totals)
  
  # Avoid real env normalization
  mockery::stub(
    run_fn, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      
      # Assert inside stub: runner must NOT appear in test_files_clean
      if (any(grepl("tinytest.R$", test_files_clean))) {
        stop("Runner was NOT removed!")
      }
      
      list(
        res_cov = list(
          name = pkg_name,
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 88
          )
        ),
        total_cov = 88,
        functions_no_tests = no_tests_df,
        tests_skipped = character(0),
        tests_passing = c(test_a_path, test_b_path)
      )
    }
  )
  
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) invisible(NULL))
  
  cov_env <- new.env(parent = emptyenv())
  
  # Act
  test_pkg_data <- tpd(has_tinytest = TRUE)
  result <- run_fn(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # Assert
  expect_type(result, "list")
  
  # Runner is excluded
  expect_setequal(result$tests_passing, c(test_a_path, test_b_path))
  expect_false(any(grepl("tinytest.R$", result$tests_passing)))
  
  expect_length(result$tests_skipped, 0L)
  
  expect_equal(result$total_cov, 88)
  expect_equal(result$res_cov$coverage$totalcoverage, 88)
})

 


test_that("base layout: success path collects coverage", {
  
  run_fn <- getFromNamespace("run_covr_skip_nstf", "test.assessr")
  
  # --- Arrange ---
  pkg_source_path <- file.path(tempdir(), "mockpkg_base_ok")
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  normalizePath_passthru <- function(path, ...) path
  
  mapping <- data.frame(
    source_file = c("a.R", "b.R"),
    test_file   = c("t-a.R", "t-b.R"),
    stringsAsFactors = FALSE
  )
  
  problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  no_tests_df <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  
  coverage_obj <- structure(list(), class = "coverage")
  
  coverage_list <- list(
    filecoverage  = matrix(1),
    totalcoverage = 82
  )
  
  compute_totals <- list(
    total_cov           = 82,
    percent_cov         = 82,
    n_testfiles         = 2,
    n_no_function_tests = 0,
    n_skipped_files     = 0
  )
  
  tf1 <- file.path(pkg_source_path, "tests", "t-a.R")
  tf2 <- file.path(pkg_source_path, "tests", "t-b.R")
  
  # ---- STUBS ----
  mockery::stub(run_fn, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(
    run_fn, "get_nstf_test_path",
    function(...) list(base = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_fn, "normalizePath", normalizePath_passthru)
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_nstf",          function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests",         function(...) no_tests_df)
  
  # Selective dir.exists stub: base layout only
  mockery::stub(
    run_fn, "dir.exists",
    function(path) {
      path %in% c(
        pkg_source_path,
        file.path(pkg_source_path, "tests")
      )
    }
  )
  
  # hermetic loader stubs
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_fn, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_fn, "asNamespace",       function(...) fake_ns)
  mockery::stub(run_fn, "ls",                function(...) character(0))
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_package_datasets",     function(...) character(0))
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE)
  
  captured <- new.env(parent = emptyenv())
  mockery::stub(
    run_fn, "create_base_tests_coverage",
    function(pkg_source_path, pkg_name, abs_tests) {
      captured$pkg_source_path <- pkg_source_path
      captured$pkg_name        <- pkg_name
      captured$abs_tests       <- abs_tests
      list(coverage = coverage_obj)
    }
  )
  
  mockery::stub(run_fn, "covr::coverage_to_list", function(...) coverage_list)
  mockery::stub(run_fn, "compute_total_coverage", function(...) compute_totals)
  
  mockery::stub(
    run_fn, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      list(
        res_cov = list(
          name = pkg_name,
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 82
          )
        ),
        total_cov = 82,
        functions_no_tests = no_tests_df,
        tests_skipped = character(0),
        tests_passing = c(tf1, tf2)
      )
    }
  )
  
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) invisible(NULL))
  
  cov_env <- new.env(parent = emptyenv())
  
  # --- Act ---
  expect_message({
    test_pkg_data <- tpd(has_tinytest = FALSE, has_testit = FALSE)
    result <- run_fn(pkg_source_path, test_pkg_data, cov_env = cov_env)
  }, sprintf("running base-style tests coverage for %s", pkg_name))
  
  # --- Assert ---
  expect_identical(captured$pkg_source_path, pkg_source_path)
  expect_identical(captured$pkg_name, pkg_name)
  
  expect_type(result, "list")
  expect_setequal(result$tests_passing, c(tf1, tf2))
  expect_length(result$tests_skipped, 0)
  expect_equal(result$total_cov, 82)
  expect_equal(result$res_cov$coverage$totalcoverage, 82)
})


test_that("returns testthat path when tests/testthat exists", {
  # IMPORTANT: if your function gates on has_testthat, set it TRUE
  test_pkg_data <- tpd(has_testthat = TRUE)

  # Make path building predictable
  mockery::stub(get_nstf_test_path, "file.path",
                function(...) paste(..., sep = "/"))

  # Normalize path as identity (handles unlimited calls safely)
  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)

  # The function should see these two directories as existing
  true_dirs <- c("/pkg/tests", "/pkg/tests/testthat")
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))

  # No base .R files in /pkg/tests (so default_tests_root should be omitted)
  mockery::stub(
    get_nstf_test_path, "list.files",
    list_files_fun(list(
      "/pkg/tests" = character(0)
    ))
  )

  # Call with the tests/ directory as per the function contract
  res <- get_nstf_test_path(test_pkg_data, "/pkg/tests")

  # Expect only the frameworks you want to expose
  expect_named(res, "testthat")
  expect_identical(res$testthat, "/pkg/tests/testthat")
})

test_that("returns testit path when tests/testit exists", {
  test_pkg_data <- tpd(has_testit = TRUE)

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)
  true_dirs <- c("/pkg/tests", "/pkg/tests/testit")
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(list()))

  res <- get_nstf_test_path(test_pkg_data, "IGNORED")

  expect_named(res, c("testit", "nonstandard_testit_root"))
  expect_identical(res$testit, "IGNORED/testit")
})

############################################





test_that("includes nonstandard testit root when test-ci and test-cran both exist", {
  test_pkg_data <- tpd(has_testit = TRUE)

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)

  true_dirs <- c(
    "/pkg/tests",
    "/pkg/tests/test-ci",
    "/pkg/tests/test-cran"
  )
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))

  # No base tests
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(list()))

  res <- get_nstf_test_path(test_pkg_data, "IGNORED")

  expect_named(res, c("testit", "nonstandard_testit_root"))
  expect_identical(res$nonstandard_testit_root, "IGNORED")
})
 
# list.files mock that respects pattern and full.names for one directory

# list.files mock that tolerates trailing slash, respects pattern/full.names/recursive,
# and returns filenames that satisfy common base-test patterns (e.g., ^test.*\\.[Rr]$)
list_files_base_tests <- function() {
  function(path, pattern = NULL, full.names = FALSE, recursive = FALSE, ...) {
    # Normalize trailing slash
    norm_path <- sub("/+$", "", path)

    # Only emulate for /pkg/tests
    if (!identical(norm_path, "/pkg/tests")) return(character(0))

    # Common base R test filenames often start with 'test-'
    files <- c("test-a.R", "test-b.R")

    # Apply pattern if provided
    if (!is.null(pattern)) {
      files <- files[grepl(pattern, files, perl = TRUE)]
    }

    # If recursive = TRUE, behavior is the same here because files are directly under tests/
    # (adjust if your production code expects nested files)

    # Apply full.names if requested
    if (full.names) {
      files <- file.path(path, files)
    }

    files
  }
}

test_that("detects base R tests when .R files are directly under tests/", {
  # If your implementation gates on has_tests_base, set TRUE here.
  # If detection is purely by file presence, leaving FALSE is fine.
  test_pkg_data <- tpd(has_tests_base = TRUE)

  # Stable paths
  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)
  mockery::stub(get_nstf_test_path, "file.path",
                function(...) paste(..., sep = "/"))

  # The function may probe /pkg, /pkg/tests, and sometimes with a trailing slash
  true_dirs <- c("/pkg", "/pkg/tests", "/pkg/tests/")
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))

  # Simulate base R tests directly in /pkg/tests
  mockery::stub(get_nstf_test_path, "list.files", list_files_base_tests())

  # IMPORTANT: pass "/pkg" (not "IGNORED") so file.path joins match your mocks
  res <- get_nstf_test_path(test_pkg_data, "/pkg")

  # Expected: base tests root is detected
  expect_named(res, "base_tests_root")
  expect_identical(res$base_tests_root, "/pkg")
})
 
test_that("detects Bioconductor unit tests under inst/unitTests", {
  # Use the correct capability flag expected by production
  test_pkg_data <- tpd(has_bioc_unit_tests = TRUE)

  # --- Pure/path-stable stubs ---
  # Collapse ".." and "." without touching the filesystem (so the test is deterministic)
  normalizePath_collapse <- function(path, winslash = "/", mustWork = FALSE) {
    collapse_one <- function(p) {
      # Keep leading slash if present
      abs <- startsWith(p, "/")
      parts <- strsplit(p, "/")[[1]]
      stack <- character()
      for (seg in parts) {
        if (seg == "" || seg == ".") next
        if (seg == "..") {
          if (length(stack) > 0) stack <- stack[-length(stack)]
        } else {
          stack <- c(stack, seg)
        }
      }
      out <- paste(stack, collapse = "/")
      if (abs) out <- paste0("/", out)
      if (out == "" && abs) out <- "/"  # root case
      out
    }
    # Support vectors if the SUT loops over paths
    if (length(path) > 1) return(vapply(path, collapse_one, character(1)))
    collapse_one(path)
  }

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_collapse)
  mockery::stub(get_nstf_test_path, "file.path", function(...) paste(..., sep = "/"))

  # Make only the Bioc unit tests path (and parents) exist.
  # IMPORTANT: Do NOT include "/pkg/tests" here, to avoid default_tests_root detection.
  true_dirs <- c("/pkg", "/pkg/inst", "/pkg/inst/unitTests", "/pkg/inst/unitTests/")
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))

  # No base .R files anywhere
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(list()))

  # Pass the tests directory (correct contract for 'testdir')
  res <- get_nstf_test_path(test_pkg_data, "/pkg/tests")

  expect_named(res, "bioc_unitTests")
  expect_identical(res$bioc_unitTests, "/pkg/inst/unitTests")
})

test_that("de-duplicates when both nonstandard_testit_root and base_tests_root resolve to tests/", {
  test_pkg_data <- tpd()

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)
  true_dirs <- c("/pkg/tests", "/pkg/tests/test-ci", "/pkg/tests/test-cran")
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))
  # base files present
  files_map <- list("/pkg/tests" = c("/pkg/tests/only_base.R"))
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(files_map))

  res <- get_nstf_test_path(test_pkg_data, "IGNORED")

  # Should keep only the first appearance (tinytest)
  expect_length(res, 1L)
  expect_named(res, "default_tests_root")
  expect_identical(res$default_tests_root, "IGNORED")
})
 
test_that("falls back to default_tests_root when nothing detected", {
  test_pkg_data <- tpd()

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)
  mockery::stub(get_nstf_test_path, "file.path",
                function(...) paste(..., sep = "/"))

  # No directories exist at all
  true_dirs <- character(0)
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(true_dirs))
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(list()))

  # Use a concrete root so default path resolves to /pkg/tests
  res <- get_nstf_test_path(test_pkg_data, "/pkg/tests")

  expect_named(res, "default_tests_root")
  expect_identical(res$default_tests_root, "/pkg/tests")
})

test_that("flag-driven inclusion: has_testthat = TRUE includes tests/testthat even if dir doesn't exist", {
  test_pkg_data <- tpd(has_testthat = TRUE)

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)
  # Simulate that no directories exist; list.files irrelevant here
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(character(0)))
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(list()))

  res <- get_nstf_test_path(test_pkg_data, "IGNORED")

  # Since flag is TRUE, testthat should be included regardless of FS state.
  expect_named(res, "testthat")
  expect_identical(res$testthat, "IGNORED/testthat")
})

test_that("flag-driven inclusion: has_bioc_unit_tests = TRUE includes inst/unitTests even if dir doesn't exist", {
  test_pkg_data <- tpd(has_bioc_unit_tests = TRUE)

  mockery::stub(get_nstf_test_path, "normalizePath", normalizePath_identity)
  mockery::stub(get_nstf_test_path, "dir.exists", dir_exists_fun(character(0)))
  mockery::stub(get_nstf_test_path, "list.files", list_files_fun(list()))

  res <- get_nstf_test_path(test_pkg_data, "IGNORED")

  expect_named(res, "bioc_unitTests")
  expect_identical(res$bioc_unitTests, "IGNORED/../inst/unitTests")
})
 
# Helpers ---------------------------------------------------------------
# A fake covr namespace that flips flags in a shared environment
make_fake_covr_ns <- function(flags_env, counters_value, coverage_value) {
  stopifnot(is.environment(flags_env))

  covr_ns <- new.env(parent = emptyenv())

  # When called, these will flip booleans in flags_env by reference
  assign("trace_environment", function(env) {
    flags_env$trace <- TRUE
    invisible(NULL)
  }, covr_ns)

  assign("reset_traces", function() {
    flags_env$reset <- TRUE
    invisible(NULL)
  }, covr_ns)

  assign("clear_counters", function() {
    flags_env$clear <- TRUE
    invisible(NULL)
  }, covr_ns)

  assign("as_coverage", function(x) coverage_value, covr_ns)
  assign(".counters", counters_value, covr_ns)

  covr_ns
}


# A safe dir.exists stub that only returns TRUE for whitelisted paths
dir_exists_only <- function(truthy) {
  function(p) p %in% truthy
}

# ----------------------------------------------------------------------


test_that("runs tinytest coverage successfully and returns structure", {
  pkg_source_path <- "/pkg"
  tiny_dir        <- "/pkg/inst/tinytest"
  pkg_name        <- "mypkg"

  # Shared flags environment (modified by reference)
  flags <- new.env(parent = emptyenv())
  flags$trace <- FALSE; flags$reset <- FALSE; flags$clear <- FALSE

  # Fake covr env & outputs
  fake_counters <- list(counter_data = 123)
  fake_coverage <- structure(list(kind = "coverage", from = "as_coverage"),
                             class = "coverage")
  fake_covr_ns  <- make_fake_covr_ns(flags,
                                     counters_value = fake_counters,
                                     coverage_value = fake_coverage)

  # Fake package namespace env (target to be traced)
  fake_pkg_ns <- new.env(parent = emptyenv())

  # ---- Stubs ----

  # Dirs exist
  mockery::stub(create_tinytest_coverage, "dir.exists",
                function(p) p %in% c(pkg_source_path, tiny_dir))

  # Namespace is "loaded"
  mockery::stub(create_tinytest_coverage, "loadedNamespaces",
                function() c(pkg_name, "covr", "tinytest"))

  # Return fake envs for covr & the target pkg
  mockery::stub(create_tinytest_coverage, "asNamespace", function(name) {
    if (identical(name, "covr"))    return(fake_covr_ns)
    if (identical(name, pkg_name))  return(fake_pkg_ns)
    stop("unexpected asNamespace(", name, ")")
  })

  # tinytest runner returns a known value
  fake_tt_result <- list(results = "all good")
  mockery::stub(create_tinytest_coverage, "tinytest::run_test_dir",
                function(dir, at_home, color) {
                  expect_identical(dir, tiny_dir)
                  expect_true(at_home)
                  expect_false(color)
                  fake_tt_result
                })

  # Ensure NOT_CRAN is restored (set to a known value before)
  Sys.setenv(NOT_CRAN = "OLD_VALUE")
  on.exit(Sys.setenv(NOT_CRAN = ""), add = TRUE)

  # ---- Run ----
  res <- create_tinytest_coverage(pkg_source_path, pkg_name, tiny_dir = tiny_dir)

  # ---- Assertions ----
  expect_type(res, "list")
  expect_true(all(c("coverage", "tinytests") %in% names(res)))
  expect_identical(res$coverage,  fake_coverage)
  expect_identical(res$tinytests, fake_tt_result)

  # covr hooks invoked
  expect_true(isTRUE(flags$trace))
  expect_true(isTRUE(flags$reset))
  expect_true(isTRUE(flags$clear))

  # NOT_CRAN restored to prior value
  expect_identical(Sys.getenv("NOT_CRAN"), "OLD_VALUE")
})
 
test_that("cleanup (reset_traces/clear_counters) runs even if tinytest fails", {
  pkg_source_path <- "/pkg"
  tiny_dir        <- "/pkg/inst/tinytest"
  pkg_name        <- "mypkg"

  # Shared flags
  flags <- new.env(parent = emptyenv())
  flags$trace <- FALSE; flags$reset <- FALSE; flags$clear <- FALSE

  fake_counters <- list(counter_data = 999)
  fake_coverage <- structure(list(kind = "coverage"), class = "coverage")
  fake_covr_ns  <- make_fake_covr_ns(flags,
                                     counters_value = fake_counters,
                                     coverage_value = fake_coverage)
  fake_pkg_ns <- new.env(parent = emptyenv())

  # Dirs exist
  mockery::stub(create_tinytest_coverage, "dir.exists",
                function(p) p %in% c(pkg_source_path, tiny_dir))

  # Namespace is "loaded"
  mockery::stub(create_tinytest_coverage, "loadedNamespaces",
                function() c(pkg_name, "covr", "tinytest"))

  # Return fake envs
  mockery::stub(create_tinytest_coverage, "asNamespace", function(name) {
    if (identical(name, "covr"))    return(fake_covr_ns)
    if (identical(name, pkg_name))  return(fake_pkg_ns)
    stop("unexpected asNamespace(", name, ")")
  })

  # tinytest fails
  mockery::stub(create_tinytest_coverage, "tinytest::run_test_dir",
                function(dir, at_home, color) stop("tinytest boom"))

  # Preserve & check NOT_CRAN restoration
  Sys.unsetenv("NOT_CRAN")
  on.exit(Sys.unsetenv("NOT_CRAN"), add = TRUE)

  expect_error(
    create_tinytest_coverage(pkg_source_path, pkg_name, tiny_dir = tiny_dir),
    regexp = "tinytest boom"
  )

  # Cleanup ran due to on.exit
  expect_true(isTRUE(flags$trace))  # trace happened before error
  expect_true(isTRUE(flags$reset))  # reset_traces() in on.exit
  expect_true(isTRUE(flags$clear))  # clear_counters() in on.exit

  # NOT_CRAN restored to unset (since it was unset before)
  expect_identical(Sys.getenv("NOT_CRAN", unset = NA), NA_character_)
})


test_that("tinytest branch: runner basenames are removed from test files", {
  
  run_fn <- getFromNamespace("run_covr_skip_nstf", "test.assessr")
  
  # Arrange: real package dir and tests/
  pkg_source_path <- file.path(tempdir(), "mockpkg_tiny_runner_basenames")
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  
  # Mapping contains three runner basenames + two real tests
  mapping <- data.frame(
    source_file = c("x.R",         "y.R",           "z.R",              "a.R",         "b.R"),
    test_file   = c("tinytest.R",  "runTinyTests.R","run_tinytest.R",   "test-a.R",    "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  no_tests_df   <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  coverage_obj  <- structure(list(), class = "coverage")
  
  # Expected paths after runner filtering
  exp_a <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  exp_b <- file.path(pkg_source_path, "inst", "tinytest", "test-b.R")
  
  # ---- Stubs ----
  mockery::stub(run_fn, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(
    run_fn, "get_nstf_test_path",
    function(...) list(tinytest = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_fn, "normalizePath", function(path, ...) path)
  
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_nstf",          function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests",         function(...) no_tests_df)
  
  # All tinytest files "exist"
  mockery::stub(run_fn, "file.exists", function(paths) TRUE)
  
  # Loader stubs
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_fn, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_fn, "asNamespace",       function(name) fake_ns)
  mockery::stub(run_fn, "ls",                function(...) character(0))
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_package_datasets",     function(...) character(0))
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE)
  
  # Tinytest coverage path
  mockery::stub(
    run_fn, "create_tinytest_coverage",
    function(...) list(coverage = coverage_obj, tinytests = list(ok = TRUE))
  )
  
  mockery::stub(
    run_fn, "covr::coverage_to_list",
    function(...) list(filecoverage = matrix(1), totalcoverage = 88)
  )
  
  # CRITICAL: Fake coverage normalization so result$total_cov == 88
  mockery::stub(
    run_fn, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      
      # Validate inside stub: runner files must not remain
      if (any(basename(test_files_clean) %in% c("tinytest.R",
                                                "runTinyTests.R",
                                                "run_tinytest.R"))) {
        stop("Runner basename was NOT removed!")
      }
      
      list(
        res_cov = list(
          name = pkg_name,
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 88
          )
        ),
        total_cov           = 88,
        functions_no_tests  = no_tests_df,
        tests_skipped       = character(0),
        tests_passing       = c(exp_a, exp_b)
      )
    }
  )
  
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) invisible(NULL))
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Act ----
  test_pkg_data <- tpd(has_tinytest = TRUE)
  result <- run_fn(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assert ----
  expect_type(result, "list")
  
  expect_setequal(result$tests_passing, c(exp_a, exp_b))
  expect_length(result$tests_skipped, 0L)
  
  expect_equal(result$total_cov, 88)
  expect_equal(result$res_cov$coverage$totalcoverage, 88)
})



test_that("tinytest branch: runner basenames are removed from test files", {
  
  run_fn <- getFromNamespace("run_covr_skip_nstf", "test.assessr")
  
  # Arrange: real package dir
  pkg_source_path <- file.path(tempdir(), "mockpkg_tiny_runner_basenames")
  dir.create(file.path(pkg_source_path, "tests"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  pkg_name <- "mockpkg"
  
  mapping <- data.frame(
    source_file = c("x.R",         "y.R",           "z.R",              "a.R",         "b.R"),
    test_file   = c("tinytest.R",  "runTinyTests.R","run_tinytest.R",   "test-a.R",    "test-b.R"),
    stringsAsFactors = FALSE
  )
  
  problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  no_tests_df   <- data.frame(source_file = character(0), stringsAsFactors = FALSE)
  coverage_obj  <- structure(list(), class = "coverage")
  
  exp_a <- file.path(pkg_source_path, "inst", "tinytest", "test-a.R")
  exp_b <- file.path(pkg_source_path, "inst", "tinytest", "test-b.R")
  
  # ---- Stubs ----
  mockery::stub(run_fn, "get_pkg_name", function(...) pkg_name)
  
  mockery::stub(
    run_fn, "get_nstf_test_path",
    function(...) list(tinytest = file.path(pkg_source_path, "tests"))
  )
  
  mockery::stub(run_fn, "normalizePath", function(path, ...) path)
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_nstf",          function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests",         function(...) no_tests_df)
  
  mockery::stub(run_fn, "file.exists", function(paths) TRUE)
  
  fake_ns <- new.env(parent = emptyenv())
  mockery::stub(run_fn, "pkgload::load_all", function(...) NULL)
  mockery::stub(run_fn, "asNamespace",       function(name) fake_ns)
  mockery::stub(run_fn, "ls",                function(...) character(0))
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_package_datasets",     function(...) character(0))
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE)
  
  mockery::stub(
    run_fn, "create_tinytest_coverage",
    function(...) list(coverage = coverage_obj, tinytests = list(ok = TRUE))
  )
  
  mockery::stub(
    run_fn, "covr::coverage_to_list",
    function(...) list(filecoverage = matrix(1), totalcoverage = 77)
  )
  
  # CRITICAL FIX — add this for your 77% case
  mockery::stub(
    run_fn, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      
      # Verify runners are removed
      if (any(basename(test_files_clean) %in% c("tinytest.R",
                                                "runTinyTests.R",
                                                "run_tinytest.R"))) {
        stop("Runner basename was NOT removed!")
      }
      
      list(
        res_cov = list(
          name = pkg_name,
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 77
          )
        ),
        total_cov           = 77,
        functions_no_tests  = no_tests_df,
        tests_skipped       = character(0),
        tests_passing       = c(exp_a, exp_b)
      )
    }
  )
  
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) invisible(NULL))
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Act ----
  test_pkg_data <- tpd(has_tinytest = TRUE)
  result <- run_fn(pkg_source_path, test_pkg_data, cov_env = cov_env)
  
  # ---- Assert ----
  expect_type(result, "list")
  expect_setequal(result$tests_passing, c(exp_a, exp_b))
  expect_length(result$tests_skipped, 0L)
  expect_equal(result$total_cov, 77)
  expect_equal(result$res_cov$coverage$totalcoverage, 77)
})


# A fake 'covr' namespace for create_base_tests_coverage() tests.
# Uses a shared environment (flags_env) to flip flags by reference so tests can assert calls.
make_fake_covr_ns_base <- function(flags_env, counters_value, coverage_value) {
  stopifnot(is.environment(flags_env))
  covr_ns <- new.env(parent = emptyenv())

  assign("trace_environment", function(env) {
    flags_env$trace <- TRUE
    invisible(NULL)
  }, covr_ns)

  assign("reset_traces", function() {
    flags_env$reset <- TRUE
    invisible(NULL)
  }, covr_ns)

  assign("clear_counters", function() {
    flags_env$clear <- TRUE
    invisible(NULL)
  }, covr_ns)

  assign("as_coverage", function(x) coverage_value, covr_ns)
  assign(".counters", counters_value, covr_ns)

  covr_ns
}


# Helper: a normalizePath stub that doesn’t error with mustWork=TRUE
normalizePath_identity_base <- function(path, ..., winslash = "/", mustWork = FALSE) path


# Helper: extract argument from ... in sys.source/source stubs
get_arg <- function(..., name, default = NULL) {
  a <- list(...)
  if (!is.null(a[[name]])) return(a[[name]])
  default
}

test_that("errors if package namespace is not loaded", {
  pkg_source_path <- "/pkg"
  pkg_name <- "mypkg"
  test_files <- c("/pkg/tests/a.R", "/pkg/tests/b.R")

  mockery::stub(create_base_tests_coverage, "loadedNamespaces", function() c("covr"))

  expect_error(
    create_base_tests_coverage(pkg_source_path, pkg_name, test_files),
    regexp = "Namespace 'mypkg' is not loaded",
    fixed  = FALSE
  )
})

test_that("traces ns, sources each file, returns coverage (happy path)", {
  pkg_source_path <- "/pkg"
  pkg_name <- "mypkg"
  test_files <- c("/pkg/tests/a.R", "/pkg/tests/b.R")

  # Flags that flip when covr hooks are called
  flags <- new.env(parent = emptyenv())
  flags$trace <- FALSE; flags$reset <- FALSE; flags$clear <- FALSE

  fake_counters <- list(n = 2)
  fake_coverage <- structure(list(total = 100), class = "coverage")
  fake_covr_ns  <- make_fake_covr_ns_base(flags, fake_counters, fake_coverage)

  # A fake pkg namespace to trace
  fake_pkg_ns <- new.env(parent = emptyenv())

  # ---- Stubs ----
  mockery::stub(create_base_tests_coverage, "loadedNamespaces",
                function() c(pkg_name, "covr"))

  mockery::stub(create_base_tests_coverage, "asNamespace", function(name) {
    if (identical(name, "covr"))   return(fake_covr_ns)
    if (identical(name, pkg_name)) return(fake_pkg_ns)
    stop("unexpected asNamespace(", name, ")")
  })

  mockery::stub(create_base_tests_coverage, "normalizePath",
                function(path, ..., winslash = "/", mustWork = TRUE) path)

  # Robust sys.source stub: handle both named and positional args.
  sourced <- character(0)
  source_stub <- function(...) {
    args <- list(...)
    file  <- if (!is.null(args$file)) args$file else args[[1]]
    envir <- if (!is.null(args$envir)) args$envir else NULL
    chdir <- if (!is.null(args$chdir)) args$chdir else NULL

    expect_true(file %in% test_files)
    expect_true(is.environment(envir))
    expect_identical(parent.env(envir), fake_pkg_ns)
    expect_true(isTRUE(chdir))

    sourced <<- c(sourced, file)
    invisible(NULL)
  }

  # IMPORTANT: stub both unqualified and namespaced calls
  mockery::stub(create_base_tests_coverage, "sys.source",        source_stub)
  mockery::stub(create_base_tests_coverage, "base::sys.source",  source_stub)

  # ---- Run ----
  res <- create_base_tests_coverage(pkg_source_path, pkg_name, test_files)

  # ---- Assertions ----
  expect_type(res, "list")
  expect_true("coverage" %in% names(res))
  expect_identical(res$coverage, fake_coverage)

  # Now both files should have been captured
  expect_setequal(sourced, test_files)

  # covr hooks invoked
  expect_true(isTRUE(flags$trace))
  expect_true(isTRUE(flags$reset))
  expect_true(isTRUE(flags$clear))
})

test_that("continues if a test file errors (try silent), still returns coverage and runs cleanup", {
  pkg_source_path <- "/pkg"
  pkg_name <- "mypkg"
  test_files <- c("/pkg/tests/a.R", "/pkg/tests/b.R")

  flags <- new.env(parent = emptyenv())
  flags$trace <- FALSE; flags$reset <- FALSE; flags$clear <- FALSE

  fake_counters <- list(n = 1)
  fake_coverage <- structure(list(total = 80), class = "coverage")
  fake_covr_ns  <- make_fake_covr_ns_base(flags, fake_counters, fake_coverage)
  fake_pkg_ns   <- new.env(parent = emptyenv())

  mockery::stub(create_base_tests_coverage, "loadedNamespaces",
                function() c(pkg_name, "covr"))

  mockery::stub(create_base_tests_coverage, "asNamespace", function(name) {
    if (identical(name, "covr"))   return(fake_covr_ns)
    if (identical(name, pkg_name)) return(fake_pkg_ns)
    stop("unexpected asNamespace(", name, ")")
  })

  # Must accept winslash/mustWork to be safe
  mockery::stub(create_base_tests_coverage, "normalizePath",
                function(path, ..., winslash = "/", mustWork = TRUE) path)

  sourced <- character(0)

  # Robust stub: handle named or positional 'file', ensure we append the actual path
  source_stub <- function(...) {
    args <- list(...)
    file  <- if (!is.null(args$file)) args$file else args[[1]]
    # First file errors; second succeeds
    if (identical(file, test_files[[1]])) stop("boom from a.R")
    sourced <<- c(sourced, file)
    invisible(NULL)
  }

  # Stub both variants to be safe
  mockery::stub(create_base_tests_coverage, "sys.source",       source_stub)
  mockery::stub(create_base_tests_coverage, "base::sys.source", source_stub)

  res <- create_base_tests_coverage(pkg_source_path, pkg_name, test_files)

  expect_type(res, "list")
  expect_true("coverage" %in% names(res))
  expect_identical(res$coverage, fake_coverage)
  # Only the second file should have been recorded
  expect_identical(sourced, test_files[[2]])

  expect_true(isTRUE(flags$trace))
  expect_true(isTRUE(flags$reset))
  expect_true(isTRUE(flags$clear))
})

test_that("empty test_files still traces/cleans and returns coverage", {
  pkg_source_path <- "/pkg"
  pkg_name <- "mypkg"
  test_files <- character(0)

  flags <- new.env(parent = emptyenv())
  flags$trace <- FALSE; flags$reset <- FALSE; flags$clear <- FALSE

  fake_counters <- list(n = 0)
  fake_coverage <- structure(list(total = 0), class = "coverage")
  fake_covr_ns  <- make_fake_covr_ns_base(flags, fake_counters, fake_coverage)
  fake_pkg_ns   <- new.env(parent = emptyenv())

  mockery::stub(create_base_tests_coverage, "loadedNamespaces",
                function() c(pkg_name, "covr"))

  mockery::stub(create_base_tests_coverage, "asNamespace", function(name) {
    if (identical(name, "covr"))   return(fake_covr_ns)
    if (identical(name, pkg_name)) return(fake_pkg_ns)
    stop("unexpected asNamespace(", name, ")")
  })

  mockery::stub(create_base_tests_coverage, "normalizePath", normalizePath_identity_base )
  mockery::stub(create_base_tests_coverage, "sys.source", function(...) {
    stop("sys.source should not be called for empty test_files")
  })

  res <- create_base_tests_coverage(pkg_source_path, pkg_name, test_files)

  expect_type(res, "list")
  expect_true("coverage" %in% names(res))
  expect_identical(res$coverage, fake_coverage)

  expect_true(isTRUE(flags$trace))
  expect_true(isTRUE(flags$reset))
  expect_true(isTRUE(flags$clear))
})

test_that("restrict_test_paths_to_framework returns single path when framework matches", {
  # Mocked multi-framework path list (as from get_nstf_test_path)
  multi_path_list <- list(
    testthat = "/pkg/tests/testthat",
    testit = "/pkg/tests/testit",
    base_tests_root = "/pkg/tests",
    tinytest = "/pkg/inst/tinytest"
  )
  fn <- test.assessr:::restrict_test_paths_to_framework

  expect_equal(fn(multi_path_list, "testit"), list(testit = "/pkg/tests/testit"))
  expect_equal(fn(multi_path_list, "tests_base"), list(base_tests_root = "/pkg/tests"))
  expect_equal(fn(multi_path_list, "tinytest"), list(tinytest = "/pkg/inst/tinytest"))
})
 
test_that("restrict_test_paths_to_framework returns full list when framework is NULL or empty", {
  multi_path_list <- list(
    testthat = "/pkg/tests/testthat",
    testit = "/pkg/tests/testit"
  )
  fn <- test.assessr:::restrict_test_paths_to_framework

  expect_identical(fn(multi_path_list, NULL), multi_path_list)
  expect_identical(fn(multi_path_list, ""), multi_path_list)
})

test_that("restrict_test_paths_to_framework returns full list when framework is unknown", {
  multi_path_list <- list(testthat = "/pkg/tests/testthat", testit = "/pkg/tests/testit")
  fn <- test.assessr:::restrict_test_paths_to_framework

  expect_identical(fn(multi_path_list, "unknown"), multi_path_list)
  expect_identical(fn(multi_path_list, "runit"), multi_path_list)
})

test_that("restrict_test_paths_to_framework falls back to nonstandard_testit_root when testit missing", {
  # Nonstandard testit layout: test-ci + test-cran, no tests/testit
  # Mocked path list has nonstandard_testit_root but no testit (or testit absent)
  fn <- test.assessr:::restrict_test_paths_to_framework

  # Case: testit key absent, nonstandard_testit_root present
  path_list_no_testit <- list(
    nonstandard_testit_root = "/pkg/tests",
    testthat = "/pkg/tests/testthat"
  )
  expect_equal(
    fn(path_list_no_testit, "testit"),
    list(nonstandard_testit_root = "/pkg/tests")
  )

  # Case: testit explicitly NULL, nonstandard_testit_root present
  path_list_testit_null <- list(
    testit = NULL,
    nonstandard_testit_root = "/pkg/tests"
  )
  expect_equal(
    fn(path_list_testit_null, "testit"),
    list(nonstandard_testit_root = "/pkg/tests")
  )
})
 
test_that("restrict_test_paths_to_framework prefers testit over nonstandard_testit_root when both present", {
  # Standard testit: tests/testit exists, nonstandard also present
  path_list <- list(
    testit = "/pkg/tests/testit",
    nonstandard_testit_root = "/pkg/tests"
  )
  fn <- test.assessr:::restrict_test_paths_to_framework

  expect_equal(fn(path_list, "testit"), list(testit = "/pkg/tests/testit"))
})

test_that("restrict_test_paths_to_framework maps testthat framework", {
  multi_path_list <- list(
    testthat = "/pkg/tests/testthat",
    testit = "/pkg/tests/testit"
  )
  fn <- test.assessr:::restrict_test_paths_to_framework

  expect_equal(fn(multi_path_list, "testthat"), list(testthat = "/pkg/tests/testthat"))
})



test_that("restrict_test_paths_to_framework returns full list when framework path is missing", {
  # testit requested but not present in list
  path_list <- list(testthat = "/pkg/tests/testthat")
  fn <- test.assessr:::restrict_test_paths_to_framework

  expect_identical(fn(path_list, "testit"), path_list)
})


test_that("restrict_test_paths_to_framework works via run_covr_skip_nstf with stubbed get_nstf_test_path", {
  
  run_fn <- getFromNamespace("run_covr_skip_nstf", "test.assessr")
  
  # Prepare a real directory structure
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testit"), recursive = TRUE, showWarnings = FALSE)
  withr::defer(unlink(pkg_source_path, recursive = TRUE, force = TRUE))
  
  # Multi-path returned by get_nstf_test_path()
  multi_path_list <- list(
    testthat = file.path(pkg_source_path, "tests", "testthat"),
    testit   = file.path(pkg_source_path, "tests", "testit")
  )
  
  mapping <- data.frame(
    source_file = "foo.R",
    test_file   = "test-foo.R",
    stringsAsFactors = FALSE
  )
  
  problems <- data.frame(
    test_file  = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  fake_coverage      <- structure(list(), class = "coverage")
  fake_coverage_list <- list(filecoverage = matrix(1), totalcoverage = 50)
  
  # ---- Required Stubs ----
  mockery::stub(run_fn, "get_pkg_name", "mockpkg")
  
  mockery::stub(run_fn, "get_nstf_test_path", function(...) multi_path_list)
  
  mockery::stub(run_fn, "get_source_test_mapping_nstf", function(...) mapping)
  mockery::stub(run_fn, "check_covr_skip_testit", function(...) problems)
  mockery::stub(run_fn, "get_function_no_tests", data.frame())
  
  mockery::stub(run_fn, "normalizePath", function(path, ...) path)
  
  # Loader stubs
  mockery::stub(run_fn, "pkgload::load_all", NULL)
  mockery::stub(run_fn, "asNamespace",       function(...) new.env(parent = emptyenv()))
  mockery::stub(run_fn, "ls",                function(...) character(0))
  mockery::stub(run_fn, "get",               function(obj, envir) obj)
  mockery::stub(run_fn, "load_dependencies_into_env", function(...) TRUE)
  mockery::stub(run_fn, "load_package_datasets",     function(...) character(0))
  mockery::stub(run_fn, "load_base_r_packages_into_env", function(...) TRUE) # Prevent tcltk failures
  
  # Force coverage path
  mockery::stub(run_fn, "covr::environment_coverage", function(...) fake_coverage)
  mockery::stub(run_fn, "covr::coverage_to_list",     function(...) fake_coverage_list)
  
  # ---- CRITICAL FIX: deterministic nstf coverage normalization ----
  mockery::stub(
    run_fn, "create_nstf_covr_list",
    function(coverage, pkg_name, no_tests_df, test_files_clean, skip_tests, problems) {
      # Validate: only testit test path should remain
      expect_true(all(grepl("testit", test_files_clean)))
      list(
        res_cov = list(
          name = "mockpkg",
          coverage = list(
            filecoverage  = matrix(1),
            totalcoverage = 50
          )
        ),
        total_cov = 50,
        functions_no_tests = data.frame(),
        tests_skipped      = character(0),
        tests_passing      = test_files_clean
      )
    }
  )
  
  # cleanup no-op
  mockery::stub(run_fn, "cleanup_and_return_null", function(...) invisible(NULL))
  
  test_pkg_data <- list(has_testit = TRUE)
  cov_env       <- new.env(parent = emptyenv())
  
  # ---- Act ----
  result <- run_fn(
    pkg_source_path,
    test_pkg_data,
    cov_env     = cov_env,
    framework   = "testit"  # restrict to testit
  )
  
  # ---- Assert ----
  expect_type(result, "list")
  expect_true("total_cov" %in% names(result))
  expect_equal(result$total_cov, 50)
})



test_that("create_nstf_covr_list returns structured list on happy path", {
  # Get the function from your package namespace
  fn <- getFromNamespace("create_nstf_covr_list", "test.assessr")

  # --- Inputs ---
  coverage <- list(package = "mockpkg")   # used by res_cov$name
  pkg_name <- "mockpkg"

  no_tests_df      <- data.frame(`function` = c("f1", "f2"), stringsAsFactors = FALSE)
  test_files_clean <- c("tests/testit/test-a.R", "tests/testit/test-b.R")
  skip_tests       <- c("tests/testit/test-skip.R")
  problems         <- data.frame(test_file = "tests/testit/test-a.R",
                                 issue_type = "none",
                                 stringsAsFactors = FALSE)

  # What covr::coverage_to_list should return (happy path)
  covr_list_mock <- list(
    filecoverage   = matrix(1, nrow = 1, dimnames = list("fileA.R", NULL)),
    totalcoverage  = 72.345
  )

  # mock coverage_to_list to return our fake list
  covr_to_list_mock <- function(coverage) covr_list_mock

  # mock compute_total_coverage and record calls
  # We'll assert on the args later.
  compute_total_mock <- mockery::mock(list(
    total_cov          = 88.8,
    percent_cov        = 72.35,
    n_testfiles        = 4L,
    n_no_function_tests= 1L,
    n_skipped_files    = 2L
  ))

  # Stub the symbols as referenced in the function body
  mockery::stub(fn, "covr::coverage_to_list", covr_to_list_mock)
  mockery::stub(fn, "compute_total_coverage", compute_total_mock)

  # Run and assert it emits a friendly top-line coverage message
  expect_message(
    result <- fn(
      coverage         = coverage,
      pkg_name         = pkg_name,
      no_tests_df      = no_tests_df,
      test_files_clean = test_files_clean,
      skip_tests       = skip_tests,
      problems         = problems
    ),
    regexp = "Total Coverage for mockpkg is:",
    fixed  = FALSE
  )

  # --- Assert compute_total_coverage was called with correct arguments ---
  mockery::expect_called(compute_total_mock, 1)
  # Argument 1 should be as.numeric(totalcoverage) from covr_list_mock
  mockery::expect_args(
    compute_total_mock, 1,
    as.numeric(covr_list_mock$totalcoverage),
    no_tests_df,
    test_files_clean,
    skip_tests
  )

  # --- Assert return structure and values ---
  expect_type(result, "list")
  expect_true(all(c("total_cov","res_cov","functions_no_tests",
                    "tests_skipped","tests_passing","tests_problems") %in% names(result)))

  # Values from the mocked compute_total_coverage()
  expect_equal(result$total_cov, 88.8)

  # res_cov fields
  expect_type(result$res_cov, "list")
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$coverage, covr_list_mock)
  expect_true(is.na(result$res_cov$errors))
  expect_true(is.na(result$res_cov$notes))

  # Pass-throughs
  expect_identical(result$functions_no_tests, no_tests_df)
  expect_identical(result$tests_skipped,      skip_tests)
  expect_identical(result$tests_passing,      test_files_clean)
  expect_identical(result$tests_problems,     problems)
})





test_that("create_nstf_covr_list handles coverage_to_list() error branch", {
  
  # ------------------------------------------
  # Arrange mock inputs
  # ------------------------------------------
  fake_coverage <- list(package = "mockpkg")  # must have $package
  class(fake_coverage) <- "coverage"
  
  no_tests_df <- data.frame(
    source_file = "x.R",
    stringsAsFactors = FALSE
  )
  
  test_files_clean <- c("test-x.R")
  skip_tests <- character(0)
  problems <- data.frame(
    test_file = NA_character_,
    issue_type = "No tests skipped",
    stringsAsFactors = FALSE
  )
  
  # ------------------------------------------
  # Stub covr::coverage_to_list to FORCE ERROR
  # ------------------------------------------
  mockery::stub(
    create_nstf_covr_list,
    "covr::coverage_to_list",
    function(...) stop("Synthetic coverage_to_list error")
  )
  
  # ------------------------------------------
  # Stub compute_total_coverage so test runs to completion
  # Returning predictable output
  # ------------------------------------------
  mockery::stub(
    create_nstf_covr_list,
    "compute_total_coverage",
    list(
      total_cov = 0,
      percent_cov = 0,
      n_testfiles = length(test_files_clean),
      n_no_function_tests = nrow(no_tests_df),
      n_skipped_files = length(skip_tests)
    )
  )
  
  # ------------------------------------------
  # Act: run the function under test
  # ------------------------------------------
  result <- create_nstf_covr_list(
    coverage        = fake_coverage,
    pkg_name        = "mockpkg",
    no_tests_df     = no_tests_df,
    test_files_clean = test_files_clean,
    skip_tests      = skip_tests,
    problems        = problems
  )
  
  # ------------------------------------------
  # Assert: verify that error branch executed
  # ------------------------------------------
  
  # Top level structure
  expect_type(result, "list")
  expect_true("res_cov" %in% names(result))
  expect_true("total_cov" %in% names(result))
  
  # Check error-handling coverage list
  expect_equal(result$res_cov$name, "mockpkg")
  
  # coverage_list should contain NA values (error block)
  expect_true(is.na(result$res_cov$coverage$filecoverage))
  expect_true(is.na(result$res_cov$coverage$totalcoverage))
  
  # errors must contain the thrown condition
  expect_true(inherits(result$res_cov$errors, "error"))
  expect_match(conditionMessage(result$res_cov$errors),
               "Synthetic coverage_to_list error")
  
  # notes = NA (as in error branch)
  expect_true(is.na(result$res_cov$notes))
})

#' Run coverage with skip detection for RUnit and Bioconductor RUnit-based packages
#'
#' `run_covr_skip_runit_nstf()` orchestrates coverage for packages using the
#' Bioconductor **RUnit** layout, such as **BiocGenerics** and packages that store
#' tests in `inst/unitTests` or ship a `tests/run_unitTests.R` wrapper.
#'
#' The function:
#' 1. Builds a source-to-test mapping (using your existing `get_source_test_mapping_nstf()`).
#' 2. Removes any `run_unitTests.R` wrapper from the mapping (via `remove_run_unitTests_mapping()`).
#' 3. Prepares a development environment **before** running RUnit tests:
#'    - Detaches commonly attached BioC infrastructure packages (`IRanges`, `S4Vectors`, `BiocGenerics`)
#'      to avoid namespace/shim conflicts.
#'    - Calls `pkgload::load_all()` to load the package under test.
#'    - Clears `R_TESTS` to avoid R CMD check harness interference when running RUnit programmatically.
#' 4. Identifies problem/skip-able tests via your RUnit-aware `check_covr_skip_nstf()` (which should use
#'    `RUnit::runTestFile()` + `RUnit::getErrors()` per file).
#' 5. Builds a list of test files **under `inst/unitTests`** excluding any that should be skipped.
#' 6. Runs coverage with `covr::environment_coverage()`: tries `tests/doRUnit.R` first when
#'    present (matches R CMD check for CRAN RUnit packages); falls back to a custom runner
#'    that calls `RUnit::runTestFile()` on each test file if doRUnit.R fails or is absent.
#'
#' @param pkg_source_path Character scalar. Path to the root of the package source.
#' @param test_pkg_data A list or environment with test metadata (e.g., fields like
#'   `bioc_run_ut_path` (path to `tests/run_unitTests.R`, if present),
#'   `has_BioG_test` (logical flag), etc.).
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run.   
#' @param framework Optional character. When provided (e.g. from
#'   \code{run_one_framework}), restricts test paths to the framework-specific
#'   directory. Use when a package has multiple frameworks (e.g. testthat + testit)
#'   to avoid mixing paths across frameworks.
#'
#' @return A `covr` coverage object on success; `NULL` on failure (via `cleanup_and_return_null()`).
#'
#' @details
#' Use this function **only** for packages that follow the Bioconductor RUnit layout:
#' - tests are located in `inst/unitTests/`, and/or
#' - a `tests/run_unitTests.R` wrapper exists that calls an internal `.test()` / `testPackage()` runner.
#'
#' @return A covr coverage object, or NULL on failure.
#' 
#' @keywords internal
run_covr_skip_runit_nstf <- function(pkg_source_path, 
                                     test_pkg_data, 
                                     cov_env,
                                     framework = NULL
                                     ) {
  
  cov_env  <- new.env(parent = globalenv())
  pkg_name <- get_pkg_name(pkg_source_path)
  
  # Ensure this really looks like a BioC/RUnit package (caller can decide too)
  # Ensure this really looks like an RUnit package (BioC or RUnit-only; caller can decide too)
  if (!is_runit_layout(pkg_source_path, test_pkg_data)) {
    message("Package does not appear to use the RUnit layout (inst/unitTests); ",
            "consider calling run_covr_skip_nstf() instead.")
  }
  
  # Assemble candidate test roots and normalize (reuses your existing helper)
  path    <- as.character(pkg_source_path)
  testdir <- file.path(path, "inst/unitTests")
  test_path_list <- get_runit_test_paths(test_pkg_data, testdir)
  
  # restrict paths 
  test_path_list <- restrict_test_paths_to_framework(test_path_list, framework)
  
  test_path <- normalizePath(
    unlist(test_path_list, use.names = TRUE),
    winslash = "/",
    mustWork = FALSE
  )
  
  message(paste0("performing source test mapping for ", pkg_name))
  mapping <- tryCatch({
    get_source_test_mapping_runit_nstf(pkg_source_path, test_path)
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error in source-test mapping for ", pkg_name, " : ", e$message))
  })
  if (is.null(mapping)) return(NULL)
  
  # Clean out run_unitTests.R (BioC wrapper) if provided
  if (!is.na(test_pkg_data$bioc_run_ut_path)) {
    message(paste0("performing source test map cleaning for ", pkg_name))
    mapping <- tryCatch({
      remove_run_unitTests_mapping(mapping, test_pkg_data)
    }, error = function(e) {
      cleanup_and_return_null(paste0("Error in cleaning mapping for ", pkg_name, " : ", e$message))
    })
    if (is.null(mapping)) return(NULL)
  }
  
  # --- BioC/RUnit preparation: detach (if attached) and load_all() BEFORE running tests ---
  detach_bioc <- isTRUE(test_pkg_data$has_BioG_test)
  
  if (detach_bioc) {
    message("[BioC RUnit] preparing package before running tests...")
  } else {
    message("[RUnit] preparing package before running tests...")
  }  
  
  prep_ok <- tryCatch({
    prepare_for_bioc_runit(pkg_name, pkg_source_path, detach_bioc_infra = detach_bioc)
    TRUE
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error preloading for BioC RUnit (", pkg_name, "): ", e$message))
  })
  if (is.null(prep_ok)) return(NULL)
  
  # Identify problematic/skip-able tests AFTER load_all()
  message(paste0("identifying problem tests for ", pkg_name))
  problems <- tryCatch({
    # Uses your RUnit-aware checker that calls RUnit::runTestFile() per file
    check_covr_skip_runit_nstf(pkg_source_path, mapping, test_path)
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error identifying skipped/problematic tests for ", pkg_name, " : ", e$message))
  })
  
  if (is.null(problems)) return(NULL)
  
  # Decide which tests to skip
  skip_tests <- problems$test_file[
    !is.na(problems$test_file) &
      !startsWith(problems$issue_type, "warning:") &
      problems$issue_type != "No tests skipped"
  ]
  
  # create df with source files with no tests
  no_tests_df <- get_function_no_tests(mapping) 
  
  # Prepare mapping and build test file list from inst/unitTests
  clean_mapping <- if (anyNA(mapping)) tidyr::drop_na(mapping) else mapping
  root <- file.path(pkg_source_path, "inst", "unitTests")
  test_files_clean <- file.path(root, clean_mapping$test_file)
  if (length(skip_tests) > 0) {
    test_files_clean <- setdiff(test_files_clean, file.path(root, skip_tests))
  }
  
  # Populate the coverage environment with exported & internal objects
  prep_env_ok <- tryCatch({
    pkg_all <- ls(envir = asNamespace(pkg_name), all.names = TRUE)
    for (obj in pkg_all) {
      try(assign(obj, get(obj, envir = asNamespace(pkg_name)), envir = cov_env), silent = TRUE)
    }
    datasets <- load_package_datasets(pkg_name, env = cov_env)
    message("Datasets loaded: ", paste(datasets, collapse = ", "))
    TRUE
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error preparing coverage env for ", pkg_name, " : ", e$message))
  })
  if (is.null(prep_env_ok)) return(NULL)
  
  
  message(paste0("running coverage manually for ", pkg_name))
  
  
  
  coverage <- tryCatch({
    
    # 0) Guard: need at least one test file for custom runner fallback
    if (length(test_files_clean) == 0L) {
      stop("No test files provided in 'test_files_clean'.")
    }
    
    # 1) Ensure 'methods' and 'RUnit' are attached (S4 + checks)
    if (!"methods" %in% loadedNamespaces()) {
      loadNamespace("methods")  # does not attach (no library())
    }
    
    # Unload first to avoid 'package in use' issues
    detach_pkg_if_attached(pkg_name)
    
    # Load the target package from SOURCE so covr instruments the package being assessed
    # (not a different installed version)
    suppressPackageStartupMessages(pkgload::load_all(pkg_source_path, quiet = TRUE))
    
    pkg_ns <- asNamespace(pkg_name)
    
    # 2) Try doRUnit.R first when available (matches R CMD check for CRAN RUnit packages)
    cov <- NULL
    doRUnit_path <- test_pkg_data$doRUnit_path
    use_doRUnit <- !is.null(doRUnit_path) && !is.na(doRUnit_path) && file.exists(doRUnit_path)
    
    if (use_doRUnit) {
      message(paste0("Trying doRUnit.R for ", pkg_name))
      cov <- tryCatch({
        doRUnit_abs <- normalizePath(doRUnit_path, winslash = "/", mustWork = TRUE)
        # doRUnit.R expects wd = tests/ so basename(dirname(wd)) = pkg name (R CMD check layout)
        runner <- build_doRUnit_wrapper(doRUnit_abs)
        covr::environment_coverage(env = pkg_ns, test_files = runner)
      }, error = function(e) {
        message(paste0("doRUnit.R failed: ", conditionMessage(e), "; falling back to custom runner"))
        NULL
      })
      # runTests.R uses defineTestSuite default testFuncRegexp = "^test.+" which does not
      # match runit.*.R files (they use runit.* functions). Retry with runTests-style
      # runner that injects testFuncRegexp = "^runit.+".
      if (!is.null(cov) && length(test_files_clean) > 0L) {
        pct <- tryCatch(covr::percent_coverage(cov), error = function(e) NA_real_)
        if (is.numeric(pct) && !is.na(pct) && pct == 0) {
          message(paste0("doRUnit.R produced 0% coverage (runTests.R uses wrong testFuncRegexp); retrying with runTests-style runner and testFuncRegexp = \"^runit.+\""))
          cov <- NULL
        }
      }
    }
    
    # 3) Fall back if doRUnit wasn't used, failed, or produced 0%
    if (is.null(cov)) {
      tests <- normalizePath(test_files_clean, winslash = "/", mustWork = TRUE)
      # Use runTests-style runner when doRUnit existed (correct testFuncRegexp);
      # otherwise custom runner. Both run only test_files_clean (filtered list).
      if (use_doRUnit) {
        message(paste0("Using runTests-style runner for ", pkg_name, " (", length(tests), " filtered test files)"))
        runner <- build_runit_runTests_runner(pkg_name, tests, pkg_root = pkg_source_path)
      } else {
        message(paste0("Using custom runner for ", pkg_name, " (", length(tests), " test files)"))
        runner <- build_runit_custom_runner(tests, pkg_name = pkg_name, pkg_root = pkg_source_path)
      }
      cov <- covr::environment_coverage(
        env        = pkg_ns,
        test_files = runner
      )
    }
    
    # 4) If environment_coverage produced 0%, try package_coverage as fallback
    #    Uses type = "none" and code = runner so we run only filtered tests (not
    #    the full suite). package_coverage installs to temp and runs our runner
    #    in a subprocess, which can succeed when env instrumentation fails.
    if (!is.null(cov)) {
      pct <- tryCatch(covr::percent_coverage(cov), error = function(e) NA_real_)
      if (is.numeric(pct) && !is.na(pct) && pct == 0) {
        message(paste0("environment_coverage produced 0%; retrying with covr::package_coverage(type = \"none\", code = runner)"))
        detach_pkg_if_attached(pkg_name)
        try(unloadNamespace(pkg_name), silent = TRUE)
        runner_lines <- readLines(runner, warn = FALSE)
        pkg_cov <- tryCatch({
          covr::package_coverage(
            path = pkg_source_path,
            type = "none",
            code = runner_lines,
            quiet = TRUE
          )
        }, error = function(e) {
          message(paste0("package_coverage fallback failed: ", conditionMessage(e)))
          NULL
        })
        if (!is.null(pkg_cov)) {
          pkg_pct <- tryCatch(covr::percent_coverage(pkg_cov), error = function(e) NA_real_)
          if (is.numeric(pkg_pct) && !is.na(pkg_pct) && pkg_pct > 0) {
            cov <- pkg_cov
          }
        }
      }
    }
    
    cov
  }, error = function(e) {
    cleanup_and_return_null(
      paste0("Error running coverage for ", pkg_name, " : ", e$message),
      env = cov_env
    )
  })
  
  if (!is.null(coverage)) {
    
    # create coverage list
    res_cov <- tryCatch({
      coverage_list <- covr::coverage_to_list(coverage)
      list(name = pkg_name, coverage = coverage_list, errors = NA, notes = NA)
    },
    error = function(cond){
      coverage_list <- list(filecoverage = NA, totalcoverage = NA_integer_)
      list(
        name = pkg_name, coverage = coverage_list,
        errors = cond,
        notes = NA
      )
    })
    
    # set up inputs for calculating package coverage
    percent_cov <- as.numeric(res_cov$coverage$totalcoverage)
    
    functions_no_tests_df <- no_tests_df  # data.frame of functions lacking tests
    tests_passing <- test_files_clean               # character vector of passing test files
    tests_skipped <- skip_tests              # character vector of skipped test files
    
    # calculate package coverage
    calc_covr_list <- compute_total_coverage(percent_cov, 
                                             functions_no_tests_df, 
                                             tests_passing,
                                             tests_skipped
    )
    
    
    total_cov <- calc_covr_list$total_cov
    
    percent_cov_round <- round(total_cov, 2)
    message(paste0("Total Coverage for ", pkg_name, " is: ", percent_cov_round, "%"))
    
    percent_cov <- calc_covr_list$percent_cov
    n_functions_tests_passing <- calc_covr_list$n_testfiles
    n_functions_no_tests <- calc_covr_list$n_no_function_tests
    n_functions_tests_skipped <- calc_covr_list$n_skipped_files
    
    message(sprintf(
      "Approx total coverage: (%.2f%%), counts: tested_functions=%d, untested_functions=%d, skipped_tests=%d",
      percent_cov, n_functions_tests_passing, n_functions_no_tests, n_functions_tests_skipped
    ))
    
    cleanup_and_return_null(env = cov_env)
    
    # structure the return list
    covr_list <- list(
      total_cov = total_cov,
      res_cov = list(
        name = res_cov$name,
        coverage = res_cov$coverage,
        errors = res_cov$errors,
        notes = res_cov$notes
      ),
      functions_no_tests = no_tests_df,
      tests_skipped = skip_tests,
      tests_passing = test_files_clean,
      n_functions_tests_passing = n_functions_tests_passing,
      n_functions_no_tests = n_functions_no_tests,
      n_functions_tests_skipped = n_functions_tests_skipped
    )
    
  } else {
    message("No test coverage for this configuration")
    covr_list <- list(
      total_cov = 0,
      res_cov = list(
        name = pkg_name,
        coverage = list(
          filecoverage = matrix(0, nrow = 1, dimnames = list("No functions tested")),
          totalcoverage = 0
        ),
        errors = "No RUnit or BiocGenerics configuration",
        notes = NA
      )
    )
    
  } 
  return(covr_list)
}


#' Detect if a package uses the Bioconductor RUnit layout
#'
#' Heuristics:
#'  - inst/unitTests exists, OR
#'  - tests/run_unitTests.R exists, OR
#'  - explicit flag in test_pkg_data$has_BioG_test
#'
#' @keywords internal
is_bioc_runit_layout <- function(pkg_source_path, test_pkg_data = NULL) {
  is_runit_layout(pkg_source_path, test_pkg_data)
}

#' Detect if a package uses RUnit layout (BioC or RUnit-only)
#'
#' Returns TRUE when the package has RUnit tests in inst/unitTests (with or without
#' the BiocGenerics .test() runner). Heuristics:
#'  - test_pkg_data$has_BioG_test (BioC full layout), OR
#'  - test_pkg_data$has_RUnit_test (inst/unitTests, no .test required), OR
#'  - inst/unitTests exists, OR
#'  - tests/run_unitTests.R exists
#'
#' @param pkg_source_path Character. Path to package source root.
#' @param test_pkg_data List or NULL. Output from check_pkg_tests_and_snaps().
#' @return Logical. TRUE if the package uses RUnit layout.
#' @keywords internal
is_runit_layout <- function(pkg_source_path, test_pkg_data = NULL) {
  inst_unitTests <- file.path(pkg_source_path, "inst", "unitTests")
  run_ut_wrapper <- file.path(pkg_source_path, "tests", "run_unitTests.R")
  
  isTRUE(test_pkg_data$has_BioG_test) ||
    isTRUE(test_pkg_data$has_RUnit_test) ||
    dir.exists(inst_unitTests) ||
    file.exists(run_ut_wrapper)
}

#' Detach a package if attached (internal)
#'
#' Safely detaches a package from the search path and optionally unloads its namespace.
#' @keywords internal
detach_pkg_if_attached <- function(pkg, unload_namespace = TRUE, quiet = TRUE) {
  
  pkg_search_name <- paste0("package:", pkg)
  
  if (pkg_search_name %in% search()) {
    try(detach(pkg_search_name, unload = TRUE, character.only = TRUE), silent = quiet)
    if (!quiet) message("Detached ", pkg_search_name)
  }
  if (isTRUE(unload_namespace) && isNamespaceLoaded(pkg)) {
    try(unloadNamespace(pkg), silent = quiet)
    if (!quiet) message("Unloaded namespace: ", pkg)
  }
  invisible(TRUE)
}

#' Prepare a BioC/RUnit package for testing
#' Detaches common BioC infrastructure packages (if attached) and load_all() the target
#' package before RUnit tests. Also clears R_TESTS to avoid interference from the check harness.
#'
#' @param pkg_name Character. Package name.
#' @param pkg_source_path Character. Path to package source.
#' @param detach_bioc_infra Logical. If TRUE (default for BioC packages), detach
#'   IRanges, S4Vectors, BiocGenerics. Set to FALSE for RUnit-only packages.
#'   
#' @keywords internal
prepare_for_bioc_runit <- function(pkg_name, 
                                   pkg_source_path, 
                                   detach_bioc_infra
                                   ) {
  
  # Detach target if already attached
  detach_pkg_if_attached(pkg_name)
  # Detach common BioC infra (optional; skip for RUnit-only packages)
  if (isTRUE(detach_bioc_infra)) {
    for (p in c("IRanges", "S4Vectors", "BiocGenerics")) {
      detach_pkg_if_attached(p)
    }
  }
  
  # Save current options and set warn = 1 temporarily
  old_opts <- options()
  options(warn = 1)
  on.exit(options(old_opts), add = TRUE)
  
  
  # Load package from source (dev)
  pkgload::load_all(pkg_source_path, quiet = TRUE)
  # Avoid R CMD check harness interference when running RUnit programmatically
  Sys.setenv("R_TESTS" = "")
  invisible(TRUE)
}

#' Check for skipped or errored test files (BiocGenerics)
#'
#' This internal function attempts to run test files mapped to source files and
#' identifies any that produce warnings or errors. It does not handle `testit`.
#'
#' @param pkg_source_path Path to the root of the package source.
#' @param mapping A data frame with a column `test_file` listing test file names.
#' @param test_path directory  
#'
#' @return A data frame listing test files and any issues encountered.
#' @keywords internal
check_covr_skip_runit_nstf <- function(pkg_source_path, mapping, test_path) {
  
  # Guard: RUnit must be available
  if (!requireNamespace("RUnit", quietly = TRUE)) {
    stop("RUnit must be installed to run RUnit tests (install.packages('RUnit')).")
  }
  
  # Some RUnit wrappers advise clearing R_TESTS when running programmatically
  # to avoid R CMD check harness interference. Restore it on exit.
  # See general RUnit runner guidance & practice.  [1](https://www.rdocumentation.org/packages/TriMatch/versions/1.0.0/topics/as.data.frame.list)
  # Save current R_TESTS, set it to ""
  
  # Save current value
  old_R_TESTS <- Sys.getenv("R_TESTS", unset = NA_character_)
  
  is_windows <- tolower(Sys.info()[["sysname"]]) == "windows"
  
  
  if (is_windows) {
    # Windows branch: quoted name in Sys.setenv works best here
    Sys.setenv("R_TESTS" = "")
    on.exit({
      if (is.na(old_R_TESTS)) {
        Sys.unsetenv("R_TESTS")
      } else {
        Sys.setenv("R_TESTS" = old_R_TESTS)
      }
    }, add = TRUE)
  } else {
    # Linux/macOS branch: unquoted name in Sys.setenv works reliably here
    Sys.setenv(R_TESTS = "")
    on.exit({
      if (is.na(old_R_TESTS)) {
        Sys.unsetenv("R_TESTS")
      } else {
        Sys.setenv(R_TESTS = old_R_TESTS)
      }
    }, add = TRUE)
  }
  
  
  test_dir <- test_path
  
  # Use provided mapping; drop rows with NA test_file only
  clean_mapping <-
    if (anyNA(mapping$test_file)) tidyr::drop_na(mapping, test_file) else mapping
  
  # Collector
  problems <- data.frame(test_file = character(),
                         issue_type = character(),
                         stringsAsFactors = FALSE)
  
  run_one_file <- function(test) {
    
    abs <- file.path(test_dir, test)
    
    # Missing file -> flag as 'missing'
    if (!file.exists(abs)) {
      return(data.frame(test_file = test, issue_type = "missing", stringsAsFactors = FALSE))
    }
    
    # load RUnit to ensure all RUnit functions are available for tests
    # library(RUnit)
    
    # Ensure RUnit assertions are visible to legacy tests (temporary attach)
    if (!"package:RUnit" %in% search()) {
      attachNamespace("RUnit")
      on.exit({
        # Detach if we attached it
        if ("package:RUnit" %in% search()) {
          try(detach("package:RUnit", unload = TRUE, character.only = TRUE), silent = TRUE)
        }
      }, add = TRUE)
    }
    
    # Execute the file's tests using RUnit (not source())
    td <- tryCatch(
      {
        RUnit::runTestFile(
          absFileName        = abs,
          useOwnErrorHandler = TRUE,
          testFuncRegexp     = "^test.+"
        )
      },
      error = function(e) {
        # parse/runtime error before RUnit could produce a test data object
        return(structure(list(.error = TRUE), class = "RUnitTestData-ERROR"))
      }
    )
    
    # If we got a hard error, flag it
    if (inherits(td, "RUnitTestData-ERROR")) {
      return(data.frame(test_file = test, issue_type = "error", stringsAsFactors = FALSE))
    }
    
    # Summarize via RUnit API
    # returns counts: nErr, nFail, nDeactivated, etc.  
    # [2](https://bioc.r-universe.dev/BiocGenerics/BiocGenerics.pdf)
    errs <- RUnit::getErrors(td)  
    
    if (!is.null(errs$nErr) && errs$nErr > 0) {
      data.frame(test_file = test, issue_type = "error", stringsAsFactors = FALSE)
    } else if (!is.null(errs$nFail) && errs$nFail > 0) {
      data.frame(test_file = test, issue_type = "failure", stringsAsFactors = FALSE)
    } else if (!is.null(errs$nDeactivated) && errs$nDeactivated > 0) {
      data.frame(test_file = test, issue_type = "skipped", stringsAsFactors = FALSE)
    } else {
      NULL  # clean run for this file
    }
  }
  
  problem_list <- lapply(clean_mapping$test_file, run_one_file)
  problems <- do.call(rbind, problem_list)
  
  if (is.null(problems) || nrow(problems) == 0) {
    problems <- data.frame(
      test_file  = NA_character_,
      issue_type = "No tests skipped",
      stringsAsFactors = FALSE
    )
  }
  
  return(problems)
}

#' Remove mapping rows that point to run_unitTests.R
#'
#' @param mapping A data.frame with columns `source_file` and `test_file`
#'                returned by get_source_test_mapping_nstf().
#' @param test_pkg_data A list or environment containing `bioc_run_ut_path`,
#'                      e.g., "/tmp/.../BiocGenerics/tests/run_unitTests.R".
#'
#' @return The filtered mapping data.frame with run_unitTests.R rows removed.
#' @keywords internal
remove_run_unitTests_mapping <- function(mapping, test_pkg_data) {
  
  # Derive the filename from the provided path, if available.
  run_ut_name <- NA_character_
  if (!missing(test_pkg_data) &&
      !is.null(test_pkg_data$bioc_run_ut_path) &&
      !is.na(test_pkg_data$bioc_run_ut_path)) {
    run_ut_name <- basename(test_pkg_data$bioc_run_ut_path)
  }
  
  # Build a logical vector marking rows to drop
  drop_rows <- rep(FALSE, nrow(mapping))
  
  # 1) Exact filename (case-insensitive) if we have it
  if (!is.na(run_ut_name)) {
    drop_rows <- drop_rows |
      (!is.na(mapping$test_file) &
         tolower(mapping$test_file) == tolower(run_ut_name))
  }
  
  # 2) Fallback regex to be resilient to small variations
  #    Matches run_unitTests.R, run-unitTests.R, RUN_UNITTESTS.R, etc.
  drop_rows <- drop_rows |
    (!is.na(mapping$test_file) &
       grepl("run[_-]?unitTests\\.R$", mapping$test_file, ignore.case = TRUE))
  
  # Return filtered mapping
  mapping <- mapping[!drop_rows, , drop = FALSE]
  
  return(mapping)
}



#' Locate active test paths for Bioconductor RUnit layout
#'
#' Returns a *list* of one or more existing test directories suitable for
#' normalization and use in subsequent mapping (e.g., `get_source_test_mapping_nstf`).
#' Designed to work with:
#'
#'   path    <- as.character(pkg_source_path)
#'   testdir <- file.path(path, "inst/unitTests")
#'   test_path_list <- find_biocgenerics_test_paths(test_pkg_data, testdir)
#'   test_path <- normalizePath(
#'     unlist(test_path_list, use.names = TRUE),
#'     winslash = "/",
#'     mustWork = FALSE
#'   )
#'
#' @param test_pkg_data list/env of test metadata; may include:
#'   - `bioc_run_ut_path` (character): path to `tests/run_unitTests.R` wrapper, if present
#' @param testdir character: expected RUnit tests directory (typically "<pkg_root>/inst/unitTests")
#' @return named list of **existing** directories (e.g., `bioc_unitTests`, `tests_root`);
#'         if nothing is found, returns a single `fallback` entry pointing to `testdir`.
#' @keywords internal
get_runit_test_paths <- function(test_pkg_data, testdir) {
  # Normalize incoming testdir
  
  testdir <- normalizePath(testdir, winslash = "/", mustWork = FALSE)
  
  # Prepare output holder
  out <- list()
  
  # If testdir already exists and is inst/unitTests, prefer it
  if (dir.exists(testdir)) {
    out$bioc_unitTests <- testdir
  } else {
    # Otherwise, try to derive the package root and reconstruct inst/unitTests
    is_pkg_root <- function(dir) file.exists(file.path(dir, "DESCRIPTION"))
    
    # Probe upwards for a package root
    candidates <- unique(c(testdir, dirname(testdir), dirname(dirname(testdir))))
    pkg_root <- NA_character_
    for (cdir in candidates) {
      if (is_pkg_root(cdir)) {
        pkg_root <- normalizePath(cdir, winslash = "/", mustWork = FALSE)
        break
      }
    }
    
    if (!is.na(pkg_root)) {
      bioc_unit <- normalizePath(file.path(pkg_root, "inst", "unitTests"),
                                 winslash = "/", mustWork = FALSE)
      if (dir.exists(bioc_unit)) {
        out$bioc_unitTests <- bioc_unit
      }
    }
    
    # Final fallback if nothing found    # Final fallback if nothing found
    if (length(out) == 0) {
      out$fallback <- testdir
    }
  }
  
  # Single return point
  return(out)
  
}



#' Map Bioconductor and RUnit source files to test files, avoiding false positives.
#' Tightens filename and content-based matching; handles short names like "t".
#'
#' @param pkg_source_path Character; path to the package root (expects R/ under it).
#' @param test_dir        Character; path to tests directory (e.g., file.path(pkg, "tests"))
#' @return data.frame with columns: source_file, test_file, evidence, has_tests
#' @keywords internal
#' @noRd
get_source_test_mapping_runit_nstf <- function(pkg_source_path, test_dir) {
  source_dir <- file.path(pkg_source_path, "R")
  
  src_files  <- list.files(source_dir, pattern = "\\.R$", full.names = TRUE, recursive = FALSE)
  test_files <- list.files(test_dir,   pattern = "\\.R$", full.names = TRUE, recursive = TRUE)
  
  if (length(src_files) == 0) {
    stop("No R files found under: ", source_dir)
  }
  if (length(test_files) == 0) {
    warning("No R test files found under: ", test_dir)
  }
  
  src_basenames  <- basename(src_files)
  test_basenames <- basename(test_files)
  
  # Split a stem into tokens by any non-alphanumeric delimiter (underscore, dash, dot, etc.)
  tokenize_stem <- function(stem) {
    toks <- unlist(strsplit(stem, "[^A-Za-z0-9]+", perl = TRUE))
    toks[nchar(toks) > 0]
  }
  
  # Stricter filename-based match: symbol must equal a token (case-insensitive).
  # Special handling for short symbols (<=2 chars) to avoid matching "test_*" prefix.
  filename_match <- function(src_stem, test_stem) {
    toks <- tolower(tokenize_stem(test_stem))
    s    <- tolower(src_stem)
    if (nchar(s) <= 2) {
      # Only accept if token equals the symbol (e.g., test_t.R or test-t.R)
      return(s %in% toks)
    } else {
      return(s %in% toks)
    }
  }
  
  # Extract declared symbols from a source file:
  # - S4 generics via setGeneric("name")
  # - function objects via `name <- function(...)`
  extract_symbols <- function(file) {
    lines <- tryCatch(readLines(file, warn = FALSE), error = function(e) character())
    generics <- character()
    fun_defs <- character()
    
    if (length(lines)) {
      gen_lines <- grep('setGeneric\\("', lines, value = TRUE, fixed = TRUE)
      if (length(gen_lines)) {
        generics <- unique(sub('.*setGeneric\\("([A-Za-z0-9._]+)".*', '\\1', gen_lines))
      }
      fun_lines <- grep('^\\s*[A-Za-z0-9._]+\\s*<-\\s*function\\s*\\(', lines, value = TRUE, perl = TRUE)
      if (length(fun_lines)) {
        fun_defs <- unique(sub('^\\s*([A-Za-z0-9._]+)\\s*<-\\s*function\\s*\\(.*', '\\1', fun_lines, perl = TRUE))
      }
    }
    unique(c(generics, fun_defs))
  }
  
  # Determine whether a test file is *primarily* about a given symbol.
  # Evidence sources:
  #  - Filename tokens contain the symbol
  #  - test_that("...sym...") description contains symbol as a whole word
  #  - RUnit-style test function name contains symbol (e.g., test_order <- function() {})
  # We deliberately IGNORE mere calls inside assertions that use the symbol as a helper.
  test_targets_symbol <- function(tf, sym) {
    lines <- tryCatch(readLines(tf, warn = FALSE), error = function(e) character())
    if (!length(lines)) return(FALSE)
    
    s <- tolower(sym)
    
    # 1) Filename tokens
    stem <- tools::file_path_sans_ext(basename(tf))
    toks <- tolower(tokenize_stem(stem))
    if (nchar(s) <= 2) {
      if (s %in% toks) return(TRUE)
    } else {
      if (s %in% toks) return(TRUE)
    }
    
    # 2) test_that descriptions (testthat)
    #    Match test_that("...") or test_that('...')
    desc_lines <- grep('test_that\\s*\\(\\s*["\']', lines, value = TRUE, perl = TRUE)
    if (length(desc_lines)) {
      # Pull the first quoted string (description)
      desc <- gsub('^.*test_that\\s*\\(\\s*"\'["\'].*$', '\\1', desc_lines, perl = TRUE)
      has_word <- any(grepl(paste0("\\b", s, "\\b"), tolower(desc), perl = TRUE))
      if (has_word) return(TRUE)
    }
    
    # 3) RUnit style test function names
    #    Looks like: test_<name> <- function() { ... }
    runit_defs <- grep('^\\s*test_[A-Za-z0-9._-]+\\s*<-\\s*function\\s*\\(', lines, value = TRUE, perl = TRUE)
    if (length(runit_defs)) {
      names <- gsub('^\\s*(test_[A-Za-z0-9._-]+)\\s*<-\\s*function\\s*\\(.*$', '\\1', runit_defs, perl = TRUE)
      name_toks <- tolower(unlist(strsplit(names, "[^A-Za-z0-9]+", perl = TRUE)))
      if (s %in% name_toks) return(TRUE)
    }
    
    # Do NOT treat incidental usage, setGeneric/setMethod inside tests, etc., as primary evidence.
    return(FALSE)
  }
  
  out_rows <- lapply(seq_along(src_files), function(i) {
    src_full <- src_files[i]
    src_base <- src_basenames[i]
    src_stem <- tools::file_path_sans_ext(src_base)
    
    # --- Pass 1: strict filename matching ---
    fname_idx <- which(vapply(test_basenames, function(tb) {
      filename_match(src_stem, tools::file_path_sans_ext(tb))
    }, logical(1)))
    
    test_found <- if (length(fname_idx) > 0) test_basenames[fname_idx[1]] else NA_character_
    evidence   <- if (!is.na(test_found)) "filename_match" else NA_character_
    
    # --- Pass 2: content-based primary-target checks ---
    # Only if filename matching failed
    if (is.na(test_found) && length(test_files) > 0) {
      # Prefer S4 generics / defined function symbols
      symbols <- extract_symbols(src_full)
      if (!length(symbols)) symbols <- src_stem
      
      # Try each symbol until we find a test file that targets it
      for (sym in symbols) {
        tf_idx <- which(vapply(test_files, function(tf) test_targets_symbol(tf, sym), logical(1)))
        if (length(tf_idx) > 0) {
          test_found <- test_basenames[tf_idx[1]]
          evidence   <- "content_match"
          break
        }
      }
    }
    
    data.frame(
      source_file = src_base,
      test_file   = ifelse(is.na(test_found), NA_character_, test_found),
      evidence    = ifelse(is.na(test_found), NA_character_, evidence),
      has_tests   = !is.na(test_found),
      stringsAsFactors = FALSE
    )
  })
  
  
  mapping <- do.call(rbind, out_rows)
  rownames(mapping) <- NULL
  return(mapping)
}

#' Detect RUnit test function prefix from test files (test., runit., or Test.)
#'
#' Scans test file content and filenames for function definitions or naming patterns
#' to avoid regex with parentheses that can cause parsing errors when written to
#' runner scripts. Matches patterns like \code{test.plot <-} (function may be on next line).
#'
#' @param test_files Character vector of paths to RUnit test files.
#' @return Character. One of "test", "runit", "Test"; defaults to "test".
#' @keywords internal
#' @noRd
detect_runit_test_prefix <- function(test_files) {
  
  prefix_from_filename <- function(f) {
    base <- basename(f)
    
    if (grepl("^runit\\.[a-zA-Z0-9_.]+\\.R$", base, ignore.case = TRUE)) return("runit")
    if (grepl("^Test\\.[a-zA-Z0-9_.]+\\.R$", base)) return("Test")
    if (grepl("^test_[a-zA-Z0-9_.]+\\.R$", base, ignore.case = TRUE)) return("test")
    if (grepl("^test\\.[a-zA-Z0-9_.]+\\.R$", base, ignore.case = TRUE)) return("test")
    NA_character_
  }
  found <- lapply(test_files, function(f) {
    
    p <- prefix_from_filename(f)
    if (!is.na(p)) return(p)
    NA_character_
  })
  for (p in found) {
    if (!is.null(p) && !is.na(p)) return(p)
  }
  "test"
}

#' Build a wrapper script for doRUnit.R that sets the working directory
#'
#' Many doRUnit.R scripts (e.g. fBasics) derive the package name from
#' \code{basename(dirname(getwd()))}, expecting to be run from \code{tests/}
#' (R CMD check layout). This wrapper sets \code{wd} to the tests directory
#' before sourcing doRUnit.R.
#'
#' @param doRUnit_path Character. Absolute path to \code{tests/doRUnit.R}.
#'
#' @return Character. Path to the temporary wrapper script.
#'
#' @keywords internal
#' @noRd
build_doRUnit_wrapper <- function(doRUnit_path) {
  
  runner <- tempfile(fileext = ".R")
  lines <- c(
    "owd <- getwd()",
    "on.exit(setwd(owd), add = TRUE)",
    sprintf('setwd("%s")', dirname(doRUnit_path)),
    sprintf('source("%s")', doRUnit_path)
  )
  writeLines(lines, runner)
  
  return(runner)
}

#' Build a runTests-style runner with explicit testFuncRegexp
#'
#' Replicates the runTests.R flow (defineTestSuite + runTestSuite) but injects
#' \code{testFuncRegexp} so runit.*.R files are executed. Package runTests.R
#' uses RUnit default \code{testFuncRegexp = "^test.+"} which does not match
#' \code{runit.*} functions.
#'
#' @param pkg_name Character. Package name (e.g. "fBasics").
#' @param test_files Character vector of test file paths (used to detect prefix).
#' @param pkg_root Character source of package
#' @return Character. Path to the temporary runner script.
#'
#' @keywords internal
#' @noRd
build_runit_runTests_runner <- function(pkg_name, test_files, pkg_root = NULL) {
  
  prefix <- detect_runit_test_prefix(test_files)
  # runit.*.R files may use runit.* or test.* functions; match both for robustness
  # Use [.] for literal dot to avoid R "unrecognized escape" when sourcing generated script
  testFuncRegexp <- switch(prefix,
                           runit = '"^runit[.]|^test.+"',
                           Test  = '"^Test[.]|^test.+"',
                           '"^test.+"'
  )
  paths <- normalizePath(test_files, winslash = "/", mustWork = TRUE)
  run_test_line <- vapply(paths, function(p) {
    sprintf('RUnit::runTestFile("%s", useOwnErrorHandler = TRUE, testFuncRegexp = %s)', p, testFuncRegexp)
  }, character(1L))
  runner <- tempfile(fileext = ".R")
  # Set wd to package root so tests can find data files (inst/unitTests/data, etc.)
  wd_lines <- if (!is.null(pkg_root) && dir.exists(pkg_root)) {
    c(
      "owd <- getwd()",
      "on.exit(setwd(owd), add = TRUE)",
      sprintf('setwd("%s")', normalizePath(pkg_root, winslash = "/", mustWork = TRUE))
    )
  } else character(0L)
  lines <- c(
    "if (!requireNamespace('RUnit', quietly = TRUE)) stop('RUnit is required')",
    "if (!'package:RUnit' %in% search()) attachNamespace('RUnit')",
    sprintf('library(package = "%s", character.only = TRUE)', pkg_name),
    wd_lines,
    run_test_line
  )
  writeLines(lines, runner)
  
  return(runner)
}

#' Build a custom RUnit runner script
#'
#' Creates a temporary R script that loads RUnit and runs each test file via
#' \code{RUnit::runTestFile()}. The test function regex is chosen based on the
#' detected prefix (runit, Test, or test) from the test filenames.
#'
#' @param pkg_name Character. Package name (e.g. "fBasics").
#' @param tests Character vector of absolute paths to RUnit test files
#'   (e.g. under \code{inst/unitTests}).
#' @param pkg_root Character source of pack
#'
#' @return Character. Path to the temporary runner script (suitable for
#'   \code{covr::environment_coverage(test_files = ...)}).
#'
#' @keywords internal
#' @noRd
build_runit_custom_runner <- function(tests, pkg_name = NULL, pkg_root = NULL) {
  runner <- tempfile(fileext = ".R")
  prefix <- detect_runit_test_prefix(tests)
  paths <- normalizePath(tests, winslash = "/", mustWork = TRUE)
  # runit.*.R files may use runit.* or test.* functions; match both (same as build_runit_runTests_runner)
  testFuncRegexp <- switch(prefix,
                           runit = '"^runit[.]|^test.+"',
                           Test  = '"^Test[.]|^test.+"',
                           '"^test.+"'
  )
  run_test_line <- vapply(paths, function(p) {
    sprintf('RUnit::runTestFile("%s", useOwnErrorHandler = TRUE, testFuncRegexp = %s)', p, testFuncRegexp)
  }, character(1L))
  wd_lines <- if (!is.null(pkg_root) && dir.exists(pkg_root)) {
    c(
      "owd <- getwd()",
      "on.exit(setwd(owd), add = TRUE)",
      sprintf('setwd("%s")', normalizePath(pkg_root, winslash = "/", mustWork = TRUE))
    )
  } else character(0L)
  lib_line <- if (!is.null(pkg_name) && nzchar(pkg_name)) {
    sprintf('library(package = "%s", character.only = TRUE)', pkg_name)
  } else character(0L)
  runner_lines <- c(
    "if (!requireNamespace('RUnit', quietly=TRUE)) stop('RUnit is required')",
    "if (!'package:RUnit' %in% search()) attachNamespace('RUnit')",
    lib_line,
    wd_lines,
    run_test_line
  )
  writeLines(runner_lines, runner)
  return(runner)
}  

#' Run coverage with skip detection for Bioconductor RUnit-based packages
#'
#' `run_covr_skip_bioc_nstf()` orchestrates coverage for packages using the
#' Bioconductor **RUnit** layout, such as **BiocGenerics** and packages that store
#' tests in `inst/unitTests` or ship a `tests/run_unitTests.R` wrapper.
#'
#' The function:
#' 1. Builds a source-to-test mapping (using your existing `get_source_test_mapping_nstf()`).
#' 2. Removes any `run_unitTests.R` wrapper from the mapping (via `remove_run_unitTests_mapping()`).
#' 3. Prepares a development environment **before** running RUnit tests:
#'    - Detaches commonly attached BioC infrastructure packages (`IRanges`, `S4Vectors`, `BiocGenerics`)
#'      to avoid namespace/shim conflicts.
#'    - Calls `pkgload::load_all()` to load the package under test.
#'    - Clears `R_TESTS` to avoid R CMD check harness interference when running RUnit programmatically.
#' 4. Identifies problem/skip-able tests via your RUnit-aware `check_covr_skip_nstf()` (which should use
#'    `RUnit::runTestFile()` + `RUnit::getErrors()` per file).
#' 5. Builds a list of test files **under `inst/unitTests`** excluding any that should be skipped.
#' 6. Exposes namespace objects + datasets to a dedicated environment and runs coverage with
#'    `covr::environment_coverage()`.
#'
#' @param pkg_source_path Character scalar. Path to the root of the package source.
#' @param test_pkg_data A list or environment with test metadata (e.g., fields like
#'   `bioc_run_ut_path` (path to `tests/run_unitTests.R`, if present),
#'   `has_BioG_test` (logical flag), etc.).
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run.
#' @param framework Optional character. When provided (e.g. from
#'   \code{run_one_framework}), restricts test paths to the framework-specific
#'   directory. Use when a package has multiple frameworks (e.g. testthat + testit)
#'   to avoid mixing paths across frameworks.      
#'
#' @return A `covr` coverage object on success; `NULL` on failure (via `cleanup_and_return_null()`).
#'
#' @details
#' Use this function **only** for packages that follow the Bioconductor RUnit layout:
#' - tests are located in `inst/unitTests/`, and/or
#' - a `tests/run_unitTests.R` wrapper exists that calls an internal `.test()` / `testPackage()` runner.
#'
#' @return A covr coverage object, or NULL on failure.
#' 
#' @keywords internal
run_covr_skip_bioc_nstf <- function(pkg_source_path, test_pkg_data, 
                                    cov_env, 
                                    framework = NULL) {
  
  cov_env  <- new.env(parent = globalenv())
  pkg_name <- get_pkg_name(pkg_source_path)
  
  # Ensure this really looks like a BioC/RUnit package (caller can decide too)
  if (!is_bioc_runit_layout(pkg_source_path, test_pkg_data)) {
    message("Package does not appear to use the BioC RUnit layout; ",
            "consider calling run_covr_skip_nstf() instead.")
  }
  
  # Assemble candidate test roots and normalize (reuses your existing helper)
  path    <- as.character(pkg_source_path)
  testdir <- file.path(path, "inst/unitTests")
  test_path_list <- get_biocgenerics_test_paths(test_pkg_data, testdir)
  test_path <- normalizePath(
    unlist(test_path_list, use.names = TRUE),
    winslash = "/",
    mustWork = FALSE
  )
  
  message(paste0("performing source test mapping for ", pkg_name))
  mapping <- tryCatch({
    get_source_test_mapping_bioc_nstf(pkg_source_path, test_path)
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error in source-test mapping for ", pkg_name, " : ", e$message))
  })
  if (is.null(mapping)) return(NULL)
  
  # Clean out run_unitTests.R (BioC wrapper) if provided
  if (!is.na(test_pkg_data$bioc_run_ut_path)) {
    message(paste0("performing source test map cleaning for ", pkg_name))
    mapping <- tryCatch({
      remove_run_unitTests_mapping(mapping, test_pkg_data)
    }, error = function(e) {
      cleanup_and_return_null(paste0("Error in cleaning mapping for ", pkg_name, " : ", e$message))
    })
    if (is.null(mapping)) return(NULL)
  }
  
  # --- BioC/RUnit preparation: detach (if attached) and load_all() BEFORE running tests ---
  detach_bioc <- isTRUE(test_pkg_data$has_BioG_test)
  
  if (detach_bioc) {
    message("[BioC RUnit] preparing package before running tests...")
  } else {
    message("[RUnit] preparing package before running tests...")
  }  
  
  prep_ok <- tryCatch({
    prepare_for_bioc_runit(pkg_name, pkg_source_path, detach_bioc_infra = detach_bioc)
    TRUE
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error preloading for BioC RUnit (", pkg_name, "): ", e$message))
  })
  if (is.null(prep_ok)) return(NULL)
  
  # Identify problematic/skip-able tests AFTER load_all()
  message(paste0("identifying problem tests for ", pkg_name))
  problems <- tryCatch({
    # Uses your RUnit-aware checker that calls RUnit::runTestFile() per file
    check_covr_skip_bioc_nstf(pkg_source_path, mapping, test_path)
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error identifying skipped/problematic tests for ", pkg_name, " : ", e$message))
  })
  
  if (is.null(problems)) return(NULL)
  
  # Decide which tests to skip
  skip_tests <- problems$test_file[
    !is.na(problems$test_file) &
      !startsWith(problems$issue_type, "warning:") &
      problems$issue_type != "No tests skipped"
  ]
  
  # create df with source files with no tests
  no_tests_df <- get_function_no_tests(mapping) 
  
  # Prepare mapping and build test file list from inst/unitTests
  clean_mapping <- if (anyNA(mapping)) tidyr::drop_na(mapping) else mapping
  root <- file.path(pkg_source_path, "inst", "unitTests")
  test_files_clean <- file.path(root, clean_mapping$test_file)
  if (length(skip_tests) > 0) {
    test_files_clean <- setdiff(test_files_clean, file.path(root, skip_tests))
  }
  
  # Populate the coverage environment with exported & internal objects
  prep_env_ok <- tryCatch({
    pkg_all <- ls(envir = asNamespace(pkg_name), all.names = TRUE)
    for (obj in pkg_all) {
      try(assign(obj, get(obj, envir = asNamespace(pkg_name)), envir = cov_env), silent = TRUE)
    }
    datasets <- load_package_datasets(pkg_name, env = cov_env)
    message("Datasets loaded: ", paste(datasets, collapse = ", "))
    TRUE
  }, error = function(e) {
    cleanup_and_return_null(paste0("Error preparing coverage env for ", pkg_name, " : ", e$message))
  })
  if (is.null(prep_env_ok)) return(NULL)
  
  
  message(paste0("running coverage manually for ", pkg_name))
  
  
  
  coverage <- tryCatch({
    
    # 0) Normalize test paths to forward slashes (Windows-safe)
    if (length(test_files_clean) == 0L) {
      stop("No test files provided in 'test_files_clean'.")
    }
    tests <- normalizePath(test_files_clean, winslash = "/", mustWork = TRUE)
    
    # 1) Ensure 'methods' and 'RUnit' are attached (S4 + checks)
    
    if (!"methods" %in% loadedNamespaces()) {
      loadNamespace("methods")  # does not attach (no library())
    }
    
    #    Unload first to avoid 'package in use' issues
    detach_pkg_if_attached(pkg_name)
    
    # Load the target package so covr can instrument its *namespace*
    suppressPackageStartupMessages(library(pkg_name, character.only = TRUE))
    
    # 3) Build a runner script that actually *executes* each RUnit file
    runner <- tempfile(fileext = ".R")
    runner_lines <- c(
      "if (!requireNamespace('RUnit', quietly=TRUE)) stop('RUnit is required')",
      # Prefer not to attach; but if tests need it:
      "if (!'package:RUnit' %in% search()) attachNamespace('RUnit')",
      sprintf(
        'RUnit::runTestFile("%s", useOwnErrorHandler = TRUE, testFuncRegexp = "^test.+")',
        tests
      )
    )
    writeLines(runner_lines, runner)
    
    # 4) Instrument the *loaded package* environment and run the runner
    #    Use the package environment so S4 generics/classes are intact.
    pkg_env <- as.environment(paste0("package:", pkg_name))
    
    cov <- covr::environment_coverage(
      env        = pkg_env,
      test_files = runner
    )
    
    cov
  }, error = function(e) {
    cleanup_and_return_null(
      paste0("Error running coverage for ", pkg_name, " : ", e$message),
      env = cov_env
    )
  })
  
  if (!is.null(coverage)) {
    
    # create coverage list
    res_cov <- tryCatch({
      coverage_list <- covr::coverage_to_list(coverage)
      list(name = pkg_name, coverage = coverage_list, errors = NA, notes = NA)
    },
    error = function(cond){
      coverage_list <- list(filecoverage = NA, totalcoverage = NA_integer_)
      list(
        name = pkg_name, coverage = coverage_list,
        errors = cond,
        notes = NA
      )
    })
    
    # set up inputs for calculating package coverage
    percent_cov <- as.numeric(res_cov$coverage$totalcoverage)
    
    functions_no_tests_df <- no_tests_df  # data.frame of functions lacking tests
    tests_passing <- test_files_clean               # character vector of passing test files
    tests_skipped <- skip_tests              # character vector of skipped test files
    
    # calculate package coverage
    calc_covr_list <- compute_total_coverage(percent_cov, 
                                             functions_no_tests_df, 
                                             tests_passing,
                                             tests_skipped
    )
    
    
    total_cov <- calc_covr_list$total_cov
    
    percent_cov_round <- round(total_cov, 2)
    message(paste0("Total Coverage for ", pkg_name, " is: ", percent_cov_round, "%"))
    
    percent_cov <- calc_covr_list$percent_cov
    n_functions_tests_passing <- calc_covr_list$n_testfiles
    n_functions_no_tests <- calc_covr_list$n_no_function_tests
    n_functions_tests_skipped <- calc_covr_list$n_skipped_files
    
    message(sprintf(
      "Approx total coverage: (%.2f%%), counts: tested_functions=%d, untested_functions=%d, skipped_tests=%d",
      percent_cov, n_functions_tests_passing, n_functions_no_tests, n_functions_tests_skipped
    ))
    
    cleanup_and_return_null(env = cov_env)
    
    # structure the return list
    covr_list <- list(
      total_cov = total_cov,
      res_cov = list(
        name = res_cov$name,
        coverage = res_cov$coverage,
        errors = res_cov$errors,
        notes = res_cov$notes
      ),
      functions_no_tests = no_tests_df,
      tests_skipped = skip_tests,
      tests_passing = test_files_clean,
      n_functions_tests_passing = n_functions_tests_passing,
      n_functions_no_tests = n_functions_no_tests,
      n_functions_tests_skipped = n_functions_tests_skipped
    )
    
  } else {
    message("No test coverage for this configuration")
    covr_list <- list(
      total_cov = 0,
      res_cov = list(
        name = pkg_name,
        coverage = list(
          filecoverage = matrix(0, nrow = 1, dimnames = list("No functions tested")),
          totalcoverage = 0
        ),
        errors = "No RUnit or BiocGenerics configuration",
        notes = NA
      )
    )
    
  } 
  return(covr_list)
}

#' Check for skipped or errored test files (BiocGenerics)
#'
#' This internal function attempts to run test files mapped to source files and
#' identifies any that produce warnings or errors. It does not handle `testit`.
#'
#' @param pkg_source_path Path to the root of the package source.
#' @param mapping A data frame with a column `test_file` listing test file names.
#' @param test_path directory  
#'
#' @return A data frame listing test files and any issues encountered.
#' @keywords internal
check_covr_skip_bioc_nstf <- function(pkg_source_path, mapping, test_path) {
  
  # Guard: RUnit must be available
  if (!requireNamespace("RUnit", quietly = TRUE)) {
    stop("RUnit must be installed to run RUnit tests (install.packages('RUnit')).")
  }
  
  # Some RUnit wrappers advise clearing R_TESTS when running programmatically
  # to avoid R CMD check harness interference. Restore it on exit.
  # See general RUnit runner guidance & practice.  [1](https://www.rdocumentation.org/packages/TriMatch/versions/1.0.0/topics/as.data.frame.list)
  # Save current R_TESTS, set it to ""
  
  # Save current value
  old_R_TESTS <- Sys.getenv("R_TESTS", unset = NA_character_)
  
  is_windows <- tolower(Sys.info()[["sysname"]]) == "windows"
  
  
  if (is_windows) {
    # Windows branch: quoted name in Sys.setenv works best here
    Sys.setenv("R_TESTS" = "")
    on.exit({
      if (is.na(old_R_TESTS)) {
        Sys.unsetenv("R_TESTS")
      } else {
        Sys.setenv("R_TESTS" = old_R_TESTS)
      }
    }, add = TRUE)
  } else {
    # Linux/macOS branch: unquoted name in Sys.setenv works reliably here
    Sys.setenv(R_TESTS = "")
    on.exit({
      if (is.na(old_R_TESTS)) {
        Sys.unsetenv("R_TESTS")
      } else {
        Sys.setenv(R_TESTS = old_R_TESTS)
      }
    }, add = TRUE)
  }
  
  
  test_dir <- test_path
  
  # Use provided mapping; drop rows with NA test_file only
  clean_mapping <-
    if (anyNA(mapping$test_file)) tidyr::drop_na(mapping, test_file) else mapping
  
  # Collector
  problems <- data.frame(test_file = character(),
                         issue_type = character(),
                         stringsAsFactors = FALSE)
  
  run_one_file <- function(test) {
    
    abs <- file.path(test_dir, test)
    
    # Missing file -> flag as 'missing'
    if (!file.exists(abs)) {
      return(data.frame(test_file = test, issue_type = "missing", stringsAsFactors = FALSE))
    }
    
    # load RUnit to ensure all RUnit functions are available for tests
    # library(RUnit)
    
    # Ensure RUnit assertions are visible to legacy tests (temporary attach)
    if (!"package:RUnit" %in% search()) {
      attachNamespace("RUnit")
      on.exit({
        # Detach if we attached it
        if ("package:RUnit" %in% search()) {
          try(detach("package:RUnit", unload = TRUE, character.only = TRUE), silent = TRUE)
        }
      }, add = TRUE)
    }
    
    # Execute the file's tests using RUnit (not source())
    td <- tryCatch(
      {
        RUnit::runTestFile(
          absFileName        = abs,
          useOwnErrorHandler = TRUE,
          testFuncRegexp     = "^test.+"
        )
      },
      error = function(e) {
        # parse/runtime error before RUnit could produce a test data object
        return(structure(list(.error = TRUE), class = "RUnitTestData-ERROR"))
      }
    )
    
    # If we got a hard error, flag it
    if (inherits(td, "RUnitTestData-ERROR")) {
      return(data.frame(test_file = test, issue_type = "error", stringsAsFactors = FALSE))
    }
    
    # Summarize via RUnit API
    # returns counts: nErr, nFail, nDeactivated, etc.  
    # [2](https://bioc.r-universe.dev/BiocGenerics/BiocGenerics.pdf)
    errs <- RUnit::getErrors(td)  
    
    if (!is.null(errs$nErr) && errs$nErr > 0) {
      data.frame(test_file = test, issue_type = "error", stringsAsFactors = FALSE)
    } else if (!is.null(errs$nFail) && errs$nFail > 0) {
      data.frame(test_file = test, issue_type = "failure", stringsAsFactors = FALSE)
    } else if (!is.null(errs$nDeactivated) && errs$nDeactivated > 0) {
      data.frame(test_file = test, issue_type = "skipped", stringsAsFactors = FALSE)
    } else {
      NULL  # clean run for this file
    }
  }
  
  problem_list <- lapply(clean_mapping$test_file, run_one_file)
  problems <- do.call(rbind, problem_list)
  
  if (is.null(problems) || nrow(problems) == 0) {
    problems <- data.frame(
      test_file  = NA_character_,
      issue_type = "No tests skipped",
      stringsAsFactors = FALSE
    )
  }
  
  return(problems)
}

#' Remove mapping rows that point to run_unitTests.R
#'
#' @param mapping A data.frame with columns `source_file` and `test_file`
#'                returned by get_source_test_mapping_nstf().
#' @param test_pkg_data A list or environment containing `bioc_run_ut_path`,
#'                      e.g., "/tmp/.../BiocGenerics/tests/run_unitTests.R".
#'
#' @return The filtered mapping data.frame with run_unitTests.R rows removed.
#' @keywords internal
remove_run_unitTests_mapping <- function(mapping, test_pkg_data) {
  
  # Derive the filename from the provided path, if available.
  run_ut_name <- NA_character_
  if (!missing(test_pkg_data) &&
      !is.null(test_pkg_data$bioc_run_ut_path) &&
      !is.na(test_pkg_data$bioc_run_ut_path)) {
    run_ut_name <- basename(test_pkg_data$bioc_run_ut_path)
  }
  
  # Build a logical vector marking rows to drop
  drop_rows <- rep(FALSE, nrow(mapping))
  
  # 1) Exact filename (case-insensitive) if we have it
  if (!is.na(run_ut_name)) {
    drop_rows <- drop_rows |
      (!is.na(mapping$test_file) &
         tolower(mapping$test_file) == tolower(run_ut_name))
  }
  
  # 2) Fallback regex to be resilient to small variations
  #    Matches run_unitTests.R, run-unitTests.R, RUN_UNITTESTS.R, etc.
  drop_rows <- drop_rows |
    (!is.na(mapping$test_file) &
       grepl("run[_-]?unitTests\\.R$", mapping$test_file, ignore.case = TRUE))
  
  # Return filtered mapping
  mapping <- mapping[!drop_rows, , drop = FALSE]
  
  return(mapping)
}



#' Locate active test paths for Bioconductor RUnit layout
#'
#' Returns a *list* of one or more existing test directories suitable for
#' normalization and use in subsequent mapping (e.g., `get_source_test_mapping_nstf`).
#' Designed to work with:
#'
#'   path    <- as.character(pkg_source_path)
#'   testdir <- file.path(path, "inst/unitTests")
#'   test_path_list <- find_biocgenerics_test_paths(test_pkg_data, testdir)
#'   test_path <- normalizePath(
#'     unlist(test_path_list, use.names = TRUE),
#'     winslash = "/",
#'     mustWork = FALSE
#'   )
#'
#' @param test_pkg_data list/env of test metadata; may include:
#'   - `bioc_run_ut_path` (character): path to `tests/run_unitTests.R` wrapper, if present
#' @param testdir character: expected RUnit tests directory (typically "<pkg_root>/inst/unitTests")
#' @return named list of **existing** directories (e.g., `bioc_unitTests`, `tests_root`);
#'         if nothing is found, returns a single `fallback` entry pointing to `testdir`.
#' @keywords internal
get_biocgenerics_test_paths <- function(test_pkg_data, testdir) {
  # Normalize incoming testdir
  
  testdir <- normalizePath(testdir, winslash = "/", mustWork = FALSE)
  
  # Prepare output holder
  out <- list()
  
  # If testdir already exists and is inst/unitTests, prefer it
  if (dir.exists(testdir)) {
    out$bioc_unitTests <- testdir
  } else {
    # Otherwise, try to derive the package root and reconstruct inst/unitTests
    is_pkg_root <- function(dir) file.exists(file.path(dir, "DESCRIPTION"))
    
    # Probe upwards for a package root
    candidates <- unique(c(testdir, dirname(testdir), dirname(dirname(testdir))))
    pkg_root <- NA_character_
    for (cdir in candidates) {
      if (is_pkg_root(cdir)) {
        pkg_root <- normalizePath(cdir, winslash = "/", mustWork = FALSE)
        break
      }
    }
    
    if (!is.na(pkg_root)) {
      bioc_unit <- normalizePath(file.path(pkg_root, "inst", "unitTests"),
                                 winslash = "/", mustWork = FALSE)
      if (dir.exists(bioc_unit)) {
        out$bioc_unitTests <- bioc_unit
      }
    }
    
    # Final fallback if nothing found    # Final fallback if nothing found
    if (length(out) == 0) {
      out$fallback <- testdir
    }
  }
  
  # Single return point
  return(out)
  
}



#' Map Bioconductor source files to test files, avoiding false positives.
#' Tightens filename and content-based matching; handles short names like "t".
#'
#' @param pkg_source_path Character; path to the package root (expects R/ under it).
#' @param test_dir        Character; path to tests directory (e.g., file.path(pkg, "tests"))
#' @return data.frame with columns: source_file, test_file, evidence, has_tests
#' @keywords internal
get_source_test_mapping_bioc_nstf <- function(pkg_source_path, test_dir) {
  source_dir <- file.path(pkg_source_path, "R")
  
  src_files  <- list.files(source_dir, pattern = "\\.R$", full.names = TRUE, recursive = FALSE)
  test_files <- list.files(test_dir,   pattern = "\\.R$", full.names = TRUE, recursive = TRUE)
  
  if (length(src_files) == 0) {
    stop("No R files found under: ", source_dir)
  }
  if (length(test_files) == 0) {
    warning("No R test files found under: ", test_dir)
  }
  
  src_basenames  <- basename(src_files)
  test_basenames <- basename(test_files)
  
  # Split a stem into tokens by any non-alphanumeric delimiter (underscore, dash, dot, etc.)
  tokenize_stem <- function(stem) {
    toks <- unlist(strsplit(stem, "[^A-Za-z0-9]+", perl = TRUE))
    toks[nchar(toks) > 0]
  }
  
  # Stricter filename-based match: symbol must equal a token (case-insensitive).
  # Special handling for short symbols (<=2 chars) to avoid matching "test_*" prefix.
  filename_match <- function(src_stem, test_stem) {
    toks <- tolower(tokenize_stem(test_stem))
    s    <- tolower(src_stem)
    if (nchar(s) <= 2) {
      # Only accept if token equals the symbol (e.g., test_t.R or test-t.R)
      return(s %in% toks)
    } else {
      return(s %in% toks)
    }
  }
  
  # Extract declared symbols from a source file:
  # - S4 generics via setGeneric("name")
  # - function objects via `name <- function(...)`
  extract_symbols <- function(file) {
    lines <- tryCatch(readLines(file, warn = FALSE), error = function(e) character())
    generics <- character()
    fun_defs <- character()
    
    if (length(lines)) {
      gen_lines <- grep('setGeneric\\("', lines, value = TRUE, fixed = TRUE)
      if (length(gen_lines)) {
        generics <- unique(sub('.*setGeneric\\("([A-Za-z0-9._]+)".*', '\\1', gen_lines))
      }
      fun_lines <- grep('^\\s*[A-Za-z0-9._]+\\s*<-\\s*function\\s*\\(', lines, value = TRUE, perl = TRUE)
      if (length(fun_lines)) {
        fun_defs <- unique(sub('^\\s*([A-Za-z0-9._]+)\\s*<-\\s*function\\s*\\(.*', '\\1', fun_lines, perl = TRUE))
      }
    }
    unique(c(generics, fun_defs))
  }
  
  # Determine whether a test file is *primarily* about a given symbol.
  # Evidence sources:
  #  - Filename tokens contain the symbol
  #  - test_that("...sym...") description contains symbol as a whole word
  #  - RUnit-style test function name contains symbol (e.g., test_order <- function() {})
  # We deliberately IGNORE mere calls inside assertions that use the symbol as a helper.
  test_targets_symbol <- function(tf, sym) {
    lines <- tryCatch(readLines(tf, warn = FALSE), error = function(e) character())
    if (!length(lines)) return(FALSE)
    
    s <- tolower(sym)
    
    # 1) Filename tokens
    stem <- tools::file_path_sans_ext(basename(tf))
    toks <- tolower(tokenize_stem(stem))
    if (nchar(s) <= 2) {
      if (s %in% toks) return(TRUE)
    } else {
      if (s %in% toks) return(TRUE)
    }
    
    # 2) test_that descriptions (testthat)
    #    Match test_that("...") or test_that('...')
    desc_lines <- grep('test_that\\s*\\(\\s*["\']', lines, value = TRUE, perl = TRUE)
    if (length(desc_lines)) {
      # Pull the first quoted string (description)
      desc <- gsub('^.*test_that\\s*\\(\\s*"\'["\'].*$', '\\1', desc_lines, perl = TRUE)
      has_word <- any(grepl(paste0("\\b", s, "\\b"), tolower(desc), perl = TRUE))
      if (has_word) return(TRUE)
    }
    
    # 3) RUnit style test function names
    #    Looks like: test_<name> <- function() { ... }
    runit_defs <- grep('^\\s*test_[A-Za-z0-9._-]+\\s*<-\\s*function\\s*\\(', lines, value = TRUE, perl = TRUE)
    if (length(runit_defs)) {
      names <- gsub('^\\s*(test_[A-Za-z0-9._-]+)\\s*<-\\s*function\\s*\\(.*$', '\\1', runit_defs, perl = TRUE)
      name_toks <- tolower(unlist(strsplit(names, "[^A-Za-z0-9]+", perl = TRUE)))
      if (s %in% name_toks) return(TRUE)
    }
    
    # Do NOT treat incidental usage, setGeneric/setMethod inside tests, etc., as primary evidence.
    return(FALSE)
  }
  
  out_rows <- lapply(seq_along(src_files), function(i) {
    src_full <- src_files[i]
    src_base <- src_basenames[i]
    src_stem <- tools::file_path_sans_ext(src_base)
    
    # --- Pass 1: strict filename matching ---
    fname_idx <- which(vapply(test_basenames, function(tb) {
      filename_match(src_stem, tools::file_path_sans_ext(tb))
    }, logical(1)))
    
    test_found <- if (length(fname_idx) > 0) test_basenames[fname_idx[1]] else NA_character_
    evidence   <- if (!is.na(test_found)) "filename_match" else NA_character_
    
    # --- Pass 2: content-based primary-target checks ---
    # Only if filename matching failed
    if (is.na(test_found) && length(test_files) > 0) {
      # Prefer S4 generics / defined function symbols
      symbols <- extract_symbols(src_full)
      if (!length(symbols)) symbols <- src_stem
      
      # Try each symbol until we find a test file that targets it
      for (sym in symbols) {
        tf_idx <- which(vapply(test_files, function(tf) test_targets_symbol(tf, sym), logical(1)))
        if (length(tf_idx) > 0) {
          test_found <- test_basenames[tf_idx[1]]
          evidence   <- "content_match"
          break
        }
      }
    }
    
    data.frame(
      source_file = src_base,
      test_file   = ifelse(is.na(test_found), NA_character_, test_found),
      evidence    = ifelse(is.na(test_found), NA_character_, evidence),
      has_tests   = !is.na(test_found),
      stringsAsFactors = FALSE
    )
  })
  
  
  mapping <- do.call(rbind, out_rows)
  rownames(mapping) <- NULL
  return(mapping)
}

globalVariables(c(
                  ".data",
                  "bioc_unit_tests_dir_path",
                  "Code_script",
                  "code_script",
                  "cov_env",
                  "Freq",
                  "Test_Coverage",
                  "Var1",
                  "coverage_percent",
                  "covered",
                  "comments",
                  "dependencies",
                  "description",
                  "descriptions",
                  "documentation",
                  "error",
                  "exported_function",
                  "exported_functions",
                  "filename",
                  "functions",
                  "function_body",
                  "function_calls",
                  "function_type",
                  "has_bug_reports_url",
                  "imp_count",
                  "is_assignment",
                  "is_parameter",
                  "join_key",
                  "link_count",
                  "matched",
                  "name",
                  "overall_risk_score",
                  "package",
                  "package_name",
                  "pkg_name",
                  "pkg_version",
                  "server",
                  "status",
                  "sug_count",
                  "suggested_function",
                  "targeted_package",
                  "test_file",
                  "testthat_summary",
                  "type",
                  "value",
                  "Date", 
                  "Month", 
                  "Nb_of_distinct_IPs",
                  "Nb_of_downloads",
                  "Year",
                  "count",
                  "cumulative_downloads",
                  "day",
                  "desc",
                  "downloads",
                  "month",
                  "n_commits",
                  "week_start",
                  "yarn",
                  "year"
                  ))

#' Null/empty coalescing operator
#'
#' Returns `b` when `a` is `NULL`, length-zero, or `NA`; otherwise returns `a`.
#'
#' @param a Any object.
#' @param b Fallback value.
#'
#' @return Either `a` or `b`.
#'
#' @keywords internal
#' @noRd
`%||%` <- function(a, b) if (is.null(a) || length(a) == 0 || is.na(a)) b else a


#' Safe DESCRIPTION field accessor
#'
#' Fetches a DESCRIPTION field from a one-row `read.dcf()` result, returning
#' `NA_character_` if the field is not present.
#'
#' @param desc A one-row matrix returned by `read.dcf("DESCRIPTION")`.
#' @param field `character(1)`. Field name to retrieve.
#'
#' @return `character(1)` value or `NA_character_` if absent.
#'
#' @keywords internal
#' @noRd
get_field <- function(desc, field) {
  if (!is.null(desc) && field %in% colnames(desc)) desc[1, field] else NA_character_
}


#' Normalize HTML-escaped version operators
#'
#' Converts HTML-escaped comparison operators (`&gt;`, `&lt;`) found in
#' DESCRIPTION constraints into their literal forms (`>`, `<`), trims
#' whitespace, and collapses internal spacing for consistent parsing.
#'
#' Returns `NULL` for empty or `NA` inputs after normalization.
#'
#' @param x `character(1)` or `NULL`. A raw constraint string (e.g., `"&gt;= 1.2.3"`).
#'
#' @return `character(1)` normalized constraint string, or `NULL` if empty/`NA`.
#'
#' @keywords internal
#' @noRd
normalize_constraint <- function(x) {
  if (is.null(x) || is.na(x)) return(NULL)
  x <- trimws(x)
  x <- gsub("&gt;", ">", x, fixed = TRUE)
  x <- gsub("&lt;", "<", x, fixed = TRUE)
  x <- gsub("\\s+", " ", x)
  if (!nzchar(x)) return(NULL)
  x
}


#' Parse a single DESCRIPTION dependency entry
#'
#' Parses entries of the form `"pkg"` or `"pkg (op version)"` from DESCRIPTION
#' fields, where `op` can be one of `>=`, `>`, `<=`, `<`, `==`, or `=` (which is
#' normalized to `==`). HTML-escaped operators (`&gt;`, `&lt;`) are handled.
#'
#' @param dep `character(1)`. A single dependency token from DESCRIPTION, e.g.,
#'   `"pkg"` or `"pkg (>= 1.2.3)"`.
#'
#' @return A `list` with elements:
#'   - `pkg`: `character(1)` package name
#'   - `op`: `character(1)` operator or `NULL`
#'   - `ver`: `character(1)` version string or `NULL`
#'   - `raw`: `character(1)` raw normalized constraint, or `NULL`
#'   Returns `NULL` if parsing fails or token is empty.
#'
#' @keywords internal
#' @noRd
parse_dep <- function(dep) {
  dep <- trimws(dep)
  if (identical(dep, "") || is.na(dep)) return(NULL)
  m <- regmatches(dep, regexec("^([A-Za-z0-9\\.]+)\\s*(?:\\(([^)]+)\\))?$", dep))[[1]]
  if (length(m) >= 2) {
    pkg <- trimws(m[2])
    raw <- if (length(m) >= 3 && !is.na(m[3])) trimws(m[3]) else NULL
    raw <- normalize_constraint(raw)
    op <- ver <- NULL
    if (!is.null(raw)) {
      # accept one of: >=, >, <=, <, ==, = 
      m2 <- regmatches(raw, regexec("^([><=!]=?|==)\\s*(\\S.*)$", raw))[[1]]
      if (length(m2) >= 3) {
        op  <- m2[2]
        ver <- trimws(m2[3])
        if (op == "=") op <- "=="  # normalize
      } else {
        # Unrecognized constraint text; keep raw but unset op/ver
        op <- ver <- NULL
      }
    }
    return(list(pkg = pkg, op = op, ver = ver, raw = raw))
  } else {
    warning("Could not parse dependency: ", dep)
    return(NULL)
  }
}


#' Check a version constraint against an installed version
#'
#' Compares an installed version string against a constraint expressed as an
#' operator and version using `utils::compareVersion()`. If `op` or `ver` is
#' `NULL`, the constraint is treated as satisfied.
#'
#' @param installed_version `character(1)` or `package_version`. The installed version.
#' @param op `character(1)` operator among `>=`, `>`, `<=`, `<`, `==`.
#' @param ver `character(1)` version to compare against.
#'
#' @return `logical(1)`. `TRUE` if satisfied; otherwise `FALSE`.
#'
#' @keywords internal
#' @noRd
constraint_satisfied <- function(installed_version, op, ver) {
  if (is.null(op) || is.null(ver)) return(TRUE)
  cmp <- utils::compareVersion(as.character(installed_version), ver)
  switch(op,
         ">=" = cmp >= 0,
         ">"  = cmp >  0,
         "<=" = cmp <= 0,
         "<"  = cmp <  0,
         "==" = cmp == 0,
         # Unknown operator -> be conservative (not satisfied)
         FALSE
  )
}


#' Install a package if missing or version-mismatched
#'
#' Ensures a package is installed and its namespace can be loaded. If a version
#' constraint is provided and the installed version does not satisfy it,
#' re-installs the package. Uses `install.packages()` for `>=`/`>` constraints
#' (assumes latest satisfies), and `remotes::install_version()` for `<=`, `<`,
#' or `==` constraints to target a specific version.
#'
#' Always attempts to `loadNamespace(pkg)` at the end (suppressed), so the
#' namespace is available even if the package is not attached.
#'
#' @param pkg `character(1)`. Package name.
#' @param op `character(1)` or `NULL`. Constraint operator.
#' @param ver `character(1)` or `NULL`. Constraint version.
#'
#' @return Invisibly returns `TRUE`.
#'
#' @keywords internal
#' @noRd
install_if_needed <- function(pkg, op = NULL, ver = NULL) {
  installed <- requireNamespace(pkg, quietly = TRUE)
  needs_install <- !installed
  
  if (installed && !is.null(op) && !is.null(ver)) {
    inst_ver <- utils::packageDescription(pkg, fields = "Version")
    if (!constraint_satisfied(inst_ver, op, ver)) {
      message("Installed version of ", pkg, " (", inst_ver,
              ") does not satisfy constraint ", op, " ", ver, ". Re-installing.")
      needs_install <- TRUE
    }
  }
  
  if (needs_install) {
    tryCatch({
      if (!is.null(op) && !is.null(ver)) {
        # Strategy:
        # - For >= / >: the latest should satisfy (install.packages)
        # - For <= / < / ==: try exact/targeted via remotes::install_version
        if (op %in% c(">=", ">")) {
          utils::install.packages(pkg)
        } else {
          remotes::install_version(pkg, version = ver, upgrade = "never", force = TRUE)
        }
      } else {
        utils::install.packages(pkg)
      }
    }, error = function(e) {
      message("Failed to install package: ", pkg, " - ", conditionMessage(e))
    })
  }
  
  # Ensure namespace is loaded
  try(suppressWarnings(loadNamespace(pkg)), silent = TRUE)
  invisible(TRUE)
}


#' Import namespace exports into an environment
#'
#' Copies all (or a filtered subset of) exported symbols from a package
#' namespace into a target environment. This enables bare symbol resolution
#' (e.g., `plot`, `lm`) inside environments that do not inherit from
#' `.GlobalEnv` (e.g., when the parent is `baseenv()`), independent of the
#' search path or attachment state.
#'
#' @param pkg `character(1)`. Package name.
#' @param env `environment`. Target environment to receive the exported symbols.
#' @param overwrite `logical(1)`. Overwrite existing bindings in `env`. Default `TRUE`.
#' @param filter `character()` or `NULL`. If provided, only these exported names
#'   are imported (intersection with `getNamespaceExports(pkg)`).
#'
#' @return Invisibly returns `TRUE`.
#'
#' @keywords internal
#' @noRd
import_namespace_exports <- function(pkg, env, overwrite = TRUE, filter = NULL) {
  ns <- asNamespace(pkg)
  exports <- getNamespaceExports(pkg)
  if (!is.null(filter)) {
    exports <- intersect(exports, filter)
  }
  for (nm in exports) {
    if (!overwrite && exists(nm, envir = env, inherits = FALSE)) next
    env[[nm]] <- getExportedValue(pkg, nm)
  }
  invisible(TRUE)
}


#' Resolve dependencies from DESCRIPTION into a data.frame
#'
#' Extracts dependency tokens from DESCRIPTION fields (default: `Depends`,
#' `Imports`, `Suggests`), optionally excluding the `R (>= x.y.z)` entry,
#' parses each token, and returns a structured data frame suitable for further
#' processing (installation, loading, and visibility control).
#'
#' The returned `field` values are trimmed of whitespace to avoid equality
#' issues when selecting rows (e.g., `"Depends "` vs `"Depends"`).
#'
#' @param desc A one-row matrix returned by `read.dcf("DESCRIPTION")`.
#' @param fields `character()`. DESCRIPTION fields to parse.
#' @param exclude_R `logical(1)`. If `TRUE`, drop `R (...)` tokens from results.
#'
#' @return A `data.frame` with columns:
#'   - `package`: package name
#'   - `op`: operator (may be `NA`)
#'   - `ver`: version string (may be `NA`)
#'   - `field`: the DESCRIPTION field source (trimmed)
#'
#' @keywords internal
#' @noRd
resolve_description_deps <- function(desc, fields = c("Depends", "Imports", "Suggests"),
                                     exclude_R = TRUE) {
  vals <- unlist(lapply(fields, function(f) {
    v <- get_field(desc, f)
    if (is.na(v) || !nzchar(v)) return(character(0))
    parts <- unlist(strsplit(v, ",\\s*"))
    parts <- parts[parts != "" & !is.na(parts)]
    if (exclude_R) {
      parts <- parts[!grepl("^R(\\s|$)", parts)]
    }
    # Tag each part with its originating field so we can decide import policy
    paste(f, ":::SEP:::", parts)
  }), use.names = FALSE)
  
  if (length(vals) == 0) {
    # create a an empty data frame that preserves the structure:
    # of column names, column types, zero rows.
    return(data.frame(package=character(), op=character(),
                      ver=character(), field=character()))
  }
  
  res <- lapply(vals, function(x) {
    sp <- strsplit(x, ":::SEP:::", fixed = TRUE)[[1]]
    field <- sp[1]
    dep   <- sp[2]
    parsed <- parse_dep(dep)
    if (is.null(parsed)) return(NULL)
    data.frame(
      package = parsed$pkg,
      op = parsed$op %||% NA_character_,
      ver = parsed$ver %||% NA_character_,
      field = trimws(field),
      stringsAsFactors = FALSE
    )
  })
  
  # Ensure this function always returns a data.frame 
  # (possibly zero-row) even when all tokens are dropped/unparseable.
  res_filtered <- res[!vapply(res, is.null, logical(1))]
  
  if (length(res_filtered) == 0) {
    return(data.frame(
      package = character(),
      op = character(),
      ver = character(),
      field = character(),
      stringsAsFactors = FALSE
    ))
  }
  do.call(rbind, res_filtered)
}


#' Validate the `R (>= x.y.z)` requirement in DESCRIPTION
#'
#' If a Depends entry of the form `R (op version)` exists, normalize the
#' constraint, parse it, and compare against the current interpreter's version
#' (`getRversion()`). Throws an error if the requirement is not satisfied.
#'
#' @param desc A one-row matrix returned by `read.dcf("DESCRIPTION")`.
#'
#' @return Invisibly returns `TRUE` (or throws on failure).
#'
#' @keywords internal
#' @noRd
validate_R_requirement <- function(desc) {
  deps <- get_field(desc, "Depends")
  if (is.na(deps) || !nzchar(deps)) return(invisible(TRUE))
  m <- regmatches(deps, gregexpr("R\\s*\\(([^)]+)\\)", deps))[[1]]
  if (length(m) == 0) return(invisible(TRUE))
  # take first constraint
  cstr <- gsub("^R\\s*\\(|\\)$", "", m[1])
  cstr <- normalize_constraint(cstr)
  if (!length(cstr) || is.null(cstr)) return(invisible(TRUE))
  m2 <- regmatches(cstr, regexec("^([><=!]=?|==)\\s*(\\S.*)$", cstr))[[1]]
  if (length(m2) >= 3) {
    op <- if (m2[2] == "=") "==" else m2[2]
    ver <- trimws(m2[3])
    rver <- getRversion()
    ok <- constraint_satisfied(rver, op, ver)
    if (!ok) {
      stop("This package requires R version ", op, " ", ver,
           " but current R is ", as.character(rver), call. = FALSE)
    }
  }
  invisible(TRUE)
}

#' Load and Install Package Dependencies from a DESCRIPTION File
#'
#' Reads a package's `DESCRIPTION` file, parses dependency fields
#' (`Depends`, `Imports`, `Suggests`), installs missing packages or those that
#' do not satisfy version constraints, loads namespaces, and—for packages listed
#' under **Depends**—imports their exported symbols into a provided evaluation
#' environment. This makes bare function names visible when evaluating code in
#' environments that do not inherit from `.GlobalEnv` (e.g., environments with
#' `baseenv()` as parent).
#'
#' @section Behavior:
#'
#' 1. **Dependency parsing**  
#'    Parses entries of the form `"pkg"` or `"pkg (op version)"`, where the
#'    operator `op` is one of `>=`, `>`, `<=`, `<`, `==`. HTML-escaped forms
#'    (`&gt;`, `&lt;`) are normalized.
#'
#' 2. **Dependency fields**  
#'    Dependencies are extracted from the `Depends`, `Imports`, and `Suggests`
#'    fields. The `Depends: R (>= x.y.z)` entry is ignored for installation
#'    purposes, but the R version constraint is validated.
#'
#' 3. **Base/recommended packages**  
#'    Unlike tools such as `R CMD check`, this function **does not skip**
#'    recommended or base packages (e.g., `graphics`, `stats`, `utils`,
#'    `grDevices`) because packages listed in `Depends:` must be made visible
#'    in the target evaluation environment even when its parent is `baseenv()`.
#'    If a package is already installed with R, no installation is attempted.
#'
#' 4. **Availability and version checks**  
#'    Package presence is determined via `requireNamespace(pkg, quietly = TRUE)`
#'    instead of `installed.packages()`, for performance and portability.  
#'    If a version constraint is declared, the installed version (from
#'    `packageDescription(pkg, fields = "Version")`) is compared using
#'    `utils::compareVersion()`.
#'
#' 5. **Package installation**  
#'    - For constraints of the form `>=` or `>`:  
#'      `install.packages()` is used (latest version is assumed valid).  
#'    - For `<=`, `<`, or `==`:  
#'      `remotes::install_version()` is used to install an exact version.  
#'
#' 6. **Loading behavior**  
#'    Each dependency is namespace-loaded via `loadNamespace()`.  
#'    Packages that appear in the `Depends` field have all their **exported
#'    symbols imported** into the supplied `env`. This makes calls such as
#'    `plot()`, `lm()`, or `head()` resolve correctly inside environments whose
#'    parent is not `.GlobalEnv`.
#'
#' @param pkg_source_path `character(1)`. Path to a package source directory
#'   containing a `DESCRIPTION` file.
#'
#' @param env An environment into which exported symbols from packages listed
#'   in the **Depends** field will be imported. This allows code evaluated in
#'   `env` to resolve functions from those packages using bare names without
#'   requiring attachment on the search path.
#'
#' @return Invisibly returns `TRUE` on success.
#'
#' @details
#' This function avoids calling `installed.packages()` because that function
#' scans all installed packages and is slow on some platforms (especially
#' Windows and network filesystems). Using `requireNamespace()` is the
#' recommended approach for checking whether a package is installed and usable.
#' `packageDescription()` is used only on a small set of packages when version
#' information or priority metadata is needed.
#'
#' When evaluating code in environments that do not inherit from `.GlobalEnv`,
#' attaching packages does not necessarily make their exported symbols available.
#' For this reason, exports from packages listed under `Depends` are explicitly
#' imported into the target environment.
#'
#' @keywords internal
load_dependencies_into_env <- function(pkg_source_path, env) {
  # Ensure 'remotes' exists for versioned installs
  
  if (!requireNamespace("remotes", quietly = TRUE)) {
    utils::install.packages("remotes")
  }
  
  desc_file <- file.path(pkg_source_path, "DESCRIPTION")
  if (!file.exists(desc_file)) {
    stop("DESCRIPTION file not found at: ", desc_file)
  }
  
  desc <- read.dcf(desc_file)
  
  # 1) Validate R version requirement if specified
  validate_R_requirement(desc)
  
  # 2) Resolve dependencies (Depends, Imports, Suggests)
  deps_df <- resolve_description_deps(desc,
                                      fields = c("Depends", "Imports", "Suggests"),
                                      exclude_R = TRUE)
  if (nrow(deps_df) == 0) return(invisible(TRUE))
  
  # 3) Install / load namespaces
  for (i in seq_len(nrow(deps_df))) {
    pkg <- deps_df$package[i]
    op  <- if (!is.na(deps_df$op[i])) deps_df$op[i] else NULL
    ver <- if (!is.na(deps_df$ver[i])) deps_df$ver[i] else NULL
    
    # Do NOT skip base packages (grDevices, graphics, stats, utils, methods)
    # Installation will be skipped automatically if already present.
    install_if_needed(pkg, op, ver)
  }
  
  # 4) Make symbols visible from 'env'
  depends_pkgs <- deps_df$package[deps_df$field %in% c("Depends", "Imports", "Suggests")]
  depends_pkgs <- unique(depends_pkgs)
  
  for (pkg in depends_pkgs) {
    # Ensure namespace is loaded (defensive)
    try(suppressWarnings(loadNamespace(pkg)), silent = TRUE)
    # Import all exports into env so code evaluated there can see them directly
    import_namespace_exports(pkg, env, overwrite = TRUE)
  }
  
  invisible(TRUE)
}

test_that("fails if DESCRIPTION file is missing", {
  env <- new.env()
  mockery::stub(load_dependencies_into_env, "requireNamespace", TRUE)
  mockery::stub(load_dependencies_into_env, "file.exists", FALSE)
  expect_error(load_dependencies_into_env("fakepkg", env), "DESCRIPTION file not found")
})

test_that("parses and installs dependencies", {
  env <- new.env()
  desc <- matrix(
    c("pkgA (>= 1.0.0), pkgB", "pkgC", NA_character_),
    nrow = 1,
    dimnames = list(NULL, c("Depends", "Imports", "Suggests"))
  )
  
  # --- FS & DESCRIPTION stubs ---
  mockery::stub(load_dependencies_into_env, "file.exists", TRUE)
  mockery::stub(load_dependencies_into_env, "read.dcf", desc)
  
  # --- Availability check (requireNamespace) ---
  # Call order expected:
  #   1) requireNamespace("remotes", quietly = TRUE)
  #   2) requireNamespace("pkgA", quietly = TRUE) -> FALSE (needs install)
  #   3) requireNamespace("pkgB", quietly = TRUE) -> TRUE (already installed)
  #   4) requireNamespace("pkgC", quietly = TRUE) -> FALSE (needs install)
  req_ns_mock <- mockery::mock(TRUE, FALSE, TRUE, FALSE)
  mockery::stub(load_dependencies_into_env, "requireNamespace", req_ns_mock)
  
  # --- packageDescription (Version checks) ---
  # Call order expected:
  #   1) packageDescription("pkgA", fields = "Version") -> "0.9.0" (fails >= 1.0.0)
  pkg_desc_mock <- mockery::mock("0.9.0")
  mockery::stub(load_dependencies_into_env, "utils::packageDescription", pkg_desc_mock)
  
  # --- Installers ---
  # install.packages() should be called for pkgA (fails >= 1.0.0) and for pkgC (missing)
  install_mock <- mockery::mock(invisible(NULL), invisible(NULL))
  mockery::stub(load_dependencies_into_env, "utils::install.packages", install_mock)
  
  # --- loadNamespace() ---
  load_ns_mock <- mockery::mock(invisible(NULL), invisible(NULL), invisible(NULL))
  mockery::stub(load_dependencies_into_env, "loadNamespace", load_ns_mock)
  
  # --- import_namespace_exports() ---
  import_mock <- mockery::mock(invisible(TRUE), invisible(TRUE), invisible(TRUE))
  mockery::stub(load_dependencies_into_env, "import_namespace_exports", import_mock)
  
  # --- Run & assertions ---
  suppressWarnings(result <- load_dependencies_into_env("fakepkg", env))
  expect_equal(result, invisible(TRUE))
  
  # import_namespace_exports() called three times: pkgA, pkgB, pkgC
  mockery::expect_called(import_mock, 3)
  mockery::expect_args(import_mock, 1, "pkgA", env, overwrite = TRUE)
  mockery::expect_args(import_mock, 2, "pkgB", env, overwrite = TRUE)
  mockery::expect_args(import_mock, 3, "pkgC", env, overwrite = TRUE)
})


test_that("parses and installs dependencies from Imports/Suggests (no Depends import)", {
  env <- new.env(parent = baseenv())
  
  # DESCRIPTION with only Imports and Suggests
  desc <- matrix(
    c("pkgA (>= 1.0.0), pkgB",  # Imports
      "pkgC",                   # Suggests
      NA_character_),           # Depends (none)
    nrow = 1,
    dimnames = list(NULL, c("Imports", "Suggests", "Depends"))
  )
  
  # --- FS & DESCRIPTION stubs ---
  mockery::stub(load_dependencies_into_env, "file.exists", TRUE)
  mockery::stub(load_dependencies_into_env, "read.dcf", desc)
  
  # --- remotes availability check ---
  # Only called once at the beginning: requireNamespace("remotes", quietly = TRUE)
  req_ns_mock <- mockery::mock(TRUE)
  mockery::stub(load_dependencies_into_env, "requireNamespace", req_ns_mock)
  
  # --- Core seams for the NEW design ---
  # Do NOT reach out to CRAN; we just assert install_if_needed is called correctly
  installer_mock <- mockery::mock(invisible(TRUE), invisible(TRUE), invisible(TRUE))
  mockery::stub(load_dependencies_into_env, "install_if_needed", installer_mock)
  
  # No Depends => we must NOT import exports into env
  import_mock <- mockery::mock()
  mockery::stub(load_dependencies_into_env, "import_namespace_exports", import_mock)
  
  # --- Run ---
  expect_invisible(load_dependencies_into_env("fakepkg", env))
  
  # --- Assertions ---
  # remotes check once
  mockery::expect_called(req_ns_mock, 1)
  
  # install_if_needed called for each token in order: pkgA, pkgB, pkgC
  mockery::expect_called(installer_mock, 3)
  mockery::expect_args(installer_mock, 1, "pkgA", ">=", "1.0.0")
  mockery::expect_args(installer_mock, 2, "pkgB", NULL, NULL)
  mockery::expect_args(installer_mock, 3, "pkgC", NULL, NULL)
  
})


test_that("imports exports from Depends into env", {
  env <- new.env(parent = baseenv())
  
  desc <- matrix(
    c("grDevices, graphics, stats, utils",  # Depends
      NA_character_,                        # Imports
      NA_character_),                       # Suggests
    nrow = 1,
    dimnames = list(NULL, c("Depends", "Imports", "Suggests"))
  )
  
  mockery::stub(load_dependencies_into_env, "file.exists", TRUE)
  mockery::stub(load_dependencies_into_env, "read.dcf", desc)
  
  # Only remotes check at the start
  req_ns_mock <- mockery::mock(TRUE)
  mockery::stub(load_dependencies_into_env, "requireNamespace", req_ns_mock)
  
  # Avoid network/installation; just assert calls happen for each Depends package
  installer_mock <- mockery::mock(invisible(TRUE), invisible(TRUE), invisible(TRUE), invisible(TRUE))
  mockery::stub(load_dependencies_into_env, "install_if_needed", installer_mock)
  
  # Should be called for each Depends pkg with (pkg, env, overwrite=TRUE, filter=NULL)
  import_mock <- mockery::mock(invisible(TRUE), invisible(TRUE), invisible(TRUE), invisible(TRUE))
  mockery::stub(load_dependencies_into_env, "import_namespace_exports", import_mock)
  
  expect_invisible(load_dependencies_into_env("fakepkg", env))
  
  mockery::expect_called(req_ns_mock, 1)
  mockery::expect_called(installer_mock, 4)
  mockery::expect_called(import_mock, 4)
  
})


test_that("resolve_description_deps returns zero-row data.frame when all tokens are unparseable", {
  # dummy desc object – its structure is irrelevant because we stub get_field()
  desc <- new.env(parent = emptyenv())
  
  # 1) Stub get_field() to return non-empty fields so vals > 0
  #    (ensures the code reaches the res/res_filtered path rather than the earlier vals==0 return)
  mockery::stub(
    resolve_description_deps, 
    "get_field",
    function(d, f) {
      switch(f,
             "Depends"  = "A (>= 1.0), B",
             "Imports"  = "C, D (>= 2.0)",
             "Suggests" = "E",
             ""         # default (unused)
      )
    }
  )
  
  # 2) Stub parse_dep() to always return NULL so every token is dropped
  mockery::stub(
    resolve_description_deps,
    "parse_dep",
    function(x) NULL
  )
  
  # Call the function – it should go through res -> res_filtered -> return zero-row df
  out <- resolve_description_deps(desc, fields = c("Depends", "Imports", "Suggests"), exclude_R = TRUE)
  
  # Assertions: structure and zero rows
  expect_true(is.data.frame(out))
  expect_identical(names(out), c("package", "op", "ver", "field"))
  expect_identical(nrow(out), 0L)
  # Ensure character columns (stringsAsFactors = FALSE)
  expect_true(all(vapply(out, typeof, character(1)) == "character"))
})

test_that("resolve_description_deps filters NULL entries and binds remaining rows", {
  # Make sure an infix %||% exists for the function body
  `%||%` <- function(a, b) if (!is.null(a)) a else b
  
  desc <- new.env(parent = emptyenv())
  
  # Return a mix of tokens across fields
  mockery::stub(
    resolve_description_deps,
    "get_field",
    function(d, f) {
      switch(f,
             "Depends"  = "A (>= 1.0), B",
             "Imports"  = "R (>= 4.0), C",  # 'R' should be excluded when exclude_R = TRUE
             "Suggests" = "D, E (>= 2.3)",
             ""
      )
    }
  )
  
  # parse_dep returns:
  #  - a proper parsed list for A, C, E
  #  - NULL for B and D (to test filtering via res_filtered line)
  mockery::stub(
    resolve_description_deps,
    "parse_dep",
    function(token) {
      token <- trimws(token)
      if (startsWith(token, "A")) list(pkg = "A", op = ">=", ver = "1.0")
      else if (token == "B")      NULL
      else if (token == "C")      list(pkg = "C", op = NULL, ver = NULL)  # will use %||% -> NA
      else if (token == "D")      NULL
      else if (startsWith(token, "E")) list(pkg = "E", op = ">=", ver = "2.3")
      else NULL
    }
  )
  
  out <- resolve_description_deps(desc, fields = c("Depends", "Imports", "Suggests"), exclude_R = TRUE)
  
  # Should have rows for A, C, E (B and D dropped; R excluded)
  expect_true(is.data.frame(out))
  expect_identical(names(out), c("package", "op", "ver", "field"))
  expect_setequal(out$package, c("A", "C", "E"))
  # Check op/ver handling with %||% -> NA_character_ where missing
  expect_true(is.character(out$op) && is.character(out$ver) && is.character(out$field))
  # Confirm that exactly 3 rows are present
  expect_equal(nrow(out), 3L)
  
  # Per-row checks (order-agnostic)
  row_A <- out[out$package == "A", , drop = FALSE]
  expect_identical(row_A$op, ">=")
  expect_identical(row_A$ver, "1.0")
  expect_identical(row_A$field, "Depends")
  
  row_C <- out[out$package == "C", , drop = FALSE]
  expect_true(is.na(row_C$op))
  expect_true(is.na(row_C$ver))
  expect_identical(row_C$field, "Imports")
  
  row_E <- out[out$package == "E", , drop = FALSE]
  expect_identical(row_E$op, ">=")
  expect_identical(row_E$ver, "2.3")
  expect_identical(row_E$field, "Suggests")
})


test_that("imports all exported symbols into environment", {
  env <- new.env()
  
  # Mock asNamespace to return a fake namespace
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  
  # Mock getNamespaceExports to return a list of exported names
  mockery::stub(
    import_namespace_exports,
    "getNamespaceExports",
    c("func1", "func2", "data1")
  )
  
  # Mock getExportedValue to return mock functions/data
  get_exported_mock <- mockery::mock(
    function(x) x + 1,  # func1
    function(x) x * 2,  # func2
    data.frame(a = 1)   # data1
  )
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  result <- import_namespace_exports("testpkg", env)
  
  # Verify return value
  expect_equal(result, invisible(TRUE))
  
  # Verify all symbols were imported into env
  expect_true(exists("func1", envir = env))
  expect_true(exists("func2", envir = env))
  expect_true(exists("data1", envir = env))
  
  # Verify getExportedValue was called 3 times
  mockery::expect_called(get_exported_mock, 3)
  mockery::expect_args(get_exported_mock, 1, "testpkg", "func1")
  mockery::expect_args(get_exported_mock, 2, "testpkg", "func2")
  mockery::expect_args(get_exported_mock, 3, "testpkg", "data1")
})

test_that("respects overwrite parameter when FALSE", {
  env <- new.env()
  
  # Pre-populate env with existing binding
  env$func1 <- "existing_value"
  
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  mockery::stub(
    import_namespace_exports,
    "getNamespaceExports",
    c("func1", "func2")
  )
  
  get_exported_mock <- mockery::mock(
    function(x) x + 1,  # func1
    function(x) x * 2   # func2
  )
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  # Import with overwrite = FALSE
  result <- import_namespace_exports("testpkg", env, overwrite = FALSE)
  
  expect_equal(result, invisible(TRUE))
  
  # func1 should NOT be overwritten
  expect_equal(env$func1, "existing_value")
  
  # func2 should be imported (didn't exist)
  expect_true(exists("func2", envir = env))
  
  # getExportedValue called only once (for func2, skipped func1)
  mockery::expect_called(get_exported_mock, 1)
  mockery::expect_args(get_exported_mock, 1, "testpkg", "func2")
})


test_that("overwrites existing bindings when overwrite = TRUE", {
  env <- new.env()
  
  # Pre-populate env with existing binding
  env$func1 <- "old_value"
  
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  mockery::stub(
    import_namespace_exports,
    "getNamespaceExports",
    c("func1")
  )
  
  get_exported_mock <- mockery::mock(
    function(x) x + 1
  )
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  # Import with overwrite = TRUE (default)
  result <- import_namespace_exports("testpkg", env, overwrite = TRUE)
  
  expect_equal(result, invisible(TRUE))
  
  # func1 should be overwritten with new value
  expect_true(is.function(env$func1))
  expect_false(identical(env$func1, "old_value"))
  
  mockery::expect_called(get_exported_mock, 1)
})


test_that("applies filter to exported symbols", {
  env <- new.env()
  
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  
  # All available exports
  mockery::stub(
    import_namespace_exports,
    "getNamespaceExports",
    c("func1", "func2", "func3", "data1")
  )
  
  get_exported_mock <- mockery::mock(
    function(x) x + 1,  # func1
    data.frame(a = 1)   # data1
  )
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  # Import only func1 and data1
  result <- import_namespace_exports(
    "testpkg",
    env,
    filter = c("func1", "data1")
  )
  
  expect_equal(result, invisible(TRUE))
  
  # Only filtered symbols should be imported
  expect_true(exists("func1", envir = env))
  expect_true(exists("data1", envir = env))
  expect_false(exists("func2", envir = env))
  expect_false(exists("func3", envir = env))
  
  # getExportedValue called only 2 times (for filtered symbols)
  mockery::expect_called(get_exported_mock, 2)
})


test_that("handles empty exports gracefully", {
  env <- new.env()
  
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  mockery::stub(import_namespace_exports, "getNamespaceExports", character(0))
  
  get_exported_mock <- mockery::mock()
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  result <- import_namespace_exports("emptypkg", env)
  
  expect_equal(result, invisible(TRUE))
  
  # getExportedValue should not be called at all
  mockery::expect_called(get_exported_mock, 0)
})


test_that("handles filter that matches no exports", {
  env <- new.env()
  
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  mockery::stub(
    import_namespace_exports,
    "getNamespaceExports",
    c("func1", "func2")
  )
  
  get_exported_mock <- mockery::mock()
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  # Filter for symbols that don't exist
  result <- import_namespace_exports(
    "testpkg",
    env,
    filter = c("nonexistent1", "nonexistent2")
  )
  
  expect_equal(result, invisible(TRUE))
  
  # getExportedValue should not be called (intersection is empty)
  mockery::expect_called(get_exported_mock, 0)
})


test_that("filter intersects with available exports", {
  env <- new.env()
  
  mockery::stub(import_namespace_exports, "asNamespace", "fake_ns")
  mockery::stub(
    import_namespace_exports,
    "getNamespaceExports",
    c("func1", "func2", "func3")
  )
  
  get_exported_mock <- mockery::mock(
    function(x) x + 1,  # func1
    function(x) x * 2   # func2
  )
  mockery::stub(import_namespace_exports, "getExportedValue", get_exported_mock)
  
  # Filter includes some that exist and some that don't
  result <- import_namespace_exports(
    "testpkg",
    env,
    filter = c("func1", "func2", "nonexistent")
  )
  
  expect_equal(result, invisible(TRUE))
  
  # Only the intersection (func1, func2) should be imported
  expect_true(exists("func1", envir = env))
  expect_true(exists("func2", envir = env))
  expect_false(exists("func3", envir = env))
  expect_false(exists("nonexistent", envir = env))
  
  # getExportedValue called 2 times (for the intersection)
  mockery::expect_called(get_exported_mock, 2)
})

test_that("installs package when not installed", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  result <- install_if_needed("missingpkg")
  
  expect_equal(result, invisible(TRUE))
  
  # install.packages should be called for missing package
  mockery::expect_called(install_packages_mock, 1)
  mockery::expect_args(install_packages_mock, 1, "missingpkg")
  
  # loadNamespace should be called
  mockery::expect_called(load_ns_mock, 1)
})


test_that("does not reinstall when package is installed with no version constraint", {
  mockery::stub(install_if_needed, "requireNamespace", TRUE)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  result <- install_if_needed("existingpkg")
  
  expect_equal(result, invisible(TRUE))
  
  # install.packages should NOT be called
  mockery::expect_called(install_packages_mock, 0)
  
  # loadNamespace should still be called
  mockery::expect_called(load_ns_mock, 1)
})


test_that("skips version check if installed but op is NULL", {
  mockery::stub(install_if_needed, "requireNamespace", TRUE)
  
  pkg_desc_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::packageDescription", pkg_desc_mock)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Call with op = NULL, ver = "1.0.0"
  result <- install_if_needed("pkg", op = NULL, ver = "1.0.0")
  
  expect_equal(result, invisible(TRUE))
  
  # packageDescription should NOT be called (op is NULL)
  mockery::expect_called(pkg_desc_mock, 0)
  
  # install.packages should NOT be called
  mockery::expect_called(install_packages_mock, 0)
})


test_that("skips version check if installed but ver is NULL", {
  mockery::stub(install_if_needed, "requireNamespace", TRUE)
  
  pkg_desc_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::packageDescription", pkg_desc_mock)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Call with op = ">=", ver = NULL
  result <- install_if_needed("pkg", op = ">=", ver = NULL)
  
  expect_equal(result, invisible(TRUE))
  
  # packageDescription should NOT be called (ver is NULL)
  mockery::expect_called(pkg_desc_mock, 0)
  
  # install.packages should NOT be called
  mockery::expect_called(install_packages_mock, 0)
})


test_that("checks version constraint when installed with both op and ver", {
  mockery::stub(install_if_needed, "requireNamespace", TRUE)
  
  # Simulate installed version "1.5.0" (satisfies >= 1.0.0)
  mockery::stub(install_if_needed, "utils::packageDescription", "1.5.0")
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(install_if_needed, "constraint_satisfied", constraint_satisfied_mock)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  result <- install_if_needed("pkg", op = ">=", ver = "1.0.0")
  
  expect_equal(result, invisible(TRUE))
  
  # constraint_satisfied should be called with correct args
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, "1.5.0", ">=", "1.0.0")
  
  # install.packages should NOT be called (constraint is satisfied)
  mockery::expect_called(install_packages_mock, 0)
})


test_that("reinstalls when version constraint is NOT satisfied", {
  mockery::stub(install_if_needed, "requireNamespace", TRUE)
  
  # Simulate installed version "0.9.0" (does NOT satisfy >= 1.0.0)
  mockery::stub(install_if_needed, "utils::packageDescription", "0.9.0")
  
  constraint_satisfied_mock <- mockery::mock(FALSE)
  mockery::stub(install_if_needed, "constraint_satisfied", constraint_satisfied_mock)
  
  install_packages_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Capture message about re-installation
  msgs <- testthat::capture_messages({
    result <- install_if_needed("pkg", op = ">=", ver = "1.0.0")
  })
  
  expect_equal(result, invisible(TRUE))
  
  # Should message about re-installation
  expect_true(
    any(grepl("does not satisfy constraint", msgs, fixed = TRUE)),
    info = paste("Messages were:", paste(msgs, collapse = "\n"))
  )
  
  # install.packages should be called (constraint NOT satisfied)
  mockery::expect_called(install_packages_mock, 1)
  mockery::expect_args(install_packages_mock, 1, "pkg")
})


test_that("uses install.packages for >= and > constraints", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  remotes_install_mock <- mockery::mock()
  mockery::stub(install_if_needed, "remotes::install_version", remotes_install_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Test with >= constraint
  result <- install_if_needed("pkg", op = ">=", ver = "1.0.0")
  
  expect_equal(result, invisible(TRUE))
  
  # install.packages should be called (>= constraint)
  mockery::expect_called(install_packages_mock, 1)
  
  # remotes::install_version should NOT be called
  mockery::expect_called(remotes_install_mock, 0)
})


test_that("uses install.packages for > constraint", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  remotes_install_mock <- mockery::mock()
  mockery::stub(install_if_needed, "remotes::install_version", remotes_install_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Test with > constraint
  result <- install_if_needed("pkg", op = ">", ver = "1.0.0")
  
  expect_equal(result, invisible(TRUE))
  
  # install.packages should be called (> constraint)
  mockery::expect_called(install_packages_mock, 1)
  
  # remotes::install_version should NOT be called
  mockery::expect_called(remotes_install_mock, 0)
})


test_that("uses remotes::install_version for <= constraint", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  remotes_install_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "remotes::install_version", remotes_install_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Test with <= constraint
  result <- install_if_needed("pkg", op = "<=", ver = "2.0.0")
  
  expect_equal(result, invisible(TRUE))
  
  # remotes::install_version should be called
  mockery::expect_called(remotes_install_mock, 1)
  mockery::expect_args(remotes_install_mock, 1, "pkg", version = "2.0.0", 
                       upgrade = "never", force = TRUE)
  
  # install.packages should NOT be called
  mockery::expect_called(install_packages_mock, 0)
})


test_that("uses remotes::install_version for < constraint", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  remotes_install_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "remotes::install_version", remotes_install_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Test with < constraint
  result <- install_if_needed("pkg", op = "<", ver = "2.0.0")
  
  expect_equal(result, invisible(TRUE))
  
  # remotes::install_version should be called
  mockery::expect_called(remotes_install_mock, 1)
  mockery::expect_args(remotes_install_mock, 1, "pkg", version = "2.0.0",
                       upgrade = "never", force = TRUE)
  
  # install.packages should NOT be called
  mockery::expect_called(install_packages_mock, 0)
})


test_that("uses remotes::install_version for == constraint", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", install_packages_mock)
  
  remotes_install_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "remotes::install_version", remotes_install_mock)
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Test with == constraint
  result <- install_if_needed("pkg", op = "==", ver = "1.5.0")
  
  expect_equal(result, invisible(TRUE))
  
  # remotes::install_version should be called
  mockery::expect_called(remotes_install_mock, 1)
  mockery::expect_args(remotes_install_mock, 1, "pkg", version = "1.5.0",
                       upgrade = "never", force = TRUE)
  
  # install.packages should NOT be called
  mockery::expect_called(install_packages_mock, 0)
})


test_that("handles installation errors gracefully", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", function(pkg, ...) {
    stop("CRAN mirror error")
  })
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Should not throw, should capture error message
  msgs <- testthat::capture_messages({
    result <- install_if_needed("pkg")
  })
  
  expect_equal(result, invisible(TRUE))
  
  # Should have error message
  expect_true(
    any(grepl("Failed to install package: pkg - CRAN mirror error", msgs, fixed = TRUE)),
    info = paste("Messages were:", paste(msgs, collapse = "\n"))
  )
})


test_that("loads namespace even if installation fails", {
  mockery::stub(install_if_needed, "requireNamespace", FALSE)
  
  install_packages_mock <- mockery::mock()
  mockery::stub(install_if_needed, "utils::install.packages", function(pkg, ...) {
    stop("install error")
  })
  
  load_ns_mock <- mockery::mock(invisible(NULL))
  mockery::stub(install_if_needed, "loadNamespace", load_ns_mock)
  
  # Suppress error messages
  suppressMessages({
    result <- install_if_needed("pkg")
  })
  
  expect_equal(result, invisible(TRUE))
  
  # loadNamespace should still be called despite install error
  mockery::expect_called(load_ns_mock, 1)
  mockery::expect_args(load_ns_mock, 1, "pkg")
})

test_that("returns TRUE when op is NULL", {
  result <- constraint_satisfied("1.5.0", op = NULL, ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("returns TRUE when ver is NULL", {
  result <- constraint_satisfied("1.5.0", op = ">=", ver = NULL)
  expect_equal(result, TRUE)
})


test_that("returns TRUE when both op and ver are NULL", {
  result <- constraint_satisfied("1.5.0", op = NULL, ver = NULL)
  expect_equal(result, TRUE)
})


test_that("handles >= operator when versions are equal", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  result <- constraint_satisfied("1.0.0", op = ">=", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles >= operator when installed is greater", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("1.5.0", op = ">=", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles >= operator when installed is less", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  
  result <- constraint_satisfied("0.9.0", op = ">=", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles > operator when versions are equal", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  result <- constraint_satisfied("1.0.0", op = ">", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles > operator when installed is greater", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("1.5.0", op = ">", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles > operator when installed is less", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  
  result <- constraint_satisfied("0.9.0", op = ">", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles <= operator when versions are equal", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  result <- constraint_satisfied("1.0.0", op = "<=", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles <= operator when installed is less", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  
  result <- constraint_satisfied("0.9.0", op = "<=", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles <= operator when installed is greater", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("1.5.0", op = "<=", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles < operator when versions are equal", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  result <- constraint_satisfied("1.0.0", op = "<", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles < operator when installed is less", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  
  result <- constraint_satisfied("0.9.0", op = "<", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles < operator when installed is greater", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("1.5.0", op = "<", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles == operator when versions are equal", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  result <- constraint_satisfied("1.0.0", op = "==", ver = "1.0.0")
  expect_equal(result, TRUE)
})


test_that("handles == operator when installed is greater", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("1.5.0", op = "==", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("handles == operator when installed is less", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  
  result <- constraint_satisfied("0.9.0", op = "==", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("returns FALSE for unknown operator", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  # Use an operator not in the switch statement
  result <- constraint_satisfied("1.0.0", op = "!=", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("returns FALSE for another unknown operator", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("1.5.0", op = "<>", ver = "1.0.0")
  expect_equal(result, FALSE)
})


test_that("converts installed_version to character before comparison", {
  compare_version_mock <- mockery::mock(0)
  mockery::stub(constraint_satisfied, "utils::compareVersion", compare_version_mock)
  
  # Pass a package_version object instead of character
  installed_ver <- package_version("1.0.0")
  result <- constraint_satisfied(installed_ver, op = ">=", ver = "1.0.0")
  
  expect_equal(result, TRUE)
  
  # Verify compareVersion was called with character version
  mockery::expect_called(compare_version_mock, 1)
  mockery::expect_args(compare_version_mock, 1, "1.0.0", "1.0.0")
})


test_that("handles numeric package versions", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  
  result <- constraint_satisfied("2.0", op = ">", ver = "1.0")
  expect_equal(result, TRUE)
})


test_that("handles complex version strings", {
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  
  result <- constraint_satisfied("1.2.3.4", op = ">=", ver = "1.2.3.4")
  expect_equal(result, TRUE)
})


test_that("calls compareVersion with correct arguments in order", {
  compare_version_mock <- mockery::mock(-1)
  mockery::stub(constraint_satisfied, "utils::compareVersion", compare_version_mock)
  
  result <- constraint_satisfied("0.5.0", op = ">=", ver = "1.0.0")
  
  expect_equal(result, FALSE)
  
  # Verify compareVersion was called with installed version first, constraint version second
  mockery::expect_called(compare_version_mock, 1)
  mockery::expect_args(compare_version_mock, 1, "0.5.0", "1.0.0")
})


test_that(">= operator with all three compareVersion outputs", {
  # Test with compareVersion returning -1 (less)
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  expect_equal(constraint_satisfied("0.5.0", op = ">=", ver = "1.0.0"), FALSE)
  
  # Test with compareVersion returning 0 (equal)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  expect_equal(constraint_satisfied("1.0.0", op = ">=", ver = "1.0.0"), TRUE)
  
  # Test with compareVersion returning 1 (greater)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  expect_equal(constraint_satisfied("1.5.0", op = ">=", ver = "1.0.0"), TRUE)
})


test_that("> operator with all three compareVersion outputs", {
  # Test with compareVersion returning -1 (less)
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  expect_equal(constraint_satisfied("0.5.0", op = ">", ver = "1.0.0"), FALSE)
  
  # Test with compareVersion returning 0 (equal)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  expect_equal(constraint_satisfied("1.0.0", op = ">", ver = "1.0.0"), FALSE)
  
  # Test with compareVersion returning 1 (greater)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  expect_equal(constraint_satisfied("1.5.0", op = ">", ver = "1.0.0"), TRUE)
})


test_that("<= operator with all three compareVersion outputs", {
  # Test with compareVersion returning -1 (less)
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  expect_equal(constraint_satisfied("0.5.0", op = "<=", ver = "1.0.0"), TRUE)
  
  # Test with compareVersion returning 0 (equal)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  expect_equal(constraint_satisfied("1.0.0", op = "<=", ver = "1.0.0"), TRUE)
  
  # Test with compareVersion returning 1 (greater)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  expect_equal(constraint_satisfied("1.5.0", op = "<=", ver = "1.0.0"), FALSE)
})


test_that("< operator with all three compareVersion outputs", {
  # Test with compareVersion returning -1 (less)
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  expect_equal(constraint_satisfied("0.5.0", op = "<", ver = "1.0.0"), TRUE)
  
  # Test with compareVersion returning 0 (equal)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  expect_equal(constraint_satisfied("1.0.0", op = "<", ver = "1.0.0"), FALSE)
  
  # Test with compareVersion returning 1 (greater)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  expect_equal(constraint_satisfied("1.5.0", op = "<", ver = "1.0.0"), FALSE)
})


test_that("== operator with all three compareVersion outputs", {
  # Test with compareVersion returning -1 (less)
  mockery::stub(constraint_satisfied, "utils::compareVersion", -1)
  expect_equal(constraint_satisfied("0.5.0", op = "==", ver = "1.0.0"), FALSE)
  
  # Test with compareVersion returning 0 (equal)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 0)
  expect_equal(constraint_satisfied("1.0.0", op = "==", ver = "1.0.0"), TRUE)
  
  # Test with compareVersion returning 1 (greater)
  mockery::stub(constraint_satisfied, "utils::compareVersion", 1)
  expect_equal(constraint_satisfied("1.5.0", op = "==", ver = "1.0.0"), FALSE)
})

test_that("returns invisibly TRUE when Depends field is NA", {
  desc <- matrix(NA_character_, nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", NA_character_)
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("returns invisibly TRUE when Depends field is empty string", {
  desc <- matrix("", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "")
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("returns invisibly TRUE when no R constraint found in Depends", {
  desc <- matrix("pkg1, pkg2", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "pkg1, pkg2")
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("returns invisibly TRUE when R constraint exists but normalize_constraint returns NULL", {
  desc <- matrix("R ()", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R ()")
  mockery::stub(validate_R_requirement, "normalize_constraint", NULL)
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("returns invisibly TRUE when normalized constraint is empty string", {
  desc <- matrix("R (   )", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (   )")
  mockery::stub(validate_R_requirement, "normalize_constraint", "")
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("returns invisibly TRUE when constraint string doesn't match operator pattern", {
  desc <- matrix("R (invalid)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (invalid)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "invalid")
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("throws error when R version constraint is not satisfied with >=", {
  desc <- matrix("R (>= 4.5.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 4.5.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 4.5.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  mockery::stub(validate_R_requirement, "constraint_satisfied", FALSE)
  
  expect_error(
    validate_R_requirement(desc),
    "This package requires R version >= 4.5.0 but current R is 4.4.0",
    fixed = TRUE
  )
})


test_that("throws error when R version constraint is not satisfied with >", {
  desc <- matrix("R (> 4.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (> 4.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "> 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.0.0"))
  mockery::stub(validate_R_requirement, "constraint_satisfied", FALSE)
  
  expect_error(
    validate_R_requirement(desc),
    "This package requires R version > 4.0.0 but current R is 4.0.0",
    fixed = TRUE
  )
})


test_that("returns invisibly TRUE when R version constraint is satisfied with >=", {
  desc <- matrix("R (>= 4.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 4.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  mockery::stub(validate_R_requirement, "constraint_satisfied", TRUE)
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("returns invisibly TRUE when R version constraint is satisfied with >", {
  desc <- matrix("R (> 4.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (> 4.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "> 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.1.0"))
  mockery::stub(validate_R_requirement, "constraint_satisfied", TRUE)
  
  result <- validate_R_requirement(desc)
  expect_equal(result, invisible(TRUE))
})


test_that("normalizes single = to == operator", {
  desc <- matrix("R (= 4.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (= 4.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "= 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.0.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  expect_equal(result, invisible(TRUE))
  
  # Verify constraint_satisfied was called with == operator
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.0.0"), "==", "4.0.0")
})


test_that("uses != operator without converting to ==", {
  desc <- matrix("R (!= 4.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (!= 4.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "!= 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.1.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  expect_equal(result, invisible(TRUE))
  
  # Verify constraint_satisfied was called with != operator (not normalized)
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.1.0"), "!=", "4.0.0")
})


test_that("trims whitespace from version string", {
  desc <- matrix("R (>=  4.0.0  )", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>=  4.0.0  )")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">=  4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.1.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  # Verify version was trimmed
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.1.0"), ">=", "4.0.0")
})


test_that("extracts first R constraint when multiple are present", {
  desc <- matrix("R (>= 4.0.0), R (>= 3.5.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 4.0.0), R (>= 3.5.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  # Should use first constraint (>= 4.0.0), not second
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.4.0"), ">=", "4.0.0")
})


test_that("handles <= operator in R requirement", {
  desc <- matrix("R (<= 4.5.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (<= 4.5.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "<= 4.5.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.4.0"), "<=", "4.5.0")
})


test_that("handles < operator in R requirement", {
  desc <- matrix("R (< 5.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (< 5.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "< 5.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.4.0"), "<", "5.0.0")
})


test_that("handles == operator in R requirement", {
  desc <- matrix("R (== 4.4.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (== 4.4.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "== 4.4.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.4.0"), "==", "4.4.0")
})


test_that("error message includes current R version", {
  desc <- matrix("R (>= 5.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 5.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 5.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.2"))
  mockery::stub(validate_R_requirement, "constraint_satisfied", FALSE)
  
  expect_error(
    validate_R_requirement(desc),
    "but current R is 4.4.2",
    fixed = TRUE
  )
})


test_that("error has attr simpleError", {
  desc <- matrix("R (>= 5.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 5.0.0)")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 5.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.4.0"))
  mockery::stub(validate_R_requirement, "constraint_satisfied", FALSE)
  
  error_obj <- tryCatch(
    validate_R_requirement(desc),
    error = function(e) e
  )
  
  # When call. = FALSE, the call is not included in the error
  expect_true(checkmate::check_class(error_obj, "simpleError"))
})


test_that("handles complex version strings with multiple dots", {
  desc <- matrix("R (>= 4.0.1.2)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 4.0.1.2)")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 4.0.1.2")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.1.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.1.0"), ">=", "4.0.1.2")
})


test_that("strips R() parentheses correctly from constraint string", {
  desc <- matrix("R (>= 4.0.0)", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R (>= 4.0.0)")
  # This stub simulates the gsub() operation
  mockery::stub(validate_R_requirement, "normalize_constraint", ">= 4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.1.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  # The constraint string passed should not have R( ) wrapper
  mockery::expect_called(constraint_satisfied_mock, 1)
  mockery::expect_args(constraint_satisfied_mock, 1, package_version("4.1.0"), ">=", "4.0.0")
})


test_that("handles whitespace variations in R constraint syntax", {
  desc <- matrix("R(  >=  4.0.0  )", nrow = 1, dimnames = list(NULL, c("Depends")))
  
  mockery::stub(validate_R_requirement, "get_field", "R(  >=  4.0.0  )")
  mockery::stub(validate_R_requirement, "normalize_constraint", ">=  4.0.0")
  mockery::stub(validate_R_requirement, "getRversion", package_version("4.1.0"))
  
  constraint_satisfied_mock <- mockery::mock(TRUE)
  mockery::stub(validate_R_requirement, "constraint_satisfied", constraint_satisfied_mock)
  
  result <- validate_R_requirement(desc)
  
  mockery::expect_called(constraint_satisfied_mock, 1)
})


test_that("all if statement branches are covered", {
  # Branch 1: is.na(deps) returns TRUE
  mockery::stub(validate_R_requirement, "get_field", NA_character_)
  result1 <- validate_R_requirement(matrix(NA_character_, nrow = 1, dimnames = list(NULL, c("Depends"))))
  expect_equal(result1, invisible(TRUE))
  
  # Branch 2: !nzchar(deps) returns TRUE
  mockery::stub(validate_R_requirement, "get_field", "")
  result2 <- validate_R_requirement(matrix("", nrow = 1, dimnames = list(NULL, c("Depends"))))
  expect_equal(result2, invisible(TRUE))
  
  # Branch 3: length(m) == 0
  mockery::stub(validate_R_requirement, "get_field", "pkg1, pkg2")
  result3 <- validate_R_requirement(matrix("pkg1, pkg2", nrow = 1, dimnames = list(NULL, c("Depends"))))
  expect_equal(result3, invisible(TRUE))
  
  # Branch 4: normalize_constraint returns NULL
  mockery::stub(validate_R_requirement, "get_field", "R ()")
  mockery::stub(validate_R_requirement, "normalize_constraint", NULL)
  result4 <- validate_R_requirement(matrix("R ()", nrow = 1, dimnames = list(NULL, c("Depends"))))
  expect_equal(result4, invisible(TRUE))
  
  # Branch 5: regexec doesn't match pattern
  mockery::stub(validate_R_requirement, "get_field", "R (malformed)")
  mockery::stub(validate_R_requirement, "normalize_constraint", "malformed")
  result5 <- validate_R_requirement(matrix("R (malformed)", nrow = 1, dimnames = list(NULL, c("Depends"))))
  expect_equal(result5, invisible(TRUE))
})

test_that("returns NULL when dep is empty string after trimming", {
  result <- parse_dep("")
  expect_null(result)
})


test_that("returns NULL when dep is empty string after trimming", {
  result <- parse_dep("")
  expect_null(result)
})


test_that("returns NULL when dep is whitespace only", {
  result <- parse_dep("   ")
  expect_null(result)
})


test_that("returns NULL when dep is NA", {
  result <- parse_dep(NA_character_)
  expect_null(result)
})


test_that("parses simple package name without version constraint", {
  result <- parse_dep("pkgA")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_null(result$op)
  expect_null(result$ver)
  expect_null(result$raw)
})


test_that("parses package name with whitespace", {
  result <- parse_dep("  pkgA  ")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_null(result$op)
  expect_null(result$ver)
  expect_null(result$raw)
})


test_that("parses package with >= constraint", {
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.0.0")
  
  result <- parse_dep("pkgA (>= 1.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, ">=")
  expect_equal(result$ver, "1.0.0")
  expect_equal(result$raw, ">= 1.0.0")
})


test_that("parses package with > constraint", {
  mockery::stub(parse_dep, "normalize_constraint", "> 1.0.0")
  
  result <- parse_dep("pkgA (> 1.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, ">")
  expect_equal(result$ver, "1.0.0")
  expect_equal(result$raw, "> 1.0.0")
})


test_that("parses package with <= constraint", {
  mockery::stub(parse_dep, "normalize_constraint", "<= 2.0.0")
  
  result <- parse_dep("pkgA (<= 2.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, "<=")
  expect_equal(result$ver, "2.0.0")
  expect_equal(result$raw, "<= 2.0.0")
})


test_that("parses package with < constraint", {
  mockery::stub(parse_dep, "normalize_constraint", "< 2.0.0")
  
  result <- parse_dep("pkgA (< 2.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, "<")
  expect_equal(result$ver, "2.0.0")
  expect_equal(result$raw, "< 2.0.0")
})


test_that("parses package with == constraint", {
  mockery::stub(parse_dep, "normalize_constraint", "== 1.5.0")
  
  result <- parse_dep("pkgA (== 1.5.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, "==")
  expect_equal(result$ver, "1.5.0")
  expect_equal(result$raw, "== 1.5.0")
})


test_that("normalizes single = to == operator", {
  mockery::stub(parse_dep, "normalize_constraint", "= 1.0.0")
  
  result <- parse_dep("pkgA (= 1.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, "==")
  expect_equal(result$ver, "1.0.0")
  expect_equal(result$raw, "= 1.0.0")
})


test_that("returns NULL for malformed dep string that doesn't match regex", {
  result <- parse_dep("invalid (name (nested)")
  expect_equal(result$pkg, "invalid")
})

test_that("handles package name with dots", {
  result <- parse_dep("pkg.A")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkg.A")
  expect_null(result$op)
})


test_that("handles package name with numbers", {
  result <- parse_dep("pkg123")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkg123")
  expect_null(result$op)
})


test_that("handles package name with mixed alphanumeric and dots", {
  result <- parse_dep("pkg.1.2.A")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkg.1.2.A")
  expect_null(result$op)
})


test_that("handles whitespace around package name in constraint", {
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.0.0")
  
  result <- parse_dep("  pkgA  (>= 1.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, ">=")
})


test_that("handles whitespace around operator and version", {
  mockery::stub(parse_dep, "normalize_constraint", ">=  1.0.0")
  
  result <- parse_dep("pkgA (>=  1.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, ">=")
  expect_equal(result$ver, "1.0.0")
})

test_that("returns op and ver as NULL when raw constraint is unrecognized", {
  mockery::stub(parse_dep, "normalize_constraint", "invalid constraint")
  
  result <- parse_dep("pkgA (invalid constraint)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_null(result$op)
  expect_null(result$ver)
  expect_equal(result$raw, "invalid constraint")
})


test_that("keeps raw constraint even when op/ver cannot be parsed", {
  mockery::stub(parse_dep, "normalize_constraint", "malformed")
  
  result <- parse_dep("pkgA (malformed)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_null(result$op)
  expect_null(result$ver)
  expect_equal(result$raw, "malformed")
})


test_that("handles != operator (not normalized)", {
  mockery::stub(parse_dep, "normalize_constraint", "!= 1.0.0")
  
  result <- parse_dep("pkgA (!= 1.0.0)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, "!=")
  expect_equal(result$ver, "1.0.0")
})


test_that("skips constraint parsing when raw is NULL after normalization", {
  mockery::stub(parse_dep, "normalize_constraint", NULL)
  
  result <- parse_dep("pkgA (   )")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_null(result$op)
  expect_null(result$ver)
  expect_null(result$raw)
})


test_that("handles complex version strings in constraint", {
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.2.3.4")
  
  result <- parse_dep("pkgA (>= 1.2.3.4)")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, ">=")
  expect_equal(result$ver, "1.2.3.4")
})


test_that("trims whitespace from extracted version string", {
  mockery::stub(parse_dep, "normalize_constraint", ">=  1.0.0  ")
  
  result <- parse_dep("pkgA (>=  1.0.0  )")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_equal(result$op, ">=")
  expect_equal(result$ver, "1.0.0")
})


test_that("returns list with all four elements", {
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.0.0")
  
  result <- parse_dep("pkgA (>= 1.0.0)")
  
  expect_true(all(c("pkg", "op", "ver", "raw") %in% names(result)))
  expect_equal(length(result), 4)
})


test_that("single = operator is specifically converted to ==", {
  mockery::stub(parse_dep, "normalize_constraint", "= 1.0.0")
  
  result <- parse_dep("pkgA (= 1.0.0)")
  
  expect_equal(result$op, "==")
  expect_false(identical(result$op, "="))
})


test_that("other operators are not converted", {
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.0.0")
  
  result <- parse_dep("pkgA (>= 1.0.0)")
  
  expect_equal(result$op, ">=")
})


test_that("handles length(m) < 2 case with warning", {
  expect_warning(
    parse_dep("()"),
    "Could not parse dependency:"
  )
})


test_that("handles pattern with no version constraint (length m < 3)", {
  result <- parse_dep("pkgA")
  
  expect_true(checkmate::check_list(result))
  expect_equal(result$pkg, "pkgA")
  expect_null(result$raw)
})

test_that("all conditional branches for empty/NA dep are covered", {
  # Branch 1: empty string
  result1 <- parse_dep("")
  expect_null(result1)
  
  # Branch 2: NA
  result2 <- parse_dep(NA_character_)
  expect_null(result2)
  
  # Branch 3: whitespace only (becomes empty after trim)
  result3 <- parse_dep("   ")
  expect_null(result3)
})


test_that("all conditional branches for op normalization are covered", {
  # Branch 1: op == "="
  mockery::stub(parse_dep, "normalize_constraint", "= 1.0.0")
  result1 <- parse_dep("pkgA (= 1.0.0)")
  expect_equal(result1$op, "==")
  
  # Branch 2: op != "=" (stays as is)
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.0.0")
  result2 <- parse_dep("pkgA (>= 1.0.0)")
  expect_equal(result2$op, ">=")
})


test_that("all conditional branches for constraint parsing are covered", {
  # Branch 1: length(m2) >= 3 (valid constraint)
  mockery::stub(parse_dep, "normalize_constraint", ">= 1.0.0")
  result1 <- parse_dep("pkgA (>= 1.0.0)")
  expect_equal(result1$op, ">=")
  expect_equal(result1$ver, "1.0.0")
  
  # Branch 2: length(m2) < 3 (unrecognized constraint)
  mockery::stub(parse_dep, "normalize_constraint", "invalid")
  result2 <- parse_dep("pkgA (invalid)")
  expect_null(result2$op)
  expect_null(result2$ver)
  expect_equal(result2$raw, "invalid")
})

#' Untar package and return installation directory
#'
#' @param pkg_tar path to tarball package
#' @param temp_file_name name of `tempfile`
#' @keywords internal
unpack_tarball <- function(pkg_tar,
                           temp_file_name = "temp_file_"){
  
  pkg_disp <- get_pkg_name(pkg_tar)
  message(paste0("unpacking ", pkg_disp, " locally"))
  # Create temporary location for package installation
  temp_pkg_dir <- tempfile(temp_file_name)
  if (!dir.create(temp_pkg_dir)) stop("unable to create ", temp_pkg_dir)
  
  source_tar_dir <- file.path(temp_pkg_dir)
  
  # unpack tarball
  # use internal tar from R to ensure untar doesn't crash with
  # unacceptable tar file
  utils::untar(pkg_tar, exdir = source_tar_dir, tar = "internal")
  
  # unpackaged package path
  pkg_source_path <- fs::dir_ls(source_tar_dir)
  
  # Confirm tar is unpackaged in expected directory
  check1 <- checkmate::check_string(pkg_source_path)
  check2 <- checkmate::check_directory_exists(pkg_source_path)
  
  if (check1 == TRUE & check2 == TRUE) {
    message(paste0("unpacked ", pkg_disp, " locally"))
  } else {
    message(paste0("not able to unpack ", pkg_disp, " locally"))
  }
  
  return(pkg_source_path)
}

test_that("Unpacking a tar file works correctly", {
  
  r = getOption("repos")
  r["CRAN"] = "http://cran.us.r-project.org"
  options(repos = r)
  
  # Copy test package to a temp file
  pkg_orig <- system.file("test-data", "here-1.0.1.tar.gz", 
                         package = "test.assessr")
  pkg <- tempfile(fileext = ".tar.gz")
  file.copy(pkg_orig, pkg)
  
  # Defer cleanup of copied tarball
  withr::defer(unlink(pkg), envir = parent.frame())
  
  extract_files <-  
    suppressWarnings(unpack_tarball(pkg))
  
  expect_true(checkmate::checkVector(extract_files))
  
  expect_true(checkmate::check_class(extract_files, "fs_path"))
  
  expect_true(checkmate::check_string(extract_files))
  
  expect_true(checkmate::check_directory_exists(extract_files))
})

test_that("Unpacking an empty tar file works correctly", {
  
  # Copy test package to a temp file
  pkg_orig <- system.file("test-data", "empty.tar.gz", 
                          package = "test.assessr")
  pkg <- tempfile(fileext = ".tar.gz")
  file.copy(pkg_orig, pkg)
  
  # Defer cleanup of copied tarball
  withr::defer(unlink(pkg), envir = parent.frame())
  
  extract_files <-  
    suppressWarnings(unpack_tarball(pkg))
  
  expect_true(checkmate::checkVector(extract_files))
  
  expect_true(checkmate::check_class(extract_files, "fs_path"))
  
  expect_identical(checkmate::check_string(extract_files), 
                   "Must have length 1"
  )
  
  expect_identical(checkmate::check_directory_exists(extract_files), 
                   "No directory provided"
  )
})  

#' Run Coverage with Skipped Failing Tests (Standard Testing Framework)
#'
#' This function runs testthat-based test execution and code coverage
#' while identifying and skipping failing tests. It is designed for use
#' with standard testing frameworks (STF), primarily `testthat`, and
#' supports automated skip mapping to enable coverage calculation even
#' when some tests fail.
#'
#' The function:
#' \itemize{
#'   \item executes tests and collects detailed test results,
#'   \item identifies failing, erroring, or skipped tests,
#'   \item constructs skip maps to temporarily comment out failing test lines,
#'   \item computes coverage on the remaining tests, and
#'   \item returns a combined object containing coverage and test summaries.
#' }
#'
#' @param pkg_source_path Character. Path to the root directory of the
#'   package source.
#' @param test_pkg_data List. Output from 
#'   \code{check_pkg_tests_and_snaps()}, containing information about
#'   the package's testing setup (presence of testthat, snapshots, etc.).
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run. This must be an environment created by
#'   `initialise_covr_env()`, which sets `cov_env$work_dir` to a unique
#'   subdirectory under `tempdir()`. `run_covr_skip_stf()` will not create or
#'   modify `cov_env` beyond reading helpers into it; all file outputs produced
#'   by downstream helpers are expected to be written under
#'   `cov_env$work_dir` (never to the package directory or the user's home).
#'   The caller is responsible for lifecycle management; register
#'   `on.exit(cleanup_and_return_null(env = cov_env), add = TRUE)` in the
#'   calling scope to ensure bindings and the temporary workspace are cleaned
#'   up even on error.

#'
#' @return 
#' A named list containing coverage results and test diagnostics. The structure
#' includes:
#' \describe{
#'   \item{total_cov}{Numeric. Overall coverage percentage.}
#'   \item{res_cov}{A nested list with file-level coverage, errors, and notes.}
#'   \item{long_summary}{Long-format test results including PASS/FAIL/ERROR/SKIPPED.}
#'   \item{test_skip}{A data frame describing test lines skipped during coverage,
#'     if applicable.}
#' }
#'
#' If no tests fail, the returned list describes full coverage based on
#' unskipped tests. If no valid STF test directory is found, an empty coverage
#' object is returned via \code{create_empty_covr_list()}.
#'
#' @rdname run_covr_skip_stf  
#' @family stf_utility
#' @keywords internal
run_covr_skip_stf <- function(pkg_source_path, 
                              test_pkg_data, 
                              cov_env
                              ) {
  
  # Input checking
  checkmate::assert_string(pkg_source_path)
  checkmate::assert_directory_exists(pkg_source_path)
  
  # Ensure the argument was provided
  checkmate::assert_true(!missing(cov_env), .var.name = "cov_env")
  
  # Ensure it is an environment
  checkmate::assert_environment(cov_env, .var.name = "cov_env")
  
  path <- as.character(pkg_source_path)
  testdir <- file.path(path, "tests")
  pkg <- get_pkg_name(pkg_source_path)
  
  test_path <- get_stf_test_path(test_pkg_data, testdir)
  
  if (is.null(test_path)) {
    message(paste0("No test directory found for ", pkg))
    return(create_empty_covr_list(pkg, "No test directory found"))
  }
  
  # --- Ensure WD is restored (CRAN reviewer request) ---
  original_wd <- getwd()
  setwd(path)
  on.exit(setwd(original_wd), add = TRUE)
  
  # --- Setup coverage environment here (no env creation; CRAN-safe) ---
  # This will NOT write to tests/; it just loads helpers into the provided env.
  covr_setup <- setup_covr_env(pkg_source_path = pkg_source_path,
                               pkg_name        = pkg,
                               test_path       = test_path,
                               cov_env         = cov_env)
  
  # Optionally stash metadata onto env for downstream consumers (non-breaking)
  cov_env$datasets_loaded <- covr_setup$datasets_loaded
  cov_env$root_dir        <- covr_setup$root_dir
  cov_env$initial_state   <- covr_setup$initial_state
  cov_env$helpers_loaded  <- covr_setup$helpers_loaded
  
  message(paste0("Starting reporter process for ", pkg))
  testthat::set_max_fails(Inf)
  
  # unload target package
  try(unloadNamespace(pkg), silent = TRUE)
  
  # create fail reporter
  create_fail_reporter(test_path, pkg, cov_env)
  
  # create default reporter
  test_results <- create_test_reporter(test_path, pkg, cov_env)
  
  if (is.null(test_results)) {
    return(create_empty_covr_list(pkg, "Default reporter execution failed"))
  } else {
    # update context column 
    test_results <- fix_test_context(test_results)
  }
  
  message(paste0("Mapping package test structure for ", pkg))
  map_path <- test_path
  test_map <- tryCatch({
    map_tests_stf(map_path)
  }, error = function(e) {
    message(paste0("Error mapping tests for ", pkg, " : ", e$message))
    return(NULL)
  })
 
  if (is.null(test_map)) {
    message(paste0("Test map could not be created for ", pkg))
    return(create_empty_covr_list(pkg, "Test map creation failed"))
  }
  
  test_results_long <- get_tests_long_sum_stf(test_results)
  
  if (is.null(test_results_long)) {
    message(paste0("Unable to create skipped tests for :", pkg))
    return(create_empty_covr_list(pkg, "Skipped tests map creation failed"))
  }
  
  lookup <- c(code_script = "context", context = "test")
  test_results_long <- dplyr::rename(test_results_long, dplyr::all_of(lookup))
  
  test_results_filtered <- test_results_long[test_results_long$status != 'PASS', ]
  test_results_filtered$file <- gsub('#(.*?)$', '', basename(test_results_filtered$file))
  test_map <- dplyr::rename(test_map, context = "test")
  
  status_filter <- c("FAILED", "ERROR", "SKIPPED")
  problem_tests <- subset(test_results_filtered, status %in% status_filter)
  
  if (nrow(problem_tests) == 0) {
    problem_tests <- data.frame(
      file = "", code_script = "", context = "", status = "",
      n = NA_integer_, time = NA_real_, stringsAsFactors = FALSE
    )
  }
  
  failed <- any(test_results_long$status %in% c('ERROR', 'FAILED'))
  
  if (failed) {
    test_skip <- get_tests_skip_stf(problem_tests, test_map)
    message(paste0("Creating skip test map for ", pkg))
    
    # --- Pass cov_env down (create_coverage_skip_stf signature updated elsewhere) ---
    cvr <- create_coverage_skip_stf(pkg_source_path = pkg_source_path,
                                    pkg             = pkg,
                                    test_path       = map_path,
                                    test_skip       = test_skip,
                                    test_map        = test_map,
                                    cov_env         = cov_env)
    
    message(paste0("Creating test coverage for ", pkg))
    res_sum <- create_results_summary(cvr)
    
    # check if test coverage result exist
    no_cov_check <- is_no_coverage_cov(res_sum)
    
    if (no_cov_check) {
      message("Detected 'no_coverage.R' sentinel. Falling back to covr::package_coverage(type = 'tests').")
      cvr <- tryCatch({
        withr::with_dir(pkg_source_path, covr::package_coverage(path = pkg_source_path, type = "tests", quiet = TRUE))
      }, error = function(e) {
        message("package_coverage() failed: ", conditionMessage(e))
        NULL
      })
      res_sum <- create_results_summary(cvr)
    }
    
    
    covr_list <- list(
      total_cov = res_sum$total_cov,
      res_cov = list(
        name = res_sum$res_cov$name,
        coverage = res_sum$res_cov$coverage,
        errors = res_sum$res_cov$errors,
        notes = res_sum$res_cov$notes
      ),
      long_summary = test_results_long,
      test_skip = test_skip
    )
  } else {
    message(paste0("Creating test results for no skipped tests for ", pkg))
    
    # --- Pass cov_env down (create_covr_list_no_skip signature updated elsewhere) ---
    covr_list <- create_covr_list_no_skip(test_map        = test_map,
                                          test_results_long = test_results,
                                          pkg_source_path = pkg_source_path,
                                          pkg             = pkg,
                                          cov_env         = cov_env)
  }
  
  return(covr_list)
}


#' @title Create Test Reporter
#' @description Executes all tests in a given directory using `testthat::test_dir()`
#' and applies context fixes to ensure each test has a valid context.
#'
#' @param test_path Character string specifying the path to the test directory.
#' @param pkg Character string specifying the package name.
#' @param cov_env environment covr environment.
#'
#' @return A list of test results with corrected contexts, or `NULL` if an error occurs.
#' @keywords internal
create_test_reporter <- function(test_path, pkg, cov_env) {
  
  message(paste0("Creating test reporter for: ", pkg))
  test_results <- tryCatch({
    sink(tempfile())
    res <- testthat::test_dir(
      path = test_path,
      reporter = testthat::default_reporter(),
      env = cov_env,
      stop_on_failure = FALSE,
      stop_on_warning = FALSE,
      package = pkg,
      load_package = "source"
    )
    sink()
    res
  }, error = function(e) {
    message(paste0("Error running tests for ", pkg, " : ", e$message))
    return(NULL)
  })
  test_results
}

#' @title Create Fail Reporter
#' @description Executes all tests in a given directory using `testthat::test_dir()`
#' and applies context fixes to ensure each test has a valid context.
#'
#' @param test_path Character string specifying the path to the test directory.
#' @param pkg Character string specifying the package name.
#' @param cov_env environment convr environment.
#'
#' @return A list of test results with corrected contexts, or `NULL` if an error occurs.
#' @keywords internal
create_fail_reporter <- function(test_path, pkg, cov_env) {
  message("Creating failing tests reporter for: ", pkg)
  
  rep <- testthat::FailReporter$new()
  
  test_results <- tryCatch({
    testthat::test_dir(
      path = test_path,
      reporter = rep,
      env = cov_env,
      stop_on_failure = FALSE,
      stop_on_warning  = FALSE,
      package = pkg,
      load_package = "source"
    )
    # rep$results contains every expectation recorded
    rep$results
  }, error = function(e) {
    message("Error running tests for ", pkg, " : ", e$message)
    NULL
  })
  
  test_results
}


#' @title Fix Missing Test Contexts
#' @description Ensures that each test result has a valid `context` field.
#' If missing or empty, the context is set to the file name (without extension).
#'
#' @param test_results A list of test result objects from `testthat`.
#'
#' @return A list of test results with updated contexts.
#' @keywords internal
fix_test_context <- function(test_results) {
  lapply(test_results, function(x) {
    if (is.null(x$context) || x$context == "") {
      file_name <- tools::file_path_sans_ext(basename(x$file))
      x$context <- file_name
    }
    x
  })
}


#' Run Coverage While Skipping Failing Tests
#'
#' This function runs coverage analysis using `covr` while temporarily
#' commenting out known failing tests. It is intended for situations where
#' specific tests must be skipped (e.g., due to upstream bugs or known
#' failures) so that coverage can still be computed on remaining tests.
#'
#' The function:
#' \itemize{
#'   \item sets up a temporary coverage environment,
#'   \item identifies lines to skip based on `test_skip`,
#'   \item comments out those test lines,
#'   \item runs `covr::environment_coverage()`, and
#'   \item restores the original test files on exit.
#' }
#'
#' @param pkg_source_path Character. Path to the root of the package source.
#' @param pkg Character. Name of the package being tested.
#' @param test_path Character. Path to the test directory.
#'   Defaults to `'tests/testthat'`.
#' @param test_skip Data frame. Mapping of failing tests, typically containing
#'   columns `file`, `line1`, and `line2`. Each row describes a span of lines
#'   that should be commented out before running coverage.
#' @param test_map Data frame. Mapping of all tests within the package (used
#'   for diagnostic messaging and consistency checks).
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run.   
#'
#' @return 
#' A coverage object (a list) returned by `covr::environment_coverage()`,
#' filtered to remove problematic entries. If coverage fails, the function
#' returns `NULL`.
#'
#' The returned structure typically contains:
#' \describe{
#'   \item{filename}{Name of the file being measured.}
#'   \item{functions}{Function-level coverage information.}
#'   \item{linecoverage}{Line-by-line execution counts.}
#'   \item{srcref}{Source references extracted for coverage.}
#' }
#'
#' @details
#' If `test_skip` is `NULL`, failing tests must be detected elsewhere before
#' calling this function. This function does not perform test detection; it
#' only applies the skip mappings supplied in `test_skip`.
#'
#' @seealso
#'   \code{\link[testthat]{test_dir}},
#'   \code{\link[covr]{package_coverage}},
#'   \code{\link[covr]{environment_coverage}}
#' 
#' @keywords internal
#' @family stf_utility
#' @importFrom testthat test_dir
#' @importFrom covr package_coverage environment_coverage
#' @importFrom stats setNames
create_coverage_skip_stf <- function(
    pkg_source_path,
    pkg,
    test_path = "tests/testthat",
    test_skip,
    test_map,
    cov_env
) {
  
  message(paste0("Setting up coverage environment for : ", pkg))
  
  # --- Assertions (checkmate) ---
  checkmate::assert_string(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_string(pkg,            .var.name = "pkg")
  checkmate::assert_directory_exists(test_path, .var.name = "test_path")
  checkmate::assert_true(!missing(cov_env), .var.name = "cov_env")
  checkmate::assert_environment(cov_env,     .var.name = "cov_env")
  
  # --- Assertions: Validate cov_env$work_dir early and explicitly ---
  if (!base::exists("work_dir", envir = cov_env, inherits = FALSE)) {
    message(
      "Missing `cov_env$work_dir`. Please set a working directory in `cov_env` ",
      "that already exists and is writable (ideally a subdirectory of tempdir()). ",
      "Exiting create_coverage_skip_stf() early."
    )
    cleanup_and_return_null(env = cov_env)
    return(invisible(NULL))
  }
  work_dir <- base::get("work_dir", envir = cov_env, inherits = FALSE)
  checkmate::assert_string(work_dir, min.chars = 1, .var.name = "cov_env$work_dir")
  
  # Must exist on disk -> message + cleanup + early return (per request)
  if (!dir.exists(work_dir)) {
    message(
      "`cov_env$work_dir` does not exist on disk: ", work_dir,
      ". Exiting create_coverage_skip_stf() early."
    )
    cleanup_and_return_null(env = cov_env)
    return(invisible(NULL))
  }
  
  # Must be writable
  wd_norm <- normalizePath(work_dir, winslash = "/", mustWork = FALSE)
  if (base::file.access(wd_norm, 2) != 0) {
    stop("`cov_env$work_dir` is not writable: ", wd_norm)
  }
  
  # Prefer (but don't require) location under tempdir(), warn if not
  td_norm <- normalizePath(tempdir(), winslash = "/", mustWork = FALSE)
  starts_with_dir <- function(path, prefix) {
    path <- sub("/+$", "", path)
    prefix <- sub("/+$", "", prefix)
    identical(path, prefix) || startsWith(paste0(path, "/"), paste0(prefix, "/"))
  }
  if (!starts_with_dir(wd_norm, td_norm)) {
    warning(
      "`cov_env$work_dir` (", wd_norm, 
      ") is not located under tempdir() (", td_norm, "). ",
      "For CRAN-safety and to avoid side effects, consider using a subdirectory of tempdir()."
    )
  }
  # --- END assertion block ---
  
  
  # Pull optional metadata from cov_env if available
  helpers_loaded <- if (base::exists("helpers_loaded", envir = cov_env, inherits = FALSE))
    base::get("helpers_loaded", envir = cov_env, inherits = FALSE) else character(0)
  
  initial_cov_env_state <- if (base::exists("initial_state", envir = cov_env, inherits = FALSE))
    base::get("initial_state", envir = cov_env, inherits = FALSE) else
      base::ls(envir = cov_env, all.names = TRUE)
  
  root_dir <- if (base::exists("root_dir", envir = cov_env, inherits = FALSE))
    base::get("root_dir", envir = cov_env, inherits = FALSE) else
      normalizePath(pkg_source_path, winslash = "/", mustWork = FALSE)
  
  # Ensure we clean the env/workspace if the caller didn't (idempotent)
  on.exit({
    message(paste0("Cleaning up environment for ", pkg))
    cleanup_and_return_null(env = cov_env)
  }, add = TRUE)
  
  
  # Report loaded helpers
  if (length(helpers_loaded) > 0) {
    message("Helper files loaded:")
    for (helper in helpers_loaded) {
      message("  - ", helper)
    }
  } else {
    message("No helper files were loaded.")
  }
  
  # Identify lines to skip
  message(paste0("Identifying test lines to skip for : ", pkg))
  test_skip_lines <- lapply(
    split(test_skip, test_skip$file),
    function(x) {
      # keep only rows with valid finite line numbers
      ok <- is.finite(x$line1) & is.finite(x$line2)
      
      if (!any(ok)) {
        return(integer(0))
      }
      
      unlist(
        mapply(
          seq,
          from = x$line1[ok],
          to   = x$line2[ok],
          SIMPLIFY = FALSE
        ),
        use.names = FALSE
      )
    }
  )
  
  # -------------------------------------------------------------------
  # Prepare test files in a TEMPORARY workspace (CRAN-safe).
  # We never modify files under `tests/`; instead, copy to work_dir.
  # -------------------------------------------------------------------
  tmp_tests_dir <- file.path(cov_env$work_dir, "stf_tests")
  dir.create(tmp_tests_dir, recursive = TRUE, showWarnings = FALSE)
  
  # Helper to copy a single test file to temp (preserving name) and return its path
  copy_to_tmp <- function(src_path) {
    tgt_path <- file.path(tmp_tests_dir, basename(src_path))
    # use file.copy to temp; overwrite to be safe on re-runs
    ok <- file.copy(src_path, tgt_path, overwrite = TRUE)
    if (!ok) stop("Failed to copy test file to temp workspace: ", src_path)
    tgt_path
  }
  
  message(paste0("Preparing test files to check for : ", pkg))
  if (length(test_skip_lines) > 0) {
    # Only copy & modify the subset of test files that have skip ranges
    test_files <- file.path(test_path, names(test_skip_lines))
    tmp_files  <- vapply(test_files, copy_to_tmp, character(1), USE.NAMES = FALSE)
    names(tmp_files) <- names(test_skip_lines)
    
    # Apply "comment-out" on the TEMP copies only
    for (nm in names(test_skip_lines)) {
      file_path <- tmp_files[[nm]]
      test_lines <- readLines(file_path, warn = FALSE)
      lines_to_comment <- test_skip_lines[[nm]]
      # Robust bounds check
      lines_to_comment <- lines_to_comment[lines_to_comment >= 1 & lines_to_comment <= length(test_lines)]
      if (length(lines_to_comment)) {
        test_lines[lines_to_comment] <- sprintf("#%s", test_lines[lines_to_comment])
        cat(test_lines, file = file_path, sep = "\n")
      }
    }
    
    # Use the temp copies for coverage
    test_files <- unname(tmp_files)
    
  } else {
    # No specific skip ranges: copy all tests to temp and use those
    all_tests <- list.files(test_path, full.names = TRUE)
    test_files <- vapply(all_tests, copy_to_tmp, character(1))
  }
  
  # Run coverage
  message(paste0("Running environment coverage for ", pkg))
  cvr <- tryCatch({
    withr::with_dir(root_dir, {
      cov <- covr::environment_coverage(env = cov_env, test_files = test_files)
      # clean problematic entries (avoid devtools vapply bug)
      cov <- Filter(function(x) length(c(x$srcref, x$value)) == 9, cov)
      cov
    })
  }, error = function(e) {
    NULL
  })
  
  
  # Clean up newly introduced globals
  message(paste0("Removing new globals from environment for ", pkg))
  remove_new_globals(env = cov_env, initial_state = initial_cov_env_state)
  
  # Final cleanup
  
  # cleanup_and_return_null(env = cov_env)
  
  return(cvr)
}



#' Create a Summary of Coverage Results
#'
#' This function extracts overall test coverage and file‑level coverage
#' information from a covr coverage object. It provides a simplified
#' summary for downstream reporting and scoring.
#'
#' @param cvr A covr coverage object, typically produced by
#'   `covr::environment_coverage()` or by higher‑level wrapper functions in
#'   this package (e.g., skip‑aware coverage for STF or NSTF workflows).
#'
#' @return 
#' A data frame containing summary statistics, including:
#' \describe{
#'   \item{total_cov}{Numeric. Overall percentage of lines covered.}
#'   \item{name}{Character. File or package name associated with the coverage result.}
#'   \item{coverage}{List or data structure summarizing file‑level or
#'         element‑level coverage from the covr object.}
#'   \item{errors}{Character vector containing any coverage‑related errors (if present).}
#'   \item{notes}{Additional diagnostic notes returned by coverage processing.}
#' }
#'
#' The summary object is used by downstream reporting functions to
#' generate test coverage summaries and scoring metrics.
#'
#' @keywords internal
#' @importFrom dplyr %>%
create_results_summary <- function(cvr) {
  
  # check for empty cvr object
  if (is.null(cvr)) {
    # create filecoverage object
    filecoverage <- structure(
      c(NA_real_),
      dim = c(1),
      dimnames = list("no_coverage.R")
    )
    # create covr_list object
    covr_list <- list(
      total_cov = NA,
      res_cov = list(
        name = NA,
        coverage = list(
          filecoverage = filecoverage,
          totalcoverage = NA
        ),
        errors = NA,
        notes = NA
      )
    )
    
  } else {
    # Extract package name
    pkg <- attr(cvr, "package")$package
    # Extract overall coverage percentage
    total_cov <- covr::percent_coverage(cvr)
    
    # convert percentage into a score between 0 and 1
    total_cov <- total_cov / 100
    
    
    # Aggregate coverage by base filename
    file_coverage <- lapply(names(cvr), function(srcref_name) {
      entries <- cvr[[srcref_name]]
      values <- entries$value
      base_file <- sub(":.*", "", srcref_name)  # Extract base filename before first colon
      data.frame(filename = base_file, covered = values > 0)
    })
    
    file_coverage_df <- do.call(rbind, file_coverage)
    
    # Summarize coverage per base file
    file_summary <- file_coverage_df %>%
      dplyr::group_by(filename) %>%
      dplyr::summarise(
        percent = round(100 * sum(covered) / length(covered), 2),
        .groups = "drop"
      )
    
    # Create 1D numeric vector with dimnames (correct structure)
    filecoverage <- setNames(file_summary$percent, file_summary$filename)
    filecoverage <- structure(
      filecoverage,
      dim = c(length(filecoverage)),
      dimnames = list(names(filecoverage))
    )
    
    # Build result structure
    covr_list <- list(
      total_cov = total_cov,
      res_cov = list(
        name = pkg,
        coverage = list(
          filecoverage = filecoverage,
          totalcoverage = total_cov
        ),
        errors = NA,
        notes = NA
      )
    )
  }
  return(covr_list)
}


#' create covr_list when no tests are skipped
#'
#' @param test_map - R object with mapped tests
#' @param test_results_long - cvr object with no tests skipped
#' @param pkg_source_path - path to the package
#' @param pkg - name of the package
#' @param cov_env Environment that holds the temporary coverage workspace and
#'   shared state for the run.   
#'
#' @returns - summary cvr object
#' @keywords internal
#' @importFrom utils data
create_covr_list_no_skip <- function(test_map, 
                                     test_results_long, 
                                     pkg_source_path, 
                                     pkg,
                                     cov_env
                                     ) {
  
  # --- Input checks ---
  checkmate::assert_string(pkg_source_path, .var.name = "pkg_source_path")
  checkmate::assert_string(pkg,            .var.name = "pkg")
  checkmate::assert_true(!missing(cov_env), .var.name = "cov_env")
  checkmate::assert_environment(cov_env,     .var.name = "cov_env")
  
  # Attempt to unload package safely
  safe_unload_package(pkg)
  
  # Create environment for coverage
  cov_env <- new.env(parent = globalenv())
  
  # Load package into cov_env
  load_package_into_env(pkg_source_path, pkg, cov_env)
  
  # Detect datasets used in test files
  available_datasets <- utils::data(package = pkg)$results[, "Item"]
  
  test_files <- unique(file.path(pkg_source_path, "tests", "testthat", test_map$file))
  
  # Validate file paths
  valid_test_files <- Filter(function(f) {
    is.character(f) && length(f) == 1 && file.exists(f)
  }, test_files)
  
  
  used_datasets <- unique(unlist(lapply(valid_test_files, function(file) {
    lines <- readLines(file, warn = FALSE)
    matches <- grep(paste(available_datasets, collapse = "|"), lines, value = TRUE)
    regmatches(matches, gregexpr(paste(available_datasets, collapse = "|"), matches))
  })))
  
  
  # Flatten and clean dataset names
  used_datasets <- unique(unlist(used_datasets))
  used_datasets <- used_datasets[used_datasets %in% available_datasets]
  
  # Load datasets if needed
  if (length(used_datasets) > 0) {
    message(paste0("Loading datasets used in testing: ", paste(used_datasets, collapse = ", ")))
    for (ds in used_datasets) {
      tryCatch({
        utils::data(list = ds, package = pkg, envir = cov_env)
      }, error = function(e) {
        message(paste0("Failed to load dataset ", ds, " : ", e$message))
      })
    }
    
    message(paste0("Running coverage manually for ", pkg))
    
    coverage <- tryCatch({
      result <- covr::environment_coverage(
        env = cov_env,
        test_files = test_files
      )
      rm(list = ls(envir = cov_env), envir = cov_env)
      # clean problematic entries (avoid devtools vapply bug)
      result <- Filter(function(x) length(c(x$srcref, x$value)) == 9, result)
      result
    }, error = function(e) {
      cleanup_and_return_null(paste0("Error running coverage for ", pkg, " : ", e$message), env = cov_env)
    })
    
  } else {
    message(paste0("Creating package coverage for ", pkg))
    # unload package before running package_coverage
    try(unloadNamespace(pkg), silent = TRUE)
    coverage <- covr::package_coverage(path = pkg_source_path, type = "tests")
    rm(list = ls(envir = cov_env), envir = cov_env)
  }
  
  # Enrich each test entry with passed, failed, and skipped counts
  test_results_long <- lapply(test_results_long, function(entry) {
    results <- entry$results
    entry$passed  <- sum(vapply(results, inherits, logical(1), "expectation_success"))
    entry$failed  <- sum(vapply(results, inherits, logical(1), "expectation_failure"))
    entry$skipped <- sum(vapply(results, inherits, logical(1), "expectation_skip"))
    return(entry)
  })
  
  if (!is.null(coverage)) {
    # Extract coverage list
    coverage_list <- covr::coverage_to_list(coverage)
    filecoverage <- coverage_list$filecoverage
    totalcoverage <- coverage_list$totalcoverage
    
    total_passed <- sum(sapply(test_results_long, function(entry) entry$passed))
    total_failed <- sum(sapply(test_results_long, function(entry) entry$failed))
    tests_skipped <- Filter(function(entry) entry$skipped > 0, test_results_long)
    
    res_cov <- list(
      name = pkg,
      coverage = list(
        filecoverage = filecoverage,
        totalcoverage = totalcoverage
      ),
      errors = NA,
      notes = NA,
      passed = total_passed,
      failed = total_failed
    )
    
    total_cov <- if (!is.na(totalcoverage)) totalcoverage / 100 else NA_real_
    
    
    covr_list <- list(
      total_cov = total_cov,
      res_cov = res_cov,
      tests_skipped = tests_skipped
    )
   } else {
      message("No test coverage for this configuration")
      covr_list <- list(
        total_cov = 0,
        res_cov = list(
          name = pkg,
          coverage = list(
            filecoverage = matrix(0, nrow = 1, dimnames = list("No functions tested")),
            totalcoverage = 0
          ),
          errors = "No testthat or testit configuration",
          notes = NA
        )
      )
    } 
    
  cleanup_and_return_null(env = cov_env)
  
  return(covr_list)
}

#' Determine the appropriate standard test path for a package
#'
#' This internal utility function inspects the `test_pkg_data` structure to identify
#' which testing framework is used in the package and returns the corresponding test path.
#'
#' @param test_pkg_data A named list containing logical flags such as `has_testthat`, `has_testit`, `has_tests_base`, etc.
#' @param testdir A character string specifying the root directory where test folders are located.
#'
#' @return A character string representing the path to the test directory, or `NULL` if no known test framework is found.
#'
#' @keywords internal
get_stf_test_path <- function(test_pkg_data, testdir) {
  if (isTRUE(test_pkg_data$has_testthat)) {
    return(file.path(testdir, "testthat"))
  } else if (isTRUE(test_pkg_data$has_testit)) {
    return(file.path(testdir, "testit"))
  } else if (
    dir.exists(file.path(testdir, "test-ci")) &&
    dir.exists(file.path(testdir, "test-cran"))
  ) {
    return(testdir)  # fallback for nonstandard testit
  } else if (isTRUE(test_pkg_data$has_tests_base)) {
    return(testdir)  # base R scripts
  }
  
  return(NULL)
}


#' Detect covr's "no_coverage" sentinel
#'
#' @description
#' `is_no_coverage_cov()` returns `TRUE` when a covr coverage object contains
#' the special **no coverage** sentinel—i.e., a 1×1 `NA` `filecoverage` whose
#' sole filename (row name) is `"no_coverage.R"`. This commonly indicates that
#' instrumentation ran but no executable lines were attributed to the target, or
#' that the target test files did not exercise any package code.
#'
#' @details
#' The function is defensive across covr versions/shapes:
#' - It looks for `cov$coverage$filecoverage`, returning `FALSE` if missing.
#' - It accepts file names from either `dimnames(filecoverage)[[1]]` or
#'   `names(filecoverage)`.
#' - It positively identifies the sentinel *only* when **both** conditions hold:
#'   (1) the coverage array is length 1 and `NA`, and (2) the associated filename
#'   is `"no_coverage.R"`.
#'
#' @param cov A coverage object returned by `covr::environment_coverage()`,
#'   `covr::package_coverage()`, or a list-like object that exposes
#'   `$coverage$filecoverage`.
#'
#' @return A single `logical`: `TRUE` if the object represents “no coverage”,
#'   otherwise `FALSE`.
#' 
#' @keywords internal
#' @noRd
is_no_coverage_cov <- function(cov) {
  # Try to retrieve the filecoverage array safely
  
  fc <- tryCatch(cov$res_cov$coverage$filecoverage, error = function(e) NULL)
  if (!is.null(fc)) {
    # Derive file labels from dimnames (prefer) or names (fallback)
    dn <- dimnames(fc)
    files <- if (!is.null(dn) && length(dn) >= 1L && !is.null(dn[[1L]])) {
      dn[[1L]]
    } else {
      names(fc)
    }
  
    # Check for the precise sentinel shape and label
    is_len1_na <- (length(fc) == 1L) && is.na(fc[[1L]])
    has_no_cov_name <- isTRUE(length(files) >= 1L) && any(files %in% "no_coverage.R")
    
    if (!is.null(fc) && is_len1_na && has_no_cov_name) {
      no_cov_check <- TRUE
    } else {
      no_cov_check <- FALSE
    }  
  } else {
    no_cov_check <- FALSE
  }
  return(no_cov_check)
}

test_that("run_covr_skip_stf filters test results and generates skip map", {
  
  # Create a temporary directory to satisfy the assertion
  pkg_source_path <- file.path(tempdir(), "mockpkg")
  dir.create(pkg_source_path, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_source_path, "tests", "testthat"), recursive = TRUE, showWarnings = FALSE)
  
  # Mock long summary
  mock_test_results_long <- data.frame(
    file    = c("test-a.R", "test-b.R", "test-c.R"),
    context = c("script1", "script2", "script3"),
    test    = c("test 1", "test 2", "test 3"),
    status  = c("FAILED", "PASS", "SKIPPED"),
    n       = c(1, 1, 1),
    time    = c(0.01, 0.02, 0.03),
    stringsAsFactors = FALSE
  )
  
  # Mock test map
  mock_test_map <- data.frame(
    test = c("test 1", "test 2", "test 3"),
    expectation = c("expect_equal", "expect_true", "expect_error"),
    line1 = c(2, 4, 6),
    line2 = c(2, 4, 6),
    file  = c("test-a.R", "test-b.R", "test-c.R"),
    stringsAsFactors = FALSE
  )
  
  mock_test_skip <- data.frame(
    file  = c("test-a.R", "test-c.R"),
    line1 = c(2, 6),
    line2 = c(2, 6),
    stringsAsFactors = FALSE
  )
  
  # Minimal pkg data
  mock_test_pkg_data <- list(
    has_testthat = TRUE,
    has_testit = FALSE,
    has_tests_base = FALSE
  )
  
  # ---- STUBS (Complete) ----
  
  mockery::stub(run_covr_skip_stf, "checkmate::assert_string", TRUE)
  mockery::stub(run_covr_skip_stf, "checkmate::assert_directory_exists", TRUE)
  mockery::stub(run_covr_skip_stf, "checkmate::assert_environment", TRUE)
  mockery::stub(run_covr_skip_stf, "get_pkg_name", function(...) "mockpkg")
  
  # test path resolver
  mockery::stub(run_covr_skip_stf, "get_stf_test_path", function(x, testdir) testdir)
  
  # avoid changing working directory
  mockery::stub(run_covr_skip_stf, "setwd", function(...) NULL)
  
  # dummy covr env setup
  mockery::stub(run_covr_skip_stf, "setup_covr_env", function(...) {
    list(datasets_loaded=TRUE, root_dir=".", initial_state=list(), helpers_loaded=TRUE)
  })
  
  # unload namespace stub
  mockery::stub(run_covr_skip_stf, "unloadNamespace", function(...) NULL)
  
  # fail reporter stub
  mockery::stub(run_covr_skip_stf, "create_fail_reporter", function(test_path, pkg, cov_env) NULL)
  
  # create_test_reporter must return valid minimal objects
  mockery::stub(run_covr_skip_stf, "create_test_reporter", function(test_path, pkg, cov_env) {
    list(
      list(file="test-a.R", context="script1"),
      list(file="test-b.R", context="script2"),
      list(file="test-c.R", context="script3")
    )
  })
  
  # VERY IMPORTANT: fix_test_context must be stubbed
  mockery::stub(run_covr_skip_stf, "fix_test_context", function(x) x)
  
  mockery::stub(run_covr_skip_stf, "map_tests_stf", function(...) mock_test_map)
  mockery::stub(run_covr_skip_stf, "get_tests_long_sum_stf", function(...) mock_test_results_long)
  mockery::stub(run_covr_skip_stf, "get_tests_skip_stf", function(...) mock_test_skip)
  
  # coverage skip stub
  mockery::stub(run_covr_skip_stf, "create_coverage_skip_stf",
                function(pkg_source_path, pkg, test_path,
                         test_skip, test_map, cov_env) {
                  structure(list(package="mockpkg", coverage=85.5),
                            class="coverage")
                })
  
  # summary stub
  mockery::stub(run_covr_skip_stf, "create_results_summary",
                function(cvr) {
                  list(
                    total_cov = 85.5,
                    res_cov = list(
                      name="mockpkg",
                      coverage=85.5,
                      errors=NULL,
                      notes="Coverage summary generated"
                    )
                  )
                })
  
  # no-skip path stub (not used in this test)
  mockery::stub(run_covr_skip_stf, "create_covr_list_no_skip",
                function(...) list())
  
  # ---- EXECUTION ----
  cov_env <- new.env(parent = emptyenv())
  
  result <- run_covr_skip_stf(
    pkg_source_path = pkg_source_path,
    test_pkg_data = mock_test_pkg_data,
    cov_env = cov_env
  )
  
  # ---- ASSERTIONS ----
  
  expect_type(result, "list")
  expect_equal(result$total_cov, 85.5)
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$coverage, 85.5)
  expect_equal(result$res_cov$notes, "Coverage summary generated")
  
  expect_equal(result$test_skip, mock_test_skip)
  
  # new context column should match the original test column
  expect_equal(result$long_summary$context, mock_test_results_long$test)
  
  # new code_script column should match the original context column
  expect_equal(result$long_summary$code_script, mock_test_results_long$context)
  
})




test_that("run_covr_skip_stf processes no test dir and generates message", {
  
  # Simulate a package with no recognizable test directories
  mock_test_pkg_data <- list(
    has_testthat   = FALSE,
    has_testit     = FALSE,
    has_tests_base = FALSE
  )
  
  # Stubs required to reach the "no test dir" branch
  mockery::stub(run_covr_skip_stf, "checkmate::assert_string", TRUE)
  mockery::stub(run_covr_skip_stf, "checkmate::assert_directory_exists", TRUE)
  mockery::stub(run_covr_skip_stf, "checkmate::assert_environment", TRUE)
  
  mockery::stub(run_covr_skip_stf, "get_pkg_name", function(...) "mockpkg")
  mockery::stub(run_covr_skip_stf, "setwd", function(...) NULL)
  mockery::stub(run_covr_skip_stf, "message", function(...) NULL)
  
  # Critical: this must return NULL or the "no test directory" branch never triggers
  mockery::stub(run_covr_skip_stf, "get_stf_test_path", function(...) NULL)
  
  cov_env <- new.env(parent = emptyenv())
  
  # Execute function
  result <- run_covr_skip_stf(
    pkg_source_path = tempdir(),
    test_pkg_data   = mock_test_pkg_data,
    cov_env         = cov_env
  )
  
  # --- Assertions ---
  expect_type(result, "list")
  expect_named(result, c("total_cov", "res_cov"))
  
  expect_equal(result$total_cov, 0)
  
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$coverage$totalcoverage, 0)
  
  expect_equal(result$res_cov$errors, "No test directory found")
  expect_equal(result$res_cov$notes,  NA)
})



test_that("run_covr_skip_stf handles missing test directory correctly", {
  # Create a temporary directory to simulate a package source path
  temp_pkg_path <- tempfile("pkg")
  dir.create(temp_pkg_path)
  
  # Mock test_pkg_data to simulate no test directories
  mock_test_pkg_data <- list(
    has_testthat = FALSE,
    has_testit = FALSE,
    has_tests_base = FALSE
  )
  
  # Stub required dependencies
  mockery::stub(run_covr_skip_stf, "checkmate::assert_string", function(...) TRUE)
  mockery::stub(run_covr_skip_stf, "checkmate::assert_directory_exists", function(...) TRUE)
  mockery::stub(run_covr_skip_stf, "get_pkg_name", function(...) "dummyPkg")
  mockery::stub(run_covr_skip_stf, "setwd", function(...) NULL)
  mockery::stub(run_covr_skip_stf, "message", function(...) NULL)
  
  cov_env <- new.env(parent = emptyenv())
  
  # Execute function
  result <- run_covr_skip_stf(
    pkg_source_path = temp_pkg_path,
    test_pkg_data   = mock_test_pkg_data,
    cov_env         = cov_env
  )
  
  # Check that the result contains NAs as expected
  expect_type(result, "list")
  testthat::expect_named(result, c("total_cov", "res_cov"))
  testthat::expect_equal(result$res_cov$coverage$totalcoverage, 0)
  testthat::expect_equal(result$res_cov$errors, "No test directory found")
})


testthat::test_that("Fallback triggers and succeeds: package_coverage() is called and res_sum recomputed", {
  skip_on_cran()
  
  # ---- Arrange minimal on-disk structure so checkmate passes ----
  pkg_dir <- tempfile("pkgx_")
  dir.create(pkg_dir, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_dir, "tests"), showWarnings = FALSE)
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Build a 'no coverage' sentinel res_sum for first summary call ----
  sentinel_fc <- structure(NA_real_, .Dim = 1L, .Dimnames = list("no_coverage.R"))
  res_sum_no_cov <- list(
    total_cov = NA,
    res_cov = list(
      name     = NA,
      coverage = list(filecoverage = sentinel_fc, totalcoverage = NA),
      errors   = NA,
      notes    = NA
    )
  )
  
  # ---- Build a normal res_sum to be returned after fallback ----
  res_sum_ok <- list(
    total_cov = 0.42,
    res_cov = list(
      name     = "pkgx",
      coverage = list(filecoverage = c("R/a.R" = 1), totalcoverage = 42),
      errors   = NULL,
      notes    = NULL
    )
  )
  
  # ---- Create mocks/stubs for the long chain above the lines under test ----
  # Plumbing helpers
  mock_get_pkg_name        <- function(path) "pkgx"
  mock_get_stf_test_path   <- function(test_pkg_data, testdir) testdir
  mock_setup_covr_env      <- function(...) list(datasets_loaded = character(), root_dir = pkg_dir,
                                                 initial_state = character(), helpers_loaded = character())
  mock_create_fail_reporter <- function(...) invisible(NULL)
  mock_create_test_reporter <- function(...) list()  # not used downstream in our stubs
  mock_fix_test_context    <- function(x) x
  
  # Map + test results to force `failed <- TRUE`
  mock_map_tests_stf <- function(map_path) data.frame(test = "t1", stringsAsFactors = FALSE)
  mock_get_tests_long_sum_stf <- function(test_results) {
    data.frame(
      context = "ctx1", test = "t1",
      status  = "FAILED",
      file    = "tests/testthat/test-foo.R#L1",
      n = 1L, time = 0.01,
      stringsAsFactors = FALSE
    )
  }
  
  mock_get_tests_skip_stf      <- function(problem_tests, test_map) data.frame(
    file = "test-foo.R", line1 = 1L, line2 = 1L, stringsAsFactors = FALSE
  )
  mock_create_coverage_skip_stf <- function(...) "cvr_env"  # initial cvr from environment_coverage path
  
  # We stub dplyr helpers because rename/all_of are invoked but irrelevant here
  mock_dplyr_rename <- function(df, ...) df
  mock_dplyr_all_of <- function(x) x
  
  # ---- Mock create_results_summary to be called twice (before and after fallback) ----
  m_create_results_summary <- mockery::mock(res_sum_no_cov, res_sum_ok)
  
  # ---- Stub everything into run_covr_skip_stf() ----
  mockery::stub(run_covr_skip_stf, "get_pkg_name",               mock_get_pkg_name)
  mockery::stub(run_covr_skip_stf, "get_stf_test_path",          mock_get_stf_test_path)
  mockery::stub(run_covr_skip_stf, "setup_covr_env",             mock_setup_covr_env)
  mockery::stub(run_covr_skip_stf, "create_fail_reporter",       mock_create_fail_reporter)
  mockery::stub(run_covr_skip_stf, "create_test_reporter",       mock_create_test_reporter)
  mockery::stub(run_covr_skip_stf, "fix_test_context",           mock_fix_test_context)
  mockery::stub(run_covr_skip_stf, "map_tests_stf",              mock_map_tests_stf)
  mockery::stub(run_covr_skip_stf, "get_tests_long_sum_stf",     mock_get_tests_long_sum_stf)
  mockery::stub(run_covr_skip_stf, "get_tests_skip_stf",         mock_get_tests_skip_stf)
  mockery::stub(run_covr_skip_stf, "create_coverage_skip_stf",   mock_create_coverage_skip_stf)
  mockery::stub(run_covr_skip_stf, "dplyr::rename",              mock_dplyr_rename)
  mockery::stub(run_covr_skip_stf, "dplyr::all_of",              mock_dplyr_all_of)
  mockery::stub(run_covr_skip_stf, "create_results_summary",     m_create_results_summary)
  
  # withr::with_dir should evaluate the `code` argument and return its value
  mock_with_dir <- function(path, code) eval.parent(substitute(code))
  mockery::stub(run_covr_skip_stf, "withr::with_dir", mock_with_dir)
  
  # covr::package_coverage should be called by the fallback and return a new cvr
  mock_pkg_cov <- function(path, type, quiet) {
    testthat::expect_identical(type, "tests")
    "cvr_pkgcov"
  }
  mockery::stub(run_covr_skip_stf, "covr::package_coverage", mock_pkg_cov)
  
  # ---- Act & Assert ----
  # We expect the specific fallback message to be emitted
  testthat::expect_message(
    run_covr_skip_stf(pkg_source_path = pkg_dir, test_pkg_data = NULL, cov_env = cov_env),
    "Detected 'no_coverage\\.R' sentinel\\.",  # regex escape the dot
    perl = TRUE
  )
  
  # create_results_summary must have been called twice:
  mockery::expect_called(m_create_results_summary, 2)
  
  # 1st call should receive the cvr from create_coverage_skip_stf()
  mockery::expect_args(m_create_results_summary, 1, "cvr_env")
  
  # 2nd call should receive the cvr from covr::package_coverage()
  mockery::expect_args(m_create_results_summary, 2, "cvr_pkgcov")
})


testthat::test_that("Fallback triggers and errors: error handler path sets cvr <- NULL and res_sum recomputed", {
  skip_on_cran()
  
  # ---- Arrange minimal on-disk structure ----
  pkg_dir <- tempfile("pkgx_")
  dir.create(pkg_dir, recursive = TRUE, showWarnings = FALSE)
  dir.create(file.path(pkg_dir, "tests"), showWarnings = FALSE)
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Sentinel first summary ----
  sentinel_fc <- structure(NA_real_, .Dim = 1L, .Dimnames = list("no_coverage.R"))
  res_sum_no_cov <- list(
    total_cov = NA,
    res_cov = list(
      name     = NA,
      coverage = list(filecoverage = sentinel_fc, totalcoverage = NA),
      errors   = NA,
      notes    = NA
    )
  )
  
  # ---- Second summary (post-fallback) can be anything; we only check call args ----
  res_sum_after_error <- list(total_cov = NA_real_, res_cov = list(name = NA, coverage = list(), errors = NULL, notes = NULL))
  
  # ---- Stubs/mocks ----
  mockery::stub(run_covr_skip_stf, "get_pkg_name",             function(path) "pkgx")
  mockery::stub(run_covr_skip_stf, "get_stf_test_path",        function(test_pkg_data, testdir) testdir)
  mockery::stub(run_covr_skip_stf, "setup_covr_env",           function(...) list(datasets_loaded = character(), root_dir = pkg_dir, initial_state = character(), helpers_loaded = character()))
  mockery::stub(run_covr_skip_stf, "create_fail_reporter",     function(...) invisible(NULL))
  mockery::stub(run_covr_skip_stf, "create_test_reporter",     function(...) list())
  mockery::stub(run_covr_skip_stf, "fix_test_context",         function(x) x)
  mockery::stub(run_covr_skip_stf, "map_tests_stf",            function(map_path) data.frame(test = "t1", stringsAsFactors = FALSE))
  mockery::stub(run_covr_skip_stf, "get_tests_long_sum_stf",   function(test_results) {
    data.frame(
      context = "ctx1", test = "t1",
      status  = "FAILED",
      file    = "tests/testthat/test-foo.R#L1",
      n = 1L, time = 0.01,
      stringsAsFactors = FALSE
    )
  })
  mockery::stub(run_covr_skip_stf, "get_tests_skip_stf",       function(problem_tests, test_map) data.frame(file = "test-foo.R", line1 = 1L, line2 = 1L))
  mockery::stub(run_covr_skip_stf, "create_coverage_skip_stf", function(...) "cvr_env")
  
  # dplyr helpers neutralized
  mockery::stub(run_covr_skip_stf, "dplyr::rename",            function(df, ...) df)
  mockery::stub(run_covr_skip_stf, "dplyr::all_of",            function(x) x)
  
  # First summary -> sentinel, second summary -> arbitrary object
  m_create_results_summary <- mockery::mock(res_sum_no_cov, res_sum_after_error)
  mockery::stub(run_covr_skip_stf, "create_results_summary",   m_create_results_summary)
  
  # withr::with_dir evaluates the code argument
  mockery::stub(run_covr_skip_stf, "withr::with_dir",          function(path, code) eval.parent(substitute(code)))
  
  # Force package_coverage to error, exercising the error handler path
  mockery::stub(run_covr_skip_stf, "covr::package_coverage",   function(...) stop("boom"))
  
  # ---- Act & Assert ----
  # We expect the failure message from the error handler to appear
  testthat::expect_message(
    run_covr_skip_stf(pkg_source_path = pkg_dir, test_pkg_data = NULL, cov_env = cov_env),
    "package_coverage\\(\\) failed: boom",
    perl = TRUE
  )
  
  # create_results_summary must be called twice even on error
  mockery::expect_called(m_create_results_summary, 2)
  
  # 1st call receives the env coverage result
  mockery::expect_args(m_create_results_summary, 1, "cvr_env")
  
  # 2nd call receives NULL (from the error handler)
  mockery::expect_args(m_create_results_summary, 2, NULL)
})


test_that("returns no-skip coverage result when all tests pass", {
  # ------------------------------------------------------------------
  # Fake mapping and summary objects
  # ------------------------------------------------------------------
  fake_test_map <- data.frame(
    test = rep(c("two working tests", "three working tests"), each = 2),
    expectation = rep("expect_equal", 4),
    line1 = c(2, 4, 8, 10),
    line2 = c(2, 4, 8, 10),
    file = rep("test-myscript1.R", 4),
    stringsAsFactors = FALSE
  )
  
  fake_test_results_long <- data.frame(
    file    = rep("test-myscript1.R#L", 2),
    context = rep("myscript1", 2),
    test    = c("two working tests", "three working tests"),
    status  = rep("PASS", 2),
    n       = c(2, 3),
    time    = c(0.01, 0.01),
    stringsAsFactors = FALSE
  )
  
  # ------------------------------------------------------------------
  # Real temporary package dir + cov_env
  # ------------------------------------------------------------------
  pkg_dir <- withr::local_tempdir()
  cov_env <- new.env(parent = emptyenv())
  
  # Pull function from namespace so stubs attach correctly
  rcs <- get("run_covr_skip_stf", envir = asNamespace("test.assessr"))
  
  # Capture what create_covr_list_no_skip() receives
  seen_args <- NULL
  
  # ------------------------------------------------------------------
  # Stubs
  # ------------------------------------------------------------------
  
  # Package/test-path helpers
  mockery::stub(rcs, "get_pkg_name", function(...) "mockpkg")
  mockery::stub(rcs, "get_stf_test_path", function(...) "tests/testthat")
  
  # New setup path: return all fields run_covr_skip_stf may stash onto cov_env
  mockery::stub(
    rcs,
    "setup_covr_env",
    function(...) {
      list(
        cov_env = cov_env,
        datasets_loaded = character(0),
        base_r_datasets_loaded = character(0),
        base_r_exports_loaded = character(0),
        root_dir = pkg_dir,
        initial_state = character(0),
        helpers_loaded = character(0)
      )
    }
  )
  
  # Reporters
  mockery::stub(
    rcs,
    "create_fail_reporter",
    function(...) {
      list()
    }
  )
  
  mockery::stub(
    rcs,
    "create_test_reporter",
    function(...) {
      list(
        list(file = "test-a.R", context = "script1"),
        list(file = "test-b.R", context = "script2")
      )
    }
  )
  
  # Downstream processing helpers
  mockery::stub(rcs, "fix_test_context", function(x) x)
  mockery::stub(rcs, "map_tests_stf", function(...) fake_test_map)
  mockery::stub(rcs, "get_tests_long_sum_stf", function(...) fake_test_results_long)
  
  # Stub the no-skip coverage creator and capture inputs
  mockery::stub(
    rcs,
    "create_covr_list_no_skip",
    function(test_map,
             test_results_long,
             pkg_source_path,
             pkg,
             cov_env,
             ...) {
      
      seen_args <<- list(
        test_map = test_map,
        test_results_long = test_results_long,
        pkg_source_path = pkg_source_path,
        pkg = pkg,
        cov_env = cov_env
      )
      
      list(
        total_cov = 0.5,
        res_cov = list(
          name = pkg,
          coverage = list(
            filecoverage = c("test-myscript1.R" = 50),
            totalcoverage = 50
          ),
          errors = NA,
          notes = NA,
          passed = 5,
          failed = 0
        ),
        tests_skipped = list()
      )
    }
  )
  
  mockery::stub(rcs, "unloadNamespace", function(...) NULL)
  
  # ------------------------------------------------------------------
  # Run
  # ------------------------------------------------------------------
  result <- rcs(
    pkg_source_path = pkg_dir,
    test_pkg_data   = NULL,
    cov_env         = cov_env
  )
  
  # ------------------------------------------------------------------
  # Assertions
  # ------------------------------------------------------------------
  expect_true(is.list(result))
  expect_named(result, c("total_cov", "res_cov", "tests_skipped"))
  
  expect_equal(result$total_cov, 0.5)
  expect_equal(result$res_cov$name, "mockpkg")
  expect_equal(result$res_cov$passed, 5)
  expect_equal(result$res_cov$failed, 0)
  expect_identical(result$tests_skipped, list())
  
  # Ensure the no-skip builder got the expected inputs
  expect_identical(seen_args$pkg_source_path, pkg_dir)
  expect_identical(seen_args$pkg, "mockpkg")
  expect_identical(seen_args$cov_env, cov_env)
  
  # Ensure setup metadata was stashed onto cov_env
  expect_true(exists("datasets_loaded", envir = cov_env, inherits = FALSE))
  expect_true(exists("root_dir", envir = cov_env, inherits = FALSE))
  expect_true(exists("initial_state", envir = cov_env, inherits = FALSE))
  expect_true(exists("helpers_loaded", envir = cov_env, inherits = FALSE))
  
  expect_identical(get("root_dir", envir = cov_env, inherits = FALSE), pkg_dir)
})


test_that("create_coverage_skip_stf runs coverage with datasets", {
  # --- Arrange ---
  pkg             <- "mockpkg"
  pkg_source_path <- file.path(tempdir(), "mockpkg-root")
  test_path       <- file.path(pkg_source_path, "tests", "testthat")
  dir.create(test_path, recursive = TRUE, showWarnings = FALSE)
  
  # Create the *actual* source test file that the function will copy
  src_test_file <- file.path(test_path, "test-file.R")
  writeLines(c(
    "test_that('example', {",
    "  expect_true(TRUE)",
    "})"
  ), src_test_file)
  
  test_skip <- data.frame(file = "test-file.R", line1 = 1, line2 = 2, stringsAsFactors = FALSE)
  test_map  <- data.frame(file = "test-file.R", stringsAsFactors = FALSE)
  
  # The function expects a working directory in cov_env
  cov_env <- new.env(parent = emptyenv())
  cov_env$work_dir <- file.path(tempdir(), "cov-work")
  dir.create(cov_env$work_dir, recursive = TRUE, showWarnings = FALSE)
  
  # --- Stubs (minimal, let real IO happen where safe) ---
  # Let assertions pass (directories exist anyway)
  mockery::stub(create_coverage_skip_stf, "checkmate::assert_string", TRUE)
  mockery::stub(create_coverage_skip_stf, "checkmate::assert_directory_exists", TRUE)
  mockery::stub(create_coverage_skip_stf, "checkmate::assert_environment", TRUE)
  
  # Keep project root stable
  mockery::stub(create_coverage_skip_stf, "find_project_root", function(...) pkg_source_path)
  
  # Avoid changing the working directory for the test
  mockery::stub(create_coverage_skip_stf, "withr::with_dir", function(new_dir, code) force(code))
  
  # No-op cleanup if used internally
  mockery::stub(create_coverage_skip_stf, "unlink", function(x, recursive = FALSE, force = FALSE) 0)
  
  # Coverage stub that *survives* the Filter(length(c(srcref,value)) == 9)
  # Here srcref has 5 parts and value has 4 parts -> 9 total.
  mock_cov_list <- list(list(
    srcref = as.list(1:5),
    value  = as.list(1:4)
  ))
  mockery::stub(create_coverage_skip_stf, "covr::environment_coverage", function(env, test_files) {
    mock_cov_list
  })
  
  # --- Act ---
  result <- create_coverage_skip_stf(
    pkg_source_path = pkg_source_path,
    pkg             = pkg,
    test_path       = test_path,
    test_skip       = test_skip,
    test_map        = test_map,
    cov_env         = cov_env
  )
  
  # --- Assert ---
  # After the Filter step, the object is a plain list (class "coverage" is not preserved).
  expect_true(is.list(result))
  expect_length(result, 1L)
  expect_true(all(c("srcref", "value") %in% names(result[[1]])))
})


test_that("create_results_summary returns correct structure and values", {
  # Mock coverage object
  mock_cvr <- list(
    "file1.R:1" = list(value = c(1, 0, 1)),
    "file2.R:2" = list(value = c(0, 0, 1)),
    "file1.R:3" = list(value = c(1, 1))
  )
  
  # Add package attribute
  attr(mock_cvr, "package") <- list(package = "mockpkg")
  
  # Stub covr::percent_coverage to return a fixed value
  mockery::stub(create_results_summary, "covr::percent_coverage", function(x) 75.0)
  
  # Run the function
  result <- create_results_summary(mock_cvr)
  
  # Check top-level structure
  expect_type(result, "list")
  expect_named(result, c("total_cov", "res_cov"))
  
  # Check total coverage
  expect_equal(result$total_cov, 0.75)
  
  # Check res_cov structure
  expect_type(result$res_cov, "list")
  expect_equal(result$res_cov$name, "mockpkg")
  expect_type(result$res_cov$coverage, "list")
  
  # Check file coverage vector
  filecoverage <- result$res_cov$coverage$filecoverage
  expect_true(is.numeric(filecoverage))
  expect_true(!is.null(dimnames(filecoverage)))
  
  # Check that coverage values are correct
  expect_equal(names(filecoverage), c("file1.R", "file2.R"))
  
  expect_equal(filecoverage["file1.R"], round(100 * 4/5, 2), ignore_attr = TRUE)
  expect_equal(filecoverage["file2.R"], round(100 * 1/3, 2), ignore_attr = TRUE)
  
  
  # Check notes and errors
  expect_true(is.na(result$res_cov$errors))
  expect_true(is.na(result$res_cov$notes))
})


test_that("create_results_summary handles NULL input correctly", {
  # Run the function with NULL input
  result <- create_results_summary(NULL)
  
  # Check top-level structure
  expect_type(result, "list")
  expect_named(result, c("total_cov", "res_cov"))
  
  # Check that total_cov is NA
  expect_true(is.na(result$total_cov))
  
  # Check res_cov structure
  expect_type(result$res_cov, "list")
  expect_true(is.na(result$res_cov$name))
  expect_type(result$res_cov$coverage, "list")
  
  # Check filecoverage is an empty numeric vector with dimnames
  filecoverage <- result$res_cov$coverage$filecoverage
  expect_true(is.numeric(filecoverage))
  expect_equal(length(filecoverage), 1)
  expect_equal(dim(filecoverage), c(1))
  expect_true(is.list(dimnames(filecoverage)))
  expect_equal(length(dimnames(filecoverage)), 1)
  expect_equal(dimnames(filecoverage)[[1]], "no_coverage.R")
  
  
  # Check totalcoverage is NA
  expect_true(is.na(result$res_cov$coverage$totalcoverage))
  
  # Check notes and errors
  expect_true(is.na(result$res_cov$errors))
  expect_true(is.na(result$res_cov$notes))
})

test_that("create_covr_list_no_skip returns structured coverage list", {
  # Mock inputs
  test_map <- data.frame(file = c("test-one.R", "test-two.R"), stringsAsFactors = FALSE)
  test_results_long <- list(
    list(results = list(structure(TRUE, class = "expectation_success"))),
    list(results = list(structure(TRUE, class = "expectation_failure")))
  )
  pkg_source_path <- "mock/path"
  pkg <- "mockpkg"
  
  # ---- Stubs inside create_covr_list_no_skip ----
  # Unload is a no-op
  mockery::stub(create_covr_list_no_skip, "safe_unload_package", function(pkg) TRUE)
  
  # Simulate package load into env
  mockery::stub(
    create_covr_list_no_skip,
    "load_package_into_env",
    function(pkg_source_path, pkg, cov_env) TRUE
  )
  
  # Simulate available datasets via `data()`
  # Keep symbol name as used in the function body: `data`, not `utils::data`
  mockery::stub(
    create_covr_list_no_skip,
    "utils::data",
    function(package, ...) {
      list(results = data.frame(Item = c("iris", "mtcars"), stringsAsFactors = FALSE))
    }
  )
  
  
  # Pretend all referenced test files exist
  mockery::stub(create_covr_list_no_skip, "file.exists", function(path) TRUE)
  
  # Fake test file contents that reference datasets
  mockery::stub(
    create_covr_list_no_skip,
    "readLines",
    function(file, warn = FALSE) c("data(iris)", "data(mtcars)")
  )
  
  # Covr coverage over environment
  mock_coverage <- structure(list(), class = "coverage")
  mockery::stub(
    create_covr_list_no_skip,
    "covr::environment_coverage",
    function(...) mock_coverage
  )
  
  # Convert coverage to list
  mockery::stub(
    create_covr_list_no_skip,
    "covr::coverage_to_list",
    function(x) list(filecoverage = matrix(1, nrow = 1), totalcoverage = 85)
  )
  
  # Cleanup helper returns NULL when asked
  mockery::stub(create_covr_list_no_skip, "cleanup_and_return_null", function(...) NULL)
  
  cov_env <- new.env(parent = emptyenv())
  
  # ---- Run with the new argument explicitly passed ----
  result <- create_covr_list_no_skip(
    test_map = test_map,
    test_results_long = test_results_long,
    pkg_source_path = pkg_source_path,
    pkg = pkg,
    cov_env = cov_env
  )
  
  # ---- Assertions ----
  expect_type(result, "list")
  expect_equal(result$total_cov, 0.85)
  expect_equal(result$res_cov$name, pkg)
  expect_equal(result$res_cov$passed, 1)
  expect_equal(result$res_cov$failed, 1)
  expect_true(is.list(result$tests_skipped))
})


test_that("create_covr_list_no_skip falls back to package_coverage when no datasets used", {
  # Mock inputs
  test_map <- data.frame(file = "test-file.R", stringsAsFactors = FALSE)
  
  # Create mock expectations using official helpers
  mock_success <- testthat::expect_success(expect_true(TRUE))
  mock_failure <- testthat::expect_failure(expect_true(FALSE))
  
  test_results_long <- list(list(results = list(mock_success, mock_failure)))
  
  
  pkg_source_path <- "fake/path"
  pkg <- "fakepkg"
  
  # Stub safe_unload_package and load_package_into_env to do nothing
  mockery::stub(create_covr_list_no_skip, "safe_unload_package", function(pkg) NULL)
  mockery::stub(create_covr_list_no_skip, "load_package_into_env", function(path, pkg, env) NULL)
  
  # Stub data() to return no datasets
  mockery::stub(
    create_covr_list_no_skip,
    "utils::data",
    function(package, ...) {
      list(results = data.frame(Item = character(0), stringsAsFactors = FALSE))
    }
  )
  
  
  # Stub file.exists to always return TRUE
  mockery::stub(create_covr_list_no_skip, "file.exists", function(f) TRUE)
  
  # Stub readLines to return lines that do not match any dataset
  mockery::stub(create_covr_list_no_skip, "readLines", function(file, warn) c("some unrelated code"))
  
  # Stub covr::package_coverage to return a mock coverage object
  mock_coverage <- structure(list(), class = "coverage")
  mockery::stub(create_covr_list_no_skip, "covr::package_coverage", function(path, type) mock_coverage)
  
  # Stub covr::coverage_to_list to return mock coverage list
  mockery::stub(create_covr_list_no_skip, "covr::coverage_to_list", function(cov) {
    list(filecoverage = list(), totalcoverage = 85)
  })
  
  # Stub cleanup_and_return_null to do nothing
  mockery::stub(create_covr_list_no_skip, "cleanup_and_return_null", function(..., env) NULL)
  
  cov_env <- new.env(parent = emptyenv())
  # Run the function
  result <- create_covr_list_no_skip(test_map, 
                                     test_results_long, 
                                     pkg_source_path, 
                                     pkg, 
                                     cov_env = cov_env
  )
  
  # Assertions
  expect_type(result, "list")
  expect_equal(result$total_cov, 0.85)
  expect_equal(result$res_cov$name, pkg)
  expect_equal(result$res_cov$coverage$totalcoverage, 85)
})


test_that("create_covr_list_no_skip handles NULL coverage", {
  
  # Mock inputs
  test_map <- data.frame(file = "test-file.R", stringsAsFactors = FALSE)
  
  # Use valid expectation objects
  mock_success <- testthat::expect_success(expect_true(TRUE))
  mock_failure <- testthat::expect_failure(expect_true(FALSE))
  mock_skip    <- structure(list(), class = "expectation_skip")
  
  test_results_long <- list(
    list(results = list(mock_success, mock_failure, mock_skip))
  )
  
  pkg_source_path <- tempdir()
  pkg <- "mockpkg"
  
  # ----------- REQUIRED STUBS -------------
  
  # No-op package unloading
  mockery::stub(create_covr_list_no_skip, "safe_unload_package", function(...) NULL)
  
  # No-op environment loader
  mockery::stub(create_covr_list_no_skip, "load_package_into_env", function(...) NULL)
  
  # CRITICAL: stub utils::data(), not data()
  mockery::stub(
    create_covr_list_no_skip,
    "utils::data",
    function(package, ...) {
      list(results = data.frame(Item = character(0), stringsAsFactors = FALSE))
    }
  )
  
  # Pretend all referenced files exist
  mockery::stub(create_covr_list_no_skip, "file.exists", function(...) TRUE)
  
  # No datasets referenced in test files
  mockery::stub(create_covr_list_no_skip, "readLines", function(...) "some unrelated code")
  
  # Force NULL coverage from covr
  mockery::stub(create_covr_list_no_skip, "covr::package_coverage", function(...) NULL)
  
  # cleanup helper as no-op
  mockery::stub(create_covr_list_no_skip, "cleanup_and_return_null", function(...) NULL)
  
  # ----------------------------------------
  
  cov_env <- new.env(parent = emptyenv())
  
  # Run function using correct signature (named args)
  result <- create_covr_list_no_skip(
    test_map          = test_map,
    test_results_long = test_results_long,
    pkg_source_path   = pkg_source_path,
    pkg               = pkg,
    cov_env           = cov_env
  )
  
  # ---- Assertions for the fallback NULL-coverage branch ----
  expect_type(result, "list")
  expect_equal(result$total_cov, 0)
  
  expect_equal(result$res_cov$coverage$totalcoverage, 0)
  
  expect_equal(
    result$res_cov$coverage$filecoverage,
    matrix(0, nrow = 1, dimnames = list("No functions tested"))
  )
  
  expect_equal(result$res_cov$errors, "No testthat or testit configuration")
})


test_that("create_coverage_skip_stf reports no helpers when none exist", {
  pkg_name <- "mockpkg"
  pkg_dir  <- tempdir()
  
  # Create the test directory expected by the function
  test_path <- file.path(pkg_dir, "tests", "testthat")
  dir.create(test_path, recursive = TRUE, showWarnings = FALSE)
  
  # Create an empty test file that matches test_skip$file
  writeLines("test_that('x',{ expect_true(TRUE) })",
             file.path(test_path, "test.R"))
  
  # ---- cov_env must exist AND must contain work_dir ----
  cov_env <- new.env(parent = emptyenv())
  cov_env$work_dir <- file.path(tempdir(), "work")
  dir.create(cov_env$work_dir, recursive = TRUE, showWarnings = FALSE)
  
  # ---- Stubs ----
  # NOTE: setup_covr_env is NOT used in this function (remove your stub)
  mockery::stub(create_coverage_skip_stf, "cleanup_and_return_null", function(...) NULL)
  mockery::stub(create_coverage_skip_stf, "remove_new_globals", function(...) NULL)
  
  mockery::stub(create_coverage_skip_stf, "covr::environment_coverage",
                function(...) structure(list(), class = "coverage"))
  
  # Prevent actual file mutation during test
  mockery::stub(create_coverage_skip_stf, "readLines", function(...) character())
  mockery::stub(create_coverage_skip_stf, "cat", function(...) NULL)
  
  mockery::stub(create_coverage_skip_stf, "safe_unload_package", function(...) NULL)
  
  # Do not actually change directories
  mockery::stub(create_coverage_skip_stf, "withr::with_dir", function(path, code) force(code))
  
  # ---- Capture messages ----
  msgs <- testthat::capture_messages({
    create_coverage_skip_stf(
      pkg_source_path = pkg_dir,
      pkg             = pkg_name,
      test_path       = test_path,
      test_skip       = data.frame(file = "test.R", line1 = 1, line2 = 1),
      test_map        = data.frame(file = "test.R"),
      cov_env         = cov_env
    )
  })
  
  # ---- Assertion ----
  expect_true(any(grepl("No helper files were loaded\\.", msgs)))
})


test_that("create_coverage_skip_stf reports loaded helpers when helpers exist", {
  pkg_name <- "mockpkg"
  pkg_dir  <- tempdir()
  
  # ------------------------------
  # Create real test directory + file
  # ------------------------------
  test_path <- file.path(pkg_dir, "tests", "testthat")
  dir.create(test_path, recursive = TRUE, showWarnings = FALSE)
  
  writeLines(
    "test_that('x',{ expect_true(TRUE) })",
    file.path(test_path, "test.R")
  )
  
  # ------------------------------
  # Create cov_env with required fields
  # ------------------------------
  cov_env <- new.env(parent = emptyenv())
  cov_env$helpers_loaded <- c("helper1.R", "helper2.R")
  cov_env$root_dir       <- pkg_dir
  cov_env$initial_state  <- character()
  cov_env$work_dir       <- file.path(tempdir(), "work")
  dir.create(cov_env$work_dir, recursive = TRUE, showWarnings = FALSE)
  
  # ------------------------------
  # Stubs for dependencies
  # ------------------------------
  mockery::stub(create_coverage_skip_stf, "remove_new_globals", function(...) NULL)
  mockery::stub(create_coverage_skip_stf, "cleanup_and_return_null", function(...) NULL)
  
  mockery::stub(
    create_coverage_skip_stf,
    "covr::environment_coverage",
    function(...) structure(list(), class = "coverage")
  )
  
  # Avoid file writing during comment-out
  mockery::stub(create_coverage_skip_stf, "readLines", function(...) character())
  mockery::stub(create_coverage_skip_stf, "cat", function(...) NULL)
  
  # Avoid directory switching
  mockery::stub(create_coverage_skip_stf, "withr::with_dir", function(path, code) force(code))
  
  # ------------------------------
  # Capture messages
  # ------------------------------
  msgs <- testthat::capture_messages({
    create_coverage_skip_stf(
      pkg_source_path = pkg_dir,
      pkg             = pkg_name,
      test_path       = test_path,
      test_skip       = data.frame(file = "test.R", line1 = 1, line2 = 1),
      test_map        = data.frame(file = "test.R"),
      cov_env         = cov_env
    )
  })
  
  # ------------------------------
  # Assertions
  # ------------------------------
  expect_true(any(grepl("Helper files loaded:", msgs)))
  expect_true(any(grepl("helper1.R", msgs)))
  expect_true(any(grepl("helper2.R", msgs)))
})


test_that("create_test_reporter returns mocked test results", {
  # Mocked test results
  fake_results <- list(
    list(file = "test-a.R", context = "script1"),
    list(file = "test-b.R", context = "script2")
  )
  
  # Stub testthat::test_dir to return fake results
  mockery::stub(create_test_reporter, "testthat::test_dir", function(...) fake_results)
  
  # Stub fix_test_context to just return the same results
  mockery::stub(create_test_reporter, "fix_test_context", function(x) x)
  
  # Call the function
  result <- create_test_reporter(test_path = "tests", pkg = "mockpkg")
  
  # Assertions
  expect_type(result, "list")
  expect_equal(length(result), 2)
  expect_equal(result[[1]]$context, "script1")
  expect_equal(result[[2]]$file, "test-b.R")
})

test_that("create_test_reporter handles errors gracefully", {
  # Stub testthat::test_dir to throw an error
  mockery::stub(create_test_reporter, "testthat::test_dir", function(...) stop("Simulated error"))
  
  # Call the function
  result <- create_test_reporter(test_path = "tests", pkg = "mockpkg")
  
  # Assertions
  expect_null(result) # Should return NULL on error
})


test_that("fix_test_context sets context when NULL", {
  # Mock input: context is NULL
  test_results <- list(list(file = "test-script.R", context = NULL))
  
  # Stub tools::file_path_sans_ext to return predictable value
  mockery::stub(fix_test_context, "tools::file_path_sans_ext", function(path) "test-script")
  
  # Call function
  result <- fix_test_context(test_results)
  
  # Assertions
  expect_equal(result[[1]]$context, "test-script")
})

test_that("fix_test_context sets context when empty string", {
  # Mock input: context is empty
  test_results <- list(list(file = "test-script.R", context = ""))
  
  # Stub tools::file_path_sans_ext
  mockery::stub(fix_test_context, "tools::file_path_sans_ext", function(path) "test-script")
  
  # Call function
  result <- fix_test_context(test_results)
  
  # Assertions
  expect_equal(result[[1]]$context, "test-script")
})

test_that("fix_test_context does not change context when already set", {
  # Mock input: context already set
  test_results <- list(list(file = "test-script.R", context = "existing-context"))
  
  # Stub tools::file_path_sans_ext (should not be called)
  mockery::stub(fix_test_context, "tools::file_path_sans_ext", function(path) "test-script")
  
  # Call function
  result <- fix_test_context(test_results)
  
  # Assertions
  expect_equal(result[[1]]$context, "existing-context")
})

test_that("get_stf_test_path returns correct paths based on test_pkg_data", {
  testdir <- tempfile()
  dir.create(testdir)
  
  # Create dummy subdirectories
  dir.create(file.path(testdir, "testthat"))
  dir.create(file.path(testdir, "testit"))
  dir.create(file.path(testdir, "test-ci"))
  dir.create(file.path(testdir, "test-cran"))
  
  # Case: has_testthat
  test_pkg_data <- list(has_testthat = TRUE, has_testit = FALSE, has_tests_base = FALSE)
  expect_equal(get_stf_test_path(test_pkg_data, testdir), file.path(testdir, "testthat"))
  
  # Case: has_testit
  test_pkg_data <- list(has_testthat = FALSE, has_testit = TRUE, has_tests_base = FALSE)
  expect_equal(get_stf_test_path(test_pkg_data, testdir), file.path(testdir, "testit"))
  
  # Case: fallback for nonstandard testit
  test_pkg_data <- list(has_testthat = FALSE, has_testit = FALSE, has_tests_base = FALSE)
  expect_equal(get_stf_test_path(test_pkg_data, testdir), testdir)
  
  # Case: has_tests_base
  unlink(file.path(testdir, "test-ci"), recursive = TRUE)
  unlink(file.path(testdir, "test-cran"), recursive = TRUE)
  test_pkg_data <- list(has_testthat = FALSE, has_testit = FALSE, has_tests_base = TRUE)
  expect_equal(get_stf_test_path(test_pkg_data, testdir), testdir)
  
  # Case: no known test framework
  test_pkg_data <- list(has_testthat = FALSE, has_testit = FALSE, has_tests_base = FALSE)
  expect_null(get_stf_test_path(test_pkg_data, testdir))
  
  # Cleanup
  unlink(testdir, recursive = TRUE)
})

test_that("create_fail_reporter returns results on success", {
  pkg <- "mockpkg"
  test_path <- tempdir()
  
  # 1. Create a mock for the FailReporter results
  mock_results <- list(
    list(file = "test-a.R", context = "script1", status = "FAILED"),
    list(file = "test-b.R", context = "script2", status = "PASS")
  )
  
  # 2. Mock the FailReporter R6 object
  # We stub the 'new' method to return an object with our fake results
  mock_rep <- list(results = mock_results)
  mockery::stub(create_fail_reporter, "testthat::FailReporter$new", mock_rep)
  
  # 3. Stub test_dir to do nothing (preventing the DESCRIPTION/path warning)
  mockery::stub(create_fail_reporter, "testthat::test_dir", function(...) NULL)
  
  # Execute
  result <- create_fail_reporter(test_path, pkg)
  
  # Assertions
  expect_equal(result, mock_results)
  expect_length(result, 2)
})

test_that("create_fail_reporter handles test_dir errors gracefully", {
  pkg <- "mockpkg"
  test_path <- "non/existent/path"
  
  # Stub FailReporter to return a valid object structure
  mockery::stub(create_fail_reporter, "testthat::FailReporter$new", list(results = list()))
  
  # Stub test_dir to throw an error
  mockery::stub(create_fail_reporter, "testthat::test_dir", function(...) {
    stop("Subprocess failure simulation")
  })
  
  # Execute and capture message
  expect_message(
    result <- create_fail_reporter(test_path, pkg),
    regexp = "Error running tests for mockpkg : Subprocess failure simulation"
  )
  
  # Assertions
  expect_null(result)
})


# ---- Helpers ----

# Create a cov_env with optional bindings
make_cov_env <- function(work_dir = NULL, helpers_loaded = character(0),
                         initial_state = character(0), root_dir = NULL) {
  e <- new.env(parent = emptyenv())
  if (!is.null(work_dir)) e$work_dir <- work_dir
  if (!is.null(helpers_loaded)) e$helpers_loaded <- helpers_loaded
  if (!is.null(initial_state)) e$initial_state <- initial_state
  if (!is.null(root_dir)) e$root_dir <- root_dir
  e
}

# Create a minimal test fixture with one dummy test file
make_test_fixture <- function() {
  base <- tempfile("pkgsrc_")
  dir.create(base, recursive = TRUE, showWarnings = FALSE)
  testthat_dir <- file.path(base, "tests", "testthat")
  dir.create(testthat_dir, recursive = TRUE, showWarnings = FALSE)
  tf <- file.path(testthat_dir, "test-dummy.R")
  writeLines("test_that('dummy', { expect_true(TRUE) })", tf)
  list(base = base, testthat = testthat_dir, testfile = tf)
}

# Build a minimal `test_skip` with proper columns but no rows
empty_test_skip <- function() {
  data.frame(file = character(), line1 = integer(), line2 = integer())
}

# ---- Tests ----

# 1) Missing cov_env$work_dir -> message + cleanup_and_return_null + invisible(NULL)
testthat::test_that("missing work_dir -> message, cleanup, invisible NULL", {
  cov_env <- make_cov_env(work_dir = NULL)  # No work_dir binding
  fx <- create_coverage_skip_stf
  cleanup_mock <- mockery::mock()
  
  # Stub the cleanup function used by early exit
  mockery::stub(fx, "cleanup_and_return_null", cleanup_mock)
  
  fixture <- make_test_fixture()
  test_skip <- empty_test_skip()
  
  testthat::expect_message(
    testthat::expect_invisible({
      res <- fx(
        pkg_source_path = fixture$base,
        pkg = "dummyPkg",
        test_path = fixture$testthat,
        test_skip = test_skip,
        test_map = NULL,
        cov_env = cov_env
      )
      testthat::expect_null(res)
    }),
    regexp = "Missing `cov_env\\$work_dir`.*Exiting create_coverage_skip_stf\\(\\) early\\.",
    all = TRUE
  )
  
  mockery::expect_called(cleanup_mock, 1)
  mockery::expect_args(cleanup_mock, 1, env = cov_env)
})

# 2) Non-existent directory -> message + cleanup_and_return_null + invisible(NULL)
testthat::test_that("non-existent work_dir -> message, cleanup, invisible NULL", {
  nonexistent <- file.path(tempdir(), paste0("no_such_dir_", as.integer(Sys.time())))
  cov_env <- make_cov_env(work_dir = nonexistent)
  fx <- create_coverage_skip_stf
  cleanup_mock <- mockery::mock()
  
  mockery::stub(fx, "cleanup_and_return_null", cleanup_mock)
  
  fixture <- make_test_fixture()
  test_skip <- empty_test_skip()
  
  testthat::expect_message(
    testthat::expect_invisible({
      res <- fx(
        pkg_source_path = fixture$base,
        pkg = "dummyPkg",
        test_path = fixture$testthat,
        test_skip = test_skip,
        test_map = NULL,
        cov_env = cov_env
      )
      testthat::expect_null(res)
    }),
    regexp = "`cov_env\\$work_dir` does not exist on disk: .*Exiting create_coverage_skip_stf\\(\\) early\\.",
    all = TRUE
  )
  
  mockery::expect_called(cleanup_mock, 1)
  mockery::expect_args(cleanup_mock, 1, env = cov_env)
})

# 3) Empty-string work_dir -> checkmate assert_string fails
testthat::test_that("empty-string work_dir -> assert_string(min.chars=1) error", {
  cov_env <- make_cov_env(work_dir = "")
  fx <- create_coverage_skip_stf
  
  fixture <- make_test_fixture()
  test_skip <- empty_test_skip()
  
  testthat::expect_error(
    fx(
      pkg_source_path = fixture$base,
      pkg = "dummyPkg",
      test_path = fixture$testthat,
      test_skip = test_skip,
      test_map = NULL,
      cov_env = cov_env
    ),
    regexp = "cov_env\\$work_dir"
  )
})

# 4) Non-writable work_dir -> error from writability check
testthat::test_that("non-writable work_dir -> error", {
  work_dir <- tempfile("work_")
  dir.create(work_dir, recursive = TRUE, showWarnings = FALSE)
  # Try to remove write permission; if we can't, skip the test
  old_mode <- suppressWarnings(file.info(work_dir)$mode)
  suppressWarnings(Sys.chmod(work_dir, mode = "0555"))  # read & execute only
  on.exit({
    suppressWarnings(Sys.chmod(work_dir, mode = "0755"))
    unlink(work_dir, recursive = TRUE, force = TRUE)
  }, add = TRUE)
  
  # If system still considers it writable, skip (platform-dependent)
  if (file.access(work_dir, 2) == 0) {
    testthat::skip("Could not make directory non-writable on this platform; skipping.")
  }
  
  cov_env <- make_cov_env(work_dir = work_dir)
  fx <- create_coverage_skip_stf
  
  fixture <- make_test_fixture()
  test_skip <- empty_test_skip()
  
  testthat::expect_error(
    fx(
      pkg_source_path = fixture$base,
      pkg = "dummyPkg",
      test_path = fixture$testthat,
      test_skip = test_skip,
      test_map = NULL,
      cov_env = cov_env
    ),
    regexp = "`cov_env\\$work_dir` is not writable"
  )
})

# 5) work_dir not under tempdir() -> warns but proceeds (stub heavy calls)
testthat::test_that("work_dir outside tempdir -> warns but proceeds", {
  work_dir <- tempfile("work_")
  dir.create(work_dir, recursive = TRUE, showWarnings = FALSE)
  
  cov_env <- make_cov_env(work_dir = work_dir)
  fixture <- make_test_fixture()
  test_skip <- empty_test_skip()
  
  fx <- create_coverage_skip_stf
  
  # Force tempdir() to be a different root to trigger the warning
  mockery::stub(fx, "tempdir", function() file.path(base::tempdir(), "different_root"))
  
  # Stub heavy/external calls and no-ops for cleanup to keep test fast & deterministic
  mockery::stub(fx, "covr::environment_coverage", function(env, test_files) list())
  mockery::stub(fx, "remove_new_globals", function(env, initial_state) invisible(NULL))
  mockery::stub(fx, "cleanup_and_return_null", function(env) invisible(NULL))
  
  testthat::expect_warning(
    {
      res <- fx(
        pkg_source_path = fixture$base,
        pkg = "dummyPkg",
        test_path = fixture$testthat,
        test_skip = test_skip,
        test_map = NULL,
        cov_env = cov_env
      )
      testthat::expect_type(res, "list")
    },
    regexp = "`cov_env\\$work_dir` .* is not located under tempdir\\("
  )
})


testthat::test_that("TRUE when sentinel has dimnames == 'no_coverage.R'", {
  # 1x1 NA with row dimname "no_coverage.R"
  fc <- structure(NA_real_, .Dim = 1L, .Dimnames = list("no_coverage.R"))
  cov <- list(res_cov = list(coverage = list(filecoverage = fc)))
  
  out <- is_no_coverage_cov(cov)
  testthat::expect_true(out)
})

testthat::test_that("TRUE when sentinel found via names() fallback and dimnames() is stubbed to NULL", {
  # 1x1 NA with *names* set (no dimnames)
  fc <- structure(NA_real_, .Dim = 1L)
  names(fc) <- "no_coverage.R"
  cov <- list(res_cov = list(coverage = list(filecoverage = fc)))
  
  # Force the code path where dimnames(fc) is NULL, so names() is used
  mockery::stub(is_no_coverage_cov, "dimnames", function(x) NULL)
  
  out <- is_no_coverage_cov(cov)
  testthat::expect_true(out)
})

testthat::test_that("FALSE when length != 1 even if 'no_coverage.R' present", {
  # Two NAs — not the sentinel even if a name matches
  fc <- c(NA_real_, NA_real_)
  names(fc) <- c("no_coverage.R", "something_else.R")
  cov <- list(res_cov = list(coverage = list(filecoverage = fc)))
  
  out <- is_no_coverage_cov(cov)
  testthat::expect_false(out)
})

testthat::test_that("FALSE when lone NA but filename is not 'no_coverage.R'", {
  fc <- structure(NA_real_, .Dim = 1L, .Dimnames = list("different_file.R"))
  cov <- list(res_cov = list(coverage = list(filecoverage = fc)))
  
  out <- is_no_coverage_cov(cov)
  testthat::expect_false(out)
})

testthat::test_that("FALSE when filecoverage retrieval fails (tryCatch path -> NULL)", {
  # Any cov will do; we stub tryCatch so the retrieval returns NULL
  cov <- list(res_cov = list(coverage = list(filecoverage = "will_not_be_used")))
  
  # Make the internal tryCatch call return NULL
  mockery::stub(is_no_coverage_cov, "tryCatch", NULL)
  
  out <- is_no_coverage_cov(cov)
  testthat::expect_false(out)
})

testthat::test_that("FALSE when filecoverage has no names or dimnames", {
  fc <- structure(NA_real_, .Dim = 1L)  # no dimnames
  # no names(fc) either
  cov <- list(res_cov = list(coverage = list(filecoverage = fc)))
  
  # Ensure dimnames returns NULL so the code checks names() next
  mockery::stub(is_no_coverage_cov, "dimnames", function(x) NULL)
  
  out <- is_no_coverage_cov(cov)
  testthat::expect_false(out)
})

#' Map the Structure of Testthat Tests (Standard Testing Framework)
#'
#' This function analyses the test directory used by the Standard Testing
#' Framework (STF), typically `tests/testthat/`, and extracts a structured
#' representation of the unit tests contained in the directory. It identifies
#' test files, expectations, contexts, test names, and associated line ranges.
#'
#' @param path Character. Path to the test directory. Defaults to
#'   `"tests/testthat"`.
#'
#' @details
#' The function scans all testthat test files matching the pattern
#' `test*.R`. For each file, it extracts structured metadata describing the
#' hierarchy of expectations, including:
#' \itemize{
#'   \item file name
#'   \item context or description
#'   \item test label
#'   \item expectation type
#'   \item line ranges associated with each expectation
#' }
#'
#' Files that contain no parseable test structure are omitted. If no valid
#' test files are found, the function returns `NULL`.
#'
#' @return 
#' A consolidated `data.frame` where each row corresponds to an expectation
#' discovered within the STF test directory. The data frame typically contains:
#' \describe{
#'   \item{file}{Name of the test file in which the expectation appears.}
#'   \item{test}{The combined description and test label assigned to the
#'         expectation.}
#'   \item{expectation}{The expectation type detected (e.g., `expect_equal`).}
#'   \item{linerange}{The associated line numbers for that expectation.}
#' }
#'
#' This structured mapping is used by downstream STF processing functions to
#' support skip mapping, coverage enhancement, and reporting.
#'
#' @seealso
#'   \code{\link[base]{list.files}},
#'   \code{\link[base]{lapply}},
#'   \code{\link[base]{rbind}}
#'
#' @family stf_utility
#' @rdname map_tests_stf
#' @keywords internal
map_tests_stf <- function(path = "tests/testthat") {
  
  # Input checking
  checkmate::assert_string(path)
  checkmate::assert_directory_exists(path)
  
  # Step 1: List all test files in the specified directory that match the pattern "test*.R"
  files <- list.files(path, full.names = TRUE, pattern = "^test(.*?)R$")
  
  # Step 2: Apply the map_test function to each file and store results in a named list
  results_list <- lapply(files, map_test)
  names(results_list) <- basename(files) # Assign file names as list names
  
  # Step 3: Remove entries with no content (length == 0)
  results_list <- results_list[sapply(results_list, length) > 0]
  
  # Step 4: If no valid test files were found, return NULL
  if (length(results_list) == 0) {
    return(NULL)
  }
  
  # Step 5: Add a 'file' column to each data.frame in the list using the file name
  results_list <- lapply(names(results_list), function(name) {
    df <- results_list[[name]]
    df$file <- name
    df
  })
  
  # Step 6: Combine all individual data.frames into one consolidated data.frame
  results <- do.call("rbind", results_list)
  
  # Step 7: Identify rows where the 'test' column is not just a space
  idx <- !(results$test == " ")
  
  # Step 8: Clean up the 'description' and 'test' columns by removing prefixes
  results$description <- gsub("describe: |test_that: |tar_test: ", "", results$description)
  results$test <- gsub("it: ", "", results$test)
  
  # Step 9: For rows without a specific test, use the description as the test name
  results$test[!idx] <- results$description[!idx]
  
  # Step 10: For rows with both description and test, combine them into a single string
  results$test[idx] <- sprintf("%s: %s", results$description[idx], results$test[idx])
  
  # Step 11: Remove the 'description' column as it's now merged into 'test'
  results$description <- NULL
  
  # Step 12: Return the final structured data.frame
  return(results)
}



#' Group Parsed Testthat Code into Logical Test Blocks (Internal)
#'
#' This internal helper groups parse data from a testthat test file into
#' logical test blocks based on calls to block‑defining functions such as
#' `context()`, `test_that()`, `describe()`, `it()`, or `tar_test()`. It
#' identifies top‑level expressions that match a given function‑name regex
#' and returns a named list of data frames—one per detected block—trimmed
#' to the relevant code region.
#'
#' @param x data.frame. Parse data produced by
#'   \code{utils::getParseData(parse(file, keep.source = TRUE), includeText = TRUE)}.
#'   Must include columns such as \code{token}, \code{text}, \code{parent},
#'   and row names that correspond to node ids.
#' @param token_text Character. A regular expression used to select the
#'   block‑starting function names (e.g., \code{"^context$"}, 
#'   \code{"^test_that$|^describe$|^tar_test$"}, or \code{"^it$"}).
#'   Defaults to \code{"^context$"}.
#'
#' @details
#' The function locates top‑level expressions whose function call tokens
#' (\code{SYMBOL_FUNCTION_CALL}) match \code{token_text}. For each matching
#' block, it:
#' \itemize{
#'   \item assigns a human‑readable name using the function name and its
#'         string constant argument (when available),
#'   \item trims the block to exclude leading parse noise, and
#'   \item returns the subset of \code{x} corresponding to that block.
#' }
#' Block names are prefixed with the function (e.g., \code{"test_that: ..."},
#' \code{"describe: ..."}, \code{"it: ..."}, \code{"tar_test: ..."}) when the
#' function is recognized.
#'
#' @return
#' A named list of data frames, where each element represents the parse rows
#' for a single detected test block. Names encode the block function and label.
#' If no matching blocks are found, an empty list is returned. This structure
#' is consumed by higher‑level STF mapping helpers (e.g., \code{map_test()},
#' \code{map_tests_stf()}) to derive expectations, contexts, and line ranges.
#'
#' @seealso
#'   \code{\link[utils]{getParseData}},
#'   \code{\link{map_test}},
#'   \code{\link{map_tests_stf}}
#'
#' @family stf_utility
#' @keywords internal
#' @importFrom utils head tail
nest_test <- function(x, token_text = "^context$") {
  
  # Store row names of the parse data
  rx <- rownames(x)
  
  # Find indices of parent expressions that match the specified token_text (e.g., context)
  idx <- which(rx %in% x$parent[grepl("^SYMBOL_FUNCTION_CALL$", x$token) & grepl(token_text, x$text) & x$terminal])
  
  # Refine indices to get top-level expressions for those matches
  idx <- which(rx %in% x$parent[idx])
  
  # Initialize a grouping vector
  x1 <- rep(0, nrow(x))
  
  # Mark the start of each new block
  x1[idx] <- 1
  
  # Create a cumulative sum to assign group IDs
  x1 <- cumsum(x1)
  
  # Split the parse data into groups based on the group IDs
  x2 <- split(x, x1)
  
  # Assign names to each group based on the function call and its string argument
  names(x2) <- sapply(x2, function(x) {
    # Get the function name (e.g., test_that, describe, it)
    add_text <- x$text[grep("^SYMBOL_FUNCTION_CALL$", x$token)[1]]
    
    # Evaluate the string constant (e.g., test description)
    results <- eval(parse(text = x$text[grepl("^STR_CONST$", x$token)][1], keep.source = TRUE))
    
    # Prefix the name with the function name if it's a known test block
    if (add_text %in% c("test_that", "describe", "it", "tar_test")) {
      results <- sprintf("%s: %s", add_text, results)
    }
    
    results
  })
  
  # Trim each group to remove leading parse noise and keep only the relevant block
  x2 <- lapply(x2, function(x) {
    # Find the second top-level expression (if available)
    idx <- utils::tail(utils::head(which(x$parent == 0), 2), 1)
    if (length(idx) == 0) {
      idx <- 1
    }
    # Return the block from the identified index to the end
    x[idx:nrow(x), ]
  })
  
  x2
}



#' Extract Expectation Calls from Parsed Testthat Code
#'
#' This internal helper identifies expectation calls (e.g.,
#' `expect_equal()`, `expect_true()`, etc.) within parse data extracted from a
#' testthat test file. The function locates expectation function calls by
#' matching their token text and returns a structured data frame containing
#' the expectation names and their corresponding source line ranges.
#'
#' @param parsed_data data.frame. Parse data produced by
#'   \code{utils::getParseData(parse(file, keep.source = TRUE), includeText = TRUE)}.
#'   Must include the columns \code{token}, \code{text}, \code{line1},
#'   and \code{line2}.
#' @param token_text Character. A regular expression used to detect
#'   expectation‑style function calls. Defaults to \code{"^expect_"}.
#'
#' @details
#' The function:
#' \itemize{
#'   \item locates all rows where the token is \code{"SYMBOL_FUNCTION_CALL"}
#'         and the text matches \code{token_text},
#'   \item identifies the expression node corresponding to each expectation,
#'   \item extracts line ranges (\code{line1}, \code{line2}) describing where the
#'         expectation appears in the source file, and
#'   \item returns a consolidated data frame of results.
#' }
#'
#' If no expectation calls are found, the function returns \code{NULL}.
#' This helper is used internally by STF mapping functions (such as
#' \code{map_test()} and \code{map_tests_stf()}) when constructing a full
#' hierarchy of test blocks and expectations.
#'
#' @return
#' A \code{data.frame} where each row corresponds to a detected expectation.
#' The data frame contains:
#' \describe{
#'   \item{expectation}{Character. Name of the expectation function (e.g.,
#'         \code{"expect_equal"}).}
#'   \item{line1}{Integer. Starting line number of the expression.}
#'   \item{line2}{Integer. Ending line number of the expression.}
#' }
#' Returns \code{NULL} if no matching expectations are found.
#'
#' @keywords internal
get_expect <- function(parsed_data, token_text = "^expect_") {
  
  # Identify indices of function calls that match the expectation pattern
  idx <- which(grepl("^SYMBOL_FUNCTION_CALL$", parsed_data$token) & grepl(token_text, parsed_data$text) & parsed_data$terminal)
  
  # Extract the matched expectation function names
  expect_data <- parsed_data$text[idx]
  
  if (length(expect_data) == 0) {
    return(NULL)
  }
  
  # For each matched expectation, find the corresponding expression's line range
  line_ <- lapply(idx, function(y) {
    # Find the last two 'expr' tokens before the expectation
    this_idx <- tail(grep("expr", parsed_data$token[1:y]), 2)
    # Determine which of those expressions contains the expectation
    this_exp <- min(grep(sprintf("%s", parsed_data$text[y]), parsed_data$text[this_idx]))
    # Extract the line range for that expression
    parsed_data[this_idx[this_exp], c("line1", "line2")]
  })
  
  # Combine all line ranges into a single data.frame
  line_ <- do.call("rbind", line_)
  
  # Construct the final data.frame with expectation names and line numbers
  expect_data_df <- data.frame(
    expectation = expect_data,
    line1 = line_$line1,
    line2 = line_$line2,
    stringsAsFactors = FALSE
  )
  
  expect_data_df
}


#' Extract and Structure Expectations from Parsed Test Blocks
#'
#' This internal helper takes a list of parsed test blocks (as produced by
#' \code{nest_test()}) and extracts expectation calls from each block using
#' \code{get_expect()}. It then assembles all detected expectations into a
#' structured data frame that associates each expectation with its
#' corresponding test label.
#'
#' @param x A named list of parsed test blocks, typically produced by
#'   \code{nest_test()}. Each element should contain parse data suitable for
#'   expectation extraction via \code{get_expect()}.
#'
#' @details
#' For each block in \code{x}, the function:
#' \itemize{
#'   \item calls \code{get_expect()} to extract expectation function calls
#'         and their line ranges,
#'   \item removes blocks with no expectations,
#'   \item attaches a \code{test} label using the block's list name, and
#'   \item binds all expectation results into a single consolidated data frame.
#' }
#'
#' If no expectations are found in any block, the function returns \code{NULL}.
#'
#' @return
#' A \code{data.frame} in which each row corresponds to an expectation
#' extracted from the supplied test blocks. Typical columns include:
#' \describe{
#'   \item{expectation}{Name of the expectation function (e.g.,
#'         \code{"expect_equal"}).}
#'   \item{line1}{Starting line number of the expectation expression.}
#'   \item{line2}{Ending line number of the expectation expression.}
#'   \item{test}{Label identifying the test block the expectation belongs to.}
#' }
#' Returns \code{NULL} if no expectations are detected.
#'
#' @keywords internal
nest_expect <- function(x) {
  
  # Apply get_expect to each test block to extract expectations
  ret <- lapply(x, get_expect)
  
  # Remove any NULL results (i.e., blocks with no expectations)
  ret <- ret[!sapply(ret, is.null)]
  
  if (length(ret) == 0) {
    return(NULL)
  }
  
  # Add a 'test' column to each expectation block using the list names
  ret <- lapply(names(ret), unrowname, ret = ret, label = "test")
  
  # Combine all expectation blocks into a single data.frame
  ret <- do.call("rbind", ret)
  
  ret
}


#' Map the Structure of Expectations in a Testthat File
#'
#' This function parses a single testthat test file and extracts a structured
#' representation of the expectations it contains. It identifies top‑level test
#' blocks (e.g., `test_that()`, `describe()`, `tar_test()`) and any nested
#' expectation blocks (e.g., `it()`). The result maps expectation location,
#' context, and source line ranges for downstream analysis.
#'
#' @param path Character. Path to a single test file (typically named
#'   `test-*.R`).
#'
#' @details
#' The function uses `utils::getParseData()` to inspect parse information from
#' the test file and constructs a hierarchical mapping of:
#' \itemize{
#'   \item test block context (e.g., `describe()`, `test_that()`)
#'   \item test labels or sub‑contexts (e.g., `it()` blocks)
#'   \item expectation type
#'   \item source line ranges associated with each expectation
#' }
#'
#' If the file does not contain any parseable test blocks or expectations,
#' `NULL` is returned.
#'
#' @return 
#' A `data.frame` in which each row corresponds to an expectation identified in
#' the test file. Typical columns include:
#' \describe{
#'   \item{description}{Context or test block label derived from the test file.}
#'   \item{expectation}{The expectation type (e.g., `expect_equal`).}
#'   \item{linerange}{Numeric vector giving the range of lines associated with
#'         the expectation.}
#'   \item{test}{Test label (combined description and nested block label, if
#'         applicable).}
#' }
#'
#' This structured output is used internally by STF test‑mapping functions
#' (e.g., `map_tests_stf()`) for coverage analysis, skip mapping, and test
#' reporting.
#'
#' @seealso
#'   \code{\link[utils]{getParseData}}
#'
#' @family stf_utility
#' @rdname map_test
#' @keywords internal
#' @importFrom utils getParseData
#' @importFrom stats setNames
map_test <- function(path) {
  
  # Parse the test file and extract parse data including source text
  parse_data <- utils::getParseData(parse(path, keep.source = TRUE), includeText = TRUE)
  if (is.null(parse_data)) return(NULL)
  
  # Treat test_that and describe as top-level test blocks
  ret <- lapply(nest_test(parse_data, token_text = "^test_that$|^describe$|^tar_test$"), function(xx) {
    # Identify the function name (e.g., test_that or describe)
    SYMB <- xx$text[grep("^SYMBOL_FUNCTION_CALL$", xx$token)[1]]
    
    # Depending on the function, process expectations differently
    switch(SYMB,
           # Nest 'it' blocks inside describe
           describe = {
             nest_expect(nest_test(xx, token_text = "^it$"))
           },
           test_that = {
             # Wrap test_that block in a list with a placeholder name
             nest_expect(setNames(list(xx), " "))
           },
           tar_test = {
             # Wrap test_that block in a list with a placeholder name
             nest_expect(setNames(list(xx), " "))
           }, {
             list()
           }
    )
  })
  
  # Remove empty elements
  ret <- ret[sapply(ret, length) > 0]
  if (length(ret) == 0) return(NULL)
  
  
  # Ensure unique names for test blocks
  names(ret) <- make.unique(names(ret), sep = "_")
  
  # Add 'description' column to each test block
  ret <- lapply(names(ret), unrowname, ret = ret, label = "description")
  
  # Combine all test blocks into a single data.frame
  ret <- do.call("rbind", ret)
  return(ret)
}


#' Add Identifier Column and Remove Row Names from a Data Frame
#'
#' This internal helper restructures a data frame by adding an identifier
#' column (e.g., file name or test label) and removing row names. It is used
#' by STF parsing functions to attach contextual labels (such as test block
#' names or file names) to expectation or test‑block structures.
#'
#' @param identifier Character. The identifier to insert into the resulting
#'   data frame (typically the name of the list element being processed).
#' @param ret List of data frames, usually the output of applying a mapping
#'   or extraction function across multiple parsed test blocks.
#' @param label Character. Name of the column used to store the identifier
#'   (e.g., \code{"test"} or \code{"file"}).
#'
#' @details
#' The function:
#' \itemize{
#'   \item extracts the data frame corresponding to \code{identifier},
#'   \item appends a new column containing the identifier value,
#'   \item removes row names for clean downstream processing, and
#'   \item reorders columns so the identifier appears first.
#' }
#'
#' This structure is used by STF mapping helpers such as
#' \code{nest_expect()} and \code{map_tests_stf()} to carry forward file or
#' test labels when constructing unified test‑mapping tables.
#'
#' @return
#' A \code{data.frame} with an identifier column prepended and row names
#' removed. The remaining columns retain their original structure.
#'
#' @keywords internal
unrowname <- function(identifier, ret, label) {
  # Extract the data.frame corresponding to the identifier
  results <- ret[[identifier]]
  
  # Get the number of columns in the original data.frame
  nc <- ncol(results)
  
  # Add a new column with the identifier value (e.g., file name)
  results[[label]] <- identifier
  
  
  # Remove row names to ensure clean output
  rownames(results) <- NULL
  
  # Reorder columns to place the identifier column first
  results <- results[, c(c(nc + 1), 1:nc)]
  
  return(results)
}

---
title: "&nbsp;"
output:
  rmarkdown::html_document:
  css: www/styles.css
editor_options: 
  chunk_output_type: console
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.path = "man/figures/",
  out.width = "100%"
)

# Suppress package loading messages
suppressMessages({
  library(dplyr)
  library(kableExtra)
  library(knitr)
})

suppressWarnings({
  library(htmltools)
  library(htmlwidgets)
})
```

# Package Test Report for `r get('pkg_name', envir = report_env)` version `r get('pkg_version', envir = report_env)` <a><img src="`r system.file('man/figures/test_assessr.png', package = 'test.assessr')`" align="right" height="100" /><br></a>

```{r test_summary, echo=FALSE}
library(knitr)
library(kableExtra)

test_summary_output <- get("test_pkg_summary_output", envir = report_env)

test_summary_output_html <- knitr::kable(test_summary_output,
                                         col.names = c("Type", "Values"),
                                         format = "html") %>%
  kableExtra::kable_styling("striped", 
                            position = "left", 
                            font_size = 12) %>%
  kableExtra::row_spec(0, 
                       bold = TRUE,
                       italic = TRUE,
                       background = "green") %>%
  kableExtra::add_header_above(c("Test Assessment Summary" = 2), 
                               bold = TRUE, 
                               background = "green", 
                               color = "black")

test_summary_output_html <- test_summary_output_html %>%
  kableExtra::row_spec(which(test_summary_output$Value == "Low"), background = "green") %>%
  kableExtra::row_spec(which(test_summary_output$Value == "Medium"), background = "yellow") %>%
  kableExtra::row_spec(which(test_summary_output$Value == "High"), background = "red")

test_summary_output_html
```

```{r setup_multi, echo=FALSE}
multi_framework <- get("multi_framework", envir = report_env)
frameworks <- if (multi_framework) get("frameworks", envir = report_env) else character(0)
```

```{r test_details_single, echo=FALSE}
if (!multi_framework) {
  library(knitr)
  library(kableExtra)
  library(dplyr)
  test_details_output <- get("test_details_output", envir = report_env)
  test_details_output_display <- test_details_output %>% select(Metric, Value)
  test_details_output_html <- knitr::kable((test_details_output_display),
                                 col.names = c("Type", "Details"), "html")
  kableExtra::kable_styling(test_details_output_html, 
                            "striped", position = "left", font_size = 12) %>%
  kableExtra::row_spec(0, bold = TRUE, italic = TRUE, background = "green") %>%
  kableExtra::add_header_above(c("Test Assessment Run Details" = 2), bold = TRUE, background = "green", color = "black")
}
```

```{r test_details_multi, echo=FALSE}
if (multi_framework) {
  library(knitr)
  library(kableExtra)
  library(dplyr)
  library(htmltools)
  framework_results <- get("framework_results", envir = report_env)
  out_parts <- list()
  for (fw in frameworks) {
    fr <- framework_results[[fw]]
    if (!is.null(fr$test_details_output)) {
      test_details_output_display <- fr$test_details_output %>% select(Metric, Value)
      test_details_output_html <- knitr::kable((test_details_output_display),
                                     col.names = c("Type", "Details"), "html") %>%
        kableExtra::kable_styling("striped", position = "left", font_size = 12) %>%
        kableExtra::row_spec(0, bold = TRUE, italic = TRUE, background = "green") %>%
        kableExtra::add_header_above(setNames(c(2), paste0("Test Assessment Run Details (", fw, ")")), bold = TRUE, background = "green", color = "black")
      out_parts <- c(out_parts, list(
        htmltools::tags$h3(paste0(fw, " Run Details")),
        htmltools::HTML(as.character(test_details_output_html))
      ))
    }
  }
  if (length(out_parts) > 0) htmltools::tagList(out_parts)
}
```

```{r test_coverage_single, echo=FALSE}
if (!multi_framework) {
  library(knitr)
  library(kableExtra)
  coverage_output <- get("coverage_output", envir = report_env)
  colnames(coverage_output) <- c("Function", "Coverage", "Errors", "Notes")
  coverage_output_html_DT <- DT::datatable(coverage_output, 
                                           rownames = FALSE, 
                                           colnames = c("Function", "Coverage", "Errors", "Notes"),
                                           options = list(pageLength = 10, lengthMenu = c(10, 20, 50),
                                                          search = list(search = ""), dom = 'Bfrtip',
                                                          buttons = c('copy', 'csv', 'excel', 'pdf', 'print')))
  coverage_output_html_DT <- coverage_output_html_DT %>%
    DT::formatStyle(columns = 'Errors',
      backgroundColor = DT::styleEqual(
        levels = c("No testthat or testit configuration", "No test coverage errors", "Other Value"),
        values = c("red", "green", "red"))) %>%
    DT::formatStyle(columns = 'Notes',
      backgroundColor = DT::styleEqual(
        levels = c("No test coverage notes", "Other Value"),
        values = c("green", "yellow")))
  htmltools::tagList(
    htmltools::tags$h2("Test Coverage", style = "font-weight: bold; background-color: green; color: black; text-align: center;"),
    coverage_output_html_DT
  )
}
```

```{r test_coverage_multi, echo=FALSE}
if (multi_framework) {
  library(knitr)
  library(kableExtra)
  library(htmltools)
  framework_results <- get("framework_results", envir = report_env)
  out_parts <- list()
  for (fw in frameworks) {
    fr <- framework_results[[fw]]
    if (!is.null(fr$coverage_output)) {
      cov_out <- fr$coverage_output
      colnames(cov_out) <- c("Function", "Coverage", "Errors", "Notes")
      cov_out <- DT::datatable(cov_out, rownames = FALSE, 
                               colnames = c("Function", "Coverage", "Errors", "Notes"),
                               options = list(pageLength = 10, lengthMenu = c(10, 20, 50),
                                              search = list(search = ""), dom = 'Bfrtip',
                                              buttons = c('copy', 'csv', 'excel', 'pdf', 'print'))) %>%
        DT::formatStyle(columns = 'Errors',
          backgroundColor = DT::styleEqual(
            levels = c("No testthat or testit configuration", "No test coverage errors", "Other Value"),
            values = c("red", "green", "red"))) %>%
        DT::formatStyle(columns = 'Notes',
          backgroundColor = DT::styleEqual(
            levels = c("No test coverage notes", "Other Value"),
            values = c("green", "yellow")))
      out_parts <- c(out_parts, list(
        htmltools::tags$h3(paste0(fw, " Coverage")),
        cov_out
      ))
    }
  }
  if (length(out_parts) > 0) htmltools::tagList(out_parts)
}
```

```{r stf_nstf_detect, echo=FALSE}
has_stf <- get("has_stf", envir = report_env)
has_nstf <- get("has_nstf", envir = report_env)
```

```{r stf_nstf_selector, echo=FALSE, results='asis'}
if (has_stf && !has_nstf) {
  cat('
<div style="margin-bottom: 20px;">
  <label for="matrixSelector" style="background-color: purple; color: white; font-size: 20px; font-weight: bold; padding: 10px; border-radius: 5px; display: inline-block; margin-bottom: 10px;">Select Standard Test Type:</label>
  <select id="matrixSelector" onchange="toggleMatrix(this.value)" style="margin-left: 10px; padding: 6px; font-size: 18px; border-radius: 4px;">
    <option value="long_summary">Tests Passed</option>
    <option value="tests_skip">Tests Skipped</option>
  </select>
</div>
<script>
  function toggleMatrix(value) {
    var ids = ["long_summary", "tests_skip"];

    // STF values map directly to div IDs
    var showId = value;  

    ids.forEach(function(id) {
      var el = document.getElementById(id);
      if (el) {
        el.style.visibility = (id === showId) ? "visible" : "hidden";
      }
    });
  }

  document.addEventListener("DOMContentLoaded", function() {
    toggleMatrix("long_summary");
  });
</script>
')
} else if (has_nstf && !has_stf) {
  cat('
<div style="margin-bottom: 20px;">
  <label for="matrixSelector" style="background-color: purple; color: white; font-size: 20px; font-weight: bold; padding: 10px; border-radius: 5px; display: inline-block; margin-bottom: 10px;">Select Non-Standard Test Type:</label>
  <select id="matrixSelector" onchange="toggleMatrix(this.value)" style="margin-left: 10px; padding: 6px; font-size: 18px; border-radius: 4px;">
    <option value="no">No Tests</option>
    <option value="skipped">Tests Skipped</option>
    <option value="passing">Tests Passing</option>
  </select>
</div>
<script>
  function toggleMatrix(value) {
    var ids = ["no_tests", "tests_skipped", "tests_passing"];
    var showId = value === "no" ? "no_tests" : value === "skipped" ? "tests_skipped" : "tests_passing";
    ids.forEach(function(id) {
      var el = document.getElementById(id);
      if (el) el.style.visibility = (id === showId) ? "visible" : "hidden";
    });
  }
  document.addEventListener("DOMContentLoaded", function() {
    var sel = document.getElementById("matrixSelector");
    toggleMatrix(sel ? sel.value : "no");
  });
</script>
')
} else if (has_stf && has_nstf) {
  cat('
<div style="margin-bottom: 20px;">
  <label for="matrixSelector" style="background-color: purple; color: white; font-size: 20px; font-weight: bold; padding: 10px; border-radius: 5px; display: inline-block; margin-bottom: 10px;">Select Test Type:</label>
  <select id="matrixSelector" onchange="toggleMatrix(this.value)" style="margin-left: 10px; padding: 6px; font-size: 18px; border-radius: 4px;">
    <option value="long_summary">Tests Passed</option>
    <option value="tests_skip">Tests Skipped</option>
    <option value="no">No Tests</option>
    <option value="skipped">Tests Skipped (NSTF)</option>
    <option value="passing">Tests Passing</option>
  </select>
</div>
<script>
  function toggleMatrix(value) {
    var showId = value === "no" ? "no_tests" : value === "skipped" ? "tests_skipped" : value === "passing" ? "tests_passing" : value;
    ["long_summary","tests_skip","no_tests","tests_skipped","tests_passing"].forEach(function(id) {
      var el = document.getElementById(id);
      if (el) el.style.display = (id === showId) ? "block" : "none";
    });
  }
  document.addEventListener("DOMContentLoaded", function() { toggleMatrix(document.getElementById("matrixSelector") ? document.getElementById("matrixSelector").value : "long_summary"); });
</script>
')
}
```

```{r stf_sections_single, echo=FALSE}

if (!multi_framework && has_stf) {
  library(DT)
  library(htmltools)

  # --- PASSING TESTS CONTENT ---
  long_summary <- get("long_summary_df", envir = report_env)

  long_summary_content <- if (is.data.frame(long_summary) && nrow(long_summary) > 0) {
    colnames(long_summary) <- gsub("_", " ", colnames(long_summary))

    DT::datatable(
      long_summary,
      elementId = "dt_stf_long_summary",
      options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = "Bfrtip"),
      colnames = c("R function", "Test", "Test Block start line", "Test Status"),
      caption = tags$caption(
        style = "caption-side: top; text-align: left;",
        tags$strong(tags$u("Passing Tests"))
      ),
      rownames = FALSE
    )
  } else {
    tags$div(style = "padding: 8px;", "No passing tests.")
  }

  # --- SKIPPED TESTS CONTENT ---
  tests_skip_df <- get("tests_skip_df", envir = report_env)

  tests_skip_content <- if (is.data.frame(tests_skip_df) && nrow(tests_skip_df) > 0) {
    DT::datatable(
      tests_skip_df,
      elementId = "dt_stf_tests_skip",
      options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = "Bfrtip"),
      colnames = c("R function", "Test", "Test Status", "Expectation type", "Test Block start line"),
      caption = tags$caption(
        style = "caption-side: top; text-align: left;",
        tags$strong(tags$u("Skipped Tests"))
      ),
      rownames = FALSE
    )
  } else {
    tags$div(style = "padding: 8px;", "No skipped tests.")
  }

  # --- WRAP ONCE (NO DUPLICATE IDs), TOGGLE BY VISIBILITY ---
  long_div <- tags$div(
    id = "long_summary",
    style = "position:absolute; top:0; left:0; right:0; visibility:visible;",
    long_summary_content
  )

  skip_div <- tags$div(
    id = "tests_skip",
    style = "position:absolute; top:0; left:0; right:0; visibility:hidden;",
    tests_skip_content
  )

  tags$div(
    style = "position:relative; min-height:300px;",
    tagList(long_div, skip_div)
  )
}


```

```{r nstf_sections_single, echo=FALSE}
if (!multi_framework && has_nstf) {
  library(knitr)
  library(kableExtra)
  library(DT)
  library(htmltools)
  functions_no_tests <- get("functions_no_tests", envir = report_env)
  no_tests_content <- if (is.data.frame(functions_no_tests) && nrow(functions_no_tests) > 0) {
    colnames(functions_no_tests) <- gsub("_", " ", colnames(functions_no_tests))
    DT::datatable(functions_no_tests, elementId = "dt_nstf_no_tests",
      options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'),
      caption = htmltools::tags$caption(style = "caption-side: top; text-align: left;",
        htmltools::tags$strong(htmltools::tags$u("Functions with No Tests"))), rownames = FALSE)
  } else {
    htmltools::tags$div(style = "padding: 8px; background:#fff3cd; border:1px solid #ffeeba; border-radius:6px;",
      htmltools::tags$strong("No data to display: "), "The 'functions_no_tests' data frame is missing or has zero rows.")
  }
  tests_skipped_df <- get("tests_skipped_df", envir = report_env)
  tests_skipped_content <- if (is.data.frame(tests_skipped_df) && nrow(tests_skipped_df) > 0) {
    DT::datatable(tests_skipped_df, elementId = "dt_nstf_tests_skipped",
      options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'),
      colnames = c("File"), caption = htmltools::tags$caption(style = "caption-side: top; text-align: left;",
        htmltools::tags$strong(htmltools::tags$u("Skipped Tests"))))
  } else {
    htmltools::tags$div(style = "padding: 8px;", "No skipped tests.")
  }
  tests_passing_df <- get("tests_passing_df", envir = report_env)
  tests_passing_content <- if (is.data.frame(tests_passing_df) && nrow(tests_passing_df) > 0) {
    DT::datatable(tests_passing_df, elementId = "dt_nstf_tests_passing",
      options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'),
      colnames = c("File"), caption = htmltools::tags$caption(style = "caption-side: top; text-align: left;",
        htmltools::tags$strong(htmltools::tags$u("Passing Tests"))))
  } else {
    htmltools::tags$div(style = "padding: 8px;", "No passing tests.")
  }
  no_tests_div <- htmltools::tags$div(id = "no_tests", style = "position:absolute; top:0; left:0; right:0; visibility:visible;", no_tests_content)
  tests_skipped_div <- htmltools::tags$div(id = "tests_skipped", style = "position:absolute; top:0; left:0; right:0; visibility:hidden;", tests_skipped_content)
  tests_passing_div <- htmltools::tags$div(id = "tests_passing", style = "position:absolute; top:0; left:0; right:0; visibility:hidden;", tests_passing_content)
  htmltools::tags$div(style = "position:relative; min-height:300px;", htmltools::tagList(no_tests_div, tests_skipped_div, tests_passing_div))
}
```

```{r stf_nstf_sections_multi, echo=FALSE}
if (multi_framework && (has_stf || has_nstf)) {
  library(knitr)
  library(kableExtra)
  library(DT)
  library(htmltools)
  framework_results <- get("framework_results", envir = report_env)
  default_visible <- if (has_nstf) "block" else "none"
  build_div <- function(id, content, display = "none") {
    if (length(content) == 0) return(htmltools::tags$div(id = id, style = paste0("display:", display, ";"), ""))
    htmltools::tags$div(id = id, style = paste0("display:", display, ";"), htmltools::tagList(content))
  }
  long_summary_parts <- list()
  tests_skip_parts <- list()
  no_tests_parts <- list()
  tests_skipped_parts <- list()
  tests_passing_parts <- list()
  for (fw in frameworks) {
    fr <- framework_results[[fw]]
    if (has_stf && !is.null(fr$long_summary_df) && is.data.frame(fr$long_summary_df) && nrow(fr$long_summary_df) > 0) {
      colnames(fr$long_summary_df) <- gsub("_", " ", colnames(fr$long_summary_df))
      cap <- htmltools::tags$caption(style = "caption-side: top;", htmltools::tags$strong(paste0("Passing Tests (", fw, ")")))
      long_summary_parts <- c(long_summary_parts, list(htmltools::tags$h3(fw), DT::datatable(fr$long_summary_df, options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'), caption = cap, rownames = FALSE)))
    }
    if (has_stf && !is.null(fr$tests_skip_df) && is.data.frame(fr$tests_skip_df) && nrow(fr$tests_skip_df) > 0) {
      cap <- htmltools::tags$caption(style = "caption-side: top;", htmltools::tags$strong(paste0("Skipped Tests (", fw, ")")))
      tests_skip_parts <- c(tests_skip_parts, list(htmltools::tags$h3(fw), DT::datatable(fr$tests_skip_df, options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'), caption = cap, rownames = FALSE)))
    }
    if (has_nstf && !is.null(fr$functions_no_tests) && is.data.frame(fr$functions_no_tests) && nrow(fr$functions_no_tests) > 0) {
      colnames(fr$functions_no_tests) <- gsub("_", " ", colnames(fr$functions_no_tests))
      cap <- htmltools::tags$caption(style = "caption-side: top;", htmltools::tags$strong(paste0("Functions with No Tests (", fw, ")")))
      no_tests_parts <- c(no_tests_parts, list(htmltools::tags$h3(fw), DT::datatable(fr$functions_no_tests, elementId = paste0("dt_multi_no_tests_", fw), options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'), caption = cap, rownames = FALSE)))
    }
    if (has_nstf && !is.null(fr$tests_skipped_df) && is.data.frame(fr$tests_skipped_df) && nrow(fr$tests_skipped_df) > 0) {
      cap <- htmltools::tags$caption(style = "caption-side: top;", htmltools::tags$strong(paste0("Skipped Tests (", fw, ")")))
      tests_skipped_parts <- c(tests_skipped_parts, list(htmltools::tags$h3(fw), DT::datatable(fr$tests_skipped_df, elementId = paste0("dt_multi_tests_skipped_", fw), options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'), caption = cap, rownames = FALSE)))
    }
    if (has_nstf && !is.null(fr$tests_passing_df) && is.data.frame(fr$tests_passing_df) && nrow(fr$tests_passing_df) > 0) {
      cap <- htmltools::tags$caption(style = "caption-side: top;", htmltools::tags$strong(paste0("Passing Tests (", fw, ")")))
      tests_passing_parts <- c(tests_passing_parts, list(htmltools::tags$h3(fw), DT::datatable(fr$tests_passing_df, elementId = paste0("dt_multi_tests_passing_", fw), options = list(pageLength = 10, lengthMenu = c(10, 20, 50), dom = 'Bfrtip'), caption = cap, rownames = FALSE)))
    }
  }
  first_visible <- if (has_nstf) "no" else "long_summary"
  d_long <- if (length(long_summary_parts) > 0) build_div("long_summary", long_summary_parts, if (first_visible == "long_summary") "block" else "none") else htmltools::tags$div(id = "long_summary", style = "display:none;")
  d_skip <- if (length(tests_skip_parts) > 0) build_div("tests_skip", tests_skip_parts, "none") else htmltools::tags$div(id = "tests_skip", style = "display:none;")
  d_no <- if (length(no_tests_parts) > 0) build_div("no_tests", no_tests_parts, if (first_visible == "no") "block" else "none") else htmltools::tags$div(id = "no_tests", style = "display:none;")
  d_skipped <- if (length(tests_skipped_parts) > 0) build_div("tests_skipped", tests_skipped_parts, "none") else htmltools::tags$div(id = "tests_skipped", style = "display:none;")
  d_passing <- if (length(tests_passing_parts) > 0) build_div("tests_passing", tests_passing_parts, "none") else htmltools::tags$div(id = "tests_passing", style = "display:none;")
  htmltools::tagList(d_long, d_skip, d_no, d_skipped, d_passing)
}
```

---
title: "How to Start"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{How to Start}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r load, echo=FALSE}
library(test.assessr)
```

Run package coverage for a package

If the following code was run,

```{r pkg_covr, eval=FALSE}
dp <- system.file("test-data", "test.package.0001_0.1.0.tar.gz",
                 package = "test.assessr")


test_package <- get_package_coverage(dp)
```
The result would be:

| R object | field | element 1 | element 2 | value |
| -------- | ----- | --------- | --------- | ------|
| test_package |pkg_name | | | "test.package.0001" |
| |  pkg_ver | | | "0.1.0" |
| | date_time | | | "2026-03-09 14:27:36.673915" |
| | executor | | | "xxxxxxx" |
| | sysname | | | "Linux" |
| | version | | | "#1 SMP PREEMPT_DYNAMIC Mon Dec  1 05:36:50 UTC 2025" |
| | release | | | "6.1.158-180.294.amzn2023.x86_64" |
| | machine | | | "x86_64" |
| | r_version | | |"4.4.1" |
| | test_framework_type | | | standard testing framework |
| | total_cov | | | 1 |
| | res_cov | name| |"test.package.0001" |
| | | coverage | filecoverage | R/myscript.R 100 | 
| | | |      totalcoverage | 100 |
| | errors | | |NA |
| | notes  | | |NA |

---
title: "STF Testing Frameworks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{STF Testing Frameworks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r load, echo=FALSE}
library(test.assessr)
```
The `test.assessr` package checks package structure and test drivers to detect and classify testing configurations.

If it detects a standard testing framework (STF) (i.e. `testthat` or `testit`),
it runs a standard testing workflow involving a number of steps:

 `Step 1`
 
  - run `callr::r_safe` which calls `covr::package_coverage`. 
 
  - The `r_safe` function has the parameter `error = "error"`. 
 
  - This means if `package_coverage` encounters an error, 
 then `r_safe` returns an error to the main testing process.

  - If `Step 1` is successful, 
 then a unit test data coverage object like this is produced:


`Step 1 unit test data coverage object`

| R object | field | element 1 | element 2 | value |
| -------- | ----- | --------- | --------- | ------|
| test_package |pkg_name | | | "test.package.0001" |
| |  pkg_ver | | | "0.1.0" |
| | date_time | | | "2026-03-09 14:27:36.673915" |
| | executor | | | "xxxxxxx" |
| | sysname | | | "Linux" |
| | version | | | "#1 SMP PREEMPT_DYNAMIC Mon Dec  1 05:36:50 UTC 2025" |
| | release | | | "6.1.158-180.294.amzn2023.x86_64" |
| | machine | | | "x86_64" |
| | r_version | | |"4.4.1" |
| | test_framework_type | | | standard testing framework |
| | total_cov | | | 1 |
| | res_cov | name| |"test.package.0001" |
| | | coverage | filecoverage | R/myscript.R 100 | 
| | | |      totalcoverage | 100 |
| | errors | | |NA |
| | notes  | | |NA |

`Step 2`
 
  - run `setup_covr_env` which 
   
    - loads helper files from `tests/testthat`,
    - loads package namespace objects into `cov_env`,
    - resolves the project root,
    - loads package datasets into `cov_env`,
    - loads Base R datasets and exports into `cov_env`,
    - optionally loads `nycflights13` datasets if available, 
    - loads dependencies into `cov_env`, and
    - unloads the package namespace afterward to reduce conflicts.
 
  - run `create fail reporter` to find failing tests
  - run `create test reporter` to create a reporter for all tests
  - map `exported functions` to test files
  - create a filter for test status i.e. tests that have `"FAILED", "ERROR", "SKIPPED"`
  - identify tests to be skipped based on status
  - comment out tests to be skipped
  - run `covr::environment_coverage` on remaining test files.

  - If `Step 2` is successful, 
 then a unit test data coverage object like this is produced:



`Step 2 unit test data coverage object`

| R object | field | element 1 | element 2 | value |
| -------- | ----- | --------- | --------- | ------|
| test_package |pkg_name | | | "test.package.0003" |
| |  pkg_ver | | | "0.1.0" |
| | date_time | | | "2026-04-09 10:17:27.71564" |
| | executor | | | "xxxxxxx" |
| | sysname | | | "Linux" |
| | version | | | "#1 SMP PREEMPT_DYNAMIC Mon Dec  1 05:36:50 UTC 2025" |
| | release | | | "6.1.158-180.294.amzn2023.x86_64" |
| | machine | | | "x86_64" |
| | r_version | | |"4.4.1" |
| | test_framework_type | | | standard testing framework |
| | total_cov | | | 0 |
| | res_cov | name| |"test.package.0003" |
| | | coverage | filecoverage | num [1(1d)] 0 
| | | | | attr(*, "dimnames")=List of 1 chr "myscript.R" | 
| | | |       totalcoverage | 0 |
| | errors | | |NA |
| | notes  | | |NA |
| long_summary        |  file | | |      "test-myscript.R#L2_L2" |
| | code_script | | | "myscript" |
| | context     | | | "this works" |
| |  status     | | | "FAILED" |
| | n           | | | 1 |
| | time        | | | 0.291 |
|   test_skip         |  file | | |  "test-myscript.R" |
| | code_script  | | | "myscript" |
| | context   | | |  "this works" |
| | status     | | | "FAILED" |
| | n          | | |  1 |
| | time       | | | 0.291 |
| | expectation: | | |  "expect_equal" |
| | line1      | | | 2 |
| | line2      | | | 2 |

This data object differs from `Step 1 data object` in that it gives information 
about which tests/test blocks were skipped and the status tells the user why 
they were skipped.

`Step 3`
 
  - run `withr::with_dir` which calls `covr::package_coverage`. 
 
  - `covr::package_coverage` has the parameters `type = "tests", quiet = TRUE`

  - If `Step 3` is successful, 
 then a unit test data coverage object like this is produced:


`Step 3 unit test data coverage object`

| R object | field | element 1 | element 2 | value |
| -------- | ----- | --------- | --------- | ------|
| test_package |pkg_name | | | "dplyr" |
| |  pkg_ver | | | "1.1.4" |
| | date_time | | | "2026-04-09 10:17:27.71564" |
| | executor | | | "xxxxxxx" |
| | sysname | | | "Linux" |
| | version | | | "#1 SMP PREEMPT_DYNAMIC Mon Dec  1 05:36:50 UTC 2025" |
| | release | | | "6.1.158-180.294.amzn2023.x86_64" |
| | machine | | | "x86_64" |
| | r_version | | |"4.4.1" |
| | test_framework_type | | | standard testing framework |
| | total_cov | | | 0 |
| | res_cov | name| |"dplyr" |
| | | coverage | filecoverage | num [1:109(1d)] 98.3 76.7 97.5 100 100 ...
| | | | | attr(*, "dimnames")=List of 1 [1:109] "across.R" "all-equal.R" "arrange.R" "bind-cols.R" ... | 
| | | |       totalcoverage | 0.907 |
| | errors | | |NA |
| | notes  | | |NA |
| long_summary        |  file | | |      [1:3432] "test-across.R#L7_L7" "test-across.R#L14_L14" "test-across.R#L20_L20" "test-across.R#L24_L24" ... |
| | code_script | | | [1:3432] "across" "across" "across" "across" ... |
| | context     | | | [1:3432] "across() works on one column data.frame" "across() does not select grouping variables" "across() corre" |
| |  status     | | | [1:3432] "PASS" "PASS" "PASS" "PASS" ... |
| | n           | | | [1:3432] 1 1 1 1 1 1 1 1 1 1 ... |
| | time        | | | [1:3432] 0.041 0.014 0.077 0.077 0.077 ...|
|   test_skip         |  file | | |  [1:448] "test-across.R" "test-across.R" "test-across.R" "test-across.R" ... |
| | code_script  | | | [1:448] "across" "across" "across" "across" ... |
| | context   | | |  "across() gives meaningful messages" "across() gives meaningful messages" "across() gives meaningful mes" |
| | status     | | | [1:448] "FAILED" "FAILED" "FAILED" "FAILED" ... |
| | n          | | | [1:448] 1 1 1 1 1 1 1 1 1 1 ... |
| | time       | | | [1:448] 2.03 2.03 2.03 2.03 2.03 ... |
| | expectation| | |  [1:448] "expect_snapshot" "expect_error" "expect_error" "expect_error" ... |
| | line1      | | | [1:448] 209 212 216 222 226 230 234 238 243 244 ... |
| | line2      | | | [1:448] 280 215 219 225 229 233 237 241 243 244 ... |

Both `Step 2 unit test data coverage object` and `Step 3 unit test data coverage object` give the user
details about the status of the tests `FAILED`, `PASSED` and `SKIPPED`.

They also contain the test block start line (i.e. `line1` and `line2`) 
and the expectation type in the `expectation` column (e.g. `expect_snapshot` `expect_error`).

---
title: "NSTF Testing Frameworks"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{NSTF Testing Frameworks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r load, echo=FALSE}
library(test.assessr)
```
The `test.assessr` package checks package structure and test drivers to detect and classify testing configurations.

If it detects a non-standard testing framework (NSTF) (i.e. `RUnit‑based`, 
`BiocGenerics`, `data.table`, `tinytest`, and `custom test layouts`),
it runs non-standard testing workflows depending on the framework type:

The following example is the workflow for `BiocGenerics` which involves 
Bioconductor **RUnit** framework

 `BiocGeneric workflow`
 
   - Builds a source-to-test mapping.
   - Removes any `run_unitTests.R` wrapper from the mapping.
   - Prepares a test environment **before** running RUnit tests:
   - Detaches commonly attached BioC infrastructure packages (`IRanges`, `S4Vectors`, `BiocGenerics`)
     to avoid namespace/shim conflicts.
   - Calls `pkgload::load_all()` to load the package under test.
   - Clears `R_TESTS` to avoid R CMD check harness interference when running RUnit programmatically.
   - Identifies problem/skip-able tests (which should use `RUnit::runTestFile()` + `RUnit::getErrors()` per file).
   - Builds a list of test files **under `inst/unitTests`** excluding any that should be skipped.
   - Exposes namespace objects + datasets to a dedicated environment and runs coverage with
    `covr::environment_coverage()`.

  - If `BiocGeneric workflow` is successful, 
 then a unit test data coverage object like this is produced:


`BiocGeneric workflow unit test data coverage object`

| R object | field | element 1 | element 2 | value |
| -------- | ----- | --------- | --------- | ------|
| test_package |pkg_name | | | "BiocGenerics" |
| |  pkg_ver | | | "0.56.0" |
| | date_time | | | "2026-03-09 14:27:36.673915" |
| | executor | | | "xxxxxxx" |
| | sysname | | | "Linux" |
| | version | | | "#1 SMP PREEMPT_DYNAMIC Mon Dec  1 05:36:50 UTC 2025" |
| | release | | | "6.1.158-180.294.amzn2023.x86_64" |
| | machine | | | "x86_64" |
| | r_version | | |"4.4.1" |
| | test_framework_type | | | non-standard testing framework - Bioconductor |
| | total_cov | | | 15.5 |
| | res_cov | name| |"BiocGenerics" |
| | | coverage | filecoverage | num [1:57(1d)] 0 0 0 0 0 0 0 0 0 0 ...
| | | | | attr(*, "dimnames")=List of 1 [1:57] "/tmp/RtmpB1gcqI/R.INSTALL8ff8254343d/S4Vectors/R/DataFrame-class.R" "/tmp/RtmpB1gcqI/R.INSTALL8ff8254343d/S4Vectors/R/DataFrame-combine.R" "/tmp/RtmpB1gcqI/R.INSTALL8ff8254343d/S4Vectors/R/DataFrame-comparison.R" "/tmp/RtmpB1gcqI/R.INSTALL8ff8254343d/S4Vectors/R/DataFrame-utils.R" ... | 
| | | |       totalcoverage | 15.5 |
| | errors | | |NA |
| | notes  | | |NA |
| functions_no_tests         |  source_file | | |      [1:67] "annotation.R" "aperm.R" "append.R" "as.data.frame.R" ... |
| | test_file | | | [1:67] NA NA NA NA ... |
| | evidence     | | | [1:67] NA NA NA NA ... |
| | has_tests     | | | [1:67] FALSE FALSE FALSE FALSE FALSE FALSE ... |
| | comment           | | | [1:67] "No tests found" "No tests found" "No tests found" "No tests found" ... |
| tests_skipped | | | | chr(0) |
| tests_passing         | | | |  chr [1:7] "/tmp/RtmpneA8hi/temp_file_53d27b76020b/BiocGenerics/inst/unitTests/test_combine.R" "/tmp/RtmpneA8hi/temp_file_53d27b76020b/BiocGenerics/inst/unitTests/test_Extremes.R" "/tmp/RtmpneA8hi/temp_file_53d27b76020b/BiocGenerics/inst/unitTests/test_format.R" "/tmp/RtmpneA8hi/temp_file_53d27b76020b/BiocGenerics/inst/unitTests/test_mapply.R" ... |
| n_functions_tests_passing |  | | | int 7 |
| n_functions_no_tests|  | | |  int 67 |
| n_functions_tests_skipped |     | | | int 0 |


The `NSTF unit test data coverage objects` are not able to give details about 
test block start and end lines and the expectation types.

---
title: "Generate Test Report"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Generate Test Report}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

```{r load, echo=FALSE}
library(test.assessr)
```

## Introduction

The `generate_test_report()` function provides a unified way to generate HTML test assessment reports for R packages. It supports both standard testing frameworks (STF) and non-standard testing frameworks (NSTF) as well as single testing frameworks and multi‑framework test setups, producing a structured report that combines test summaries, detailed results, and coverage information.

## Step 1 - Run package coverage for a package

Run the following code

```{r pkg_covr, eval=FALSE}
dp <- system.file("test-data", "test.package.0001_0.1.0.tar.gz",
                 package = "test.assessr")


test_results <- get_package_coverage(dp)

generate_test_report(
  test_results,
  output_dir = getwd()
)

```
Important:
`output_dir` is required. It must be an existing, writable directory, such as:

`getwd()`
`"path/to/output_dir"`


If output_dir is missing or invalid, the function will error.

The generated HTML file will be written to output_dir. 
and named using the package name and version, for example:

`test_report_mypackage_1.2.3.html`

## What the Report Contains
Regardless of setup, every report includes:

- Package name and version
- Execution metadata (platform, R version, date/time)
- A package‑level summary of test status
- Test coverage results
- Detailed test results, per framework if applicable

The function automatically normalizes input so that single‑ and multi‑framework results share the same report structure.

## STF vs NSTF Reporting

The report distinguishes between STF and NSTF testing frameworks and presents framework‑appropriate detail sections when available.

## STF (Standard Testing Frameworks)

STF results come from `testthat` and/or `testit` frameworks. 

When STF data are present, the report may include:

- Long summaries of test outcomes
- Detailed test skip reasons
- expectation types
- test block start lines

These sections appear only when STF elements are detected, keeping reports concise when they are not relevant.

See the `dplyr` test report in the `inst/examples` folder.

## NSTF (Non‑Standard Testing Frameworks)

NSTF reporting supports lighter‑weight or base‑R‑style testing approaches, such as:

- RUnit
- data.table
- Manual or script‑based tests without rich metadata

For NSTF frameworks, the report can include:

- Functions with no associated tests
- Lists of skipped tests
- Lists of passing tests
- File‑level context with shortened paths when needed for readability

NSTF sections are generated only when relevant elements are present in test_results.

See the `MASS` test report in the `inst/examples` folder.

## Multi‑Framework Reports

When test_results indicates a multi‑framework setup:

- Each framework (e.g. testthat, covr, custom runners) gets its own subsection
- Coverage is reported per framework
- STF and NSTF sections are evaluated independently per framework

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{detect_frameworks}
\alias{detect_frameworks}
\title{Detect which test frameworks are present in a package}
\usage{
detect_frameworks(test_pkg_data)
}
\arguments{
\item{test_pkg_data}{List from \code{check_pkg_tests_and_snaps}.}
}
\value{
Character vector of framework IDs: "testthat", "testit", "tests_base",
  "tinytest", "data.table", "bioconductor".
}
\description{
Detect which test frameworks are present in a package
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core_test_functions.R
\name{run_coverage}
\alias{run_coverage}
\title{Run Coverage and Return Structured Coverage Results}
\usage{
run_coverage(pkg_source_path, timeout = Inf)
}
\arguments{
\item{pkg_source_path}{Character. Path to the installed package directory
from which coverage should be computed.}

\item{timeout}{Numeric. Timeout (in seconds) passed to
\code{callr::r_safe()} when running coverage. This limits the maximum
time allowed for executing the tests underlying the coverage analysis.}
}
\value{
A named list containing:
\describe{
  \item{total_cov}{Numeric. Aggregated coverage percentage for the package.}
  \item{res_cov}{A list containing file‑level or function‑level coverage
        results as returned by `covr`. This includes per‑file coverage,
        errors, and diagnostic notes if present.}
}

Returns \code{NULL} if coverage could not be computed.
}
\description{
This function executes code coverage analysis for a package using the
`covr` framework. It is typically used after a package has been installed
locally and test files are available. The function runs coverage in an
isolated process (using `callr`) and returns a structured summary of
overall and file‑level coverage.
}
\details{
The function invokes covr’s coverage evaluation in a clean R session and
extracts both:
\itemize{
  \item total coverage: percentage of lines covered across the package, and
  \item function/file‑level coverage: coverage data for individual files.
}

It is used internally by higher‑level functions such as
\code{run_covr_modes()} and skip‑aware coverage wrappers in the Standard
Testing Framework (STF).
}
\examples{
\donttest{
# Save and adjust CRAN mirror for reproducibility
r <- getOption("repos")
old <- options(repos = r)
r["CRAN"] <- "http://cran.us.r-project.org"
options(repos = r)

# Example package from test.assessr
dp <- system.file(
  "test-data",
  "test.package.0001_0.1.0.tar.gz",
  package = "test.assessr"
)

# Set up package source directory
install_list <- set_up_pkg(dp)
pkg_source_path <- install_list$pkg_source_path

# Install locally to enable testing and coverage
package_installed <- install_package_local(pkg_source_path)

if (isTRUE(package_installed)) {
  coverage_results <- run_coverage(pkg_source_path)
}

# Restore user's original repository settings
options(old)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{create_empty_covr_list}
\alias{create_empty_covr_list}
\title{Create an empty covr list for fallback}
\usage{
create_empty_covr_list(pkg_name, error_message)
}
\arguments{
\item{pkg_name}{A character string representing the name of the package.}

\item{error_message}{A character string describing the reason for the fallback.}
}
\value{
A named list containing default coverage values, the provided error message,
and placeholder notes.
}
\description{
This internal utility function returns a default coverage list structure
used when test execution or mapping fails. It ensures consistent output
format for downstream processing even when no valid test results are available.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{convert_vectors_to_df}
\alias{convert_vectors_to_df}
\title{This internal function takes named character vectors (e.g., `tests_skipped`, `tests_passing`)
from a list and converts them into data frames for easier reporting or processing.}
\usage{
convert_vectors_to_df(x, elems)
}
\arguments{
\item{x}{A list containing character vectors (e.g., `test_results`).}

\item{elems}{A character vector of element names to convert (must exist in `x`).}
}
\value{
A named list of data frames, where each data frame has one column:
\describe{
  \item{file}{The file names or paths from the original character vector.}
}
}
\description{
This internal function takes named character vectors (e.g., `tests_skipped`, `tests_passing`)
from a list and converts them into data frames for easier reporting or processing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_bioc_nstf.R
\name{get_source_test_mapping_bioc_nstf}
\alias{get_source_test_mapping_bioc_nstf}
\title{Map Bioconductor source files to test files, avoiding false positives.
Tightens filename and content-based matching; handles short names like "t".}
\usage{
get_source_test_mapping_bioc_nstf(pkg_source_path, test_dir)
}
\arguments{
\item{pkg_source_path}{Character; path to the package root (expects R/ under it).}

\item{test_dir}{Character; path to tests directory (e.g., file.path(pkg, "tests"))}
}
\value{
data.frame with columns: source_file, test_file, evidence, has_tests
}
\description{
Map Bioconductor source files to test files, avoiding false positives.
Tightens filename and content-based matching; handles short names like "t".
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_tests_stf.R
\name{get_expect}
\alias{get_expect}
\title{Extract Expectation Calls from Parsed Testthat Code}
\usage{
get_expect(parsed_data, token_text = "^expect_")
}
\arguments{
\item{parsed_data}{data.frame. Parse data produced by
\code{utils::getParseData(parse(file, keep.source = TRUE), includeText = TRUE)}.
Must include the columns \code{token}, \code{text}, \code{line1},
and \code{line2}.}

\item{token_text}{Character. A regular expression used to detect
expectation‑style function calls. Defaults to \code{"^expect_"}.}
}
\value{
A \code{data.frame} where each row corresponds to a detected expectation.
The data frame contains:
\describe{
  \item{expectation}{Character. Name of the expectation function (e.g.,
        \code{"expect_equal"}).}
  \item{line1}{Integer. Starting line number of the expression.}
  \item{line2}{Integer. Ending line number of the expression.}
}
Returns \code{NULL} if no matching expectations are found.
}
\description{
This internal helper identifies expectation calls (e.g.,
`expect_equal()`, `expect_true()`, etc.) within parse data extracted from a
testthat test file. The function locates expectation function calls by
matching their token text and returns a structured data frame containing
the expectation names and their corresponding source line ranges.
}
\details{
The function:
\itemize{
  \item locates all rows where the token is \code{"SYMBOL_FUNCTION_CALL"}
        and the text matches \code{token_text},
  \item identifies the expression node corresponding to each expectation,
  \item extracts line ranges (\code{line1}, \code{line2}) describing where the
        expectation appears in the source file, and
  \item returns a consolidated data frame of results.
}

If no expectation calls are found, the function returns \code{NULL}.
This helper is used internally by STF mapping functions (such as
\code{map_test()} and \code{map_tests_stf()}) when constructing a full
hierarchy of test blocks and expectations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{create_nstf_covr_list}
\alias{create_nstf_covr_list}
\title{Create the NSTF Coverage List}
\usage{
create_nstf_covr_list(
  coverage,
  pkg_name,
  no_tests_df,
  test_files_clean,
  skip_tests,
  problems
)
}
\arguments{
\item{coverage}{A coverage object returned from covr.}

\item{pkg_name}{A Character string - package name.}

\item{no_tests_df}{A dataframe - functions with no tests.}

\item{test_files_clean}{A Character vector - functions with working tests}

\item{skip_tests}{A Character vector - functions skipped}

\item{problems}{A Character vector - test problems}
}
\value{
A named list (`covr_list`) containing:
  \describe{
    \item{total_cov}{Numeric; overall package coverage percentage.}
    \item{res_cov}{List; raw coverage extraction results with coverage data, errors, notes.}
    \item{functions_no_tests}{Data frame of functions lacking tests.}
    \item{tests_skipped}{Character vector of skipped test files.}
    \item{tests_passing}{Character vector of passing test files.}
    \item{tests_problems}{Character vector describing test problems.}
  }
}
\description{
Internal helper that constructs the final `covr_list` structure from a
coverage object. It wraps extraction of coverage information, performs
total coverage computation, and assembles all diagnostic components.
}
\seealso{
Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_base_tests_coverage}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_nstf_test_path}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{restrict_test_paths_to_framework}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_stf.R
\name{create_fail_reporter}
\alias{create_fail_reporter}
\title{Create Fail Reporter}
\usage{
create_fail_reporter(test_path, pkg, cov_env)
}
\arguments{
\item{test_path}{Character string specifying the path to the test directory.}

\item{pkg}{Character string specifying the package name.}

\item{cov_env}{environment convr environment.}
}
\value{
A list of test results with corrected contexts, or `NULL` if an error occurs.
}
\description{
Executes all tests in a given directory using `testthat::test_dir()`
and applies context fixes to ensure each test has a valid context.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{run_covr_modes}
\alias{run_covr_modes}
\title{Run Coverage Analysis with Test Detection}
\usage{
run_covr_modes(pkg_source_path, covr_timeout = 60)
}
\arguments{
\item{pkg_source_path}{Character. Path to the root directory of the 
package source.}

\item{covr_timeout}{Numeric. Timeout in seconds for running coverage 
analysis. Default is 60.}
}
\value{
A named list containing coverage results, package metadata, and test
configuration details. The returned list includes (but is not limited to):
\describe{
  \item{pkg_name}{Character. Package name extracted from the DESCRIPTION file.}
  \item{pkg_ver}{Character. Package version.}
  \item{date_time}{Timestamp of when the analysis was run.}
  \item{executor}{User or environment running the analysis.}
  \item{sysname, version, release, machine}{System metadata.}
  \item{r_version}{R version used during analysis.}
  \item{test_framework_type}{Character. Detected testing framework type.}
  \item{covr_list}{A nested list containing: 
    \describe{
      \item{total_cov}{Numeric. Aggregated coverage percentage.}
      \item{res_cov}{File‑level and line‑level coverage details.}
      \item{errors, notes}{Any warnings or notes detected during testing.}
    }
  }
  \item{test_pkg_data}{A list describing the test configuration of the
    package (presence of testthat, testit, base tests, snapshots, etc.).}
}

This function always returns a value. When no supported testing framework is
detected, a default object with zero coverage and diagnostic information is
returned.
}
\description{
This function inspects the test configuration of an R package and runs 
code coverage analysis using any available testing framework, including 
`testthat`, `testit`, base R test scripts, or Bioconductor-style tests. 
If no recognised testing configuration is found, a default zero‑coverage
result is returned.
}
\examples{
\donttest{

dp <- system.file("test-data", 
  "test.package.0001_0.1.0.tar.gz", 
  package = "test.assessr")

# set up package
install_list <- set_up_pkg(dp)

package_installed <- install_list$package_installed
pkg_source_path <- install_list$pkg_source_path

# install package locally to ensure test works
package_installed <- install_package_local(pkg_source_path)
package_installed <- TRUE

covr_mode_list <- run_covr_modes(pkg_source_path)

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_runit_nstf.R
\name{prepare_for_bioc_runit}
\alias{prepare_for_bioc_runit}
\title{Prepare a BioC/RUnit package for testing
Detaches common BioC infrastructure packages (if attached) and load_all() the target
package before RUnit tests. Also clears R_TESTS to avoid interference from the check harness.}
\usage{
prepare_for_bioc_runit(pkg_name, pkg_source_path, detach_bioc_infra)
}
\arguments{
\item{pkg_name}{Character. Package name.}

\item{pkg_source_path}{Character. Path to package source.}

\item{detach_bioc_infra}{Logical. If TRUE (default for BioC packages), detach
IRanges, S4Vectors, BiocGenerics. Set to FALSE for RUnit-only packages.}
}
\description{
Prepare a BioC/RUnit package for testing
Detaches common BioC infrastructure packages (if attached) and load_all() the target
package before RUnit tests. Also clears R_TESTS to avoid interference from the check harness.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{setup_covr_env}
\alias{setup_covr_env}
\title{Set up a coverage environment for isolated test execution}
\usage{
setup_covr_env(
  pkg_source_path,
  pkg_name,
  test_path = "tests/testthat",
  cov_env
)
}
\arguments{
\item{pkg_source_path}{Character scalar. Path to the source package root.}

\item{pkg_name}{Character scalar. Package name.}

\item{test_path}{Character scalar. Path to the test directory.
Defaults to `"tests/testthat"`.}

\item{cov_env}{Environment. A pre-existing coverage environment, usually
created by an initializer such as `initialise_covr_env()`.}
}
\value{
A named list with components:
\describe{
  \item{cov_env}{The populated coverage environment.}
  \item{datasets_loaded}{Character vector of dataset objects loaded into
  `cov_env` from the target package, `datasets`, and optionally
  `nycflights13`.}
  \item{base_r_datasets_loaded}{Character vector of datasets loaded from
  the `datasets` package.}
  \item{base_r_exports_loaded}{Character vector of exported objects copied
  from Base R packages other than `base` and `datasets`.}
  \item{root_dir}{Resolved project root path.}
  \item{initial_state}{Character vector of object names present in
  `cov_env` before setup.}
  \item{helpers_loaded}{Character vector of helper files successfully
  sourced into `cov_env`.}
}
}
\description{
Prepare an existing coverage environment for use with coverage workflows
that evaluate tests in an isolated environment, typically one created
with `new.env(parent = baseenv())`.
}
\details{
The setup process:

- loads helper files from `tests/testthat`,
- loads package namespace objects into `cov_env`,
- resolves the project root,
- loads package datasets into `cov_env`,
- loads Base R datasets and exports into `cov_env`,
- optionally loads `nycflights13` datasets if available,
- loads dependencies into `cov_env`, and
- unloads the package namespace afterward to reduce conflicts.

This function does **not** create `cov_env`; it expects an existing
environment to be supplied.


This function is designed for workflows where `cov_env` is deliberately
isolated (for example with `parent = baseenv()`). In such cases, objects
from packages like `stats` and `datasets` are not automatically visible,
so they are explicitly loaded into `cov_env`.

The package `nycflights13` is treated as optional: if it is installed,
its datasets are loaded into `cov_env`; otherwise it is silently skipped.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_up_pkg.R
\name{set_up_pkg}
\alias{set_up_pkg}
\title{Create Information on Local Package Installation}
\usage{
set_up_pkg(dp)
}
\arguments{
\item{dp}{Character string. Path to a package tarball or package directory.}
}
\value{
A named list with the following elements:
\itemize{
  \item \code{package_installed} — Logical. \code{TRUE} if the unpacked
        package directory exists, otherwise \code{FALSE}.
  \item \code{pkg_source_path} — Character string giving the unpacked
        package source directory, or \code{""} if unpacking failed.
}

The returned object is always a base R \code{list}. It contains no side
effects besides calling \code{unpack_tarball()} and checking filesystem paths.
}
\description{
This function unpacks a package tarball using \code{unpack_tarball()} and
returns information about whether the unpacked directory exists. It is a
lightweight preparatory step used before attempting a local package
installation.
}
\section{Output Meaning}{

A value of \code{package_installed = TRUE} indicates that the unpacked
directory exists on disk and can be used for local installation.  

\code{package_installed = FALSE} indicates either:
\itemize{
  \item \code{unpack_tarball()} returned an empty result, or
  \item the unpacked directory does not exist on disk.
}
}

\examples{
\dontrun{
set_up_pkg(path/to/package, "mypackage")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_bioc_nstf.R,
%   R/run_covr_skip_runit_nstf.R
\name{remove_run_unitTests_mapping}
\alias{remove_run_unitTests_mapping}
\title{Remove mapping rows that point to run_unitTests.R}
\usage{
remove_run_unitTests_mapping(mapping, test_pkg_data)

remove_run_unitTests_mapping(mapping, test_pkg_data)
}
\arguments{
\item{mapping}{A data.frame with columns `source_file` and `test_file`
returned by get_source_test_mapping_nstf().}

\item{test_pkg_data}{A list or environment containing `bioc_run_ut_path`,
e.g., "/tmp/.../BiocGenerics/tests/run_unitTests.R".}
}
\value{
The filtered mapping data.frame with run_unitTests.R rows removed.

The filtered mapping data.frame with run_unitTests.R rows removed.
}
\description{
Remove mapping rows that point to run_unitTests.R

Remove mapping rows that point to run_unitTests.R
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{check_covr_skip_nstf}
\alias{check_covr_skip_nstf}
\title{Check for skipped or errored test files (non-testit)}
\usage{
check_covr_skip_nstf(
  pkg_source_path,
  mapping,
  base_dir = file.path(pkg_source_path, "tests")
)
}
\arguments{
\item{pkg_source_path}{Path to the root of the package source.}

\item{mapping}{A data frame with a column `test_file` listing test file names.}

\item{base_dir}{Directory containing the test files to probe. Defaults to
`<pkg>/tests`, but may also be `inst/tinytest` or `tests/testthat`
depending on package layout.}
}
\value{
A data frame listing test files and any issues encountered.
}
\description{
This internal function attempts to run test files mapped to source files and
identifies any that produce warnings or errors. It does not handle `testit`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{generate_coverage_section}
\alias{generate_coverage_section}
\title{Generate Coverage Section (single framework)}
\usage{
generate_coverage_section(assessment_results, pkg_name)
}
\arguments{
\item{assessment_results}{- input data with \code{res_cov}}

\item{pkg_name}{- name of the package}
}
\description{
Generates the Coverage section for the HTML report from a single
  framework's \code{res_cov}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_runit_nstf.R
\name{is_runit_layout}
\alias{is_runit_layout}
\title{Detect if a package uses RUnit layout (BioC or RUnit-only)}
\usage{
is_runit_layout(pkg_source_path, test_pkg_data = NULL)
}
\arguments{
\item{pkg_source_path}{Character. Path to package source root.}

\item{test_pkg_data}{List or NULL. Output from check_pkg_tests_and_snaps().}
}
\value{
Logical. TRUE if the package uses RUnit layout.
}
\description{
Returns TRUE when the package has RUnit tests in inst/unitTests (with or without
the BiocGenerics .test() runner). Heuristics:
 - test_pkg_data$has_BioG_test (BioC full layout), OR
 - test_pkg_data$has_RUnit_test (inst/unitTests, no .test required), OR
 - inst/unitTests exists, OR
 - tests/run_unitTests.R exists
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_runit_nstf.R
\name{is_bioc_runit_layout}
\alias{is_bioc_runit_layout}
\title{Detect if a package uses the Bioconductor RUnit layout}
\usage{
is_bioc_runit_layout(pkg_source_path, test_pkg_data = NULL)
}
\description{
Heuristics:
 - inst/unitTests exists, OR
 - tests/run_unitTests.R exists, OR
 - explicit flag in test_pkg_data$has_BioG_test
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{load_base_r_packages_into_env}
\alias{load_base_r_packages_into_env}
\title{Load Base R packages into an isolated coverage environment}
\usage{
load_base_r_packages_into_env(env, overwrite = FALSE)
}
\arguments{
\item{env}{An environment to populate.}

\item{overwrite}{Logical; if `TRUE`, overwrite objects already present in
`env`. If `FALSE` (default), existing objects are left unchanged.}
}
\value{
A named list with two character vectors:
\describe{
  \item{exports_loaded}{Exported objects copied from Base R packages
  other than `base` and `datasets`.}
  \item{datasets_loaded}{Dataset objects loaded from the `datasets`
  package.}
}
}
\description{
Populate an isolated environment (typically created with
`new.env(parent = baseenv())`) with objects needed to mimic the
standard Base R package surface during coverage or test evaluation.
}
\details{
The `base` package itself is already available via `baseenv()` and is
therefore not copied into `env`. The function instead:

- loads datasets from the `datasets` package into `env`, and
- copies exported objects from other Base R packages into `env`
  (e.g. `stats`, `utils`, `graphics`, `grDevices`, `methods`, etc.).

This is useful when tests are evaluated in a deliberately isolated
environment whose parent is `baseenv()`, because objects from packages
such as `stats` (e.g. `rnorm`) and `datasets` (e.g. `mtcars`, `iris`)
are not automatically visible there.


The Base R package set handled here is:
\itemize{
  \item `base`
  \item `compiler`
  \item `datasets`
  \item `graphics`
  \item `grDevices`
  \item `grid`
  \item `methods`
  \item `parallel`
  \item `splines`
  \item `stats`
  \item `stats4`
  \item `tcltk`
  \item `tools`
  \item `utils`
}

Because `base` is already reachable through `baseenv()`, it is excluded
from the explicit copy step.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{get_source_test_mapping_nstf}
\alias{get_source_test_mapping_nstf}
\title{Map R Source Files to Corresponding Test Files}
\usage{
get_source_test_mapping_nstf(pkg_source_path, test_dir)
}
\arguments{
\item{pkg_source_path}{Character string. Path to the root of the package source directory.}

\item{test_dir}{character string. directory where tests are}
}
\value{
A data frame with two columns:
\describe{
  \item{source_file}{Name of the source file in the `R/` directory.}
  \item{test_file}{Name of the matched test file in the `tests/` directory, or \code{NA} if no match is found.}
}
}
\description{
This internal helper function scans the `R/` and `tests/` directories of a package
and attempts to match each source file with a corresponding test file based on filename similarity.
}
\seealso{
Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_base_tests_coverage}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_nstf_test_path}()},
\code{\link{restrict_test_paths_to_framework}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{normalize_test_results}
\alias{normalize_test_results}
\title{Normalize test results for single or multi-framework structure}
\usage{
normalize_test_results(test_results)
}
\arguments{
\item{test_results}{List from \code{get_package_coverage} or \code{run_covr_modes}.}
}
\value{
A list with \code{multi_framework} (logical), \code{frameworks}
  (character, multi only), and \code{results} preserved for multi-framework.
}
\description{
For single-framework results, passes through unchanged. For
  multi-framework, sets \code{multi_framework} and \code{frameworks} while
  keeping \code{results} as a named list of per-framework blocks (no flattening).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{create_base_tests_coverage}
\alias{create_base_tests_coverage}
\title{Create coverage by running base-style tests under covr (internal)}
\usage{
create_base_tests_coverage(pkg_source_path, pkg_name, test_files)
}
\arguments{
\item{pkg_source_path}{Character; package source root.}

\item{pkg_name}{Character; package name.}

\item{test_files}{Character vector of test file paths (absolute or relative).}
}
\value{
list(coverage = <covr coverage object>)
}
\description{
Traces the package namespace with covr, then sources each test script
from its own directory (so relative file references like 'load("solr.rda")'
in MASS/tests work). Converts covr counters to a coverage object.
}
\seealso{
Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_nstf_test_path}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{restrict_test_paths_to_framework}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{load_test_helpers}
\alias{load_test_helpers}
\title{Load Testthat Helper Files into Environment}
\usage{
load_test_helpers(env, helper_dir = "tests/testthat")
}
\arguments{
\item{env}{An environment into which the helper files should be sourced.
Typically this is the same environment used for coverage, e.g., `cov_env`.}

\item{helper_dir}{Path to the directory containing helper files. Defaults to `"tests/testthat"`.}
}
\value{
Invisibly returns `TRUE` if all files are sourced successfully.
}
\description{
This function searches for all files in the `tests/testthat` directory
that match the pattern `helper-*.R` and sources them into the provided environment.
This is useful when running coverage with `covr::environment_coverage()` to ensure
helper functions are available during test execution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_stf.R
\name{create_coverage_skip_stf}
\alias{create_coverage_skip_stf}
\title{Run Coverage While Skipping Failing Tests}
\usage{
create_coverage_skip_stf(
  pkg_source_path,
  pkg,
  test_path = "tests/testthat",
  test_skip,
  test_map,
  cov_env
)
}
\arguments{
\item{pkg_source_path}{Character. Path to the root of the package source.}

\item{pkg}{Character. Name of the package being tested.}

\item{test_path}{Character. Path to the test directory.
Defaults to `'tests/testthat'`.}

\item{test_skip}{Data frame. Mapping of failing tests, typically containing
columns `file`, `line1`, and `line2`. Each row describes a span of lines
that should be commented out before running coverage.}

\item{test_map}{Data frame. Mapping of all tests within the package (used
for diagnostic messaging and consistency checks).}

\item{cov_env}{Environment that holds the temporary coverage workspace and
shared state for the run.}
}
\value{
A coverage object (a list) returned by `covr::environment_coverage()`,
filtered to remove problematic entries. If coverage fails, the function
returns `NULL`.

The returned structure typically contains:
\describe{
  \item{filename}{Name of the file being measured.}
  \item{functions}{Function-level coverage information.}
  \item{linecoverage}{Line-by-line execution counts.}
  \item{srcref}{Source references extracted for coverage.}
}
}
\description{
This function runs coverage analysis using `covr` while temporarily
commenting out known failing tests. It is intended for situations where
specific tests must be skipped (e.g., due to upstream bugs or known
failures) so that coverage can still be computed on remaining tests.
}
\details{
The function:
\itemize{
  \item sets up a temporary coverage environment,
  \item identifies lines to skip based on `test_skip`,
  \item comments out those test lines,
  \item runs `covr::environment_coverage()`, and
  \item restores the original test files on exit.
}


If `test_skip` is `NULL`, failing tests must be detected elsewhere before
calling this function. This function does not perform test detection; it
only applies the skip mappings supplied in `test_skip`.
}
\seealso{
\code{\link[testthat]{test_dir}},
  \code{\link[covr]{package_coverage}},
  \code{\link[covr]{environment_coverage}}

Other stf_utility: 
\code{\link{map_test}()},
\code{\link{map_tests_stf}()},
\code{\link{nest_test}()},
\code{\link{run_covr_skip_stf}()}
}
\concept{stf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{generate_test_details}
\alias{generate_test_details}
\title{Generate Test Details}
\usage{
generate_test_details(assessment_results, framework_id = NULL)
}
\arguments{
\item{assessment_results}{- input data}

\item{framework_id}{Optional character; when provided (multi-framework), the
"Testing Framework" row shows this framework name.}
}
\description{
Generates the Test Details table for the HTML report.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_tests_stf.R
\name{nest_test}
\alias{nest_test}
\title{Group Parsed Testthat Code into Logical Test Blocks (Internal)}
\usage{
nest_test(x, token_text = "^context$")
}
\arguments{
\item{x}{data.frame. Parse data produced by
\code{utils::getParseData(parse(file, keep.source = TRUE), includeText = TRUE)}.
Must include columns such as \code{token}, \code{text}, \code{parent},
and row names that correspond to node ids.}

\item{token_text}{Character. A regular expression used to select the
block‑starting function names (e.g., \code{"^context$"}, 
\code{"^test_that$|^describe$|^tar_test$"}, or \code{"^it$"}).
Defaults to \code{"^context$"}.}
}
\value{
A named list of data frames, where each element represents the parse rows
for a single detected test block. Names encode the block function and label.
If no matching blocks are found, an empty list is returned. This structure
is consumed by higher‑level STF mapping helpers (e.g., \code{map_test()},
\code{map_tests_stf()}) to derive expectations, contexts, and line ranges.
}
\description{
This internal helper groups parse data from a testthat test file into
logical test blocks based on calls to block‑defining functions such as
`context()`, `test_that()`, `describe()`, `it()`, or `tar_test()`. It
identifies top‑level expressions that match a given function‑name regex
and returns a named list of data frames—one per detected block—trimmed
to the relevant code region.
}
\details{
The function locates top‑level expressions whose function call tokens
(\code{SYMBOL_FUNCTION_CALL}) match \code{token_text}. For each matching
block, it:
\itemize{
  \item assigns a human‑readable name using the function name and its
        string constant argument (when available),
  \item trims the block to exclude leading parse noise, and
  \item returns the subset of \code{x} corresponding to that block.
}
Block names are prefixed with the function (e.g., \code{"test_that: ..."},
\code{"describe: ..."}, \code{"it: ..."}, \code{"tar_test: ..."}) when the
function is recognized.
}
\seealso{
\code{\link[utils]{getParseData}},
  \code{\link{map_test}},
  \code{\link{map_tests_stf}}

Other stf_utility: 
\code{\link{create_coverage_skip_stf}()},
\code{\link{map_test}()},
\code{\link{map_tests_stf}()},
\code{\link{run_covr_skip_stf}()}
}
\concept{stf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{load_package_into_env}
\alias{load_package_into_env}
\title{Load package objects into a coverage environment}
\usage{
load_package_into_env(pkg_source_path, pkg, cov_env)
}
\arguments{
\item{pkg_source_path}{Path to the package source directory.}

\item{pkg}{Name of the package to load.}

\item{cov_env}{Environment into which package objects will be loaded.}
}
\value{
Invisibly returns TRUE if successful, otherwise stops with an error.
}
\description{
This internal function loads all objects from a package namespace into a specified environment.
It is used by `create_covr_list_no_skip()` to prepare the environment for coverage analysis.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{safe_unload_package}
\alias{safe_unload_package}
\title{Safely unload a package to avoid coverage conflicts}
\usage{
safe_unload_package(pkg)
}
\arguments{
\item{pkg}{Character string. The name of the package to unload.}
}
\value{
NULL (invisibly). Messages are printed to indicate success or failure.
}
\description{
This internal function attempts to unload a specified package if it is currently loaded.
It is used to prevent coverage conflicts when reloading the package in a custom environment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_package_coverage.R
\name{get_package_coverage}
\alias{get_package_coverage}
\title{get package test coverage}
\usage{
get_package_coverage(path = NULL)
}
\arguments{
\item{path}{(optional) path of locally stored package source code}
}
\value{
An object of class \code{"coverage"} as produced by 
\code{test.assessr::run_covr_modes()}.  
This object is a structured list containing detailed test coverage
information, including:
\itemize{
  \item \strong{File-level coverage:} percentage of lines covered in each file.
  \item \strong{Line-by-line coverage:} number of executions for each line.
  \item \strong{Overall coverage:} aggregated coverage across the package.
  \item \strong{Metadata:} source file paths, expressions, and summary stats.
}

The resulting object can be printed, summarized, or passed to 
\code{test.assessr::generate_test_report()} to produce a human-readable test
coverage report.

Returns \code{NULL} if the package cannot be installed or if the specified
path does not exist.
}
\description{
simplified input to assess package for test coverage
}
\examples{
\donttest{
# Example STF package included in test.assessr
pkg_source_path <- system.file(
  "test-data",
  "test.package.0001_0.1.0.tar.gz",
  package = "test.assessr"
)

# Run get_package_coverage
pkg_test_coverage <- get_package_coverage(pkg_source_path)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_pkg_name}
\alias{get_pkg_name}
\title{get package name for display}
\usage{
get_pkg_name(input_string)
}
\arguments{
\item{input_string}{- string containing package name}
}
\value{
pkg_disp - package name for display
}
\description{
get package name for display
}
\examples{
\donttest{
pkg_source_path <- "/home/user/R/test.package.0001_0.1.0.tar.gz"
pkg_disp_1 <- get_pkg_name(pkg_source_path)
print(pkg_disp_1)

pkg <- "TxDb.Dmelanogaster.UCSC.dm3.ensGene_3.2.2.tar.gz"
pkg_disp_2 <- get_pkg_name(pkg)
print(pkg_disp_2)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{initialise_covr_env}
\alias{initialise_covr_env}
\title{Initialise a temporary coverage environment (framework-aware)}
\usage{
initialise_covr_env(pkg_source_path, pkg, framework = "none", work_dir = NULL)
}
\arguments{
\item{pkg_source_path}{Character; package root.}

\item{pkg}{Character; package name.}

\item{framework}{Character; e.g. "testthat", "testit", "bioc", "none".}

\item{work_dir}{Optional explicit temp workspace; if NULL it is composed from inputs.}
}
\value{
environment cov_env with $work_dir and $pkg_root set.
}
\description{
Initialise a temporary coverage environment (framework-aware)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_dt_nstf.R
\name{run_covr_skip_dt_nstf}
\alias{run_covr_skip_dt_nstf}
\title{Run Coverage Analysis on data.table Package Source and Tests}
\usage{
run_covr_skip_dt_nstf(pkg_source_path, test_pkg_data, cov_env)
}
\arguments{
\item{pkg_source_path}{Character string. Path to the root of the package source directory.}

\item{test_pkg_data}{List. Output from `check_pkg_tests_and_snaps()` indicating test framework presence.}

\item{cov_env}{Environment that holds the temporary coverage workspace and
shared state for the run.}
}
\value{
returns a list with coverage object and source files with no tests.
}
\description{
This function maps source files to test files, checks for skipped/problematic tests,
runs the remaining test files manually, calculates coverage, and optionally generates a report.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{handle_null}
\alias{handle_null}
\title{Helper function to replace NULL with "N/A"}
\usage{
handle_null(x)
}
\arguments{
\item{x}{- input value}
}
\description{
Helper function to replace NULL with "N/A"
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_tests_stf.R
\name{map_tests_stf}
\alias{map_tests_stf}
\title{Map the Structure of Testthat Tests (Standard Testing Framework)}
\usage{
map_tests_stf(path = "tests/testthat")
}
\arguments{
\item{path}{Character. Path to the test directory. Defaults to
`"tests/testthat"`.}
}
\value{
A consolidated `data.frame` where each row corresponds to an expectation
discovered within the STF test directory. The data frame typically contains:
\describe{
  \item{file}{Name of the test file in which the expectation appears.}
  \item{test}{The combined description and test label assigned to the
        expectation.}
  \item{expectation}{The expectation type detected (e.g., `expect_equal`).}
  \item{linerange}{The associated line numbers for that expectation.}
}

This structured mapping is used by downstream STF processing functions to
support skip mapping, coverage enhancement, and reporting.
}
\description{
This function analyses the test directory used by the Standard Testing
Framework (STF), typically `tests/testthat/`, and extracts a structured
representation of the unit tests contained in the directory. It identifies
test files, expectations, contexts, test names, and associated line ranges.
}
\details{
The function scans all testthat test files matching the pattern
`test*.R`. For each file, it extracts structured metadata describing the
hierarchy of expectations, including:
\itemize{
  \item file name
  \item context or description
  \item test label
  \item expectation type
  \item line ranges associated with each expectation
}

Files that contain no parseable test structure are omitted. If no valid
test files are found, the function returns `NULL`.
}
\seealso{
\code{\link[base]{list.files}},
  \code{\link[base]{lapply}},
  \code{\link[base]{rbind}}

Other stf_utility: 
\code{\link{create_coverage_skip_stf}()},
\code{\link{map_test}()},
\code{\link{nest_test}()},
\code{\link{run_covr_skip_stf}()}
}
\concept{stf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{compute_multi_framework_total}
\alias{compute_multi_framework_total}
\title{Compute aggregate total coverage across multiple frameworks}
\usage{
compute_multi_framework_total(per_framework)
}
\arguments{
\item{per_framework}{Named list of covr results from each framework.}
}
\value{
Numeric. Mean of per-framework total_cov (0-100), or 0 if none.
}
\description{
Compute aggregate total coverage across multiple frameworks
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_stf.R
\name{fix_test_context}
\alias{fix_test_context}
\title{Fix Missing Test Contexts}
\usage{
fix_test_context(test_results)
}
\arguments{
\item{test_results}{A list of test result objects from `testthat`.}
}
\value{
A list of test results with updated contexts.
}
\description{
Ensures that each test result has a valid `context` field.
If missing or empty, the context is set to the file name (without extension).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_stf.R
\name{run_covr_skip_stf}
\alias{run_covr_skip_stf}
\title{Run Coverage with Skipped Failing Tests (Standard Testing Framework)}
\usage{
run_covr_skip_stf(pkg_source_path, test_pkg_data, cov_env)
}
\arguments{
\item{pkg_source_path}{Character. Path to the root directory of the
package source.}

\item{test_pkg_data}{List. Output from 
\code{check_pkg_tests_and_snaps()}, containing information about
the package's testing setup (presence of testthat, snapshots, etc.).}

\item{cov_env}{Environment that holds the temporary coverage workspace and
shared state for the run. This must be an environment created by
`initialise_covr_env()`, which sets `cov_env$work_dir` to a unique
subdirectory under `tempdir()`. `run_covr_skip_stf()` will not create or
modify `cov_env` beyond reading helpers into it; all file outputs produced
by downstream helpers are expected to be written under
`cov_env$work_dir` (never to the package directory or the user's home).
The caller is responsible for lifecycle management; register
`on.exit(cleanup_and_return_null(env = cov_env), add = TRUE)` in the
calling scope to ensure bindings and the temporary workspace are cleaned
up even on error.}
}
\value{
A named list containing coverage results and test diagnostics. The structure
includes:
\describe{
  \item{total_cov}{Numeric. Overall coverage percentage.}
  \item{res_cov}{A nested list with file-level coverage, errors, and notes.}
  \item{long_summary}{Long-format test results including PASS/FAIL/ERROR/SKIPPED.}
  \item{test_skip}{A data frame describing test lines skipped during coverage,
    if applicable.}
}

If no tests fail, the returned list describes full coverage based on
unskipped tests. If no valid STF test directory is found, an empty coverage
object is returned via \code{create_empty_covr_list()}.
}
\description{
This function runs testthat-based test execution and code coverage
while identifying and skipping failing tests. It is designed for use
with standard testing frameworks (STF), primarily `testthat`, and
supports automated skip mapping to enable coverage calculation even
when some tests fail.
}
\details{
The function:
\itemize{
  \item executes tests and collects detailed test results,
  \item identifies failing, erroring, or skipped tests,
  \item constructs skip maps to temporarily comment out failing test lines,
  \item computes coverage on the remaining tests, and
  \item returns a combined object containing coverage and test summaries.
}
}
\seealso{
Other stf_utility: 
\code{\link{create_coverage_skip_stf}()},
\code{\link{map_test}()},
\code{\link{map_tests_stf}()},
\code{\link{nest_test}()}
}
\concept{stf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{get_nstf_test_path}
\alias{get_nstf_test_path}
\title{Collect all applicable nonstandard test paths for a package (internal)}
\usage{
get_nstf_test_path(test_pkg_data, testdir)
}
\arguments{
\item{test_pkg_data}{\code{list}. A list of booleans/fields describing detected
test frameworks and layouts. All fields are optional and default to \code{FALSE}
or \code{NULL} if missing:
\itemize{
  \item \code{has_testthat}: logical; whether \code{tests/testthat/} is present.
  \item \code{has_tinytest} (or \code{has_tiny_test}): logical; whether
        tinytest is present. If true, the function will look for
        \code{inst/tinytest/} (or use \code{tinytest_dir} if provided).
  \item \code{tinytest_dir}: character; optional explicit path to the tinytest
        directory. May be absolute or relative to the package root; when
        missing, defaults to \code{inst/tinytest/}.
  \item \code{has_testit}: logical; whether \code{tests/testit/} or the
        nonstandard \code{tests/test-ci/ + tests/test-cran/} layout is present.
  \item \code{has_tests_base}: logical; whether \code{*.R} test files exist
        directly under \code{tests/}.
  \item \code{has_bioc_unit_tests}: logical; whether \code{inst/unitTests/}
        exists (Bioconductor layout).
}}

\item{testdir}{\code{character}. Path to the package \code{tests/} directory. The function
also derives the package root as \code{dirname(testdir)} to locate paths such as
\code{inst/unitTests/} and \code{inst/tinytest/}.}
}
\value{
\code{list}. A named list of character paths to all applicable test directories found.
  Possible names include:
  \itemize{
    \item \code{tinytest} — \code{inst/tinytest/}
    \item \code{testthat} — \code{tests/testthat/}
    \item \code{testit} — \code{tests/testit/}
    \item \code{nonstandard_testit_root} — \code{tests/} (when both \code{test-ci} and \code{test-cran} exist)
    \item \code{base_tests_root} — \code{tests/} (when base \code{*.R} files exist in \code{tests/} root)
    \item \code{bioc_unitTests} — \code{inst/unitTests/}
    \item \code{default_tests_root} — \code{tests/} (returned only if no other paths are detected)
  }
}
\description{
Detects and returns all relevant testing directories used by an R package,
combining framework-specific and filesystem-based checks. Supported layouts:
\itemize{
  \item \strong{testthat}: \code{tests/testthat/}
  \item \strong{tinytest}: \code{inst/tinytest/} (primary location).
        Some packages also include a small runner under \code{tests/}
        (e.g., \code{tests/tinytest.R}) which this function does not return
        as a path, but it will still detect \code{inst/tinytest/}.
  \item \strong{testit} (standard): \code{tests/testit/}
  \item \strong{testit} (nonstandard): both \code{tests/test-ci/} and
        \code{tests/test-cran/} present, using \code{tests/} as the root
  \item \strong{base R tests}: \code{*.R} files directly under \code{tests/}
  \item \strong{Bioconductor unit tests}: \code{inst/unitTests/}
}
}
\details{
The function returns a named list of paths for all detected testing setups.
It uses both the flags contained in \code{test_pkg_data} and direct filesystem
checks for robustness. If no paths are detected, the returned list contains
a single entry \code{default_tests_root = tests/}.


Identical paths arising from multiple detection routes (e.g., \code{tests/} playing
multiple roles) are de-duplicated while preserving the first descriptive label.
If \code{tinytest_dir} is provided and relative, it is resolved against the
package root (i.e., \code{dirname(testdir)}).
}
\seealso{
Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_base_tests_coverage}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{restrict_test_paths_to_framework}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{create_long_summary_details}
\alias{create_long_summary_details}
\title{Generate Long Summary Details}
\usage{
create_long_summary_details(df)
}
\arguments{
\item{df}{- input data}
}
\description{
Generates the  Long Summary table for the HTML report.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aaa-cov-env.R
\docType{data}
\name{cov_env}
\alias{cov_env}
\title{Package-Private Coverage Environment}
\format{
An object of class \code{environment} of length 0.
}
\usage{
cov_env
}
\description{
`cov_env` is a package-internal environment used by test/coverage helpers
(e.g., `cleanup_and_return_null()`) to store temporary state without ever
touching the user's workspace (`.GlobalEnv`). It is created at package load
time so that a binding for `cov_env` exists *inside the locked package
namespace*, which allows tests to temporarily replace its value (mock) while
still complying with CRAN’s “no .GlobalEnv modifications” rule.
}
\details{
**When is it built?**  
During package setup, all `R/` files are parsed and evaluated as part of
namespace loading. At that point this line is executed:

```
cov_env <- new.env(parent = baseenv())
```

which creates the binding `cov_env` in the package namespace **before** the
namespace is locked by R. Once a namespace is locked, you cannot add or
remove bindings—you may only change the *values* of existing bindings. By
pre-creating `cov_env`, tests can safely and temporarily replace it (e.g.,
with `testthat::local_mocked_bindings()`), because they modify an *existing*
binding rather than creating a new one.  (See: R’s environment/namespace
locking rules and testthat’s guidance that mocked bindings must already
exist.)  [1](https://colinfay.me/writing-r-extensions/writing-r-documentation-files.html)[2](https://contributor.r-project.org/cran-cookbook/code_issues.html)

**Why `parent = baseenv()`?**  
Using `baseenv()` isolates lookups so they do not fall through to the user’s
workspace. This keeps internal state separate and avoids accidental reads or
writes to `.GlobalEnv`.  [3](https://github.com/coolbutuseless/CRAN-checks)

**Testing tip:**  
In unit tests, you can temporarily replace this binding without editing the
global environment:

```
testthat::local_mocked_bindings(
  .package = "yourpkg",
  cov_env  = new.env(parent = baseenv())
)
```

This works because the `cov_env` *binding already exists* in the namespace;
`local_mocked_bindings()` cannot create new bindings in a locked namespace.  [2](https://contributor.r-project.org/cran-cookbook/code_issues.html)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{check_elements_simple}
\alias{check_elements_simple}
\title{Check Which Elements Exist in a List}
\usage{
check_elements_simple(x, elems, nested = FALSE)
}
\arguments{
\item{x}{A list object to check (e.g., `test_results`).}

\item{elems}{A character vector of element names to check.}

\item{nested}{If TRUE and \code{x} has \code{results}, check inside each
framework block. Returns \code{present} (flattened unique) and
\code{by_framework} (named list: framework -> present elements).}
}
\value{
A list with:
\describe{
  \item{present}{Character vector of elements that exist.}
  \item{missing}{Character vector of elements that do not exist (top-level only).}
  \item{by_framework}{Named list (framework -> present elements), when nested.}
}
}
\description{
This internal function checks whether specified elements exist in a given list
and returns two vectors: elements that exist and elements that do not. When
\code{nested = TRUE} and \code{x} has \code{results} (multi-framework), checks
inside each \code{x$results$<framework>}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_dt_nstf.R
\name{compute_test_files_clean_dt}
\alias{compute_test_files_clean_dt}
\title{Compute wrapper data.table test files to run for coverage (internal)}
\usage{
compute_test_files_clean_dt(
  pkg_name,
  pkg_source_path,
  clean_mapping,
  test_pkg_data,
  skip_tests = NULL
)
}
\arguments{
\item{pkg_name}{`character(1)`. Package name.}

\item{pkg_source_path}{`character(1)`. Absolute or relative path to the
package source root.}

\item{clean_mapping}{A data frame or list containing at least a
`test_file` column/element with **relative** wrapper paths to run.
Paths are resolved against `tests/` or `tests/testit/` depending on
`test_pkg_data$has_testit`.}

\item{test_pkg_data}{`list`. Must include logical flag `has_testit`
indicating whether the package uses a `tests/testit/` layout.}

\item{skip_tests}{`NULL` or `character()`. Relative wrapper paths to
exclude (matched within the chosen base directory).}
}
\value{
`character()` - **normalized**, de-duplicated vector of wrapper
  file paths that **exist** on disk. May be zero length if no candidates
  exist and no fallback is available.
}
\description{
Selects and sanitizes the list of **wrapper** test files to execute during
coverage, supporting both the standard `tests/` layout and the `testit`
layout (`tests/testit/`). When `pkg_name` is `"data.table"`, the function
adds package-aware behavior: it falls back to `tests/main.R` if the mapping
yields no existing files and optionally prioritizes `froll.R` so rolling
function tests run early.
}
\details{
This helper **only resolves file paths**; it does **not** modify the
evaluation environment, copy/mirror test assets, or toggle data.table's
test harness mode. Any package-specific preparation (e.g., mirroring
`inst/tests/` to `tests/` or exposing `test.data.table()` in dev mode)
must be handled elsewhere (see `prepare_tests_for_package()`).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_bioc_nstf.R
\name{get_biocgenerics_test_paths}
\alias{get_biocgenerics_test_paths}
\title{Locate active test paths for Bioconductor RUnit layout}
\usage{
get_biocgenerics_test_paths(test_pkg_data, testdir)
}
\arguments{
\item{test_pkg_data}{list/env of test metadata; may include:
- `bioc_run_ut_path` (character): path to `tests/run_unitTests.R` wrapper, if present}

\item{testdir}{character: expected RUnit tests directory (typically "<pkg_root>/inst/unitTests")}
}
\value{
named list of **existing** directories (e.g., `bioc_unitTests`, `tests_root`);
        if nothing is found, returns a single `fallback` entry pointing to `testdir`.
}
\description{
Returns a *list* of one or more existing test directories suitable for
normalization and use in subsequent mapping (e.g., `get_source_test_mapping_nstf`).
Designed to work with:
}
\details{
path    <- as.character(pkg_source_path)
  testdir <- file.path(path, "inst/unitTests")
  test_path_list <- find_biocgenerics_test_paths(test_pkg_data, testdir)
  test_path <- normalizePath(
    unlist(test_path_list, use.names = TRUE),
    winslash = "/",
    mustWork = FALSE
  )
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_runit_nstf.R
\name{detach_pkg_if_attached}
\alias{detach_pkg_if_attached}
\title{Detach a package if attached (internal)}
\usage{
detach_pkg_if_attached(pkg, unload_namespace = TRUE, quiet = TRUE)
}
\description{
Safely detaches a package from the search path and optionally unloads its namespace.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_stf.R
\name{create_results_summary}
\alias{create_results_summary}
\title{Create a Summary of Coverage Results}
\usage{
create_results_summary(cvr)
}
\arguments{
\item{cvr}{A covr coverage object, typically produced by
`covr::environment_coverage()` or by higher‑level wrapper functions in
this package (e.g., skip‑aware coverage for STF or NSTF workflows).}
}
\value{
A data frame containing summary statistics, including:
\describe{
  \item{total_cov}{Numeric. Overall percentage of lines covered.}
  \item{name}{Character. File or package name associated with the coverage result.}
  \item{coverage}{List or data structure summarizing file‑level or
        element‑level coverage from the covr object.}
  \item{errors}{Character vector containing any coverage‑related errors (if present).}
  \item{notes}{Additional diagnostic notes returned by coverage processing.}
}

The summary object is used by downstream reporting functions to
generate test coverage summaries and scoring metrics.
}
\description{
This function extracts overall test coverage and file‑level coverage
information from a covr coverage object. It provides a simplified
summary for downstream reporting and scoring.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/load_dependencies_into_env.R
\name{load_dependencies_into_env}
\alias{load_dependencies_into_env}
\title{Load and Install Package Dependencies from a DESCRIPTION File}
\usage{
load_dependencies_into_env(pkg_source_path, env)
}
\arguments{
\item{pkg_source_path}{`character(1)`. Path to a package source directory
containing a `DESCRIPTION` file.}

\item{env}{An environment into which exported symbols from packages listed
in the **Depends** field will be imported. This allows code evaluated in
`env` to resolve functions from those packages using bare names without
requiring attachment on the search path.}
}
\value{
Invisibly returns `TRUE` on success.
}
\description{
Reads a package's `DESCRIPTION` file, parses dependency fields
(`Depends`, `Imports`, `Suggests`), installs missing packages or those that
do not satisfy version constraints, loads namespaces, and—for packages listed
under **Depends**—imports their exported symbols into a provided evaluation
environment. This makes bare function names visible when evaluating code in
environments that do not inherit from `.GlobalEnv` (e.g., environments with
`baseenv()` as parent).
}
\details{
This function avoids calling `installed.packages()` because that function
scans all installed packages and is slow on some platforms (especially
Windows and network filesystems). Using `requireNamespace()` is the
recommended approach for checking whether a package is installed and usable.
`packageDescription()` is used only on a small set of packages when version
information or priority metadata is needed.

When evaluating code in environments that do not inherit from `.GlobalEnv`,
attaching packages does not necessarily make their exported symbols available.
For this reason, exports from packages listed under `Depends` are explicitly
imported into the target environment.
}
\section{Behavior}{


1. **Dependency parsing**  
   Parses entries of the form `"pkg"` or `"pkg (op version)"`, where the
   operator `op` is one of `>=`, `>`, `<=`, `<`, `==`. HTML-escaped forms
   (`&gt;`, `&lt;`) are normalized.

2. **Dependency fields**  
   Dependencies are extracted from the `Depends`, `Imports`, and `Suggests`
   fields. The `Depends: R (>= x.y.z)` entry is ignored for installation
   purposes, but the R version constraint is validated.

3. **Base/recommended packages**  
   Unlike tools such as `R CMD check`, this function **does not skip**
   recommended or base packages (e.g., `graphics`, `stats`, `utils`,
   `grDevices`) because packages listed in `Depends:` must be made visible
   in the target evaluation environment even when its parent is `baseenv()`.
   If a package is already installed with R, no installation is attempted.

4. **Availability and version checks**  
   Package presence is determined via `requireNamespace(pkg, quietly = TRUE)`
   instead of `installed.packages()`, for performance and portability.  
   If a version constraint is declared, the installed version (from
   `packageDescription(pkg, fields = "Version")`) is compared using
   `utils::compareVersion()`.

5. **Package installation**  
   - For constraints of the form `>=` or `>`:  
     `install.packages()` is used (latest version is assumed valid).  
   - For `<=`, `<`, or `==`:  
     `remotes::install_version()` is used to install an exact version.  

6. **Loading behavior**  
   Each dependency is namespace-loaded via `loadNamespace()`.  
   Packages that appear in the `Depends` field have all their **exported
   symbols imported** into the supplied `env`. This makes calls such as
   `plot()`, `lm()`, or `head()` resolve correctly inside environments whose
   parent is not `.GlobalEnv`.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/core_test_functions.R
\name{run_covr}
\alias{run_covr}
\title{Run covr in a Subprocess with a Timeout}
\usage{
run_covr(path, timeout)
}
\arguments{
\item{path}{Character. Path to the package source directory for which
coverage should be computed.}

\item{timeout}{Numeric. Maximum time (in seconds) allowed for the coverage
computation. Defaults to \code{Inf}, meaning no timeout is applied unless
explicitly set.}
}
\value{
A coverage object returned by the underlying covr call, or \code{NULL} if
the subprocess execution fails or times out.
}
\description{
This internal helper runs coverage evaluation in a separate R subprocess
using `callr`. It allows coverage to be computed with a specified timeout,
ensuring that long‑running or hanging test processes do not block further
analysis.
}
\details{
The function launches a safe R session via \code{callr::r_safe()} and
executes the relevant covr routines within that isolated environment.
This is used internally by higher‑level STF/NSTF coverage functions to
ensure reproducible and controlled coverage execution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_tests_stf.R
\name{unrowname}
\alias{unrowname}
\title{Add Identifier Column and Remove Row Names from a Data Frame}
\usage{
unrowname(identifier, ret, label)
}
\arguments{
\item{identifier}{Character. The identifier to insert into the resulting
data frame (typically the name of the list element being processed).}

\item{ret}{List of data frames, usually the output of applying a mapping
or extraction function across multiple parsed test blocks.}

\item{label}{Character. Name of the column used to store the identifier
(e.g., \code{"test"} or \code{"file"}).}
}
\value{
A \code{data.frame} with an identifier column prepended and row names
removed. The remaining columns retain their original structure.
}
\description{
This internal helper restructures a data frame by adding an identifier
column (e.g., file name or test label) and removing row names. It is used
by STF parsing functions to attach contextual labels (such as test block
names or file names) to expectation or test‑block structures.
}
\details{
The function:
\itemize{
  \item extracts the data frame corresponding to \code{identifier},
  \item appends a new column containing the identifier value,
  \item removes row names for clean downstream processing, and
  \item reorders columns so the identifier appears first.
}

This structure is used by STF mapping helpers such as
\code{nest_expect()} and \code{map_tests_stf()} to carry forward file or
test labels when constructing unified test‑mapping tables.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{generate_test_report}
\alias{generate_test_report}
\title{Generate HTML Report for Package Test Assessment}
\usage{
generate_test_report(test_results, output_dir = NULL)
}
\arguments{
\item{test_results}{List containing the results from get_package_coverage function.}

\item{output_dir}{(required) Character string path to an existing directory where the report will be saved.}
}
\value{
Path to the generated HTML report.
}
\description{
Generates an HTML report for the package test assessment results using rmarkdown.
}
\examples{
\dontrun{
test_results <- get_package_coverage()
# Always provide a directory; tempdir() is convenient in examples.
generate_test_report(test_results, output_dir = tempdir())
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_runit_nstf.R
\name{check_covr_skip_runit_nstf}
\alias{check_covr_skip_runit_nstf}
\title{Check for skipped or errored test files (BiocGenerics)}
\usage{
check_covr_skip_runit_nstf(pkg_source_path, mapping, test_path)
}
\arguments{
\item{pkg_source_path}{Path to the root of the package source.}

\item{mapping}{A data frame with a column `test_file` listing test file names.}

\item{test_path}{directory}
}
\value{
A data frame listing test files and any issues encountered.
}
\description{
This internal function attempts to run test files mapped to source files and
identifies any that produce warnings or errors. It does not handle `testit`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{convert_number_to_abbreviation}
\alias{convert_number_to_abbreviation}
\title{Convert number to abbreviation}
\usage{
convert_number_to_abbreviation(value)
}
\arguments{
\item{value}{}
}
\description{
Convert number to abbreviation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_tests_skip_stf.R
\name{get_tests_skip_stf}
\alias{get_tests_skip_stf}
\title{Safely generate skip test mapping from problem tests and test map}
\usage{
get_tests_skip_stf(problem_tests, test_map)
}
\arguments{
\item{problem_tests}{A data frame of problematic tests, including columns `file` and `context`.}

\item{test_map}{A data frame representing the test structure, including columns `file` and `context`.}
}
\value{
A data frame containing the merged skip test mapping, or `NULL` if validation or processing fails.
}
\description{
This function filters and joins test metadata to identify which tests
should be skipped based on failed or problematic test results.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unittest_methods.R
\name{compute_total_coverage}
\alias{compute_total_coverage}
\title{Compute approximate total coverage combining line coverage and test breadth}
\usage{
compute_total_coverage(
  percent_cov,
  functions_no_tests_df,
  tests_passing,
  tests_skipped
)
}
\arguments{
\item{percent_cov}{Numeric scalar. Line coverage percentage (0–100).}

\item{functions_no_tests_df}{A `data.frame` (or `NULL`) of functions lacking tests.}

\item{tests_passing}{A character vector (or `NULL`) of passing test file paths/names.}

\item{tests_skipped}{A character vector (or `NULL`) of skipped test file paths/names.}
}
\value{
A named list with:
\itemize{
  \item \code{total_cov} (numeric): Approximate total coverage (proportion 0–1).
  \item \code{percent_cov_round} (numeric): \code{total_cov} rounded to 2 decimals (still 0–1).
  \item \code{n_testfiles} (integer): Count of passing test files.
  \item \code{n_no_function_tests} (integer): Count of functions lacking tests.
  \item \code{n_skipped_files} (integer): Count of skipped test files.
}
}
\description{
Computes an approximate "total coverage" metric by combining:
1) line coverage percentage (`percent_cov`, 0–100), and
2) a conservative proxy for breadth of tests based on counts of passing
   test files, functions without tests, and skipped test files.

The final metric scales line coverage (as a proportion) by the fraction of
tested over total function-space, where:
- `n_testfiles` = number of passing test files (if 0, treated as 1 to avoid divide-by-zero),
- `n_no_function_tests` = number of functions lacking tests (`nrow(functions_no_tests_df)`),
- `n_skipped_files` = number of skipped test files (`length(tests_skipped)`),
- `total_functions` = `n_tested + n_untested` with `n_tested = max(n_testfiles, 1)`,
- `total_cov` = (percent_cov / 100) * (n_tested / total_functions).
}
\details{
This function is intentionally conservative:
- Passing test *files* are used as a proxy for tested functions.
- Skipped test files contribute to "untested breadth".
- If there are zero passing test files, `n_tested` is set to 1 to avoid
  divide-by-zero and produce a small non-zero denominator.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_test_metadata}
\alias{get_test_metadata}
\title{Get test metadata}
\usage{
get_test_metadata(executor = NULL)
}
\arguments{
\item{executor}{- user who executes the riskmetrics process 

adapted from mrgvalprep::get_sys_info() and mpn.scorecard}
}
\value{
list with metadata
}
\description{
Get test metadata
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_dt_nstf.R
\name{dt_find_wrapper_tests}
\alias{dt_find_wrapper_tests}
\title{Locate data.table wrapper test scripts (internal)}
\usage{
dt_find_wrapper_tests(pkg_source_path)
}
\arguments{
\item{pkg_source_path}{character(1). Path to the package source root that
contains the \code{tests/} directory.}
}
\value{
character(). A vector of existing files among the known candidates,
  with absolute paths normalized via \code{normalizePath(..., winslash = "/")}.
  The vector is empty if none of the candidates exist.
}
\description{
Returns absolute, normalized paths to known wrapper test scripts under
\code{tests/} for a data.table source checkout. This is used to determine
which wrapper files exist locally so that callers can distinguish between
the canonical entrypoint \code{tests/main.R} and other wrapper scripts
(e.g., \code{froll.R}, \code{programming.R}, etc.).
}
\section{Notes}{

The candidate list is intentionally small and stable. If the upstream
repository adds additional wrappers, extend this function’s list.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{restrict_test_paths_to_framework}
\alias{restrict_test_paths_to_framework}
\title{Restrict a list of test paths to a single framework}
\usage{
restrict_test_paths_to_framework(test_path_list, framework)
}
\arguments{
\item{test_path_list}{\code{list}. Named list of candidate test paths. Known
names include \code{"testthat"}, \code{"testit"}, \code{"tinytest"},
\code{"base_tests_root"}, and optionally \code{"nonstandard_testit_root"}.}

\item{framework}{\code{character(1)} or \code{NULL}. Framework identifier
(e.g., \code{"testthat"}, \code{"testit"}, \code{"tests_base"},
\code{"tinytest"}). If \code{NULL} or empty (\code{""}), the list is
returned unchanged.}
}
\value{
\code{list}. Either:
\itemize{
  \item a single-element named list containing only the resolved framework
        path (name is the selected key), or
  \item the original \code{test_path_list} if \code{framework} is
        \code{NULL}/empty, maps to an unknown key, or the resolved key is
        absent in \code{test_path_list}.
}
}
\description{
Filters a named list of candidate test directories (typically produced by
\code{get_nstf_test_path()}) down to the entry that corresponds to a requested
framework. If no usable mapping can be made, the input list is returned
unchanged.
}
\details{
The function maps \code{framework} to a primary key in \code{test_path_list}
as follows:
\itemize{
  \item \code{"testthat"}   → \code{"testthat"}
  \item \code{"testit"}     → \code{"testit"} (falls back to
        \code{"nonstandard_testit_root"} if \code{"testit"} is \code{NULL})
  \item \code{"tests_base"} → \code{"base_tests_root"}
  \item \code{"tinytest"}   → \code{"tinytest"}
}


For \code{framework == "testit"}, if \code{test_path_list[["testit"]]} is
\code{NULL} and \code{test_path_list[["nonstandard_testit_root"]]} is present,
the latter is used instead. No directories are created or validated here; the
function only selects from the provided list.
}
\seealso{
\code{\link{get_nstf_test_path}}

Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_base_tests_coverage}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_nstf_test_path}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{run_covr_skip_nstf}
\alias{run_covr_skip_nstf}
\title{Run Coverage Analysis on Package Source and Non‑Standard Tests (NSTF)}
\usage{
run_covr_skip_nstf(pkg_source_path, test_pkg_data, cov_env, framework = NULL)
}
\arguments{
\item{pkg_source_path}{Character. Path to the root of the package source
directory.}

\item{test_pkg_data}{List. Output from \code{check_pkg_tests_and_snaps()}
indicating the detected testing infrastructure (e.g., base tests,
Bioconductor tests) and the presence of snapshots or related artifacts.}

\item{cov_env}{Environment that holds the temporary coverage workspace and
shared state for the run.}

\item{framework}{Optional character. When provided (e.g. from
\code{run_one_framework}), restricts test paths to the framework-specific
directory. Use when a package has multiple frameworks (e.g. testthat + testit)
to avoid mixing paths across frameworks.}
}
\value{
A named list containing coverage results and diagnostics:
\describe{
  \item{total_cov}{Numeric. Overall percentage of lines covered across the package.}
  \item{res_cov}{List. File‑ or element‑level coverage results (as derived from
        covr output), including per‑file coverage, any errors, and diagnostic notes.}
  \item{long_summary}{Long‑format summary of test outcomes (e.g., PASS/FAIL/ERROR/SKIPPED),
        when available from the NSTF execution.}
  \item{test_skip}{A data frame describing lines or regions skipped during coverage
        (if any). \code{NULL} if no skipping was required.}
  \item{untested_sources}{Character vector of source files for which no tests were found.}
}
If no valid NSTF test directory is detected, an “empty” coverage object may be
returned via \code{create_empty_covr_list()}.
}
\description{
This function maps package source files to non‑standard test scripts, detects
failing or skipped tests, executes the remaining tests manually, computes code
coverage, and returns structured results. It is intended for packages that do
not use the standard `testthat` framework (e.g., base R tests under `tests/`,
Bioconductor‑style tests in `inst/tests/`, or other legacy/custom scripts).
}
\details{
This NSTF workflow does not rely on `testthat` reporters. Instead, it:
\itemize{
  \item maps source files to available NSTF test scripts,
  \item identifies failing or problematic tests and derives a skip map,
  \item runs the remaining test scripts, and
  \item computes coverage on the successfully executed regions.
}
The output mirrors the structure returned by the STF skip‑aware workflow so
that downstream reporting remains consistent across frameworks.
}
\seealso{
Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_base_tests_coverage}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_nstf_test_path}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{restrict_test_paths_to_framework}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{find_project_root}
\alias{find_project_root}
\title{Find R project root directory}
\usage{
find_project_root(start = getwd())
}
\arguments{
\item{start}{character, starting path (e.g., test directory or script path)}
}
\value{
character path to project root, or NULL if not found
}
\description{
Searches upward from a starting path to locate the root of an R package or project.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{extract_short_path}
\alias{extract_short_path}
\title{Helper to extract "R/<file>" from any path by taking the last two components}
\usage{
extract_short_path(long_file_name)
}
\arguments{
\item{long_file_name}{A string containing the full file path (supports '/' or '\')}
}
\value{
A character string composed of the last two path components, e.g., "R/add.R"
}
\description{
Helper to extract "R/<file>" from any path by taking the last two components
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_dt_nstf.R
\name{prepare_tests_for_package_dt}
\alias{prepare_tests_for_package_dt}
\title{Prepare package-specific test environment for data.table (internal)}
\usage{
prepare_tests_for_package_dt(
  pkg_name,
  pkg_source_path,
  test_files,
  mode = c("mirror")
)
}
\description{
For data.table, Only "mirror" mode is supported here: 
copy inst/tests/* -> tests/ and clean up.
}
\details{
Returns a list with:
  - test_files: character() of wrapper files to execute
  - teardown:   function() to revert any changes (delete mirrored files, unset env)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{load_package_exports_into_env}
\alias{load_package_exports_into_env}
\title{Load exported objects from installed packages into an environment}
\usage{
load_package_exports_into_env(pkg_names, env, overwrite = FALSE)
}
\arguments{
\item{pkg_names}{Character vector of package names whose exported objects
should be copied into `env`.}

\item{env}{Environment into which exported objects should be assigned.}

\item{overwrite}{Logical; if `TRUE`, overwrite objects already present in
`env`. If `FALSE` (default), existing objects are left unchanged.}
}
\value{
A character vector of fully qualified object names that were loaded,
  in the form `"pkg::name"`.
}
\description{
Copy exported objects (typically functions, but also exported constants or
other objects) from one or more installed package namespaces into a target
environment.
}
\details{
This is useful when evaluating code in an isolated environment, such as one
created with `new.env(parent = baseenv())`, where package exports that are
normally available on the search path (for example from `stats` or `utils`)
are not automatically visible.

The function does not attach packages to the search path. Instead, it copies
exported objects directly into `env`.


For each package in `pkg_names`, this function:
\enumerate{
  \item checks whether the package namespace is available,
  \item retrieves the package's exported names,
  \item obtains each exported object with `getExportedValue()`,
  \item assigns the object into `env`.
}

Packages that are not installed or cannot be loaded are skipped with a
message.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{create_tinytest_coverage}
\alias{create_tinytest_coverage}
\title{Create coverage from tinytest under covr instrumentation (internal)}
\usage{
create_tinytest_coverage(
  pkg_source_path,
  pkg_name,
  tiny_dir = file.path(pkg_source_path, "inst", "tinytest"),
  at_home = TRUE,
  color = FALSE
)
}
\arguments{
\item{pkg_source_path}{Character scalar. Path to the package source root.}

\item{pkg_name}{Character scalar. Package name (as in \code{DESCRIPTION}).}

\item{tiny_dir}{Optional override for the tinytest directory. Defaults
to \code{file.path(pkg_source_path, "inst", "tinytest")}.}

\item{at_home}{Logical; passed to \code{tinytest::run_test_dir()}.
Defaults to \code{TRUE}.}

\item{color}{Logical; color output from tinytest. Defaults to \code{FALSE}.}
}
\value{
A list with components:
\describe{
  \item{coverage}{An object of class \code{"coverage"} (from \pkg{covr}).}
  \item{tinytests}{The \code{tinytests} object returned by tinytest.}
}
}
\description{
Instruments the package **namespace** with \pkg{covr}, runs all
\pkg{tinytest} files in \code{inst/tinytest}, and returns a covr coverage
object together with the tinytest results. This mirrors tinytest's execution
model (expectation helpers available; tests run from each file's directory)
while ensuring calls into the package namespace are traced.
}
\details{
Intended for **source-tree** coverage of tinytest suites:
\itemize{
  \item The package namespace is traced via \code{covr::trace_environment()}
        so that calls made by tests to package functions are recorded
        for coverage.
  \item \code{NOT_CRAN} is set to \code{"true"} during the run (restored on exit)
        so tinytest \code{at_home()} logic treats the run as local.
  \item Test files are driven by \code{tinytest::run_test_dir()} from
        \code{file.path(pkg_source_path, "inst", "tinytest")}.
  \item Coverage is built from covr's counters using \code{covr::as_coverage()}.
}

\strong{Preconditions}
\itemize{
  \item The package under test should already be \emph{loaded} (e.g.,
        \code{pkgload::load_all(pkg_source_path)}), so that
        \code{asNamespace(pkg_name)} exists and reflects the current source.
  \item The directory \code{inst/tinytest} must exist and contain test files.
}
}
\seealso{
Other nstf_utility: 
\code{\link{check_covr_skip_testit}()},
\code{\link{create_base_tests_coverage}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{get_nstf_test_path}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{restrict_test_paths_to_framework}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_bioc_nstf.R
\name{run_covr_skip_bioc_nstf}
\alias{run_covr_skip_bioc_nstf}
\title{Run coverage with skip detection for Bioconductor RUnit-based packages}
\usage{
run_covr_skip_bioc_nstf(
  pkg_source_path,
  test_pkg_data,
  cov_env,
  framework = NULL
)
}
\arguments{
\item{pkg_source_path}{Character scalar. Path to the root of the package source.}

\item{test_pkg_data}{A list or environment with test metadata (e.g., fields like
`bioc_run_ut_path` (path to `tests/run_unitTests.R`, if present),
`has_BioG_test` (logical flag), etc.).}

\item{cov_env}{Environment that holds the temporary coverage workspace and
shared state for the run.}

\item{framework}{Optional character. When provided (e.g. from
\code{run_one_framework}), restricts test paths to the framework-specific
directory. Use when a package has multiple frameworks (e.g. testthat + testit)
to avoid mixing paths across frameworks.}
}
\value{
A `covr` coverage object on success; `NULL` on failure (via `cleanup_and_return_null()`).

A covr coverage object, or NULL on failure.
}
\description{
`run_covr_skip_bioc_nstf()` orchestrates coverage for packages using the
Bioconductor **RUnit** layout, such as **BiocGenerics** and packages that store
tests in `inst/unitTests` or ship a `tests/run_unitTests.R` wrapper.
}
\details{
The function:
1. Builds a source-to-test mapping (using your existing `get_source_test_mapping_nstf()`).
2. Removes any `run_unitTests.R` wrapper from the mapping (via `remove_run_unitTests_mapping()`).
3. Prepares a development environment **before** running RUnit tests:
   - Detaches commonly attached BioC infrastructure packages (`IRanges`, `S4Vectors`, `BiocGenerics`)
     to avoid namespace/shim conflicts.
   - Calls `pkgload::load_all()` to load the package under test.
   - Clears `R_TESTS` to avoid R CMD check harness interference when running RUnit programmatically.
4. Identifies problem/skip-able tests via your RUnit-aware `check_covr_skip_nstf()` (which should use
   `RUnit::runTestFile()` + `RUnit::getErrors()` per file).
5. Builds a list of test files **under `inst/unitTests`** excluding any that should be skipped.
6. Exposes namespace objects + datasets to a dedicated environment and runs coverage with
   `covr::environment_coverage()`.


Use this function **only** for packages that follow the Bioconductor RUnit layout:
- tests are located in `inst/unitTests/`, and/or
- a `tests/run_unitTests.R` wrapper exists that calls an internal `.test()` / `testPackage()` runner.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unittest_methods.R
\name{check_pkg_tests_and_snaps}
\alias{check_pkg_tests_and_snaps}
\title{Check for tests/testthat and _snaps folder and count golden tests}
\usage{
check_pkg_tests_and_snaps(pkg_source_path)
}
\arguments{
\item{pkg_source_path}{Path to the root of the package source}
}
\value{
A list with:
  - `has_testthat`: Does tests/testthat exist?
  - `has_testit`: Does tests/testit exist?
  - `has_tests_base`: Do tests exist in tests subdirectory?
  - `has_BioG_test`: Do BioGenerics tests exist in inst/tests subdirectory?
  - `has_RUnit_test`: Do RUnit tests exist (inst/unitTests; optionally tests/run_unitTests.R)?
  - `bioc_unit_tests_dir` : location of BioGenerics unit tests
  - `bioc_run_ut_path`: location of BioGenerics `run_unitTests.R`
  - `has_snaps`: Does _snaps exist inside tests/testthat?
  - `n_golden_tests`: Number of snapshot files inside _snaps
  - `n_test_files`: Number of test-*.R files inside tests/testthat
  - `has_tinytest` : Does tinytest exist? 
  - `tinytest_dir` : location of tinytest  unit tests
  - `tinytest_runner_paths` : location of tinytest runner paths
  - `n_tinytest_files` :  Number of test_*.R files under inst/tinytest  
  - `has_data_table_tests`: Do data.table tests exist in directories?
  - `data_table_tests_path` : location of data.table unit tests
}
\description{
Check for tests/testthat and _snaps folder and count golden tests
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_tests_stf.R
\name{nest_expect}
\alias{nest_expect}
\title{Extract and Structure Expectations from Parsed Test Blocks}
\usage{
nest_expect(x)
}
\arguments{
\item{x}{A named list of parsed test blocks, typically produced by
\code{nest_test()}. Each element should contain parse data suitable for
expectation extraction via \code{get_expect()}.}
}
\value{
A \code{data.frame} in which each row corresponds to an expectation
extracted from the supplied test blocks. Typical columns include:
\describe{
  \item{expectation}{Name of the expectation function (e.g.,
        \code{"expect_equal"}).}
  \item{line1}{Starting line number of the expectation expression.}
  \item{line2}{Ending line number of the expectation expression.}
  \item{test}{Label identifying the test block the expectation belongs to.}
}
Returns \code{NULL} if no expectations are detected.
}
\description{
This internal helper takes a list of parsed test blocks (as produced by
\code{nest_test()}) and extracts expectation calls from each block using
\code{get_expect()}. It then assembles all detected expectations into a
structured data frame that associates each expectation with its
corresponding test label.
}
\details{
For each block in \code{x}, the function:
\itemize{
  \item calls \code{get_expect()} to extract expectation function calls
        and their line ranges,
  \item removes blocks with no expectations,
  \item attaches a \code{test} label using the block's list name, and
  \item binds all expectation results into a single consolidated data frame.
}

If no expectations are found in any block, the function returns \code{NULL}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_tests_stf.R
\name{map_test}
\alias{map_test}
\title{Map the Structure of Expectations in a Testthat File}
\usage{
map_test(path)
}
\arguments{
\item{path}{Character. Path to a single test file (typically named
`test-*.R`).}
}
\value{
A `data.frame` in which each row corresponds to an expectation identified in
the test file. Typical columns include:
\describe{
  \item{description}{Context or test block label derived from the test file.}
  \item{expectation}{The expectation type (e.g., `expect_equal`).}
  \item{linerange}{Numeric vector giving the range of lines associated with
        the expectation.}
  \item{test}{Test label (combined description and nested block label, if
        applicable).}
}

This structured output is used internally by STF test‑mapping functions
(e.g., `map_tests_stf()`) for coverage analysis, skip mapping, and test
reporting.
}
\description{
This function parses a single testthat test file and extracts a structured
representation of the expectations it contains. It identifies top‑level test
blocks (e.g., `test_that()`, `describe()`, `tar_test()`) and any nested
expectation blocks (e.g., `it()`). The result maps expectation location,
context, and source line ranges for downstream analysis.
}
\details{
The function uses `utils::getParseData()` to inspect parse information from
the test file and constructs a hierarchical mapping of:
\itemize{
  \item test block context (e.g., `describe()`, `test_that()`)
  \item test labels or sub‑contexts (e.g., `it()` blocks)
  \item expectation type
  \item source line ranges associated with each expectation
}

If the file does not contain any parseable test blocks or expectations,
`NULL` is returned.
}
\seealso{
\code{\link[utils]{getParseData}}

Other stf_utility: 
\code{\link{create_coverage_skip_stf}()},
\code{\link{map_tests_stf}()},
\code{\link{nest_test}()},
\code{\link{run_covr_skip_stf}()}
}
\concept{stf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_test_report.R
\name{generate_coverage_sections}
\alias{generate_coverage_sections}
\title{Generate Coverage Sections (multi-framework)}
\usage{
generate_coverage_sections(test_results, pkg_name)
}
\arguments{
\item{test_results}{Multi-framework list with \code{results} (named list of
per-framework blocks, each with \code{res_cov}).}

\item{pkg_name}{- name of the package}
}
\value{
Named list of data frames (Function, Coverage, Errors, Notes).
}
\description{
Returns a named list of coverage data frames, one per framework.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_nstf.R
\name{check_covr_skip_testit}
\alias{check_covr_skip_testit}
\title{Run testit tests individually, capturing all warnings and errors, with dynamic package root detection}
\usage{
check_covr_skip_testit(pkg_name, mapping, test_path)
}
\arguments{
\item{pkg_name}{The package name (not used in this function, but kept for compatibility)}

\item{mapping}{A data frame with a column `test_file` listing test file names}

\item{test_path}{The path to the test directory (e.g., tests/testit)}
}
\value{
A data frame with test_file and issue_type columns
}
\description{
Run testit tests individually, capturing all warnings and errors, with dynamic package root detection
}
\seealso{
Other nstf_utility: 
\code{\link{create_base_tests_coverage}()},
\code{\link{create_nstf_covr_list}()},
\code{\link{create_tinytest_coverage}()},
\code{\link{get_nstf_test_path}()},
\code{\link{get_source_test_mapping_nstf}()},
\code{\link{restrict_test_paths_to_framework}()},
\code{\link{run_covr_skip_nstf}()}
}
\concept{nstf_utility}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_tests_skip_stf.R
\name{safe_left_join}
\alias{safe_left_join}
\title{Safely perform a left join with error handling}
\usage{
safe_left_join(x, y, by, relationship = "many-to-many", ...)
}
\arguments{
\item{x}{A data frame or tibble. The left-hand side of the join.}

\item{y}{A data frame or tibble. The right-hand side of the join.}

\item{by}{A character vector specifying the join keys. Passed to
`dplyr::left_join()`.}

\item{relationship}{A string specifying the expected relationship between keys:
`"one-to-one"`, `"one-to-many"`, `"many-to-one"`, or `"many-to-many"`.
Defaults to `NULL` (no check).}
}
\value{
A data frame resulting from the left join if successful,
  otherwise `NULL` when an error occurs.
}
\description{
This helper function attempts to join two data frames using
`dplyr::left_join()`. If an error occurs during the join (for example,
due to missing columns or incompatible types), the function issues a
warning and returns `NULL` instead of stopping execution.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{remove_new_globals}
\alias{remove_new_globals}
\title{Remove New Globals from Environment Based on Initial Snapshot}
\usage{
remove_new_globals(env, initial_state, patterns = NULL, dry_run = FALSE)
}
\arguments{
\item{env}{Environment to clean.}

\item{initial_state}{Character vector of object names before setup.}

\item{patterns}{Optional regex patterns to filter objects.}

\item{dry_run}{Logical. If TRUE, only prints what would be removed.}
}
\value{
Invisibly returns a character vector of removed object names.
}
\description{
This function removes objects from an environment that were added after
a recorded initial state. Optionally filters by pattern.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{load_package_datasets}
\alias{load_package_datasets}
\title{Load Datasets from a Package}
\usage{
load_package_datasets(pkg_names, env, source_paths = NULL, overwrite = FALSE)
}
\arguments{
\item{pkg_names}{A character string specifying the name of the package.}

\item{env}{A character string specifying the covr env}

\item{source_paths}{A character string specifying the source paths}

\item{overwrite}{A logical specifying whether datasets should be overwritten}
}
\value{
A character vector of dataset names that were attempted to be loaded.
Returns `NULL` if the package is not installed.
}
\description{
Loads all datasets from a specified installed package into the cov environment.
It first attempts to load `.rda` files from the package's `data/` directory.
If none are found, it falls back to using `data()` to load datasets listed in the package metadata.
}
\details{
This function is intended for internal use. It silently attempts to load each dataset
and suppresses errors if loading fails. Datasets are loaded into the cov environment.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_skip_dt_nstf.R
\name{dt_compute_untested_functions}
\alias{dt_compute_untested_functions}
\title{Compute untested functions from coverage using srcref spans (internal)}
\usage{
dt_compute_untested_functions(pkg_name, coverage)
}
\arguments{
\item{pkg_name}{character(1). The package name whose namespace is inspected
for functions (\code{asNamespace(pkg_name)}).}

\item{coverage}{A coverage object as returned by \code{covr} (e.g.,
\code{covr::environment_coverage()} or \code{covr::package_coverage()}),
consumable by \code{covr::tally_coverage()}.}
}
\value{
\link[base:data.frame]{data.frame} with one row per untested function
  and columns:
  \describe{
    \item{function_name}{Function symbol in the package namespace.}
    \item{file}{Absolute, normalized path to the source file containing the function.}
    \item{start_line}{Start line (inclusive) from the function's \code{srcref}.}
    \item{end_line}{End line (inclusive) from the function's \code{srcref}.}
  }
  Returns an empty data.frame with these columns if (a) the coverage table is
  empty, (b) no functions have an \code{srcref}, or (c) all functions had at
  least one executed line within their span.
}
\description{
Identifies package functions that were not executed by the provided coverage
object. A function is considered \emph{untested} if \strong{no} line within
its \code{srcref} span (from \code{start_line} to \code{end_line}) has a
positive execution count in \code{covr::tally_coverage()}.
}
\section{Details}{

\itemize{
  \item Functions without an \code{srcref} are skipped.
  \item Files are matched by \emph{basename} between the coverage table and the
        function’s \code{srcref} source path to make the join resilient to
        differing absolute paths in temporary build trees.
  \item A line is considered executed when its coverage \code{value} is a finite
        number greater than zero.
}
}

\seealso{
\code{\link[covr]{tally_coverage}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{run_one_framework}
\alias{run_one_framework}
\title{Run coverage for a single test framework}
\usage{
run_one_framework(framework, pkg, test_pkg_data, pkg_source_path, covr_timeout)
}
\arguments{
\item{framework}{Character. Framework ID.}

\item{pkg}{Character. Package name.}

\item{test_pkg_data}{List from \code{check_pkg_tests_and_snaps}.}

\item{pkg_source_path}{Character. Path to package source.}

\item{covr_timeout}{Numeric. Timeout for coverage.}
}
\value{
List with total_cov, res_cov, test_framework_type, framework_id.
}
\description{
Run coverage for a single test framework
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/install_package_local.R
\name{install_package_local}
\alias{install_package_local}
\title{Install a Package from a Local Source Directory}
\usage{
install_package_local(pkg_source_path)
}
\arguments{
\item{pkg_source_path}{Character string. Path to the local package source
directory (e.g., an unpacked package or extracted tarball path).}
}
\value{
A logical value indicating whether the package is installed after running
the function.  

The returned object is always of class \code{logical}:
\itemize{
  \item \code{TRUE} — The package is already installed or was successfully installed.
  \item \code{FALSE} — Installation failed or the path does not exist.
}
}
\description{
Attempts to install an R package from a local source directory using
\code{remotes::install_local()}. The function reports on whether installation
succeeded, whether the package was already installed, or whether the provided
source path does not exist.
}
\details{
The display name of the package is derived from the input path using
\code{get_pkg_name()}.
}
\section{Output Meaning}{

\code{TRUE} does not necessarily imply that the installation occurred during
this function call—it may also mean the package was already installed.  

\code{FALSE} indicates a failure to install or an invalid path. All diagnostic
messages are printed via \code{message()} for user visibility.
}

\examples{
\dontrun{
results <- install_package_local("pkg_source_path")
print(results)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/run_covr_modes.R
\name{get_test_framework_type}
\alias{get_test_framework_type}
\title{Get human-readable test framework type label}
\usage{
get_test_framework_type(test_pkg_data, framework_id)
}
\arguments{
\item{test_pkg_data}{List from \code{check_pkg_tests_and_snaps}.}

\item{framework_id}{Character. Framework ID from \code{detect_frameworks}.}
}
\value{
Character. Label for test_framework_type.
}
\description{
Get human-readable test framework type label
}
\keyword{internal}
